#define INCL_DOSMISC
#define INCL_DOSFILEMGR
#define INCL_DOSPROCESS
#define INCL_DOSDATETIME
#define INCL_DOSEXCEPTIONS
#define INCL_DOSSEMAPHORES
#include <os2.h>

#include <os2.h>

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <process.h>

#include "except.h"

#include "language.h"

#include "heap.h"
#include "config.h"
#include "misc.h"
#include "timer.h"
#include "buffer.h"
#include "fidoaddr.h"
#include "queue.h"
#include "nodelist.h"
#include "process.h"
#include "sched.h"
#include "zinter.h"
#include "rexxhook.h"
#include "key.h"

static int argNum;
static char **argPtr;
char **envPtr;
char *prfName;
char *brakeLoadPath;

void terminalEmulator();
int firstInit( int, char** );
static char defaultCtlName[] = "Brake!.Ctl";
char *ctlName = defaultCtlName;
static volatile int RC = 0;
char needDumpOnly = 0;

static HEV mainTidSema = 0xffffffffl;
static unsigned long mainStartTime;


void complete()
{
  if( systemIsReady )
  {
    showAction( Msg_ScrTheEnd );

    if( RC>0 && RC<10 && RC!=4 )
      execSchedule( melodies.error );

    execSchedule( melodies.stop );
  }

  if( !systemStartTime )
    systemStartTime = getTimeMark();

  writeToBinaryLog( 7, 0, 0 );

  long elapsed( getTimeMark()- systemStartTime );
  long percent( makePercent( elapsed, systemUsageTime ) );
  char *t( strTime( getTimeMark()-systemStartTime, 0 ) );
  int usa( nodeListPtr==NULL ? 0 : nodeList.getCacheUsage() );
  unsigned long hint( nodeListPtr==NULL ? 0 : nodeList.getCacheHints() );

  if( logPtr!=NULL )
  {
    log( Msg_InfoComplStat, t, percent, usa, hint );
    log( Msg_InfoComplRC, RC );
    logPtr->empty();
  }

  dataDeInit();

  log( Msg_InfoComplStat, t, percent, usa, hint );
  log( Msg_InfoComplRC, RC );

  delete t;
}

void makeDone()
{
  APIRET rc = DosPostEventSem( mainTidSema );

  if( rc )
  {
    log( Msg_ErrUnablePostMainSema, rc );
    os2Message( rc );
  }
}

void _Optlink mainThread( void * )
{
  registerExceptionHandler

  time_t initStep1( getMSec() ),
         initStep2( initStep1 ),
         initStep3( initStep1 ),
         initStep4( initStep1 );

  if( firstInit( argNum, argPtr ) )
    goto CtlError1;

  heapState( "Start" );
  initStep1 = getMSec();

  cfgStack = new CfgStack;

  if( processConfig( ctlName ) )
  {
CtlError:
    log( Msg_ErrCtlError );
CtlError1:
    RC = 2;
    goto okComplete;
  }

  initStep2 = getMSec();

  if( compileSchedule() )
    goto CtlError;

  log( Msg_InfoReadingDone );

  initStep3 = getMSec();

  if( dataInit() )
    goto CtlError1;

  writeToBinaryLog( 7, 0, 0 );
  RC = terminalActive;
  terminalActive = 0;

  if( comPort==NULL || !comPort->ok() )
  {
    log( Msg_ErrNoPort );
    goto abortProgramExecution;
  }

  terminalActive = char(RC);
  RC = 0;

  initStep4 = getMSec();

  dataInitMsg( initStep1-mainStartTime, Msg_InitStat1 );
  dataInitMsg( initStep2-initStep1,     Msg_InitStat2 );
  dataInitMsg( initStep3-initStep2,     Msg_InitStat3 );
  dataInitMsg( initStep4-initStep3,     Msg_InitStat4 );
  dataInitMsg( initStep4-mainStartTime, Msg_InitStat5 );

  if( terminalActive )
  {
    terminalEmulator();
    comPort->init();
  }
  else
  {
    delete cfgStack;

    extern char *todayStrings[];

    log( todayStrings[ todayMessage( 5, 2, 2 ) ] );
    execSchedule( melodies.start );

    if( needDumpOnly )
    {
      queue->rescanQueue();
      debugDump();
      goto okComplete;
    }

    if( getRexxHook( "Mailer" )!=NULL )
    {
      char *rez=processRexxHook( "Mailer", NULL, NULL, NULL,
                                           NULL, NULL, NULL, 1 );

      if( rez==NULL )
        RC=0;
      else
        sscanf( rez, "%d", &RC );
    }
    else
    {
//      int testEmsi();
//      RC = testEmsi();

      int mailer();
      RC = mailer();
    }

    if( RC==1 )
      goto abortProgramExecution;
  }

  goto okComplete;

abortProgramExecution:
  RC = 1;

okComplete:
  complete();
  makeDone();

  deRegisterExceptionHandler
}

long mainTid;

void abortAndExit()
{
  if( logPtr )
    log( Msg_InfoAbortProcess );

  terminalActive = 0;
  OSReturnCode = 4;

  if( comPort )
    comPort->hangUp();
}

ULONG isNowTotalFh;

int __cdecl BrakeMain( int argNumIni, char **argPtrIni, char **envPtrIni )
{
  int firstFirstInit();
  void signalTERM( int );

  registerExceptionHandler
  mainStartTime = getMSec();

  LONG addFh( 20 );

  DosSetRelMaxFH( &addFh, &isNowTotalFh );

  ULONG times;
  DosAcknowledgeSignalException( XCPT_SIGNAL_KILLPROC );
  DosAcknowledgeSignalException( XCPT_SIGNAL_BREAK );
  DosAcknowledgeSignalException( XCPT_SIGNAL_INTR );

  DosSetSignalExceptionFocus( SIG_SETFOCUS, &times );

  argNum = argNumIni;
  argPtr = argPtrIni;
  envPtr = envPtrIni;

  char drive[_MAX_DRIVE ],
       dir  [_MAX_DIR ],
       fname[_MAX_FNAME],
       ext  [_MAX_EXT],
       prf[] = "Brake!.Ini";

  _splitpath( argPtr[0], drive, dir, fname, ext );
  prfName = new char [ strlen(drive)+strlen(dir)+sizeof(prf) ];
  strcpy( prfName, drive );
  strcat( prfName, dir );
  strcat( prfName, prf );
  capitalize( prfName );

  brakeLoadPath = new char [ strlen(drive)+strlen(dir)+2 ];
  strcpy( brakeLoadPath, drive );
  strcat( brakeLoadPath, dir );

  int lenPath = strlen(brakeLoadPath);

  if( !lenPath )
    brakeLoadPath = ".\\";
  else if( brakeLoadPath[ lenPath-1 ]!='\\' )
    strcat( brakeLoadPath, "\\" );

  if( firstFirstInit() )
    RC = 2;

  if( !RC )
  {
    APIRET rc = createEventSemaphore( &mainTidSema, "MainTid" );

    if( rc )
    {
      log( Msg_ErrUnableCreateMainSema, rc );
      os2Message( rc );
      RC = 2;
    }
    else if( mainTid = _beginthread( mainThread, 0, 40960, 0 ), mainTid ==-1 )
    {
      log( Msg_ErrUnableStartMain, strError() );
      RC = 2;
    }
    else
      do
        rc = DosWaitEventSem( mainTidSema, (ULONG)SEM_INDEFINITE_WAIT );
      while( rc );
  }

  heapState( "Done!" );
  deRegisterExceptionHandler
  return RC;
}
