#define INCL_DOS
#define INCL_WIN
#define INCL_WINSWITCHLIST
#define INCL_WINPROGRAMLIST
#define INCL_WINWORKPLACE
#include <os2.h>

#include <string.h>

#include "brakepm.h"

class WinTitle
{
  char *winTitle,
       *switchTitle,
       *prevTitle;
  PTIB tib;
  PPIB pib;
  ULONG oldType;
  HSWITCH hSw;
  SWCNTRL swData;
  HAB hab;
  HMQ hmq;
  char forceTitle;
  HPOINTER iconPtr;

protected:
  int goToPm();
  void backFromPm();
  void setBoth( char *wint, char *switcht );

public:
  WinTitle();
  ~WinTitle();

  void setTitle( char * );
  void setIcon();
  void setFrame( int x, int y, int cx, int cy );
  void switchToProgram();
};

WinTitle::WinTitle() :
  hab( 0 ),
  hmq( 0 ),
  iconPtr( 0 ),
  forceTitle( 1 ),
  prevTitle( 0 )
{
  DosGetInfoBlocks( &tib, &pib );

  oldType = pib->pib_ultype;

  memset( &swData, 0, sizeof swData );

  hSw = WinQuerySwitchHandle( 0, pib->pib_ulpid );

  if( hSw!=NULLHANDLE && !WinQuerySwitchEntry( hSw, &swData ) )
  {
    switchTitle = new char [ strlen( swData.szSwtitle )+1 ];
    
    strcpy( switchTitle, swData.szSwtitle );

    if( !goToPm() )
    {
      winTitle = new char [ MAXNAMEL+1 ];
      WinQueryWindowText( swData.hwnd, MAXNAMEL+1, winTitle );
      iconPtr = WinLoadPointer( HWND_DESKTOP, 0, BRAKE_ICON );
    }

    backFromPm();
  }
}

WinTitle::~WinTitle()
{
  setBoth( winTitle, switchTitle );

  delete winTitle;
  delete switchTitle;
  delete prevTitle;
}

int WinTitle::goToPm()
{
  if( hab!=NULLHANDLE && hmq!=NULLHANDLE )
    return 0;

  pib->pib_ultype = 3;

  hab = WinInitialize( 0 );
  hmq = WinCreateMsgQueue( hab, 0 );

  return hab==NULLHANDLE || hmq==NULLHANDLE;
}

void WinTitle::backFromPm()
{
  if( hmq )
  {
    WinDestroyMsgQueue( hmq );
    hmq = 0;
  }

  if( hab )
  {
    WinTerminate( hab );
    hab = 0;
  }

  pib->pib_ultype = oldType;
}

void WinTitle::setTitle( char *title )
{
  if( title == 0 )
    forceTitle = 1;
  else
  {
    char force( char(forceTitle || prevTitle==0) );

    forceTitle = 0;

    if( !force )
      force = (char)(strcmp( prevTitle, title ) != 0);

    if( force )
    {
      setBoth( title, title );
      delete prevTitle;
      prevTitle = new char [ strlen(title)+1 ];
      strcpy( prevTitle, title );
    }
  }
}

void WinTitle::setBoth( char *wint, char *switcht )
{
  if( hSw!=NULLHANDLE && !goToPm() )
  {
    strncpy( swData.szSwtitle, switcht, MAXNAMEL );
    WinChangeSwitchEntry( hSw, &swData );

    if( swData.hwnd )
      WinSetWindowText( swData.hwnd, wint );
  }

  backFromPm();
}

void WinTitle::setFrame( int x, int y, int cx, int cy )
{
  if( hSw != NULLHANDLE )
    WinSetWindowPos( swData.hwnd, 0, x, y, cx, cy, SWP_SIZE|SWP_MOVE|SWP_SHOW );
}

void WinTitle::switchToProgram()
{
  if( hSw != NULLHANDLE )
    WinSwitchToProgram( hSw );
}

void WinTitle::setIcon()
{
  if( hSw!=NULLHANDLE && !goToPm() )
    if( swData.hwnd )
      if( iconPtr!=NULLHANDLE )
        WinSendMsg( swData.hwnd, WM_SETICON, (MPARAM)iconPtr, NULL );

  backFromPm();
}

HAB APIENTRY _Export BrakeWinQueryAnchorBlock( HWND hwnd )
{
  return WinQueryAnchorBlock( hwnd );
}

BOOL APIENTRY _Export BrakePrfCloseProfile( HINI hini )
{
  return PrfCloseProfile( hini );
}

HINI APIENTRY _Export BrakePrfOpenProfile( HAB hab, PCSZ pszFileName )
{
  return PrfOpenProfile( hab, pszFileName );
}

BOOL APIENTRY _Export BrakePrfWriteProfileString( HINI hini,
                                                  PCSZ pszApp,
                                                  PCSZ pszKey,
                                                  PCSZ pszData )
{
  return PrfWriteProfileString( hini, pszApp, pszKey, pszData );
}

BOOL APIENTRY _Export BrakePrfWriteProfileData( HINI hini,
                                                PCSZ  pszApp,
                                                PCSZ  pszKey,
                                                PVOID pData,
                                                ULONG cchDataLen )
{
  return PrfWriteProfileData( hini, pszApp, pszKey, pData, cchDataLen );
}

ULONG APIENTRY _Export BrakePrfQueryProfileString( HINI hini,
                                                   PCSZ  pszApp,
                                                   PCSZ  pszKey,
                                                   PCSZ  pszDefault,
                                                   PVOID pBuffer,
                                                   ULONG cchBufferMax )
{
  return PrfQueryProfileString( hini, pszApp, pszKey, pszDefault,
                                pBuffer, cchBufferMax );
}

BOOL APIENTRY _Export BrakePrfQueryProfileData( HINI hini,
                                                PCSZ  pszApp,
                                                PCSZ  pszKey,
                                                PVOID pBuffer,
                                                PULONG pulBuffLen )
{
  return PrfQueryProfileData( hini, pszApp, pszKey, pBuffer, pulBuffLen );
}

HOBJECT APIENTRY _Export BrakeWinQueryObject( PCSZ objName )
{
  return WinQueryObject( objName );
}

BOOL APIENTRY _Export BrakeWinOpenObject( HOBJECT hObject,
                                          ULONG   ulView,
                                          BOOL    Flag )
{
  return WinOpenObject( hObject, ulView, Flag );
}

int APIENTRY _Export BrakeWinMessage( char *title, char *message )
{
  PTIB tib;
  PPIB pib;

  DosGetInfoBlocks( &tib, &pib );

  ULONG oldType( pib->pib_ultype );

  pib->pib_ultype = 3;

  HAB hab( WinInitialize( 0 ) );
  HMQ hmq( WinCreateMsgQueue( hab, 0 ) );
  
  struct Info
  {
    MB2INFO mb;
    MB2D    i[2];
  };

  Info info = 
  {
    {
      sizeof(Info),
      NULLHANDLE, 3,
      MB_ICONQUESTION,
      NULLHANDLE,
      { "~Continue", 1, 0 }
    },
    { "Online ~registration", 2, 0,
      "Panic ~quit", 3, BS_DEFAULT
    }
  };

  ULONG ret( WinMessageBox2( HWND_DESKTOP, 0, message, title,
                             100, (PMB2INFO)&info ) );
  WinDestroyMsgQueue( hmq );
  WinTerminate( hab );

  pib->pib_ultype = oldType;
  
  return int( ret );
}

char * APIENTRY _Export BrakeGetFromClipboard()
{
  PTIB tib;
  PPIB pib;

  DosGetInfoBlocks( &tib, &pib );

  ULONG oldType( pib->pib_ultype );

  pib->pib_ultype = 3;

  HAB hab( WinInitialize( NULL ) );
  HMQ hmq( WinCreateMsgQueue( hab, 0 ) );

  char *ret( 0 );

  if( WinOpenClipbrd( hab ) )
  {
    ULONG fmtInfo( 0 );

    if( WinQueryClipbrdFmtInfo( hab, CF_TEXT, &fmtInfo )!=FALSE )
    {
      ULONG selClipText( WinQueryClipbrdData( hab, CF_TEXT ) );

      if( selClipText )
      {
        PCHAR pchClipText( (PCHAR)selClipText );
        APIRET rc( DosAllocMem( (PPVOID)&ret, strlen(pchClipText)+1,
                                PAG_READ|PAG_WRITE|PAG_COMMIT ) );

        if( !rc )
          strcpy( ret, pchClipText );
        else
          ret = 0;
      }
    }

    WinCloseClipbrd( hab );
  }

  WinDestroyMsgQueue( hmq );
  WinTerminate( hab );

  pib->pib_ultype = oldType;

  return ret;
}

int APIENTRY _Export BrakeOpenObject( char *obj )
{
  HOBJECT hObject = WinQueryObject( (PCSZ)obj );

  if( !hObject )
    return 1;

  if( !WinOpenObject( hObject, 0, FALSE ) )
    return 2;

  return 0;
}

void * APIENTRY _Export BrakeInitTitle()
{
  return new WinTitle;
}

void APIENTRY _Export BrakeDeInitTitle( void *p )
{
  delete (WinTitle *)p;
}

void APIENTRY _Export BrakeSetTitle( void *p, char *str )
{
  ((WinTitle *)p)->setTitle( str );
}

void APIENTRY _Export BrakeSetFrame( void *p, int x, int y, int cx, int cy )
{
  ((WinTitle *)p)->setFrame( x, y, cx, cy );
}

void APIENTRY _Export BrakeSwitchToForeground( void *p )
{
  ((WinTitle *)p)->switchToProgram();
}

void APIENTRY _Export BrakeSetIcon( void *p )
{
  ((WinTitle *)p)->setIcon();
}

/*
void main()
{
  void *p = BrakeInitTitle();

  BrakeSetTitle( p, "Test" );
  BrakeSetIcon( p );
  DosSleep( 5000 );

  BrakeDeInitTitle( p );
}
*/