#ifndef __MACRO_H

#define __MACRO_H

#include <string.h>

#include "class.h"
#include "newstr.h"

class Macro : public BaseObject
{
  char *name,
       *val;
  char imported;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:
  Macro( char *n, char *v, char im=0 );
  ~Macro();

  char *getName() const { return name; }
  char *getValue() const { return val; }

  char isImported() const { return imported; }
};

class MacroList : public Collection
{
public:
  MacroList();

  Macro* Find( Macro& item, LockMode lock=LockModeNeed ) const;
  Macro& operator [] ( const int pos ) const;
};

class MacroIterator : public CollectionIterator
{
public:
  MacroIterator( MacroList& ini );
  
  Macro& operator () () const;
};

enum ParserState
{
  ParserStateNormal,
  ParserStateIfDef,
  ParserStateIfNDef,
  ParserStateFor
};

class ParserStateStackEntry : public BaseObject
{
private:
  int beginLine;
  char inSkipMode;
  ParserState state;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:
  ParserStateStackEntry( int l, ParserState st, char skip );

  ParserState getState() const { return state; }
  int getLine() const { return beginLine; }
  char getSkipMode() const { return inSkipMode; }
};

class ParserStateStack : public Collection		//ParserStateStackEntry
{
public:
  ParserStateStack();
  
  ParserStateStackEntry* Find( ParserStateStackEntry& item, LockMode lock=LockModeNeed ) const;
  ParserStateStackEntry& operator [] ( const int pos ) const;
};

#endif