//#define __DEMO__CLASSLIB__

#include "class.h"

#include <string.h>
#include <builtin.h>

#ifdef __DEMO__CLASSLIB__

#define INCL_DOSSEMAPHORES
#define INCL_DOSERRORS
#define INCL_DOSPROCESS
#include <os2.h>

#include <time.h>
#include <stdio.h>
#include <stdlib.h>

void SemaphoreCreate( unsigned long *sem )
{
  char sname[ 128 ];
  APIRET rc;

  for( rc=ERROR_DUPLICATE_NAME; rc==ERROR_DUPLICATE_NAME;
       rc = DosCreateMutexSem( (PSZ)sname, sem, DC_SEM_SHARED, 0 ) )
  {
    strcpy( sname, "\\SEM32\\ClassLib\\" );
    ltoa( time( NULL ), sname+strlen(sname), 2 );
  }

  if( rc )
  {
    cerr << "Unable create semaphore: " << rc << endl;
    *sem = 0xffffffffl;
  }
}

inline void SemaphoreRemove( unsigned long sem )
{
  APIRET rc;

  if( sem!=0xffffffffl )
    if( rc=DosCloseMutexSem( sem ), rc )
      cerr << "Unable remove semaphore: " << rc << endl;
}

inline void SemaphoreRequest( unsigned long sem )
{
  APIRET rc;

  if( sem!=0xffffffffl )
    if( rc=DosRequestMutexSem( sem, (ULONG)SEM_INDEFINITE_WAIT ), rc )
      cerr << "Unable request semaphore: " << rc << endl;
}

void SemaphoreRelease( unsigned long sem )
{
  APIRET rc;

  if( sem!=0xffffffffl )
    if( rc=DosReleaseMutexSem( sem ), rc )
      cerr << "Unable release semaphore: " << rc << endl;
}

#else // __DEMO__CLASSLIB__

unsigned long  createMutexSemaphore( char *tag );
void           deleteMutexSemaphore( unsigned long sema );
void           lockMutexSemaphore( unsigned long sema );
void           unlockMutexSemaphore( unsigned long sema );

#endif // __DEMO__CLASSLIB__

#ifdef __DEMO__CLASSLIB__
#define checkCondition(p) \
                  if(!(p)) \
                  { \
                    printf("*BUG* %d: %s\n", __LINE__, #p );\
                    __interrupt( 0x3 ); \
                  }
#else
#define checkCondition(p) {}
#endif

char classLibLockNeed = 0;

// Collection -----------------------------------------------------------------

Collection::Collection( const int isize, const int idelta,
                        const SemaphoreMode sema,
                        const DeleteMode del ) :
  used( 0 ),
  size( isize ),
  delta( idelta ),
  deleteMode( del ),
  semaphore( 0xffffffffl ),
  needSema( sema )
{
  list = Allocate( size );

  if( sema==SemaphoreModeNeed )
#ifdef __DEMO__CLASSLIB__
    SemaphoreCreate( &semaphore );
#else
  semaphore = createMutexSemaphore( "i don't know now" );
#endif
}

Collection::~Collection()
{
  Lock();

  if( deleteMode==DeleteModeDelete )
    for( int i=0; i<used; i++ )
      delete list[i];

  delete list;
  UnLock();

  if( needSema==SemaphoreModeNeed )
#ifdef __DEMO__CLASSLIB__
    SemaphoreRemove( semaphore );
#else
    deleteMutexSemaphore( semaphore );
#endif
}

void Collection::Lock() const
{
  if( needSema==SemaphoreModeNeed && classLibLockNeed )
#ifdef __DEMO__CLASSLIB__
    SemaphoreRequest( semaphore );
#else
    lockMutexSemaphore( semaphore );
#endif
}

void Collection::UnLock() const
{
  if( needSema==SemaphoreModeNeed && classLibLockNeed )
#ifdef __DEMO__CLASSLIB__
    SemaphoreRelease( semaphore );
#else
    unlockMutexSemaphore( semaphore );
#endif
}

BaseObject** Collection::Allocate( const int size ) const
{
  checkCondition( size!=0 )

  BaseObject** newList = new BaseObject* [size];

  memset( newList, 0, sizeof(BaseObject**)*size );
  return newList;
}

void Collection::Expand()
{
  size += delta;

  BaseObject **newList = Allocate( size );

  memcpy( newList, list, sizeof(BaseObject**)*used );
  delete list;
  list = newList;
}

void Collection::MoveRight( const int pos )
{
  checkCondition( pos<used && pos>=0 )

  memmove( list+pos+1, list+pos, sizeof(BaseObject**)*(used-pos) );
}

void Collection::Add( BaseObject& item, const int pos )
{
  checkCondition( pos<=used && pos>=0 )
  checkCondition( used+1<=size )
  checkCondition( size!=0 )

  Lock();

  if( used+1==size )
    Expand();

  if( list[pos]!=0 )
    MoveRight( pos );

  list[pos] = &item;
  used++;
  UnLock();
}

void Collection::Replace( BaseObject& item, DeleteMode del )
{
  Lock();

  BaseObject *found = Find( item, LockModeNoNeed );

  if( found!=NULL )
    Remove( *found, del );

  AddAtTail( item );
  UnLock();
}

BaseObject& Collection::operator [] ( const int pos ) const
{
  checkCondition( pos<used && pos>=0 )

  return *list[pos];
}

int Collection::GetIndex( BaseObject& item, LockMode lock ) const
{
  if( !used )
    return -1;

  if( lock==LockModeNeed )
    Lock();

  int iter;

  for( iter=0; iter<used; iter++ )
    if( *(list[iter]) == item )
      break;

  if( lock==LockModeNeed )
    UnLock();

  return iter==used ? -1 : iter;
}

void Collection::MoveLeft( const int ind )
{
  checkCondition( ind<used && ind>=0 )

  if( ind!=used-1 )
    memmove( list+ind, list+ind+1, sizeof(BaseObject**)*(used-1-ind) );
}

void Collection::Remove( int ind, DeleteMode del, LockMode lock )
{
  checkCondition( ind>=0 && ind<used )

  if( lock==LockModeNeed )
    Lock();

  if( del==DeleteModeDelete ||
      (del==DeleteModeDefault && deleteMode==DeleteModeDelete ) )
    delete list[ind];

  MoveLeft( ind );

  --used;
  list[used] = 0;

  if( lock==LockModeNeed )
    UnLock();
}

void Collection::Remove( BaseObject& item, DeleteMode del, LockMode lock )
{
  if( used )
  {
    if( lock==LockModeNeed )
      Lock();

    int ind = GetIndex( item, LockModeNoNeed );

    if( ind!=-1 )
    {
      if( del==DeleteModeDelete ||
          (del==DeleteModeDefault && deleteMode==DeleteModeDelete ) )
        delete list[ind];

      MoveLeft( ind );

      --used;
      list[used] = 0;
    }

    if( lock==LockModeNeed )
      UnLock();
  }
}

void Collection::Flush( DeleteMode del )
{
  if( used )
  {
    Lock();

    while( used )
      Remove( used-1, del, LockModeNoNeed );

    UnLock();
  }
}

BaseObject* Collection::Find( BaseObject& item, LockMode lock ) const
{
  int ind = GetIndex( item, lock );

  return ind==-1 ? (BaseObject *)0 : list[ind];
}

int Collection::HasMember( BaseObject& item ) const
{
  return Find( item )!=0;
}

void Collection::AddAtHead( BaseObject& item )
{
  Add( item, 0 );
}

void Collection::AddAtTail( BaseObject& item )
{
  Lock();
  Add( item, used );
  UnLock();
}

void Collection::Add( BaseObject& item )
{
  AddAtTail( item );
}

ostream & operator << ( ostream& o, const Collection& c )
{
  c.Lock();
  o << "  Total " << c.used << " items (sem=" ;

  if( c.semaphore==0xffffffffl )
    o << "none";
  else
    o << c.semaphore;

  o << char(')') << endl;

  for( int i=0; i<c.used; i++ )
    o << "  "
      << i
      << "> "
      << *(c.list[i])
      << endl;

  c.UnLock();
  return o;
}

// BaseObject -----------------------------------------------------------------

BaseObject::BaseObject()
{
}

BaseObject::~BaseObject()
{
}

int operator == ( const BaseObject &obj1, const BaseObject &obj2 )
{ 
  if( !strcmp( obj1.nameOfClass(), obj2.nameOfClass() ) )
    return obj1.equal( obj2 );
    
  return 0;
}

ostream& operator << ( ostream& stream, const BaseObject& obj )
{
  stream << char('{') << obj.nameOfClass() << char(':');
  obj.print( stream );
  stream << char('}');
  return stream;
}

// CollectionIterator ---------------------------------------------------------

CollectionIterator::CollectionIterator( const Collection& _c ) :
  current(0),
  c( _c )
{
   _c.Lock();
}

CollectionIterator::~CollectionIterator()
{
  c.UnLock();
}

BaseObject& CollectionIterator::operator () () const
{
  return c[current];
}

CollectionIterator::operator int() const
{
  return current<c.used;
}

void CollectionIterator::operator ++()
{
  ++current;
}

void CollectionIterator::operator ++( int )
{
  ++current;
}

void CollectionIterator::restart()
{
  current=0;
}

// VOID MAIN ------------------------------------------------------------------

#ifdef __DEMO__CLASSLIB__

class Object : public BaseObject 
{
  int i;
    
public:
  Object( int _i ) : i(_i) {}
    
  int equal( const BaseObject& t ) const { return i == ((const Object *)&t)->i; }
  void print( ostream& s) const { s << i; }
  char *nameOfClass() const { return "Object"; }
};

void main()
{
  Collection aa;
  
  for( int i(0); i<100; i++ )
    aa.Add( *new Object( i ) );

  int tot( aa.GetItemsInContainer() );
    
  for( i=0; i<tot; i+=5 )
  {
    Object rem( i );

    aa.Remove( *(BaseObject*)&rem );
  }
    
  cout << aa;
}
#endif