#ifndef __MY_CLASS_LIB

#define  __MY_CLASS_LIB

#include <iostream.h>

extern char classLibLockNeed;

class BaseObject
{
  virtual int equal( const BaseObject &t ) const = 0;
  virtual void print( ostream& s ) const = 0;
  virtual char *nameOfClass() const = 0;
  
public:
  BaseObject();
  virtual ~BaseObject();

  friend int operator == ( const BaseObject &obj1, const BaseObject &obj2 );
  friend ostream& operator << ( ostream& stream, const BaseObject& obj );
};

enum SemaphoreMode
{
  SemaphoreModeNoNeed = 0,
  SemaphoreModeNeed = 1
};

enum LockMode
{
  LockModeNoNeed = 0,
  LockModeNeed = 1
};

enum DeleteMode
{
  DeleteModeNoDelete = 0,
  DeleteModeDelete = 1,
  DeleteModeDefault = 2
};

class Collection
{
protected:

  BaseObject **list;
  int used,
      size,
      delta;
  DeleteMode deleteMode;
  SemaphoreMode needSema;
  unsigned long semaphore;

  BaseObject** Allocate( const int size ) const;
  void Expand();
  void MoveRight( const int pos );
  void MoveLeft( const int ind );

  BaseObject* Find( BaseObject& item, LockMode lock=LockModeNeed ) const;
  BaseObject& operator[]( const int pos ) const;

public:

  Collection( const int size=10, const int delta=5,
	      const SemaphoreMode needSema=SemaphoreModeNeed,
	      const DeleteMode del=DeleteModeDelete );
  ~Collection();

  void Lock() const;
  void UnLock() const;

  int GetIndex( BaseObject& item, LockMode lock=LockModeNeed ) const;

  int HasMember( BaseObject& item ) const;

  void Add( BaseObject& item, const int pos );
  void AddAtHead( BaseObject& item );
  void AddAtTail( BaseObject& item );
  void Add( BaseObject& item );

  void Replace( BaseObject& item, DeleteMode del=DeleteModeDefault );
  void Remove( BaseObject& item, DeleteMode del=DeleteModeDefault, LockMode lock=LockModeNeed );
  void Remove( int item, DeleteMode del=DeleteModeDefault, LockMode lock=LockModeNeed );

  void Flush( DeleteMode del=DeleteModeDefault );

  int GetItemsInContainer() const { return used; }

  friend ostream& operator << ( ostream& o, const Collection& c );
  friend class CollectionIterator;
};

class CollectionIterator
{
private:
  int current;
  const Collection& c;

protected:
  BaseObject& operator()() const;

public:
  CollectionIterator( const Collection& _c );
  ~CollectionIterator();

  operator int() const;
  void operator ++();
  void operator ++( int );
  void restart();
};

#endif

