#define INCL_DOSPROCESS
#include <os2.h>

#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>
#include <process.h>

#include "language.h"

#include "misc.h"
#include "process.h"
#include "buffer.h"
#include "fidoaddr.h"
#include "nodelist.h"
#include "translat.h"
#include "pair.h"
#include "zmodem.h"
#include "zinter.h"
#include "freq.h"
#include "calls.h"
#include "enable.h"
#include "config.h"
#include "rexxhook.h"
#include "timeout.h"
#include "hydra.h"
#include "addinqu.h"
#include "nolog.h"
#include "fax.h"

static char miscPresent = 0,
            modemPresent = 0;

void unexpMsg( char *what )
{
  log( Msg_ErrCfgUnExpect, what, cfgLine );
}

int storeWord( ifstream **cfg, char **where, char *what )
{
  if( getWord( cfg ) )
  {
    unexpMsg( what );
    return 1;
  }

  delete *where;
  *where = newStr( inputWord.str );
  return 0;

}

int storeNumber( ifstream **cfg, int &number, char *msg )
{
  number = 0;

  if( getWord( cfg ) )
  {
    unexpMsg( msg );
    return 1;
  }

  for( int i=0; i<inputWord.length()-1; i++ )
    if( isdigit( inputWord[i] ) )
      number = number*10 + (inputWord[i]-'0');
    else
    {
      log( Msg_ErrCfgExpect, msg, inputWord.str, cfgLine );
      return 2;
    }

  if( !number )
  {
    log( Msg_ErrCfgNoZero, cfgLine );
    return 2;
  }

  return 0;
}

int storeLong( ifstream **cfg, long &number, char *msg )
{
  number = 0;

  if( getWord( cfg ) )
  {
    unexpMsg( msg );
    return 1;
  }

  for( int i=0; i<inputWord.length()-1; i++ )
    if( isdigit( inputWord[i] ) )
      number = number*10 + (inputWord[i]-'0');
    else
    {
      log( Msg_ErrCfgExpect, msg, inputWord.str, cfgLine );
      return 2;
    }

  if( !number )
  {
    log( Msg_ErrCfgNoZero, cfgLine );
    return 2;
  }

  return 0;
}

int storePath( ifstream **cfg, char **where, char *what, int flag )
{
  if( getWord( cfg ) )
  {
    unexpMsg( what );
    return 1;
  }

  if( inputWord[ inputWord.length()-2 ]!='\\' )
  {
    inputWord.stripByte();
    inputWord.putInto( '\\' );
    inputWord.putInto( 0 );
  }

  delete *where;
  *where = newStr( inputWord.str );

  if( flag )
    if( makePath( *where ) )
    {
      log( Msg_ErrUnableMkDir, *where, strError() );
      return 3;
    }

  return 0;
}

int strIsYes( char *str )
{
  if( !stricmp( str, "Yes" )  || !stricmp( str, "On" ) ||
      !stricmp( str, "True" ) || !stricmp( str, "Rulez" ) )
    return 1;

  return 0;
}

int strIsNo( char *str )
{
  if( !stricmp( str, "No" )    || !stricmp( str, "Off" ) ||
      !stricmp( str, "False" ) || !stricmp( str, "Suxx" ) )
    return 1;

  return 0;
}

int storeLogic( ifstream **cfg, int& yesno, char *who )
{
  if( getWord( cfg ) )
  {
    unexpMsg( who );
    return 1;
  }

  yesno = 0;

  if( strIsYes( inputWord.str ) )
    yesno = 1;
  else
    if( strIsNo( inputWord.str ) )
      yesno = 0;
    else
      return need( Msg_ErrCfgNeedYesNo );

  return 0;
}

int storeSpeed( ifstream **cfg, FOSSILbaud& speed, char *msg )
{
  if( getWord( cfg ) )
  {
    unexpMsg( msg );
    return 1;
  }

  miscInfo.baud = 0;

  for( int i=0; i<inputWord.length()-1; i++ )
    if( isdigit( inputWord[i] ) )
      miscInfo.baud = miscInfo.baud*10 + (inputWord[i]-'0');
    else
      return 3;

  switch( miscInfo.baud )
  {
  case 300:
    speed = B300;
    break;

  case 600:
    speed = B600;
    break;

  case 1200:
    speed = B1200;
    break;

  case 2400:
    speed = B2400;
    break;

  case 4800:
    speed = B4800;
    break;

  case 9600:
    speed = B9600;
    break;

  case 19200:
    speed = B19200;
    break;

  case 38400l:
    speed = B38400;
    break;

  case 57600l:
    speed = B57600;
    break;

  case 76800l:
    speed = B76800;
    break;

  case 115200l:
    speed = B115200;
    break;

  case 230400l:
    speed = B230400;
    break;

  case 345600l:
    speed = B345600;
    break;

  default:
    log( Msg_ErrCfgUnkSpeed, inputWord.str, cfgLine );
    return 2;
  }

  return 0;
}

int storeMode( ifstream **cfg, Flavour& flavour, char *msg )
{
  if( getWord( cfg ) )
  {
    unexpMsg( msg );
    return 1;
  }

  if( !stricmp( inputWord.str, "Hold" ) )
    flavour = holdFlavour;
  else if( !stricmp( inputWord.str, "Normal" ) )
    flavour = normalFlavour;
  else if( !stricmp( inputWord.str, "Crash" ) )
    flavour = crashFlavour;
  else if( !stricmp( inputWord.str, "Immed" ) )
    flavour = immedFlavour;
  else if( !stricmp( inputWord.str, "Direct" ) )
    flavour = directFlavour;
  else
    return need( Msg_ErrCfgUnkFlavour );

  return 0;
}

int processAKA( ifstream **cfg, FIDOAddressList **list, int *aLine )
{
  uShort zone, net, node, point;
  char *domain( NULL );
  int rc( 0 );

  if( aLine!=NULL && !*aLine  )
    *aLine = cfgLine;

  if( getWord( cfg ) )
    return 1;

  if( checkAddress( inputWord.str, zone, net, node, point, &domain ) )
  {
    if( *list==NULL )
      *list = new FIDOAddressList;

    (*list)->AddAtTail( *new FIDOAddress( zone, net, node, point, domain ) );
  }
  else
    rc = need( Msg_ErrCfgNeedAddr );

  return rc;
}

int processEmsi( ifstream **cfg )
{

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Address" ) )
     rc = processAKA( cfg, &systemInfo.addr, NULL );
   else if( !stricmp( inputWord.str, "SysOp" ) )
     rc = storeWord( cfg, &systemInfo.sysopName, "SysOp" );
   else if( !stricmp( inputWord.str, "System" ) )
     rc = storeWord( cfg, &systemInfo.name, "System" );
   else if( !stricmp( inputWord.str, "Place" ) )
     rc = storeWord( cfg, &systemInfo.place, "Place" );
   else if( !stricmp( inputWord.str, "Phone" ) )
     rc = storeWord( cfg, &systemInfo.phone, "Phone" );
   else if( !stricmp( inputWord.str, "Speed" ) )
     rc = storeWord( cfg, &systemInfo.baud, "Speed" );
   else if( !stricmp( inputWord.str, "Flags" ) )
     rc = storeWord( cfg, &systemInfo.flags, "Flags" );
  else
     return 3;
  }

  if( rc )
    return rc;

  if( systemInfo.addr->GetItemsInContainer() )
    defaultAddress = &(*systemInfo.addr)[0];

  return 0;
}

int storePriority( ifstream **cfg, unsigned long &priority )
{
  int rc( 0 );

  if( getWord( cfg ) )
  {
    unexpMsg( Msg_ErrCfgPryClass );
    rc = 1;
  }
  else if( !stricmp( inputWord.str, "IdleTime" ) )
    priority = PRTYC_IDLETIME;
  else if( !stricmp( inputWord.str, "Regular" ) )
    priority = PRTYC_REGULAR;
  else if( !stricmp( inputWord.str, "TimeCritical" ) )
    priority = PRTYC_TIMECRITICAL;
  else if( !stricmp( inputWord.str, "ForeGroundServer" ) )
    priority = PRTYC_FOREGROUNDSERVER;
  else
    rc = need( "IdleTime/Regular/TimeCritical/ForeGroundServer" );

  return rc;
}

int processAddInQueue( ifstream **cfg )
{
  char *name( NULL ),
       *pars( NULL );

  int rc = storeWord( cfg, &name, "DLL Name" );

  if( rc )
    return rc;

  rc = storeWord( cfg, &pars, "DLL Parametrs" );

  if( rc )
    return rc;

  addInQueueList->AddAtTail( *new AddInQueue( name, pars ) );
  return 0;
}

int processNoLog( ifstream **cfg, NoLogList **list, char *who )
{
  char *str( NULL );

  int rc = storeWord( cfg, &str, who );

  if( rc )
    return rc;

  if( *list==NULL )
    *list = new NoLogList;

  (*list)->AddAtTail( *new NoLog( str ) );

  delete str;
  return 0;
}

int processAddressForm( ifstream **cfg )
{
  while( 1 )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "ZeroPoint" ) )
      miscInfo.addrForm[1] = 'P';
    else if( !stricmp( inputWord.str, "NoZeroPoint" ) )
      miscInfo.addrForm[1] = 'I';
    else if( !stricmp( inputWord.str, "Domain" ) )
    {
      if( getWord( cfg ) )
        return 1;

      if( strlen( inputWord.str )!=1 )
        return need( Msg_ErrCfgIllDomainChar );

      miscInfo.addrForm[0] = inputWord.str[0];
    }
    else
    {
      parserNoRead = 1;
      break;
    }
  }

  return 0;
}

int processMisc( ifstream **cfg )
{
  char *zmh( NULL ),
       *answerTime( NULL ),
       *callTime( NULL ),
       *callOnlyImmed( NULL ),
       *callOnlyCrash( NULL ),
       *callOnlyDirect( NULL ),
       *callOnlyNormal( NULL );
  int zmhLine,
      callLine,
      callImmedLine,
      callCrashLine,
      callDirectLine,
      callNormalLine,
      answerLine,
      zd( -1 ),
      ed( -1 ),
      discard( -1 ),
      adjust( -1 ),
      hold( -1 ),
      bsyabort( -1 ),
      retry( -1 ),
      EMSItime( 0 ),
      maxIn( 0 ),
      maxOut( 0 ),
      fulllog( -1 ),
      rexx( -1 ),
      sema( -1 ),
      taskin( -1 ),
      hpfs( -1 ),
      status( -1 ),
      chName( -1 ),
      bink( -1 ),
      fastScan( -1 ),
      full5D( -1 ),
      scanOnLine( -1 ),
      scanBefore( -1 ),
      EA( -1 ),
      termlog( -1 ),
      callAfterAnswer( -1 ),
      binkDRecover( -1 ),
      exportConfig( -1 );

  miscInfo.semaphore = newStr( miscInfo.semaphore );

  miscPresent = 1;

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Circle" ) )
    {
      rc = storeNumber( cfg, miscInfo.circleMax, "CircleMax" );
      if( rc )
	return rc;

      rc = storeNumber( cfg, miscInfo.circleMin, "CircleMin" );
      if( rc )
	return rc;

      if( miscInfo.circleMax < miscInfo.circleMin )
      {
        log( Msg_ErrCfgCircle, cfgLine );
	rc = miscInfo.circleMax;
	miscInfo.circleMax = miscInfo.circleMin;
	miscInfo.circleMin = rc;
	rc = 0;
      }
    }
    else if( !stricmp( inputWord.str, "AddInQueue" ) )
      rc = processAddInQueue( cfg );
    else if( !stricmp( inputWord.str, "TimeSync" ) )
    {
      if( getWord( cfg ) )
        return 1;

      EnableItem *item = new EnableItem( inputWord.str, EnableItemAddress );

      if( !item->isCorrect() )
      {
        log( Msg_ErrCfgIllRegExp, inputWord.str, cfgLine, item->getRegErrorMsg() );
        delete item;
        return 2;
      }

      if( timeSyncList==NULL )
        timeSyncList = new EnableList;

      timeSyncList->AddAtTail( *item );
    }
    else if( !stricmp( inputWord.str, "TryToSend" ) )
      rc = storeNumber( cfg, miscInfo.sendTry, "TryToSend" );
    else if( !stricmp( inputWord.str, "InFileBufferSize" ) )
      rc = storeNumber( cfg, miscInfo.inFileBufferSize, "InFileBufferSize" );
    else if( !stricmp( inputWord.str, "OutFileBufferSize" ) )
      rc = storeNumber( cfg, miscInfo.outFileBufferSize, "OutFileBufferSize" );
    else if( !stricmp( inputWord.str, "HoldTime" ) )
      rc = storeNumber( cfg, miscInfo.holdTime, "HoldTime" );
    else if( !stricmp( inputWord.str, "TimeSyncDelta" ) )
      rc = storeNumber( cfg, miscInfo.timeSyncDelta, "TimeSyncDelta" );
    else if( !stricmp( inputWord.str, "AutoReScan" ) )
      rc = storeNumber( cfg, miscInfo.rescan, "AutoReScan" );
    else if( !stricmp( inputWord.str, "ReScanSleep" ) )
      rc = storeNumber( cfg, miscInfo.rescanPeriod, "ReScanSleep" );
    else if( !stricmp( inputWord.str, "ImmedPriority" ) )
      rc = storeNumber( cfg, immedFlavourPriority, "ImmedPriority" );
    else if( !stricmp( inputWord.str, "CrashPriority" ) )
      rc = storeNumber( cfg, crashFlavourPriority, "CrashPriority" );
    else if( !stricmp( inputWord.str, "DirectPriority" ) )
      rc = storeNumber( cfg, directFlavourPriority, "DirectPriority" );
    else if( !stricmp( inputWord.str, "NormalPriority" ) )
      rc = storeNumber( cfg, normalFlavourPriority, "NormalPriority" );
    else if( !stricmp( inputWord.str, "HoldPriority" ) )
      rc = storeNumber( cfg, holdFlavourPriority, "HoldPriority" );
    else if( !stricmp( inputWord.str, "ConfigCheck" ) )
    {
      rc = storeLong( cfg, miscInfo.configCheck, "ConfigCheck" );
      miscInfo.configCheck *= 1000;
    }
    else if( !stricmp( inputWord.str, "LogBuffer" ) )
    {
      rc = storeNumber( cfg, miscInfo.logBuffer, "LogBuffer" );
      miscInfo.logBuffer <<= 10;
    }
    else if( !stricmp( inputWord.str, "BSYMaxAge" ) )
    {
      rc = storeNumber( cfg, miscInfo.bsyAge, "BSYMaxAge" );
      miscInfo.bsyAge *= 60*60;
    }
    else if( !stricmp( inputWord.str, "BSYCheck" ) )
      rc = storeNumber( cfg, miscInfo.bsyInterval, "BSYCheck" );
    else if( !stricmp( inputWord.str, "MaxCallCost" ) )
      rc = storeLong( cfg, miscInfo.maxCost, "MaxCallCost" );
    else if( !stricmp( inputWord.str, "TerminalTimeOut" ) )
      rc = storeLong( cfg, miscInfo.terminalTimeOut, "TerminalTimeOut" );
    else if( !stricmp( inputWord.str, "FlagsCheck" ) )
      rc = storeNumber( cfg, miscInfo.checkFlags, "FlagsCheck" );
    else if( !stricmp( inputWord.str, "ScheduleCheck" ) )
      rc = storeNumber( cfg, miscInfo.checkSched, "ScheduleCheck" );
    else if( !stricmp( inputWord.str, "InBound" ) )
      rc = storePath( cfg, &miscInfo.inbound, "InBound" );
    else if( !stricmp( inputWord.str, "BsySemaPath" ) )
      rc = storePath( cfg, &miscInfo.bsySemaPath, "BsySemaPath", 0 );
    else if( !stricmp( inputWord.str, "Protected" ) )
      rc = storePath( cfg, &miscInfo.protectedInbound, "Protected" );
    else if( !stricmp( inputWord.str, "Flags" ) )
      rc = storePath( cfg, &miscInfo.flags, "Flags", 0 );
    else if( !stricmp( inputWord.str, "RexxPath" ) )
      rc = storePath( cfg, &miscInfo.rexxPath, "RexxPath", 0 );
    else if( !stricmp( inputWord.str, "TerminalInBound" ) )
      rc = storePath( cfg, &miscInfo.terminalInbound, "TerminalInBound" );
    else if( !stricmp( inputWord.str, "SentList" ) )
      rc = storeWord( cfg, &miscInfo.sentList, "SentList" );
    else if( !stricmp( inputWord.str, "TBoxes" ) )
      rc = storePath( cfg, &miscInfo.tbox, "TBoxes", 0 );
    else if( !stricmp( inputWord.str, "LongBoxes" ) )
      rc = storePath( cfg, &miscInfo.longBox, "LongBoxes", 0 );
    else if( !stricmp( inputWord.str, "LongOutbound" ) )
      rc = storePath( cfg, &miscInfo.longOutbound, "LongOutbound" );
    else if( !stricmp( inputWord.str, "BinkDRecover" ) )
      rc = storeLogic( cfg, binkDRecover, "BinkDRecover" );
    else if( !stricmp( inputWord.str, "HPFS" ) )
      rc = storeLogic( cfg, hpfs, "HPFS" );
    else if( !stricmp( inputWord.str, "4DCompatible" ) )
      rc = storeLogic( cfg, full5D, "4DCompatible" );
    else if( !stricmp( inputWord.str, "FastScan" ) )
      rc = storeLogic( cfg, fastScan, "FastScan" );
    else if( !stricmp( inputWord.str, "ScanOnLine" ) )
      rc = storeLogic( cfg, scanOnLine, "ScanOnLine" );
    else if( !stricmp( inputWord.str, "ScanBefore" ) )
      rc = storeLogic( cfg, scanBefore, "ScanBefore" );
    else if( !stricmp( inputWord.str, "ExportConfig" ) )
      rc = storeLogic( cfg, exportConfig, "ExportConfig" );
    else if( !stricmp( inputWord.str, "EA" ) )
      rc = storeLogic( cfg, EA, "EA" );
    else if( !stricmp( inputWord.str, "TaskInLog" ) )
      rc = storeLogic( cfg, taskin, "TaskInLog" );
    else if( !stricmp( inputWord.str, "FullLog" ) )
      rc = storeLogic( cfg, fulllog, "FullLog" );
    else if( !stricmp( inputWord.str, "StatusLine" ) )
      rc = storeLogic( cfg, status, "StatusLine" );
    else if( !stricmp( inputWord.str, "StatusPeriod" ) )
      rc = storeLong( cfg, miscInfo.statusLine, "StatusPeriod" );
    else if( !stricmp( inputWord.str, "RexxDebug" ) )
      rc = storeLogic( cfg, rexx, "RexxDebug" );
    else if( !stricmp( inputWord.str, "RetryAfterExtern" ) )
      rc = storeLogic( cfg, retry, "RetryAfterExtern" );
    else if( !stricmp( inputWord.str, "BSYSemaphore" ) )
      rc = storeLogic( cfg, sema, "BSYSemaphore" );
    else if( !stricmp( inputWord.str, "CheckName" ) )
      rc = storeLogic( cfg, chName, "CheckName" );
    else if( !stricmp( inputWord.str, "BinkLog" ) )
      rc = storeLogic( cfg, bink, "BinkLog" );
    else if( !stricmp( inputWord.str, "CallAfterAnswer" ) )
      rc = storeLogic( cfg, callAfterAnswer, "CallAfterAnswer" );
    else if( !stricmp( inputWord.str, "TerminalLog" ) )
      rc = storeLogic( cfg, termlog, "TerminalLog" );
    else if( !stricmp( inputWord.str, "ZMH" ) )
    {
      zmhLine = cfgLine;
      rc = storeWord( cfg, &zmh, "ZMH" );
    }
    else if( !stricmp( inputWord.str, "CallTime" ) )
    {
      callLine = cfgLine;
      rc = storeWord( cfg, &callTime, "CallTime" );
    }
    else if( !stricmp( inputWord.str, "CallOnlyImmed" ) )
    {
      callImmedLine = cfgLine;
      rc = storeWord( cfg, &callOnlyImmed, "CallOnlyImmed" );
    }
    else if( !stricmp( inputWord.str, "CallOnlyCrash" ) )
    {
      callCrashLine = cfgLine;
      rc = storeWord( cfg, &callOnlyCrash, "CallOnlyCrash" );
    }
    else if( !stricmp( inputWord.str, "CallOnlyDirect" ) )
    {
      callDirectLine = cfgLine;
      rc = storeWord( cfg, &callOnlyDirect, "CallOnlyDirect" );
    }
    else if( !stricmp( inputWord.str, "CallOnlyNormal" ) )
    {
      callNormalLine = cfgLine;
      rc = storeWord( cfg, &callOnlyNormal, "CallOnlyNormal" );
    }
    else if( !stricmp( inputWord.str, "AnswerTime" ) )
    {
      answerLine = cfgLine;
      rc = storeWord( cfg, &answerTime, "AnswerTime" );
    }
    else if( !stricmp( inputWord.str, "ReDial" ) )
      rc = storeNumber( cfg, miscInfo.redial, "ReDial" );
    else if( !stricmp( inputWord.str, "MaxTry" ) )
      rc = storeNumber( cfg, miscInfo.maxTry, "MaxTry" );
    else if( !stricmp( inputWord.str, "WaitMailer" ) )
      rc = storeNumber( cfg, EMSItime, "WaitMailer" );
    else if( !stricmp( inputWord.str, "MaxInComing" ) )
      rc = storeNumber( cfg, maxIn, "MaxInComing" );
    else if( !stricmp( inputWord.str, "MaxOutGoing" ) )
      rc = storeNumber( cfg, maxOut, "MaxOutGoing" );
    else if( !stricmp( inputWord.str, "Debug" ) )
      rc = storeLogic( cfg, zd, "Debug" );
    else if( !stricmp( inputWord.str, "EMSIDebug" ) )
      rc = storeLogic( cfg, ed, "EMSIDebug" );
    else if( !stricmp( inputWord.str, "EMSILog" ) )
      rc = storeWord( cfg, &miscInfo.emsiLog, "EMSILog" );
    else if( !stricmp( inputWord.str, "LogName" ) )
      rc = storeWord( cfg, &miscInfo.logName, "LogName" );
    else if( !stricmp( inputWord.str, "Pipe" ) )
      rc = storeWord( cfg, &miscInfo.pipe, "Pipe" );
    else if( !stricmp( inputWord.str, "BinaryLog" ) )
      rc = storeWord( cfg, &miscInfo.binLogName, "BinaryLog" );
    else if( !stricmp( inputWord.str, "Schedule" ) )
      rc = storeWord( cfg, &miscInfo.schedule, "Schedule" );
    else if( !stricmp( inputWord.str, "Compiled" ) )
      rc = storeWord( cfg, &miscInfo.schedComp, "Compiled" );
    else if( !stricmp( inputWord.str, "Threshold" ) )
      rc = storeNumber( cfg, miscInfo.trash, "Threshold" );
    else if( !stricmp( inputWord.str, "CompilerPriority" ) )
      rc = storePriority( cfg, miscInfo.compilerPriority );
    else if( !stricmp( inputWord.str, "CompilerPriorityDelta" ) )
      rc = storeLong( cfg, miscInfo.compilerPriorityDelta, "CompilerPriorityDelta" );
    else if( !stricmp( inputWord.str, "ScanPriority" ) )
      rc = storePriority( cfg, miscInfo.reScanPriority );
    else if( !stricmp( inputWord.str, "ScanPriorityDelta" ) )
      rc = storeLong( cfg, miscInfo.reScanPriorityDelta, "ScanPriorityDelta" );
    else if( !stricmp( inputWord.str, "Priority" ) )
      rc = storePriority( cfg, miscInfo.priority );
    else if( !stricmp( inputWord.str, "ForeGroundDelta" ) )
      rc = storeLong( cfg, miscInfo.priorityDelta, "ForeGroundDelta" );
    else if( !stricmp( inputWord.str, "DiscardPartialPKT" ) )
      rc = storeLogic( cfg, discard, "DiscardPartialPKT" );
    else if( !stricmp( inputWord.str, "AbortOnBSY" ) )
      rc = storeLogic( cfg, bsyabort, "AbortOnBSY" );
    else if( !stricmp( inputWord.str, "EmptyExit" ) )
    {
      rc = storeNumber( cfg, miscInfo.emptyExit, "EmptyExit" );

      if( !rc )
        if( miscInfo.emptyExit<10 || miscInfo.emptyExit>255 )
        {
          log( Msg_ErrCfgIllNumb, inputWord.str, cfgLine );
          rc = 2;
        }
    }
    else if( !stricmp( inputWord.str, "DOSTime" ) )
      rc = storeLogic( cfg, adjust, "DOSTime" );
    else if( !stricmp( inputWord.str, "HoldOnCall" ) )
      rc = storeLogic( cfg, hold, "HoldOnCall" );
    else if( !stricmp( inputWord.str, "Semaphore" ) )
      rc = storeWord( cfg, &miscInfo.semaphore, "Semaphore" );
    else if( !stricmp( inputWord.str, "Separator" ) )
    {
      if( getWord( cfg ) )
        return 1;

      miscInfo.separator = *inputWord.str;
    }
    else if( !stricmp( inputWord.str, "QueueSort" ) )
    {
      if( getWord( cfg ) )
        return 1;

      if( !stricmp( inputWord.str, "Address" ) )
        miscInfo.sortMode = QueueSortAddr;
      else if( !stricmp( inputWord.str, "+Address" ) )
        miscInfo.sortMode = QueueSortAddr;
      else if( !stricmp( inputWord.str, "-Address" ) )
        miscInfo.sortMode = QueueSortAddrDec;

      else if( !stricmp( inputWord.str, "Size" ) )
        miscInfo.sortMode = QueueSortSize;
      else if( !stricmp( inputWord.str, "+Size" ) )
        miscInfo.sortMode = QueueSortSize;
      else if( !stricmp( inputWord.str, "-Size" ) )
        miscInfo.sortMode = QueueSortSizeDec;

      else if( !stricmp( inputWord.str, "Priority" ) )
        miscInfo.sortMode = QueueSortPriority;
      else if( !stricmp( inputWord.str, "+Priority" ) )
        miscInfo.sortMode = QueueSortPriority;
      else if( !stricmp( inputWord.str, "-Priority" ) )
        miscInfo.sortMode = QueueSortPriorityDec;

      else if( !stricmp( inputWord.str, "Age" ) )
        miscInfo.sortMode = QueueSortAge;
      else if( !stricmp( inputWord.str, "+Age" ) )
        miscInfo.sortMode = QueueSortAge;
      else if( !stricmp( inputWord.str, "-Age" ) )
        miscInfo.sortMode = QueueSortAgeDec;

      else if( !stricmp( inputWord.str, "NoSort" ) )
        miscInfo.sortMode = QueueSortNoSort;
      else
        return need( "Address/-Address/Size/-Size/Priority/-Priority/Age/-Age/Nosort" );
    }
    else if( !stricmp( inputWord.str, "FileCase" ) )
    {
      if( getWord( cfg ) )
        return 1;

      if( !stricmp( inputWord.str, "Lower" ) )
        miscInfo.fileCase = FileCaseLower;
      else if( !stricmp( inputWord.str, "Upper" ) )
        miscInfo.fileCase = FileCaseUpper;
      else if( !stricmp( inputWord.str, "Mixed" ) )
        miscInfo.fileCase = FileCaseMixed;
      else if( !stricmp( inputWord.str, "AsIs" ) )
        miscInfo.fileCase = FileCaseAsIs;
      else
        return need( "AsIs/Lower/Mixed/Upper" );
    }
    else if( !stricmp( inputWord.str, "NoLOG" ) )
      rc = processNoLog( cfg, &noLogList, "NoLOG" );
    else if( !stricmp( inputWord.str, "NoSCR" ) )
      rc = processNoLog( cfg, &noScrList, "NoSCR" );
    else if( !stricmp( inputWord.str, "IconMail" ) )
      rc = storeWord( cfg, &miscInfo.iconMail, "IconMail" );
    else if( !stricmp( inputWord.str, "IconXMail" ) )
      rc = storeWord( cfg, &miscInfo.iconXMail, "IconXMail" );
    else if( !stricmp( inputWord.str, "IconRIP" ) )
      rc = storeWord( cfg, &miscInfo.iconRIP, "IconRIP" );
    else if( !stricmp( inputWord.str, "AddressForm" ) )
      rc = processAddressForm( cfg );
    else
      return 3;
  }

  if( rc )
    return rc;

  if( EMSItime )
    miscInfo.EMSItime = long( EMSItime )*1000;

  if( maxIn )
    miscInfo.maxIn = short( maxIn );

  if( maxOut )
    miscInfo.maxOut = short( maxOut );

  if( zmh!=NULL )
  {
    miscInfo.zmh = new WorkTime( zmh );

    if( !miscInfo.zmh->isValid() )
    {
      log( Msg_ErrCfgIllTime, zmh, zmhLine );
      return 2;
    }

    delete zmh;
  }

  if( callTime!=NULL )
  {
    miscInfo.callTime = new WorkTime( callTime );

    if( !miscInfo.callTime->isValid() )
    {
      log( Msg_ErrCfgIllTime, callTime, callLine );
      return 2;
    }

    delete callTime;
  }

  if( callOnlyImmed!=NULL )
  {
    miscInfo.callOnlyImmed = new WorkTime( callOnlyImmed );

    if( !miscInfo.callOnlyImmed->isValid() )
    {
      log( Msg_ErrCfgIllTime, callOnlyImmed, callImmedLine );
      return 2;
    }

    delete callOnlyImmed;
  }

  if( callOnlyCrash!=NULL )
  {
    miscInfo.callOnlyCrash = new WorkTime( callOnlyCrash );

    if( !miscInfo.callOnlyCrash->isValid() )
    {
      log( Msg_ErrCfgIllTime, callOnlyCrash, callCrashLine );
      return 2;
    }

    delete callOnlyCrash;
  }

  if( callOnlyDirect!=NULL )
  {
    miscInfo.callOnlyDirect = new WorkTime( callOnlyDirect );

    if( !miscInfo.callOnlyDirect->isValid() )
    {
      log( Msg_ErrCfgIllTime, callOnlyDirect, callDirectLine );
      return 2;
    }

    delete callOnlyDirect;
  }

  if( callOnlyNormal!=NULL )
  {
    miscInfo.callOnlyNormal = new WorkTime( callOnlyNormal );

    if( !miscInfo.callOnlyNormal->isValid() )
    {
      log( Msg_ErrCfgIllTime, callOnlyNormal, callNormalLine );
      return 2;
    }

    delete callOnlyNormal;
  }

  if( answerTime!=NULL )
  {
    miscInfo.answerTime = new WorkTime( answerTime );

    if( !miscInfo.answerTime->isValid() )
    {
      log( Msg_ErrCfgIllTime, answerTime, answerLine );
      return 2;
    }

    delete answerTime;
  }

  switch( exportConfig )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~ExportConfig );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|ExportConfig );
    break;
  }

  switch( binkDRecover )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~BinkDRecover );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|BinkDRecover );
    break;
  }

  switch( termlog )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~TerminalLog );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|TerminalLog );
    break;
  }

  switch( scanBefore )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~ScanBefore );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|ScanBefore );
    break;
  }

  switch( scanOnLine )
  {
  case 0:
    miscInfo.state3 = SysState3( int(miscInfo.state3)&~ScanOnLine );
    break;

  case 1:
    miscInfo.state3 = SysState3( int(miscInfo.state3)|ScanOnLine );
    break;
  }

  switch( fastScan )
  {
  case 0:
    miscInfo.state3 = SysState3( int(miscInfo.state3)&~FastScan );
    break;

  case 1:
    miscInfo.state3 = SysState3( int(miscInfo.state3)|FastScan );
    break;
  }

  switch( bsyabort )
  {
  case 0:
    miscInfo.state3 = SysState3( int(miscInfo.state3)&~BsyAbort );
    break;

  case 1:
    miscInfo.state3 = SysState3( int(miscInfo.state3)|BsyAbort );
    break;
  }

  switch( zd )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~ZDebug );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|ZDebug );
    break;
  }

  switch( full5D )
  {
  case 1:
    miscInfo.state = SysState( int(miscInfo.state)&~Full5D );
    break;

  case 0:
    miscInfo.state = SysState( int(miscInfo.state)|Full5D );
    break;
  }

  switch( bink )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~BinkLog );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|BinkLog );
    break;
  }

  switch( ed )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~EMSIDebug );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|EMSIDebug );
    break;
  }

  switch( callAfterAnswer )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~CallAfterAnswer );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|CallAfterAnswer );
    break;
  }

  switch( discard )
  {
  case 0:
    miscInfo.state1 = SysState1( int(miscInfo.state1)&~DiscardPKT );
    break;

  case 1:
    miscInfo.state1 = SysState1( int(miscInfo.state1)|DiscardPKT );
    break;
  }

  switch( status )
  {
  case 0:
    miscInfo.state1 = SysState1( int(miscInfo.state1)&~StatusLine );
    break;

  case 1:
    miscInfo.state1 = SysState1( int(miscInfo.state1)|StatusLine );
    break;
  }

  switch( retry )
  {
  case 0:
    miscInfo.state1 = SysState1( int(miscInfo.state1)&~RetryAfterExtern );
    break;

  case 1:
    miscInfo.state1 = SysState1( int(miscInfo.state1)|RetryAfterExtern );
    break;
  }

  switch( hold )
  {
  case 0:
    miscInfo.state1 = SysState1( int(miscInfo.state1)&~HoldOnCall );
    break;

  case 1:
    miscInfo.state1 = SysState1( int(miscInfo.state1)|HoldOnCall );
    break;
  }

  switch( adjust )
  {
  case 0:
    miscInfo.state1 = SysState1( int(miscInfo.state1)&~AdjustTime );
    break;

  case 1:
    miscInfo.state1 = SysState1( int(miscInfo.state1)|AdjustTime );
    break;
  }

  switch( fulllog )
  {
  case 0:
    miscInfo.state2 = SysState2( int(miscInfo.state2)&~FullLog );
    break;

  case 1:
    miscInfo.state2 = SysState2( int(miscInfo.state2)|FullLog );
    break;
  }

  switch( rexx )
  {
  case 0:
    miscInfo.state2 = SysState2( int(miscInfo.state2)&~RexxDebug );
    break;

  case 1:
    miscInfo.state2 = SysState2( int(miscInfo.state2)|RexxDebug );
    break;
  }

  switch( sema )
  {
  case 0:
    miscInfo.state2 = SysState2( int(miscInfo.state2)&~SemaphoreBsy );
    break;

  case 1:
    miscInfo.state2 = SysState2( int(miscInfo.state2)|SemaphoreBsy );
    break;
  }

  switch( taskin )
  {
  case 0:
    miscInfo.state2 = SysState2( int(miscInfo.state2)&~TaskInLog );
    break;

  case 1:
    miscInfo.state2 = SysState2( int(miscInfo.state2)|TaskInLog );
    break;
  }

  switch( hpfs )
  {
  case 0:
    miscInfo.state3 = SysState3( int(miscInfo.state3)&~HPFS );
    break;

  case 1:
    miscInfo.state3 = SysState3( int(miscInfo.state3)|HPFS );
    break;
  }

  switch( chName )
  {
  case 0:
    miscInfo.state3 = SysState3( int(miscInfo.state3)&~CheckName );
    break;

  case 1:
    miscInfo.state3 = SysState3( int(miscInfo.state3)|CheckName );
    break;
  }

  switch( EA )
  {
  case 1:
    miscInfo.state = SysState( int(miscInfo.state)&~NoEA );
    break;

  case 0:
    miscInfo.state = SysState( int(miscInfo.state)|NoEA );
    break;
  }

  if( miscInfo.pipe!=NULL )
    if( *miscInfo.pipe!='\\' )
    {
      char pref[] = "\\PIPE\\Brake!\\",
	   *pipe = new char [ strlen(miscInfo.pipe)+sizeof(pref) ];

      strcpy( pipe, pref );
      strcat( pipe, miscInfo.pipe );
      delete miscInfo.pipe;
      miscInfo.pipe = pipe;
    }

  int len( strlen( miscInfo.semaphore ) );

  if( miscInfo.semaphore[ len-1 ]!='\\' )
  {
    char *c = new char [ len+2 ];

    strcpy( c, miscInfo.semaphore );
    strcat( c, "\\" );

    delete miscInfo.semaphore;
    miscInfo.semaphore = c;
  }

  return 0;
}

int processTimeoutCps( ifstream **cfg, TimeOutList **list, char *who )
{
  long speed( 0 );
  int timeOut( 0 );

  int rc = storeLong( cfg, speed, Msg_ErrCfgSpeed );

  if( rc )
    return rc;

  rc = storeNumber( cfg, timeOut, who );

  if( rc )
    return rc;

  TimeOut *t( new TimeOut( speed, timeOut ) ),
          *found( NULL );

  if( *list==NULL )
    *list = new TimeOutList;
  else
    found = (*list)->Find( *t );

  if( found!=NULL )
    (*list)->Remove( *found );

  (*list)->Add( *t );

  return 0;
}

int processProtocol( ifstream **cfg )
{
  int dirzap( -1 ),
      zedzap( -1 ),
      hydra( -1 ),
      hydraChat( -1 );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "ChatTimeOut" ) )
      rc = storeNumber( cfg, miscInfo.chatTimeout, "ChatTimeOut" );
    else if( !stricmp( inputWord.str, "ZStartBlock" ) )
      rc = storeNumber( cfg, miscInfo.zBlock, "ZStartBlock" );
    else if( !stricmp( inputWord.str, "ZBaseTimeOut" ) )
      rc = storeNumber( cfg, miscInfo.timeOut, "ZBaseTimeOut" );
    else if( !stricmp( inputWord.str, "ZCheckCpsAfter" ) )
      rc = storeLong( cfg, zTransferCpsAfter, "ZCheckCpsAfter" );
    else if( !stricmp( inputWord.str, "ZTimeOut" ) )
      rc = processTimeoutCps( cfg, &timeOuts, Msg_ErrCfgTimeOut );
    else if( !stricmp( inputWord.str, "ZMinOutCps" ) )
      rc = processTimeoutCps( cfg, &zMinOutCps, Msg_ErrCfgOutMinCPS );
    else if( !stricmp( inputWord.str, "ZMinInCps" ) )
      rc = processTimeoutCps( cfg, &zMinInCps, Msg_ErrCfgInMinCPS );
    else if( !stricmp( inputWord.str, "DirZap" ) )
      rc = storeLogic( cfg, dirzap, "DirZap" );
    else if( !stricmp( inputWord.str, "ZedZap" ) )
      rc = storeLogic( cfg, zedzap, "ZedZap" );
    else if( !stricmp( inputWord.str, "Hydra" ) )
      rc = storeLogic( cfg, hydra, "Hydra" );
    else if( !stricmp( inputWord.str, "HydraChat" ) )
      rc = storeLogic( cfg, hydraChat, "HydraChat" );
    else if( !stricmp( inputWord.str, "HBaseTimeOut" ) )
      rc = storeNumber( cfg, hydra_timeout, "HBaseTimeOut" );
    else if( !stricmp( inputWord.str, "HCheckCpsAfter" ) )
      rc = storeLong( cfg, hTransferCpsAfter, "HCheckCpsAfter" );
    else if( !stricmp( inputWord.str, "HTimeOut" ) )
      rc = processTimeoutCps( cfg, &timeOutsHydra, Msg_ErrCfgTimeOut );
    else if( !stricmp( inputWord.str, "HMinOutCps" ) )
      rc = processTimeoutCps( cfg, &hMinOutCps, Msg_ErrCfgOutMinCPS );
    else if( !stricmp( inputWord.str, "HMinInCps" ) )
      rc = processTimeoutCps( cfg, &hMinInCps, Msg_ErrCfgInMinCPS );
    else if( !stricmp( inputWord.str, "HTxWindow" ) )
      rc = storeLong( cfg, hydra_txwindow, "HTxWindow" );
    else if( !stricmp( inputWord.str, "HRxWindow" ) )
      rc = storeLong( cfg, hydra_rxwindow, "HRxWindow" );
    else if( !stricmp( inputWord.str, "HOption" ) )
    {
      if( getWord( cfg ) )
        return need( "HOption" );

      hydra_options = hydraGetFlags( inputWord.str );
    }
    else if( !stricmp( inputWord.str, "NoStartHydraChat" ) )
    {
      if( getWord( cfg ) )
	return 1;

      noHydraChatList->Add( *new NoHydraChat( inputWord.str ) );
    }
    else
      rc = 3;
  }

  switch( hydraChat )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~HydraChat );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|HydraChat );
    break;
  }

  switch( dirzap )
  {
  case 0:
    miscInfo.state2 = SysState2( int(miscInfo.state2)&~DirZapAvail );
    break;

  case 1:
    miscInfo.state2 = SysState2( int(miscInfo.state2)|DirZapAvail );
    break;
  }

  switch( zedzap )
  {
  case 0:
    miscInfo.state2 = SysState2( int(miscInfo.state2)&~ZedZapAvail );
    break;

  case 1:
    miscInfo.state2 = SysState2( int(miscInfo.state2)|ZedZapAvail );
    break;
  }

  switch( hydra )
  {
  case 0:
    miscInfo.state2 = SysState2( int(miscInfo.state2)&~HydraAvail );
    break;

  case 1:
    miscInfo.state2 = SysState2( int(miscInfo.state2)|HydraAvail );
    break;
  }

  return rc;
}

int storeFlowControl( ifstream **cfg, int& flow )
{
  if( getWord( cfg ) )
  {
    strcpy( inputWord.str, "EOF" );
    goto illegalWord;
  }
  else
    if( !stricmp( inputWord.str, "HardWare" ) )
      flow = 1;
    else
      if( !stricmp( inputWord.str, "SoftWare" ) )
	flow = 0;
      else
illegalWord:
	return need( Msg_ErrCfgHardSoft );

  return 0;
}

int processPrefSuf( ifstream **cfg, CallsList **list, char *msg )
{
  char *flag( NULL ),
       *pref( NULL );

  int rc = storeWord( cfg, &flag, Msg_ErrCfgFlag );

  if( rc )
    return rc;

  rc = storeWord( cfg, &pref, msg );

  if( rc )
    return rc;

  if( *list==NULL )
    *list = new CallsList;

  (*list)->AddAtTail( *new Calls( flag, pref ) );

  return 0;
}

int processModem( ifstream **cfg )
{
  int collision( -1 ),
      telebit( -1 ),
      window( -1 ),
      report( -1 ),
      stats( 0 ),
      hard( 0 );

  modemPresent = 1;

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Port" ) )
      rc = storeNumber( cfg, miscInfo.port, "Port" );
    else if( !stricmp( inputWord.str, "MaxRRing" ) )
      rc = storeNumber( cfg, miscInfo.maxRRing, "MaxRRing" );
    else if( !stricmp( inputWord.str, "AutoInit" ) )
      rc = storeNumber( cfg, miscInfo.autoInit, "AutoInit" );
    else if( !stricmp( inputWord.str, "RingCount" ) )
      rc = storeNumber( cfg, miscInfo.ringCount, "RingCount" );
    else if( !stricmp( inputWord.str, "RingWait" ) )
      rc = storeNumber( cfg, miscInfo.ringWait, "RingWait" );
    else if( !stricmp( inputWord.str, "ARQ" ) )
    {
      if( getWord( cfg ) )
	return 1;

      arqList->Add( *new Calls( inputWord.str, NULL ) );
    }
    else if( !stricmp( inputWord.str, "HalfDuplex" ) )
    {
      if( getWord( cfg ) )
	return 1;

      halfDuplexList->Add( *new Calls( inputWord.str, NULL ) );
    }
    else if( !stricmp( inputWord.str, "WaitCall" ) )
      rc = storeNumber( cfg, portInfo.callWait, "WaitCall" );
    else if( !stricmp( inputWord.str, "WaitAnswer" ) )
      rc = storeNumber( cfg, portInfo.answerWait, "WaitAnswer" );
    else if( !stricmp( inputWord.str, "Ignore" ) )
    {
      if( getWord( cfg ) )
        return 1;

      ignoreString->AddAtTail( *new ModemString( inputWord.str ) );
    }
    else if( !stricmp( inputWord.str, "BadCounter" ) )
    {
      if( getWord( cfg ) )
        return 1;

      badCounterString->AddAtTail( *new ModemString( inputWord.str ) );
    }
    else if( !stricmp( inputWord.str, "FailSafe" ) )
      rc = storeLong( cfg, portInfo.failSafe, "FailSafe" );
    else if( !stricmp( inputWord.str, "Delay" ) )
      rc = storeNumber( cfg, portInfo.pause, "Delay" );
    else if( !stricmp( inputWord.str, "Baud" ) )
      rc = storeSpeed( cfg, portInfo.speed, "Baud" );
    else if( !stricmp( inputWord.str, "Init" ) )
      rc = storeWord( cfg, &portInfo.init, "Init" );
    else if( !stricmp( inputWord.str, "Answer" ) )
      rc = storeWord( cfg, &portInfo.answer, "Answer" );
    else if( !stricmp( inputWord.str, "HangUp" ) )
      rc = storeWord( cfg, &portInfo.hangup, "HangUp" );
    else if( !stricmp( inputWord.str, "DialPrefix" ) )
      rc = storeWord( cfg, &portInfo.pref, "DialPrefix" );
    else if( !stricmp( inputWord.str, "DialSuffix" ) )
      rc = storeWord( cfg, &portInfo.suff, "DialSuffix" );
    else if( !stricmp( inputWord.str, "AfterDial" ) )
      rc = storeWord( cfg, &portInfo.after, "AfterDial" );
    else if( !stricmp( inputWord.str, "AbortDial" ) )
      rc = storeWord( cfg, &portInfo.dialterm, "AbortDial" );
    else if( !stricmp( inputWord.str, "OnExit" ) )
      rc = storeWord( cfg, &portInfo.onExit, "OnExit" );
    else if( !stricmp( inputWord.str, "Statistics" ) )
    {
      rc = storeWord( cfg, &modemString.statCmd, "Statistics Command" );

      if( !rc )
        rc = storeWord( cfg, &modemString.statStop, "Statistics Stop" );
    }
    else if( !stricmp( inputWord.str, "Fax" ) )
      rc = storeWord( cfg, &modemString.fax, "Fax" );
    else if( !stricmp( inputWord.str, "Ring" ) )
      rc = storeWord( cfg, &modemString.ring, "Ring" );
    else if( !stricmp( inputWord.str, "Busy" ) )
      rc = storeWord( cfg, &modemString.busy, "Busy" );
    else if( !stricmp( inputWord.str, "Ok" ) )
      rc = storeWord( cfg, &modemString.ok, "Ok" );
    else if( !stricmp( inputWord.str, "Connect" ) )
      rc = storeWord( cfg, &modemString.connect, "Connect" );
    else if( !stricmp( inputWord.str, "CallersID" ) )
      rc = storeWord( cfg, &modemString.callersId, "CallersID" );
    else if( !stricmp( inputWord.str, "NoDialTone" ) )
      rc = storeWord( cfg, &modemString.nodialtone, "NoDialTone" );
    else if( !stricmp( inputWord.str, "RRing" ) )
      rc = storeWord( cfg, &modemString.rring, "RRing" );
    else if( !stricmp( inputWord.str, "NoCarrier" ) )
      rc = storeWord( cfg, &modemString.nocarrier, "NoCarrier" );
    else if( !stricmp( inputWord.str, "Error" ) )
      rc = storeWord( cfg, &modemString.error, "Error" );
    else if( !stricmp( inputWord.str, "Voice" ) )
      rc = storeWord( cfg, &modemString.voice, "Voice" );
    else if( !stricmp( inputWord.str, "NoAnswer" ) )
      rc = storeWord( cfg, &modemString.noanswer, "NoAnswer" );
    else if( !stricmp( inputWord.str, "NoConnect" ) )
      rc = storeWord( cfg, &modemString.noconnect, "NoConnect" );
    else if( !stricmp( inputWord.str, "Monitor" ) )
      rc = storeNumber( cfg, portInfo.carrierDelay, "Monitor" );
    else if( !stricmp( inputWord.str, "FlowControl" ) )
      rc = storeFlowControl( cfg, portInfo.hardwareFlow );
    else if( !stricmp( inputWord.str, "BreakMode" ) )
      rc = storeLogic( cfg, portInfo.breakMode, "BreakMode" );
    else if( !stricmp( inputWord.str, "ModemReport" ) )
      rc = storeLogic( cfg, report, "ModemReport" );
    else if( !stricmp( inputWord.str, "VirtualDCD" ) )
      rc = storeLogic( cfg, portInfo.virtualDCD, "VirtualDCD" );
    else if( !stricmp( inputWord.str, "Window" ) )
      rc = storeLogic( cfg, window, "Window" );
    else if( !stricmp( inputWord.str, "QueueStats" ) )
      rc = storeLogic( cfg, stats, "QueueStats" );
    else if( !stricmp( inputWord.str, "Collision" ) )
      rc = storeLogic( cfg, collision, "Collision" );
    else if( !stricmp( inputWord.str, "Special" ) )
      rc = processPrefSuf( cfg, &preffixList, Msg_ErrCfgPreff );
    else if( !stricmp( inputWord.str, "SpecialSuf" ) )
      rc = processPrefSuf( cfg, &suffixList, Msg_ErrCfgSuff );
    else if( !stricmp( inputWord.str, "WindowAt" ) )
    {
      rc = storeNumber( cfg, miscInfo.wndModemX, Msg_ErrCfgNeedXPos );

      if( !rc )
	rc = storeNumber( cfg, miscInfo.wndModemY, Msg_ErrCfgNeedYPos );
    }
    else if( !stricmp( inputWord.str, "WindowOff" ) )
      rc = storeNumber( cfg, miscInfo.wndModemOff, "WindowOff" );
    else if( !stricmp( inputWord.str, "MinInBound" ) )
      rc = storeLong( cfg, miscInfo.minIn, "MinInBound" );
    else if( !stricmp( inputWord.str, "MinOutBound" ) )
      rc = storeLong( cfg, miscInfo.minOut, "MinOutBound" );
    else if( !stricmp( inputWord.str, "FuckingTelebit" ) )
      rc = storeLogic( cfg, telebit, "FuckingTelebit" );
    else if( !stricmp( inputWord.str, "HardWareRing" ) )
      rc = storeLogic( cfg, hard, "HardWareRing" );
    else
      return 3;
  }

  if( rc )
    return rc;

  extern char startInputQueueMonitor;

  if( stats )
    startInputQueueMonitor = 1;

  switch( window )
  {
  case 1:
    miscInfo.state = SysState( miscInfo.state|ModemWindow );
    break;

  case 0:
    miscInfo.state = SysState( miscInfo.state&(~ModemWindow) );
    break;
  }

  switch( hard )
  {
  case 1:
    miscInfo.state3 = SysState3( miscInfo.state3|HardRing );
    break;

  case 0:
    miscInfo.state3 = SysState3( miscInfo.state3&(~HardRing) );
    break;
  }

  switch( report )
  {
  case 1:
    miscInfo.state = SysState( miscInfo.state|ModemReport );
    break;

  case 0:
    miscInfo.state = SysState( miscInfo.state&(~ModemReport) );
    break;
  }

  switch( collision )
  {
  case 1:
    miscInfo.state = SysState( miscInfo.state|Collision );
    break;

  case 0:
    miscInfo.state = SysState( miscInfo.state&(~Collision) );
    break;
  }

  switch( telebit )
  {
  case 0:
    miscInfo.state1 = SysState1( int(miscInfo.state1)&~Telebit );
    break;

  case 1:
    miscInfo.state1 = SysState1( int(miscInfo.state1)|Telebit );
    break;
  }

  return 0;
}

int processHidden( ifstream **cfg, HiddenList **list,
                   char *dialIni, int waitIni,
                   WorkTime *workIni, int& hLine,
                   char *dialSufIni )
{
  char *phone( NULL ),
       *dial( NULL ),
       *dialSuf( NULL ),
       *wt( NULL );

  int wait( 0 );

  WorkTime *work( NULL );

  if( !hLine  )
    hLine = cfgLine;

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Phone" ) )
      rc = storeWord( cfg, &phone, "Phone" );
    else if( !stricmp( inputWord.str, "Dial" ) )
      rc = storeWord( cfg, &dial, "Dial" );
    else if( !stricmp( inputWord.str, "DialSuf" ) )
      rc = storeWord( cfg, &dialSuf, "DialSuf" );
    else if( !stricmp( inputWord.str, "Carrier" ) )
      rc = storeNumber( cfg, wait, "Carrier" );
    else if( !stricmp( inputWord.str, "Time" ) )
    {
      rc = storeWord( cfg, &wt, "Time" );
      work = new WorkTime( wt );

      if( !work->isValid() )
      {
        log( Msg_ErrCfgIllTime, wt, cfgLine );
	return 2;
      }
    }
    else
      rc = need( Msg_ErrCfgHiddenDirs );
  }

  if( rc )
    return rc;

  if( !wait )
    wait = waitIni;

  if( dial==NULL )
    dial = newStr( dialIni );

  if( dialSuf==NULL )
    dialSuf = newStr( dialSufIni );

  if( wt==NULL )
    work = new WorkTime( workIni );

  if( phone==NULL )
  {
    log( Msg_ErrCfgNoPhone, cfgLine );
    return 2;
  }

  if( *list==NULL )
    *list = new HiddenList;

  (*list)->AddAtTail( *new Hidden( wait, phone, dial, work, dialSuf ) );

  delete phone;
  delete dial;
  delete dialSuf;
  delete wt;
  delete work;

  return 0;
}

int processNode( ifstream **cfg,
                 uShort zone, uShort net, uShort node,
                 uShort  point, char *domain,
                 char *mask )
{
  if( !miscPresent || !modemPresent )
  {
    log( Msg_ErrCfgNoMisc, cfgLine );
    return 2;
  }

  char *phone( NULL ),
       *password( NULL ),
       *normal( NULL ),
       *hold( NULL ),
       *crash( NULL ),
       *immed( NULL ),
       *direct( NULL ),
       *inbound( NULL ),
       *pref( NULL ),
       *suff( NULL ),
       *timeStr( NULL ),
       *flags( NULL ),
       *rexxDial( NULL ),
       *alias( NULL );

  int maxIn( 0 ),
      maxOut( 0 ),
      maxTry( 0 ),
      wt( 0 ),
      cost( 0 ),
      baud( 0 ),
      prio( 0 );

  FIDOAddressList *aka( NULL );
  HiddenList *hidden( NULL );

  Schedule *before( NULL ),
	   *after( NULL ),
	   *doit( NULL );

  Flavour flavour( unknownFlavour );
  WorkTime *workTime( NULL );
  int noDZA( -1 );
  int noZAP( -1 );
  int noHydra( -1 );
  int chain( -1 );

  int hLine( 0 ),
      aLine( 0 );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Mode" ) )
      rc = storeMode( cfg, flavour, "Mode" );
    else if( !stricmp( inputWord.str, "Password" ) )
      rc = storeWord( cfg, &password, "Password" );
    else if( !stricmp( inputWord.str, "Flags" ) )
      rc = storeWord( cfg, &flags, "Flags" );
    else if( !stricmp( inputWord.str, "RexxDial" ) )
      rc = storeWord( cfg, &rexxDial, "RexxDial" );
    else if( !stricmp( inputWord.str, "Dial" ) )
      rc = storeWord( cfg, &pref, "Dial" );
    else if( !stricmp( inputWord.str, "DialSuf" ) )
      rc = storeWord( cfg, &suff, "DialSuf" );
    else if( !stricmp( inputWord.str, "Phone" ) )
      if( mask==NULL )
        rc = storeWord( cfg, &phone, "Phone" );
      else
      {
        log( Msg_CfgErrNodeMaskPhone, cfgLine );
        return 2;
      }
    else if( !stricmp( inputWord.str, "Hidden" ) )
      if( mask==NULL )
        rc = processHidden( cfg, &hidden, pref, wt, workTime, hLine, suff );
      else
      {
        log( Msg_CfgErrNodeMaskHidden, cfgLine );
        return 2;
      }
    else if( !stricmp( inputWord.str, "Priority" ) )
      rc = storeNumber( cfg, prio, "Priority" );
    else if( !stricmp( inputWord.str, "Carrier" ) )
      rc = storeNumber( cfg, wt, "Carrier" );
    else if( !stricmp( inputWord.str, "Cost" ) )
      rc = storeNumber( cfg, cost, "Cost" );
    else if( !stricmp( inputWord.str, "CPS" ) )
      rc = storeNumber( cfg, baud, "CPS" );
    else if( !stricmp( inputWord.str, "MaxInComing" ) )
      rc = storeNumber( cfg, maxIn, "MaxInComing" );
    else if( !stricmp( inputWord.str, "MaxOutGoing" ) )
      rc = storeNumber( cfg, maxOut, "MaxOutGoing" );
    else if( !stricmp( inputWord.str, "MaxTry" ) )
      rc = storeNumber( cfg, maxTry, "MaxTry" );
    else if( !stricmp( inputWord.str, "DirZap" ) )
      rc = storeLogic( cfg, noDZA, "DirZap" );
    else if( !stricmp( inputWord.str, "ZedZap" ) )
      rc = storeLogic( cfg, noZAP, "ZedZap" );
    else if( !stricmp( inputWord.str, "Hydra" ) )
      rc = storeLogic( cfg, noHydra, "Hydra" );
    else if( !stricmp( inputWord.str, "Chain" ) )
      rc = storeLogic( cfg, chain, "Chain" );
    else if( !stricmp( inputWord.str, "Hold" ) )
      rc = storePath( cfg, &hold, "Hold", 0 );
    else if( !stricmp( inputWord.str, "Crash" ) )
      rc = storePath( cfg, &crash, "Crash", 0 );
    else if( !stricmp( inputWord.str, "Normal" ) )
      rc = storePath( cfg, &normal, "Normal", 0 );
    else if( !stricmp( inputWord.str, "Immed" ) )
      rc = storePath( cfg, &immed, "Immed", 0 );
    else if( !stricmp( inputWord.str, "Direct" ) )
      rc = storePath( cfg, &direct, "Direct", 0 );
    else if( !stricmp( inputWord.str, "InBound" ) )
      rc = storePath( cfg, &inbound, "InBound" );
    else if( !stricmp( inputWord.str, "Address" ) )
      rc = processAKA( cfg, &aka, &aLine );
    else if( !stricmp( inputWord.str, "BeforeCall" ) )
      rc = processSch( cfg, &before );
    else if( !stricmp( inputWord.str, "AfterCall" ) )
      rc = processSch( cfg, &after );
    else if( !stricmp( inputWord.str, "Do" ) )
      rc = processSch( cfg, &doit );
    else if( !stricmp( inputWord.str, "Time" ) )
    {
      rc = storeWord( cfg, &timeStr, "Time" );

      if( !rc )
      {
	workTime = new WorkTime( timeStr );

	if( !workTime->isValid() )
	{
	  delete workTime;
          log( Msg_ErrCfgIllTime, timeStr, cfgLine );
	  return 2;
	}
      }
    }
    else if( !stricmp( inputWord.str, "Alias" ) )
      rc = storeWord( cfg, &alias, "Alias" );
    else
      return 3;
  }

  if( rc )
    return rc;

  NodeConfigList *list = mask==NULL ? systemList : maskList;
  FIDOAddress *nodeAddr( new FIDOAddress( zone, net, node, point, domain ) );
  NodeConfig test( nodeAddr );

  test.changeMask( mask );

  NodeConfig *newNode = list->Find( test );

  if( newNode==NULL )
  {
    if( noDZA==-1 )
      noDZA = 0;
    else
      noDZA = !noDZA;

    if( noZAP==-1 )
      noZAP = 0;
    else
      noZAP = !noZAP;

    if( noHydra==-1 )
      noHydra = 0;
    else
      noHydra = !noHydra;

    if( chain==-1 )
      chain = 0;

    list->AddAtTail( *new NodeConfig( nodeAddr, workTime, flavour, phone,
		                      password, normal, hold, crash,
		                      immed, direct,
		                      inbound, pref, wt, flags,
		                      short( maxIn ), short( maxOut ),
			              maxTry, aka, hidden,
			              before, after, doit, noDZA, rexxDial,
			              cost, baud, noHydra, chain, mask, prio,
                                      suff, alias, noZAP ) );
  }
  else
  {
    if( prio )
      newNode->changePriority( prio );

    if( phone!=NULL )
      newNode->changePhone( phone );

    if( password!=NULL )
      newNode->changePassword( password );

    if( normal!=NULL )
      newNode->changeNormalBox( normal );

    if( hold!=NULL )
      newNode->changeHoldBox( hold );

    if( crash!=NULL )
      newNode->changeCrashBox( crash );

    if( immed!=NULL )
      newNode->changeImmedBox( immed );

    if( direct!=NULL )
      newNode->changeDirectBox( direct );

    if( inbound!=NULL )
      newNode->changeInbound( inbound );

    if( pref!=NULL )
      newNode->changeDial( pref );

    if( suff!=NULL )
      newNode->changeDialSuff( suff );

    if( flags!=NULL )
      newNode->changeFlags( flags );

    if( rexxDial!=NULL )
      newNode->changeRexxDial( rexxDial );

    if( wt )
      newNode->changeWait( wt );

    if( baud )
      newNode->changeBaud( baud );

    if( flavour!=unknownFlavour )
      newNode->changeFlavour( flavour );

    if( maxIn )
      newNode->changeMaxIn( short( maxIn ) );

    if( maxOut )
      newNode->changeMaxOut( short( maxOut ) );

    if( maxTry )
      newNode->changeMaxTry( maxTry );

    if( cost )
      newNode->changeCost( cost );

    if( noDZA!=-1 )
      newNode->changeNoDZAenabled( !noDZA );

    if( noZAP!=-1 )
      newNode->changeNoZAPenabled( !noZAP );

    if( noHydra!=-1 )
      newNode->changeNoHydraEnabled( !noHydra );

    if( chain!=-1 )
      newNode->changeChain( chain );

    if( after!=NULL )
      newNode->changeAfter( after );

    if( before!=NULL )
      newNode->changeBefore( before );

    if( doit!=NULL )
      newNode->changeDoit( doit );

    if( aka!=NULL )
      newNode->changeAka( aka, aLine );

    if( hidden!=NULL )
      newNode->changeHidden( hidden, hLine );

    if( workTime!=NULL )
      newNode->changeWork( workTime );

    if( alias!=NULL )
      newNode->changeAlias( alias );
  }

  delete phone;
  delete password;
  delete normal;
  delete hold;
  delete crash;
  delete immed;
  delete direct;
  delete inbound;
  delete pref;
  delete suff;
  delete timeStr;
  delete flags;
  delete rexxDial;
  delete alias;

  delete workTime;

  deleteSched( after );
  deleteSched( before );
  deleteSched( doit );

  delete aka;
  delete hidden;

  return 0;
}

int processArea( ifstream **cfg )
{
  char *pref( NULL ),
       *as( NULL );
  int cost( 0 );

  int rc = storeWord( cfg, &pref, Msg_ErrCfgCodeArea );

  if( rc )
    return rc;

  rc = getWord( cfg );

  if( rc )
    return rc;

  if( !stricmp( inputWord.str, "Cost" ) )
  {
    char *cc( NULL ),
         *e;

    rc = storeWord( cfg, &cc, Msg_ErrCfgCodeCost );

    if( rc )
      return rc;

    cost = int( strtol( inputWord.str, &e, 10 ) );

    if( *e )
      return need( Msg_ErrCfgCodeCost );

    delete cc;
  }
  else
    parserNoRead = 1;

  rc = storeWord( cfg, &as, Msg_ErrCfgCodeTranslated );

  char *slash = strchr( as, '/' );

  if( slash!=NULL )
    *(slash++) = 0;

  translationTable->append( pref, as, slash, cost );
  delete pref;
  delete as;
  return 0;
}

int processTranslate( ifstream **cfg )
{
  if( translationTable==NULL )
    translationTable = new TranslationTable;

  char *code( NULL ),
       *inter( NULL ),
       *loc( NULL );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Local" ) )
      rc = storeWord( cfg, &loc, Msg_ErrCfgCallLoc );
    else if( !stricmp( inputWord.str, "Country" ) )
      rc = storeWord( cfg, &code, Msg_ErrCfgCodeCountry );
    else if( !stricmp( inputWord.str, "International" ) )
      rc = storeWord( cfg, &inter, Msg_ErrCfgCallInter );
    else if( !stricmp( inputWord.str, "Area" ) )
      rc = processArea( cfg );
    else
      return 3;

  }

  if( rc )
    return rc;

  translationTable->setCountry( code, inter, loc );

  return 0;
}

int processXLATPair( ifstream **cfg, short *tab )
{
  int orig, as, rc( storeNumber( cfg, orig, "ਣ " ) );

  if( rc )
    return rc;

  if( orig>255 )
    log( Msg_ErrCfgCharOrig, orig, cfgLine );

  rc = storeNumber( cfg, as, Msg_ErrCfgCharWho );
  if( rc )
    return rc;

  if( as>255 )
    log( Msg_ErrCfgCharTrans, as, cfgLine );

  if( as<256 && orig<256 )
    tab[ orig ] = short(as);

  return 0;
}

class InputWordsList : public Collection		//Pair
{
public:
  InputWordsList() : Collection() {}

  Pair* Find( Pair& item, LockMode lock=LockModeNeed ) const
    { return (Pair*)Collection::Find( item, lock ); }
    
  Pair& operator [] ( const int pos ) const
    { return *(Pair *)&(Collection::operator[]( pos )); }
};

class InputWordsIterator : public CollectionIterator	//Pair
{
public:
  InputWordsIterator( InputWordsList& ini ) :
    CollectionIterator( *(Collection *)&ini )
  {}
  
  Pair& operator () () const
    { return *(Pair *)&(CollectionIterator::operator()()); }
};

int processXLATWord( ifstream **cfg, InputWordsList **inputWords,
                     Pair::DoType type )
{
  char *who( NULL ),
       *as( NULL );

  int rc = storeWord( cfg, &who, Msg_ErrCfgWordOrig );
  if( rc )
    return rc;

  rc = storeWord( cfg, &as, Msg_ErrCfgWordTrans );
  if( rc )
    return rc;

  if( *inputWords==NULL )
    *inputWords = new InputWordsList;

  (*inputWords)->AddAtTail( *new Pair( who, as, type ) );

  return 0;
}

int processXLAT( ifstream **cfg )
{
  InputWordsList *inputWords( NULL );
  char *xlatName( NULL );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "In" ) )
      rc = processXLATPair( cfg, terminalInXLAT );
    else if( !stricmp( inputWord.str, "Out" ) )
      rc = processXLATPair( cfg, terminalOutXLAT );
    else if( !stricmp( inputWord.str, "Answer" ) )
      rc = processXLATWord( cfg, &inputWords, Pair::doModem );
    else if( !stricmp( inputWord.str, "Display" ) )
      rc = processXLATWord( cfg, &inputWords, Pair::doOut );
    else if( !stricmp( inputWord.str, "XLATfile" ) )
      rc = storeWord( cfg, &xlatName, "XLATfile" );
    else
      rc = 3;
  }

  if( rc )
    return rc;

  if( xlatName!=NULL )
  {
    ifstream xlat( xlatName, ios::in|ios::binary|ios::nocreate );

    cfgFileList.registerFile( xlatName );

    if( !xlat.good() )
      log( Msg_ErrCfgOpenXLAT, xlatName, strError() );
    else
    {
      xlat.read( (char *)terminalInXLAT, sizeof(terminalInXLAT)/sizeof(terminalInXLAT[0]) );

      int c1( xlat.gcount() );

      xlat.read( (char *)terminalOutXLAT, sizeof(terminalOutXLAT)/sizeof(terminalOutXLAT[0]) );

      if( c1!=sizeof(terminalInXLAT)/sizeof(terminalInXLAT[0]) ||
	  xlat.gcount()!=sizeof(terminalOutXLAT)/sizeof(terminalOutXLAT[0]) )
      {
        log( Msg_ErrCfgReadXLAT, xlatName, strError() );
	memset( terminalOutXLAT, -1, sizeof(terminalOutXLAT) );
	memset( terminalInXLAT, -1, sizeof(terminalInXLAT) );
      }
      else
      {
	char *p( (char*)terminalInXLAT );

	for( int i( sizeof(terminalInXLAT)/sizeof(terminalInXLAT[0]) ); i>=0; i-- )
	{
	  p[i*2]=p[i];
	  p[i*2+1]=0;
	}

	p = (char*)terminalOutXLAT;

	for( i=sizeof(terminalOutXLAT)/sizeof(terminalOutXLAT[0]); i>=0; i-- )
	{
	  p[i*2]=p[i];
	  p[i*2+1]=0;
	}
      }
    }

    delete xlatName;
  }

  if( inputWords!=NULL )
  {
    ansiHow = ansiHowInternal + inputWords->GetItemsInContainer();
    ansiTab = new Pair [ ansiHow ];

    for( rc=0; rc<ansiHowInternal; rc++ )
      ansiTab[rc] = ansiTabInternal[rc];

    rc=ansiHow-1;

    for( InputWordsIterator lst( *inputWords ); lst; rc--, lst++ )
      ansiTab[ rc ] = lst();

    delete inputWords;
  }

  return 0;
}

int processTableEntry( ifstream **cfg, ProcessList *table )
{
  if( getWord( cfg ) )
    return 1;

  ProcessItemType isA( unknownProcessItemType );

  if( !stricmp( inputWord.str, "File" ) )
    isA = fileProcessItemType;
  else if( !stricmp( inputWord.str, "Mask" ) )
    isA = maskProcessItemType;
  else if( !stricmp( inputWord.str, "NetMail" ) )
    isA = netMailProcessItemType;
  else if( !stricmp( inputWord.str, "ArcMail" ) )
    isA = arcMailProcessItemType;
  else if( !stricmp( inputWord.str, "Files" ) )
    isA = filesProcessItemType;
  else return need( Msg_ErrCfgNeedProcType );

  char *mask( NULL );
  int rc, line;

  if( isA==fileProcessItemType || isA==maskProcessItemType )
    if( rc = storeWord( cfg, &mask, Msg_ErrCfgNeedProcTempl ), rc )
      return rc;
    else
      line = cfgLine;

  Schedule *sh;

  rc = processSch( cfg, &sh );

  if( rc )
    return rc;

  SessionItem *item( new SessionItem( isA, mask, sh ) );

  if( !item->isOK() )
  {
    log( Msg_ErrCfgIllRegExp, mask, line, item->getRegErrorMsg() );
    return 2;
  }

  table->AddAtTail( *item );
  delete mask;
  return 0;
}

int storeOnceMulti( ifstream **cfg, int& flow )
{
  if( getWord( cfg ) )
  {
    strcpy( inputWord.str, "EOF" );
    goto illegalWord;
  }
  else
    if( !stricmp( inputWord.str, "Once" ) )
      flow = 1;
    else
      if( !stricmp( inputWord.str, "Multi" ) )
	flow = 0;
      else
illegalWord:
	return need( Msg_ErrCfgNeedProcMulti );

  return 0;
}

int processExitFlag( ifstream **cfg )
{
  char *nm( NULL ),
       number[32];

  int rc = storeWord( cfg, &nm, "ExitFlag" );

  if( rc )
    return rc;

  char drive[_MAX_DRIVE],
       dir[_MAX_DIR],
       name[_MAX_FNAME],
       ext[_MAX_EXT];

  _splitpath( nm, drive, dir, name, ext );

  if( *ext )
  {
    log( Msg_ErrCfgIllFlagExt, cfgLine );
    return 2;
  }

  if( !*dir && !*drive )
    if( miscInfo.flags==NULL )
    {
      log( Msg_ErrCfgIllFlagLoc, cfgLine );
      return 2;
    }
    else
    {
      char *name = new char [ strlen(miscInfo.flags)+strlen(nm)+1 ];
      strcpy( name, miscInfo.flags );
      strcat( name, nm );
      delete nm;
      nm = name;
    }

  sprintf( number, ".%d", miscInfo.taskNumber );
  delete miscInfo.exitFlag;
  miscInfo.exitFlag = new char [ strlen(nm)+strlen(number)+1 ];
  strcpy( miscInfo.exitFlag, nm );
  strcat( miscInfo.exitFlag, number );
  delete nm;

  return 0;
}

int processProcess( ifstream **cfg )
{
  int rr( 0 );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "ExitFlag" ) )
      rc = processExitFlag( cfg );
    else if( !stricmp( inputWord.str, "Online" ) )
      rc = processTableEntry( cfg, onlineTable );
    else if( !stricmp( inputWord.str, "After" ) )
      rc = processTableEntry( cfg, processTable );
    else if( !stricmp( inputWord.str, "Immediate" ) )
      rc = processTableEntry( cfg, immediateTable );
    else if( !stricmp( inputWord.str, "Process" ) )
      rc = storeOnceMulti( cfg, rr );
    else if( !stricmp( inputWord.str, "Flag" ) )
      rc = processBothFlag( cfg, &flagList, &flagCount );
    else if( !stricmp( inputWord.str, "NoFlag" ) )
      rc = processBothFlag( cfg, &noFlagList, &noFlagCount );
    else
      return 3;

  }

  if( rr )
    miscInfo.state = SysState( miscInfo.state|ProcessOnce );
  else
    miscInfo.state = SysState( miscInfo.state&(~ProcessOnce) );

  return rc;
}

int processFreqAlias( ifstream **cfg )
{
  char *alias( NULL );
  char *path( NULL );

  int rc( storeWord( cfg, &alias, Msg_ErrCfgNeedAlias ) );

  if( rc )
    return rc;

  rc = storeWord( cfg, &path, Msg_ErrCfgNeedPath );

  if( rc )
    return rc;

  if( freqAlias==NULL )
    freqAlias = new AliasList;

  freqAlias->AddAtTail( *new Alias( path, alias ) );

  delete path;
  delete alias;

  return 0;
}

int processFreqPath( ifstream **cfg, int scan, char *msg )
{
  char *path( NULL );

  int rc( storeWord( cfg, &path, msg ) );

  if( rc )
    return rc;

  if( freqList==NULL )
    freqList = new AliasList;

  freqList->AddAtTail( *new Alias( path, NULL, scan ) );

  delete path;

  return 0;
}

int processFreq( ifstream **cfg )
{
  int check( -1 );
  char *freqTime( NULL );
  int freqTimeLine( 0 );

  long sessionFreqTime = -1,
       sessionFreqSize = -1,
       dayFreqTime = -1,
       dayFreqSize = -1,
       weekFreqTime = -1,
       weekFreqSize = -1,
       monFreqTime = -1,
       monFreqSize = -1;

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "FREQ" ) )
    {
      if( getWord( cfg ) )
	return 1;

      if( strIsYes( inputWord.str ) )
	miscInfo.freqMode = AllFreq;
      else if( strIsNo( inputWord.str ) )
	miscInfo.freqMode = NoFreq;
      else if( !stricmp( inputWord.str, "Listed" ) )
	miscInfo.freqMode = ListedFreq;
      else if( !stricmp( inputWord.str, "Protected" ) )
	miscInfo.freqMode = ProtectedFreq;
      else
	rc = need( Msg_ErrCfgFreqMode );
    }
    else if( !stricmp( inputWord.str, "SessionTime" ) )
      rc = storeLong( cfg, sessionFreqTime, "SessionTime" );
    else if( !stricmp( inputWord.str, "SessionSize" ) )
      rc = storeLong( cfg, sessionFreqSize, "SessionSize" );
    else if( !stricmp( inputWord.str, "DayTime" ) )
      rc = storeLong( cfg, dayFreqTime, "DayTime" );
    else if( !stricmp( inputWord.str, "DaySize" ) )
      rc = storeLong( cfg, dayFreqSize, "DaySize" );
    else if( !stricmp( inputWord.str, "WeekTime" ) )
      rc = storeLong( cfg, weekFreqTime, "WeekTime" );
    else if( !stricmp( inputWord.str, "WeekSize" ) )
      rc = storeLong( cfg, weekFreqSize, "WeekSize" );
    else if( !stricmp( inputWord.str, "MonthTime" ) )
      rc = storeLong( cfg, monFreqTime, "MonthTime" );
    else if( !stricmp( inputWord.str, "MonthSize" ) )
      rc = storeLong( cfg, monFreqSize, "MonthSize" );
    else if( !stricmp( inputWord.str, "Time" ) )
    {
      freqTimeLine = cfgLine;
      rc = storeWord( cfg, &freqTime, "Time" );
    }
    else if( !stricmp( inputWord.str, "Path" ) )
      rc = processFreqPath( cfg, 0, Msg_FreqPathToFiles );
    else if( !stricmp( inputWord.str, "Tree" ) )
      rc = processFreqPath( cfg, 1, Msg_FreqTreeOfFiles );
    else if( !stricmp( inputWord.str, "Alias" ) )
      rc = processFreqAlias( cfg );
    else if( !stricmp( inputWord.str, "BaseFile" ) )
      rc = storeWord( cfg, &miscInfo.freqBase, "BaseFile" );
    else if( !stricmp( inputWord.str, "MinBaud" ) )
      rc = storeLong( cfg, miscInfo.minFreqBaud, "MinBaud" );
    else if( !stricmp( inputWord.str, "Policy" ) )
      rc = storeWord( cfg, &miscInfo.freqPolicy, "Policy" );
    else if( !stricmp( inputWord.str, "CheckREQ" ) )
      rc = storeLogic( cfg, check, "CheckREQ" );
    else if( !stricmp( inputWord.str, "Index" ) )
      rc = storeWord( cfg, &miscInfo.freqIdx, "Index" );
    else
      return 3;


  }

  if( rc )
    return rc;

  if( sessionFreqTime!=-1 )
    miscInfo.sessionFreqTime = (unsigned long)sessionFreqTime;

  if( sessionFreqSize!=-1 )
    miscInfo.sessionFreqSize = (unsigned long)sessionFreqSize;

  if( dayFreqTime!=-1 )
    miscInfo.dayFreqTime     = (unsigned long)dayFreqTime;

  if( dayFreqSize!=-1 )
    miscInfo.dayFreqSize     = (unsigned long)dayFreqSize;

  if( weekFreqTime!=-1 )
    miscInfo.weekFreqTime    = (unsigned long)weekFreqTime;

  if( weekFreqSize!=-1 )
    miscInfo.weekFreqSize    = (unsigned long)weekFreqSize;

  if( monFreqTime!=-1 )
    miscInfo.monFreqTime     = (unsigned long)monFreqTime;

  if( monFreqSize!=-1 )
    miscInfo.monFreqSize     = (unsigned long)monFreqSize;

  if( freqTime!=NULL )
  {
    miscInfo.freqTime = new WorkTime( freqTime );

    if( !miscInfo.freqTime->isValid() )
    {
      log( Msg_ErrCfgIllTime, freqTime, freqTimeLine );
      return 2;
    }

    delete freqTime;
  }

  switch( check )
  {
  case 0:
    miscInfo.state3 = SysState3( int(miscInfo.state3)&~CheckREQ );
    break;

  case 1:
    miscInfo.state3 = SysState3( int(miscInfo.state3)|CheckREQ );
    break;
  }

  return 0;
}

int processSplit( ifstream **cfg )
{
  int del( -1 ),
      lspl( -1 ),
      notfound( -1 ),
      fromlo( -1 );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Circle" ) )
      rc = storeNumber( cfg, miscInfo.splitCircle, "Circle" );
    else if( !stricmp( inputWord.str, "MinEchoSize" ) )
      rc = storeLong( cfg, miscInfo.splitSize, "MinEchoSize" );
    else if( !stricmp( inputWord.str, "MinPKTSize" ) )
      rc = storeLong( cfg, miscInfo.splitPktSize, "MinPKTSize" );
    else if( !stricmp( inputWord.str, "MaxFileSize" ) )
      rc = storeLong( cfg, miscInfo.splitFileSize, "MaxFileSize" );
    else if( !stricmp( inputWord.str, "Delete" ) )
      rc = storeLogic( cfg, del, "Delete" );
    else if( !stricmp( inputWord.str, "PKTFromLO" ) )
      rc = storeLogic( cfg, fromlo, "PKTFromLO" );
    else if( !stricmp( inputWord.str, "LongNames" ) )
      rc = storeLogic( cfg, lspl, "LongNames" );
    else if( !stricmp( inputWord.str, "RemoveNotFound" ) )
      rc = storeLogic( cfg, notfound, "RemoveNotFound" );
    else
      return 3;

  }

  if( rc )
    return rc;

  switch( del )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~SplitDelete );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|SplitDelete );
    break;
  }

  switch( fromlo )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~SplitPKTFromLO );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|SplitPKTFromLO );
    break;
  }

  switch( lspl )
  {
  case 0:
    miscInfo.state2 = SysState2( int(miscInfo.state2)&~SplitterLong );
    break;

  case 1:
    miscInfo.state2 = SysState2( int(miscInfo.state2)|SplitterLong );
    break;
  }

  switch( notfound )
  {
  case 0:
    miscInfo.state2 = SysState2( int(miscInfo.state2)&~SplitDelNoFound );
    break;

  case 1:
    miscInfo.state2 = SysState2( int(miscInfo.state2)|SplitDelNoFound );
    break;
  }

  return 0;
}

int storeVideoMode( ifstream **cfg )
{
  int rc = storeNumber( cfg, colorSet.x, Msg_ErrCfgNeedVideoX );

  if( rc )
    return rc;

  rc = storeNumber( cfg, colorSet.y, Msg_ErrCfgNeedVideoY );

  return rc;
}

int storeVideoAttr( ifstream **cfg, unsigned char& att,
					 char *who )
{
  if( getWord( cfg ) )
  {
    unexpMsg( who );
    return 1;
  }

  char *end;
  unsigned long r = strtoul( inputWord.str, &end, 0 );

  if( *end )
    return need( Msg_ErrCfgNeedNumb );

  if( r>255 || !r )
    return need( Msg_ErrCfgNeedNumb255 );

  att = (unsigned char)r;

  return 0;
}

int processFrame( ifstream **cfg )
{
  int rc = storeNumber( cfg, miscInfo.x, Msg_ErrCfgNeedXPos );

  if( rc )
    return rc;

  rc = storeNumber( cfg, miscInfo.y, Msg_ErrCfgNeedYPos );

  if( rc )
    return rc;

  rc = storeNumber( cfg, miscInfo.cx, Msg_ErrCfgNeedWidth );

  if( rc )
    return rc;

  rc = storeNumber( cfg, miscInfo.cy, Msg_ErrCfgNeedHeight );

  if( rc )
    return rc;

  return rc;
}

extern char *switchListTitle;

int processTitle( ifstream **cfg )
{
  if( getWord( cfg ) )
    return 1;

  switchListTitle = newStr( inputWord.str );

  return 0;
}

int processVideo( ifstream **cfg )
{
  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "VideoMode" ) )
      rc = storeVideoMode( cfg );
    else if( !stricmp( inputWord.str, "Frame" ) )
      rc = processFrame( cfg );
    else if( !stricmp( inputWord.str, "Title" ) )
      rc = processTitle( cfg );
    else if( !stricmp( inputWord.str, "Status" ) )
      rc = storeVideoAttr( cfg, colorSet.status, "Status" );
    else if( !stricmp( inputWord.str, "StatusLampOn" ) )
      rc = storeVideoAttr( cfg, colorSet.statusLampOn, "StatusLampOn" );
    else if( !stricmp( inputWord.str, "StatusLampOff" ) )
      rc = storeVideoAttr( cfg, colorSet.statusLampOff, "StatusLampOff" );
    else if( !stricmp( inputWord.str, "Border" ) )
      rc = storeVideoAttr( cfg, colorSet.border, "Border" );
    else if( !stricmp( inputWord.str, "Action" ) )
      rc = storeVideoAttr( cfg, colorSet.action, "Action" );
    else if( !stricmp( inputWord.str, "ActionBold" ) )
      rc = storeVideoAttr( cfg, colorSet.actionBold, "ActionBold" );
    else if( !stricmp( inputWord.str, "Abort" ) )
      rc = storeVideoAttr( cfg, colorSet.abort, "Abort" );
    else if( !stricmp( inputWord.str, "PopUp" ) )
      rc = storeVideoAttr( cfg, colorSet.popUp, "PopUp" );
    else if( !stricmp( inputWord.str, "PopUpItem" ) )
      rc = storeVideoAttr( cfg, colorSet.popUpItem, "PopUpItem" );
    else if( !stricmp( inputWord.str, "PopUpBold" ) )
      rc = storeVideoAttr( cfg, colorSet.popUpBold, "PopUpBold" );
    else if( !stricmp( inputWord.str, "Terminal" ) )
      rc = storeVideoAttr( cfg, colorSet.terminal, "Terminal" );
    else if( !stricmp( inputWord.str, "Help" ) )
      rc = storeVideoAttr( cfg, colorSet.help, "Help" );
    else if( !stricmp( inputWord.str, "HelpBold" ) )
      rc = storeVideoAttr( cfg, colorSet.helpBold, "HelpBold" );
    else if( !stricmp( inputWord.str, "QueueCrash" ) )
      rc = storeVideoAttr( cfg, colorSet.queueCrash, "QueueCrash" );
    else if( !stricmp( inputWord.str, "QueueNormal" ) )
      rc = storeVideoAttr( cfg, colorSet.queueNormal, "QueueNormal" );
    else if( !stricmp( inputWord.str, "QueueHold" ) )
      rc = storeVideoAttr( cfg, colorSet.queueHold, "QueueHold" );
    else if( !stricmp( inputWord.str, "QueueImmed" ) )
      rc = storeVideoAttr( cfg, colorSet.queueImmed, "QueueImmed" );
    else if( !stricmp( inputWord.str, "QueueDirect" ) )
      rc = storeVideoAttr( cfg, colorSet.queueDirect, "QueueDirect" );
    else if( !stricmp( inputWord.str, "QueuePtr" ) )
      rc = storeVideoAttr( cfg, colorSet.queuePtr, "QueuePtr" );
    else if( !stricmp( inputWord.str, "ModemTo" ) )
      rc = storeVideoAttr( cfg, colorSet.modemTo, "ModemTo" );
    else if( !stricmp( inputWord.str, "ModemFrom" ) )
      rc = storeVideoAttr( cfg, colorSet.modemFrom, "ModemFrom" );
    else if( !stricmp( inputWord.str, "ModemCtlTo" ) )
      rc = storeVideoAttr( cfg, colorSet.modemCtlTo, "ModemCtlTo" );
    else if( !stricmp( inputWord.str, "ModemCtlFrom" ) )
      rc = storeVideoAttr( cfg, colorSet.modemCtlFrom, "ModemCtlFrom" );
    else if( !stricmp( inputWord.str, "Color" ) )
      rc = storeLogic( cfg, colorSet.isColor, "Color" );
    else if( !stricmp( inputWord.str, "ScreenSaverModule" ) )
      rc = storeWord( cfg, &miscInfo.saverModule, "ScreenSaverModule" );
    else if( !stricmp( inputWord.str, "ScreenSaver" ) )
      rc = storeNumber( cfg, miscInfo.saverInterval, "ScreenSaver" );
    else
      return need( Msg_ErrCfgVideoDirs );

  }

  return rc;
}

int processRemote( ifstream **cfg )
{
  int rc = storeWord( cfg, &miscInfo.remoteName, "RemoteName" );

  if( rc )
    return rc;

  rc = storeLong( cfg, miscInfo.remotePassword, "RemotePassword" );

  return rc;
}

int processExternal( ifstream **cfg )
{
  int processExternalItem( ifstream **, Pair**, int& );
  int processBBSRun( ifstream **cfg );

  int bbs( -1 );
  int autoBbs( -1 );
  char *bbsTime( NULL );
  int bbsTimeLine( 0 );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "BBS" ) )
      rc = storeLogic( cfg, bbs, "BBS" );
    else if( !stricmp( inputWord.str, "AutoBBS" ) )
      rc = storeLogic( cfg, autoBbs, "AutoBBS" );
    else if( !stricmp( inputWord.str, "Time" ) )
    {
      bbsTimeLine = cfgLine;
      rc = storeWord( cfg, &bbsTime, "Time" );
    }
    else if( !stricmp( inputWord.str, "MinBaud" ) )
      rc = storeLong( cfg, miscInfo.minBBSBaud, "MinBaud" );
    else if( !stricmp( inputWord.str, "Connect" ) )
      rc = processExternalItem( cfg, &connectList, connectCount );
    else if( !stricmp( inputWord.str, "HandShake" ) )
      rc = processExternalItem( cfg, &handShakeList, handShakeCount );
    else if( !stricmp( inputWord.str, "RunBBS" ) )
      rc = processBBSRun( cfg );
    else if( !stricmp( inputWord.str, "Remote" ) )
      rc = processRemote( cfg );
    else if( !stricmp( inputWord.str, "TwiceESC" ) )
      rc = storeWord( cfg, &bbsInfo.twiceEsc, "TwiceESC" );
    else if( !stricmp( inputWord.str, "MailOnly" ) )
      rc = storeWord( cfg, &bbsInfo.mailOnly, "MailOnly" );
    else if( !stricmp( inputWord.str, "SlowModem" ) )
      rc = storeWord( cfg, &bbsInfo.slowModem, "SlowModem" );
    else if( !stricmp( inputWord.str, "MailNow" ) )
      rc = storeWord( cfg, &bbsInfo.mailNow, "MailNow" );
    else if( !stricmp( inputWord.str, "Loading" ) )
      rc = storeWord( cfg, &bbsInfo.loading, "Loading" );
    else if( !stricmp( inputWord.str, "GoodBye" ) )
      rc = storeWord( cfg, &bbsInfo.goodBye, "GoodBye" );
    else
      return 3;

  }

  if( rc )
    return rc;

  if( bbsTime!=NULL )
  {
    miscInfo.BBSTime = new WorkTime( bbsTime );

    if( !miscInfo.BBSTime->isValid() )
    {
      log( Msg_ErrCfgIllTime, bbsTime, bbsTimeLine );
      return 2;
    }

    delete bbsTime;
  }

  switch( bbs )
  {
  case 0:
    miscInfo.state1 = SysState1( int(miscInfo.state1)&~BBSenable );
    break;

  case 1:
    miscInfo.state1 = SysState1( int(miscInfo.state1)|BBSenable );
    break;
  }

  switch( autoBbs )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~AutoBBS );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|AutoBBS );
    break;
  }

  return 0;
}

int processCalls( ifstream **cfg )
{
  EnableList **list( &enabledList );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    EnableItemType type( EnableItemUnknown );

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Allowed" ) )
    {
      list = &enabledList;
      continue;
    }
    else if( !stricmp( inputWord.str, "Except" ) )
    {
      list = &disabledList;
      continue;
    }
    else if( !stricmp( inputWord.str, "Address" ) )
      type = EnableItemAddress;
    else if( !stricmp( inputWord.str, "Flag" ) )
      type = EnableItemFlag;
    else
      return need( Msg_ErrCfgAddrFlag );

    if( getWord( cfg ) )
      return 1;

    EnableItem *item = new EnableItem( inputWord.str, type );

    if( !item->isCorrect() )
    {
      log( Msg_ErrCfgIllRegExp, inputWord.str, cfgLine, item->getRegErrorMsg() );
      delete item;
      return 2;
    }

    if( *list==NULL )
      *list = new EnableList;

    (*list)->AddAtTail( *item );
  }

  return rc;
}

int processAccept( ifstream **cfg )
{
  EnableList **list( &acceptedList );
  int abortSess( -1 );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    EnableItemType type( EnableItemUnknown );

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Abort" ) )
    {
      rc = storeLogic( cfg, abortSess, "Abort" );
      continue;
    }
    else if( !stricmp( inputWord.str, "Allowed" ) )
    {
      list = &acceptedList;
      continue;
    }
    else if( !stricmp( inputWord.str, "Except" ) )
    {
      list = &unAcceptedList;
      continue;
    }
    else if( !stricmp( inputWord.str, "Address" ) )
      type = EnableItemAddress;
    else
      return need( Msg_ErrCfgAddr );

    if( getWord( cfg ) )
      return 1;

    EnableItem *item = new EnableItem( inputWord.str, type );

    if( !item->isCorrect() )
    {
      log( Msg_ErrCfgIllRegExp, inputWord.str, cfgLine, item->getRegErrorMsg() );
      delete item;
      return 2;
    }

    if( *list==NULL )
      *list = new EnableList;

    (*list)->AddAtTail( *item );
  }

  switch( abortSess )
  {
  case 0:
    miscInfo.state1 = SysState1( int(miscInfo.state1)&~NotAcceptAbort );
    break;

  case 1:
    miscInfo.state1 = SysState1( int(miscInfo.state1)|NotAcceptAbort );
    break;
  }

  return rc;
}

int processHook( ifstream **cfg, char *hook )
{
  if( getWord( cfg ) )
    return need( Msg_ErrCfgHook );

  if( rexxHookList==NULL )
    rexxHookList = new HookList;

  RexxHook *newHook = new RexxHook( hook, inputWord.str ),
           *test = rexxHookList->Find( *newHook );

  if( test!=NULL )
    rexxHookList->Remove( *test );

  rexxHookList->Add( *newHook );
  return 0;
}

int processRexx( ifstream **cfg )
{
  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "BeforeScan" ) )
      rc = processHook( cfg, "BeforeScan" );
    else if( !stricmp( inputWord.str, "AfterScan" ) )
      rc = processHook( cfg, "AfterScan" );
    else if( !stricmp( inputWord.str, "BeforeScanNode" ) )
      rc = processHook( cfg, "BeforeScanNode" );
    else if( !stricmp( inputWord.str, "AfterScanNode" ) )
      rc = processHook( cfg, "AfterScanNode" );
    else if( !stricmp( inputWord.str, "Translate" ) )
      rc = processHook( cfg, "Translate" );
    else if( !stricmp( inputWord.str, "CallAvailable" ) )
      rc = processHook( cfg, "CallAvailable" );
    else if( !stricmp( inputWord.str, "AnswerAvailable" ) )
      rc = processHook( cfg, "AnswerAvailable" );
    else if( !stricmp( inputWord.str, "CallableNode" ) )
      rc = processHook( cfg, "CallableNode" );
    else if( !stricmp( inputWord.str, "AcceptableNode" ) )
      rc = processHook( cfg, "AcceptableNode" );
    else if( !stricmp( inputWord.str, "Terminal" ) )
      rc = processHook( cfg, "Terminal" );
    else if( !stricmp( inputWord.str, "Mailer" ) )
      rc = processHook( cfg, "Mailer" );
    else if( !stricmp( inputWord.str, "RecoverFile" ) )
      rc = processHook( cfg, "RecoverFile" );
    else if( !stricmp( inputWord.str, "BeforeSession" ) )
      rc = processHook( cfg, "BeforeSession" );
    else if( !stricmp( inputWord.str, "AfterSession" ) )
      rc = processHook( cfg, "AfterSession" );
    else if( !stricmp( inputWord.str, "Statistics" ) )
      rc = processHook( cfg, "Statistics" );
    else if( !stricmp( inputWord.str, "SendThisFile" ) )
      rc = processHook( cfg, "SendThisFile" );
    else if( !stricmp( inputWord.str, "FREQ" ) )
      rc = processHook( cfg, "FREQ" );
    else if( !stricmp( inputWord.str, "Answer" ) )
      rc = processHook( cfg, "Answer" );
    else if( !stricmp( inputWord.str, "InitModem" ) )
      rc = processHook( cfg, "InitModem" );
    else if( !stricmp( inputWord.str, "BadModem" ) )
      rc = processHook( cfg, "BadModem" );
    else if( !stricmp( inputWord.str, "FromEMSI_DAT" ) )
      rc = processHook( cfg, "FromEMSI_DAT" );
    else if( !stricmp( inputWord.str, "ToEMSI_DAT" ) )
      rc = processHook( cfg, "ToEMSI_DAT" );
    else if( !stricmp( inputWord.str, "ModemString" ) )
      rc = processHook( cfg, "ModemString" );
    else if( !stricmp( inputWord.str, "FileSent" ) )
      rc = processHook( cfg, "FileSent" );
    else if( !stricmp( inputWord.str, "LogFilter" ) )
      rc = processHook( cfg, "LogFilter" );
    else
      return 3;
  }

  return rc;
}

int processDomain( ifstream **cfg )
{
  char *name,
       *outb = NULL,
       *inb = NULL,
       *prot = NULL;
  int mainZone = 0;
  short *list = NULL;
  int size=0;
  NDLList *ndl = new NDLList;
  int outLine = cfgLine;

  if( getWord( cfg ) )
    return 1;

  name = newStr( inputWord.str );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

no_get_word:

    NDLItemType type;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "OutBound" ) )
    {
      rc = storePath( cfg, &outb, "OutBound", 0 );
      outLine = cfgLine;
    }
    else if( !stricmp( inputWord.str, "InBound" ) )
      rc = storePath( cfg, &inb, "InBound", 0 );
    else if( !stricmp( inputWord.str, "Protected" ) )
      rc = storePath( cfg, &prot, "Protected", 0 );
    else if( !stricmp( inputWord.str, "MainZone" ) )
      rc = storeNumber( cfg, mainZone, "MainZone" );
    else if( !stricmp( inputWord.str, "Zones" ) )
    {
      if( list!=NULL )
        return 3;

      if( getWord( cfg ) )
        return 1;

      for( char *p=inputWord.str; *p; p++ )
        if( *p==',' )
          size++;

      list = new short [ ++size ];

      int i=0;

      for( p=inputWord.str; *p; )
      {
        long num = strtoul( p, &p, 10 );

        if( *p && *p!=',' )
        {
          strcpy( inputWord.str, p );
          return need( "," );
        }

        list[i++] = (short)num;

        if( *p )
          p++;
      }
    }
    else if( !stricmp( inputWord.str, "Add" ) )
    {
      type = NDLItemAdd;
process_item:

      if( getWord( cfg ) )
        return 1;

      short defZone( 0 ),
            defNet( 0 ),
            defNode( 0 );

      if( !stricmp( inputWord.str, "Default" ) )
      {
        if( getWord( cfg ) )
          return 1;

        char *defaul = inputWord.str;

        defZone = (short)strtoul( defaul, &defaul, 10 );

        if( !defZone || (*defaul && *defaul!=':') )
        {
          log( Msg_CfgErrIllDefaultZone, cfgLine, inputWord.str );
          return 2;
        }

        if( *defaul )
        {
          defNet = (short)strtoul( defaul+1, &defaul, 10 );

          if( !defNet || (*defaul && *defaul!='/') )
          {
            log( Msg_CfgErrIllDefaultNet, cfgLine, inputWord.str );
            return 2;
          }
        }

        if( *defaul )
        {
          defNode = (short)strtoul( defaul+1, &defaul, 10 );

          if( !defNode || *defaul )
          {
            log( Msg_CfgErrIllDefaultNode, cfgLine, inputWord.str );
            return 2;
          }
        }

        if( getWord( cfg ) )
          return 1;
      }

      char *ndlName = newStr( inputWord.str ),
           *ndlDiff = NULL;

      if( getWord( cfg ) )
        return 1;

      if( !stricmp( inputWord.str, "Diff" ) )
      {
        if( getWord( cfg ) )
          return 1;

        ndlDiff =  newStr( inputWord.str );
      }

      ndl->AddAtTail( *new NDLItem( ndlName, ndlDiff, type,
                                    defZone, defNet, defNode ) );

      delete ndlName;
      delete ndlDiff;

      if( ndlDiff==NULL )
        goto no_get_word;
    }
    else if( !stricmp( inputWord.str, "Replace" ) )
    {
      type = NDLItemReplace;
      goto process_item;
    }
    else
      rc = 3;
  }

  if( rc )
    return rc;

  if( outb==NULL )
  {
    char su[] = "\\";

    outb = new char [ strlen(name)+sizeof(su) ];
    strcpy( outb, name );
    strcat( outb, su );
  }

  char drive[_MAX_DRIVE],
       dir[_MAX_DIR],
       fname[_MAX_FNAME],
       ext[_MAX_EXT];

  outb[ strlen(outb)-1 ] = 0;
  _splitpath( outb, drive, dir, fname, ext );
  outb[ strlen(outb) ] = '\\';

  if( *ext || *drive || *dir )
  {
    log( Msg_ErrCfgOutExt, outLine );
    return 2;
  }

  if( list==NULL )
  {
    log( Msg_ErrCfgDomainNoZones, name );
    return 2;
  }

  if( mainZone )
  {
    char inList = 0;

    for( int i(0); i<size && !inList; i++ )
      if( list[i]==mainZone )
        inList = 1;

    if( !inList )
    {
      log( Msg_ErrCfgDomainIllMainZones, name, mainZone );
      return 2;
    }
  }

  outboundInfo.domains->AddAtTail( *new Domain( name, list, size,
						(short)mainZone,
                                                outb, ndl,
                                                inb, prot ) );

  delete name;
  delete outb;
  return 0;
}

int processNodeList( ifstream **cfg )
{
  char *defDomain = NULL;
  int defLine;
  char *exclude = NULL;
  int  exclLine;

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Path" ) )
      rc = storePath( cfg, &outboundInfo.path, "Path" );
    else if( !stricmp( inputWord.str, "OutboundRoot" ) )
      rc = storePath( cfg, &outboundInfo.root, "OutboundRoot" );
    else if( !stricmp( inputWord.str, "Indexes" ) )
      rc = storeWord( cfg, &outboundInfo.index, "Indexes" );
    else if( !stricmp( inputWord.str, "DefaultDomain" ) )
    {
      rc = storeWord( cfg, &defDomain, "DefaultDomain" );
      defLine = cfgLine;
    }
    else if( !stricmp( inputWord.str, "Exclude" ) )
    {
      rc = storeWord( cfg, &exclude, "Exclude" );
      exclLine = cfgLine;
    }
    else if( !stricmp( inputWord.str, "Domain" ) )
      rc = processDomain( cfg );
    else
      rc = 3;
  }

  if( rc )
    return rc;

  if( !outboundInfo.domains->GetItemsInContainer() )
  {
    const HOW = 6;
    short *arr = new short [HOW];

    for( int i=0; i<HOW; i++ )
      arr[i] = short(i+1);

    log( Msg_ErrCfgNoDomains );
    outboundInfo.domains->Add( *new Domain( "FidoNet", arr, HOW, 2,
					    "FidoNet", NULL ) );
  }

  if( defDomain==NULL )
  {
    log( Msg_ErrCfgNoDefDomain );
    outboundInfo.defDomain = &(*outboundInfo.domains)[0];
  }
  else
  {
    Domain test( defDomain ),
           *found = outboundInfo.domains->Find( test );

    if( found==NULL )
    {
      log( Msg_ErrCfgIllDomain, defLine, defDomain );
      return 2;
    }

    outboundInfo.defDomain = found;
    delete defDomain;
  }

  if( exclude!=NULL )
  {
    for( char *end = exclude; *end; )
    {
      unsigned long ret = strtoul( end, &end, 10 );

      if( !ret || (*end && *end!=',') )
      {
        log( Msg_ErrCfgIllExclude, exclLine, exclude );
        return 2;
      }

      if( *end )
        end++;
    }

    miscInfo.exclude = new char [ strlen(exclude)+3 ];
    strcpy( miscInfo.exclude, "," );
    strcat( miscInfo.exclude, exclude );
    strcat( miscInfo.exclude, "," );
    delete exclude;
  }
  else
    miscInfo.exclude = NULL;

  if( nodeListPtr!=NULL )
    log( "~BUG~ Internal error: NodeList object already exist!" );

  extern long nodeListCacheSize;

  Nodelist::lock();
  nodeListPtr = new Nodelist( int( nodeListCacheSize ) );
  Nodelist::unlock();

  return 0;
}

int processMelody( ifstream **cfg )
{
  char *noSound = NULL;
  int noSoundLine = 0;

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Connect" ) )
      rc = processSch( cfg, &melodies.connect );
    else if( !stricmp( inputWord.str, "BeginSession" ) )
      rc = processSch( cfg, &melodies.begSess );
    else if( !stricmp( inputWord.str, "EndSession" ) )
      rc = processSch( cfg, &melodies.endSess );
    else if( !stricmp( inputWord.str, "BadSession" ) )
      rc = processSch( cfg, &melodies.fail );
    else if( !stricmp( inputWord.str, "EndTransfer" ) )
      rc = processSch( cfg, &melodies.abort );
    else if( !stricmp( inputWord.str, "Human" ) )
      rc = processSch( cfg, &melodies.human );
    else if( !stricmp( inputWord.str, "ErrorAbort" ) )
      rc = processSch( cfg, &melodies.error );
    else if( !stricmp( inputWord.str, "Ring" ) )
      rc = processSch( cfg, &melodies.ring );
    else if( !stricmp( inputWord.str, "Alarm" ) )
      rc = processSch( cfg, &melodies.alarm );
    else if( !stricmp( inputWord.str, "Start" ) )
      rc = processSch( cfg, &melodies.start );
    else if( !stricmp( inputWord.str, "Stop" ) )
      rc = processSch( cfg, &melodies.stop );
    else if( !stricmp( inputWord.str, "Bell" ) )
      rc = processSch( cfg, &melodies.bell );
    else if( !stricmp( inputWord.str, "Chat" ) )
      rc = processSch( cfg, &melodies.chat );
    else if( !stricmp( inputWord.str, "AudioPath" ) )
      rc = storePath( cfg, &miscInfo.wavePath, "AudioPath" );
    else if( !stricmp( inputWord.str, "Volume" ) )
      rc = storeNumber( cfg, miscInfo.volume, "Volume" );
    else if( !stricmp( inputWord.str, "NoSound" ) )
    {
      rc = storeWord( cfg, &noSound, "NoSound" );
      noSoundLine = cfgLine;
    }
    else
      rc = need( "Connect/BeginSession/EndSession/EntTransfer/Human/ErrorAbort/Ring/Alarm/Start/Stop/Bell/Volume/AudioPath/End" );
  }

  if( noSound!=NULL )
  {
    miscInfo.noSound = new WorkTime( noSound );

    if( !miscInfo.noSound->isValid() )
    {
      log( Msg_ErrCfgIllTime, noSound, noSoundLine );
      return 2;
    }

    delete noSound;
  }

  return rc;
}

int processMask( ifstream **cfg )
{
  char *mask = NULL;
  int rc = storeWord( cfg, &mask, "Mask" );

  if( !rc )
  {
    EnableItem item( inputWord.str, EnableItemAddress );

    if( !item.isCorrect() )
    {
      log( Msg_ErrCfgIllRegExp, inputWord.str, cfgLine, item.getRegErrorMsg() );
      rc = 2;
    }

    if( !rc )
      rc = processNode( cfg, 0, 0, 0, 0, NULL, mask );
  }

  delete mask;
  return rc;
}

int processFAX( ifstream **cfg )
{
  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Inbound" ) )
      rc = storePath( cfg, &faxConfig.inbound, "Inbound" );
    else if( !stricmp( inputWord.str, "Baud" ) )
      rc = storeNumber( cfg, faxConfig.baud, "Baud" );
    else
      rc = need( "End/Baud/Inbound" );
  }

  return rc;
}

int processSecurity( ifstream **cfg )
{
  int pabort( -1 );

  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;
    else if( !stricmp( inputWord.str, "Password" ) )
    {
      if( getWord( cfg ) )
        return 1;

      NodeConfig *nn = new NodeConfig( inputWord.str );

      if( !nn->isValid() )
      {
        log( Msg_ErrCfgIllAddr, inputWord.str, cfgLine );
        delete nn;
        return 2;
      }

      if( getWord( cfg ) )
        return 1;

      NodeConfig *newNode = systemList->Find( *nn );

      if( newNode==NULL )
      {
        nn->changePassword( inputWord.str );
        systemList->AddAtTail( *nn );
      }
      else
        newNode->changePassword( inputWord.str );
    }
    else if( !stricmp( inputWord.str, "PasswordAbort" ) )
      rc = storeLogic( cfg, pabort, "PasswordAbort" );
    else
      rc = need( "End/Password/PasswordAbort" );
  }

  if( rc )
    return rc;

  switch( pabort )
  {
  case 0:
    miscInfo.state = SysState( int(miscInfo.state)&~PasswordAbort );
    break;

  case 1:
    miscInfo.state = SysState( int(miscInfo.state)|PasswordAbort );
    break;
  }

  return rc;
}


int processSection( ifstream **cfg )
{
  char needBegin( 1 );
  char *needWord( Msg_ErrCfgDefComp );

  while( !cfgStack->end() )
  {

    int rc = getWord( cfg );

    if( rc )
      if( rc==1 )
        return needBegin ? 0 : 1;
      else
        return rc;

    if( inputWord[0] )
      if( needBegin )
	if( !stricmp( inputWord.str, "Define" ) )
	  needBegin = 0;
	else
	  if( !stricmp( inputWord.str, "Complete." ) )
	    return 999;
	  else
err:  	    return need( needWord );
      else
      {
        static char first = 1;

        if( first )
        {
          first = 0;

          if( !stricmp( inputWord.str, "NodeList" ) )
            return processNodeList( cfg );
          else
            return need( "NodeList" );
        }
        else if( !stricmp( inputWord.str, "EMSI" ) )
 	  return processEmsi( cfg );
        else if( !stricmp( inputWord.str, "Miscellaneous" ) )
	  return processMisc( cfg );
        else if( !stricmp( inputWord.str, "Modem" ) )
	  return processModem( cfg );
        else if( !stricmp( inputWord.str, "Translate" ) )
	  return processTranslate( cfg );
        else if( !stricmp( inputWord.str, "XLAT" ) )
	  return processXLAT( cfg );
        else if( !stricmp( inputWord.str, "Process" ) )
	  return processProcess( cfg );
        else if( !stricmp( inputWord.str, "FREQ" ) )
	  return processFreq( cfg );
        else if( !stricmp( inputWord.str, "EchoSplit" ) )
	  return processSplit( cfg );
        else if( !stricmp( inputWord.str, "Keys" ) )
	  return processKeys( cfg );
        else if( !stricmp( inputWord.str, "StartUp" ) )
	  return processExitStartUp( cfg, &startUpList, &startUpCount );
        else if( !stricmp( inputWord.str, "Exit" ) )
	  return processExitStartUp( cfg, &exitList, &exitCount );
        else if( !stricmp( inputWord.str, "Video" ) )
	  return processVideo( cfg );
        else if( !stricmp( inputWord.str, "External" ) )
	  return processExternal( cfg );
        else if( !stricmp( inputWord.str, "Calls" ) )
	  return processCalls( cfg );
        else if( !stricmp( inputWord.str, "Accept" ) )
  	  return processAccept( cfg );
        else if( !stricmp( inputWord.str, "Protocol" ) )
	  return processProtocol( cfg );
        else if( !stricmp( inputWord.str, "RexxHook" ) )
	  return processRexx( cfg );
        else if( !stricmp( inputWord.str, "Melody" ) )
	  return processMelody( cfg );
        else if( !stricmp( inputWord.str, "Actions" ) )
	  return processActions( cfg );
        else if( !stricmp( inputWord.str, "Mask" ) )
	  return processMask( cfg );
        else if( !stricmp( inputWord.str, "FAX" ) )
	  return processFAX( cfg );
        else if( !stricmp( inputWord.str, "Security" ) )
	  return processSecurity( cfg );
        else
        {
	  uShort zone, net, node, point;
	  char *domain;

	  if( checkAddress( inputWord.str, zone, net, node, point, &domain ) )
	    return processNode( cfg, zone, net, node, point, domain, NULL );
	  else
	  {
	    needWord = Msg_ErrCfgSectAddr;
	    goto err;
	  }
        }
      }
  }

  return 0;
}

int processConfig( char *name )
{
  log( Msg_InfoReading, name );

  ifstream *cfg = new ifstream( name, ios::in|ios::nocreate );
  cfgFileList.registerFile( name );

  if( !cfg->good() )
  {
    log( Msg_ErrUnableOpen, name, strError() );
    return 2;
  }

  currentCfg = cfg;
  cfgLine = 1;

  for( int rc(0); !cfgStack->end() && !rc; rc = processSection( &cfg ) )
    continue;

  closeAllConfig( cfg );

  switch( rc )
  {
  case 1:
    inputWord.reset();
    inputWord.putString( "EOF" );
    inputWord.putInto( 0 );

// fall!
  case 3:

    log( Msg_ErrCfgNoNeed, inputWord.str, cfgLine );

  default:
    return 2;

  case 0:
  // fall
  case 999:
    if( !miscInfo.port )
      log( Msg_ErrCfgNeedPort );

    if( !systemInfo.addr->GetItemsInContainer() )
    {
      log( Msg_ErrCfgNeedSysAddr );
      return 2;
    }

    if( outboundInfo.defDomain==NULL )
    {
      log( Msg_ErrCfgNeedDomain );
      return 2;
    }

    break;        //ok processed
  }

  char task[80];

  sprintf( task, "Task\\Task.%lu.Is.Already.Active", miscInfo.taskNumber );

  if( checkUsage( task ) )
  {
    sprintf( task, "#%lu", miscInfo.taskNumber );
    log( Msg_ErrActiveLine, task );
    return 2;
  }

  if( checkUsage( miscInfo.schedComp ) )
  {
    log( Msg_ErrFileUsed, miscInfo.schedComp );
    return 2;
  }

  if( miscInfo.binLogName!=NULL )
    if( checkUsage( miscInfo.binLogName ) )
    {
      log( Msg_ErrFileUsed, miscInfo.binLogName );
      return 2;
    }

  if( systemList && maskList && 
      maskList->GetItemsInContainer() &&
      systemList->GetItemsInContainer() )
    systemList->postProcess();

  return 0;
}
