#define INCL_KBD
#define INCL_MOU
#define INCL_VIO
#define INCL_DOSDEVIOCTL
#define INCL_DOSDEVICES
#define INCL_DOSMEMMGR
#define INCL_DOSNMPIPES
#define INCL_DOSMISC
#define INCL_DOSFILEMGR
#define INCL_DOSPROCESS
#include <os2.h>

#include <fstream.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include <localdef.h>

#include "rexx.h"

#include "heap.h"
#include "misc.h"
#include "sched.h"
#include "queue.h"
#include "translat.h"
#include "zmodem.h"
#include "zinter.h"
#include "process.h"
#include "freq.h"
#include "calls.h"
#include "enable.h"
#include "config.h"
#include "rexxhook.h"
#include "timeout.h"
#include "hydra.h"
#include "addinqu.h"
#include "nolog.h"
#include "fax.h"

#include "language.h"

void dumpHandles( ostream* o, ULONG totalFh )
{
  char *hd = "*** File handles (size, mdate, type)";

  if( o!=NULL )
    (*o) << endl << hd << ", MaxAvail handles = " << totalFh << endl;
  else
  {
    log( hd );
    log( "*** MaxAvail handles = %lu", totalFh );
  }

  for( int j=0; j<totalFh; j++ )
  {
    ULONG type,
          attr;

    if( !DosQueryHType( j, &type, &attr ) )
    {
      char buf[512];

      {
        FILESTATUS3 fil;

        memset( &fil, 0, sizeof(fil) );

//        if( (type&0xf)==0 )
          DosQueryFileInfo( j, FIL_STANDARD, (PVOID)&fil, sizeof(fil) );

        sprintf( buf, "%3d -> %9lu %02lu:%02lu:%02lu %02lu/%02lu/%02lu",
                 j, fil.cbFile,
                 fil.ftimeLastWrite.hours,
                 fil.ftimeLastWrite.minutes,
                 fil.ftimeLastWrite.twosecs<1,
                 fil.fdateLastWrite.day,
                 fil.fdateLastWrite.month,
                 (fil.fdateLastWrite.year+80)%100
               );
      }

      if( (type&HANDTYPE_DEVICE)==HANDTYPE_DEVICE )
        strcat( buf, " Device" );
      else if( (type&HANDTYPE_PIPE)==HANDTYPE_PIPE )
        strcat( buf, " Pipe" );
      else
        strcat( buf, " File" );

      {
        USHORT err;
        ULONG errSize( sizeof(err) );

        if( !DosDevIOCtl( j, IOCTL_ASYNC, ASYNC_GETMODEMINPUT,
                          0, 0, 0, &err, errSize, &errSize ) ) 
          strcat( buf, " Communication" );
      }

      {
        char ch;
        ULONG readed = 0,
              state;
        AVAILDATA avail;

        if( !DosPeekNPipe( j, &ch, 1, &readed, &avail, &state ) )
          strcat( buf, " NamedPipe" );
      }

      {
        KBDKEYINFO k;

        if( !KbdPeek( &k, (HKBD)j ) )
          strcat( buf, " Keyboard" );
      }

      {
        USHORT buttons;

        if( !MouGetNumButtons( &buttons, (HMOU)j ) )
          strcat( buf, " Mouse" );
      }

      {
        USHORT codePageId;

        if( !VioGetCp( 0, &codePageId, (HVIO)j ) )
          strcat( buf, " Screen" );
      }

      {
        ULONG state;

        if( !DosQueryFHState( j, &state ) )
        {
          if( (state&OPEN_FLAGS_DASD)==OPEN_FLAGS_DASD )
            strcat( buf, " DASD" );

          if( (state&OPEN_FLAGS_WRITE_THROUGH)==OPEN_FLAGS_WRITE_THROUGH )
            strcat( buf, " WriteThrough" );

          if( (state&OPEN_FLAGS_FAIL_ON_ERROR)==OPEN_FLAGS_FAIL_ON_ERROR )
            strcat( buf, " FailOnError" );

          if( (state&OPEN_FLAGS_NO_CACHE)==OPEN_FLAGS_NO_CACHE )
            strcat( buf, " NoCache" );

          if( (state&OPEN_FLAGS_NOINHERIT)==OPEN_FLAGS_NOINHERIT )
            strcat( buf, " NoInherit" );

          if( (state&OPEN_SHARE_DENYNONE)==OPEN_SHARE_DENYNONE )
            strcat( buf, " DenyNone" );
          else if( (state&OPEN_SHARE_DENYREAD)==OPEN_SHARE_DENYREAD )
            strcat( buf, " DenyRead" );
          else if( (state&OPEN_SHARE_DENYREADWRITE)==OPEN_SHARE_DENYREADWRITE )
            strcat( buf, " DenyReadWrite" );
          else
            strcat( buf, " DenyWrite" );

          if( (state&OPEN_ACCESS_READWRITE)==OPEN_ACCESS_READWRITE )
            strcat( buf, " ReadWrite" );
          else if( (state&OPEN_ACCESS_WRITEONLY)==OPEN_ACCESS_WRITEONLY )
            strcat( buf, " WriteOnly" );
          else
            strcat( buf, " ReadOnly" );
        }
      }

      if( (type&HANDTYPE_NETWORK)==HANDTYPE_NETWORK )
        strcat( buf, " Remote" );

      if( (type&HANDTYPE_PROTECTED)==HANDTYPE_PROTECTED )
        strcat( buf, " Protected" );

      if( o!=NULL )
        (*o) << buf << endl;
      else
        log( buf );
    }
  }
}

void showSched( ostream& o, char *who, Schedule *sh )
{
  o << "  " << who << ": ";

  if( sh==NULL )
    o << "<None>";
  else
    o << *sh;

  o << endl;
}

extern ULONG isNowTotalFh;

void debugDump()
{
  showAction( "Debug dump" );

  char b1[80];

  sprintf( b1, "$Dump.%d", miscInfo.taskNumber );

  ofstream o( b1, ios::out|ios::app );

  if( !o.good() )
    log( Msg_ErrUnableOpen, b1, strError() );

  time_t elapsed( time( NULL ) );
  struct tm *tt( localTime( &elapsed ) );

  strftime( b1, sizeof(b1)-1, "%a %d %b %H:%M:%S", tt );

  int i, j;

  o << endl << "****** " << M_NAME << char(' ') << M_VERSION << char('/') << M_NUMBER
    << " Dump at " << b1 << endl;

  char *tz( getenv("TZ") );

  {
    ULONG b[26],
          &up( b[QSV_MS_COUNT-1] );

    DosQuerySysInfo( QSV_MAX_PATH_LENGTH, 26, b, sizeof(b) );

    char *tt( strTime( up/1000, 0 ) );

    o << endl << "*** OS: " << b[QSV_VERSION_MAJOR-1] << char('.')
                            << b[QSV_VERSION_MINOR-1] << char('.')
                            << b[QSV_VERSION_REVISION-1] << " from "
                            << char(b[QSV_BOOT_DRIVE-1]+'A'-1) << ", "
                            << b[QSV_TOTPHYSMEM-1] << " memory and "
                            << b[26-1] << " CPU(s), up " << tt << endl;
    delete tt;
  }

  o << endl << "*** TZ: " << (tz==NULL ? "No TZ!" : tz)
    << ", Current:" << getTimeZone()
    << endl << "        "
       << (__lc_tod->tname==NULL ? "*None*" : __lc_tod->tname )
       << __lc_tod->tzdiff
       << (__lc_tod->tzname==NULL ? "*None*" : __lc_tod->tzname )
       << ",ms="    << __lc_tod->dststm
       << ",me="    << __lc_tod->dstenm
       << ",ws="    << __lc_tod->dststw
       << ",we="    << __lc_tod->dstenw
       << ",ds="    << __lc_tod->dststd
       << ",de="    << __lc_tod->dstend
       << ",ss="    << __lc_tod->starttm
       << ",se="    << __lc_tod->endtm
       << ",shift=" << __lc_tod->shift
       << ",UTC:"   << (__lc_tod->uctname==NULL ? "*None*" : __lc_tod->uctname )
    << endl;

  o << endl << "*** Domains:" << endl << endl
    << " Exclude: ";
  strItem( o, miscInfo.exclude );
  o << endl
    << "    Path: " << outboundInfo.path << endl
    << "    Root: " << outboundInfo.root << endl
    << "   Index: " << outboundInfo.index << endl
    << " Default: " << outboundInfo.defDomain->getName() << endl
    << "    List: " << endl << *outboundInfo.domains << endl;

  extern int fstblklen;

  o << endl << "*** Miscellenous:" << endl;
  o << "  AddrForm:      " << char(miscInfo.addrForm[0])
                           << char(miscInfo.addrForm[1])
                           << endl;
  o << "  TaskNumber:    " << miscInfo.taskNumber << endl;
  o << "  ConfigCheck:   " << miscInfo.configCheck << endl;
  o << "  DefAddress:    " << defaultAddress->toString( b1, sizeof(b1)-1 ) << endl;
  o << "  Thrashold:     " << miscInfo.trash << endl;
  o << "  CircleMin:     " << miscInfo.circleMin << endl;
  o << "  CircleMax:     " << miscInfo.circleMax << endl;
  o << "  ReDial:        " << miscInfo.redial << endl;
  o << "  AutoReScan:    " << miscInfo.rescan << endl;
  o << "  ReScanPeriod:  " << miscInfo.rescanPeriod << endl;
  o << "  EmptyExit:     " << miscInfo.emptyExit << endl;
  o << "  WaitMailer:    " << miscInfo.EMSItime << endl;
  o << "  MaxInComing:   " << miscInfo.maxIn << endl;
  o << "  MaxOutGoing:   " << miscInfo.maxOut << endl;
  o << "  MaxOutTry:     " << miscInfo.maxTry << endl;
  o << "  MaxCallCost:   " << miscInfo.maxCost << endl;
  o << "  HoldTime:      " << miscInfo.holdTime << endl;
  o << "  Port:          " << miscInfo.port << endl;
  o << "  Baud:          " << miscInfo.baud << endl;
  o << "  ChatTimeout:   " << miscInfo.chatTimeout << endl;
  o << "  TryToSend:     " << miscInfo.sendTry << endl;
  o << "  HeapDebug:     " << int(heapDebug) << endl;
  o << "  Detached:      " << (detachedSession ? "Yes" : "No") << endl;
  o << "  ImmedPrio:     " << immedFlavourPriority << endl;
  o << "  CrashPrio:     " << crashFlavourPriority << endl;
  o << "  DirectPrio:    " << directFlavourPriority << endl;
  o << "  NormalPrio:    " << normalFlavourPriority << endl;
  o << "  HoldPrio:      " << holdFlavourPriority << endl;
  o << "  ProcessOnce:   " << ((int(miscInfo.state)&ProcessOnce) ? "Yes" : "No") << endl;
  o << "  ModemWindow:   " << ((int(miscInfo.state)&ModemWindow) ? "Yes" : "No") << endl;
  o << "  Debug:         " << ((int(miscInfo.state)&ZDebug) ? "Yes" : "No") << endl;
  o << "  EMSIDebug:     " << ((int(miscInfo.state)&EMSIDebug) ? "Yes" : "No") << endl;
  o << "  Collision:     " << ((int(miscInfo.state)&Collision) ? "Yes" : "No") << endl;
  o << "  SplitDelete:   " << ((int(miscInfo.state)&SplitDelete) ? "Yes" : "No") << endl;
  o << "  PasswdAbort:   " << ((int(miscInfo.state)&PasswordAbort) ? "Yes" : "No") << endl;
  o << "  BinkLog:       " << ((int(miscInfo.state)&BinkLog) ? "Yes" : "No") << endl;
  o << "  AutoBBS:       " << ((int(miscInfo.state)&AutoBBS) ? "Yes" : "No") << endl;
  o << "  Full5D:        " << ((int(miscInfo.state)&Full5D) ? "Yes" : "No") << endl;
  o << "  ScanBefore:    " << ((int(miscInfo.state)&ScanBefore) ? "Yes" : "No") << endl;
  o << "  HydraChat:     " << ((int(miscInfo.state)&HydraChat) ? "Yes" : "No") << endl;
  o << "  ModemReport:   " << ((int(miscInfo.state)&ModemReport) ? "Yes" : "No") << endl;
  o << "  EA:            " << ((int(miscInfo.state)&NoEA) ? "No" : "Yes") << endl;
  o << "  CallAfterAns:  " << ((int(miscInfo.state)&CallAfterAnswer) ? "Yes" : "No") << endl;
  o << "  TerminalLog:   " << ((int(miscInfo.state)&TerminalLog) ? "Yes" : "No") << endl;
  o << "  PKTFromLO:     " << ((int(miscInfo.state)&SplitPKTFromLO) ? "Yes" : "No") << endl;
  o << "  BinkDRecover:  " << ((int(miscInfo.state)&BinkDRecover) ? "Yes" : "No") << endl;
  o << "  ExportConfig:  " << ((int(miscInfo.state)&ExportConfig) ? "Yes" : "No") << endl;
  o << "  DaemonMode:    " << daemonMode << endl;
  o << "  DiscardPKT:    " << ((int(miscInfo.state1)&DiscardPKT) ? "Yes" : "No") << endl;
  o << "  HoldOnCall:    " << ((int(miscInfo.state1)&HoldOnCall) ? "Yes" : "No") << endl;
  o << "  DOSTime:       " << ((int(miscInfo.state1)&AdjustTime) ? "Yes" : "No") << endl;
  o << "  RetryExtern:   " << ((int(miscInfo.state1)&RetryAfterExtern) ? "Yes" : "No") << endl;
  o << "  StatusLine:    " << ((int(miscInfo.state1)&StatusLine) ? "Yes" : "No") << endl;
  o << "  StatusPeriod:  " << miscInfo.statusLine << endl;
  o << "  NotAcceptAbo:  " << ((int(miscInfo.state1)&NotAcceptAbort) ? "Yes" : "No") << endl;
  o << "  FullLog:       " << ((int(miscInfo.state2)&FullLog) ? "Yes" : "No") << endl;
  o << "  RexxDebug:     " << ((int(miscInfo.state2)&RexxDebug) ? "Yes" : "No") << endl;
  o << "  DirZap:        " << ((int(miscInfo.state2)&DirZapAvail) ? "Yes" : "No") << endl;
  o << "  ZedZap:        " << ((int(miscInfo.state2)&ZedZapAvail) ? "Yes" : "No") << endl;
  o << "  Hydra:         " << ((int(miscInfo.state2)&HydraAvail) ? "Yes" : "No") << endl;
  o << "  BSYSema:       " << ((int(miscInfo.state2)&SemaphoreBsy) ? "Yes" : "No") << endl;
  o << "  TaskInLog:     " << ((int(miscInfo.state2)&TaskInLog) ? "Yes" : "No") << endl;
  o << "  SplitterLng:   " << ((int(miscInfo.state2)&SplitterLong) ? "Yes" : "No") << endl;
  o << "  SplitterRemov: " << ((int(miscInfo.state2)&SplitDelNoFound) ? "Yes" : "No") << endl;
  o << "  HPFS:          " << ((int(miscInfo.state3)&HPFS) ? "Yes" : "No") << endl;
  o << "  EnableMMPM:    " << (miscInfo.enableMMPM ? "Yes" : "No") << endl;
  o << "  FileCase:      " << miscInfo.fileCase << endl;

  o << "  Separator:     ";

  if( miscInfo.separator )
    o << char(miscInfo.separator) << endl;
  else
    o << "*None*" << endl;

  o << "  CheckREQ:      " << ((int(miscInfo.state3)&CheckREQ) ? "Yes" : "No") << endl;
  o << "  CheckName:     " << ((int(miscInfo.state3)&CheckName) ? "Yes" : "No") << endl;
  o << "  FastScan:      " << ((int(miscInfo.state3)&FastScan) ? "Yes" : "No") << endl;
  o << "  ScanOnLine:    " << ((int(miscInfo.state3)&ScanOnLine) ? "Yes" : "No") << endl;
  o << "  Full Screen:   " << (isFullScreen ? "Yes" : "No") << endl;
  o << "  PM Features:   " << (!noPMFeatures ? "Yes" : "No") << endl;
  o << "  StartBlock:    " << miscInfo.zBlock << endl;
  o << "  ZBaseTimeOut:  " << miscInfo.timeOut << endl;
  o << "  HBaseTimeOut:  " << hydra_timeout << endl;
  o << "  HRxWindow:     " << hydra_rxwindow << endl;
  o << "  HTxWindow:     " << hydra_txwindow << endl;

  char hOption[ 512 ] = {0};

  hydraPutFlags( hOption, hydra_options );

  if( !*hOption )
    strcpy( hOption, "*None*" );

  o << "  HOption:       " << hOption << endl;
  o << "  ZCheckAfter:   " << zTransferCpsAfter << endl;
  o << "  HCheckAfter:   " << hTransferCpsAfter << endl;
  o << "  SplitEchoSize: " << miscInfo.splitSize << endl;
  o << "  SplitPktSize:  " << miscInfo.splitPktSize << endl;
  o << "  SplitFileSize: " << miscInfo.splitFileSize << endl;
  o << "  SplitCircle:   " << miscInfo.splitCircle << endl;
  o << "  Priority:      " << miscInfo.priority << endl;
  o << "  PriorDelta:    " << miscInfo.priorityDelta << endl;
  o << "  RescanPrior:   " << miscInfo.reScanPriority << endl;
  o << "  RescanPriDel:  " << miscInfo.reScanPriorityDelta << endl;
  o << "  CopilerPrior:  " << miscInfo.compilerPriority << endl;
  o << "  CopilerPriDel: " << miscInfo.compilerPriorityDelta << endl;
  o << "  LogBuffer:     " << miscInfo.logBuffer << endl;
  o << "  BsyAge:        " << miscInfo.bsyAge << endl;
  o << "  BsyCheck:      " << miscInfo.bsyInterval << endl;
  o << "  FlagsCheck:    " << miscInfo.checkFlags << endl;
  o << "  ScheduleCheck: " << miscInfo.checkSched << endl;
  o << "  TermTimeout:   " << miscInfo.terminalTimeOut << endl;
  o << "  Volume:        " << miscInfo.volume << endl;
  o << "  InFilBufSz:    " << miscInfo.inFileBufferSize << endl;
  o << "  OutFilBufSz:   " << miscInfo.outFileBufferSize << endl;
  o << "  ScrSaverWait:  " << miscInfo.saverInterval << endl;
  o << "  ScrSaverType:  " << (miscInfo.saverModule ? miscInfo.saverModule : "*BUILTIN*") << endl;

  o << "  InBound:       ";
  strItem( o, miscInfo.inbound );

  o << endl << "  TerminalIn:    ";
  strItem( o, miscInfo.terminalInbound );

  o << endl << "  Protected:     ";
  strItem( o, miscInfo.protectedInbound );

  o << endl << "  SentList:      ";
  strItem( o, miscInfo.sentList );

  o << endl << "  TBoxes:        ";
  strItem( o, miscInfo.tbox );

  o << endl << "  LongBoxes:     ";
  strItem( o, miscInfo.longBox );

  o << endl << "  LongOutbound:  ";
  strItem( o, miscInfo.longOutbound );

  o << endl << "  Flags:         ";
  strItem( o, miscInfo.flags );

  o << endl << "  RexxPath:      ";
  strItem( o, miscInfo.rexxPath );

  o << endl << "  WavePath:      ";
  strItem( o, miscInfo.wavePath );

  o << endl << "  Semaphores:    ";
  strItem( o, miscInfo.semaphore );

  o << endl << "  EMSILog:       ";
  strItem( o, miscInfo.emsiLog );

  o << endl << "  Pipe:          ";
  strItem( o, miscInfo.pipe );

  o << endl << "  Schedule:      ";
  strItem( o, miscInfo.schedule );

  o << endl << "  Compiled:      ";
  strItem( o, miscInfo.schedComp );

  o << endl << "  RemoteName:    ";
  strItem( o, miscInfo.remoteName );

  o << endl << "  ExitFlag:      ";
  strItem( o, miscInfo.exitFlag );

  o << endl << "  IconMail:      ";
  strItem( o, miscInfo.iconMail );

  o << endl << "  IconXMail:     ";
  strItem( o, miscInfo.iconXMail );

  o << endl << "  IconRIP:       ";
  strItem( o, miscInfo.iconRIP );

  o << endl << "  RemotePasswd:  " << miscInfo.remotePassword;

  o << endl << "  ModemWindow:   " << miscInfo.wndModemX << ", "
				   << miscInfo.wndModemY << ", "
				   << miscInfo.wndModemOff;

  o << endl << "  Frame:         " << miscInfo.x  << ", "
				   << miscInfo.y  << ", "
				   << miscInfo.cx << ", "
				   << miscInfo.cy;

  o << endl;

  o << endl << "*** FAX" << endl
    << "  Inbound: ";

  strItem( o, faxConfig.inbound );

  o << endl << "  Baud:    " << faxConfig.baud << endl;


  o << endl << "*** Video" << endl;

#define HEXPAIR(x) hexString[((x)>>4)&0xf] << hexString[(x)&0xf]

  o << "  VideoMode:     " << colorSet.x << "x" << colorSet.y << endl;
  o << "  Color:         " << (colorSet.isColor ? "Yes" : "No") << endl;
  o << "  X/Y:           " << Z_ScreenWidth << "/" << Z_ScreenHeight << endl;
  o << "  Status:        " << HEXPAIR(colorSet.status) << endl;
  o << "  StatusLampOn:  " << HEXPAIR(colorSet.statusLampOn) << endl;
  o << "  StatusLampOff: " << HEXPAIR(colorSet.statusLampOff) << endl;
  o << "  Border:        " << HEXPAIR(colorSet.border) << endl;
  o << "  Action:        " << HEXPAIR(colorSet.action) << endl;
  o << "  ActionBold:    " << HEXPAIR(colorSet.actionBold) << endl;
  o << "  Abort:         " << HEXPAIR(colorSet.abort) << endl;
  o << "  PopUp:         " << HEXPAIR(colorSet.popUp) << endl;
  o << "  PopUpItem:     " << HEXPAIR(colorSet.popUpItem) << endl;
  o << "  PopUpBold:     " << HEXPAIR(colorSet.popUpBold) << endl;
  o << "  Terminal:      " << HEXPAIR(colorSet.terminal) << endl;
  o << "  Help:          " << HEXPAIR(colorSet.help) << endl;
  o << "  HelpBold:      " << HEXPAIR(colorSet.helpBold) << endl;
  o << "  QueueCrash:    " << HEXPAIR(colorSet.queueCrash) << endl;
  o << "  QueueNormal:   " << HEXPAIR(colorSet.queueNormal) << endl;
  o << "  QueueHold:     " << HEXPAIR(colorSet.queueHold) << endl;
  o << "  QueueImmed:    " << HEXPAIR(colorSet.queueImmed) << endl;
  o << "  QueueDirect:   " << HEXPAIR(colorSet.queueDirect) << endl;
  o << "  QueuePtr:      " << HEXPAIR(colorSet.queuePtr) << endl;
  o << "  ModemTo:       " << HEXPAIR(colorSet.modemTo) << endl;
  o << "  ModemFrom:     " << HEXPAIR(colorSet.modemFrom) << endl;
  o << "  ModemCtlTo:    " << HEXPAIR(colorSet.modemCtlTo) << endl;
  o << "  ModemCtlFrom:  " << HEXPAIR(colorSet.modemCtlFrom) << endl;

#undef HEXPAIR

  o << endl << "*** ZMH ";
  if( miscInfo.zmh==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << miscInfo.zmh << ":\n" << *miscInfo.zmh << endl;

  o << endl << "*** AnswerTime ";
  if( miscInfo.answerTime==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << miscInfo.answerTime << ":\n" << *miscInfo.answerTime << endl;

  o << endl << "*** CallTime ";
  if( miscInfo.callTime==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << miscInfo.callTime << ":\n" << *miscInfo.callTime << endl << endl;

  o << endl << "*** CallOnlyImmed ";
  if( miscInfo.callOnlyImmed==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << miscInfo.callOnlyImmed << ":\n" << *miscInfo.callOnlyImmed << endl << endl;

  o << endl << "*** CallOnlyCrash ";
  if( miscInfo.callOnlyCrash==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << miscInfo.callOnlyCrash << ":\n" << *miscInfo.callOnlyCrash << endl << endl;

  o << endl << "*** CallOnlyDirect ";
  if( miscInfo.callOnlyDirect==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << miscInfo.callOnlyDirect << ":\n" << *miscInfo.callOnlyDirect << endl << endl;

  o << endl << "*** CallOnlyNormal ";
  if( miscInfo.callOnlyNormal==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << miscInfo.callOnlyNormal << ":\n" << *miscInfo.callOnlyNormal << endl << endl;

  o << endl << "*** NoLOG ";
  if( noLogList==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << noLogList << ":\n" << *noLogList << endl << endl;

  o << endl << "*** NoSCR ";
  if( noScrList==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << noScrList << ":\n" << *noScrList << endl << endl;

  o << endl << "*** BBS info:" << endl << endl
    << "  BBS:         " << ((int(miscInfo.state1)&BBSenable) ? "Yes" : "No") << endl
    << "  MinBaud:     " << miscInfo.minBBSBaud << endl
    << "  TwiceEsc:    " << bbsInfo.twiceEsc << endl
    << "  MailOnly:    " << bbsInfo.mailOnly << endl
    << "  SlowModem:   " << bbsInfo.slowModem << endl
    << "  MailNow:     " << bbsInfo.mailNow << endl
    << "  Loading:     " << bbsInfo.loading << endl
    << "  GoodBye:     " << bbsInfo.goodBye << endl;

  o << endl << endl << "*** BBSTime ";
  if( miscInfo.BBSTime==NULL )
    o << "<NULL>" << endl << endl;
  else
    o << "at " << miscInfo.BBSTime << ":\n" << *miscInfo.BBSTime << endl << endl;

  o << endl << endl << "*** RunBBS ";
  if( processBBS==NULL )
    o << "<NULL>" << endl << endl;
  else
    o << "at " << processBBS << ":\n" << *processBBS << endl << endl;

  o << "*** Freq info:" << endl << endl
    << "  FREQ:        ";

  if( miscInfo.freqMode==NoFreq )
    o << "No";

  if( miscInfo.freqMode==AllFreq )
    o << "Yes";

  if( miscInfo.freqMode==ProtectedFreq )
    o << "Protected";

  if( miscInfo.freqMode==ListedFreq )
    o << "Listed";

  o << endl;

  o << "  MinBaud:     " << miscInfo.minFreqBaud << endl
    << "  FreqBase:    " << miscInfo.freqBase << endl
    << "  Index:       " << miscInfo.freqIdx << endl
    << "  SessionTime: " << miscInfo.sessionFreqTime << " min." << endl
    << "  SessionSize: " << miscInfo.sessionFreqSize << "Kb" << endl
    << "  DayTime:     " << miscInfo.dayFreqTime << " min." << endl
    << "  DaySize:     " << miscInfo.dayFreqSize << "Kb" << endl
    << "  WeekTime:    " << miscInfo.weekFreqTime << " min." << endl
    << "  WeekSize:    " << miscInfo.weekFreqSize << "Kb" << endl
    << "  MonTime:     " << miscInfo.monFreqTime << " min." << endl
    << "  MonSize:     " << miscInfo.monFreqSize << "Kb" << endl
    << "  FreqPolicy:  ";
  strItem( o, miscInfo.freqPolicy );

  o << endl << endl << "*** FREQTime ";
  if( miscInfo.freqTime==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << miscInfo.freqTime << ":\n" << *miscInfo.freqTime << endl << endl;

  o << "*** FREQAliases ";
  if( freqAlias==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << freqAlias << ":\n" << *freqAlias << endl << endl;

  o << "*** FREQList ";
  if( freqList==NULL )
    o << "<NULL>" << endl;
  else
    o << "at " << freqList << ":\n" << *freqList << endl << endl;

  o << "*** Melody: " << endl;
  showSched( o, "Connect     ", melodies.connect );
  showSched( o, "BegSess     ", melodies.begSess );
  showSched( o, "BadSess     ", melodies.fail );
  showSched( o, "EndSess     ", melodies.endSess );
  showSched( o, "EndTransfer ", melodies.abort );
  showSched( o, "Human       ", melodies.human );
  showSched( o, "ErrorAbort  ", melodies.error );
  showSched( o, "Ring        ", melodies.ring );
  showSched( o, "Alarm       ", melodies.alarm );
  showSched( o, "Start       ", melodies.start );
  showSched( o, "Stop        ", melodies.stop );
  showSched( o, "Bell        ", melodies.bell );
  showSched( o, "Chat        ", melodies.chat );

  o << endl 
    << "*** arqList at " << arqList << char( ':' ) << endl
    << *arqList << endl << endl;
  
  o << endl 
    << "*** noHydraChatList at " << noHydraChatList << char( ':' ) << endl
    << *noHydraChatList << endl << endl;

  o << "*** halfDuplexList at " << halfDuplexList << char( ':' ) << endl
    << *halfDuplexList << endl << endl;

  o << "*** addInQueueList at " << addInQueueList << char( ':' ) << endl
    << *addInQueueList << endl << endl;

  o << "*** SystemInfo:";

  o << endl << "  Name:    ";
  strItem( o, systemInfo.name );

  o << endl << "  SysOp:   ";
  strItem( o, systemInfo.sysopName );

  o << endl << "  Place:   ";
  strItem( o, systemInfo.place );

  o << endl << "  Phone:   ";
  strItem( o, systemInfo.phone );

  o << endl << "  Baud:    ";
  strItem( o, systemInfo.baud );

  o << endl << "  Flags:   ";
  strItem( o, systemInfo.flags );

  o << endl << "  Addresses: " << endl << *systemInfo.addr << endl;


  o << endl << endl << "*** Modem at " << comPort << ":\n";

  if( comPort->getPort() )
    o << *comPort;
  else
    o << *(FictionModem *)comPort;

  o << endl << "  MinInBound:  " << miscInfo.minIn << endl
	    << "  MinOutBound: " << miscInfo.minOut << endl
	    << "  Telebit:     " << ((int(miscInfo.state1)&Telebit) ? "Yes" : "No") << endl
            << "  MaxRRing:    " << miscInfo.maxRRing  << endl
            << "  RingCount:   " << miscInfo.ringCount << endl
            << "  RingWait:    " << miscInfo.ringWait << endl
            << "  HardRing:    " << ((int(miscInfo.state3)&HardRing) ? "Yes" : "No") << endl
            << "  AutoInit:    " << miscInfo.autoInit << endl
            << "  Fax:         " << modemString.fax << endl
            << "  Ring:        " << modemString.ring << endl
            << "  Busy:        " << modemString.busy << endl
            << "  Ok:          " << modemString.ok << endl
            << "  Connect:     " << modemString.connect << endl
            << "  NoDialTone:  " << modemString.nodialtone << endl
            << "  RRing:       " << modemString.rring << endl
            << "  NoCarrier:   " << modemString.nocarrier << endl
            << "  Error:       " << modemString.error << endl
            << "  Voice:       " << modemString.voice << endl
            << "  NoAnswer:    " << modemString.noanswer << endl
            << "  NoConnect:   " << modemString.noconnect << endl
            << "  Caller's ID: " << modemString.callersId;

  o << endl << endl << "*** Ignored string list at " << ignoreString << ":\n"
					             << *ignoreString << endl;

  o << endl << endl << "*** Bad counter string list at " << badCounterString << ":\n"
					                 << *badCounterString << endl;

  if( timeOuts==NULL )
    o << endl << endl << "*** No personel z-timeouts!" << endl;
  else
    o << endl << endl << "*** ZTimeOuts list at " << timeOuts << ":\n"
					       << *timeOuts << endl;

  if( zMinOutCps==NULL )
    o << endl << endl << "*** No zMinOutCps!" << endl;
  else
    o << endl << endl << "*** zMinOutCps list at " << zMinOutCps << ":\n"
					       << *zMinOutCps << endl;

  if( zMinInCps==NULL )
    o << endl << endl << "*** No zMinInCps!" << endl;
  else
    o << endl << endl << "*** zMinInCps list at " << zMinInCps << ":\n"
						<< *zMinInCps << endl;

  if( timeOutsHydra==NULL )
    o << endl << endl << "*** No personel h-timeouts!" << endl;
  else
    o << endl << endl << "*** HTimeOuts list at " << timeOutsHydra << ":\n"
					       << *timeOutsHydra << endl;

  if( hMinOutCps==NULL )
    o << endl << endl << "*** No hMinOutCps!" << endl;
  else
    o << endl << endl << "*** hMinOutCps list at " << hMinOutCps << ":\n"
					       << *hMinOutCps << endl;

  if( hMinInCps==NULL )
    o << endl << endl << "*** No hMinInCps!" << endl;
  else
    o << endl << endl << "*** hMinInCps list at " << hMinInCps << ":\n"
						<< *hMinInCps << endl;

  if( preffixList==NULL )
    o << endl << endl << "*** No special preffixes!" << endl;
  else
    o << endl << endl << "*** preffix list at " << preffixList << ":\n"
						<< *preffixList << endl;

  if( suffixList==NULL )
    o << endl << endl << "*** No special suffixes!" << endl;
  else
    o << endl << endl << "*** suffix list at " << suffixList << ":\n"
					       << *suffixList << endl;

  if( enabledList==NULL )
    o << endl << endl << "*** No enabled calls specified!" << endl;
  else
    o << endl << endl << "*** Enabled calls at " << enabledList << ":\n"
						<< *enabledList << endl;

  if( disabledList==NULL )
    o << endl << endl << "*** No disabled calls specified!" << endl;
  else
    o << endl << endl << "*** Disabled calls at " << disabledList << ":\n"
						<< *disabledList << endl;

  if( acceptedList==NULL )
    o << endl << endl << "*** No accepted list specified!" << endl;
  else
    o << endl << endl << "*** Accepted list at " << acceptedList << ":\n"
						<< *acceptedList << endl;

  if( unAcceptedList==NULL )
    o << endl << endl << "*** No unaccepted list specified!" << endl;
  else
    o << endl << endl << "*** Unaccepted list at " << unAcceptedList << ":\n"
						   << *unAcceptedList << endl;

  if( timeSyncList==NULL )
    o << endl << endl << "*** No TimeSync list specified!" << endl;
  else
    o << endl << endl << "*** TimeSync list at " << timeSyncList << ":\n"
                                                 << *timeSyncList << endl;

  o << "    TimeSyncDelta: " << miscInfo.timeSyncDelta;

  o << endl << endl << "*** Online list at " << onlineTable << ":\n"
					       << *onlineTable << endl;

  o << endl << endl << "*** Process list at " << processTable << ":\n"
					       << *processTable << endl;

  o << endl << endl << "*** Immediate list at " << immediateTable << ":\n"
					       << *immediateTable << endl;

  o << endl << endl << "*** System list at " << systemList << ":\n"
					     << *systemList << endl;

  o << endl << endl << "*** Mask list at " << maskList << ":\n"
				           << *maskList << endl;

  o << endl << "*** Translation table at " << translationTable << ":\n"
					   << *translationTable << endl;

  o << endl << "*** Config file list: " << cfgFileList << endl;

  o << endl << "*** Events list at " << schedList << ", " << schedCount << " events:\n";

  if( schedList==NULL || !schedCount )
    o << "*None*";
  else
    for( i=0; i<schedCount; i++ )
    {
      char buf[80];

      sprintf( buf, "%02d-%02d.%02d:%02d:%02d%c%c \"", schedList[i].dayOfMonth,
               schedList[i].dayOfWeek, schedList[i].start/60/60,
               schedList[i].start/60%60, schedList[i].start%60,
               schedList[i].isGmt ? 'G' : 'L',
               schedList[i].done ? '*' : ' ' );
      o << buf << schedList[i].name << "\" ->"
        << schedList[i] << endl;
    }

  o << endl;

  o << endl << "*** Keys list at " << keyList << " " << keyCount << " keys:\n";

  if( keyList==NULL || !keyCount )
    o << "*None*";
  else
    for( i=0; i<keyCount; i++ )
      o << i << "> " << keyList[i] << endl;

  o << endl;

  o << endl << "*** StartUp list at " << startUpList << " "
    << startUpCount << " items:\n";

  if( startUpList==NULL || !startUpCount )
    o << "*None*";
  else
    for( i=0; i<startUpCount; i++ )
      o << i << "> " << startUpList[i] << endl;

  o << endl;

  o << endl << "*** Exit list at " << exitList << " "
    << exitCount << " items:\n";

  if( exitList==NULL || !exitCount )
    o << "*None*";
  else
    for( i=0; i<exitCount; i++ )
      o << i << "> " << exitList[i] << endl;

  o << endl;

  o << endl << "*** Action list at " << actionList << " "
    << actionCount << " items:\n";

  if( actionList==NULL || !actionCount )
    o << "*None*";
  else
    for( i=0; i<actionCount; i++ )
      o << i << "> " << actionList[i] << endl;

  o << endl;

  o << endl << "*** Flags list at " << flagList << " " << flagCount << " flags:\n";

  if( flagList==NULL || !flagCount )
    o << "*None*";
  else
    for( i=0; i<flagCount; i++ )
      o << i << "> " << flagList[i] << endl;

  o << endl;

  o << endl << "*** Connect list at " << connectList << " " << connectCount
    << " items:\n";

  if( connectList==NULL || !connectCount )
    o << "*None*";
  else
    for( i=0; i<connectCount; i++ )
      o << i << "> " << connectList[i] << endl;

  o << endl;

  o << endl << "*** Handshake list at " << handShakeList << " " << handShakeCount
    << " items:\n";

  if( handShakeList==NULL || !handShakeCount )
    o << "*None*";
  else
    for( i=0; i<handShakeCount; i++ )
      o << i << "> " << handShakeList[i] << endl;

  if( rexxHookList==NULL )
    o << endl << endl << "*** No REXX Hooks!" << endl;
  else
    o << endl << endl << "*** Rexx Hooks list at " << rexxHookList << ":\n"
					   << *rexxHookList << endl;

  queue->read();

  o << endl << "*** Queue at " << queue << ":\n"
			       << *queue
			       << (queue->GetItemsInContainer() ? "" : "  <Empty>\n")
			       << endl;

  o << "*** Nodelist cache:" << endl;
  if( nodeListPtr==NULL )
    o << "  <No NodeList found>";
  else
  {
    nodeList.lock();
    o << nodeList;
    nodeList.unlock();
  }
  o << endl;

  o << "*** Memory:" << endl;

  void *pp;
  DosAllocMem( &pp, 10, PAG_COMMIT|PAG_READ|PAG_WRITE );
  DosFreeMem( pp );
  o << "  Virtual mem: " << long(pp) << endl;

  o << "  Heap status: " << _heapchk() << endl;

  HeapStat heap;

  o << "  Heap info: " << heap.isUsed << " bytes, "
                       << heap.totUsed << " nodes is used," << endl
    << "             " << heap.isFree << " bytes, "
                       << heap.totFree << " nodes is free" << endl
    << "             " << heap.isUndef << " bytes, "
                       << heap.totUndef << " nodes is unknown" << endl
    << "  HeapWalk:  " << heap.heapEnd << endl
    << "  BadObject: " << heap.badObject << endl;

  if( heap.badLine && heap.badFile!=NULL )
    o << "  Bad alloc: " << heap.badLine << " line from " << heap.badFile << endl;

  o << endl << "*** XLAT Input:\n";

  for( j=0, i=0; i<256; i++ )
    if( terminalInXLAT[i]!=-1 )
    {
      j=1;
      o << "  " << i << " -> " << terminalInXLAT[i] << endl;
    }

  if( !j )
    o << "  <None>" << endl;

  o << endl << "*** XLAT Output:\n";

  for( j=0, i=0; i<256; i++ )
    if( terminalOutXLAT[i]!=-1 )
    {
      j=1;
      o << "  " << i << " -> " << terminalOutXLAT[i] << endl;
    }

  if( !j )
    o << "  <None>" << endl;

  o << endl << "*** XLAT " << ansiHow << " words at " << ansiTab
    << ":\n";

  for( i=0; i<ansiHow; i++ )
    o << i << ": " << ansiTab[i];

  o << endl << "*** Mutex Semaphores list:" << endl << mutexSemList;
  o << endl << "*** Event Semaphores list:" << endl  << eventSemList;
  o << endl << "*** Queues list:" << endl  << queueList;

  dumpHandles( &o, isNowTotalFh );
}
