#include <share.h>
#include <fcntl.h>
#include <ctype.h>
#include <time.h>
#include <builtin.h>

#include "language.h"

#include "emsi.h"
#include "pair.h"
#include "timer.h"
#include "crc.h"
#include "misc.h"
#include "key.h"
#include "session.h"
#include "buffer.h"
#include "ea.h"
#include "rexxhook.h"

#define debugEmsi( X ) if( (miscInfo.state&EMSIDebug)==EMSIDebug ) log X

static volatile int oldEMSIrest = 0,
                    needIndicator = 1;

char bbsAtThisCall = 0,
     sessionProtected = 0,
     sessionListed = 0,
     wasBusyEncountred = 0;

QueueEntryList *illegalAka( NULL );

static Timer *EMSItimer;

static const char * const emsiStr[ 6 ] =
{
  "EMSI_REQ", "EMSI_INQ", "EMSI_DAT" ,
  "EMSI_HBT", "EMSI_ACK", "EMSI_NAK"
};

static Pair emsiTab[] =
{
  Pair( "EMSI_REQA77E", EMSI_REQ ),
  Pair( "EMSI_INQC816", EMSI_INQ ),
  Pair( "EMSI_HBTEAEE", EMSI_HBT ),
  Pair( "EMSI_ACKA490", EMSI_ACK ),
  Pair( "EMSI_NAKEEC3", EMSI_NAK ),
  Pair( "EMSI_DAT",     EMSI_DAT )
};

static int emsiHow = sizeof( emsiTab )/sizeof( Pair );

Pair *EMSIuserAction = NULL;

// =========================== writeEMSIDat ================================

void writeEMSIDat( Buffer& dat, int off, char *who )
{
  if( miscInfo.emsiLog==NULL )
    return;

  makeFilePath( miscInfo.emsiLog );

  FILE *to( _fsopen( miscInfo.emsiLog, "a" ) );

  if( to==NULL )
  {
    log( Msg_ErrUnableOpen, miscInfo.emsiLog, strError() );
    return;
  }

  time_t tim = time( NULL );
  struct tm *tm = localTime( &tim );

  fprintf( to, "%02d/%02d/%02d %02d:%02d:%02d [%d] %4s `",
	   tm->tm_mday, tm->tm_mon+1, tm->tm_year%100,
	   tm->tm_hour, tm->tm_min, tm->tm_sec,
	   miscInfo.taskNumber, who );

  for( register int ch( off ); ch<dat.length(); ch++ )
    if( dat.str[ch]!='\r' )
      fputc( dat.str[ch], to );
    else
      break;

  fputs( "'\n", to );
  fclose( to );
  setExtendedAttribute( miscInfo.emsiLog, ".TYPE", Msg_EAEMSILog );
}

// ============================== restEMSI ===============================

static char *currentEmsiMode = Msg_ScrInfoNone;

void restEMSI()
{
  int rest( EMSItimer->isTimeOut() ? 0 : int(EMSItimer->rest()/1000) );

  if( rest!=oldEMSIrest && needIndicator )
  {
    char *buf = (char *)_alloca( strlen(currentEmsiMode)+
                                 strlen(Msg_ScrInfoSecs)+16 );

    oldEMSIrest = rest;
    sprintf( buf, "%s, %d %s", currentEmsiMode, rest, Msg_ScrInfoSecs );
    showAction( buf );
  }

  if( abortAction( 0 ) )
  {
    showAction( Msg_ScrEmsiAbort );
    log( Msg_LogEmsiAborted );
    comPort->hangUp();
    needIndicator = 0;
  }

}

// ============================= showEmsi ===============================

void showEmsi( const char * const str )
{
  if( str!=NULL && *str )
    currentEmsiMode = (char *)str;
  else
    currentEmsiMode = Msg_ScrInfoNone;

  restEMSI();
}

// =========================== emsiFormPacket ============================

void emsiFormPacket( char *buf, EMSI_PACKETS type )
{
  strcpy( buf, "**" );
  strcat( buf, emsiStr[ int(type) ] );
  strcat( buf, stringCRC( buf+2 ) );
  strcat( buf, "\r" );
}

// ==================== emsiHandShakeReceiveDATbody =======================

int emsiHandShakeReceiveDATbody( Buffer **answerEmsi, long left )
{
  debugEmsi( (Msg_ScrEmsiDeb1) );
  char lenStr[5];
  Timer t( left );

  for( register int i(0); !t.isTimeOut()&&comPort->isOnline()&&i<4; )
  {
    int ch( comPort->get() );

    if( ch!=-1 )
      lenStr[i++] = char(ch);

    restEMSI();
  }

  lenStr[i] = 0;

  if( i!=4 )
  {
    debugEmsi( (Msg_ScrEmsiDeb2) );
    return t.isTimeOut() ? 1 : 3;
  }

  char *endPtr;
  int len = int( strtoul( lenStr, &endPtr, 16) );

  if( *endPtr )
  {
    debugEmsi( (Msg_ScrEmsiDeb3) );
    return t.isTimeOut() ? 1 : 3;
  }

  Buffer *buf( new Buffer );

  buf->putString( (char *)emsiStr[ int(EMSI_DAT) ] );
  buf->putString( lenStr );

  for( i = 0; !t.isTimeOut()&&comPort->isOnline()&&i<len+4; )
  {
    int ch( comPort->get() );

    if( ch!=-1 )
    {
      buf->putInto( char(ch) );
      i++;
    }

    restEMSI();
  }

  if( i!=len+4 )
  {
    debugEmsi( (Msg_ScrEmsiDeb4, len+4, i) );
    delete buf;
    return t.isTimeOut() ? 1 : 3;
  }

  writeEMSIDat( *buf, 0, "From" );

  len = buf->length()-4;
  strncpy( lenStr, buf->str+len, 4 );
  lenStr[4] = 0;

  buf->str[ len ] = 0;
  endPtr = stringCRC( buf->str );

  if( stricmp( lenStr, endPtr ) )
  {
    debugEmsi( (Msg_ScrEmsiDeb5, lenStr, endPtr) );
    delete buf;
    return t.isTimeOut() ? 1 : 3;
  }

  debugEmsi( (Msg_ScrEmsiDeb6) );

// rexx hook: received emsi

  char *newEmsi = processRexxHook( "FromEMSI_DAT", buf->str, NULL, NULL,
                                   NULL, NULL, NULL, 1 );

  if( newEmsi!=NULL && *newEmsi )
  {
    buf->reset();
    buf->putString( newEmsi );
    writeEMSIDat( *buf, 0, "UFrm" );
    buf->putInto( 0 );
    buf->putInto( 0 );
    buf->putInto( 0 );
  }

  delete newEmsi;

  buf->stripByte();
  buf->stripByte();
  buf->stripByte();
  buf->stripByte();
  *answerEmsi = buf;

  return 0;
}

// ========================= locateEMSIpacket ============================

EMSI_PACKETS locateEMSIpacket( long restEmsiTime, int oneTable,
			       Buffer **answerEmsi )
{
//  debugEmsi( ("EMSI: locateEMSIpacket started") );

  for( int j(0); j<emsiHow; j++ )
    emsiTab[j].reset();

  if( !oneTable )
    for( int j(0); j<handShakeCount; j++ )
      handShakeList[j].reset();

  Timer left( restEmsiTime );

  while( !left.isTimeOut() && comPort->isOnline() )
  {
    int ch( comPort->get() );

    if( ch!=-1 )
    {
      for( int i(0); i<emsiHow; i++ )
	if( emsiTab[i].compare( (char)ch ) == Pair::Completed )
	{
	  if( emsiTab[i].getEmsi()==EMSI_DAT )
	  {
	    Buffer *buf( NULL );
	    int rc = emsiHandShakeReceiveDATbody( &buf, left.rest() );

	    if( rc || answerEmsi==NULL )
	    {
	      if( buf!=NULL )
		delete buf;

	      return EMSI_ERR;
	    }
	    else
	      *answerEmsi = buf;
	  }

	  return emsiTab[i].getEmsi();
	}

      if( !oneTable )
	for( int i(0); i<handShakeCount; i++ )
	  if( handShakeList[i].compare( (char)ch ) == Pair::Completed )
	  {
	    EMSIuserAction = handShakeList+i;
	    return EMSI_USER;
	  }
    }

    restEMSI();
  }

  return EMSI_ERR;
}

// ============================ emsiSendPacket ===========================

int emsiSendPacket( EMSI_PACKETS type, long time )
{
  char buf[ 16 ];
  emsiFormPacket( buf, type );

  int rc;

  if( rc=comPort->putOnLine( buf, 15, time ), rc )
  {
    debugEmsi( (Msg_ScrEmsiDeb7, buf, rc) );
    return rc;
  }

  debugEmsi( (Msg_ScrEmsiDeb8, buf) );
  return 0;
}

// ========================== emsiHandShakeInitAnswer ======================

int emsiHandShakeInitAnswer( long time )
{
  showEmsi( Msg_ScrEmsiModeA1 );
  debugEmsi( (Msg_ScrEmsiDeb9) );

  Timer t( time );
  int rc;

  char banner[256];

  strcpy( banner, "\r\n\n" );
  strcat( banner, M_NAME );
  strcat( banner, " Version " );
  strcat( banner, M_VERSION );
  strcat( banner, ", serial #" );
  strcat( banner, M_NUMBER );
  strcat( banner, ".\r\n" );
  strcat( banner, M_COPYRIGHT );
  strcat( banner, "\r\n\n" );

/*
  extern char *todayStrings[];
  int offInTodaystring( todayMessage( 5, 2, 2 ) );

  if( offInTodaystring )
  {
    if( rc=comPort->putOnLine( banner, 3, t.rest() ), rc )
      return rc;

    if( rc=comPort->putOnLine( todayStrings[ offInTodaystring ],
			      strlen( todayStrings[ offInTodaystring ] ),
			      t.rest() ), rc )
      return rc;
  }
*/

  if( rc=comPort->putOnLine( banner, strlen(banner), t.rest() ), rc )
    return rc;

  extern char *ownerName;

  Buffer owner;

  if( ownerName!=NULL )
  {
    owner.putString( Msg_RegisterdTo );
    owner.stripByte();
    owner.stripByte();
    owner.stripByte();
    owner.putString( ownerName );
  }
  else
  {
    owner.putString( Msg_RegisterdNone );
    owner.stripByte();
  }

  owner.putInto( 0 );

  if( rc=comPort->putOnLine( owner.str, strlen(owner.str), t.rest() ), rc )
    return rc;

  if( rc=comPort->putOnLine( banner, 3, t.rest() ), rc )
    return rc;

/*
  char address[65];

  defaultAddress->toString( address, 65, FIDOAddressShowModeEMSI );

  rc=comPort->putOnLine( address, strlen(address), t.rest() );

  if( rc )
    return rc;

  char sep[]( ", " );

  if( rc=comPort->putOnLine( sep, sizeof(sep)-1, t.rest() ), rc )
    return rc;

  if( rc=comPort->putOnLine( systemInfo.name, strlen(systemInfo.name), t.rest() ), rc )
    return rc;

  if( rc=comPort->putOnLine( sep, sizeof(sep)-1, t.rest() ), rc )
    return rc;

  if( rc=comPort->putOnLine( systemInfo.sysopName, strlen(systemInfo.sysopName), t.rest() ), rc )
    return rc;

  if( rc=comPort->putOnLine( banner, 3, t.rest() ), rc )
    return rc;
*/

  if( processBBS==NULL || (miscInfo.state1&BBSenable)!=BBSenable ||
      miscInfo.BBSTime==NULL )
  {
    bbsAtThisCall = 0;

    if( rc=comPort->putMessage( bbsInfo.mailOnly ), rc )
      return rc;
  }
  else if( miscInfo.BBSTime!=NULL && !miscInfo.BBSTime->isWorked() )
  {
    bbsAtThisCall = 0;

    if( rc=comPort->putMessage( bbsInfo.mailNow ), rc )
      return rc;
  }
  else if( comPort->getLastBaud() < miscInfo.minBBSBaud )
  {
    bbsAtThisCall = 0;

    if( rc=comPort->putMessage( bbsInfo.slowModem ), rc )
      return rc;
  }
  else
  {
    bbsAtThisCall = 1;

    if( rc=comPort->putMessage( bbsInfo.twiceEsc ), rc )
      return rc;
  }

  if( rc=comPort->putOnLine( banner, 3, t.rest() ), rc )
    return rc;

  while( !t.isTimeOut()&&comPort->isOnline() )
  {
    if( rc=emsiSendPacket( EMSI_REQ, t.rest() ), rc )
      return rc;

    Timer t1( 5000 );
    EMSI_PACKETS pak = locateEMSIpacket( minTimer( t, t1 ), 0, NULL );

    if( pak == EMSI_INQ )
      return 0;

    if( pak == EMSI_USER )
      return 255;

    restEMSI();
  }

  if( t.isTimeOut() )
    if( (miscInfo.state&AutoBBS) == AutoBBS && bbsAtThisCall )
    {
      log( Msg_InfoAutoLoadingBBS );
      EMSIuserAction = handShakeList;
      return 255;
    }
    else
      return 1;

  return 3;
}

// ======================== emsiHandShakeInitCall ===========================

int emsiHandShakeInitCall( long time )
{
  showEmsi( Msg_ScrEmsiModeC1 );
  debugEmsi( (Msg_ScrEmsiDebA) );

  Timer t( time );
  int rc;

  debugEmsi( (Msg_ScrEmsiDebB) );
  while( !t.isTimeOut() && !comPort->haveInput() && comPort->isOnline() )
  {
    if( rc=comPort->putOnLine( " \r", 2, t.rest() ), rc )
      return rc;

    restEMSI();

    if( !comPort->haveInput() )
      sleep( 1000 );
    else
      break;

    restEMSI();
  }

  while( !t.isTimeOut()&&comPort->isOnline() )
  {
    Timer t1( 5000 );
    EMSI_PACKETS pak = locateEMSIpacket( minTimer( t, t1 ), 0, NULL );

    if( pak == EMSI_USER )
      return 255;

    if( pak == EMSI_REQ )
      if( rc=emsiSendPacket( EMSI_INQ, t.rest() ), rc )
	return rc;
      else
	return 0;

    for( int i(0); i<2; i++ )
      if( rc=emsiSendPacket( EMSI_INQ, t.rest() ), rc )
        return rc;

    if( rc=comPort->putOnLine( "\r", 1, t.rest() ), rc )
      return rc;

    restEMSI();
  }

  return t.isTimeOut() ? 1 : 3;
}

// ======================= emsiHandShakeTransmitDat =======================

static char firstAKAmessage = 1;

int emsiHandShakeTransmitDat( long left, Compatibility prot,
			    long netTraf, long xTraf,
			    char *password, FIDOAddressList *akas )
{

  ourMail = netTraf;
  ourFiles = xTraf;

  debugEmsi( (Msg_ScrEmsiDebC) );
  Buffer b( 320, 80 );

  b.putString( "**EMSI_DAT----" );
  b.putString( "{EMSI}{" );

  char address[65],
       needInfo( 1 );

  if( akas==NULL )
  {
    needInfo = 0;
    akas = systemInfo.addr;
  }

  Buffer info;

  for( FIDOAddressIterator list( *akas ); list; list++ )
  {
    FIDOAddress& item( *(FIDOAddress *)&(list()) );

    item.toString( address, sizeof(address), FIDOAddressShowModeEMSI );
    b.putEMSI( address );
    b.putInto( ' ' );

    if( needInfo )
    {
      info.putEMSI( address );
      info.putInto( ',' );
      info.putInto( ' ' );
    }
  }

  if( needInfo )
  {
    info.stripByte();
    info.stripByte();
    info.putInto( 0 );

    if( firstAKAmessage )
    {
      log( Msg_ScrEmsiSubstAKA, info.str );
      firstAKAmessage = 0;
    }
  }

  b.stripByte();

  b.putString( "}{" );

  if( password!=NULL && illegalAka==NULL )
    b.putEMSI( password );

  b.putString( "}{8N1" );

  if( (prot&PUA)==PUA )
    b.putString( ",PUA" );

  if( (prot&PUP)==PUP )
    b.putString( ",PUP" );

  if( (prot&NPU)==NPU )
    b.putString( ",NPU" );

  if( (prot&HAT)==HAT )
    b.putString( ",HAT" );

  if( (prot&HXT)==HXT )
    b.putString( ",HXT" );

  if( (prot&HRQ)==HRQ )
    b.putString( ",HRQ" );

  if( (prot&HYD)==HYD )
    b.putString( ",RH1" );

  b.putString("}{" );

  if( (prot&HYD)==HYD )
    b.putString( "HYD," );

  if( (prot&JAN)==JAN )
    b.putString( "JAN," );

  if( (prot&DZA)==DZA )
    b.putString( "DZA," );

  if( (prot&ZAP)==ZAP )
    b.putString( "ZAP," );

  if( (prot&ZMO)==ZMO )
    b.putString( "ZMO," );

  if( (prot&NCP)==NCP )
    b.putString( "NCP," );

  if( bbsAtThisCall )
    b.putString( "BBS," );

  b.stripByte();

  if( (prot&NRQ)==NRQ )
    b.putString( ",NRQ" );

  if( (prot&ARC)==ARC )
    b.putString( ",ARC" );

  if( (prot&XMA)==XMA )
    b.putString( ",XMA" );

  if( (prot&FNC)==FNC )
    b.putString( ",FNC" );

  char buf[64];

  sprintf( buf, "}{%x}", M_CODE );
  b.putString(  buf );

  b.putInto( '{' );
  b.putEMSI( (char *)M_NAME );
  b.putInto( '}' );

  b.putInto( '{' );
  b.putEMSI( (char *)M_VERSION );
  b.putInto( '}' );

  b.putInto( '{' );
  b.putEMSI( (char *)M_NUMBER );
  b.putInto( '}' );

  extern char *ownerName;
  extern char isRegisterdVersion;

  if( isRegisterdVersion && ownerName!=NULL )
  {
    b.putString( "{TB!REG}{[" );
    b.putEMSI( ownerName );
    b.putString( "]}" );
  }

  sprintf( buf, "%lx", comPort->getTranx()-getTimeZone() );
  b.putString( "{TRX#}{[" );
  b.putEMSI( buf );
  b.putString( "]}" );

  long myTZ( getTimeZone()/60 );

  sprintf( buf, "%ld", myTZ );
  b.putString( "{TZ}{[" );
  b.putEMSI( buf );
  b.putString( "]}" );

  sprintf( buf, "%c%02d%02d", myTZ>0 ? '-' : '+',
                              abs(int(myTZ))/60, abs(int(myTZ))%60 );
  b.putString( "{TZUTC}{[" );
  b.putEMSI( buf );
  b.putString( "]}" );

  sprintf( buf, "%lx %lx", netTraf, xTraf );
  b.putString( "{TRAF}{" );
  b.putEMSI( buf );
  b.putString( "}" );

  sprintf( buf, "%lx", netTraf+xTraf );
  b.putString( "{MOH#}{[" );
  b.putEMSI( buf );
  b.putString( "]}" );

  extern char *todayStrings[];
  int who = todayMessage( 5, 2, 2 );

  if( who )
  {
    b.putString( "{YooHoo!}{[" );
    b.putEMSI( (char *)todayStrings[ who ] );
    b.putString( "]}" );
  }

  b.putString( "{IDENT}{[" );
  b.putEMSI( systemInfo.name );

  char *II( "][" );

  b.putString( II );
  b.putEMSI( systemInfo.place );
  b.putString( II );
  b.putEMSI( systemInfo.sysopName );
  b.putString( II );
  b.putEMSI( systemInfo.phone );
  b.putString( II );
  b.putEMSI( systemInfo.baud );
  b.putString( II );
  b.putEMSI( systemInfo.flags );
  b.putString( "]}" );

  sprintf( buf, "%04X", ((unsigned short)b.length()-14)&0xffff );
  memcpy( b.str+10, buf, 4 );

  b.putInto( 0 );

  char *crc( stringCRC( b.str+2 ) );

  b.stripByte();
  b.putString( crc );
  b.putInto( '\r' );

  writeEMSIDat( b, 2, "To" );

// rexx hook: send emsi_dat

  b.putInto( 0 );

  char *newEmsi = processRexxHook( "ToEMSI_DAT", b.str, NULL, NULL,
                                     NULL, NULL, NULL, 1 );

  if( newEmsi!=NULL && *newEmsi )
  {
    b.reset();
    b.putString( newEmsi );
    writeEMSIDat( b, 2, "UTo" );
    b.putInto( 0 );
  }

  delete newEmsi;

  int rc;

  if( rc=comPort->putOnLine( b.str, b.length()-1, left ), rc )
  {
    debugEmsi( (Msg_ScrEmsiDebD, rc) );
    return rc;
  }

  restEMSI();

  debugEmsi( (Msg_ScrEmsiDebE) );
  return 0;
}

// =========================== emsiHandShakeTransmit =====================

int emsiHandShakeTransmit( long netTraf, long xTraf,
			 Compatibility pro, char *password,
			 FIDOAddressList *akas )
{
  showEmsi( Msg_ScrEmsiModeT2 );
  debugEmsi( (Msg_ScrEmsiDebF) );

  firstAKAmessage = 1;

  int rc;
  int tries( 0 );
  Timer t1( EMSItimer->rest() );
  Timer t2( 20000 );
  int againReq(0);

step_2:

  restEMSI();

  if( !comPort->isOnline() )
  {
    debugEmsi( (Msg_ScrEmsiDeb_NoCD) );
    return 3;
  }

  if( ++tries>6 )
  {
    debugEmsi( (Msg_ScrEmsiDebG) );
    return 250;
  }

  rc = emsiHandShakeTransmitDat( t1.rest(), pro, netTraf, xTraf,
                                 password, akas );
  
  if( rc )
    return rc;

  t2.startTimer( 20000 );

step_4:

  restEMSI();

  if( !comPort->isOnline() )
  {
    debugEmsi( (Msg_ScrEmsiDeb_NoCD) );
    return 3;
  }

  EMSI_PACKETS pak = locateEMSIpacket( minTimer( t1, t2 ), 1, NULL );

  if( pak == EMSI_ACK )
  {
    debugEmsi( (Msg_ScrEmsiDebH) );
    return 0;
  }

  restEMSI();

  if( t1.isTimeOut() )
  {
    debugEmsi( (Msg_ScrEmsiDebI) );
    return 1;
  }

  if( t2.isTimeOut() )
  {
    debugEmsi( (Msg_ScrEmsiDebJ) );
    goto step_2;
  }

  if( pak==EMSI_REQ || pak==EMSI_HBT )
  {
    if( pak==EMSI_REQ && ++againReq>2 )
    {
      debugEmsi( (Msg_ScrEmsiDebP) );
      rc=emsiSendPacket( EMSI_INQ, minTimer( t1, t2 ) );
      
      if( rc )
	return rc;
    }
    
    debugEmsi( (Msg_ScrEmsiDebK) );
    goto step_4;
  }

  debugEmsi( (Msg_ScrEmsiDebO, (int)pak) );

  if( pak==EMSI_ERR )
  {
    Timer t;

    t.sleep( 1 );
  }

  goto step_2;
}

// ========================== emsiHandShakeReceive =========================

int emsiHandShakeReceive( Buffer **answerEmsi, int answer )
{
  showEmsi( Msg_ScrEmsiModeR2 );
  debugEmsi( (Msg_ScrEmsiDebL) );

  *answerEmsi = NULL;

  int tries( 0 );
  Timer t1( 20000l ), t2( EMSItimer->rest() );
  int rc;

step_2:

  if( !comPort->isOnline() )
  {
    debugEmsi( (Msg_ScrEmsiDeb_NoCD) );
    return 3;
  }

  if( ++tries>6 )
  {
    debugEmsi( (Msg_ScrEmsiDebG) );
    return 250;
  }

  if( answer )
  {
//    answer = 0;         // new hit!

    if( rc=emsiSendPacket( EMSI_REQ, t2.rest() ), rc )
      return rc;

    goto step_3;
  }

  restEMSI();

  if( tries>1 )
  {
    if( rc=emsiSendPacket( EMSI_NAK, t2.rest() ), rc )
      return rc;

    goto step_3;
  }

  goto step_4;

step_3:

  if( !comPort->isOnline() )
  {
    debugEmsi( (Msg_ScrEmsiDeb_NoCD) );
    return 3;
  }

  t1.startTimer( 20000l );

step_4:

  if( !comPort->isOnline() )
  {
    debugEmsi( (Msg_ScrEmsiDeb_NoCD) );
    return 3;
  }

  restEMSI();

  EMSI_PACKETS pak = locateEMSIpacket( minTimer( t1, t2 ), 1, answerEmsi );

  if( pak==EMSI_DAT )
    goto step_5;

  if( t2.isTimeOut() )
  {
    debugEmsi( (Msg_ScrEmsiDebJ) );
    return 1;
  }

  if( t1.isTimeOut() )
  {
    debugEmsi( (Msg_ScrEmsiDebI) );
    goto step_2;
  }

  if( pak==EMSI_HBT )
  {
    debugEmsi( (Msg_ScrEmsiDebM) );
    goto step_3;
  }

  if( pak==EMSI_ERR )
  {
    debugEmsi( (Msg_ScrEmsiDebN) );
    goto step_2;
  }

  restEMSI();

  debugEmsi( (Msg_ScrEmsiDebO, (int)pak) );

  if( pak==EMSI_ERR )
  {
    Timer t;

    t.sleep( 1 );
  }

  goto step_4;

step_5:

  for( int i(0); i<2; i++ )
    if( rc=emsiSendPacket( EMSI_ACK, minTimer( t1, t2 ) ), rc )
      return rc;

  restEMSI();

  debugEmsi( (Msg_ScrEmsiDebH) );
  return 0;
}

// ========================== comparePassword ===========================

Boolean comparePassword( char *his, char *my )
{
  if( my==NULL )
    return True;  // not expected

  // expected!

  if( his==NULL )
    return False;

  return stricmp( his, my ) ? False : True;
}

// ========================== getEmsiItem ===============================

enum ItemIs { ItemOK=0, NoMore=1, ItemError=2 };

inline int hexToBin( char ch )
{
  return ch > '9' ? toupper(ch)-'A'+10 : ch-'0';
}

void removeEscaped( char *s )
{
  if( s==NULL )
    return;

  int sum = 0,
      slash = 0;

  for( register char *n( s ); *s; s++ )
    if( *s=='\\' )
      if( s[1]=='\\' )
      {
        *(n++) = '\\';
        s++;
      }
      else
      {
        if( slash )
          *(n++) = char( sum );

        slash = 1;
        sum = 0;
      }
    else
      if( !slash )
        *(n++) = *s;
      else
        if( !isxdigit( toupper( *s ) ) )
        {
          if( !sum )
            *(n++) = '?';
          else
            *(n++) = char( sum );

          *(n++) = *s;
          slash = 0;
        }
        else
          if( slash==3 )
          {
            *(n++) = char( sum );
            slash = 0;
          }
          else
          {
            slash++;
            sum = sum*16 + hexToBin( *s );
          }

  if( slash )
    if( sum )
      *(n++) = char( sum );
    else
      *(n++) = '?';

  *(n++) = 0;
  *n = 0;
}

ItemIs getEmsiItem( char **strIni, char end, char beg,
				       char **item )
{
  if( *item!=NULL )
  {
    delete *item;
    *item = NULL;
  }

  register char *str( *strIni );

  if( !*str )
  {
    debugEmsi( (Msg_ScrEmsiItemDeb1) );
    return NoMore;
  }

  if( *(str++)!=beg )
  {
    debugEmsi( (Msg_ScrEmsiItemDeb2, beg, --str) );
    return ItemError;
  }

  Buffer ret;
  ItemIs rc( NoMore );

  while( *str && rc!=ItemOK )
  {
    if( *str==end )
      if( str[1]==end )
      {
        ret.putInto( end );
        str++;
      }
      else
        rc = ItemOK;
    else
      ret.putInto( *str );
      
    str++;
  }

  ret.putInto( 0 );
  ret.putInto( 0 );        //    !!

  if( rc==ItemOK || ret.length()>2 )
  {
    *item = new char [ ret.length() ];
    memcpy( *item, ret.str, ret.length() );
    rc = ItemOK;
  }

  if( *item!=NULL )
    debugEmsi( (Msg_ScrEmsiItemDeb3, *item, beg, end, rc) );
  else
    debugEmsi( (Msg_ScrEmsiItemDeb4) );

  *strIni = str;
  return rc;
}

// =============================== itemPrepare ==========================

void itemPrepare( char *str, char ch )
{
  for( int i( 0 ); str[i]; i++ )
    if( str[i]==ch )
      str[i] = 0;
}

// ============================ processAka ==============================

void processAka( register char *str, QueueEntryList *akas,
		        QueueEntry *called, char *hisPassword,
	                int answer, char **password )
{
  char buf[65];
  const long HBTtime( 5000l );
  Timer scanTimer( HBTtime );

  sessionListed = 0;
  sessionProtected = 0;

  for( itemPrepare( str, 0x20 ); *str; str+=strlen(str)+1 )
  {
    uShort zone( 0 ), net( 0 ), node( 0 ), point( 0 );
    char *domain( NULL );

    if( !checkAddress( str, zone, net, node, point, &domain ) )
      log( Msg_LogAddrIgnWrong, str );
    else
    {
      FIDOAddress *addr( new FIDOAddress( zone, net, node, point, domain ) );
      NodeConfig *addNode( new NodeConfig( addr ) ),
                 *found( systemList->Find( *addNode ) );
      QueueEntry *cur( new QueueEntry( *addNode ) );

      if( found!=NULL )
      {
	delete cur;
	cur = new QueueEntry( *found );
      }

      QueueEntry *foundCur = queue->Find( *cur );

      if( foundCur!=NULL )
      {
	delete cur;
	cur = new QueueEntry( *foundCur );
      }

      char rescan( 1 ),
           remove( 0 );

      cur->toString( buf, sizeof(buf) );

      if( *password==NULL )
        *password = newStr( cur->getPassword() );

      if( !cur->isListed() )
        log( Msg_LogUnkAddr, buf );
      else
        sessionListed = 1;

      if( answer )                     // ஢ઠ ஫
      {
        if( cur->getPassword()!=NULL )
          if( comparePassword( hisPassword, cur->getPassword() ) == False )
          {
            log( Msg_LogIllPasswd, buf,
                 cur->getPassword()==NULL ? Msg_LogWordNone : cur->getPassword(),
	         hisPassword==NULL ? Msg_LogWordNone : hisPassword );

  	    if( illegalAka==NULL )
	      illegalAka = new QueueEntryList;

	    illegalAka->AddAtTail( *cur );
	    rescan = 0;
          }
          else
            if( hisPassword!=NULL && *hisPassword )
               sessionProtected = 1;
      }
      else
        if( password!=NULL && *password )
	  sessionProtected = 1;

      if( rescan && cur->isOurs() )
      {
        log( Msg_LogAddrIgnOur, buf );
        rescan = 0;
        remove = 1;
      }

      if( rescan )
      {
        int rc( 0 );

        if( called==NULL || ( called!=NULL && !((*called)==(*cur)) ) )
          rc = cur->makeBsy( 1 );

        if( rc )
        {
          log( Msg_LogAddrIgnBusy, buf );
          wasBusyEncountred = 1;
          rescan = 0;
          remove = 1;
        }
      }

      if( rescan )
      {
        akas->AddAtTail( *cur );

        if( (miscInfo.state3&ScanOnLine)==ScanOnLine )
          cur->rescan();
      }

      if( remove )
        delete cur;

      delete addNode;
      delete addr;
    }

    if( scanTimer.isTimeOut() && answer )
    {
      emsiSendPacket( EMSI_HBT, EMSItimer->rest() );
      scanTimer.startTimer( HBTtime );
    }
  }
}

// =========================== processLink ==============================

void processLink( register char *str, Compatibility& hisCode )
{
  for( itemPrepare( str, ',' ); *str; str+=strlen(str)+1 )
    if( stricmp( str, "8N1" ) )
      if( !stricmp( str, "PUA" ) )
	hisCode = Compatibility( hisCode|PUA );
      else if( !stricmp( str, "PUP" ) )
	hisCode = Compatibility( hisCode|PUP );
      else if( !stricmp( str, "NPU" ) )
        hisCode = Compatibility( hisCode|NPU );
      else if( !stricmp( str, "HAT" ) )
        hisCode = Compatibility( hisCode|HAT );
      else if( !stricmp( str, "HXT" ) )
        hisCode = Compatibility( hisCode|HXT );
      else if( !stricmp( str, "HRQ" ) )
        hisCode = Compatibility( hisCode|HRQ );
      else if( !stricmp( str, "RH1" ) )
        hisCode = Compatibility( hisCode|RH1 );
      else
        log( Msg_LogIllFlagIgn, str );
}

// ======================== processCompatibility ===========================

void processCompatibility( register char *str, Compatibility& hisCode )
{
  for( itemPrepare( str, ',' ); *str; str+=strlen(str)+1 )
    if( !stricmp( str, "ZMO" ) )
      hisCode = Compatibility( hisCode|ZMO );
    else if( !stricmp( str, "ZAP" ) )
      hisCode = Compatibility( hisCode|ZAP );
    else if( !stricmp( str, "DZA" ) )
      hisCode = Compatibility( hisCode|DZA );
    else if( !stricmp( str, "JAN" ) )
      hisCode = Compatibility( hisCode|JAN );
    else if( !stricmp( str, "NCP" ) )
      hisCode = Compatibility( hisCode|NCP );
    else if( !stricmp( str, "NRQ" ) )
      hisCode = Compatibility( hisCode|NRQ );
    else if( !stricmp( str, "ARC" ) )
      hisCode = Compatibility( hisCode|ARC );
    else if( !stricmp( str, "XMA" ) )
      hisCode = Compatibility( hisCode|XMA );
    else if( !stricmp( str, "FNC" ) )
      hisCode = Compatibility( hisCode|FNC );
    else if( !stricmp( str, "HYD" ) )
      hisCode = Compatibility( hisCode|HYD );
    else if( !stricmp( str, "BBS" ) )
      log( Msg_LogBBSAvail );
    else
      log( Msg_LogIllCompatIgn, str );
}

// ============================ showSesStr ==============================

void showSesStr( char *msg, int line )
{
  register wid = wndQueue.getWidth()-2;

  if( strlen(msg)>wid )
    msg[ wid ]=0;

  wndQueue = stdColor;
  wndQueue.centre( line+((miscInfo.state1&StatusLine)==StatusLine ? 0 : 1),
                   msg );
}

// ============================= emsiParse ==============================

/*

    EMSI_DAT
	fingerprint,            "EMSI"
	system_address_list,
	password,
	link_codes,
	compatibility_codes,
	mailer_product_code,
	mailer_name,
	mailer_version,
	mailer_serial_number:    ASCII1;
	extra_field_1,
	    ..
	    ..
	extra_field_n:            EMSI_addon; (optional fields)
    end;

    The EMSI_addon structure is defined as follows:

    EMSI_addon
	product_ID,
	specific_data:            ASCII1;
    end;

    PRODUCT_IDENTIFIER   :  IDENT
    SPECIFIC_DATA        :  system_name,
			    city,
			    operator_name,
			    phone_number,
			    baud_rate,
			    flags:            ASCII2;

    {EMSI}{2:270/17}{}{8N1,PUA}{ZAP,ZMO,ARC,XMA}{44}{AirMail}{0.10}
    {Beta-2}{IDENT}{[Advanced Engineering S.A.R.L.][Luxembourg]
    [Joaquim Homrighausen][-Unpublished-][9600][MO,XA,HST,V32B,V42B]}

*/

int emsiParse( Buffer& answerEmsi,
		      Compatibility& hisCode,
		      QueueEntryList **akaList,
		      int answer,
		      QueueEntry *called,
		      char **password )
{
  QueueEntryList *akas( new QueueEntryList );

  *akaList = akas;
  hisCode = Compatibility( 0 );

  char *str( answerEmsi.str );
  int len( answerEmsi.length() );

  for( register int i( 0 ); i<len; i++ )
    if( int(str[i])<0x20 || int(str[i])>127  )
      str[i] = '?';

  answerEmsi.putInto( 0 );

  str += 12;

  char *item( NULL );
  int rc = 0;
  char *msg = Msg_LogEmsiExpEMSI;

  if( strnicmp( str, "{EMSI}{", 7 ) )
  {
Error:
    rc = 250;
    log( Msg_LogIsNotEmsiDat, answerEmsi.str );
    log( msg );

OkComplete:
    delete item;
    return rc;
  }

  str += 6;

  if( getEmsiItem( &str, '}', '{', &item )!=ItemOK || item==NULL )
  {
    msg = Msg_LogEmsiExpAKA;
    goto Error;
  }

  removeEscaped( item );

  char *akaField = (char *)_alloca( strlen(item)+2 );

  strcpy( akaField, item );
  akaField[ strlen( akaField )+1 ] = 0;

  if( getEmsiItem( &str, '}', '{', &item )!=ItemOK )
  {
    msg = Msg_LogEmsiExpPasswd;
    goto Error;
  }

  removeEscaped( item );
  hisPassword = item;
  item = NULL;

  if( hisPassword!=NULL && !strlen( hisPassword ) )
  {
    delete hisPassword;
    hisPassword = NULL;
  }

  processAka( akaField, akas, called, hisPassword, answer, password );

  if( getEmsiItem( &str, '}', '{', &item )!=ItemOK || item==NULL )
  {
    msg = Msg_LogEmsiExpLink;
    goto Error;
  }

  removeEscaped( item );
  processLink( item, hisCode );

  if( getEmsiItem( &str, '}', '{', &item )!=ItemOK || item==NULL )
  {
    msg = Msg_LogEmsiExpCompat;
    goto Error;
  }

  removeEscaped( item );
  processCompatibility( item, hisCode );

  if( getEmsiItem( &str, '}', '{', &item )!=ItemOK || item==NULL )
  {
    msg = Msg_LogEmsiExpFTSCCode;
    goto Error;
  }

  removeEscaped( item );
  mailerProductCode = item;
  remoteMailerCode = int( strtoul( mailerProductCode, &item, 16 ) );
  item = NULL;

  if( getEmsiItem( &str, '}', '{', &item )!=ItemOK || item==NULL )
  {
    msg = Msg_LogEmsiExpMName;
    goto Error;
  }

  removeEscaped( item );
  mailerName = item;
  item = NULL;

  if( getEmsiItem( &str, '}', '{', &item )!=ItemOK || item==NULL )
  {
    msg = Msg_LogEmsiExpMVer;
    goto Error;
  }

  removeEscaped( item );
  mailerVersion = item;
  item = NULL;

  if( getEmsiItem( &str, '}', '{', &item )!=ItemOK || item==NULL )
  {
    msg = Msg_LogEmsiExpMSerNo;
    goto Error;
  }

  removeEscaped( item );
  mailerSerialNumber = item;
  item = NULL;

  char *errName( NULL );

  hisTime = 0;
  hisMail = 0;
  hisFiles = 0;
  hisMOH = 0;
  hisTZ = 0;

  int presHisTZ( 0 ),
      presHisTZUTC( 0 );

  while( 1 )
  {
    if( getEmsiItem( &str, '}', '{', &item )!=ItemOK || item==NULL )
      break;

    removeEscaped( item );

    if( !stricmp( item, "IDENT" ) )
      if( hisName!=NULL )
      {
	errName = "IDENT";
Duplicate:
	log( Msg_LogEmsiDblAddon, errName );
      }
      else
      {
	char *arg( NULL ),
	     *posIni( NULL );

	if( getEmsiItem( &str, '}', '{', &posIni )==ItemOK && posIni!=NULL )
	{
	  char *pos( posIni );

	  if( getEmsiItem( &pos, ']', '[', &arg )==ItemOK && arg!=NULL )
	  {
            removeEscaped( arg );
	    hisName = arg;
	    arg = NULL;

	    if( getEmsiItem( &pos, ']', '[', &arg )==ItemOK && arg!=NULL )
	    {
              removeEscaped( arg );
	      hisPlace = arg;
	      arg = NULL;

	      if( getEmsiItem( &pos, ']', '[', &arg )==ItemOK && arg!=NULL )
	      {
                removeEscaped( arg );
		hisSysOp = arg;
		arg = NULL;

		if( getEmsiItem( &pos, ']', '[', &arg )==ItemOK && arg!=NULL )
		{
                  removeEscaped( arg );
		  hisPhone = arg;
		  arg = NULL;

		  if( getEmsiItem( &pos, ']', '[', &arg )==ItemOK && arg!=NULL )
		  {
                    removeEscaped( arg );
		    hisBaud = arg;
		    arg = NULL;

		    if( getEmsiItem( &pos, ']', '[', &arg )==ItemOK && arg!=NULL )
		    {
                      removeEscaped( arg );
		      hisFlags = arg;
		      arg = NULL;
		    }
		  }
		}
	      }
	    }
	  }

	}

	delete posIni;
	delete arg;
      }                                         // ident
    else if( !stricmp( item, "TRX#" ) )
      if( hisTime )
      {
        errName = "TRX#";
        goto Duplicate;
      }
      else
      {
        char *arg( NULL ),
             *q;

        if( getEmsiItem( &str, '}', '{', &arg )==ItemOK && arg!=NULL )
        {
          removeEscaped( arg );
          hisTime=strtoul( arg+1, &q, 16 );
        }

        delete arg;
      }                                        //trx#
    else if( !stricmp( item, "TRAF" ) )
      if( hisMail || hisFiles )
      {
        errName = "TRAF";
        goto Duplicate;
      }
      else
      {
        char *arg( NULL ),
             *q;

        if( getEmsiItem( &str, '}', '{', &arg )==ItemOK && arg!=NULL )
        {
          removeEscaped( arg );
          hisMail=strtoul( arg, &q, 16 );
          hisFiles=strtoul( q+1, &q, 16 );
        }

        delete arg;
      }                                          //traf
    else if( !stricmp( item, "MOH#" ) )
      if( hisMOH )
      {
        errName = "MOH#";
        goto Duplicate;
      }
      else
      {
        char *arg( NULL ),
             *q;

        if( getEmsiItem( &str, '}', '{', &arg )==ItemOK && arg!=NULL )
        {
          removeEscaped( arg );
          hisMOH=strtoul( arg+1, &q, 16 );
        }

        delete arg;
      }                                        //moh
    else if( !stricmp( item, "TZ" ) )
      if( presHisTZ )
      {
        errName = "TZ";
        goto Duplicate;
      }
      else
      {
        char *arg( NULL ),
             *q;

        if( getEmsiItem( &str, '}', '{', &arg )==ItemOK && arg!=NULL )
        {
          removeEscaped( arg );
          hisTZ=strtol( arg+1, &q, 10 );
          presHisTZ = 1;
          debugEmsi( (Msg_ScrEmsiDebQ, hisTZ) );
        }

        delete arg;
      }                                        //TZ
    else if( !stricmp( item, "TZUTC" ) )
      if( presHisTZUTC )
      {
        errName = "TZUTC";
        goto Duplicate;
      }
      else
      {
        char *arg( NULL );

        if( getEmsiItem( &str, '}', '{', &arg )==ItemOK && arg!=NULL )
        {
          removeEscaped( arg );

          char *p( arg+1 ),
               sign( '+' );

          if( *p == '+' || *p == '-' )
          {
            sign = *p;
            p++;
          }

          if( strlen( p ) != 5 || *arg != '[' )
IllegalTZUTC:
            log( Msg_LogEmsiIllUTCTZ, arg );
          else
          {
            char *q;

            hisTZ = strtoul( p, &q, 10 );
            debugEmsi( (Msg_ScrEmsiDebQ, hisTZ) );

            if( *q != ']' )
              goto IllegalTZUTC;

            hisTZ = hisTZ/100*60 + hisTZ%100;
            presHisTZUTC = 1;

            if( sign == '+' )
              hisTZ = -hisTZ;

            debugEmsi( (Msg_ScrEmsiDebQ, hisTZ) );
          }
        }

        delete arg;
      }                                        //TZUTC
    else if( !stricmp( item, "TB!REG" ) )
      if( mailerOwner!=NULL )
      {
        errName = "TB!REG";
        goto Duplicate;
      }
      else
      {
        if( getEmsiItem( &str, '}', '{', &mailerOwner )==ItemOK &&
            mailerOwner!=NULL )
        {
          strcpy( mailerOwner, mailerOwner+1 );
          mailerOwner[ strlen(mailerOwner)-1 ] = 0;
          removeEscaped( mailerOwner );
        }
      }                                        //TB!REG
    else
    {
      char *arg( NULL );

      if( getEmsiItem( &str, '}', '{', &arg )!=ItemOK )
      {
        delete arg;
        arg = NULL;
      }

      char *buf = (char *)_alloca( strlen(item)+16+
                                   ( arg==NULL ? 0 : strlen(arg) ) );

      strcpy( buf, "{" );
      strcat( buf, item );
      strcat( buf, "}" );

      if( arg!=NULL )
      {
        removeEscaped( arg );
        strcat( buf, "{" );
        strcat( buf, arg );
        strcat( buf, "}" );
      }

      log( Msg_LogEmsiIllAddon, buf );
      delete arg;
    }
  }

  delete item;
  item = NULL;

  if( *akaList==NULL || !(*akaList)->GetItemsInContainer() )
    goto OkComplete;

  const bufSize = 128;

  char b1[bufSize];
  Buffer buf;

  buf.putString( Msg_InfoSessWith );
  buf.putString( (**akaList)[0].toString( b1, bufSize ) );

  buf.putInto( 0 );
  log( buf.str );
  buf.reset();

  if( (*akaList)->GetItemsInContainer()>1 )
  {
    Buffer akaBuf;
    char first( 1 ),
         akaEmpty( 1 );

    akaBuf.putString( Msg_InfoSessAKA );
    buf.putString( Msg_InfoSessAKA );

    for( QueueEntryIterator list( **akaList ); list; list++ )
      if( first )
	first = 0;
      else
      {
	buf.putString( list().toString( b1, bufSize ) );
	buf.putString( ", " );
	akaBuf.putString( list().toString( b1, bufSize ) );
	akaBuf.putString( ", " );
        akaEmpty = 0;

        if( akaBuf.length()>40 )
        {
          akaBuf.stripByte();
          akaBuf.stripByte();
          akaBuf.putInto( 0 );
          log( akaBuf.str );
          akaBuf.reset();
          akaBuf.putString( Msg_InfoSessAKA );
          akaEmpty = 1;
        }
      }

    buf.stripByte();
    buf.stripByte();
    buf.putInto( 0 );

    if( !akaEmpty )
    {
      akaBuf.stripByte();
      akaBuf.stripByte();
      akaBuf.putInto( 0 );
      log( akaBuf.str );
    }
  }

  if( !buf.empty() )
    showSesStr( buf.str, 6 );

  buf.reset();
  buf.putString( (**akaList)[0].toString( b1, bufSize ) );
  buf.putString( Msg_InfoSessUsed );
  buf.putString( mailerName );

  if( *mailerVersion )
  {
    buf.putString( " " );
    buf.putString( mailerVersion );
  }

  if( *mailerSerialNumber )
  {
    buf.putString( "/" );
    buf.putString( mailerSerialNumber );
  }

  buf.putInto( 0 );

  showSesStr( buf.str, 3 );

  log( Msg_LogSessMailer, mailerName, mailerProductCode,
                          *mailerVersion ? " " : "",
                          mailerVersion,
                          *mailerSerialNumber ? "/" : "",
                          mailerSerialNumber,
                          mailerOwner!=NULL ? "/" : "",
                          mailerOwner!=NULL ? mailerOwner : "" );

  buf.reset();

  if( hisName!=NULL || hisPlace!=NULL || hisSysOp!=NULL )
  {
    if( hisName!=NULL )
    {
      buf.putString( hisName );
      buf.putString( ", " );
    }

    if( hisPlace!=NULL )
    {
      buf.putString( hisPlace );
      buf.putString( ", " );
    }

    if( hisSysOp!=NULL )
    {
      buf.putString( hisSysOp );
      buf.putString( ", " );
    }

    buf.stripByte();
    buf.stripByte();
    buf.putInto( 0 );

    log( Msg_LogSessSystem, buf.str );

    buf.stripByte();
    buf.putString( ", " );
  }

  if( hisPhone!=NULL || hisBaud!=NULL || hisFlags!=NULL )
  {
    log( Msg_LogSessPhone,
              hisPhone!=NULL ? " " : "",
              hisPhone!=NULL ? hisPhone : "",
              hisBaud!=NULL ? "," : "",
              hisBaud!=NULL ? hisBaud : "",
              hisFlags!=NULL ? "," : "",
              hisFlags!=NULL ? hisFlags : "" );
 
    if( hisPhone!=NULL )
    {
      buf.putString( hisPhone );
      buf.putString( ", " );
    }

    if( hisBaud!=NULL )
    {
      buf.putString( hisBaud );
      buf.putString( ", " );
    }

    if( hisFlags!=NULL )
    {
      buf.putString( hisFlags );
      buf.putString( ", " );
    }
  }

  if( buf.length()==0 )
    buf.putString( Msg_InfoSessNoInfo );

  buf.stripByte();
  buf.stripByte();

  buf.putInto( 0 );

  showSesStr( buf.str, 4 );

  if( hisTime || hisTZ )
  {
    char r[ 32 ] = {0};
    char t[ 32 ] = {0};

    if( hisTime )
    {
      struct tm *tt( gmTime( (long *)&hisTime ) );

      strftime( r, 31, "%a %d %b %H:%M:%S", tt );
    }

    if( hisTZ )
    {
      register char tz[2];

      if( hisTZ<0 )
        tz[0] = '-';
      else if( hisTZ>0 )
        tz[0] = '+';
      else
        tz[0] = 0;

      tz[1] = 0;

      sprintf( t, "TZ: %s%02d:%02u", tz, abs(int(hisTZ/60)), abs(int(hisTZ%60)) );
    }

    log( Msg_LogSessTime, r, (*r && *t) ? ", " : "", t );
  }

  if( hisMOH && !(hisMail || hisFiles) )
  {
    char *t = strNumber( hisMOH );

    log( Msg_LogTraffMOH, t );
    delete t;
  }

  if( hisMail || hisFiles )
  {
    char *t1 = strNumber( hisMail );
    char *t2 = strNumber( hisFiles );

    log( Msg_LogTraffDetail, t1, t2 );
    delete t1;
    delete t2;
  }

  if( hisMOH && !hisFiles && !hisMail )
    hisFiles = hisMOH;

  if( sessionProtected )
  {
    log( Msg_InfoSessProtect );
    showSesStr( Msg_InfoSessProtect, 5 );
  }
  else
    showSesStr( Msg_InfoSessNoProtect, 5 );

  goto OkComplete;
}

// ============================= answerMode ================================

int answerMode( QueueEntryList **akaList,
				   Compatibility myCode,
				   Compatibility& hisCode,
				   char **password )
{
  Buffer *answerEmsi( NULL );

  int rc( emsiHandShakeReceive( &answerEmsi, 1 ) );

  if( !rc )
    if( answerEmsi!=NULL )
    {
      rc = emsiParse( *answerEmsi, hisCode, akaList, 1, NULL, password );

      if( rc )
        goto Done;

      long netTraf( 0 ),
	   xTraf( 0 );
      char *password( NULL );

      Compatibility newCode = Compatibility( myCode & ~(HYD|JAN|DZA|ZAP|ZMO) ),
		    code( NCP );

      FIDOAddressList *myAkaList( NULL );
      char noDZA( 0 );
      char noZAP( 0 );
      char noHydra( 0 );

      if( *akaList!=NULL )
	for( QueueEntryIterator list( **akaList ); list; list++ )
	{
	  QueueEntry *cur = &list();

	  netTraf += cur->getMail();
	  xTraf += cur->getFiles();

/*
    FIDOAddress testBase( cur->getZone(), cur->getNet(), cur->getNode(),
			  cur->getPoint(), cur->getDomain() );
    NodeConfig test( &testBase ),
               *found( systemList->Find( test ) );

    if( found==NULL )
      found = cur;
*/

          NodeConfig *found( cur );

          if( found->disableDZA() )
            noDZA = 1;

          if( found->disableZAP() )
            noZAP = 1;

          if( found->disableHydra() )
            noHydra = 1;

          if( password==NULL )
            password = found->getPassword();

          if( myAkaList==NULL )
            myAkaList = found->getAkaList();
        }

      if( noDZA ||
          !comPort->getHardWareMode() ||
          (miscInfo.state2&DirZapAvail)!=DirZapAvail )
        hisCode = Compatibility( hisCode&(~DZA) );

      if( noZAP || (miscInfo.state2&ZedZapAvail)!=ZedZapAvail )
        hisCode = Compatibility( hisCode&(~ZAP) );

      if( noHydra || (miscInfo.state2&HydraAvail)!=HydraAvail )
        hisCode = Compatibility( hisCode&(~HYD) );

      if( (hisCode&HYD)==HYD )
	code = HYD;
      else if( (hisCode&DZA)==DZA )
	code = DZA;
      else if( (hisCode&ZAP)==ZAP )
        code = ZAP;
      else if( (hisCode&ZMO)==ZMO )
        code = ZMO;

      newCode = Compatibility( newCode | code );
      hisCode = Compatibility( (hisCode & (~(HYD|JAN|DZA|ZAP|ZMO))) | code );

      if( *akaList==NULL ||
	  illegalAka!=NULL ||
	  !(*akaList)->GetItemsInContainer() )
	newCode = Compatibility( newCode | HAT );

      rc = emsiHandShakeTransmit( netTraf, xTraf, newCode, password, myAkaList );
    }
    else
      log( Msg_LogEmsiFailed );

Done:
  delete answerEmsi;
  return rc;
}

// ============================= callMode ================================

int callMode( QueueEntryList **akaList,
				   Compatibility myCode,
				   Compatibility& hisCode,
				   QueueEntry& called,
				   char **password )
{
  int rc( emsiHandShakeTransmit( called.getMail(), called.getFiles(),
				 myCode, called.getPassword(), called.getAkaList() ) );

  Buffer *answerEmsi( NULL );

  if( !rc )
  {
    rc = emsiHandShakeReceive( &answerEmsi, 0 );

    if( !rc )
      if( answerEmsi!=NULL )
      {
	rc = emsiParse( *answerEmsi, hisCode, akaList, 0, &called, password );

        if( rc )
          goto Done;

        ourMail = 0;
        ourFiles = 0;

	if( *akaList!=NULL && (*akaList)->GetItemsInContainer() )
	{
	  char Yes( 0 );

	  for( QueueEntryIterator list( **akaList ); list && !Yes; list++ )
	    if( list()==called )
	      Yes = 1;

	  if( Yes )
   	    for( QueueEntryIterator lst( **akaList ); lst; lst++ )
            {
              QueueEntry& cur = lst();

              ourMail += cur.getMail();
              ourFiles += cur.getFiles();
            }
          else
	  {
            for( QueueEntryIterator i( **akaList ); i; i++ )
              i().makeBsy( 0 );

	    delete *akaList;
	    *akaList = NULL;

            char buf[65];

	    log( Msg_LogEmsiNeedThisAKA, called.toString( buf, 65 ) );
	  }
	}
      }
      else
	log( Msg_LogEmsiFailed );
  }

Done:
  delete answerEmsi;
  return rc;
}

// =========================== emsiHandShake =============================

int emsiHandShake( QueueEntryList **akaList, Compatibility myCode,
			  Compatibility& hisCode, QueueEntry* called,
			  char **password )
{
  oldEMSIrest = 0;
  needIndicator = 1;
  EMSItimer = new Timer( miscInfo.EMSItime );

  *akaList = NULL;

  int rc( called==NULL ? emsiHandShakeInitAnswer( EMSItimer->rest() ) :
			 emsiHandShakeInitCall( EMSItimer->rest() ) );

  if( !rc )
    rc = called==NULL ? answerMode( akaList, myCode, hisCode, password )
		      : callMode( akaList, myCode, hisCode, *called, password );

  showAction( rc ? Msg_ScrInfoEmsiErr : Msg_ScrInfoEmsiDone );

  delete EMSItimer;

  return rc;
}

/*
int testEmsi()
{
  char *emsi =
"EMSI_DAT01F4{EMSI}{2:5020/1475@FidoNet "
"2:5020/395.27@Fidonet 275:5020/3.2@HackNet 167:5020/38.4@MadNet 39:2/0@DrugsNet"
" 39:2/1@DrugsNet"
" 268:20/5.1@Mgapi}{}{8N1,PUA,RH1}{HYD,DZA,CHT,ZAP,ZMO,FNC}{E0}{T-Mail}{2603.DOS/"
"rcialConnection baudra{TRAF}{22C 0}{Unregistered}{TRAF}{22C 0}{OHFR}{6:00-5:59"
"}{IDENT}{[Ni(k0N'$ $t/\\b4\\c2i0N][Mo\\e1\\aa\\a2\\a0, \\93"
"\\97\\a5\\e0\\e2\\a0\\ad\\ae\\a2\\ae \\ad\\a0 \\aa\\e3\\ab\\a8\\e7\\a8\\aa\\a0\\e5"
":)][Ni(k0N][7-095-XpeH-Bam][14400][USR,V42B,V32B|BB$|DRUGS|DrugsNet|Wheels|MJ|a"
"ACiD]}{TRX#}{[351B0B0D]}4EC2";

  Buffer answerEmsi;

  answerEmsi.putString( emsi );
  answerEmsi.putInto( 0 );
  answerEmsi.stripByte();

  QueueEntryList *akaList = NULL;
  char *password = NULL;
  Compatibility comp( Compatibility(0) );

  emsiParse( answerEmsi, comp, &akaList, 1, NULL, &password );
  return 0;
}
*/