#include <stdlib.h>

#include "enable.h"
#include "regexp.h"

EnableItem::EnableItem( char *istr, EnableItemType itype ) :
  type( itype ),
  str( newStr( istr ) )
{
  RegularExpresion mask( istr );

  isOk = mask.isValid();
  errMsg = newStr( isOk ? "" : mask.getRegErrorMsg() );
}

EnableItem::~EnableItem()
{
  delete str;
  delete errMsg;
}

int EnableItem::equal( const BaseObject &t ) const
{
  return !stricmp( str, (*(const EnableItem *)&t).str );
}

void EnableItem::print( ostream& s ) const
{
  s << stringType() << char(':') << str;
}

char *EnableItem::nameOfClass() const
{
  return "EnableItem";
}

int EnableItem::isUnify( char *test )
{
  RegularExpresion mask( str );

  return mask.maskMach( test );
}

char *EnableItem::stringType() const
{
  switch( type )
  {
  case EnableItemUnknown:
    return "Unknown";

  case EnableItemAddress:
    return "Address";

  case EnableItemFlag:
    return "Flag";
  }

  return "*Who Unknown?*";
}

char *EnableItem::toString( char *buf, int size ) const
{
  strncat( buf, stringType(), size );
  strncat( buf, " ", size );
  return strncat( buf, str, size );
}

EnableList::EnableList() :
  Collection( 5, 5, SemaphoreModeNoNeed )
{
}
  
EnableItem* EnableList::Find( EnableItem& item, LockMode lock ) const
{
  return (EnableItem*)Collection::Find( item, lock );
}
    
EnableItem& EnableList::operator [] ( const int pos ) const
{
  return *(EnableItem *)&(Collection::operator[]( pos ));
}

EnableIterator::EnableIterator( EnableList& ini ) : 
  CollectionIterator( *(Collection *)&ini )
{
}
  
EnableItem& EnableIterator::operator () () const
{
  return *(EnableItem *)&(CollectionIterator::operator()());
}
