// Constants for Class 2 commands

#ifndef __FAX_H

#define __FAX_H

#include <stdio.h>

#include "ascii.h"

enum FaxRC		// exit codes
{
  FaxRCSent   = 0,
  FaxRCInSync = 0,
  FaxRCNoSync = 1,
  FaxRCNoDial = 2,
  FaxRCBusy   = 3,
  FaxRCHang   = 4,
  FaxRCError  = 5
};

enum PageRC		// My own page reception codes
{
  PageRCGood            = 0,
  PageRCAnotherDocument = 1,
  PageRCEndOfDocument   = 2,
  PageRCHangUp          = 4,
  PageRCError           = 5
};

// Class 2 session parameters
//
// Set desired transmission params with +FDT=DF,VR,WD,LN
// DF = Data Format :   0  1-d huffman
//                      *1 2-d modified Read
//                      *2 2-d uncompressed mode
//                      *3 2-d modified modified Read
//
// VR = Vertical Res :  0 Normal, 98 lpi
//                      1 Fine, 196 lpi
//
// WD = width :         0  1728 pixels in 215 mm
//                      *1 2048 pixels in 255 mm
//
// LN = page length :   0 A4, 297 mm
//                      1 B4, 364 mm
//                      2  Unlimited
//
// EC = error correction :      0 disable ECM
//
// BF = binary file transfer :  0 disable BFT
//
// ST = scan time/line :        VR = normal     VR = fine
//                         0    0 ms            0 ms
//

struct T30Params		// A T.30 DIS frame, as sent by
				//  the remote fax machine
{
  enum VerticalResolution	// vertical resolution
  {
    VRNormal = 0,		// Normal, 98 lpi
    VRFine   = 1		// Fine, 196 lpi
  } vr;

  enum BaudRate			// Baud rate
  {
    BaudRate2400 = 0,
    BaudRate4800 = 1,
    BaudRate7200 = 2,
    BaudRate9600 = 3
  } br;				// BR = Bit Rate : br * 2400BPS

  enum PageWidth		// Page Width
  {
    PageWid1728 = 0,		// 1728 pixels in 215 mm
    PageWid2048 = 1,		// 2048 pixels in 255 mm
    PageWid2432 = 2		// 2432 pixels in 303 mm
  } wd;

  enum PageLength		// page length
  {
    PageLenA4        = 0,	// A4, 297 mm
    PageLenB4        = 1,	// B4, 364 mm
    PageLenUnLimited = 2	// unlimited
  } ln;

  enum DataFormat		// data format
  {
    DataFormat1DHUFFMAN = 0,	// 1-D modified Huffman
    DataFormat2DMREAD   = 1,	// 2-D modified Read
    DataFormat2DUNCOMP  = 2,	// 2-D uncompressed mode
    DataFormat2DMMREAD  = 3	// 2-D modified modified Read
  } df;

  int ec;	// Error Correction :
		//     0 disable ECM
		//     1 enable ECM 64 bytes/frame
		//     2 enable CM 256B/frame

  int bf;	// Binary File Transfer
		//     0 disable BFT
		//     1 enable BFT

  int st;	// Scan Time (ms) :
		//     VR   Normal  Fine
		//     0    0       0 ms
		//     1    5       5
		//     2    10      5
		//     3    10      10
		//     4    20      10
		//     5    20      20
		//     6    40      20
		//     7    40      40
};

struct FaxModemResponse
{
  char remote_id[51];		// +FCSI remote id
  int hangup_code;		// +FHNG code
  int post_page_response_code;	// +FPTS code
  int post_page_message_code;	// +FET code
  int fcon;			// Boolean; TRUE if +FCON seen
  int connect;			// Boolean; TRUE if CONNECT msg seen
  int ok;			// Boolean; TRUE if OK seen
  int error;			// Boolean; TRUE if ERROR or NO CARRIER seen
  struct T30Params T30;		// Session params; parsed from +FDCS
};

class Window;

class FaxReceiver
{
private:
  Modem *port;
  Window *wnd;
  int tid;
  volatile int monitor;

  int gEnd_of_document;
  unsigned char swaptable[ 256 ];
  volatile struct FaxModemResponse response;
  volatile unsigned long faxsize,
                         totalsize,
                         filesWritten;
  unsigned long startReceive;
  volatile char *fileName,
                *modemState;

private:
  PageRC get_fax_file( char *path, int page, int taskNumber );
  PageRC read_g3_stream( FILE *, int );
  void get_faxline( char *p, int nbytes, int wtime );
  void get_modem_result_code();
  void parse_text_response( char * );

  PageRC faxmodem_receive_page( int );

public:
  FaxReceiver( Modem * );
  ~FaxReceiver();

  int receive( char *path, int taskNumber=0 );
  friend void _Optlink faxMonitor( void * );
};

struct FaxConfig
{
  char *inbound;
  int baud;
};

extern FaxConfig faxConfig;
extern char faxActive;

void receiveFax();

#endif