#define INCL_DOSSEMAPHORES
#define INCL_DOSDATETIME
#define INCL_DOSPROCESS
#define INCL_DOSEXCEPTIONS
#include <os2.h>

#include <io.h>
#include <share.h>
#include <fcntl.h>

#include "except.h"

#include "language.h"
#include "misc.h"
#include "semaphor.h"
#include "ea.h"

int process()
{
  int RC=0;

  if( miscInfo.exitFlag!=NULL && !access( miscInfo.exitFlag, 0 ) )
  {
    FILE *in = _fsopen( miscInfo.exitFlag, "r" );

    if( in!=NULL )
    {
      int count = fscanf( in, "%d", &RC );
      int err = errno;

      fclose( in );

      if( !forceDelete( miscInfo.exitFlag ) )
        if( count==1 )
        {
          log( Msg_InfoFlagExitFound, RC, miscInfo.exitFlag );
          return RC;
        }
        else
        {
          errno = err;
          log( Msg_ErrUnableRead, miscInfo.exitFlag, strError() );
        }
    }
    else
      log( Msg_ErrUnableOpen, miscInfo.exitFlag, strError() );
  }

  if( flagCount && flagList!=NULL )
    for( int i=0; i<flagCount; i++ )
      if( !access( flagList[i].name, 0 ) )
      {
        if( miscInfo.flags==NULL || !begin( flagList[i].name, miscInfo.flags ) )
          log( Msg_InfoFlagFound, flagList[i].name );
        else
          log( Msg_InfoFlagFound, flagList[i].name+strlen(miscInfo.flags) );

        if( !forceDelete( flagList[i].name ) )
        {
          if( miscInfo.flags==NULL || !begin( flagList[i].name, miscInfo.flags ) )
            log( Msg_InfoFlagDeleted, flagList[i].name );
          else
            log( Msg_InfoFlagDeleted, flagList[i].name+strlen(miscInfo.flags) );

          RC = doEventKey( flagList[i], -1 );

          if( RC )
            return RC;
        }
      }

  if( noFlagCount && noFlagList!=NULL )
    for( int i=0; i<noFlagCount; i++ )
      if( access( noFlagList[i].name, 0 ) )
      {
        if( miscInfo.flags==NULL || !begin( noFlagList[i].name, miscInfo.flags ) )
          log( Msg_InfoFlagNotFound, noFlagList[i].name );
        else
          log( Msg_InfoFlagNotFound, noFlagList[i].name+strlen(miscInfo.flags) );

        makeFilePath( noFlagList[i].name );

        FILE *fl( _fsopen( noFlagList[i].name, "a", SH_DENYRW ) );

        if( fl==NULL )
          log( Msg_ErrUnableCreateFlag, noFlagList[i].name, strError() );
        else
        {
          fclose( fl );
          setExtendedAttribute( noFlagList[i].name, ".TYPE", Msg_EAFlag );

          if( miscInfo.flags==NULL || !begin( noFlagList[i].name, miscInfo.flags ) )
            log( Msg_InfoFlagCreated, noFlagList[i].name );
          else
            log( Msg_InfoFlagCreated, noFlagList[i].name+strlen(miscInfo.flags) );

          RC = doEventKey( noFlagList[i], -1 );

          if( RC )
            return RC;
        }
      }

  return 0;
}

volatile char flagsProcessorActive = 0;
static volatile char abortFlag = 0;
static HEV sema;

void _Optlink processFlags( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  unsigned long RC( createEventSemaphore( &sema, "FlagsProcess" ) );

  if( RC )
  {
    log( Msg_ErrUnableCreateFlagsSema, RC );
    os2Message( RC );
    log( Msg_ErrFlagsFailed );
    deRegisterExceptionHandler
    return;
  }

  while( 1 )
  {
    if( abortFlag )
    {
      deRegisterExceptionHandler
      return;
    }

    if( !terminalActive )
      flagsProcessorActive = 1;

    RC = process();

    flagsProcessorActive = 0;

    if( RC )
      break;

    HTIMER timer;

    RC = DosAsyncTimer( miscInfo.checkFlags*1000, (HSEM)sema, &timer );

    if( RC )
    {
      log( Msg_ErrUnableFlagsAsync, RC );
      os2Message( RC );
      log( Msg_ErrFlagsFailed );
      deRegisterExceptionHandler
      return;
    }

    RC = DosWaitEventSem( sema, (ULONG)SEM_INDEFINITE_WAIT );

    if( RC )
    {
      log( Msg_ErrUnableWaitFlagsSema, RC );
      os2Message( RC );
      log( Msg_ErrFlagsFailed );
      deRegisterExceptionHandler
      return;
    }

    ULONG cnt;

    RC = DosResetEventSem( sema, &cnt );

    if( RC )
    {
      log( Msg_ErrUnableResetFlagsSema, RC );
      os2Message( RC );
      log( Msg_ErrFlagsFailed );
      deRegisterExceptionHandler
      return;
    }
  }

  deRegisterExceptionHandler

  if( RC>OSReturnCode )
    OSReturnCode = int( RC );
}

void abortFlagProcess()
{
  abortFlag = 1;

  APIRET rc = DosPostEventSem( sema );

  if( rc )
  {
    log( Msg_ErrUnablePostFlagsSema, rc );
    os2Message( rc );
  }
}
