#define INCL_DOS
#define INCL_DOSEXCEPTIONS
#include <os2.h>

#include <stdio.h>
#include <io.h>
#include <stdlib.h>
#include <string.h>
#include <share.h>
#include <fcntl.h>
#include <ctype.h>
#include <builtin.h>
#include <errno.h>

#include "language.h"

#include "newstr.h"
#include "direct.h"
#include "buffer.h"
#include "misc.h"
#include "freq.h"
#include "session.h"
#include "except.h"
#include "rexxhook.h"

static HEV volatile os2FreqHEV = 0xffffffffl;

static long freqCntSesTime = 0,
	    freqCntSesSize = 0;

static NodeFreqInfo saveNodeFreqInfo,
		    nodeFreqInfo;

static long nodeFreqInfoPos = -1;
static char nodeFreqInfoWasModified = 0;

static const InfoPosition = 32;

AliasList *freqList( NULL ),
          *freqAlias( NULL );

// ---------------------- Alias ----------------------

Alias::Alias( char *ip, char *ia, int is ) :
  alias( newStr( ia ) ),
  scan( is )
{
  path = new char [ strlen( ip )+2 ];
  strcpy( path, ip );

  if( ia==NULL && path[ strlen(path) -1 ] != '\\' )
    strcat( path, "\\" );
}

Alias::~Alias()
{
  delete path;
  delete alias;
}

int Alias::equal( const BaseObject &t ) const
{
  return !stricmp( path, (*(const Alias *)&t).path );
}

void Alias::print( ostream& s ) const
{
  if( alias==NULL )
    s << (scan ? "Tree" : "Path" ) << " `" << path;
  else
    s << char('`') << alias << "' as `" << path << char( '\'');
}

char *Alias::nameOfClass() const
{
  return "Alias";
}

// ---------------------- AliasList ----------------------

AliasList::AliasList() :
  Collection( 5, 5, SemaphoreModeNoNeed )
{
}

Alias* AliasList::Find( Alias& item, LockMode lock ) const
{
  return (Alias*)Collection::Find( item, lock );
}
    
Alias& AliasList::operator [] ( const int pos ) const
{
  return *(Alias *)&(Collection::operator[]( pos ));
}

// ---------------------- AliasIterator ----------------------

AliasIterator::AliasIterator( AliasList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}
  
Alias& AliasIterator::operator () () const
{
  return *(Alias *)&(CollectionIterator::operator()());
}

// ---------------------- FreqFile ----------------------

FreqFile::FreqFile( const char *_name, unsigned long _size,
                    unsigned long _attrib, unsigned long _modify ) :
  name( newStr( _name ) ),
  size( _size ),
  attrib( _attrib ),
  modify( _modify )
{}

FreqFile::~FreqFile() 
{
  delete name;
}

int FreqFile::dump( FILE *file )
{ 
  char buf[ 32 ];

  sprintf( buf, "%08lu %08lx %04lx", size, modify, attrib );

  return putPair1( file, buf, name );
}

int FreqFile::equal( const BaseObject &t ) const
{
  return !stricmp( name, (*(const FreqFile *)&t).name );
}

char *FreqFile::nameOfClass() const
{
  return "FreqFile";
}

int FreqFile::match( RegularExpresion& exp ) const
{
  return exp.maskMach( strlwr( tmpStr( name ) ) );
}

// ---------------------- FreqDirectory ----------------------

FreqDirectory::FreqDirectory( const char *_path, FreqDirectoryList& freq,
                              int recurse ) :
  Collection( 100, 100, SemaphoreModeNeed ),
  path( newStr( _path ) )
{
   build( freq, recurse );
}

FreqDirectory::~FreqDirectory()
{
  delete path;
}

int FreqDirectory::equal( const BaseObject &t ) const
{
  return !stricmp( path, (*(const FreqDirectory *)&t).path );
}

char *FreqDirectory::nameOfClass() const
{
  return "FreqDirectory";
}

FreqFile* FreqDirectory::Find( FreqFile& item, LockMode lock ) const
{
  return (FreqFile*)Collection::Find( item, lock );
}
    
FreqFile& FreqDirectory::operator [] ( const int pos ) const
{
  return *(FreqFile *)&(Collection::operator[]( pos ));
}

void FreqDirectory::build( FreqDirectoryList& freq, int recurse )
{
  int i;

  for( i=0; path[i]; i++ )
    if( path[i]=='/' )
      path[i] = '\\';

  if( path[ i-1 ]!='\\' )
  {
    char *old = path;

    path = new char [ i+2 ];
    strcpy( path, old );
    path[i] = '\\';
    path[i+1] = 0;
    delete old;
  }

  char mask[]( "*" ),
       *full = (char *)_alloca( strlen( path )+sizeof( mask ) );

  strcpy( full, path );
  strcat( full, mask );

  SearchAttributes attr = SearchAttributes( recurse ? MayBeDirectory|MayBeReadOnly|MayBeArchived :
                                            MayBeReadOnly|MayBeArchived );

  for( Directory rc( full, attr ); rc; rc++ )
  {
    char *name = rc.getName();

    if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory )
    {
      if( strcmp( name, "." ) && strcmp( name, ".." ) )
      {
        char *more = new char [ strlen( path )+strlen( name )+2 ];

        strcpy( more, path );
        strcat( more, name );
        strcat( more, "\\" );
        freq.build( more, 1 );
        delete more;
      }
    }
    else
      AddAtTail( *new FreqFile( name, rc.getSize(), rc.getAttrib(),
                                rc.getUNIXModify() ) );
  }
}

unsigned long FreqDirectory::size() const
{ 
  unsigned long total( 0 );

  Lock();

  for( int i=0; i<used; i++ )
    total += ((FreqFile*)list[i])->getSize();

  UnLock();

  return total;
}

int FreqDirectory::dump( FILE *file )
{ 
  char buf[ 16 ];

  sprintf( buf, "#%lu", used );

  if( putPair1( file, buf, path ) )
    return 1;

  Lock();

  int rc(0);

  for( int i=0; i<used && !rc; i++ )
    if( ((FreqFile*)list[i])->dump( file ) )
      rc = 1;

  UnLock();

  return rc;
}

// ---------------------- FreqDirectoryIterator ----------------------

FreqDirectoryIterator::FreqDirectoryIterator( FreqDirectory& ini ) : 
  CollectionIterator( *(Collection *)&ini )
{
}

FreqFile& FreqDirectoryIterator::operator () () const
{
  return *(FreqFile *)&(CollectionIterator::operator()());
}

// ---------------------- FreqDirectoryList ----------------------

FreqDirectoryList::FreqDirectoryList() :
  Collection( 100, 100, SemaphoreModeNeed )
{
}

FreqDirectory* FreqDirectoryList::Find( FreqDirectory& item, LockMode lock ) const
{
  return (FreqDirectory*)Collection::Find( item, lock );
}
    
FreqDirectory& FreqDirectoryList::operator [] ( const int pos ) const
{
  return *(FreqDirectory *)&(Collection::operator[]( pos ));
}

int FreqDirectoryList::equal( const BaseObject &t ) const
{
  return this == (const FreqDirectoryList*)&t;
}

char *FreqDirectoryList::nameOfClass() const
{
  return "FreqDirectoryList";
}

void FreqDirectoryList::build( char *path, int recurse )
{
  FreqDirectory *freq = new FreqDirectory( path, *this, recurse );

  AddAtTail( *freq );
}

void FreqDirectoryList::total( unsigned long& fil, unsigned long& dir,
                               unsigned long& byt ) const
{
  Lock();

  for( fil = 0, dir = 0, byt = 0; dir<used; dir++ )
  {
    FreqDirectory *cur( (FreqDirectory *)list[ dir ] );

    fil += cur->GetItemsInContainer();
    byt += cur->size();
  }

  UnLock();
}

int FreqDirectoryList::dump( FILE *file )
{
  int rc(0);

  Lock();

  for( int i=0; i<used && !rc; i++ )
    if( ((FreqDirectory *)list[i])->dump( file ) )
      rc = 1;

  UnLock();

  return rc;
}

// ---------------------- FreqDirectoryListIterator ----------------------

FreqDirectoryListIterator::FreqDirectoryListIterator( FreqDirectoryList& ini ) : 
  CollectionIterator( *(Collection *)&ini )
{
}
  
FreqDirectory& FreqDirectoryListIterator::operator () () const
{
  return *(FreqDirectory *)&(CollectionIterator::operator()());
}

// ---------------------- FreqResolve ----------------------

FreqResolve::FreqResolve( const char *_name, const char *_as,
                          unsigned long _size, unsigned long _attrib,
                          unsigned long _modify ) :
  name( newStr( _name ) ),
  as( newStr( _as ) ),
  path( 0 ),
  size( _size ),
  attrib( _attrib ),
  modify( _modify )
{}

FreqResolve::~FreqResolve()
{
  delete name;
  delete as;
  delete path;
}

int FreqResolve::equal( const BaseObject &t ) const
{
  return !stricmp( name, (*(const FreqResolve *)&t).name );
}

char *FreqResolve::nameOfClass() const
{
  return "FreqResolve";
}

// ---------------------- FreqFileList ----------------------

FreqFileList::FreqFileList() :
  Collection( 10, 10, SemaphoreModeNoNeed )
{
}

FreqResolve* FreqFileList::Find( FreqResolve& item, LockMode lock ) const
{
  return (FreqResolve*)Collection::Find( item, lock );
}
    
FreqResolve& FreqFileList::operator [] ( const int pos ) const
{
  return *(FreqResolve *)&(Collection::operator[]( pos ));
}

int FreqFileList::equal( const BaseObject &t ) const
{
  return this == (const FreqFileList* const)&t;
}

char *FreqFileList::nameOfClass() const
{
  return "FreqFileList";
}

// ---------------------- FreqFileListIterator ----------------------

FreqFileListIterator::FreqFileListIterator( FreqFileList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}

FreqResolve& FreqFileListIterator::operator () () const
{
  return *(FreqResolve *)&(CollectionIterator::operator()());
}

// --------------------------------------------

void _Optlink buildFreqIdxThr( void* thread )
{
  conditionalRegisterExceptionHandler( thread )

  FreqDirectoryList lst;
  unsigned long t( getTimeMark() );

  if( freqList!=NULL )
    for( AliasIterator cur( *freqList ); cur; cur++ )
      lst.build( cur().getPath(), cur().isScan() );

  t = getTimeMark()-t;

  FILE *file( NULL );

  for( int i(0); i<10; i++ )
  {
    file = _fsopen( miscInfo.freqIdx, "w", SH_DENYRW );

    if( file==NULL )
    {
      if( errno!=EACCESS )
        break;
    }
    else
      break;

    sleep( 500 );
  }

  if( file==NULL )
    log( Msg_ErrUnableCreate, miscInfo.freqIdx, strError() );

  unsigned long fil;
  unsigned long dir;
  unsigned long bytes;

  lst.total( fil, dir, bytes );

  char *f( strNumber( fil ) ),
       *d( strNumber( dir ) ),
       *b( strNumber( bytes ) ),
       *tt( strTime( t, 0 ) );

  if( file!=NULL )
  {
    if( lst.dump( file ) )
      log( Msg_ErrUnableWrite, miscInfo.freqIdx, strError() );
    else
      fprintf( file, "*** Elapsed time: %s, total %s bytes in %s files and %s dirs\n",
                     tt, b, f, d );
    fclose( file );
    setExtendedAttribute( miscInfo.freqIdx, ".TYPE", Msg_EAFREQBase1 );
  }

  log( Msg_InfoIndexFREQ, tt, b, f, d );
  delete tt;
  delete b;
  delete d;
  delete f;

  if( thread )
    deRegisterExceptionHandler
}

void buildFreqIdx()
{
  int tid = _beginthread( buildFreqIdxThr, 0, 40960, (void *)1 );

  if( tid==-1 )
    log( Msg_ErrIdxThreadError, strError() );
  else
    log( Msg_ErrIdxThreadStarted, tid );
}

Buffer *getDescription( char *path )
{
  char drive[_MAX_DRIVE],
       dir[_MAX_DIR],
       name[_MAX_FNAME],
       ext[_MAX_EXT];
  static char mask[] = "files.bbs";

  _splitpath( path, drive, dir, name, ext );

  char *full = (char *)_alloca( strlen(drive)+strlen(dir)+sizeof(mask) );

  strcpy( full, drive );
  strcat( full, dir );
  strcat( full, mask );

  char *search = (char *)_alloca( strlen(name)+strlen(ext)+1 );

  strcpy( search, name );
  strcat( search, ext );

  Buffer *ret = NULL;
  FILE *bbs = _fsopen( full, "r" );

  if( bbs!=NULL )
  {
    const lineSize = 512;
    char line[ lineSize ];

    while( !feof(bbs) )
    {
      *line = 0;
      fgets( line, lineSize, bbs );

      if( !*line )
	break;

      if( !strnicmp( line, search, strlen(search) ) )
      {
	ret = new Buffer;

	int searchLen = strlen( search );

	while( !feof(bbs) || *line )
	{
	  register int len = strlen( line )-1;
	  char nl( 0 );

	  if( line[ len ]=='\n' )
	    nl = 1;

	  for( char *pos( line+searchLen+1 ); *pos==' ' && *pos; pos++ ) continue;

	  searchLen = 0;
	  ret->putString( pos );

	  if( !nl )
	    while( !feof(bbs) )
	      switch( len = fgetc( bbs ), len )
	      {
	      case '\n':
		ret->putInto( '\n' );
	      case EOF:
		goto Done;

	      default:
		ret->putInto( char( len ) );
	      }
Done:
	  if( feof(bbs) )
	    break;

	  do
	  {
	    *line = 0;
	    fgets( line, lineSize, bbs );
	  } while( !*line && !feof( bbs ) );

	  if( !isspace(line[0]) )
	    break;
	}

	break;
      }

      if( line[ strlen(line)-1 ]!='\n' )
	while( !feof(bbs) )
	  if( fgetc( bbs )=='\n' )
	    break;

    }

    fclose( bbs );
  }

  if( ret==NULL )
  {
    ret = new Buffer;
    ret->putString( Msg_InfoFREQNoDesc );
  }

  ret->putInto( 0 );

  int len = strlen( ret->str )-1;

  if( len!=-1 )
    if( ret->str[ len ]=='\n' )
      ret->str[ len ] = 0;

  return ret;
}

// ============================ freq Base Func =============================

void freqBaseInfoClear()
{
  freqCntSesTime = 0;
  freqCntSesSize = 0;
  memset( &nodeFreqInfo, 0, sizeof(nodeFreqInfo) );
  nodeFreqInfoPos = -1;
  nodeFreqInfoWasModified = 0;
}

void freqBaseInit()
{
  freqBaseInfoClear();

  short zone( primaryAka->getZone() ),
	net( primaryAka->getNet() ),
	node( primaryAka->getNode() ),
	point( primaryAka->getPoint() );

  char *mode( access( miscInfo.freqBase, 0 ) ? "w+b" : "r+b" ),
       found( 0 );

  FILE *base = _fsopen( miscInfo.freqBase, mode );

  if( base==NULL )
  {
    log( Msg_ErrUnableOpen, miscInfo.freqBase, strError() );
    log( Msg_SesNoFREQLim );
  }
  else
  {
    while( !found )
    {
      nodeFreqInfoPos = ftell( base );

      int readed = fread( &nodeFreqInfo, 1, sizeof(nodeFreqInfo), base );

      if( !readed )
	break;
      else
	if( readed && readed!=sizeof(nodeFreqInfo) )
	{
	  log( Msg_ErrUnableRead, miscInfo.freqBase, strError() );
	  break;
	}

      if( zone==nodeFreqInfo.zone && net==nodeFreqInfo.net &&
	  node==nodeFreqInfo.node && point==nodeFreqInfo.point )
	found = 1;
    }

    fclose( base );

    if( !stricmp( mode, "w+b" ) )
      setExtendedAttribute( miscInfo.freqBase, ".TYPE", Msg_EAFREQBase );
  }

  if( !found )
  {
    freqBaseInfoClear();

    nodeFreqInfo.zone = zone;
    nodeFreqInfo.net = net;
    nodeFreqInfo.node = node;
    nodeFreqInfo.point = point;
  }

  long tim = time( NULL );
  struct tm *tm = localTime( &tim );
  long day;

  tm->tm_mon++;

  day = long(long(tm->tm_year)<<16) | long(short(tm->tm_mon)<<8) | tm->tm_mday;

  if( day != nodeFreqInfo.lastDay )
  {
    nodeFreqInfo.lastDay = day;
    nodeFreqInfo.dayTime = 0;
    nodeFreqInfo.daySize = 0;
  }

  if( tm->tm_mon != nodeFreqInfo.lastMon )
  {
    nodeFreqInfo.lastMon = short(tm->tm_mon);
    nodeFreqInfo.monTime = 0;
    nodeFreqInfo.monSize = 0;
  }

  short week = short((tim/60/60/24-4)/7);

  if( week != nodeFreqInfo.lastWeek )
  {
    nodeFreqInfo.lastWeek = week;
    nodeFreqInfo.weekTime = 0;
    nodeFreqInfo.weekSize = 0;
  }

  memcpy( &saveNodeFreqInfo, &nodeFreqInfo, sizeof(saveNodeFreqInfo) );
}

void freqBaseDeInit()
{
  freqCntSesTime = 0;
  freqCntSesSize = 0;
  nodeFreqInfoWasModified = 0;
  memcpy( &nodeFreqInfo, &saveNodeFreqInfo, sizeof(nodeFreqInfo) );
}

int freqBaseUpDate( long FreqTransferTime, long FreqTransferSize,
                    char *name, char *password )
{
  freqCntSesTime       += FreqTransferTime;
  freqCntSesSize       += FreqTransferSize;
  nodeFreqInfo.dayTime += FreqTransferTime;
  nodeFreqInfo.daySize += FreqTransferSize;
  nodeFreqInfo.monTime += FreqTransferTime;
  nodeFreqInfo.monSize += FreqTransferSize;
  nodeFreqInfo.weekTime += FreqTransferTime;
  nodeFreqInfo.weekSize += FreqTransferSize;

  int rc( 0 );

  if( FreqTransferTime || FreqTransferSize )
  {
    nodeFreqInfoWasModified = 1;

    Buffer buf;

    if( freqCntSesTime > miscInfo.sessionFreqTime*60 && miscInfo.sessionFreqTime )
    {
      log( Msg_SesTimeFREQLim );
      buf.putString( Msg_SesTimeFREQLimAns );
abort_msg:
      buf.putString( fileName( name ) );
      buf.putString( Msg_SesFREQLimAort );
      rc = 1;
    }
    else if( nodeFreqInfo.dayTime > miscInfo.dayFreqTime*60 && miscInfo.dayFreqTime )
    {
      log( Msg_SesDayTimeFREQLim );
      buf.putString( Msg_SesDayTimeFREQLimAns );
      goto abort_msg;
    }
    else if( nodeFreqInfo.weekTime > miscInfo.weekFreqTime*60 && miscInfo.weekFreqTime )
    {
      log( Msg_SesWeekTimeFREQLim );
      buf.putString( Msg_SesWeekTimeFREQLimAns );
      goto abort_msg;
    }
    else if( nodeFreqInfo.monTime > miscInfo.monFreqTime*60 && miscInfo.monFreqTime )
    {
      log( Msg_SesMonthTimeFREQLim);
      buf.putString( Msg_SesMonthTimeFREQLimAns );
      goto abort_msg;
    }

    buf.putInto( 0 );

    if( strlen(buf.str) )
      generateFreqAnswer( buf.str, *primaryAka, hisSysOp, password, &nodeFreqInfo );
  }

  return rc;
}

void freqBaseInfoSave()
{
  if( !nodeFreqInfoWasModified )
    return;

  char *mode( nodeFreqInfoPos==-1 ? "ab" :
				    (access( miscInfo.freqBase, 0 ) ? "w+b" :
								      "r+b") );

  FILE *base = _fsopen( miscInfo.freqBase, mode );

  if( base==NULL )
  {
    log( Msg_ErrUnableOpen, miscInfo.freqBase, strError() );
    log( Msg_SesNoFREQLim );
  }
  else
  {
    if( nodeFreqInfoPos!=-1 )
      fseek( base, nodeFreqInfoPos, SEEK_SET );

    int written = fwrite( &nodeFreqInfo, 1, sizeof(nodeFreqInfo), base );

    if( written != sizeof(nodeFreqInfo) )
      log( Msg_ErrUnableWrite, miscInfo.freqBase, strError() );

    fclose( base );

    if( !stricmp( mode, "w+b" ) )
      setExtendedAttribute( miscInfo.freqBase, ".TYPE", Msg_EAFREQBase );
  }
}

// ===================== freqIOerror & putFreqFound ==========================

void freqIOerror( Buffer& buf, int err )
{
  buf.putString( Msg_SesFREQPrcErr );
  errno = err;
  buf.putString( strError() );
  buf.putString( Msg_SesFREQPrcErrEnd );
}

void putFreqFound( char *full, FILE *rloFile, Buffer& buf,
                   char& first, long size )
{
  log( "?FREQ: found `%s' %lu", full, size );

  char *nm = fileName(full);

  if( first )
  {
    buf.putString( Msg_SesFREQPrcSendAs );
    first = 0;
  }

  char *fmt = (char *)_alloca( strlen(nm)+32 );

  sprintf( fmt, "%-12s %11d", nm, size );

  buf.putString( fmt );

  for( int len( strlen(fmt) ); len<InfoPosition; len++ )
    buf.putInto( ' ' );

  long aprTime( size/(comPort->getLastBaud()/10) );

  if( freqCntSesTime+aprTime > miscInfo.sessionFreqTime*60 && miscInfo.sessionFreqTime )
    buf.putString( Msg_SesFREQPrcSesTime );
  else if( freqCntSesSize+size > miscInfo.sessionFreqSize*1024 && miscInfo.sessionFreqSize )
    buf.putString( Msg_SesFREQPrcSesSize );
  else if( nodeFreqInfo.dayTime+aprTime > miscInfo.dayFreqTime*60 && miscInfo.dayFreqTime )
    buf.putString( Msg_SesFREQPrcDayTime );
  else if( nodeFreqInfo.daySize+size > miscInfo.dayFreqSize*1024 && miscInfo.dayFreqSize )
    buf.putString( Msg_SesFREQPrcDaySize );
  else if( nodeFreqInfo.weekTime+aprTime > miscInfo.weekFreqTime*60 && miscInfo.weekFreqTime )
    buf.putString( Msg_SesFREQPrcWeekTime );
  else if( nodeFreqInfo.weekSize+size > miscInfo.weekFreqSize*1024 && miscInfo.weekFreqSize )
    buf.putString( Msg_SesFREQPrcWeekSize );
  else if( nodeFreqInfo.monTime+aprTime > miscInfo.monFreqTime*60 && miscInfo.monFreqTime )
    buf.putString( Msg_SesFREQPrcMonthTime );
  else if( nodeFreqInfo.monSize+size > miscInfo.monFreqSize*1024 && miscInfo.monFreqSize )
    buf.putString( Msg_SesFREQPrcMonthSize );
  else
  {
    Buffer& ret = *getDescription( full );

    for( int pos( InfoPosition ), i(0); ret.str[i]; i++ )
      if( ret.str[i]=='\n' || ++pos==78 )
      {
        buf.putInto( '\r' );
        buf.putInto( '\n' );

        for( pos=0; pos<InfoPosition; pos++ )
          buf.putInto( ' ' );

        if( ret.str[i]!='\n' )
          buf.putInto( ret.str[i] );
      }
      else
        buf.putInto( ret.str[i] );

    delete &ret;

    fputs( full, rloFile );
    fputc( '\n', rloFile );

    freqCntSesTime        += aprTime;
    freqCntSesSize        += size;
    nodeFreqInfo.dayTime  += aprTime;
    nodeFreqInfo.daySize  += size;
    nodeFreqInfo.monTime  += aprTime;
    nodeFreqInfo.monSize  += size;
    nodeFreqInfo.weekTime += aprTime;
    nodeFreqInfo.weekSize += size;
  }

  buf.putString( "\r\n" );
}

// ============================ freqScanPath ==============================

#ifndef FREQ_Idx

void freqScanPath( char *fullIni, FILE *rloFile, Buffer& buf,
                   char& first, int isScan, char& found,
                   char *maskIni )
{
  if( !comPort->isOnline() )
    return;

  char *mask = fileName( maskIni );
  char *full = (char *)_alloca( strlen(fullIni)+strlen(mask)+10 );

  strcpy( full, fullIni );

  int i( strlen(full) );

  if( full[i-1]!='\\' )
    strcat( full, "\\" );

  char *path = tmpStr( full );

  strcat( full, mask );

  char *full1 = (char *)_alloca( strlen(path)+_MAX_PATH+1 );

  for( Directory dir( full ); dir; dir++ )
  {
    if( !comPort->isOnline() )
      return;

    strcpy( full1, path );
    strcat( full1, dir.getName() );
    putFreqFound( full1, rloFile, buf, first, dir.getSize() );
    found = 1;
  }

  if( isScan )
  {
    strcpy( full, path );
    strcat( full, "*" );

    for( Directory i( full,
                      SearchAttributes( MustBeDirectory|MayBeReadOnly|MayBeArchived ) );
                   i; i++ )
    {
      char *name = i.getName();

      if( !comPort->isOnline() )
        break;

      if( (i.getAttrib()&MayBeDirectory)==MayBeDirectory &&
	  stricmp( name, "." ) && stricmp( name, ".." ) )
      {
	strcpy( full1, path );
	strcat( full1, name );
	freqScanPath( full1, rloFile, buf, first, 1, found, maskIni );
      }
    }
  }
}

#endif

// ============================= killFreq ===============================
static char *receivedFreqFileName = NULL;

void killFreq()
{
  if( receivedFreqFileName!=NULL && !access( receivedFreqFileName, 0 ) )
    forceDelete( receivedFreqFileName );
}

// ============================= availFreq ===============================

int availFreq( Compatibility myCode, char *password )
{
  if( miscInfo.freqMode==ProtectedFreq && !sessionProtected )
  {
    log( Msg_SesFREQOnlyProtect );
    generateFreqAnswer( Msg_SesFREQOnlyProtectAns, *primaryAka,
                        hisSysOp, password, NULL );
    return 0;
  }
  else if( miscInfo.freqMode==ListedFreq && !sessionListed )
  {
    log( Msg_SesFREQOnlyListed );
    generateFreqAnswer( Msg_SesFREQOnlyListedAns,
                        *primaryAka, hisSysOp, password, NULL );
    return 0;
  }
  else if( (myCode&NRQ)==NRQ )
  {
    log( Msg_SesFREQUsNotSupp );
    generateFreqAnswer( Msg_SesFREQUsNotSuppAns,
                        *primaryAka, hisSysOp, password, NULL );
    return 0;
  }
  else if( (myCode&HRQ)==HRQ )
  {
    log( Msg_SesFREQUsNotAtThis );
    generateFreqAnswer( Msg_SesFREQUsNotAtThisAns,
                        *primaryAka, hisSysOp, password, NULL );
    return 0;
  }
  else if( miscInfo.minFreqBaud && comPort->getLastBaud() < miscInfo.minFreqBaud )
  {
    log( Msg_SesFREQToSlow );

    Buffer b;

    b.putString( Msg_SesFREQToSlowAns );
    b.putEMSI( comPort->getSuff() );
    b.putString( "\r\n\r\n" );
    b.putInto( 0 );

    generateFreqAnswer( b.str, *primaryAka, hisSysOp, password, NULL );
    return 0;
  }

  return 1;
}

// ============================= processFreq ===============================

void freqSearch( char *str, Buffer& buf, FILE *rloFile, int len )
{
  char found( 0 ),
       first( 1 );

  if( freqAlias!=NULL )
    for( AliasIterator cur( *freqAlias ); cur; cur++ )
    {
      Alias& item( cur() );
      char *full( item.getPath() );

      if( !stricmp( str, item.getAlias() ) )
        for( Directory rc( full ); rc; rc++ )
        {
          char name[ _MAX_PATH*2 ];

          strcpy( name, rc.getPath() );
          strcat( name, rc.getName() );
          putFreqFound( name, rloFile, buf, first, rc.getSize() );
          found = 1;
        }
    }

  if( found || freqList==NULL || !freqList->GetItemsInContainer() )
    return;

#ifndef FREQ_Idx

  for( AliasIterator cur( *freqList ); cur; cur++ )
  {
    if( !comPort->isOnline() )
      break;

    Alias& item( cur() );
    char *full( item.getPath() );

    freqScanPath( full, rloFile, buf, first, item.isScan(), found, str );
  }

#else			// FREQ_Idx

  FILE *req = _fsopen( miscInfo.freqIdx, "r", SH_DENYWR );

  if( req==NULL )

#endif			// FREQ_Idx

  if( !found )
  {
    while( len<InfoPosition )
    {
      buf.putInto( ' ' );
      len++;
    }

    buf.putString( Msg_SesFREQSearchFail );
  }
}

void processFreqOS2( int ignore, Compatibility myCode, char *password,
                     char *freqFileName )
{
  receivedFreqFileName = newStr( freqFileName );

  if( receivedFreqFileName==NULL || access( receivedFreqFileName, 0 ) )
    return;

  FILE *freq = NULL;
  Buffer buf;

  if( ignore )
  {
    log( Msg_SesFREQOnOut );
    goto UnlinkFreq;
  }

  if( !availFreq( myCode, password ) )
    goto UnlinkFreq;

  freq = _fsopen( receivedFreqFileName, "r" );

  if( freq==NULL )
  {
    int err = errno;

    log( Msg_ErrUnableOpen, receivedFreqFileName, strError() );
    freqIOerror( buf, err );
  }
  else
  {
    char str[_MAX_PATH],
         *bink = primaryAka->buildOutboundName( ".Rlo", ".Request.List" );

    if( bink!=NULL )
    {
      char *rloName( bink );

      makeFilePath( rloName );

      FILE *rloFile = _fsopen( rloName, "w" );

      if( rloFile==NULL )
      {
        int err = errno;

        log( Msg_ErrUnableCreate, rloName, strError() );
        freqIOerror( buf, err );
      }
      else
      {
        freqBaseInit();

        char firstFilePut = 1;

        while( !feof(freq) && !ferror(freq) )
        {
          if( !comPort->isOnline() )
            break;

          *str = 0;
          fgets( str, _MAX_PATH, freq );

          int len = strlen(str);

          if( !len )
            break;

          char *sp = strchr( str, ' ' );

          if( sp!=NULL )
            *sp = 0;

          len = strlen(str);

          if( str[len-1]=='\n' )
            str[len-1] = 0;

          len = strlen(str);

          if( !len )
            continue;

          if( !firstFilePut )
            buf.putString( "\r\n" );
          else
            firstFilePut = 0;

          buf.putString( str );
          log( Msg_SesFREQSearch, str );

          freqSearch( str, buf, rloFile, len );
        }			//while( !feof(freq) && !ferror(freq) )

        freqBaseDeInit();
        fclose( rloFile );
        setExtendedAttribute( rloName, ".TYPE", Msg_EARLO );
      }				//if(rloFile==NULL) else

      delete bink;
    }

    fclose( freq );
  }				//if(freq==NULL)

  buf.putInto( 0 );
  generateFreqAnswer( buf.str, *primaryAka,
                      hisSysOp, password, &nodeFreqInfo );

UnlinkFreq:
  forceDelete( receivedFreqFileName );
}

struct os2freq freqThrInfo;

#ifndef FREQ_Idx
void _Optlink os2ProcessFreq( void *ptr )
#else
void processFreq( char *ptr )
#endif
{
#ifndef FREQ_Idx
  registerExceptionHandler
#endif

  char addr[65];
  char *hook = processRexxHook( "FREQ", (char *)ptr,
                                primaryAka->toString( addr, sizeof(addr)-1,
                                                      FIDOAddressShowModeEMSI ),
                                NULL,
                                NULL, NULL, NULL, 1 );

  if( hook==NULL || !stricmp( hook, "YourSelf" ) )
    processFreqOS2( freqThrInfo.ignore, freqThrInfo.myCode, freqThrInfo.password,
                    (char *)ptr );
  else if( !stricmp( hook, "NO" ) )
  {
RemoveFreqFile:
    if( !access( (char *)ptr, 0 ) )
      forceDelete( (char *)ptr );
  }
  else if( stricmp( hook, "YES" ) )
    log( Msg_ErrIllegalFREQMode, hook );
  else
    goto RemoveFreqFile;

  delete hook;

#ifndef FREQ_Idx
  DosPostEventSem( os2FreqHEV );

  delete (char *)ptr;

  deRegisterExceptionHandler
#endif
}

#ifndef FREQ_Idx

void processFreq( char *freqFileName )
{
  if( os2FreqHEV!=0xffffffffl )
  {
    log( "BUG? processFreq already started!" );
    return;
  }

  if( terminalActive || rexxTransfer )
    return;

  os2FreqHEV = 0xffffffffl;

  APIRET rc;

  if( rc=createEventSemaphore( (PULONG)&os2FreqHEV, "Freq" ), rc )
  {
    log( Msg_ErrFREQPrcSema, rc );
    os2Message( rc );

    log( Msg_ErrFREQUnableProcess );
    generateFreqAnswer( Msg_ErrFREQUnableProcessAns,
		       *primaryAka, hisSysOp, freqThrInfo.password, NULL );
    return;
  }

  char *fil = newStr( freqFileName );
  register long ttt = _beginthread( os2ProcessFreq, 0, 40960, (void *)fil );

  if( ttt==-1 )
  {
   log( Msg_ErrFREQPrcThr, strError() );
   log( Msg_ErrFREQUnableProcess );

   generateFreqAnswer( Msg_ErrFREQUnableProcessAns1,
		       *primaryAka, hisSysOp, freqThrInfo.password, NULL );
   deleteEventSemaphore( os2FreqHEV );
   os2FreqHEV = 0xffffffffl;
   delete fil;
  }
  else
    log( Msg_ThrFREQPrcThr, ttt );
}

#endif

void initFreqProcessor()
{
  os2FreqHEV = 0xffffffffl;
  delete receivedFreqFileName;
  receivedFreqFileName = NULL;
}

void waitFreqProcessor()
{
  if( os2FreqHEV!=0xffffffffl )
  {
    DosWaitEventSem( os2FreqHEV, (ULONG)SEM_INDEFINITE_WAIT );
    deleteEventSemaphore( os2FreqHEV );
    os2FreqHEV = 0xffffffffl;
  }
}
