#ifndef __FREQIDX_H

#define __FREQIDX_H

#include <stdio.h>

#include "regexp.h"
#include "class.h"
#include "session.h"

class Alias : public BaseObject
{
private:
  char *path,
       *alias;
  int scan;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:

  Alias( char *, char* =NULL, int =0 );
  ~Alias();

  char *getPath() const { return path; }
  char *getAlias() const { return alias; }

  int isScan() const { return scan; }
};

class AliasList : public Collection
{
public:
  AliasList();

  Alias* Find( Alias& item, LockMode lock=LockModeNeed ) const;
  Alias& operator [] ( const int pos ) const;
};

class AliasIterator : public CollectionIterator
{
public:
  AliasIterator( AliasList& ini );
  
  Alias& operator () () const;
};

extern AliasList *freqList,
	         *freqAlias;

#pragma pack(1)
struct NodeFreqInfo
{
  short zone, net, node, point;
  short lastMon,
	lastWeek;
  unsigned long lastDay,
		dayTime,
		daySize,
		weekTime,
		weekSize,
		monTime,
		monSize;
};
#pragma pack()

class FreqFile : public BaseObject
{
  char *name;
  unsigned long size,
                attrib,
                modify;

  int equal( const BaseObject &t ) const;
  void print( ostream& ) const {}
  char *nameOfClass() const;

public:
  FreqFile( const char *_name, unsigned long _size,
            unsigned long _attrib, unsigned long _modify );
  ~FreqFile();

  int dump( FILE * );

  char *getName() const { return name; }
  unsigned long getSize() const { return size; }
  unsigned long getAttrib() const { return attrib; }
  unsigned long getModify() const { return modify; }

  int match( RegularExpresion& exp ) const;
};

class FreqDirectoryList;

class FreqDirectory : public Collection, public BaseObject
{
  char *path;

  int equal( const BaseObject &t ) const;
  void print( ostream& ) const {}
  char *nameOfClass() const;

protected:
  void build( FreqDirectoryList& freq, int recurse );

public:
  FreqDirectory( const char *_path, FreqDirectoryList& freq, int recurse );
  ~FreqDirectory();

  FreqFile* Find( FreqFile& item, LockMode lock=LockModeNeed ) const;
  FreqFile& operator [] ( const int pos ) const;

  int dump( FILE * );

  char *getPath() const { return path; }
  unsigned long size() const;
};

class FreqDirectoryIterator : public CollectionIterator
{
public:
  FreqDirectoryIterator( FreqDirectory& ini );

  FreqFile& operator () () const;
};

class FreqDirectoryList : public Collection, public BaseObject //FreqDirectory
{
  int equal( const BaseObject &t ) const;
  void print( ostream& ) const {}
  char *nameOfClass() const;

public:
  FreqDirectoryList();

  FreqDirectory* Find( FreqDirectory& item, LockMode lock=LockModeNeed ) const;
  FreqDirectory& operator [] ( const int pos ) const;

  int dump( FILE * );

  void build( char *path, int recurse );
  void build( istream& stream );
  void total( unsigned long& fil, unsigned long& dir, unsigned long& byt ) const;
};

class FreqDirectoryListIterator : public CollectionIterator
{
public:
  FreqDirectoryListIterator( FreqDirectoryList& ini );
  
  FreqDirectory& operator () () const;
};

class FreqResolve : public BaseObject
{
  char *name,
       *as,
       *path;
  unsigned long size,
                attrib,
                modify;

  int equal( const BaseObject &t ) const;
  void print( ostream& ) const {}
  char *nameOfClass() const;

public:
  FreqResolve( const char *_name, const char *_as,
               unsigned long _size, unsigned long _attrib,
               unsigned long _modify );
  ~FreqResolve();

  char *getName() const { return name; }
  char *getAs() const { return as; }
  char *getPath() const { return path; }
  unsigned long getSize() const { return size; }
  unsigned long getAttrib() const { return attrib; }
  unsigned long getModify() const { return modify; }

  void setResolve( char *_dir, char *_as );
};

class FreqFileList : public Collection, public BaseObject //FreqResolve
{
  int equal( const BaseObject &t ) const;
  void print( ostream& ) const {}
  char *nameOfClass() const;

public:
  FreqFileList();

  FreqResolve* Find( FreqResolve& item, LockMode lock=LockModeNeed ) const;
  FreqResolve& operator [] ( const int pos ) const;
};

class FreqFileListIterator : public CollectionIterator
{
public:
  FreqFileListIterator( FreqFileList& ini );

  FreqResolve& operator () () const;
};

void buildFreqIdx();
void resolveFreq( FreqFileList& );
void processFreq( char *freqFileName );

struct os2freq
{
  int ignore;
  Compatibility myCode;
  char *password;
};

extern struct os2freq freqThrInfo;

void initFreqProcessor();
void waitFreqProcessor();
void killFreq();

void freqBaseInit();
void freqBaseInfoClear();
void freqBaseDeInit();
int freqBaseUpDate( long FreqTransferTime, long FreqTransferSize,
                    char *name, char *password );
void freqBaseInfoSave();

#endif