#include "globlvar.h"
#include "newstr.h"

GlobalVariable::GlobalVariable( char *_name, char *_value ) :
  name( newStr( _name ) ),
  value( newStr( _value ) )
{}

GlobalVariable::~GlobalVariable()
{
  delete name;
  delete value;
}

int GlobalVariable::less( const GlobalVariable &t ) const
{
  return stricmp( name, (*(const GlobalVariable *)&t).name ) < 0;
}

int GlobalVariable::equal( const BaseObject &t ) const
{
  return !stricmp( name, (*(const GlobalVariable *)&t).name);
}

void GlobalVariable::print( ostream& s ) const
{
}

char *GlobalVariable::nameOfClass() const
{
  return "GlobalVariable";
}

char *GlobalVariable::getValue() const
{
  return value ? value : "";
}

void GlobalVariable::setValue( char *_value )
{
  delete value;
  value = newStr( _value );
}

GlobalVariableList::GlobalVariableList() :
  Collection( 10, 10, SemaphoreModeNeed )
{
}

GlobalVariable* GlobalVariableList::Find( GlobalVariable& item, LockMode lock ) const
{
  return (GlobalVariable*)Collection::Find( item, lock );
}
    
GlobalVariable& GlobalVariableList::operator [] ( const int pos ) const
{
  return *(GlobalVariable *)&(Collection::operator[]( pos ));
}

void GlobalVariableList::sort()
{
  int i;

  Lock();

  for( i=1; i<used; i++ )
  {
    GlobalVariable *x = (GlobalVariable *)list[i];
    int j = i-1;

    while( j!=-1 && x->less( *(GlobalVariable *)list[j] ) )
    {
      list[j+1] = list[j];
      --j;
    }

    list[j+1] = x;
  }

  UnLock();
}

GlobalVariableIterator::GlobalVariableIterator( GlobalVariableList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}

GlobalVariable& GlobalVariableIterator::operator () () const
{
  return *(GlobalVariable *)&(CollectionIterator::operator()());
}
