//#define HYDRA_DELAY
//#define DEBUG_IDLE

#define INCL_DOSPROCESS
#define INCL_DOSMISC
#define INCL_DOSEXCEPTIONS
#include <os2.h>

#include <stdio.h>
#include <limits.h>
#include <ctype.h>
#include <io.h>
#include <time.h>
#include <string.h>
#include <sys\stat.h>
#include <builtin.h>

#include "except.h"
#include "hydra.h"
#include "hydraint.h"
#include "hdialog.h"
#include "zmodem.h"
#include "zinter.h"
#include "misc.h"
#include "timer.h"
#include "crc.h"
#include "key.h"
#include "recover.h"
#include "direct.h"
#include "process.h"
#include "session.h"

#include "language.h"

NoHydraChat::NoHydraChat( char *f ) :
  str( newStr( f ) )
{}

NoHydraChat::~NoHydraChat()
{
  delete str;
}

int NoHydraChat::equal( const BaseObject &t ) const
{
  return !stricmp( str, (*(const NoHydraChat *)&t).str );
}

void NoHydraChat::print( ostream& s ) const
{
  s << str;
}

char *NoHydraChat::nameOfClass() const
{
  return "NoHydraChat";
}

NoHydraChatList::NoHydraChatList() :
  Collection( 10, 5, SemaphoreModeNoNeed )
{
}

NoHydraChat* NoHydraChatList::Find( NoHydraChat& item, LockMode lock ) const
{
  return (NoHydraChat*)Collection::Find( item, lock );
}
    
NoHydraChat& NoHydraChatList::operator [] ( const int pos ) const
{
  return *(NoHydraChat *)&(Collection::operator[]( pos ));
}

NoHydraChatIterator::NoHydraChatIterator( NoHydraChatList& ini ) : 
  CollectionIterator( *(Collection *)&ini )
{
}

NoHydraChat& NoHydraChatIterator::operator () () const
{
  return *(NoHydraChat *)&(CollectionIterator::operator()());
}

#define inteli(x) (x)
#define intell(x) (x)
//#define PROLONGATE_CHAT                     //BUG! prolongate on 1st batch on RH1
                                              //  no restart to 2nd batch.

	// HYDRA's memory

static int originator;				// are we the orig side?
static int batchesdone;				// No. HYDRA batches done
static int hdxlink;				// hdx link & not orig side
static unsigned long options;			// INIT options HydraInit()
static word timeout;				// general timeout in secs
static char abortstr[] =
{24, 24, 24, 24, 24, 24, 24, 24, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 0};

static char *pktprefix = "";
static char *autostr = "hydra\r";

static byte *txbuf, *rxbuf;			// packet buffers
static unsigned long txoptions, rxoptions;	// HYDRA options (INIT seq)
static char txpktprefix[H_PKTPREFIX + 1];	// pkt prefix str they want
static long txwindow, rxwindow;			// window size (0=streaming)
static Timer braindead;				// braindead timer
static byte *txbufin;				// read data from disk here
static byte txlastc;				// last byte put in txbuf
static byte rxdle;				// count of received H_DLEs
static byte rxpktformat;			// format of pkt receiving
static byte *rxbufptr;				// current position in rxbuf
static byte *rxbufmax;				// highwatermark of rxbuf
static char txfname[_MAX_PATH], rxfname[_MAX_PATH];
						// fname of current files
static char rxpathname[_MAX_PATH*2];		// pathname of currrent rx file
static char *SendFileName = NULL;
static char *SendFileAlias = NULL;
static long txftime, rxftime;			// file timestamp (UNIX)
static long txfsize, rxfsize;			// file length
static int txfd, rxfd;				// file handles
static FILE *txFile, *rxFile;			// file ptr
static word rxpktlen;				// length of last packet
static word rxblklen;				// len of last good data blk
static int txstate, rxstate;			// xmit/recv states
static long txpos, rxpos;			// current position in files
static word txblklen;				// length of last block sent
static word txmaxblklen;			// max block length allowed
static long txlastack;				// last dataack received
static Timer txtimer, rxtimer;			// retry timers
static word txretries, rxretries;		// retry counters
static long rxlastsync;				// filepos last sync retry
static long txsyncid, rxsyncid;			// id of last resync
static word txgoodneeded;			// to send before larger blk
static word txgoodbytes;
			// no. sent at this blk size
static long txStart;
static long rxStart;
static int readTruncate;
static char *hInbound;
extern char hydraCpsOld[80];
static ReceivedFileList *xferInfo;
static FILE *sentList;
static int needSkipAll;
static int hCanOptions = 0;

static long txFileNumber = 0;
static long rxFileNumber = 0;
static long deltaFileNumber = 0;

unsigned long hydra_options  = 0;
long hydra_txwindow = 0;
long hydra_rxwindow = 0;

long hTransferCpsAfter = 60;
int hydra_timeout = 40;
char hydraIsActive = 0;

struct _h_flags
{
  char *str;
  unsigned long val;
};

static _h_flags h_flags[] =
{
  {"XON", HOPT_XONXOFF},
  {"TLN", HOPT_TELENET},
  {"CTL", HOPT_CTLCHRS},
  {"HIC", HOPT_HIGHCTL},
  {"HI8", HOPT_HIGHBIT},
  {"BRK", HOPT_CANBRK},
  {"ASC", HOPT_CANASC},
  {"UUE", HOPT_CANUUE},
  {"C32", HOPT_CRC32},
  {"DEV", HOPT_DEVICE},
  {"FPT", HOPT_FPT},
  {"NFI", HOPT_NFI},
  {NULL, 0x0L}
};

static char *mtext[12] =
{
  "Jan", "Feb", "Mar", "Apr", "May", "Jun",
  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

int hGetTimeOut()
{
  long timeOut( long(hydra_timeout)*300l*100l );
  int Rx( 0 );

  if( timeOutsHydra!=NULL )
  {
    TimeOut test( comPort->getLastBaud(), 0 ),
            *found( timeOutsHydra->Find( test ) );

    if( found!=NULL )
      Rx = found->getTimeOut()*100;
  }

  if( !Rx )
    Rx = int( timeOut / comPort->getLastBaud() );

  if( Rx < 200 )
    Rx = 200;

  if( comPort->isArq() )
    Rx *= 3;

  return Rx*10;
}

char *h_revdate( long revstamp )
{
  static char buf[32];
  struct tm *t = localTime( &revstamp );

  sprintf( buf, "%02d %s %d",
           t->tm_mday, mtext[t->tm_mon], t->tm_year+1900 );
  return buf;
}

static char chatAvailable = 0;
static int devtxstate;				// dev xmit state
static Timer devtxtimer;			// dev xmit retry timer
static Timer chatTimer;
static word devtxretries;			// dev xmit retry counter
static long devtxid, devrxid;			// id of last devdata pkt
static char devtxdev[H_FLAGLEN + 1];		// xmit device ident flag
static byte *devtxbuf;				// ptr to usersupplied dbuf
static word devtxlen;				// len of data in xmit buf

struct _h_dev
{
  char *dev;
  void (*func)( byte *data, word len );
};

void chatterKeyboardSend( Chatter& chat, int ch )
{
  if( chat.curPos<ChatBufferSize )
    chat.curBuf[ chat.curPos++ ] = char( ch );
  else
    execSchedule( melodies.bell );
}

void _Optlink chatterKeyboard( void *ptr )
{
  registerExceptionHandler

  Chatter& chat( *(Chatter *)ptr );
  
  execSchedule( melodies.chat );

  for( int done(0); !done; )
  {
    if( chat.abortKeyboard || chatTimer.isTimeOut() )
      break;

    if( !keyPressed() )
    {
      checkPoint();
      continue;
    }
    
    chatTimer.startTimer( miscInfo.chatTimeout*1000 );

    int ch = keyGet();

    switch( ch )
    {
    case key_Alt_C:
      done = 1;
      break;

    case '\b':
      chatterKeyboardSend( chat, '\b' );
      chatterKeyboardSend( chat, ' ' );
      chatterKeyboardSend( chat, '\b' );
        
      {
        int x = chatter.getCurThisX(),
            y = chatter.getCurThisY();

        if( x==1 && y==1 )
        {
          execSchedule( melodies.bell );
          break;
        }

        if( x==1 )
        {
          x = chatter.getThisWindow()->getWidth();
          y--;
        }
        else
          x--;

        chatter.cursorThis( x, y );
        chatter.This( ' ' );
        chatter.cursorThis( x, y );
      }
      break;

    case '\r':
    case '\n':
      chatterKeyboardSend( chat, '\n' );
      chatter.This( "\r\n" );
      break;
      
    case 7:
      execSchedule( melodies.bell );
      chatterKeyboardSend( chat, 7 );
      break;

    default:
      if( ch>=' ' && ch<256 )
      {
        chatterKeyboardSend( chat, ch );
        chatter.This( char(ch) );
      }
      break;
    }

    chatter.setCursorThis( 1 );
  }

  char *msg = "\07\07\n * Chat mode end\n";

  if( !chatTimer.isTimeOut() )
    HydraDevSend( "CON", (byte *)msg, (word)strlen(msg) );
  else
    msg = "\07\07\n * Chat mode end - timeout\n";
  
  chatter.This( "\r\n" );
  chatter.This( msg+3 );
  chatter.This( '\r' );
  chat.removeWindow();
  chat.tid = -1;  
  chat.abortKeyboard = 0;
  deRegisterExceptionHandler
}

Chatter::Chatter() :
  chatFrame(0),
  chatThis(0),
  chatFrom(0),
  tid(-1),
  curBuf( buf1 ),
  curPos( 0 ),
  abortKeyboard( 0 ),
  active( 0 )
{
  int width  = 72,
      half   = 7,
      height = half*2+3;
  int top = (Z_ScreenHeight-height)/2+1,
      left = (Z_ScreenWidth-width)/2+1;

  chatFrame = new Window( left, top, width, height, popUpColor,
                          Msg_ScrChat, BS_DOUBLELINE, DS_LOWERRIGHT );
  chatFrom  = new Window( left+1, top+1, width-2, half, popUpColor );
  chatThis  = new Window( left+1, top+1+half+1, width-2, half, popUpColor );
  chatFrom->setScroll( 1 );
  chatThis->setScroll( 1 );

  (*chatFrame)( 1, half+1 );

  for( int i(0); i<width-2; i++ )
    (*chatFrame) << char('');

  (*chatFrame)( 3, half+1 ) << Msg_ScrChatFrom;
  (*chatFrame)( width-2-2-strlen(Msg_ScrChatThis ), half+1 )
              << Msg_ScrChatThis;
}

Chatter::~Chatter()
{
  delete chatThis;
  delete chatFrom;
  delete chatFrame;
}

void Chatter::stop()
{
  if( !active )
    return;
  
  abortKeyboard = 1;
  DosWaitThread( (PULONG)&tid,  DCWW_WAIT );
  tid = -1;
  removeWindow();
}

void Chatter::removeWindow()
{
  if( !terminalActive )
    -screenCursor;
  
  chatFrame->push();
  chatFrom->push();
  chatThis->push();
  -(*chatThis);
  -(*chatFrom);
  -(*chatFrame);
  active = 0;
  chatTimer.endTimer();
}

void Chatter::start()
{
  if( active )
    return;
  
  +(*chatFrame);
  +(*chatFrom);
  +(*chatThis);
  chatFrame->pop();
  chatFrom->pop();
  chatThis->pop();
  chatter.setCursorThis( 1 );
  +screenCursor;
  active = 1;
  
  curBuf = buf1;
  curPos = 0;
  abortKeyboard = 0;
  chatTimer.startTimer( miscInfo.chatTimeout*1000 );
  tid = _beginthread( chatterKeyboard, 0, 40960, (void *)this );

  if( tid==-1 )
  {
    log( Msg_ErrChatThr, strError() );
    removeWindow();
  }
  else
    log( Msg_ErrChatThreadStarted, tid );
}

void Chatter::init()
{
  --(*chatFrom);
  --(*chatThis);
}

int Chatter::abort()
{
  if( isActive() )
  {
    if( curPos > 0 && HydraDevSend( "CON", (byte *)curBuf, curPos ) )
    {
      curBuf = curBuf==buf1 ? buf2 : buf1;
      curPos = 0;
    }
    
    return 0;
  }

  register int ch = Zabort();

  if( ch=='c' )
  {
    if( chatAvailable )
    {
      char *msg = "\07\07 * Chat mode start\n";

      start();
      HydraDevSend( "CON", (byte *)msg, (word)strlen(msg) );
      chatter.This( msg+2 );
      chatter.This( '\r' );
    }
    else
      execSchedule( melodies.bell );
    
    ch = 0;
  }

  return ch;
}
  
void HydraLog( char *prf, byte *data, word len )
{
  char *msg = (char *)_alloca( len+1 );

  memcpy( msg, data, len );
  msg[ len ] = 0;
  log( "%s: %s", prf, msg );
}

void HydraMsgDev( byte* data, word len )
{
  HydraLog( "HydraMsg", data, len  );
}

void HydraErrDev( byte *data, word len )
{
  HydraLog( "HydraErr", data, len );
}

void HydraConDev( byte *data, word len )
{
  if( !chatAvailable )
    return;

  if( !chatter.isActive() )
  {
    char *str = (char *)_alloca( len+1 );

    memcpy( str, data, len );
    str[ len ] = 0;
    
    for( NoHydraChatIterator cur( *noHydraChatList ); cur; cur++ )
      if( !stricmp( str, cur().getStr() ) )
        return;
    
    chatter.start();
  }
  
  chatTimer.startTimer( miscInfo.chatTimeout*1000 );
  
  for( int crlf(0); len; len--, data++ )
    switch( *data )
    {
    case '\b':
      crlf = 0;
      {
        int x = chatter.getCurFromX(),
            y = chatter.getCurFromY();

        if( x==1 && y==1 )
          break;

        if( x==1 )
        {
          x = chatter.getFromWindow()->getWidth();
          y--;
        }
        else
          x--;

        chatter.cursorFrom( x, y );
      }
      break;

    case '\n':
    case '\r':
      if( !crlf )
        chatter.from( "\r\n" );

      crlf = 1;
      break;

    case 7:
      execSchedule( melodies.bell );
      crlf = 0;
      break;

    default:
      chatter.from( (char)*data );
      crlf = 0;
    }

  chatter.setCursorFrom( 1 );
}

static struct _h_dev h_dev[] =
{
  { "MSG", HydraMsgDev },			// internal protocol msg
  { "CON", HydraConDev },			// text to console (chat)
  { "PRN", NULL },				// data to printer
  { "ERR", HydraErrDev },      			// text to error output
  { NULL,  NULL }
};

int HydraDevFree()
{
  if( devtxstate || !(txoptions & HOPT_DEVICE) || txstate >= HTX_END )
    return FALSE;				// busy or not allowed
  else
    return TRUE;				// allowed to send a new pkt
}

int HydraDevSend( char *dev, byte *data, word len )
{
  if( dev==NULL || data==NULL || !len || !HydraDevFree() )
    return FALSE;

  strncpy( devtxdev, dev, H_FLAGLEN );
  devtxdev[H_FLAGLEN] = '\0';
  strupr( devtxdev );
  devtxbuf = data;
  devtxlen = (word)((len > H_MAXBLKLEN) ? H_MAXBLKLEN : len);

  devtxid++;
  devtxtimer.endTimer();
  devtxretries = 0;
  hydraDisplaySendTry( devtxretries );
  devtxstate = HTD_DATA;

#ifdef PROLONGATE_CHAT
  if( chatTimer.isStarted() && !strcmp( dev,"CON" ) && txstate == HTX_REND )
    braindead.startTimer( H_BRAINDEAD );
#endif
    
  return TRUE;
}

int HydraDevFunc( char *dev, void (*func)( byte *data, word len ) )
{
  register int i;

  for( i = 0; h_dev[i].dev; i++ )
    if( !strnicmp( dev, h_dev[i].dev, H_FLAGLEN ) )
    {
      h_dev[i].func = func;
      return TRUE;
    }

  return FALSE;
}

void HydraDevRecv()
{
  register char *p = (char *) rxbuf;
  word len = rxpktlen;

  p += (int)sizeof(long);			// skip the id long
  len -= (int)sizeof(long);

  for( register int i = 0; h_dev[i].dev; i++ )
    if( !strncmp( p, h_dev[i].dev, H_FLAGLEN ) )
    {
      if( h_dev[i].func )
      {
        len -= ((int)strlen( p )) + 1;		// sub devstr len
        p += ((int) strlen( p )) + 1;		// skip devtag
        p[len] = '\0';				// NUL terminate
	(*h_dev[i].func)( (byte *) p, len );	// call output func
      }

      break;
    }
}

void hydraPutFlags( char *buf, long val )
{
  register char *p = buf;

  for( register int i = 0; h_flags[i].val; i++ )
    if( val & h_flags[i].val )
    {
      if( p > buf )
        *p++ = ',';

      strcpy( p, h_flags[i].str );
      p += H_FLAGLEN;
    }

  *p = '\0';
}

unsigned long hydraGetFlags( char *buf )
{
  register unsigned long val = 0x0L;

  for( register char *p = strtok( buf, "," ); p!=NULL; p = strtok( NULL, "," ) )
    for( register int i = 0; h_flags[i].val; i++ )
      if( !strcmp( p, h_flags[i].str ) )
      {
        val |= h_flags[i].val;
        break;
      }

  return val;
}

byte *put_binbyte( register byte *p, register byte c )
{
  register byte n = c;

  if( txoptions & HOPT_HIGHCTL )
    n &= 0x7f;

  if( n == H_DLE ||
      ((txoptions & HOPT_XONXOFF) && (n == XON || n == XOFF)) ||
      ((txoptions & HOPT_TELENET) && n == '\r' && txlastc == '@') ||
      ((txoptions & HOPT_CTLCHRS) && (n < 32 || n == 127))
    )
  {
    *p++ = H_DLE;
    c ^= 0x40;
  }

  *p++ = c;
  txlastc = n;

  return p;
}

void debugTransmit( int type, int may32crc, int len, int format,
                    unsigned long crcDebug )
{
  char *s1, *s2, *s3, *s4;
  char buf[4096];

  switch( type )
  {
  case HPKT_START:
    sprintf( buf, "START" );
    break;

  case HPKT_INIT:
    s1 = ((char *) txbufin) + ((int) strlen ((char *) txbufin)) + 1;
    s2 = s1 + ((int) strlen (s1)) + 1;
    s3 = s2 + ((int) strlen (s2)) + 1;
    s4 = s3 + ((int) strlen (s3)) + 1;
    sprintf( buf, "INIT '%s' can='%s' want='%s' opt='%s' pktprf='%s'",
         (char *)txbufin, s1, s2, s3, s4 );
    break;

   case HPKT_INITACK:
    sprintf( buf, "INITACK" );
    break;
//intell( h_long4(txbufin) )
  case HPKT_FINFO:
    sprintf( buf, "FINFO '%s', txNFI=%d (", txbufin, txFileNumber );
    {
      for( int iii( strlen( (char *)txbufin )+1 ); iii<len; iii++ )
        sprintf( buf+strlen(buf), " %02x", ( (char *)txbufin )[iii] );
    }
    strcat( buf, ")" );
    break;

  case HPKT_FINFOACK:
    if( rxfd >= 0 )
    {
      if( rxpos > 0L )
        s1 = "RES";
      else
        s1 = "BOF";
    }
    else if( rxpos == -1L )
      s1 = "HAVE";
    else if( rxpos == -2L )
      s1 = "SKIP";
    else
      s1 = "EOB";

    sprintf( buf, "FINFOACK pos=%ld/%ld %s rxst=%d rxfd=%d rxNFI=%ld/%ld",
         intell( h_long1(txbufin) ), rxpos, 
         s1, (int)rxstate, rxfd, 
         intell( h_long2(txbufin) ), rxFileNumber );
    break;

  case HPKT_DATA:
    sprintf( buf, "DATA at=%ld len=%d",
         intell( h_long1(txbufin) ), len-5 );
    break;

  case HPKT_DATAACK:
    sprintf( buf, "DATAACK at=%ld", intell( h_long1(txbufin) ) );
    break;

  case HPKT_RPOS:
    sprintf( buf, "RPOS pos=%ld/%ld%s blk=%ld sync=%ld/%ld rxNFI=%ld/%ld",
         intell( h_long1(txbufin) ), rxpos, 
         rxpos < 0L ? " SKIP" : "",
         intell( h_long2(txbufin) ), 
         intell( h_long3(txbufin) ), rxsyncid, 
         intell( h_long4(txbufin) ), rxFileNumber );
    break;

  case HPKT_EOF:
    sprintf( buf, "EOF at=%ld%s", txpos, txpos < 0L ? " SKIP" : "" );
    break;

  case HPKT_EOFACK:
    sprintf( buf, "EOFACK" );
    break;

  case HPKT_IDLE:
    sprintf( buf, "IDLE" );
    break;

  case HPKT_END:
    sprintf( buf, "END" );
    break;

  case HPKT_DEVDATA:
    sprintf( buf, "DEVDATA id=%ld dev='%s' len=%u",
         devtxid, devtxdev, devtxlen );
  break;

  case HPKT_DEVDACK:
    sprintf( buf, "DEVDACK id=%ld", intell( h_long1 (rxbuf) ) );
    break;

  default:	// This couldn't possibly happen! ;-)
    sprintf( buf, "*NoNe*" );
    break;
  }

  int buflen = strlen( buf );

  log( "?tx pkt %d:%08x fmt=%c type=%c len=%d%s%s",
       may32crc ? 32 : 16, crcDebug, (char)format, (char)type, (int)len-1,
       buflen>8 ? "" : " -> ",
       buflen>8 ? "" : buf );

  if( buflen>8 )
    log( "?       %s", buf );

  if( type != HPKT_DATA )
  {
    register byte *in( txbufin );
    Buffer deb;

    for( buflen=0; buflen<len; buflen++, in++ )
    {
      char bb[16];

      sprintf( bb, "%02x ", *in );
      deb.putString( bb );
    }

    deb.putInto( 0 );
    log( "?tx pkt %s", deb.str );
  }
}

void txpkt( register word len, int type )
{
  register byte *in,
                *out;
  register word c,
                n;
  byte format;

  txbufin[len++] = (byte)type;

  switch( type )
  {
  case HPKT_START:
  case HPKT_INIT:
  case HPKT_INITACK:
  case HPKT_END:
  case HPKT_IDLE:
    format = HCHR_HEXPKT;
    break;

  default:
		// COULD do smart format selection depending on data and options!
    if( txoptions & HOPT_HIGHBIT )
    {
      if( (txoptions & HOPT_CTLCHRS) && (txoptions & HOPT_CANUUE) )
        format = HCHR_UUEPKT;
      else if( txoptions & HOPT_CANASC )
        format = HCHR_ASCPKT;
      else
        format = HCHR_HEXPKT;
    }
    else
     format = HCHR_BINPKT;

    break;
  }

  char fuckinTelebit(0);

  if( format!=HCHR_HEXPKT && miscInfo.state1&Telebit )
  {
    int pCount( 0 );
    in = txbufin;

    for( int cnt(0); cnt<len && format!=HCHR_HEXPKT; cnt++ )
      switch( *(in++) )
      {
      case '+':
      case '+'|0x80:
        pCount++;
        break;

      case 'a':
      case 'A':
      case 'a'|0x80:
      case 'A'|0x80:
        if( pCount>=3 )
        {
          format = HCHR_HEXPKT;
          fuckinTelebit = 1;
          telebitFuckinCorrectionTimes = 1;
          log( "?txpkt: need FuckinTelebit correction" );
          break;
        }

      // fall

      default:
        pCount = 0;
      }
  }

  int may32crc(0);

  if( format != HCHR_HEXPKT && (txoptions & HOPT_CRC32) )
  {
    Crc32 = 1;
    may32crc = 1;
  }

  unsigned long crcDebug;

  if( may32crc )
  {
    unsigned long crc = CRC32POST( crc32block( hCrc32Tab, CRC32INIT, txbufin,
                                               len ) );

    crcDebug = crc;
    txbufin[ len++ ] = (byte)crc;
    txbufin[ len++ ] = (byte)(crc >> 8);
    txbufin[ len++ ] = (byte)(crc >> 16);
    txbufin[ len++ ] = (byte)(crc >> 24);
  }
  else
  {
    word crc( (word) CRC16POST( crc16block( hCrc16Tab, CRC16INIT, txbufin, len ) ) );

    crcDebug = crc;
    txbufin[len++] = (byte)crc;
    txbufin[len++] = (byte)(crc >> 8);
  }

  if( (miscInfo.state&ZDebug)==ZDebug )
    debugTransmit( type, may32crc, len, format, crcDebug );

  in = txbufin;
  out = txbuf;
  txlastc = 0;
  *out++ = H_DLE;
  *out++ = format;

  switch( format )
  {
  case HCHR_HEXPKT:
    {
      for( int pCount( 0 ); len > 0; len--, in++ )
        if( *in & 0x80 || (pCount>=3 && (*in=='a' || *in=='A') ) )
        {
          if( !(*in & 0x80) )
            log( "?txpkt: FuckinTelebit correction at %lu", in-txbufin );

          *out++ = '\\';
          *out++ = hexDigit[ ((*in) >> 4) & 0x0f ];
          *out++ = hexDigit[ (*in) & 0x0f ];
          pCount = 0;
        }
        else if( *in < 32 || *in == 127 )
        {
          *out++ = H_DLE;
          *out++ = (byte)((*in) ^ 0x40);
        }
        else if( *in == '\\' )
        {
          *out++ = '\\';
          *out++ = '\\';
        }
        else
        {
          if( fuckinTelebit )
            if( *in=='+' )
              pCount++;
            else
              pCount = 0;

          *out++ = *in;
        }
    }
    break;

  case HCHR_BINPKT:
    while( len > 0 )
    {
      out = put_binbyte( out, *in++ );
      len--;
    }

    break;

  case HCHR_ASCPKT:
    for( n = c = 0; len > 0; len-- )
    {
      c |= ((*in++) << n);
      out = put_binbyte( out, (byte)(c & 0x7f) );
      c >>= 7;

      if( ++n >= 7 )
      {
        out = put_binbyte( out, (byte)(c & 0x7f) );
 	n = c = 0;
      }
    }

    if( n > 0 )
      out = put_binbyte( out, (byte)(c & 0x7f) );

    break;

  case HCHR_UUEPKT:
    while( len >= 3  )
    {
      *out++ = (byte)h_uuenc( in[0] >> 2 );
      *out++ = (byte)h_uuenc( ((in[0] << 4) & 0x30) | ((in[1] >> 4) & 0x0f) );
      *out++ = (byte)h_uuenc( ((in[1] << 2) & 0x3c) | ((in[2] >> 6) & 0x03) );
      *out++ = (byte)h_uuenc( in[2] & 0x3f );
      in += 3; 
      len -= 3;
    }

    if( len > 0 )
    {
      *out++ = (byte)h_uuenc( in[0] >> 2 );
      *out++ = (byte)h_uuenc( ((in[0] << 4) & 0x30) | ((in[1] >> 4) & 0x0f) );

      if( len == 2 )
        *out++ = (byte)h_uuenc( (in[1] << 2) & 0x3c );

    }
    break;
  }

  *out++ = H_DLE;
  *out++ = HCHR_PKTEND;

  if( type != HPKT_DATA && format != HCHR_BINPKT )
  {
    *out++ = '\r';
    *out++ = '\n';
  }

  for( in = (byte *)txpktprefix; *in; in++ )
    switch( *in )
    {
    case 221:				// transmit break signal for one second
      comPort->do_break( TRUE );
      sleep( 1000 );
      comPort->do_break( FALSE );
      break;

    case 222:
      sleep( 1000 );
      break;

    case 223:
      comPort->putWait( 0, 1 );
      break;

    default:
      comPort->putWait( *in, 1 );
      break;
    }

  in = (byte *)txbuf;

  for( int i=0, j=out-txbuf; i<j && !braindead.isTimeOut(); i++, in++ )
  {
    comPort->bufferByte( *in );

#ifdef HYDRA_DELAY
  {
    static count = 0;

    if( ++count%25==0 )
      checkPoint();
  }
#endif
  }

  comPort->unBufferBytes();
// //  checkPoint();
}

void debugReceived( unsigned long crcDebug, int crcType )
{
  char *s1, *s2, *s3, *s4;
  char buf[4096];

  switch( rxbuf[rxpktlen] )
  {
  case HPKT_START:
    sprintf( buf, "START" );
    break;

  case HPKT_INIT:
    s1 = ((char *) rxbuf) + ((int) strlen ((char *) rxbuf)) + 1;
    s2 = s1 + ((int) strlen (s1)) + 1;
    s3 = s2 + ((int) strlen (s2)) + 1;
    s4 = s3 + ((int) strlen (s3)) + 1;
    sprintf( buf, "INIT '%s' can='%s' want='%s' opt='%s' pktpre='%s'",
         (char *) rxbuf, s1, s2, s3, s4 );
    break;

  case HPKT_INITACK:
    sprintf( buf, "INITACK" );
    break;

  case HPKT_FINFO:
    sprintf( buf, "FINFO '%s' rxs=%d, rxNFI=%ld (", 
             rxbuf, (int)rxstate, rxFileNumber );
    {
      for( int iii( strlen( (char *)rxbuf )+1 ); iii<rxpktlen; iii++ )
        sprintf( buf+strlen(buf), " %02x", ( (char *)rxbuf )[iii] );
    }
    strcat( buf, ")" );
    break;

  case HPKT_FINFOACK:
    sprintf( buf, "FINFOACK pos=%ld txs=%d txfd=%d txNFI=%d/%d",
         intell( h_long1(rxbuf)), 
         (int)txstate, txfd, 
         intell( h_long2(rxbuf)), txFileNumber );
    break;

  case HPKT_DATA:
    sprintf( buf, "DATA rxs=%d pos=%ld len=%u",
         (int)rxstate, intell( h_long1( rxbuf ) ),
         (word)(rxpktlen - ((int)sizeof(long))));
    break;

  case HPKT_DATAACK:
    sprintf( buf, "DATAACK rxs=%d pos=%ld",
         (int)rxstate, intell( h_long1(rxbuf)));
    break;

  case HPKT_RPOS:
    sprintf( buf, "RPOS pos=%ld%s blk=%hu->%ld sync=%ld%s txs=%d txfd=%d txNFI=%d/%d",
         intell( h_long1(rxbuf) ),
         intell( h_long1(rxbuf) ) < 0L ? " SKIP" : "",
         txblklen, intell( h_long2(rxbuf) ),
         intell( h_long3(rxbuf) ),
         intell( h_long3(rxbuf) ) == rxsyncid ? " DUP" : "",
         (int)txstate, txfd, 
         intell( h_long4(rxbuf)), txFileNumber );
    break;

  case HPKT_EOF:
    sprintf( buf, "EOF rxs=%d pos=%ld%s",
         (int)rxstate, intell( h_long1(rxbuf) ),
         intell( h_long1(rxbuf) ) < 0L ? " SKIP" : "" );
    break;

  case HPKT_EOFACK:
    sprintf( buf, "EOFACK txs=%d", (int)txstate );
    break;

  case HPKT_IDLE:
    sprintf( buf, "IDLE" );
    break;

  case HPKT_END:
    sprintf( buf, "END" );
    break;

  case HPKT_DEVDATA:
    s1 = ((char *) rxbuf) + ((int) sizeof (long));
    sprintf( buf, "DEVDATA id=%ld dev=%s len=%u",
         intell( h_long1(rxbuf) ), s1,
         (int)rxpktlen - (((int)sizeof(long)) + ((int)strlen(s1)) + 1) );
    break;

  case HPKT_DEVDACK:
    sprintf( buf, "DEVDACK devtxs=%d id=%ld",
         (int)devtxstate, intell( h_long1(rxbuf)) );
    break;

  default:
    sprintf( buf, "Unk pkt=%d txs=%d rxs=%d",
         rxbuf[rxpktlen], (int)txstate, (int)rxstate );
    break;
  }

  int buflen = strlen( buf );

  log( "?rx pkt %d:%08x fmt=%c type=%c len=%d%s%s",
       crcType, crcDebug, (char)rxpktformat, rxbuf[rxpktlen], rxpktlen,
       buflen>8 ? "" : " -> ",
       buflen>8 ? "" : buf );

  if( buflen>8 )
    log( "?       %s", buf );
}

int rxpkt()
{
  register byte *p,
                *q = rxbuf;
  register word n, i;
  short c;
  long stored = 0;

  switch( chatter.abort() )
  {
  case 27:
    return H_SYSABORT;

  case 's':
    return H_SKIPTHISFILE;
  }

  if( !comPort->isOnline() )
    return H_CARRIER;

  if( braindead.isStarted() && braindead.isTimeOut() )
  {
    log( "?rx BrainDead timeout: H_BRAINTIME" );
    return H_BRAINTIME;
  }

  if( txtimer.isStarted() && txtimer.isTimeOut() )
  {
    log( "?rx TxTimer timeout: H_TXTIME" );
    return H_TXTIME;
  }

  if( devtxtimer.isStarted() && devtxtimer.isTimeOut() )
  {
    log( "?rx DevTxTimer timeout: H_DEVTXTIME" );
    return H_DEVTXTIME;
  }

  p = rxbufptr;

  while( 1 )
  {
    c = (short)comPort->get();

    if( c==-1 )
      break;

#ifdef HYDRA_DELAY
  {
    static count = 0;

    if( ++count%25==0 )
      checkPoint();
  }
#endif

    if( rxoptions & HOPT_HIGHBIT )
      c &= 0x7f;

    n = c;

    if( rxoptions & HOPT_HIGHCTL )
      n &= 0x7f;

    if( n != H_DLE &&
	(((rxoptions & HOPT_XONXOFF) && (n == XON || n == XOFF)) ||
         ((rxoptions & HOPT_CTLCHRS) && (n < 32 || n == 127)) )
      )
      continue;

    if( rxdle || c == H_DLE )
    {
      switch( c )
      {
      case H_DLE:
        if( ++rxdle >= 5 )
          return H_CANCEL;

        break;

      case HCHR_PKTEND:
        rxbufptr = p;

        switch( rxpktformat )
	{
	case HCHR_BINPKT:
	  q = rxbufptr;
	  break;

	case HCHR_HEXPKT:
	  for( p = q = rxbuf; p < rxbufptr; p++ )
            if( *p == '\\' && *++p != '\\' )
            {
              i = *p;
              n = *++p;

              if( (i -= '0') > 9 )
                i -= ('a' - ':');

              if( (n -= '0') > 9 )
                n -= ('a' - ':');

              if( (i & ~0x0f) || (n & ~0x0f) )
              {
                c = H_NOPKT;
                break;
              }

              *q++ = (byte) ((i << 4) | n);
            }
            else
              *q++ = *p;

          if( p > rxbufptr )
            c = H_NOPKT;
            break;

        case HCHR_ASCPKT:
          n = i = 0;

          for( p = q = rxbuf; p < rxbufptr; p++ )
          {
            i |= ((*p & 0x7f) << n);

            if( (n += 7) >= 8 )
            {
              *q++ = (byte) (i & 0xff);
              i >>= 8;
              n -= 8;
            }
          }
          break;

        case HCHR_UUEPKT:
          n = (int) (rxbufptr - rxbuf);

          for( p = q = rxbuf; n >= 4; n -= 4, p += 4 )
          {
            if( p[0] <= ' ' || p[0] >= 'a' ||
		p[1] <= ' ' || p[1] >= 'a' ||
		p[2] <= ' ' || p[2] >= 'a' ||
		p[3] <= ' ' || p[3] >= 'a')
            {
              c = H_NOPKT;
              break;
            }

            *q++ = (byte)((h_uudec(p[0]) << 2) | (h_uudec(p[1]) >> 4));
            *q++ = (byte)((h_uudec(p[1]) << 4) | (h_uudec(p[2]) >> 2));
            *q++ = (byte)((h_uudec(p[2]) << 6) | h_uudec(p[3]));
          }

          if( n >= 2 )
          {
            if( p[0] <= ' ' || p[0] >= 'a' ||
                p[1] <= ' ' || p[1] >= 'a')
            {
              c = H_NOPKT;
              break;
            }

            *q++ = (byte)((h_uudec(p[0]) << 2) | (h_uudec(p[1]) >> 4));

            if( n == 3 )
            {
              if( p[2] <= ' ' || p[2] >= 'a' )
              {
                c = H_NOPKT;
                break;
              }

              *q++ = (byte)((h_uudec(p[1]) << 4) | (h_uudec(p[2]) >> 2));
            }
          }
          break;

        default:		// This'd mean internal fluke
          log( "?rx PKTEND pktfmt=%d ???", rxpktformat );
          c = H_NOPKT;
          break;
        }

        rxbufptr = NULL;

        if( c == H_NOPKT )
          break;

        rxpktlen = (word) (q - rxbuf);

        unsigned long crcDebug;
        int crcType;

        if( rxpktformat != HCHR_HEXPKT && (rxoptions & HOPT_CRC32) )
        {
          if( rxpktlen < 5 )
          {
            c = H_NOPKT;
            break;
          }

          unsigned long crc = crc32block( hCrc32Tab, CRC32INIT, rxbuf, rxpktlen );
          crcDebug = crc;

          crcType = 32;
          n = h_crc32test( crc );
          rxpktlen -= (int)sizeof(long);	// remove CRC-32
        }
        else
        {
          if( rxpktlen < 3 )
          {
            c = H_NOPKT;
            break;
          }

          unsigned short crc = crc16block( hCrc16Tab, CRC16INIT, rxbuf, rxpktlen );
          crcDebug = crc;

          crcType = 16;
          n = h_crc16test( crc );
          rxpktlen -= (int)sizeof(word);	// remove CRC-16
        }

        rxpktlen--;				// remove type

        if( (miscInfo.state&ZDebug)==ZDebug )
          debugReceived( crcDebug, crcType );

// //        checkPoint();

#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: rxpkt: %s, txtimer==%lu, %s (c:%lu m:%lu)",
             txtimer.isStarted() ? "str" : "end", txtimer.rest(),
             txtimer.isTimeOut() ? "out" : "wait",
             clock(), txtimer.getMark() );
#endif
        if( n )
          return ((int) rxbuf[rxpktlen]);

// badcrc??????

        log( Msg_HydraCRC, rxpos );
        log( "?rx Bad CRC fmt=%c type=%c len=%d",
             (char)rxpktformat, rxbuf[rxpktlen], (int) rxpktlen );
        break;

      case HCHR_BINPKT:
      case HCHR_HEXPKT:
      case HCHR_ASCPKT:
      case HCHR_UUEPKT:
//        log( "?rx PKTSTART fmt=%c", (char) c);
        rxpktformat = (byte) c;
        p = rxbufptr = rxbuf;
        rxdle = 0;
        break;

      default:
        c ^= 0x40;
        rxdle = 0;
StoreIt:
        if( p )
          if( p < rxbufmax )
          {
            *p++ = (byte)c;
            ++stored;
          }
          else
          {
            log( Msg_HydraTooLong );
            p = NULL;
          }

        break;
      }					// case
    }					// if( rxdle || c == H_DLE )
    else
      goto StoreIt;
  }					// for( ;; )

  rxbufptr = p;

//  checkPoint();
  return H_NOPKT;
}

void HydraInit( unsigned long want_options, int isOrig,
                int trunc, char *inbound, ReceivedFileList& xfer,
                int skipAll,
                FILE *sent )
{
  Filename = rxpathname;
  Crc32 = 0;
  hydraIsActive = 1;
  hSendStartTime = 0;
  hReceiveStartTime = 0;
  needSkipAll = skipAll;
  xferInfo = &xfer;
  sentList = sent;
  *hydraCpsOld = 0;
  timeDelimiter = '/';
  txbuf = new byte [ (H_BUFLEN+128)*2 ];
  rxbuf = txbuf + H_BUFLEN;
  txbufin = txbuf + ((H_MAXBLKLEN + H_OVERHEAD + 5) * 2);
  rxbufmax = rxbuf + H_MAXPKTLEN;

  chatAvailable = 0;
  batchesdone = 0;
  readTruncate = trunc;
  
  int inbLen( strlen(inbound) );
  
  hInbound = new char [ inbLen+2 ];
  strcpy( hInbound, inbound );
  
  if( inbLen>1 && hInbound[ inbLen-1 ]!='\\' && hInbound[ inbLen-1 ]!='/' )
    strcat( hInbound, "\\" );
  
  hCanOptions = HOPT_XONXOFF|HOPT_TELENET|HOPT_CTLCHRS|HOPT_HIGHCTL|
                HOPT_HIGHBIT|HOPT_CANBRK|HOPT_CANASC|HOPT_CANUUE|
                HOPT_CRC32|HOPT_NFI;

  if( miscInfo.state&HydraChat )
    hCanOptions |= HOPT_DEVICE;

  originator = isOrig;

  txFileNumber = 0;
  rxFileNumber = 0;

  if( originator )
  {
    hdxlink = FALSE;
    deltaFileNumber = -1;
  }
  else
  {
    hdxlink = comPort->isHalfDuplex() ? TRUE : FALSE;
    deltaFileNumber = 1;
  }
  
  rxFileNumber -= deltaFileNumber;
  txFileNumber += deltaFileNumber;

  if( !want_options )
    want_options = hydra_options;

  options = (want_options & hCanOptions) & ~HUNN_OPTIONS;

  timeout = hGetTimeOut();
  log( "?INIT Timeout=%d, sesCpsMinimum=%d", timeout, sesCpsMinimum );

  txmaxblklen = (short)((comPort->getLastBaud() / 300) * 128);

  if( txmaxblklen < 256 )
    txmaxblklen = 256;
  else if( txmaxblklen > H_MAXBLKLEN )
    txmaxblklen = H_MAXBLKLEN;

  rxblklen = txblklen = (comPort->getLastBaud() < 2400U) ? 256 : 512;

  log( "?INIT startblklen=%d, maxblklen=%d", rxblklen, txmaxblklen );

  txgoodbytes = 0;
  txgoodneeded = txmaxblklen;

  txstate = HTX_DONE;

  hydraDisplayStart();
  chatter.init();
}

void HydraDeInit()
{
  chatter.stop();
  hydraIsActive = 0;
  delete txbuf;
  Filename = NULL;
  hydraDisplayEnd();
  delete hInbound;
  hInbound = NULL;
}

void endReceive( int flag )
{
#ifndef STREAM_IO
  long pos = _tell(rxfd);
#else
  long pos = 0;
#endif

  if( rxFile!=NULL )
  {
#ifdef STREAM_IO
    pos = ftell(rxFile)
#endif

    fclose( rxFile );
    rxfd = -1;
    rxFile = NULL;
    endReceiveRecover( flag );

    if( flag==OK )
    {
      xferInfo->AddAtTail( *new ReceivedFile( realname ) );

      if( !terminalActive && !rexxTransfer )
        processImmediate( realname );
    }
  }

  if( hReceive )
    hydraDisplayDone( pos-rxStart, rxStart, 1 );
}

int Hydra( char *txpathname, char *txalias )
{
  int res = 0;
  int pkttype;
  char *p, *q;
  int i;
  struct stat f;

  delete SendFileName;
  delete SendFileAlias;

  SendFileName = newStr( txpathname );
  SendFileAlias = newStr( txalias );
  telebitFuckinCorrectionTimes = 0;

  if( txstate == HTX_DONE )
  {
    txstate = HTX_START;
    txoptions = HTXI_OPTIONS;
    txpktprefix[0] = '\0';

    rxstate = HRX_INIT;
    rxoptions = HRXI_OPTIONS;
    rxfd = -1;
    rxdle = 0;
    rxbufptr = NULL;
    rxtimer.endTimer();

    devtxid = devrxid = 0L;
    devtxtimer.endTimer();
    devtxstate = HTD_DONE;

    braindead.startTimer( H_BRAINDEAD );
  }
  else
    txstate = HTX_FINFO;

#ifdef DEBUG_IDLE
  log( "?DEBUG_IDLE: hydra(), end" );
#endif
  txtimer.endTimer();
  txretries = 0;
  hydraDisplaySendTry( txretries );

  if( txalias )
    makeFileCase( txalias );

  if( txpathname!=NULL )
  {
    makeFileCase( txpathname );
    stat( txpathname, &f );

    if( (miscInfo.state1&AdjustTime)==AdjustTime )
    {
      f.st_mtime &= ~1;
      log( "?MSDOS file-time adjust done" );
    }

    f.st_mtime -= getTimeZone();

    txfsize = f.st_size;
    txftime = f.st_mtime;
    txFile = _fsopen( txpathname, "rb", SH_DENYWR );

    if( txFile==NULL )
    {
      log( Msg_ErrUnableOpen, txpathname, strError() );
      log( "?complete: XFER_OK" );
      return XFER_OK;
    }

    txfd = fileno( txFile );

    if( isatty(txfd) )
    {
      log( Msg_ZErrIsDev, fileName( txpathname ) );
      fclose( txFile );
      txFileNumber += deltaFileNumber;
      log( "?complete: XFER_OK" );
      return XFER_OK;
    }

    if( miscInfo.outFileBufferSize )
      setvbuf( txFile, NULL, _IOFBF, miscInfo.outFileBufferSize*1024 );

    for( p = txpathname, q = txfname; *p; p++ )
      if( *q = *p, *p == '\\' || *p == ':' || *p == '/' )
        q = txfname;
      else
        q++;

    *q = '\0';

    txsyncid = 0L;
  }
  else
  {
    txfd = -1;
    txFile = NULL;
    strcpy( txfname, "" );
  }

  do
  {
    if( !terminalActive && !rexxTransfer )
    {
      int needAbort(0);

      if( sesCpsMinimum && (hSend || hReceive) )
      {
        time_t ttt( getTimeMark() ),
               txUp( startTxTime+hTransferCpsAfter ),
               rxUp( startRxTime+hTransferCpsAfter );
        unsigned c1( (hSend && ttt > txUp) ? hydraDisplaySendCps() : INT_MAX ),
                 c2( (hReceive && ttt > rxUp) ? hydraDisplayReceiveCps() : INT_MAX ),
                 cps;

        if( c1==INT_MAX )
          cps = c2;
        else if( c2==INT_MAX )
          cps = c1;
        else
          cps = ( c1 > c2 ) ? c1 : c2;

        if( cps!=INT_MAX )
          timeDelimiter = '|';
        else
          timeDelimiter = '/';

        if( cps < sesCpsMinimum )
        {
          log( "LowCPS: cps=%d s=%lu r=%lu", cps, c1, c2 );
          log( Msg_HInfoLowCPS );
          needAbort = 1;
        }
      }

      if( !needAbort )
        if( sessionLimitExceeded() )
          needAbort = 1;

      if( needAbort )
      {
        txstate = HTX_DONE;
        res = XFER_ABORT;
        break;
      }
    }

    switch( devtxstate )
    {
    case HTD_DATA:

      if( txstate > HTX_RINIT )
      {
        h_long1( txbufin ) = intell( devtxid );
        p = ((char *)txbufin) + sizeof(long);

        strcpy( p, devtxdev );
        p += H_FLAGLEN + 1;
        memcpy( p, devtxbuf, devtxlen );
        txpkt( sizeof(long) + H_FLAGLEN + 1 + devtxlen, HPKT_DEVDATA );

        devtxtimer.startTimer( (!rxstate && txstate == HTX_REND) ? timeout/2 :
                                                                   timeout );
        devtxstate = HTD_DACK;
      }
      break;

    default:
      break;
    }

//    log( "?1 txs=%d rxs=%d", rxstate );

    switch( txstate )
    {
    case HTX_START:
      comPort->putOnLine( autostr, strlen(autostr), 5000 );
      txpkt( 0, HPKT_START );
#ifdef DEBUG_IDLE
      log( "?DEBUG_IDLE: txstate==HTX_START, start( H_START )" );
#endif
      txtimer.startTimer( H_START );
      txstate = HTX_SWAIT;
      break;

    case HTX_INIT:
      p = (char *)txbufin;
      sprintf( p, "%08lx%s,%s/%s", H_REVSTAMP, M_NAME, M_VERSION, M_NUMBER );
      p += strlen(p) + 1;			// our app info & HYDRA rev
      hydraPutFlags( p, hCanOptions );		// what we CAN
      p += strlen(p) + 1;
      hydraPutFlags( p, options );		// what we WANT
      p += strlen(p) + 1;
      sprintf( p, "%08lx%08lx",	hydra_txwindow, hydra_rxwindow );
      p += strlen(p) + 1;
      strcpy( p, pktprefix );			// pkt prefix string we want
      p += strlen(p) + 1;

      txoptions = HTXI_OPTIONS;
      txpkt( (word)(((byte *)p) - txbufin), HPKT_INIT );
      txoptions = rxoptions;
#ifdef DEBUG_IDLE
      log( "?DEBUG_IDLE: txstate==HTX_INIT, start( timeout/2 )" );
#endif
      txtimer.startTimer( timeout/2 );
      txstate = HTX_INITACK;
      break;

    case HTX_FINFO:
      {
        int sizeInfo;
        
        if( txfd >= 0)
        {
          char *d( (char *)txbufin+sprintf( (char *)txbufin,
                                            "%08lx%08lx%08lx%08lx%08lx",
                                            txftime, txfsize, 0L, 0L, 0L ) ),
               *ss( newStr( fileName( SendFileAlias!=NULL ? SendFileAlias
                                                          : SendFileName ) ) ),
               *sl( newStr( ss ) ),
               *ssPtr( ss ),
               *slPtr( sl );
             
          zModemFNC( ss, readTruncate );
          makeFileCase( sl );

          while( *ss )
            *(d++) = tolower( *(ss++) );

          *(d++) = 0;
          
          while( *sl )
            *(d++) = *(sl++);

          *(d++) = 0;
          
          sizeInfo = d-(char *)txbufin;

          delete ssPtr;
          delete slPtr;
        }
        else
        {
          if( !txretries )
            log( "?HSEND: End of batch" );

//          memset( txbufin, 0, 2 );
//          sizeInfo = 2;
          *txbufin = 0;
          sizeInfo = 1;
        }
        
        if( rxoptions & HOPT_NFI )
        {
          *(long *)(((char *)txbufin)+sizeInfo) = txFileNumber;
          sizeInfo += 4;
        }
        
        txpkt( sizeInfo, HPKT_FINFO );
      }
#ifdef DEBUG_IDLE
      log( "?DEBUG_IDLE: txstate==HTX_FINFO, start( timeout : /2 )" );
#endif
      txtimer.startTimer( txretries ? timeout / 2 : timeout );
      txstate = HTX_FINFOACK;
      break;

    case HTX_XDATA:

/*
      if( comPort->hasOutputQueue()>txmaxblklen )    //JOHN:  very new
        break;
*/

      if( comPort->hasOutputQueue() )    //JOHN:  very new
        break;

      if( txpos < 0L )
        i = -1;						// Skip
      else
      {
        h_long1( txbufin ) = intell( txpos );
#ifndef STREAM_IO
        i = read( txfd, txbufin+sizeof(long), txblklen );

        if( i<0 )
#else
        i = fread( txbufin+sizeof(long), 1, txblklen, txFile );

        if( ferror(txFile) )
#endif
        {
          log( Msg_ErrUnableRead, txfname, strError() );

          if( hSend )
            hydraDisplayDone( txpos-txStart, txStart, 0 );

          fclose( txFile );
          txFileNumber += deltaFileNumber;
          txfd = -1;
          txFile = NULL;
          txpos = -2L;					// Skip
        }
      }

      if( i > 0 )
      {
        txpos += i;
        txpkt( sizeof(long)+i, HPKT_DATA );
        hydraDisplaySendPos( txpos );

        if( txblklen < txmaxblklen && (txgoodbytes += i) >= txgoodneeded )
        {
          txblklen <<= 1;

          if( txblklen >= txmaxblklen )
          {
            txblklen = txmaxblklen;
            txgoodneeded = 0;
          }

          txgoodbytes = 0;
        }

        if( txwindow && txpos>=txlastack+txwindow )
        {
#ifdef DEBUG_IDLE
          log( "?DEBUG_IDLE: txstate==HTX_XDATA&&window, start( timeout : /2 )" );
#endif
          txtimer.startTimer( txretries ? timeout/2 : timeout );
          txstate = HTX_DATAACK;
        }

        break;
      }

    // fallthrough to HTX_EOF

    case HTX_EOF:
      h_long1( txbufin ) = intell( txpos );
      txpkt( sizeof(long), HPKT_EOF );

#ifdef DEBUG_IDLE
      log( "?DEBUG_IDLE: txstate==HTX_EOF, start( timeout : /2 )" );
#endif
      txtimer.startTimer( txretries ? timeout/2 : timeout );
      txstate = HTX_EOFACK;
      break;

    case HTX_END:
      txpkt( 0, HPKT_END );
      txpkt( 0, HPKT_END );
#ifdef DEBUG_IDLE
      log( "?DEBUG_IDLE: txstate==HTX_END, start( timeout/2 )" );
#endif
      txtimer.startTimer( timeout/2 );
      txstate = HTX_ENDACK;
      break;

    default:
//      log( "?unhandled txstate" );
      break;
    }

//    log( "?2 txs=%d rxs=%d", rxstate );
    pkttype = rxpkt();
//    log( "?type=%d", pkttype );

    switch( pkttype )
    {
    case H_CARRIER:
    case H_CANCEL:
    case H_SYSABORT:
    case H_BRAINTIME:
      switch( pkttype )
      {
      case H_CARRIER:
        log( Msg_ZErrCarrierLost );
        break;

      case H_CANCEL:
        log( Msg_ZInfoCanceledRemote );
        break;

      case H_SYSABORT:
        log( Msg_ZInfoESC );
        break;

      case H_BRAINTIME:
        log( Msg_ZInfoBrain );
        break;
      }

      txstate = HTX_DONE;
      res = XFER_ABORT;
      break;

    case H_SKIPTHISFILE:
      log( Msg_ZInfoAltS );

      if( remoteMailerCode!=M_CODE &&
          remoteMailerCode!=0x112 &&		//fh-mail
          remoteMailerCode!=0xe0		//t-mail
        )
      {
        log( "?Remote is not TB! compatible" );
        rxretries = 1;
        rxsyncid++;
        hydraDisplayReceiveTry( rxretries );
        rxpos = rxfsize;
        h_long1( txbufin ) = intell( rxpos        );
        h_long2( txbufin ) = intell( 0            );
        h_long3( txbufin ) = intell( rxsyncid     );
        h_long4( txbufin ) = intell( rxFileNumber );
        txpkt( ((rxoptions & HOPT_NFI) ? 4 : 3)*sizeof(long), HPKT_RPOS );
        rxtimer.startTimer( timeout );
        break;
      }
      else
      {
        log( "?Remote is full TB! compatible" );
        rxpos = -1L;
        goto SkipThisFile;
      }

    case H_TXTIME:
      if( txstate == HTX_XWAIT || txstate == HTX_REND )
      {
#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: rxpkt()==H_TXTIME, send, start" );
#endif
        txpkt( 0, HPKT_IDLE );
        txtimer.startTimer( H_IDLE );
        break;
      }

      if( ++txretries > H_RETRIES )
      {
        log( Msg_ZErrTimeOut );
//        log( Msg_ZErrManyError );
        txstate = HTX_DONE;
        res = XFER_ABORT;
        break;
      }

      hydraDisplaySendTry( txretries );

      log( "?Timeout at retry %u", txretries );

#ifdef DEBUG_IDLE
      log( "?DEBUG_IDLE: rxpkt()==H_TXTIME, end" );
#endif
      txtimer.endTimer();

      switch( txstate )
      {
      case HTX_SWAIT:
        txstate = HTX_START;
        break;

      case HTX_INITACK:
        txstate = HTX_INIT;
        break;

      case HTX_FINFOACK:
        txstate = HTX_FINFO;
        break;

      case HTX_DATAACK:
        txstate = HTX_XDATA;
        break;

      case HTX_EOFACK:
        txstate = HTX_EOF;
        break;

      case HTX_ENDACK:
        txstate = HTX_END;
        break;
      }
      break;

    case H_DEVTXTIME:
      if( ++devtxretries > H_RETRIES )
      {
        log( Msg_ZErrDevTimeOut );
        txstate = HTX_DONE;
        res = XFER_ABORT;
        break;
      }

      hydraDisplaySendTry( devtxretries );

      log( "?Device timeout at retry %u", devtxretries );
      devtxtimer.endTimer();
      devtxstate = HTD_DATA;
      break;

    case HPKT_START:
      if( txstate == HTX_START || txstate == HTX_SWAIT )
      {
#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: rxpkt()==HPKT_START, end" );
#endif
        txtimer.endTimer();
        txretries = 0;
        hydraDisplaySendTry( txretries );
        txstate = HTX_INIT;
        braindead.startTimer( H_BRAINDEAD );
      }
      break;

    case HPKT_INIT:
      if( rxstate == HRX_INIT )
      {
        p = (char *) rxbuf;
        p += strlen(p) + 1;
        q = p + strlen(p) + 1;
        rxoptions = options | HUNN_OPTIONS;
        rxoptions |= hydraGetFlags( q );
        rxoptions &= hydraGetFlags( p );
        rxoptions &= hCanOptions;

        if( rxoptions < (options & HNEC_OPTIONS) )
        {
          log( Msg_HydraIncomp );
          txstate = HTX_DONE;
          res = XFER_ABORT;
          break;
        }

        chatAvailable = rxoptions & HOPT_DEVICE ? 1 : 0;

        p = q + strlen(q) + 1;
        rxwindow = txwindow = 0L;
        sscanf( p, "%08lx%08lx", &rxwindow, &txwindow );

        if( rxwindow<0 )
          rxwindow = 0;

        if( hydra_rxwindow && (!rxwindow || hydra_rxwindow < rxwindow) )
          rxwindow = hydra_rxwindow;

        if( txwindow<0 )
          txwindow = 0;

        if( hydra_txwindow && (!txwindow || hydra_txwindow < txwindow) )
          txwindow = hydra_txwindow;

        p += strlen(p) + 1;

        strncpy( txpktprefix, p, H_PKTPREFIX );
        txpktprefix[H_PKTPREFIX] = '\0';

        if( !batchesdone )
        {
          long revstamp;

          p = (char *)rxbuf;
          sscanf( p, "%08lx", &revstamp );
//          log( Msg_HydraRev, h_revdate(revstamp) );
          p += 8;
          q = strchr( p, ',' );

          if( q!=NULL )
            *q = ' ';

          q = strchr( p, ',' );

          if( q!=NULL)
            *q = '/';

          log( Msg_HydraApp, p );
          hydraPutFlags( (char *)rxbuf, rxoptions );
          log( Msg_HydraLink, rxbuf );

          if( txwindow || rxwindow )
            log( Msg_HydraTxRxWindow, txwindow, rxwindow );

        }

        txoptions = rxoptions;
        rxstate = HRX_FINFO;

        char *msg = chatAvailable ? " * Remote has chat facility with bell enabled\n" :
                                    " * Remote has no chat facility available\n";
        if( !batchesdone )
          HydraDevSend( "CON", (byte *)msg, strlen(msg) );
      }

      txpkt( 0, HPKT_INITACK );
      break;

    case HPKT_INITACK:
      if( txstate == HTX_INIT || txstate == HTX_INITACK )
      {
        braindead.startTimer( H_BRAINDEAD );
#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: rxpkt()==HPKT_INITACK, end" );
#endif
        txtimer.endTimer();
        txretries = 0;
        hydraDisplaySendTry( txretries );
        txstate = HTX_RINIT;
      }
      break;

    case HPKT_FINFO:
      if( rxstate == HRX_FINFO )
      {
        braindead.startTimer( H_BRAINDEAD );

        if( !rxbuf[0] )
        {
          log( "?End of batch" );
          rxpos = 0;
          rxstate = HRX_DONE;
          batchesdone++;
        }
        else
        {
          char *rxbufChar( (char *)rxbuf );  //type conversion

//get file info

          rxfsize = rxftime = 0;
          sscanf( rxbufChar, "%08lx%08lx", //%*08lx%*08lx%*08lx
                  &rxftime, &rxfsize );

//get short file name

          char *fileNamePosition( rxbufChar+8*5 );
          
          strncpy( rxfname, fileNamePosition, sizeof(rxfname)-1 );
          rxfname[ sizeof(rxfname)-1 ] = 0;
          
          log( "?Get SFN: %s -> %s", fileNamePosition, rxfname );

          fileNamePosition += strlen( fileNamePosition )+1;

          int restSize( rxpktlen-(fileNamePosition-rxbufChar) );

          log( "?Rest pkt size: %d", restSize );

//Get NFI

          if( rxoptions & HOPT_NFI )
          {
            rxFileNumber = *(long *)( rxbufChar+rxpktlen-4 );
            restSize -= 4;
            log( "?NFI: %d, LFN size: %d", rxFileNumber, restSize );
          }
          else
            log( "?NFI: NONE, LFN size: %d", restSize );

//Get LFN
          if( restSize )            
            if( (miscInfo.state3&HPFS)==HPFS )
            {
              strncpy( rxfname, fileNamePosition, sizeof(rxfname)-1 );
              rxfname[ sizeof(rxfname)-1 ] = 0;
          
              log( "?Get LFN name: %s -> %s", fileNamePosition, rxfname );
            }
            else
              log( "?No HPFS, ignore LFN: %s", fileNamePosition );
          else
            log( "?No LFN present" );

// post process
          
          log( "?Hydra: for us %s 0x%08x %lu", rxfname, rxftime, rxfsize );
          makeFileCase( rxfname );

// i don't know what is it ;)
//          i = strlen( rxfname ) - 1;

          zModemFNC( (char *)rxfname, (miscInfo.state3&HPFS)!=HPFS );

          strcpy( Filename, hInbound );
          strcat( Filename, fileName( rxfname ) );
          makeFileCase( Filename );

          log( "?Hydra: store to %s", Filename );

          unsigned long diskfree = zfree( Filename );

          filetime = rxftime;
          file_length = rxfsize;

          if( (miscInfo.state1&AdjustTime)==AdjustTime )
          {
            filetime &= ~1;
            log( "?MSDOS file-time adjust done" );
          }

          byte *openMode;
          int recoverMode( OK );

          if( needSkipAll )
          {
            log( "? needSkipAll is active" );
            recoverMode = ZSKIP;
          }
          else
            recoverMode = startReceiveRecover( &openMode );

          switch( recoverMode )
          {
          case ZSKIP:
            rxpos = -1;
            break;

          default:
	    log( "Hydra-BUG: unknown recover state %d", recoverMode );

          // fall

          case ERROR:
            rxpos = -2;
            break;

          case OK:
            makeFilePath( Filename );
            rxFile = _fsopen( Filename, (char *)openMode, SH_DENYWR );

            if( rxFile==NULL )
            {
              log( Msg_ErrUnableOpen, Filename, strError() );
              endReceive( ERROR );
              rxpos = -2;
              break;
            }

            if( miscInfo.inFileBufferSize )
              setvbuf( rxFile, NULL, _IOFBF, miscInfo.inFileBufferSize*1024 );

            rxfd = fileno( rxFile );

            if( isatty( rxfd ) )
            {
              log( Msg_ZErrIsDev, fileName( Filename ) );
              endReceive( ERROR );
              rxpos = -2;
              break;
            }

            rxpos = rxStart = (Resume_WaZOO) ? filelength( rxfd ) : 0;

#ifndef STREAM_IO
            if( lseek( rxfd, rxStart, SEEK_SET ) < 0 )
#else
            if( fseek( rxFile, rxStart, SEEK_SET ) < 0 )
#endif
            {
              log( Msg_ErrUnableSeek, rxpathname, 0, strError() );
              endReceive( ERROR );
              rxpos = -2;
              break;
            }

            if( ((unsigned long)file_length)+10240u > diskfree && 
                ((unsigned long)file_length)+10240u-rxStart > diskfree )
            {
              log( Msg_ZErrNoSpace, rxfname, diskfree, file_length-rxStart );
              endReceive( ERROR );
              rxpos = -2;
              break;
            }

            hydraDisplayReceiveName( wasFileName, file_length, rxStart );

            rxtimer.endTimer();
            rxretries = 0;
            hydraDisplayReceiveTry( rxretries );
            rxlastsync = 0;
            rxsyncid = 0;

            if( rxStart>0 )
            {
              char *ss = strNumber( rxpos );

              log( Msg_ZReceiveFrom, ss );
              delete ss;
            }

            rxstate = HRX_DATA;
          }					// switch( recoverMode )
        }					// !rxbuf[0]
      }						// rxstate == HRX_FINFO
      else if( rxstate == HRX_DONE )
        rxpos = (!rxbuf[0]) ? 0L : -2L;

      h_long1( txbufin ) = intell( rxpos );
      h_long2( txbufin ) = intell( rxFileNumber );
      txpkt( ((rxoptions & HOPT_NFI) ? 2 : 1)*sizeof(long), HPKT_FINFOACK );
      break;

    case HPKT_FINFOACK:
      if( !txfname && (rxoptions & HOPT_NFI) && txFileNumber!=intell( h_long2(rxbuf) ) )
      {
        log( "?FINFOACK not in sequence %d/%d", 
             intell( h_long2(rxbuf) ), txFileNumber );
        break;
      }
      
      if( txstate == HTX_FINFO || txstate == HTX_FINFOACK )
      {
        braindead.startTimer( H_BRAINDEAD );
        txretries = 0;
        hydraDisplaySendTry( txretries );
          
        if( !txfname[0] )
        {
#ifdef DEBUG_IDLE
          log( "?DEBUG_IDLE: rxpkt()==HPKT_FINFOACK&&!txfname[0], start" );
#endif
          txtimer.startTimer( H_IDLE );
          txstate = HTX_REND;
        }
        else
        {
#ifdef DEBUG_IDLE
          log( "?DEBUG_IDLE: rxpkt()==HPKT_FINFOACK&&txfname[0], end" );
#endif
          txtimer.endTimer();
          txpos = intell( h_long1(rxbuf) );

          if( txpos>=0 )
          {
            txlastack = txpos;
            txStart   = txpos;

            hydraDisplaySendName( SendFileName, txfsize, SendFileAlias,
                                  txStart );
            if( txpos>0 )
            {
              char *ss = strNumber( txpos );
              log( Msg_ZInfoSendFrom, ss );
              delete ss;

#ifndef STREAM_IO
              if( lseek( txfd, txpos, SEEK_SET )<0 )
#else
              if( fseek( txFile, txpos, SEEK_SET )<0 )
#endif
              {
                log( Msg_ErrUnableSeek, txfname, txpos, strError() );

                if( hSend )
                  hydraDisplayDone( txpos-txStart, txStart, 0 );

                fclose( txFile );
                txFileNumber += deltaFileNumber;
                txfd = -1;
                txFile = NULL;
                txpos = -2L;
                txstate = HTX_EOF;
                break;
              }
            }

            txstate = HTX_XDATA;
          }
          else					// now txpos<0
          {
#ifndef STREAM_IO
            long pos = _tell( txfd );
#else
            long pos = ftell( txFile );
#endif
            int rc;

            fclose( txFile );
            txFileNumber += deltaFileNumber;
            txfd = -1;
            txFile = NULL;

            switch( txpos )
            {
            case -1:
              log( Msg_ZInfoSkipped, txfname );
              rc = XFER_OK;
              break;

            case -2:
              log( Msg_ZInfoZFERR, txfname );
              rc = XFER_SKIP;
              break;

            default:
              log( "Hydra: Unknown FINFOACK positon %d for %s",txpos,  txfname );
              rc = XFER_SKIP;
            }

            if( hSend )
              hydraDisplayDone( pos-txStart, txStart, 0 );

            log( "?complete: rc=%d", rc );
            return rc;
          }			// txpos>=0
        }			// !txfname[0]
      }				// txstate==HTX_FINFO || txstate==HTX_FINFOACK
      break;

    case HPKT_DATA:
      if( rxstate == HRX_DATA )
      {
        if( intell(h_long1(rxbuf))!=rxpos || intell(h_long1(rxbuf))<0 )
        {
          log( "?bad pos rx/want %ld/%ld", h_long1(rxbuf), rxpos );

          if( intell(h_long1 (rxbuf)) <= rxlastsync )
          {
            rxtimer.endTimer();
            rxretries = 0;
            hydraDisplayReceiveTry( rxretries );
          }

          rxlastsync = intell( h_long1(rxbuf) );

          if( !rxtimer.isStarted() || rxtimer.isTimeOut() )
          {
            if( rxretries>4 )
              if( txstate<HTX_REND && !originator && !hdxlink )
              {
                hdxlink = TRUE;
                rxretries = 0;
                hydraDisplayReceiveTry( rxretries );
              }

            if( ++rxretries > H_RETRIES )
            {
              log( Msg_ZErrManyError );
              txstate = HTX_DONE;
              res = XFER_ABORT;
              break;
            }

            hydraDisplayReceiveTry( rxretries );

            if( rxretries==1 || rxretries==4 )
              rxsyncid++;

            rxblklen >>= 1;
            i = rxblklen;

            if( i<=64 )
              i = 64;
            else if( i<=128 )
              i = 128;
            else if( i<=256 )
              i = 256;
            else if( i<=512 )
              i = 512;
            else
              i = 1024;

            log( Msg_HydraBadPacket, rxpos );
            log( "?ill pkt at %ld, retry %u, newblklen=%u",
                 rxpos, rxretries, i );

            h_long1( txbufin ) = intell( rxpos    );
            h_long2( txbufin ) = intell( (long)i  );
            h_long3( txbufin ) = intell( rxsyncid );
            h_long4( txbufin ) = intell( rxFileNumber );
            txpkt( ((rxoptions & HOPT_NFI) ? 4 : 3)*sizeof(long), HPKT_RPOS );

            rxtimer.startTimer( timeout );
          }
        }		//intell(h_long1(rxbuf))!=rxpos || intell(h_long1(rxbuf))
        else
        {
          braindead.startTimer( H_BRAINDEAD );
          rxpktlen -= sizeof(long);

          rxblklen = rxpktlen;
#ifndef STREAM_IO
          if( write( rxfd, rxbuf+sizeof(long), rxpktlen )<0 )
#else
          if( fwrite( rxbuf+sizeof(long), 1, rxpktlen, rxFile )!=rxpktlen )
#endif
          {
            log( Msg_ErrUnableWrite, rxfname, strError() );
            rxpos = -2L;
SkipThisFile:
            endReceive( ERROR );
            rxretries = 1;
            hydraDisplayReceiveTry( rxretries );
            rxsyncid++;
            h_long1( txbufin ) = intell( rxpos    );
            h_long2( txbufin ) = intell( 0        );
            h_long3( txbufin ) = intell( rxsyncid );
            h_long4( txbufin ) = intell( rxFileNumber );
            txpkt( ((rxoptions & HOPT_NFI) ? 4 : 3)*sizeof(long), HPKT_RPOS );
            rxtimer.startTimer( timeout );
            break;
          }

          rxretries = 0;
          hydraDisplayReceiveTry( rxretries );
          rxtimer.endTimer();
          rxlastsync = rxpos;
          rxpos += rxpktlen;
          hydraDisplayReceivePos( rxpos );

          if( rxwindow )
          {
            h_long1( txbufin ) = intell( rxpos );
            txpkt( sizeof(long), HPKT_DATAACK );
          }
        }				// badpkt
      }					// rxstate==HRX_DATA
      break;

    case HPKT_DATAACK:
      if( txstate == HTX_XDATA || txstate == HTX_DATAACK ||
          txstate == HTX_XWAIT || txstate == HTX_EOF || txstate == HTX_EOFACK )
        if( txwindow && intell( h_long1(rxbuf) ) > txlastack )
        {
          txlastack = intell( h_long1(rxbuf) );

          if( txstate == HTX_DATAACK &&	txpos < txlastack+txwindow )
          {
            txstate = HTX_XDATA;
            txretries = 0;
            hydraDisplaySendTry( txretries );
#ifdef DEBUG_IDLE
            log( "?DEBUG_IDLE: rxpkt()==HPKT_DATAACK&&window, end" );
#endif
            txtimer.endTimer();
          }
        }
      break;

    case HPKT_RPOS:
      if( (rxoptions & HOPT_NFI) && txFileNumber!=intell( h_long4(rxbuf) ) )
      {
        log( "?RPOS not in sequence %d/%d", 
             intell( h_long4(rxbuf) ), txFileNumber );
        break;
      }
        
      if( txstate == HTX_XDATA || txstate == HTX_DATAACK ||
          txstate == HTX_XWAIT || txstate == HTX_EOF || txstate == HTX_EOFACK )
      {
        if( intell( h_long3(rxbuf) ) != txsyncid )
        {
          txsyncid = intell( h_long3(rxbuf) );
          txretries = 1;
          hydraDisplaySendTry( txretries );
        }
        else
        {
          if( ++txretries > H_RETRIES )
          {
            log( Msg_ZErrManyError );
            txstate = HTX_DONE;
            res = XFER_ABORT;
            break;
          }

          hydraDisplaySendTry( txretries );

          if( txretries!=4 )
            break;
        }

#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: rxpkt()==HPKT_RPOS, end" );
#endif
        txtimer.endTimer();
        txpos = intell( h_long1(rxbuf) );

        if( txpos<0 )
        {
          if( txfd >= 0 )
          {
            log( Msg_ZInfoSkipRemote );

            if( hSend )
#ifndef STREAM_IO
              hydraDisplayDone( _tell(txfd)-txStart, txStart, 0 );
#else
              hydraDisplayDone( ftell(txFile)-txStart, txStart, 0 );
#endif

            fclose( txFile );
            txFileNumber += deltaFileNumber;
            txfd = -1;
            txFile = NULL;
            txstate = HTX_EOF;
          }

//          txpos = -2L;		// for skip
          break;
        }

        if( txblklen > (word)intell( h_long2(rxbuf) ) )
          txblklen = (word)intell( h_long2(rxbuf) );
        else
          txblklen >>= 1;

        if( txblklen <= 64 )
          txblklen = 64;
        else if( txblklen <= 128 )
          txblklen = 128;
        else if( txblklen <= 256 )
          txblklen = 256;
        else if( txblklen <= 512 )
          txblklen = 512;
        else
          txblklen = 1024;

        txgoodbytes = 0;
        txgoodneeded += txmaxblklen*2;

        if( txgoodneeded > txmaxblklen*8 )
          txgoodneeded = txmaxblklen*8;

        {
          char *ss = strNumber( txpos );

          log( Msg_ZInfoResendFrom, ss );
          hydraDisplaySendPos( txpos );
          delete ss;
        }

#ifndef STREAM_IO
        if( lseek( txfd, txpos, SEEK_SET ) < 0 )
#else
        if( fseek( txFile, txpos, SEEK_SET ) < 0 )
#endif
        {
          log( Msg_ErrUnableSeek, txfname, txpos, strError() );

          if( hSend )
            hydraDisplayDone( txpos-txStart, txStart, 0 );

          fclose( txFile );
          txFileNumber += deltaFileNumber;
          txfd = -1;
          txFile = NULL;
          txpos = -2L;
          txstate = HTX_EOF;
          break;
        }

        if( txstate != HTX_XWAIT )
          txstate = HTX_XDATA;
      }					// txstate
      break;

    case HPKT_EOF:
      if( rxstate == HRX_DATA )
      {
        if( intell( h_long1(rxbuf) ) < 0 )
        {
          endReceive( ERROR );
          log( Msg_InfoSkipOnLine, rxfname );
          rxstate = HRX_FINFO;
          braindead.startTimer( H_BRAINDEAD );
        }
	else if( intell( h_long1(rxbuf) ) != rxpos )
        {
          if( intell( h_long1(rxbuf) ) <= rxlastsync )
          {
            rxtimer.endTimer();
            rxretries = 0;
            hydraDisplayReceiveTry( rxretries );
          }

          rxlastsync = intell( h_long1(rxbuf) );

          if( !rxtimer.isStarted() || rxtimer.isTimeOut() )
          {
            if( ++rxretries > H_RETRIES )
            {
              log( Msg_ZErrManyError );
              txstate = HTX_DONE;
              res = XFER_ABORT;
              break;
            }

            hydraDisplayReceiveTry( rxretries );

            if( rxretries == 1 || rxretries == 4 )
              rxsyncid++;

            rxblklen >>= 1;
            i = rxblklen;

            if( i <= 64 )
              i = 64;
            else if( i <= 128 )
              i = 128;
            else if( i <= 256 )
              i = 256;
            else if( i <= 512 )
              i = 512;
            else
              i = 1024;

            log( Msg_ZErrUnExpEOF );
            log( "?bad EOF at %ld, retry %u, newblklen=%u",
                 rxpos, rxretries, i );

            h_long1( txbufin ) = intell( rxpos    );
            h_long2( txbufin ) = intell( (long)i  );
            h_long3( txbufin ) = intell( rxsyncid );
            h_long4( txbufin ) = intell( rxFileNumber );
            txpkt( ((rxoptions & HOPT_NFI) ? 4 : 3)*sizeof(long), HPKT_RPOS );

            rxtimer.startTimer( timeout );
          }					// rxtimeout
        }
        else					// ok EOF pos
        {
          endReceive( OK );
          rxstate = HRX_FINFO;
          braindead.startTimer( H_BRAINDEAD );
        }					// skip/badeof/eof
      }						//rxstate==HRX_DATA

      if( rxstate == HRX_FINFO )
        txpkt( 0, HPKT_EOFACK );

      break;

    case HPKT_EOFACK:
      if( txstate == HTX_EOF || txstate == HTX_EOFACK )
      {
        braindead.startTimer( H_BRAINDEAD );

        if( txfd >= 0 )
        {
          fclose( txFile );
          txFileNumber += deltaFileNumber;
          txfd = -1;
          txFile = NULL;

          if( hSend )
            hydraDisplayDone( txfsize-txStart, txStart, 0 );

          if( sentList!=NULL )
          {
            fprintf( sentList, "%s %s as %s at %lu %d %s\n",
                     msgOfFileType( SendFileAlias!=NULL ? SendFileAlias :
                                                          SendFileName, NULL ),
                     SendFileName,
                     fileName( SendFileAlias!=NULL ? SendFileAlias :SendFileName ),
                     time( NULL ),
                     miscInfo.taskNumber,
                     isProtocol( HydraSes ) );
            real_flush( sentList );
          }

          log( "?complete: XFER_OK" );
          return XFER_OK;
        }
        else
        {
          if( txpos==-1 )			// for skip
          {
            log( "?complete: XFER_OK and skip" );
            return XFER_OK;
          }

          log( "?complete: XFER_SKIP" );
          return XFER_SKIP;
        }
      }
      break;

    case HPKT_IDLE:
      if( txstate == HTX_XWAIT )
      {
#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: rxpkt()==HPKT_IDLE, end" );
#endif
        hdxlink = FALSE;
        txtimer.endTimer();
        txretries = 0;
        hydraDisplaySendTry( txretries );
        txstate = HTX_XDATA;
      }
      else if( txstate >= HTX_FINFO && txstate < HTX_REND )
        braindead.startTimer( H_BRAINDEAD );

      break;

    case HPKT_END:
      if( txstate == HTX_END || txstate == HTX_ENDACK )
      {
        txpkt( 0, HPKT_END );
        txpkt( 0, HPKT_END );
        txpkt( 0, HPKT_END );
        log( "?done" );
        txstate = HTX_DONE;
        res = XFER_OK;
      }
      break;

    case HPKT_DEVDATA:
      if( devrxid != intell( h_long1(rxbuf) ) )
      {
        HydraDevRecv();
        devrxid = intell( h_long1(rxbuf) );
      }

      h_long1( txbufin ) = h_long1( rxbuf );
      txpkt( sizeof(long), HPKT_DEVDACK );
      break;

    case HPKT_DEVDACK:
      if( devtxstate && devtxid == intell( h_long1(rxbuf) ) )
      {
        devtxtimer.endTimer();
        devtxstate = HTD_DONE;
      }
      break;

    default:
      if( pkttype!=H_NOPKT )
        log( "?unknown packet" );
      break;
    }						// switch(pkttype)

    switch( txstate )
    {
    case HTX_START:
    case HTX_SWAIT:
      if( rxstate == HRX_FINFO )
      {
        txstate = HTX_INIT;
        txretries = 0;
        hydraDisplaySendTry( txretries );
#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: txstate==HTX_START|HTX_SWAIT, end" );
#endif
        txtimer.endTimer();
      }
      break;

    case HTX_RINIT:
      if( rxstate == HRX_FINFO )
      {
        txstate = HTX_FINFO;
        txretries = 0;
        hydraDisplaySendTry( txretries );
#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: txstate==HTX_RINIT, end" );
#endif
        txtimer.endTimer();
      }
      break;

    case HTX_XDATA:
      if( rxstate && hdxlink )
      {
        static char msg[] = "Remote Hydra now in one way mode";

        HydraDevSend( "MSG", (byte *)msg, strlen(msg) );
        log( Msg_HydraOneWay );
        txstate = HTX_XWAIT;
#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: txstate==HTX_XDATA, start" );
#endif
        txtimer.startTimer( H_IDLE );
      }
      break;

    case HTX_XWAIT:
      if( !rxstate )
      {
        txstate = HTX_XDATA;
        txretries = 0;
        hydraDisplaySendTry( txretries );
#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: txstate==HTX_XWAIT, end" );
#endif
        txtimer.endTimer();
      }
      break;

    case HTX_REND:
      if( !rxstate && !devtxstate )
      {

#ifdef PROLONGATE_CHAT
        if( chatTimer.isStarted() )
          break;
#endif
        
        txstate = HTX_END;
        txretries = 0;
        hydraDisplaySendTry( txretries );
#ifdef DEBUG_IDLE
        log( "?DEBUG_IDLE: txstate==HTX_REND, end" );
#endif
        txtimer.endTimer();
      }
      break;

    default:
      break;
    }						// switch(txstate)

  } while( txstate );

  if( res == XFER_ABORT )
  {
    comPort->clearout();
    comPort->clearin();

    if( comPort->isOnline() )
    {
      braindead.startTimer( 10000 );		// wait max. 10s after abort
      comPort->putOnLine( abortstr, strlen(abortstr), braindead.rest() );

      while( !comPort->outEmpty() && comPort->isOnline() &&
             !braindead.isTimeOut() )
        checkPoint();

      comPort->clearout();
    }

    if( !terminalActive )
      comPort->hangUp();

    comPort->clearout();
    comPort->clearin();
  }
  else
  {
    braindead.startTimer( 10000 );		// wait max. 10s after abort

    while( !comPort->outEmpty() && comPort->isOnline() &&
           !braindead.isTimeOut() )
      checkPoint();
  }

  if( txfd>=0 )
  {
    if( hSend )
#ifndef STREAM_IO
      hydraDisplayDone( _tell(txfd)-txStart, txStart, 0 );
#else
      hydraDisplayDone( ftell(txFile)-txStart, txStart, 0 );
#endif

    fclose( txFile );
    txFileNumber += deltaFileNumber;
    txfd = -1;
    txFile = NULL;
  }

  endReceive( ERROR );

  log( "?complete: rc=%d", res );
  return res;
}

int HydraBatch( char *name, char *inbound,
                ReceivedFileList& list,
                CallBackFunc callBack, ZModemRexxCallBackContext *context )
{
  char *trnName = (char *)_alloca( strlen(name)+2 ),
       *ori( trnName );

  changeBlankWithZero( trnName, name );

  int zret( XFER_SKIP );

  char drive[_MAX_DRIVE ],
       dir  [_MAX_DIR ],
       fname[_MAX_FNAME],
       ext  [_MAX_EXT],
       full [_MAX_PATH];

  log( "?HydraBatch: ou=%s in=%s", name, inbound );

  HydraInit( 0, 1, 0, inbound, list, 0, NULL );

  for( trnName=ori; *trnName && zret!=ERROR; trnName += strlen(trnName)+1 )
  {
    char action( 0 );
    
    if( *trnName=='^' )
    {
      action = '^';
      trnName++;
    }
    else if( *trnName=='#' )
    {
      action = '#';
      trnName++;
    }
    else if( *trnName=='~' )
      continue;

    Directory rc( trnName );

    if( !rc )
      log( Msg_TrmUpLoadNotFound, trnName );
    else
      do
      {
	_splitpath( trnName, drive, dir, fname, ext );
	strcpy( full, drive);
	strcat( full, dir );
	strcat( full, rc.getName() );
	makeFileCase( full );

	zret=Hydra( full, NULL );

        if( zret==XFER_OK )
        {
          if( callBack )
            callBack( context, full );

          postProcessAttach( action, full );
        }
        
        rc++;

      } while( zret!=XFER_ABORT && rc );
  }

  if( zret!=XFER_ABORT )
    Hydra( NULL, NULL );

  HydraDeInit();

  return zret;
}
