#ifndef __HYDRA_H

#define __HYDRA_H

#include <stdio.h>

#include "class.h"
#include "timeout.h"
#include "window.h"
#include "rcvdlist.h"
#include "zmodem.h"

#define XFER_ABORT    (-1)		// Failed on this file & abort xfer
#define XFER_SKIP      (0)		// Skip this file but continue xfer
#define XFER_OK        (1)		// File was sent, continue transfer

void HydraInit( unsigned long wantOptions, int isOrig, int trunc,
                char *inbound, ReceivedFileList& xfer,
                int skipAll, FILE *sent );
void HydraDeInit();
int Hydra( char *txpathname, char *txalias );
int HydraBatch( char *name, char *inbound,
                ReceivedFileList& list,
                CallBackFunc =0, ZModemRexxCallBackContext* =0 );

void hydraPutFlags( char *buf, long val );
unsigned long hydraGetFlags( char *buf );

extern unsigned long hydra_options;
extern long hydra_txwindow;
extern long hydra_rxwindow;
extern int hydra_timeout;
extern long hTransferCpsAfter;
extern int sesCpsMinimum;
extern char hydraIsActive;

#define ChatBufferSize 256

class Chatter
{
  Window *chatFrame,
         *chatThis,
         *chatFrom;

  int tid;
  char abortKeyboard;
  char active;

  char buf1[ ChatBufferSize ];
  char buf2[ ChatBufferSize ];
  char *curBuf;
  int  curPos;

  void removeWindow();

public:
  Chatter();
  ~Chatter();
  
  void init();
  void stop();
  void start();
  int abort();
  
  friend void _Optlink chatterKeyboard( void* );
  friend void chatterKeyboardSend( Chatter&, int );

  void cursorFrom( int x, int y ) { if( chatFrom ) (*chatFrom)( x, y ); }
  void cursorThis( int x, int y ) { if( chatThis ) (*chatThis)( x, y ); }
  void setCursorFrom( int fl ) { if( chatFrom ) chatFrom->setPhysicalCursor( fl ); }
  void setCursorThis( int fl ) { if( chatThis ) chatThis->setPhysicalCursor( fl ); }
  void from( char c ) { if( chatFrom ) (*chatFrom) << c; }
  void This( char c ) { if( chatThis ) (*chatThis) << c; }
  void from( char *str ) { if( chatFrom ) (*chatFrom) << str; }
  void This( char *str ) { if( chatThis ) (*chatThis) << str; }

  int isActive() const { return active; }
  int getCurFromX() const { return chatFrom ? chatFrom->getCurrentX() : 0; }
  int getCurFromY() const { return chatFrom ? chatFrom->getCurrentY() : 0; }
  int getCurThisX() const { return chatThis ? chatThis->getCurrentX() : 0; }
  int getCurThisY() const { return chatThis ? chatThis->getCurrentY() : 0; }

  Window *getFromWindow() const { return chatFrom; }
  Window *getThisWindow() const { return chatThis; }
};

extern Chatter *chatterPointer;
#define chatter (*chatterPointer)

class NoHydraChat : public BaseObject
{
private:
  char *str;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:

  NoHydraChat( char *f );
  ~NoHydraChat();

  char *getStr() const { return str; }
};

class NoHydraChatList : public Collection		//NoHydraChat
{
public:
  NoHydraChatList();

  NoHydraChat* Find( NoHydraChat& item, LockMode lock=LockModeNeed ) const;
  NoHydraChat& operator [] ( const int pos ) const;
};

class NoHydraChatIterator : public CollectionIterator	//NoHydraChat
{
public:
  NoHydraChatIterator( NoHydraChatList& ini );

  NoHydraChat& operator () () const;
};

extern NoHydraChatList *noHydraChatList;

#endif