#define INCL_DOS
#include <os2.h>

#include <malloc.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <share.h>
#include <direct.h>
#include <io.h>
#include <builtin.h>
#include <monetary.h>
#include <sys/types.h>
#include <time.h>
#include <localdef.h>

#include "except.h"

#include "language.h"

#include "heap.h"
#include "misc.h"
#include "timer.h"
#include "queue.h"
#include "nodelist.h"
#include "translat.h"
#include "key.h"
#include "newstr.h"
#include "direct.h"
#include "zinter.h"
#include "hdialog.h"
#include "ea.h"
#include "newstr.h"
#include "buffer.h"

void showAction( char *str )
{
  if( str==NULL || !*str )
  {
    log( "*BUG* showAction str==NULL" );
    return;
  }

  static int pos = strlen(Msg_ScrAction)+3;

  int len = wndStat.getWidth()-strlen( Msg_ScrAction )-3;
  char *str1 = NULL;

  if( strlen(str)>len )
  {
    str1 = tmpStr( str );
    str1[len] = 0;
    str = str1;
  }

  Color cl( colorSet.actionBold );
  wndStat.drawTransact( pos, 2, cl, str );
}

void putToStatusLine( int c, Color& atr, char *s )
{
  static int videoPrevColumn = 1;

  if( c!=-1 )
    videoPrevColumn = c;

  statusLineWindow = atr;
  statusLineWindow( c, 1 ) << s;
  videoPrevColumn += strlen( s );
}

void registerUsageTime( int fl )
{
  if( fl )
    systemUsageBegin = getTimeMark();
  else
  {
    systemUsageTime += getTimeMark()-systemUsageBegin;
    systemUsageBegin = 0;
  }
}

void hideModemWindow( long newTime )
{
  if( (miscInfo.state&ModemWindow)==ModemWindow &&
      wndModemLast && wndModemLast+miscInfo.wndModemOff < newTime )
  {
    -wndModem;
    -(*wndModemFrame);
    wndModemLast = 0;
  }
}

void _Optlink statusFiction( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  extern char statusAbort;

  while( 1 )
  {
    if( statusAbort )
      break;

    hideModemWindow( getTimeMark() );
    keyPressed();
    sleep( 500 );
  }

  deRegisterExceptionHandler
}

void _Optlink status( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  extern char statusAbort;

  long oldTime( 0 );
  unsigned long oldCore( 0l );
  char oldTerminal( char(-1) );
  int oldCapture( -1 );
  char oldTerminalLocalEcho( char(-1) ),
       oldTerminalNeedTranslate( char(-1) ),
       oldTerminalStopOutput( char(-1) );

  int  oldZMH( 65 );
  long oldOnline( -1 );
  long oldSpeed( -1 );
  long oldElapsed( -1 );

  int statusWidth = statusLineWindow.getWidth(),
      headWidth = headLineWindow.getWidth();

  char c[80],
       bl[]( "    " );

  char heapErrorWasDetected( 0 );
  char heapErrorWasDetectedAndDisplayed( 0 );
  int heapCount( 0 );

  char newStat[] = "---- ";
  char oldStr[ sizeof(newStat)+1 ] = { 0 };

  long startOnline( 0 );
  char firstPortSpeed( 1 );
  int callCount( 1 );
  char needShowHeadLine( 0 );
  
  while( 1 )
  {
    if( statusAbort )
      break;
      
    long newTime( getTimeMark() );

    hideModemWindow( newTime );

    heapCount += miscInfo.statusLine;

    if( heapCount>5000 )
    {

      if( heapDebug>=2 )
      {
        int i=_heapchk();

        if( i<0 )
        {
          heapErrorWasDetected = 1;

          if( !heapErrorWasDetectedAndDisplayed )
          {
            log( "~~~ HEAP Corrupted, error: %d ~~~", i );
            OSReturnCode = 2;
          }
        }
        else
          _heapmin();
      }

      heapState( "stats" );
      heapCount = 0;
    }

    if( heapErrorWasDetected && !heapErrorWasDetectedAndDisplayed )
    {
      showInd( "*", 1, 1 );
      heapErrorWasDetectedAndDisplayed = 1;
    }

    long elapsed = newTime-systemStartTime;

    if( !elapsed )
      elapsed = 1;

    void showHeadLine();
 
    if( oldElapsed!=elapsed )
    {
      char needShowHeadLineIfOldElapsedGreatElapsed( char( oldElapsed>elapsed ) );

      oldElapsed = elapsed;

      if( elapsed>0 )
      {
        char *t( 0 ),
             t1[32];

        unsigned long percent;

        t = strTime( elapsed, 0 );

        if( !elapsed )
          percent = 99;
        else
          percent = makePercent( elapsed,
            systemUsageTime+(systemUsageBegin ? newTime-systemUsageBegin : 0) );

        if( long(percent)<0 )
          strcpy( t1, "<0%" );
        else
          sprintf( t1, "%02lu%%", percent );

        if( needShowHeadLineIfOldElapsedGreatElapsed || needShowHeadLine )
        {
          showHeadLine();
          needShowHeadLine = 0;
        }

        headLineWindow( headWidth-strlen(t)-strlen(t1)-6, 1 )
                      << "  " << t << "  ";
                      
        Color cl( keyPressed() ? Color( colorSet.statusLampOn ) : scrColor );

        headLineWindow = cl;
        headLineWindow << t1;
        headLineWindow = scrColor;
        delete t;
      }
      else if( !needShowHeadLine )
      {
        int makeColor( int cl, int mn );

        needShowHeadLine = 1;
        headLineWindow( 1, 1 );
        headLineWindow = Color( makeColor( BRIGHTWHITE, BLACK ),
			        makeColor( RED|0x8, WHITE|0x8 ) );

        headLineWindow.clearLine();
        headLineWindow.centre( 1, Msg_ScrCorruptedTime );
        headLineWindow = scrColor;
      }
    }

    if( oldTime!=newTime )
    {
      char *t( strTime( time( NULL ), 1 ) );

      oldTime = newTime;

      putToStatusLine( 10, scrColor, t );
      putToStatusLine( -1, scrColor, " " );

      delete t;
    }

    if( terminalActive!=oldTerminal )
    {
      char *termType[] ={ bl, "TTY ", "ANS " };

      oldTerminal = terminalActive;
      oldCore = 0l;
      oldCapture = -1;
      oldZMH = 65;
      oldOnline = -1;

      putToStatusLine( 31, scrColor, termType[ terminalActive ] );

      if( !terminalActive )
        firstPortSpeed = 1;
    }

    if( terminalActive )
    {
      if( oldCapture!=terminalCapture ||
          oldTerminalLocalEcho!=terminalLocalEcho ||
          oldTerminalNeedTranslate!=terminalNeedTranslate ||
          oldTerminalStopOutput!=terminalStopOutput )
      {
        oldCapture               = terminalCapture;
        oldTerminalLocalEcho     = terminalLocalEcho;
        oldTerminalNeedTranslate = terminalNeedTranslate;
        oldTerminalStopOutput    = terminalStopOutput;

        char msg[]="____ ";

        msg[0] = terminalCapture ? 'C' : '-';
        msg[1] = terminalNeedTranslate ? 'T' : '-';
        msg[2] = terminalStopOutput&1 ? 'S' : (terminalStopOutput&2 ? 'P' : '-');
        msg[3] = terminalLocalEcho ? 'E' : '-';

        putToStatusLine( 37, scrColor, msg );
      }
    }
    else
    {
      int ZMH( 0 );

      if( miscInfo.zmh!=NULL && miscInfo.zmh->isWorked( time(NULL) ) )
        ZMH = 1;

      if( oldZMH!=ZMH  )
      {
        oldZMH = ZMH;
        putToStatusLine( 31, scrColor, ZMH ? "ZMH " :  bl );
      }

      int isSplitterActive();
      extern char processBsyActive;
      extern char flagsProcessorActive;
      extern char nowInScan;

      newStat[0] = processBsyActive ? 'B' : ' ';
      newStat[1] = flagsProcessorActive ? 'F' : ' ';
      newStat[2] = isSplitterActive() ? 'S' : ' ';
      newStat[3] = nowInScan ? 'R' : ' ';

      if( strcmp( oldStr, newStat ) )
      {
        strcpy( oldStr, newStat );
        putToStatusLine( 37, scrColor, newStat );
      }
    }

    if( comPort!=NULL && comPort->isOpened() )
    {
      if( firstPortSpeed || (terminalActive && oldSpeed!=comPort->getSpeed()) )
      {
        firstPortSpeed = 0;
        int delta=12;

        if( comPort->getPort() )
        {
          char prf[]( "COM" );

          if( comPort->getPort()>100 )
            if( statusWidth>80 )
              delta += 2;
            else
              prf[1] = 0;
          else if( comPort->getPort()>10 )
            if( statusWidth>80 )
              delta++;
            else
              prf[2] = 0;

          sprintf( c, " %s%d %-6ld", prf, comPort->getPort(), comPort->getSpeed() );
        }
        else
          sprintf( c, " %-10s", Msg_MdmNoPort );

        putToStatusLine( statusWidth-delta, scrColor, c );

        oldSpeed = comPort->getSpeed();
      }

      comPort->stc();

      if( comPort->isOnline() )
      {
        if( !startOnline )
  	  startOnline = getTimeMark();
      }
      else
        startOnline = 0;

      if( !startOnline )
      {
        oldOnline = -1;
        oldCore = 0l;
        goto processCoreLeft;
      }
      else if( oldOnline!=newTime-startOnline )
      {
        oldOnline = newTime-startOnline;

        char ttFlag( 0 );
        long tt( oldOnline );

        if( tt/60 > 59 )
        {
          ttFlag = 1;
          tt /= 60;
        }

        if( oldOnline<0 )
        {
          ttFlag = 1;
          strcpy( c, "??:??" );
        }
        else
          makeTime( c, tt, 1 );

        showInd( c, 2, ttFlag );
        putToStatusLine( 7, scrColor, " " );
      }
    }
    else
    {
processCoreLeft:
      unsigned long newCore( oldCore );

      if( --callCount<=0 )
      {
        newCore = zfree( miscInfo.inbound );
        callCount = 10;
      }

      if( oldCore!=newCore )
      {
        char *s( strBig( newCore ) );
        
        *c = 0;
        
        for( int i(5-strlen(s)); i>0; --i )
          strcat( c, " " );
        
        strcat( c, s );
        strcat( c, " " );
        
        delete s;
        
        putToStatusLine( 2, scrColor, c );
        oldCore = newCore;
      }
    }

    sleep( miscInfo.statusLine );
  }

  deRegisterExceptionHandler
}

int abortAction( int fl )
{
  if( keyPressed() )
    switch( keyGet() )
    {
    case 0x20:
      if( fl!=1 )
        goto Abzac;

      return 0x20;

    case key_Alt_S:
      if( fl!=2 )
        goto Abzac;

      return 's';

    case 0x01:
      return 27;

    case 27:
      if( detachedSession )
        return 0;

      {
        Color clColor( colorSet.abort );
        Window *w( new Window( (Z_ScreenWidth-33)/2+1, (Z_ScreenHeight-5)/2+1,
                     33, 5,
                     clColor,
                     Msg_ScrEscPress, BS_SINGLELINE, DS_LOWERRIGHT ) );

        w->centre( 1, Msg_ScrEscQuest );
        +*w;

        char ch( 0 );
        int oldLeft( 0 );
        char msg[32];
        Timer t( 10000 );

        while( !t.isTimeOut() && ch!='Y' && ch!='N' )
        {
          int newLeft( t.secsLeft() );

          if( oldLeft!=newLeft )
          {
            oldLeft = newLeft;
            sprintf( msg, Msg_ScrEscLeft, newLeft );
            (*w)( 1, 3 );
            w->clearLine();
            w->centre( 3, msg );
          }

          if( keyPressed() )
            ch = upperChar( keyGet() );

          checkPoint();
        }

        delete w;

        return ch=='Y' ? 27 : 0;
      }

    case key_Alt_H:
      if( zmodemActive || hSend || hReceive )
      {
        log( Msg_SesHangUp );
        comPort->hangUp();
        return 0;
      }
      break;

    case key_Alt_C:
      if( hSend || hReceive )
        return 'c';

      break;

    case key_Alt_F12:
      if( zmodemActive || hSend || hReceive )
      {
        miscInfo.state = SysState( int(miscInfo.state)^int(ZDebug) );
        break;
      }

    // fall

    default:
Abzac:
      execSchedule( melodies.alarm );
    }

  return 0;
}

void bigPause( int secs )
{
  Timer t( secs*1000l );

  while( !t.isTimeOut() && !comPort->haveInput() )
    checkPoint();
}

long makeCircle()
{
  return long( miscInfo.circleMin +
	       random( miscInfo.circleMax-miscInfo.circleMin+1 ) );
}

void  makeTime( char *buf, long rest, int fl )
{
  if( rest<0 )
    rest = 0;

  long dd( rest/60 );

  if( dd>99 && fl )
    sprintf( buf, "%5lu", dd );
  else
    sprintf( buf, "%02u:%02u", int(dd), int(rest)%60 );
}

char *strTime( long elapsed, int flag )
{
  char *r = new char[ 64 ];

  if( flag && flag!=-1 )
  {
    struct tm *tt( localTime( &elapsed ) );
    char r1[32];

    strftime( r, 31, "%a", tt );

    if( strlen(r)>3 )
      r[3] = 0;

    strftime( r1, 31, " %d/%m %H:%M:%S", tt );

    strcat( r, r1 );
    capitalize( r );
  }
  else
  {
    char w[ 32 ] = {0};
    int hund( int(elapsed%1000) );

    if( flag==-1 )
      elapsed /= 1000;

    int secs( int(elapsed%60) ),
	mins( int( (elapsed/60)%60 ) ),
	hours( int( (elapsed/60/60)%24 ) ),
	days( int( elapsed/60/60/24 ) );

    *r = 0;

    if( days )
    {
      sprintf( w, "%d-", days );
      strcat( r, w );
    }

    if( hours || days )
    {
      sprintf( w, "%02d:", hours );
      strcat( r, w );
    }

    sprintf( w, "%02d:%02d", mins, secs );
    strcat( r, w );

    if( flag==-1 )
    {
      sprintf( w, ".%03d", hund );
      strcat( r, w );
    }
  }

  return r;
}

char *strNumber( unsigned long dsize )
{
  char str[64];

  if( !dsize )
  {
    str[0] = '0';
    str[1] = 0;
  }
  else if( miscInfo.separator )
  {
    strfmon( str, sizeof(str), "%+!-#1.0n", (double)dsize );

    for( int i(0); str[i]; i++ )
      if( !isdigit(str[i]) )
        str[i] = miscInfo.separator;
  }
  else
    sprintf( str, "%lu", dsize );

  return newStr( str );
}

char *strBig( unsigned long quotient )
{
  int i = 0;
  char tempstr[6],
       e_stuff[6];
  long oldq = 0L;

  static char *rgchSize = "bKMG";

  while( quotient >= 1024L )
  {
    oldq = quotient;
    quotient = oldq / 1024L;
    i++;
  }

  int intq = int( quotient );

  if( intq > 999 )
  {
    intq = (intq * 25) / 256;	/* 100/1024 */
    sprintf( e_stuff, ".%2d%c", intq, rgchSize[++i] );
  }
  else if( (intq < 10) && (i != 0) )
  {
    intq = int( (oldq * 5L) / 512L );	/* 10/1024 */
    sprintf( tempstr, "%02d", intq );
    sprintf( e_stuff, "%c.%c%c", tempstr[0], tempstr[1], rgchSize[i] );
  }
  else
   sprintf( e_stuff, "%3d%c", intq, rgchSize[i] );

  for( i=0, intq=0; e_stuff[i]; i++ )
    if( e_stuff[i]!=' ' )
      e_stuff[intq++] = e_stuff[i];

  e_stuff[intq] = 0;

  return newStr( e_stuff );
}

/*
char *strBig( unsigned long dsize )
{
  char *r = new char[ 64 ],
       *lett = "";
  float size = dsize;

  if( dsize>=1000 )
  {
     size /= 1024;
     lett = "K";

     if( size>=1000 )
     {
       size = float(dsize)/1024/1024;
       lett = "M";

       if( size>=1000 )
       {
         size = float(dsize)/1024/1024/1024;
         lett = "G";
       }
     }
  }

  sprintf( r, "%3.3f", size );
  dsize = strlen( r );

  if( dsize>4 )
  {
    r[4] = 0;
    dsize = 4;
  }

  --dsize;

  if( r[ dsize ]=='.' )
    r[ dsize ] = 0;

  strcat( r, lett );

  return r;
}
*/

/*
char *strBig( unsigned long dsize )
{
  char *r = new char[ 64 ];

  if( dsize<1000 )
    sprintf( r, "%u", int(dsize) );
  else
  {
    unsigned long rem( dsize%1024 ),
	     dsize1( dsize / 1024 );

    if( dsize1<1000 )
      sprintf( r, "%uK", (unsigned )int( dsize1 + (rem > 511 ? 1 : 0) ) );
    else
    {
      rem = dsize%(1024l*1024);
      dsize1 = dsize / (1024l*1024);

      sprintf( r, "%uM", (unsigned )int( dsize1 + (rem > 511 ? 1 : 0) ) );
    }

  }

  return r;
}
*/

int makeFilePath( char *file )
{
  char drive[_MAX_DRIVE],
       dir[_MAX_DIR],
       name[_MAX_FNAME],
       ext[_MAX_EXT],
       path[_MAX_PATH];

  _splitpath( file, drive, dir, name, ext );

  strcpy( path, drive );
  strcat( path, dir );

  if( path[ strlen(path)-1 ]!='\\' )
    strcat( path , "\\" );

 int rc( makePath( path ) );

 return rc;
}

int makePath( char *path )
{

/*
  if( end( path, "Outbound.2c7" ) || end( path, "Outbound.2c7\\" ) )
    log( "*TRACE* Try to create %s", path );
*/

 char drive[_MAX_DRIVE],
      dir[_MAX_DIR],
      name[_MAX_FNAME],
      ext[_MAX_EXT];

 path[ strlen( path )-1 ]=0;
 _splitpath( path, drive, dir, name, ext );

 Directory no( path,
               SearchAttributes( MustBeDirectory|MayBeReadOnly|MayBeArchived ) );

 path[ strlen( path ) ]='\\';

 if( no )
  if( (no.getAttrib()&MayBeDirectory)==MayBeDirectory )
  {
   errno = 0;
   return 0;
  }
  else
  {
   errno = EACCES;
   return EACCES;
  }

 if( *dir==0 )
   if( *name )
   {
     int rc = mkdir( name );

     if( !rc )
       setExtendedAttribute( name, ".COMMENTS", Msg_EACreatedDir );

     return rc;
   }
   else
   {
    errno = 0;
    return 0;
   }

 char *isPath = (char *)_alloca( strlen( drive ) + strlen( dir ) + 1 );

 strcpy( isPath, drive );
 strcat( isPath, dir );

 isPath[ strlen( isPath )-1 ]=0;

 Directory no1( isPath,
               SearchAttributes( MustBeDirectory|MayBeReadOnly|MayBeArchived ) );

 isPath[ strlen( isPath ) ]='\\';

 if( no1 )
  if( (no1.getAttrib()&MayBeDirectory)!=MayBeDirectory )
  {
   errno = EACCES;
   return EACCES;
  }

 int rc( 0 );

 if( !no1 )
  rc=makePath( isPath );

 if( rc )
 {
  errno = rc;
  return rc;
 }

 path[ strlen( path )-1 ]=0;
 rc=mkdir( path );
 path[ strlen( path ) ]='\\';

 if( rc )
  rc=errno;

 errno = rc;
 return rc;
}

void strItem( ostream& o, char *s )
{
  if( s==NULL )
    o << "<NULL>";
  else
    while( *s )
    {
      if( (unsigned char)(*s)<0x20 )
	o << "\\x" << hexString[ ( int(*s)>>4 ) & 0xf ]
		   << hexString[ int(*s) & 0xf ];
      else
	o << *s;
	
      s++;
    }
}

void capitalize( char *originName )
{
  if( originName==NULL )
    return;

  strlwr( originName );

  for( int i(0), cap(1); originName[i]; i++ )
  {
    if( cap && isalpha( originName[i] ) )
    {
      originName[i] = (char)toupper( originName[i] );
      cap = 0;
    }

    cap = !isalpha( originName[i] );
  }
}

int todayMessage( int how, int max, int limit )
{
  int ret( 0 );

  if( random( max )<limit )
  {
    time_t tim( time( NULL ) );
    struct tm *tmm = localTime( &tim );

    if( tmm->tm_mday==29 && tmm->tm_mon==4 )
      ret = 1;
    else if( tmm->tm_mday==1 && tmm->tm_mon==3 )
      ret = 2;
    else if( tmm->tm_mday==1 && tmm->tm_mon==0 )
      ret = 3;
    else if( tmm->tm_mday==7 && tmm->tm_mon==0 )
      ret = 4;
  }

  return ret>=how ? 0 : ret;
}

int begin( register char *orig, register char *patt )
{
  while( *patt&&*orig )
  {
    if( toupper( *orig )!=toupper( *patt ) )
      return 0;
      
    orig++;
    patt++;
  }

  return *patt && !*orig ? 0 : 1;
}

int end( char *orig, char *patt )
{
  register int origLen( strlen(orig) ),
	       pattLen( strlen(patt) );

  if( pattLen>origLen )
   return 0;

  return begin( orig+origLen-pattLen, patt );
}

int includeStr( register char *pat, register char *where )
{
  while( strlen(where)>=strlen(pat) )
  {
    if( begin( where, pat ) )
      return 1;
      
    where++;
  }

  return 0;
}

char *fileName( char *s )
{
  for( register char *p=s; *p; p++ )
    if( *p==':' || *p=='\\' || *p=='/' )
      s = p+1;

  return s;
}

int checkAddress( char *addr, uShort& zone, uShort& net,
                  uShort& node, uShort& point,
                  char **domain, FIDOAddress *def )
{
  if( systemList!=NULL )
  {
    NodeConfig *found( systemList->FindAlias( addr ) );

    if( found!=NULL )
    {
      zone = found->getZone();
      net = found->getNet();
      node = found->getNode();
      point = found->getPoint();
      *domain = found->getDomain();

      return 1;
    }
  }

  uShort num;
  char *term( NULL );

  zone = net = node = point = 0;
  *domain = NULL;

  char state( 0 ),
       zof( 0 ),
       nef( 0 ),
       nof( 0 );

  while( 1 )
  {
    num = uShort( strtoul( addr, &term, 10 ) );

    switch( *term )
    {
     case ':':
       zone = num;
       state = 1;
       zof = 1;
       break;
     case '/':
       net = num;
       state = 2;
       nef = 1;
       break;
     case '.':
       node = num;
       state = 3;
       nof = 1;
       break;
     case 0:
     case '@':
       switch( state )
       {
       case 0:
	 node = num;
	 nof = 1;
	 break;
       case 1:
	 net = num;
	 nef = 1;
	 break;
       case 2:
	 node = num;
	 nof = 1;
	 break;
       case 3:
	 point = num;
	 break;
       }

       if( *term=='@' )
	 *domain = term+1;

       if( point && (!zone && !node && !net) && def!=NULL )
       {
	 zone = def->getZone();
	 node = def->getNode();
	 net = def->getNet();
       }
       else
       {
	 if( !zof && def!=NULL  )
	   zone = def->getZone();

	 if( !nof && def!=NULL  )
	   node = def->getNode();

	 if( !nef && def!=NULL  )
	   net = def->getNet();
       }

       if( zone>4095 )
         return 0;

       goto CheckDomain;

     default:
       return 0;
    }
    
    addr = term+1;
  }

CheckDomain:

  if( !*domain )
    return 1;

  FIDOAddress testAddr( zone, net, node, point, *domain );

  if( testAddr.isUnknowDomain() )
    return 0;

  return 1;
}

int checkSpecLongName( char *end, LongNameType& type )
{
  if(      !stricmp( end, "Busy" ) )
    type = busyLongName;
  else if( !stricmp( end, "Request" ) )
    type = requestLongName;
  else if( !stricmp( end, "Freeze" ) )
    type = holdLongName;
  else if( !stricmp( end, "Status" ) )
    type = tryLongName;
  else
    return 0;

  return 1;
}

int checkLongNameType( char *end, char *patt, LongNameType& type )
{
  if( !begin( end, patt ) )
    return 0;

  end += strlen( patt );

  if( !*end )
    type = boxLongName;
  else if( !stricmp( end, ".Mail" ) )
    type = mailLongName;
  else if( !stricmp( end, ".List" ) )
    type = listLongName;
  else
    return 0;

  return 1;
}

int splitLongName( char *box, uShort& zone, uShort& net,
                   uShort& node, uShort& point,
                   char **domain, Flavour& flavour,
                   LongNameType& type )
{
  *domain = box;
  zone = net = node = point = 0;
  flavour = unknownFlavour;
  type = unknownLongName;

  while( *box && *box!='.'  )
    box++;

  if( !*box )
    return 0;

  char *end;

  *(box++)=0;
  zone = (uShort)strtoul( box, &end, 10 );

  if( *end!='.' )
    return 0;

  net = (uShort)strtoul( end+1, &end, 10 );

  if( *end!='.' )
    return 0;

  node = (uShort)strtoul( end+1, &end, 10 );

  if( *end!='.' )
    return 0;

  point = (uShort)strtoul( end+1, &end, 10 );

  if( *(end++)!='.' )
    return 0;

  if( !checkSpecLongName( end, type ) )
    if( checkLongNameType( end, flavourString( immedFlavour ),   type ) )
      flavour = immedFlavour;
    else if( checkLongNameType( end, flavourString( crashFlavour ),   type ) )
      flavour = crashFlavour;
    else if( checkLongNameType( end, flavourString( directFlavour ),  type ) )
      flavour = directFlavour;
    else if( checkLongNameType( end, flavourString( normalFlavour ),  type ) )
      flavour = normalFlavour;
    else if( checkLongNameType( end, flavourString( holdFlavour ),    type ) )
      flavour = holdFlavour;
    else if( checkLongNameType( end, flavourString( onlineFlavour ),  type ) )
      flavour = onlineFlavour;
    else if( checkLongNameType( end, flavourString( requestFlavour ), type ) )
      flavour = requestFlavour;
    else
      return 0;

  return 1;
}

void log( char *fmti ... )
{
  register char *fmt( fmti );

  if( *fmt=='?' && (miscInfo.state&ZDebug)!=ZDebug )
    return;

  if( *fmt=='@' )
    if( (miscInfo.state2&RexxDebug)!=RexxDebug )
      return;
    else
      fmt++;

  va_list argptr;
  char *buf = new char[ 40960 ];

  va_start( argptr, fmti );
  vsprintf( buf, fmt, argptr );
  va_end( argptr );

  if( systemIsReady )
    (*logPtr)( buf );
  else
  {
    register char *msg = buf[0]=='#' ? buf+1 : buf;

    strcat( msg, "\r\n" );
    DosPutMessage( 1, strlen(msg), msg );
  }

  delete buf;
}

int isBundle( char *b )
{
  if( strlen( b )!=12 )
    return 0;

  b += 9;

  if( !stricmp( b, "req" ) )
    return 3;
  else
    if( !stricmp( b+1, "lo" ) )
      switch( toupper( *b ) )
      {
      case 'I':
      case 'C':
      case 'D':
      case 'F':
      case 'N':
      case 'H':
	return 2;
      default:
	return 0;
      }
    else
      if( !stricmp( b+1, "ut" ) )
	switch( toupper( *b ) )
	{
	case 'I':
	case 'C':
	case 'D':
	case 'O':
	case 'N':
	case 'H':
	  return 1;
	default:
	  return 0;
	}

  return 0;
}

FILE *_fsopen( const char *filename, const char *orimode, int shflag )
{
  int open_mode = 0;
  int mode = 0;
  const char *fmode = orimode;

  switch( *fmode )
  {
  case 'w':
    open_mode = O_CREAT | O_WRONLY | O_TRUNC;
    mode = S_IWRITE;
    break;

  case 'r':
    open_mode = O_RDONLY;
    break;

  case 'a':
    open_mode = O_WRONLY | O_CREAT | O_APPEND;
    mode = S_IWRITE;
    break;

  default:
Illegal:
    errno = 255;
    return (FILE *)NULL;
  }

  for( fmode++; *fmode; fmode++ )
    switch( *fmode )
    {
    case 't':
      open_mode = (open_mode & ~O_BINARY) | O_TEXT;
      break;

    case 'b':
      open_mode = (open_mode & ~O_TEXT) | O_BINARY;
      break;

    case '+':
      open_mode = (open_mode & ~(O_WRONLY | O_RDONLY)) | O_RDWR;
      mode = S_IREAD | S_IWRITE;
      break;

    default:
      goto Illegal;
    }

  int fd = sopen( filename, open_mode|O_NOINHERIT, shflag, mode );

  if( fd==-1 )
    return (FILE *) NULL;

  FILE *stream = fdopen( fd, orimode );

  if( stream==NULL )
    close( fd );

  return stream;
}

void real_flush( FILE *f )
{
  if( f==NULL )
    return;

  fflush( (FILE *)f );
  DosResetBuffer( (ULONG)fileno( (FILE *)f ) );
}

long getTaskNumber()
{
  return miscInfo.taskNumber;
}

void makeFileCase( char *Filename )
{
  switch( miscInfo.fileCase )
  {
  case FileCaseMixed:
    capitalize( Filename );
    break;

  case FileCaseUpper:
    strupr( Filename );
    break;

  case FileCaseLower:
    strlwr( Filename );
    break;

  default:
    break;
  }
}

void changeBlankWithZero( char *trnName, char *name )
{
  for( int i(0), j(0), space(1); name[i]; i++ )
    if( isspace( name[i] ) )
    {
      if( !space )
      {
        space = 1;

        if( j )
          trnName[ j++ ] = 0;
      }
    }
    else
    {
      space = 0;
      trnName[ j++ ] = name[i];
    }

  if( j && trnName[j-1]==' ' )
    j--;

  trnName[ j++ ] = 0;
  trnName[ j ] = 0;
}

void leaveOneBlank( char *str )
{
  for( int i(0), j(0), space(1); str[i]; i++ )
    if( isspace( str[i] ) )
    {
      if( !space )
      {
        space = 1;

        if( j )
          str[ j++ ] = ' ';
      }
    }
    else
    {
      space = 0;
      str[ j++ ] = str[i];
    }

  if( j && str[j-1]==' ' )
    j--;

  str[ j ] = 0;
}

static struct tm crazyTimeAtThisTime =
{
  0,     // seconds after the minute [0-61]
  0,     // minutes after the hour [0-59]
  0,     // hours since midnight [0-23]
  1,     // day of the month [1-31]
  0,     // months since January [0-11]
  99,     // years since 1900
  0,     // days since Sunday [0-6]
  0,     // days since January 1 [0-365]
  0      // Daylight Saving Time flag
};

struct tm *gmTime( long *t )
{
  register struct tm *a = gmtime( t );

  return a ? a : &crazyTimeAtThisTime;
}

struct tm *localTime( long *t )
{
  register struct tm *a = localtime( t );

  return a ? a : &crazyTimeAtThisTime;
}

int setFileStamp( char *name, unsigned long filetime, int toLocal )
{
  if( toLocal )
    filetime += getTimeZone();

  struct tm *tmi = localtime( (time_t *)&filetime );

  if( tmi==NULL )
  {
    log( "*VAC++ BUG* Correction of mtime for %s at %08x", fileName( name ),
         filetime );
    tmi = &crazyTimeAtThisTime;
  }

  FILESTATUS3 status;
  APIRET rc = DosQueryPathInfo( (PSZ)name, FIL_STANDARD, &status, sizeof status );

  if( rc )
  {
    log( Msg_ErrQueryPathInfo, name, rc );
    os2Message( rc );
    return int( rc );
  }

  status.fdateLastWrite.day     = tmi->tm_mday;
  status.fdateLastWrite.month   = tmi->tm_mon+1;
  status.fdateLastWrite.year	= tmi->tm_year+1900-1980;	//status since 1980, tm since 1900
  status.ftimeLastWrite.twosecs = tmi->tm_sec/2;
  status.ftimeLastWrite.minutes = tmi->tm_min;
  status.ftimeLastWrite.hours   = tmi->tm_hour;

  rc = DosSetPathInfo( (PSZ)name, FIL_STANDARD, &status, sizeof status, 0 );

  if( rc )
  {
    log( Msg_ErrSetPathInfo, name, rc );
    os2Message( rc );
    return int( rc );
  }

  return 0;

/*
  struct utimbuf tb;

  tb.actime = filetime;
  tb.modtime = filetime;

  return utime( name, &tb );
*/
}

long getTimeZone()
{
  time_t curTime( time( 0 ) );
  register struct tm *tm( localTime( &curTime ) );
  register int dayLight( 0 );

  if( tm->tm_isdst>0 )
    dayLight = 1;

  return __lc_tod->tzdiff*60-(dayLight ? __lc_tod->shift : 0);
}

int forceDelete( char *bsyFull )
{
  APIRET rc = DosForceDelete( bsyFull );

  if( rc )
  {
    log( Msg_ErrUnableDelete1, bsyFull );
    os2Message( rc );
  }

  return int( rc );
}

int putString( FILE *file, char *buf )
{
  Buffer b;

  b.putQuotted( buf );
  b.putInto( 0 );

  return fputs( b.str, file )==EOF ? 1 : 0;
}

int putPairBoth( FILE *file, char *variable, char *value, int fl )
{
  int rc;

  if( fl )
    rc = putString( file, variable );
  else
    rc = fputs( variable, file )==EOF ? 1 :0;

  if( !rc )
    rc = fputc( ' ', file )==EOF ? 1 :0;

  if( !rc )
    rc = putString( file, value );

  if( !rc )
    return fputc( '\n', file )==EOF ? 1 :0;

  return 1;
}

int putPair( FILE *file, char *variable, short value )
{
  char val[16];

  sprintf( val, "%u", value );

  return putPair( file, variable, val );
}

int putPair( FILE *file, char *variable, unsigned long value )
{
  char val[16];

  sprintf( val, "%lu", value );

  return putPair( file, variable, val );
}

void getQuottedString( FILE *file, Buffer &ret, int delimitter )
{
  ret.reset();

  int ch;

  while( 1 )
  {
    ch = fgetc( file );

    if( ch==EOF || ch==delimitter || ch=='\n' )
      break;

    ret.putInto( char(ch) );
  }

  ret.deQuote();
  ret.putInto( 0 );
}

unsigned long makePercent( const unsigned long size,
                           const unsigned long current )
{
  if( size==current )
    return 100;

  if( !size )
    return 0;

  return (unsigned long)( float( current ) * 100. / float( size ) );
}

unsigned long getMSec()
{
  ULONG mark(0);

  DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &mark, sizeof mark );
  return mark;
}

void allTrim( char *name )
{
/*
  for( int i(0), j(0); name[i]; i++ )
    if( isprint(name[i]) )
      name[j++] = name[i];

   name[j] = 0;
*/

  register int len( strlen( name )-1 );

  if( name[len]=='\n' )
    name[len] = 0;
}

void dataInitMsg( unsigned long how, char *msg )
{
  if( how )
  {
    char *r( strTime( how, -1 ) );

    log( msg, r );
    delete r;
  }
}

void dataDeInitMsg( unsigned long how, char *msg )
{
  if( how )
  {
    char *r( strTime( how, -1 ) );

    log( Msg_DeInitZZZ, msg, r );
    delete r;
  }
}

void showInd( char *n, int pos, int fl )
{
  Color cl( fl ? colorSet.statusLampOn : colorSet.statusLampOff );

  putToStatusLine( pos, cl, n );
}

char upperChar( int c )
{
  return char( c>255 ? c : toupper( c ) );
}

int isDigit( int c )
{
  return (c>='0' && c<='9') ? 1 : 0;
}

int isSpace( int c )
{
  return (c=='\r' || c=='\n' || c=='\t' || c==' ' || c=='\f') ? 1 : 0;
}

int putPair( FILE *file, char *variable, char *value )
{
  return putPairBoth( file, variable, value, 1 );
}

int putPair1( FILE *file, char *variable, char *value )
{
  return putPairBoth( file, variable, value, 0 );
}

void getKeyword( FILE *file, Buffer &ret )
{
  getQuottedString( file, ret, ' ' );
}

void getValue( FILE *file, Buffer &ret )
{
  getQuottedString( file, ret, '\n' );
}

unsigned long getTimeMark()
{
  return getMSec()/1000;
}

int getFileSize( int handle )
{
  FILESTATUS3 status;
  APIRET rc( DosQueryFileInfo( handle, FIL_STANDARD, &status, sizeof status ) );

  if( rc )
  {
    log( Msg_ErrUnableQueryFileInfo, handle, rc );
    os2Message( rc );
  }

  return int( status.cbFile );
}
