#define INCL_DOS
#include <os2.h>

#include <stdlib.h>
#include <stdio.h>
#include <share.h>
#include <builtin.h>

#include "language.h"

#include "misc.h"
#include "timer.h"
#include "buffer.h"
#include "key.h"
#include "calls.h"
#include "hydra.h"
#include "rexxhook.h"
#include "fax.h"

void controlChar( int ch, int to )
{
  if( terminalActive || 
      zmodemActive || hydraIsActive ||
      faxActive ||
      (miscInfo.state&ModemWindow)!=ModemWindow )
    return;

  char buf[2], *snd( buf ), fl(1);

  switch( ch )
  {
  case 0:
    snd = "\\00";
    break;

  case '\r':
    snd = "\\r";
    break;

  case '\n':
    snd = "\\n";
    break;

  case '\b':
    snd = "\\b";
    break;

  case 17:
    snd = "^S";
    break;

  case 19:
    snd = "^Q";
    break;

  case 0xff:
    snd = "\\ff";
    break;

  case -1:
    snd = "~";
    break;

  case -2:
    snd = "v";
    break;

  case -3:
    snd = "^";
    break;

  case -4:
    snd = "`";
    break;

  case -5:
    snd = "|";
    break;

  case -6:
    snd = ":";
    break;

  default:
    fl = 0;
    buf[0] = (char)ch;
    buf[1] = 0;
  }

  if( !wndModemLast )
  {
    +(*wndModemFrame);
    +wndModem;
  }

  wndModemLast = getTimeMark();

  wndModem = Color( to ? (fl ? colorSet.modemCtlTo : colorSet.modemTo)
                       : (fl ? colorSet.modemCtlFrom : colorSet.modemFrom) );

  wndModem << snd;

/*
  if( fl )
    wndModem.clearLine();
*/
}

void dumpTo( int ch, int fl=0 )
{
  controlChar( fl ? ch : ch&0xff, 1 );
}

void dumpFrom( int ch, int fl )
{
  controlChar( fl ? ch : ch&0xff, 0 );
}


void Modem::initSpeed()
{
  switch( speed )
  {
  case B300:
    actualSpeed = 300;
    break;

  case B600:
    actualSpeed = 600;
    break;

  case B1200:
    actualSpeed = 1200;
    break;

  case B4800:
    actualSpeed = 4800;
    break;

  case B9600:
    actualSpeed = 9600;
    break;

  case B19200:
    actualSpeed = 19200;
    break;

  case B38400:
    actualSpeed = 38400l;
    break;

  case B57600:
    actualSpeed = 57600l;
    break;

  case B76800:
    actualSpeed = 76800l;
    break;

  case B115200:
    actualSpeed = 115200l;
    break;

  case B230400:
    actualSpeed = 230400l;
    break;

  case B345600:
    actualSpeed = 345600l;
    break;

  default:
    actualSpeed = 2400;
  }
}

void Modem::clearout()
{
  ClearOutQue();

  if( isOnline() && breakMode )   //check
  {
    do_break( 1 );
    sleep( 1000 );
    do_break( 0 );
  }
}

int Modem::isOnline()
{
  if( checkOnline() )
  {
    monitorTimer.endTimer();
    return 1;
  }

  if( carrierMonitor && monitorOn )
  {
    if( !monitorTimer.isStarted() )
      monitorTimer.startTimer( carrierMonitor);

    if( !monitorTimer.isTimeOut() )
      return 1;
  }

  actualSpeed = 0;
  return 0;
}

Buffer *Modem::getOffLine( unsigned long time )
{
  Buffer *buf( new Buffer );
  Timer t( time );

  for( int ch(0), done(0), chars(0); (!t.isTimeOut()&&done!=2); )
    if( (ch=get())!=-1 )
      if( chars )
        if( ch=='\r'||ch=='\n' )
          done++;
        else
          buf->put( ch );
      else
        if( ch!='\r' && ch!='\n' )
        {
          chars = 1;
          buf->put( ch );
        }

  buf->putInto( 0 );
   
  char *newStr = processRexxHook( "ModemString", buf->str, NULL, NULL,
                                  NULL, NULL, NULL, 1 );

  if( newStr!=NULL )
  {
    if( !*newStr )
      goto Ignore;
    else
    {
      buf->reset();
      buf->putString( newStr );
      buf->putInto( 0 );
    }
    
    delete newStr;
  }

  if( *buf->str )
    if( (miscInfo.state3&HardRing)==HardRing &&
        !stricmp( buf->str, modemString.ring ) )
    {
Ignore:
      buf->reset();
      buf->putInto( 0 );
    }
    else
      for( ModemStringIterator it( *ignoreString ); it; it++ )
        if( begin( buf->str, it().getString() ) )
          goto Ignore;

  return buf;
}

void Modem::stripEcho( int time )
{
  sleep( 500 );

  Timer t( time );

  for( int done(0); !t.isTimeOut() && done<1 && haveInput(); )
    if( get()=='\r' )
      done++;
}

int Modem::waitOk( char *who )
{
  register int i = 1;

  for( Timer t(5000); !t.isTimeOut() && i; )
  {
    Buffer *fromModem = getOffLine( t.rest() );

    i = stricmp( fromModem->str, modemString.ok );

    if( i && who )
      i = stricmp( fromModem->str, who );

    if( i )
      i = stricmp( fromModem->str, modemString.ring );

    delete fromModem;
  }

  return i;
}

int Modem::putOffLine( register char *str, char *ok, char echo )
{
  char ch,
       needWait(0);

  for( int rc(0); (ch=*str, ch) && !rc; str++ )
    switch(ch)
    {
    case 'v':
      dumpTo( -2, 1 );
      dtr( 0 );
      break;

    case '^':
      dumpTo( -3, 1 );
      dtr( 1 );
      break;

    case '~':
      dumpTo( -1, 1 );
      sleep( 500 );
      break;

    case '`':
      dumpTo( -4, 1 );
      sleep( 100 );
      break;

    case ':':
      dumpTo( -6, 1 );
      ch = '\r';
      needWait = 2;
      goto SendIt;

    case '|':
      dumpTo( -5, 1 );
      ch = '\r';
      needWait = 1;
      goto SendIt;

    default:
SendIt:

      if( (ch=(char)putWait( ch, 0 ), ch) )
      {
Fail:
        rc = 1;
	break;
      }

      if( needWait )
      {
        if( needWait==1 )
          if( waitOk( ok ) )
            goto Fail;

        if( needWait==2 )
          stripEcho( 5000 );

        needWait = 0;
      }

      if( pause )
	sleep( pause );
    }

  if( echo && haveInput() )
    delete getOffLine( 5000 );

  return rc;
}

int Modem::putOnLine( register char *str, int len, long time )
{
  Timer t( time );

  for( register int i(0); !t.isTimeOut()&&isOnline()&&i<len; )
    if( put( str[i] ) )
      i++;
/*
    else
      checkPoint();
*/

  return (i==len ? 0 : (t.isTimeOut() ? 1 : 3) );
}

// High Level

int Modem::initOnce()
{
  clearin();
  clearout();

  return putOffLine( initStr );
}

int Modem::offHook( char *what )
{
  clearin();
  clearout();

  virtualDCDOn = 0;
  monitorOn = 0;

  if( putOffLine( offHookStr, what ) )
    return 1;

  initSpeed();
  return 0;
}

void Modem::noCallersId()
{
  delete callersId;
  callersId = NULL;
}

int Modem::hangUp()
{
  for( int i=0, j=0; i<3&&isOnline(); i++ )
    j = offHook();

  virtualDCDOn = 0;
  monitorOn = 0;
  actualSpeed = 0;
  noCallersId();
  initSpeed();
  return j;
}

int processBadModem()
{
  char *hook = processRexxHook( "BadModem", NULL, NULL, NULL,
                                NULL, NULL, NULL, 1 );

  if( hook!=NULL )
  {
    if( !stricmp( hook, "No" ) )
    {
      delete hook;
      return 0; 
    }
    else if( !stricmp( hook, "Yes" ) )
    {
      delete hook;
      return 1; 
    }
    else
      log( Msg_ErrIllegalBadHook, hook );

    delete hook;
  }

  return 0;
}

int Modem::init()
{
  char *hook = processRexxHook( "InitModem", initStr, NULL, NULL,
                                NULL, NULL, NULL, 1 );

  if( hook!=NULL )
  {
    if( !stricmp( hook, "No" ) )
    {
      delete hook;
      return 1; 
    }
    else if( !stricmp( hook, "Yes" ) )
    {
      delete hook;
      return 0; 
    }
    else if( stricmp( hook, "YourSelf" ) )
      log( Msg_ErrIllegalInitHook, hook );

    delete hook;
  }

  clearin();
  clearout();

  for( int i=0, j=1; i<3 && j; i++ )
    if( j = initOnce(), j )
      if( keyPressed() && keyGet()==key_Alt_T )
      {
        void terminalEmulator();

        if( !terminalActive )
        {
          setup();
          terminalEmulator();
          --i;
        }
      }
      else
      {
        setup();
        offHook();
      }

  initSpeed();

  if( j && processBadModem() )
    j = 0;

  return j;
}

static int oldSecs = 0,
           oldMove = 0,
           oldRRing = 0;

static Window *carrier = NULL;
static int oldX = 25,
	   oldY = 10;
static char *carrierFormat = NULL;
static int rring = 0;

void waitIs( int secs )
{
  if( carrier!=NULL )
  {
    if( !oldSecs )
      oldMove = secs;

    if( oldSecs!=secs || oldRRing!=rring )
    {
      char *msg = (char *)_alloca( strlen( carrierFormat ) + 16 );

      sprintf( msg, carrierFormat, secs, rring );
      oldSecs = secs;
      oldRRing = rring;
      (*carrier)(1,2).clearLine();
      carrier->centre( 2, msg  );

      if( secs+1 < oldMove )
      {
	oldMove = secs;
	oldX = random(Z_ScreenWidth-carrier->getWidth()-4)+1;
	oldY = random(Z_ScreenHeight-carrier->getHeight()-3)+1;
	carrier->move( oldX, oldY );
      }
    }
  }
  else
    if( oldSecs!=secs || oldRRing!=rring )
    {
      char *msg = (char *)_alloca( strlen( carrierFormat ) + 32 );

      sprintf( msg, carrierFormat, secs, rring );
      oldSecs = secs;
      oldRRing = rring;
      showAction( msg );
    }

//  checkPoint();
}

static char binLogCode;
Buffer *fromModem = new Buffer;

inline int display( Timer& t, int& oldCount, int i, int& oldRest, char *buf )
{
  register int secs = abortAction( 0 );

  if( secs )
    return secs;

  secs  = int( t.secsLeft() );

  if( oldCount!=i || oldRest!=secs )
  {
    oldCount=i;
    oldRest=secs;
    sprintf( buf, Msg_ScrRing, i, secs );
    showAction( buf );
  }

  return 0;
}

void Modem::getRing()
{
  delete fromModem;
  fromModem = getOffLine( 1000 );

  if( *fromModem->str )
    if( stricmp( fromModem->str, modemString.ring ) )
    {
//log( "*DEBUG* Modem::getRing() %d %d %d", fromModem->str[0], fromModem->str[1], fromModem->str[2]);
      if( fromModem->length()>1 && (miscInfo.state&ModemReport) )
        log( Msg_ModemReport, fromModem->str );
    }
    else
      if( (miscInfo.state3&HardRing)!=HardRing )
        ringSignal();
}

int Modem::answer( int skipRing )
{
  int oldRest( -1 );
  int oldCount( 0 );
  char *buf = (char *)_alloca( strlen( Msg_ScrRing )+32 );

  actualSpeed = 0;

  if( !stricmp( connectSuff, Msg_ModemConnectDCD ) )
    skipRing = 1;

  for( int i=1; !skipRing && i<miscInfo.ringCount; )
  {
    Timer t( miscInfo.ringWait*1000l );

    if( (miscInfo.state3&HardRing)!=HardRing )
      while( !t.isTimeOut() && !haveInput() )
        if( display( t, oldCount, i, oldRest, buf ) )
        {
no_need:
          return 1;
        }
        else
          checkPoint();
    else
    {
      while( !t.isTimeOut() && isRinging() )
        if( display( t, oldCount, i, oldRest, buf ) )
          goto no_need;
        else
          checkPoint();

      if( t.isTimeOut() )
        break;

      while( !t.isTimeOut() && !isRinging() )
        if( display( t, oldCount, i, oldRest, buf ) )
          goto no_need;
        else
          checkPoint();

      if( t.isTimeOut() )
        break;

      ringSignal();
      i++;
    }

    if( t.isTimeOut() )
      break;

    if( haveInput() )
    {
      getRing();

      if( (miscInfo.state3&HardRing)!=HardRing &&
          !stricmp( fromModem->str, modemString.ring ) )
        i++;
    }
  }

  if( !skipRing && i!=miscInfo.ringCount )
  {
    log( Msg_ModemTooLowRing, i );
    return 1;
  }

  if( haveInput() )
    getRing();

  log( Msg_ModemAnswer, answerStr );

  putOffLine( answerStr, NULL, 0 );
  binLogCode = 0;

  int rc = waitConnect( 0, answerWait );

  if( rc==2 && !terminalActive && getSpeed()<miscInfo.minIn )
  {
    log( Msg_ModemTooSlow );
    hangUp();
    rc = 1;
  }

  return rc;
}

void steb()
{
  if( random(100)>=10 )
    return;

  char *p( strchr( fromModem->str, '/' ) );

  if( p!=NULL )
  {
    *p = 0;
    int len( strlen( fromModem->str ) );

    while( fromModem->length()!=len && fromModem->length() )
      fromModem->stripByte();

    fromModem->putInto( 0 );
  }

  fromModem->stripByte();

  if( random(10)<4 )
    fromModem->putString( Msg_MdmJoke1 );

  if( random(10)<4 )
    fromModem->putInto( char(random(10)+'0') );

  if( random(10)<4 )
  {
    fromModem->putString( Msg_MdmJoke2 );
    fromModem->putInto( char( random( '9'-'1'+1 ) + '1' ) );
    fromModem->putInto( char( random( '9'-'1'+1 ) + '1' ) );

    if( random(10)<4 )
      fromModem->putInto( 'B' );
  }

  if( random(10)<4 )
    fromModem->putString( Msg_MdmJoke3 );

  if( random(10)<4 )
    fromModem->putString( Msg_MdmJoke4 );

  if( random(10)<4 )
    fromModem->putString( Msg_MdmJoke5 );

  if( random(10)<4 )
  {
    fromModem->putInto( '/' );
    fromModem->putInto( char( random( 'Z'-'A'+1 ) + 'A' ) );
    fromModem->putInto( char( random( 'Z'-'A'+1 ) + 'A' ) );
    fromModem->putInto( char( random( 'Z'-'A'+1 ) + 'A' ) );
  }

  fromModem->putInto( 0 );
}

void steb1()
{
  if( random(100)>=10 )
    return;

  if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke6 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke7 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke8 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke9 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke10 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke11 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke12 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke13 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke14 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke15 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke16 );
  else if( todayMessage( 5, 30000, 250 )==2 )
    log( Msg_ModemReport, Msg_MdmJoke17 );
  else if( todayMessage( 5, 30000, 250 )==2 )
  {
    log( Msg_ModemReport, Msg_MdmJoke18 );
    log( Msg_ModemReport, Msg_MdmJoke19 );
    log( Msg_ModemReport, Msg_MdmJoke20 );
  }
}

int Modem::waitConnect( int dialingMode, int wait, char *connectStringUp )
{
  long startWait( getTimeMark() );
  long beginWait( time( 0 ) );

  int done( 0 );

  initSpeed();

  if( !wait )
  {
    wait = callWait;
    log( "~BUG~ waitConnect() wait==0" );
  }

  oldSecs = 0;
  oldMove = 0;
  oldRRing = 0;
  rring = 0;

  static char *formatStrings1[] =
  {
    Msg_ModemWaitConn,
    Msg_ModemCallJohn,
    Msg_ModemJoke,
    Msg_ModemWaitNewYear
  };

  static char *formatStrings0[] =
  {
    Msg_ModemWaitConn,
    Msg_ModemAnswerJohn,
    Msg_ModemReboot,
    Msg_ModemWaitNewYear
  };

  if( dialingMode )
    carrierFormat = formatStrings1[ todayMessage( 4, 30000, 2000 ) ];
  else
    carrierFormat = formatStrings0[ todayMessage( 4, 30000, 2000 ) ];

  int width( strlen(carrierFormat)+10 );

  while( oldX+width>Z_ScreenWidth-2 && oldX>=0 )
    oldX--;

  if( carrierFormat==Msg_ModemWaitConn && !terminalActive )
  {
    char *msg = dialingMode ? Msg_ModemCallScr : Msg_ModemAnswerScr;

    carrier = NULL;
    carrierFormat = new char [ strlen( msg )+ strlen( Msg_ModemWaitFmt )+1 ];
    strcpy( carrierFormat, msg );
    strcat( carrierFormat, Msg_ModemWaitFmt );
  }
  else
  {
    carrierFormat = newStr( carrierFormat );
    carrier = new Window( oldX, oldY, width, 5, popUpColor,
			  dialingMode ? Msg_ModemCallTtl : Msg_ModemAnswerTtl,
			  BS_SINGLELINE, DS_LOWERRIGHT );

    +(*carrier);
  }

  arq = 0;
  halfDuplex = 0;
  lastBaud = 0;
  noCallersId();

  fromModem->reset();
  fromModem->putInto( 0 );
  fromModem->reset();

  int needSteb1 = todayMessage( 5, 2, 2 )==2;
  Timer t( wait*1000l );
  char changedFormat = 0;
  char ringing = 0;
  char doneConnect(0);
  char *lastModemString( NULL );

  while( !done )
  {
    if( !t.isTimeOut() )
      waitIs( t.secsLeft() );
      
    if( doneConnect==3 )
    {
      done = 2;
      break;
    }

    if( isOnline() && (doneConnect&2)!=2 )
    {
      tranx = time( NULL );
      tranxMark = getTimeMark();
      lastBaud = actualSpeed;
      doneConnect |= 2;
      continue;
    }

    if( needSteb1 )
    {
      steb1();
      needSteb1 = 0;
    }

    switch( abortAction( dialingMode==2 ? 1 :0 ) )
    {
    case 27:
      done = 3;
      break;

    case 0x20:
      done = 7;
      
      if( !terminalActive )
        showAction( Msg_ScrWaitConnForce );
      break;
    }

    if( dialingMode && (miscInfo.state3&HardRing)==HardRing &&
        isRinging() )
      ringing = 1;
        
    if( done )
      break;

    if( t.isTimeOut() )
    {
      done=6;
      break;
    }

    fromModem->reset();

    if( connectStringUp!=NULL )
    {
      fromModem->putString( connectStringUp );
      fromModem->putInto( 0 );
      connectStringUp = NULL;
    }
    else if( haveInput() )
    {
      delete fromModem;
      fromModem = getOffLine( 1000 );
    }

    if( !*fromModem->str && !ringing )
      continue;

    if( miscInfo.state&ModemReport )
      log( Msg_ModemReport, fromModem->str );
      
    if( connectList!=NULL && connectCount && !terminalActive )
      for( short i=0; i<connectCount; i++ )
	if( begin( fromModem->str, connectList[i].getPattern() ) )
	{
	  done = 8;
	  currentConnectItem = i;
	  break;
	}

    if( begin( fromModem->str, modemString.connect ) && !done )
      if( todayMessage( 5, 2, 2 )==2 )
	steb();

    delete lastModemString;
    lastModemString = newStr( fromModem->str );

    if( done )
      break;

    if( !stricmp( fromModem->str, modemString.nocarrier )     ||
	!stricmp( fromModem->str, modemString.busy )          ||
	!stricmp( fromModem->str, modemString.voice )         ||
	!stricmp( fromModem->str, modemString.noanswer )      ||
	!stricmp( fromModem->str, modemString.noconnect ) )
      done = 1;
    else if( !stricmp( fromModem->str, modemString.fax ) )
      done = 5;
    else if( !stricmp( fromModem->str, modemString.error ) )
      done = 98;
    else if( dialingMode && (miscInfo.state3&HardRing)!=HardRing &&
             !stricmp( fromModem->str, modemString.ring ) )
      done = 1;
    else if( ringing )
    {
      fromModem->reset();
      fromModem->putString( modemString.ring );
      fromModem->putInto( 0 );
      done = 1;
    }
    else if( !stricmp( fromModem->str, modemString.rring ) )
    {
      rring++;

      if( miscInfo.maxRRing && dialingMode && (rring >= miscInfo.maxRRing) )
        done = 99;
    }
    else if( !stricmp( fromModem->str, modemString.nodialtone ) )
      done = 4;
    else if( begin( fromModem->str, modemString.connect ) )
    {
      setConnectString( fromModem->str );
      doneConnect |= 1;
      continue;
    }
    else if( begin( fromModem->str, modemString.callersId ) )
    {
      delete callersId;
      callersId = newStr( fromModem->str+strlen( modemString.callersId ) );
      log( Msg_ModemCallerId, callersId );
    }

    if( rring && !changedFormat )
    {
      changedFormat = 1;

      if( carrier==NULL )
      {
        char *msg = dialingMode ? Msg_ModemCallScr : Msg_ModemAnswerScr;

        delete carrierFormat;

        carrierFormat = new char [ strlen( msg ) +
                                   strlen( Msg_ModemWaitFmtRRing ) + 1 ];

        strcpy( carrierFormat, msg );
        strcat( carrierFormat, Msg_ModemWaitFmtRRing );
      }
      else
      {
        char *msg = carrierFormat;

        carrierFormat = new char [ strlen( msg ) +
                                   strlen( Msg_ModemWaitFmtRRing1 ) + 1 ];

        strcpy( carrierFormat, msg );
        strcat( carrierFormat, Msg_ModemWaitFmtRRing1 );
        delete msg;
      }
    }
  }
  
  fromModem->reset();
  
  if( lastModemString==NULL )
    fromModem->putString( Msg_ModemNothing );
  else
  {
    fromModem->putString( lastModemString );
    delete lastModemString;
  }
  
  fromModem->putInto( 0 );

  if( !terminalActive )
  {
    long end( getTimeMark()-startWait );

    writeToBinaryLog( binLogCode, &end, sizeof( end ), beginWait );
  }

  if( carrier!=NULL )
  {
    delete carrier;
    carrier = NULL;
  }

  delete carrierFormat;
  carrierFormat = NULL;

  int i;

  switch( done )
  {
  case 3:
  case 6:
  case 7:
  case 99:
    if( !terminalActive )
      showAction( Msg_ScrWaitConnAbort );

    for( i=3; i && putOffLine( abortDial, modemString.nocarrier ); --i )
      continue;
    break;

  case 1:
    putOffLine( afterDial );
    break;
  }

  switch( done )
  {
  case 99:
    done = 1;
    log( Msg_ModemRRingAbort, fromModem->str );
    break;

  case 98:
    done = 3;
    log( Msg_ModemErrorAbort, fromModem->str );
    break;

  case 1:
    log( Msg_ModemNotConnected, fromModem->str );
    break;

  case 2:
    log( Msg_ModemConnected,
          arq ? Msg_ModemConnectedArq : Msg_ModemConnectedNoArq, fromModem->str );
    monitorOn = 1;
    showAction( Msg_ScrConnected );
    break;

  case 3:
    log( Msg_ModemBreak, dialingMode ? Msg_ModemCallScr : Msg_ModemAnswerScr );
    fromModem->reset();
    fromModem->putString( Msg_ModemAborted );
    fromModem->putInto( 0 );
    break;

  case 4:
    log( Msg_ModemIncomCall, fromModem->str );
    break;

  case 5:
    log( Msg_ModemFAX, fromModem->str );
    break;

  case 6:
    log( Msg_ModemTimeout, fromModem->str );
    fromModem->reset();
    fromModem->putString( Msg_ModemNothing );
    fromModem->putInto( 0 );
    break;

  case 7:
    log( Msg_ModemForced, dialingMode ? Msg_ModemCallScr : Msg_ModemAnswerScr );
    break;

  case 8:
    log( Msg_ModemExternal, fromModem->str );
    break;

  default:
    log( Msg_ModemCrazy, fromModem->str );
  }

  return done;
}

void Modem::setConnectInfo( int baud )
{
  actualSpeed = baud;

  if( !actualSpeed )
    actualSpeed = 300;

  char buf[80];

  sprintf( buf, "Virtual connect at %d BPS", actualSpeed );
  connectSuff = newStr( buf );

  arq = 1;
  halfDuplex = 0;
  tranx = time( NULL );
  tranxMark = getTimeMark();
  lastBaud = actualSpeed;

  if( virtualDCD )
    virtualDCDOn = 1;
}

void Modem::setConnectString( char *str )
{
  delete connectSuff;
  connectSuff = NULL;

  int delta = 0;

  if( begin( str, modemString.connect ) )
    delta = strlen( modemString.connect );
  else
  {
    char *bl = strchr( str, ' ' );

    if( bl!=NULL )
      delta = bl-str+1;
  }

  if( strnicmp( str+delta, "Fast", 4 ) )
    actualSpeed = strtoul( str+delta, &connectSuff, 10 );
  else
    actualSpeed = 19200;

  if( !actualSpeed )
    actualSpeed = 300;

  connectSuff = newStr( str );

  tranx = time( NULL );
  tranxMark = getTimeMark();
  lastBaud = actualSpeed;

  for( CallsIterator list( *arqList ); list && !arq; list++ )
    if( includeStr( list().getFlag(), connectSuff ) )
      arq = 1;

  for( CallsIterator list1( *halfDuplexList ); list1 && !halfDuplex; list1++ )
    if( includeStr( list1().getFlag(), connectSuff ) )
      halfDuplex = 1;

  if( virtualDCD )
    virtualDCDOn = 1;
}

int Modem::dial( char *number, char *pref, char *suff, int wait, int blAvail )
{
  actualSpeed = 0;

  if( !wait )
    wait = callWait;

  if( pref==NULL )
    pref = dialPref;

  if( suff==NULL )
    suff = dialSuff;

  if( terminalActive || (miscInfo.state2&FullLog)==FullLog )
  {
    char *dialStr = (char *)_alloca( strlen(pref)+strlen(number)+strlen(suff)+1 );

    strcpy( dialStr, pref );
    strcat( dialStr, number );
    strcat( dialStr, suff );
    log( Msg_ModemDialing, dialStr );
  }

  putOffLine( pref, NULL, 0 );
  putWaitStr( number, 0 );
  putOffLine( suff, NULL, 0 );

  binLogCode = 1;

  int rc = waitConnect( blAvail ? 2 : 1, wait );

  if( rc==2 && !terminalActive && getSpeed()<miscInfo.minOut )
  {
    log( Msg_ModemTooSlow );
    hangUp();
    rc = 1;
  }

  return rc;
}

int Modem::putWait( int c, int dcd )
{
  Timer t( failSafe );

  while( !put( c ) && !t.isTimeOut() && (!dcd || (dcd && isOnline())) )
    continue;

  return ( t.isTimeOut() || (dcd && !isOnline()) ) ? 1 : 0;
}

int Modem::putWaitStr( char *str, int dcd )
{
  if( str != NULL )
  {
    Timer t( failSafe );

    while( *str )
    {
      while( !put( *str ) )
        if( ( t.isTimeOut() || (dcd && !isOnline()) ) ? 1 : 0 )
          return 1;

      str++;
    }
  }

  return 0;
}

int Modem::put( int c )
{
  register int r( putFos( c ) );

  if( r )
    dumpTo( c&0xff );

  return r;
}

int Modem::bufferByte( int c )
{
  int ok( 0 );

  if( curOutPointer == OutBufferSize )
    ok = unBufferBytes();
    
  if( !ok )
    outBuffer[ curOutPointer++ ] = (char)(unsigned)c;

  return !ok;
}

int Modem::bufferBytes( char *buf, int len )
{
  while( len )
  {
    if( !bufferByte( *buf ) )
      return 0;
      
    --len;
    ++buf;
  }
  
  return 1;
}

int Modem::unBufferBytes()
{
  Timer t( failSafe );
  register int i( 0 );

  if( blockMode )

    while( i<curOutPointer && !t.isTimeOut() && isOnline() )
    {
      int j( putFos( (void *)(outBuffer+i), curOutPointer-i ) );

      i += j;

/*
      if(  i < curOutPointer )
	checkPoint();
*/
    }

  else

    while( i<curOutPointer && !t.isTimeOut() && isOnline() )
      if( put( (unsigned)outBuffer[i] ) )
	i++;
/*
      else
	checkPoint();
*/

  if( i==curOutPointer )
  {
    curOutPointer = 0;
    return 0;
  }
  else
  {
    extern volatile long modemOs2Rc;

    if( comPort->isOnline() )
    {
      log( "Send stalled: er=%d bt=%d/%d oq=%lu/%lu, iq=%lu/%lu",
           modemOs2Rc, i, curOutPointer,
           hasOutputQueue(), sizeOutputQueue(),
           hasInputQueue(), sizeInputQueue() );

      if( modemOs2Rc )
      {
        os2Message( modemOs2Rc );
        modemOs2Rc = 0;
      }
      
      logState();
    }

    curOutPointer = 0;
    clearout();
    return 1;
  }
}

void Modem::byDCD()
{
  actualSpeed = 0;
  initSpeed();
  delete connectSuff;
  connectSuff = newStr( Msg_ModemConnectDCD );
  arq = 0;
  tranx = time( NULL );
  tranxMark = getTimeMark();
  monitorOn = 1;
  lastBaud = actualSpeed;

  if( virtualDCD )
    virtualDCDOn = 1;
}

int Modem::promptString( char *pr, char *st, int len, long secs, int flag )
{
  int rc;

  if( pr!=NULL && *pr )
  {
    rc=putOnLine( pr, strlen(pr), secs );

    if( rc )
      return 1;
  }

  Timer t( secs ? secs : 24*60*60*1000 );

  memset( st, 0, len );

  for( int i=0; i<len-1; )
  {
    if( !isOnline() )   //check
      return 1;

    if( t.isTimeOut() )
      return 1;

    int ch = get();

    if( ch==-1 )
      continue;

    switch( ch )
    {
    case '\r':
    case '\n':
      return 0;

    case '\b':
      if( i )
      {
        st[ --i ] = 0;

        rc=putOnLine( "\b \b", 3, t.rest() );

        if( rc )
          return 1;
      }
      else
        put( 7 );

      break;

    default:
      if( ch>=0x20 && ch<0xff )
      {
        st[ i++ ] = char(ch);
        put( flag ? '*' : char(ch) );
      }
      break;
    }
  }

  put( 7 );
  return 0;
}

void Modem::ringSignal()
{
  execSchedule( melodies.ring );
}

int Modem::putMessage( char *file )
{
  if( *file!='@' )
    return putOnLine( file, strlen(file), 5000 );

  FILE *in = _fsopen( file+1, "rb" );

  if( in==NULL )
  {
    log( Msg_ErrUnableOpen, file+1, strError() );
    return 0;
  }

  int ch(0);

  for( Timer t(10000); !t.isTimeOut() && isOnline(); )
  {
    ch = ch=='\n' ? '\r' : fgetc( in );

    if( ch==EOF )
      break;

    while( !t.isTimeOut() && !put( ch ) && isOnline() )
//      checkPoint()
      continue
      ;
  }

  fclose( in );

  if( t.isTimeOut() )
    return 1;

  if( !isOnline() )
    return 3;

  return 0;
}

void Modem::statistics()
{
  if( modemString.statCmd==NULL || modemString.statStop==NULL )
    return;

  sleep( 500 );

  while( haveInput() )
  {
    clearin();
    checkPoint();
  }

  if( putOffLine( modemString.statCmd, NULL, 0 ) )
    return;

  for( Timer t( 15000 ); !t.isTimeOut(); )
  {
    Buffer *b = getOffLine( 5000 );

    if( strlen( b->str ) )
      if( !stricmp( b->str, modemString.statStop ) )
      {
        delete b;
        break;
      }
      else
        log( Msg_InfoModemStat, b->str );

    delete b;
  }

  if( t.isTimeOut() )
    log( Msg_InfoModemStatFailed );
}

void Modem::flowControl( int f )
{
  if( hardWare )
    ctsrts( f );
  else
    xonxoff( f );
}

void Modem::com_kick()
{
  flowControl(0);
  flowControl(1);
}

int Modem::CheckOnline()
{
  return checkOnline();
}

int Modem::putFos( int _c )
{
  char c = char(_c);

  return putFos( (void *)&c, 1 );
}

long Modem::getSpeed()
{ 
  if( !actualSpeed )
    initSpeed();

  return actualSpeed;
}

long Modem::getLastBaud()
{
  return lastBaud ? lastBaud : getSpeed();
}

int ModemString::equal( const BaseObject &t ) const
{
  return !stricmp( str, (*(const ModemString *)&t).str );
}

void ModemString::print( ostream& s ) const
{
  s << str;
}

char *ModemString::nameOfClass() const
{
  return "ModemString";
}

ModemString::ModemString( char *f ) :
  str( newStr( f ) )
{}

ModemString::~ModemString()
{
  delete str;
}

ModemStringList::ModemStringList() :
  Collection( 5, 5, SemaphoreModeNoNeed )
{
}

ModemString* ModemStringList::Find( ModemString& item, LockMode lock ) const
{
  return (ModemString*)Collection::Find( item, lock );
}
    
ModemString& ModemStringList::operator [] ( const int pos ) const
{
  return *(ModemString *)&(Collection::operator[]( pos ));
}

ModemStringIterator::ModemStringIterator( ModemStringList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}

ModemString& ModemStringIterator::operator () () const
{
  return *(ModemString *)&(CollectionIterator::operator()());
}
