#include <string.h>

#include "misc.h"
#include "pair.h"

Pair::Pair()
{
  who.pattern = NULL;
  action.out = NULL;
  doType = DoType( 999 );
  reset();
}

Pair::Pair( char *ip, char *io )
{
  who.pattern=ip;
  action.out=io;
  doType=doOut;
  reset();
}

Pair::Pair( char *ip, EMSI_PACKETS em )
{
  who.pattern=ip;
  action.emsi=em;
  doType=doEMSI;
  reset();
}

Pair::Pair( int key, char *io )
{
  who.key=key;
  action.out=io;
  doType=doKey;
}

Pair::Pair( char *ip, char *io, Pair::DoType is )
{
  who.pattern=ip;
  action.out=io;
  doType=is;
  reset();
  if( is!=doOut && is!=doModem )
  {
    doType = doOut;
//    log( " Pair::Pair( char*, char*, Pair::doType )  Pair::doType" );
  }
}

Pair::Pair( char *ip, void (*id)() )
{
  who.pattern=ip;
  action.doFunc=id;
  doType=doDo;
  reset();
}

Pair::Pair( int key, void (*id)() )
{
  who.key=key;
  action.doFunc=id;
  doType=doKeyDo;
}

Pair::Pair( char *ip, Schedule *ish )
{
  who.pattern = ip;
  doType = doSched;
  action.sh = ish;
  reset();
}

int Pair::equal( const BaseObject &t ) const
{
  if( doType!=(*(const Pair *)&t).doType )
    return 0;

  if( doType==doKey || doType==doKeyDo )
    return who.key == (*(const Pair *)&t).who.key;

  return !strcmp( who.pattern, (*(const Pair *)&t).who.pattern );
}

void Pair::print( ostream& s ) const
{
  if( doType==Pair::doKey || doType==Pair::doKeyDo )
    s << "Key:" << who.key << ",NoCurrent";
  else
  {
    s << "Pattern:";
    strItem( s, who.pattern );
    s << ",Current:";
    strItem( s, cur );
  }

  s << ",Do:";

  switch( doType )
  {
  case Pair::doOut:
    s << "doOut:";
    strItem( s, action.out );
    break;

  case Pair::doDo:
    s << "doDo:" << action.doFunc;
    break;

  case Pair::doEMSI:
    s << "doEMSI:" << action.emsi;
    break;

  case Pair::doKeyDo:
    s << "doKeyDo:" << action.doFunc;
    break;

  case Pair::doModem:
    s << "doModem:";
    strItem( s, action.out );
    break;

  case Pair::doKey:
    s << "doKey:";
    strItem( s, action.out );
    break;

  case Pair::doSched:
    s << "doSched:";

    if( action.sh==NULL )
      s << "<Null>";
    else
      s << *action.sh;

    break;

  default:
    s << "<Unknown>" << action.doFunc;
  }
}

char *Pair::nameOfClass() const
{
  return "Pair";
}

void Pair::operator = ( Pair &ini )
{
  who = ini.who;
  action = ini.action;
  cur = ini.cur;
  doType = ini.doType;
}
  
int Pair::operator == ( int how )
{
  return (doType==doKey || doType==doKeyDo) && who.key==how;
}

int Pair::operator == ( char *how )
{
  return doType==doSched && !strcmp( who.pattern, how );
}

int Pair::operator [] ( int i )
{
  return int( who.pattern[i] );
}

Schedule *Pair::getSchedule() const
{
  return doType==doSched ? action.sh : NULL;
}

void Pair::reset()
{
  cur=who.pattern;
}

int Pair::matchedLen() const
{
  return int(cur-who.pattern);
}

char *Pair::process()
{
  int rc;

  if( action.out==NULL )
    log( " Pair::process() action==NULL" );
  else
    switch( doType )
    {
    case doOut:
      terminalWindow << action.out;
      break;

    case doDo:
    case doKeyDo:
      (*action.doFunc)();
      break;

    case doModem:
      terminalWindow << who.pattern;          // no break!

    case doKey:
      return action.out;

    case doSched:
      if( rc=doEventKey( *action.sh, -1 ), rc )
	return (char *)rc;
      break;

    case doEMSI:
      return (char *)&action.emsi;

    default:
      log( "~BUG~ Pair::process() ⭮ Pair::doType=%d", int(doType) );
    }

  return NULL;
}

Pair::PairRez Pair::compare( char c )
{
  if( *cur==c )
    if( !*(++cur) )
    {
      reset();
      return Completed;
    }
    else
      return InProgress;
  else
    if( !*(++cur) )
    {
      reset();
      return Reset;
    }
    else
    {
      reset();
      return NoMatch;
    }
}
