#ifndef __SESSION_PROCESS

#define __SESSION_PROCESS

#include <stdio.h>

#include "class.h"
#include "regexp.h"

#include "rcvdlist.h"
#include "fidoaddr.h"
#include "sched.h"

enum ProcessItemType
{
  unknownProcessItemType=0,
  fileProcessItemType=1,
  maskProcessItemType=2,
  netMailProcessItemType=3,
  arcMailProcessItemType=4,
  filesProcessItemType=5
};

class SessionItem : public BaseObject
{
  ProcessItemType type;
  Schedule *sh;
  int ok;
  char *mask;
  char *errMsg;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:

  SessionItem( ProcessItemType _type, char *_mask, Schedule *_sh );
  ~SessionItem();

  ProcessItemType getType() const { return type; }
  Schedule *getSchedule() const { return sh; }

  char *getRegErrorMsg() const { return errMsg; }
  
  char *toString( char *buf, size_t size ) const;

  int compare( char * ) const;

  int process( FIDOAddress &adr, long elapsed, char *name )
    { return doEventKey( *sh, -1, &adr, elapsed, name ); }

  int isOK() const { return ok; }
};

class ProcessList : public Collection			//SessionItem
{
public:
  ProcessList();

  SessionItem* Find( SessionItem& item, LockMode lock=LockModeNeed ) const;
  SessionItem& operator [] ( const int pos ) const;
};

class ProcessIterator : public CollectionIterator	//SessionItem
{
public:
  ProcessIterator( ProcessList& ini );

  SessionItem& operator () () const;
};

extern ProcessList *processTable;
extern ProcessList *onlineTable;
extern ProcessList *immediateTable;

void processInbound( ReceivedFileList& , ProcessList*, FIDOAddress&, long );
void processImmediate( char *name );

#endif
