#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <io.h>
#include <ctype.h>
#include <fstream.h>
#include <time.h>
#include <share.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <builtin.h>
#include <errno.h>

#include "language.h"

#include "misc.h"
#include "queue.h"
#include "translat.h"
#include "enable.h"
#include "rexxhook.h"
#include "cost.h"
#include "direct.h"
#include "ea.h"

Try::Try( char *path, int flag ) :
  last( newStr( Msg_ScrNoCall ) ),
  ok( 0 ),
  bad( 0 ),
  file( NULL ),
  rest( 256, 256 ),
  full( newStr( path ) ),
  lastTime( 0 ),
  okTime( 0 ),
  holdTo( 0 ),
  changed( 0 ),
  line( 0 )
{
  makeFilePath( full );

  char *mode( !flag ? "r" : (access( full, 0 ) ? "w+" : "r+") );

  for( int i(0); i<10; i++ )
  {
    file = _fsopen( full, mode, SH_DENYRW );

    if( file!=NULL || (errno==ENOENT && !flag) )
      break;

    sleep( 100 );
  }

  if( file==NULL )
  {
    if( errno!=ENOENT )
      log( Msg_ErrUnableOpen, full, strError() );

    return;
  }

  Buffer buf1,
         buf2;

  while( 1 )
  {
    getKeyword( file, buf1 );

    if( buf1.length()<2 )
      break;

    getValue( file, buf2 );

    if( !stricmp( buf1.str, "OkCount" ) )
      ok = (short)decode( buf2.str, 0, "OkCount" );
    else if( !stricmp( buf1.str, "BadCount" ) )
      bad = (short)decode( buf2.str, 0, "BadCount" );
    else if( !stricmp( buf1.str, "Line" ) )
      line = decode( buf2.str, 0, "Line" );
    else if( !stricmp( buf1.str, "OkTime" ) )
      okTime = decode( buf2.str, '-', "OkTime" );
    else if( !stricmp( buf1.str, "LastTime" ) )
      lastTime = decode( buf2.str, '-', "LastTime" );
    else if( !stricmp( buf1.str, "HoldTo" ) )
      holdTo = decode( buf2.str, '-', "HoldTo" );
    else if( !stricmp( buf1.str, "LastCall" ) )
    {
      delete last;
      last = newStr( buf2.str );
    }
    else if( buf1.length()>1 )
    {
      rest.putQuotted( buf1.str );

      if( buf2.length()>1 )
      {
        rest.putInto( ' ' );
        rest.putQuotted( buf2.str );
      }

      rest.putInto( '\n' );
    }
  }
}

Try::~Try()
{
  int rc(0);

  if( file )
  {
    if( changed )
    {
      rewind( file );

      if( ok )
        rc = putPair( file, "OkCount", ok );

      if( !rc )
        rc = putPair( file, "BadCount", bad );

      if( !rc )
        rc = putPair( file, "LastCall", last );

      if( !rc )
        rc = putPair( file, "Line", line );

       saveTime( rc, "LastTime", lastTime );
       saveTime( rc, "OkTime", okTime );
       saveTime( rc, "HoldTo", holdTo );

      if( !rc )
        rc = fputs( rest.str, file )==EOF ? 1 : 0;

      if( rc )
        log( Msg_ErrUnableWrite, full, strError() );
      else
      {
        fflush( file );

        long pos( _tell( fileno(file) ) );

        if( pos!=-1 )
          _chsize( fileno(file), pos );
      }
    }

    fclose( file );

    if( rc )
      forceDelete( full );
    else
      setExtendedAttribute( full, ".TYPE", Msg_EATry );
  }

  if( !rc )
    setExtendedAttribute( full, ".TYPE", Msg_EATry );

  delete last;
  delete full;
}

void Try::saveTime( int& rc, char *key, unsigned long lastTime )
{
  if( rc )
    return;

  char buf[128];

  sprintf( buf, "%lu-", lastTime );

  int len( strlen(buf) );

  if( lastTime )
  {
    struct tm *tt( localTime( (long *)&lastTime ) );  

    strftime( buf+len, sizeof buf - len, "%d/%m/%Y-%H:%M:%S", tt );
  }
  else
    strcpy( buf+len, "00/00/00-00:00:00" );

  rc = putPair( file, key, buf );
}

void Try::setInfo( int badFlag, int okFlag, char *l )
{
  delete last;
  last = newStr( l );

  if( badFlag )
    bad++;

  if( okFlag )
  {
    ok++;
    bad = 0;
    okTime = time( NULL );
  }

  lastTime = time( NULL );
  line = miscInfo.taskNumber;
  changed = 1;
}

unsigned long Try::decode( char *str, char last, char *msg )
{
  char *e;
  unsigned long ret( strtoul( str, &e, 10 ) );

  if( *e!=last )
  {
    log( Msg_ErrIllegalInTry, full, msg, str );
    ret = 0;
  }

  return ret;
}

QueueEntry::QueueEntry( const NodeConfig& cfg ) :
  NodeConfig ( &cfg ),
  mapLine( 0 ),
  iMenable( 1 ),
  iAcceptable( 1 ),
  tryLast( NULL ),
  costFault( 0 ),
  ours( 0 ),
  imCur( 0 )
{
  clear();

  if( enabledList!=NULL )
    if( existIn( enabledList ) )
      iMenable = 1;
    else
      iMenable = 0;

  if( iMenable && disabledList!=NULL )
    if( existIn( disabledList ) )
      iMenable = 0;

  if( acceptedList!=NULL )
    if( existIn( acceptedList ) )
      iAcceptable = 1;
    else
      iAcceptable = 0;

  if( iAcceptable && unAcceptedList!=NULL )
    if( existIn( unAcceptedList ) )
      iAcceptable = 0;

  FIDOAddress addr( getZone(), getNet(), getNode(), getPoint(), getDomain() );

  if( systemInfo.addr->Find( addr )!=NULL )
    ours = 1;

  loadTry();
}

QueueEntry::QueueEntry( const QueueEntry& cfg ) :
  NodeConfig ( (const NodeConfig * const)&cfg )
{
  memcpy( &traf, &cfg.traf, sizeof(traf) );

  mapLine = cfg.mapLine;
  iMenable = cfg.iMenable;
  iAcceptable = cfg.iAcceptable;
  costFault = cfg.costFault;

  tryOK = cfg.tryOK;
  tryBad = cfg.tryBad;
  tryLast = newStr( cfg.tryLast );

  ours = cfg.ours;
  imCur = cfg.imCur;
}

QueueEntry::~QueueEntry()
{
  delete tryLast;
}

int QueueEntry::equal( const BaseObject &t ) const
{
  return FIDOAddress::equal( *(FIDOAddress const *)&t );
}

int QueueEntry::equal( const QueueEntry &t ) const
{
  switch( miscInfo.sortMode )
  {
  case QueueSortSizeDec:
  case QueueSortSize:
    return getMail()+getFiles() == t.getMail()+t.getFiles();

  case QueueSortPriorityDec:
  case QueueSortPriority:
    return getPriority() == t.getPriority();

  case QueueSortAgeDec:
  case QueueSortAge:
    return getAge() == t.getAge();
  }

  return FIDOAddress::equal( *(FIDOAddress const *)&t );
}

int QueueEntry::less( const QueueEntry &t ) const
{
  int less,
      dec = 0;

  switch( miscInfo.sortMode )
  {
  case QueueSortSizeDec:
    dec = 1;
    less = getMail()+getFiles() < t.getMail()+t.getFiles();
    break;

  case QueueSortSize:
    less = getMail()+getFiles() > t.getMail()+t.getFiles();
    break;

  case QueueSortPriorityDec:
    dec = 1;
    less = getPriority() < t.getPriority();
    break;

  case QueueSortPriority:
    less = getPriority() > t.getPriority();
    break;

  case QueueSortAgeDec:
    dec = 1;
    less = getAge() < t.getAge();
    break;

  case QueueSortAge:
    less = getAge() > t.getAge();
    break;

  case QueueSortAddrDec:
DecAddr:
    if( FIDOAddress::less( *(FIDOAddress const *)&t ) )
      return 0;

    return FIDOAddress::equal( *(FIDOAddress const *)&t ) ? 0 : 1 ;

  default:
NormAddr:
    return FIDOAddress::less( *(FIDOAddress const *)&t ) ? 1 : 0;
  }

  if( less )
    return less;

  if( equal( *(const QueueEntry *)&t ) )
    if( dec )
      goto DecAddr;
    else
      goto NormAddr;
  else
    return 0;
}

void QueueEntry::print( ostream& s ) const
{
  QueueEntry c( *this );

  NodeConfig::print( s );

  s << endl
    << "    QFlav:" << flavourName( c.getQueuedFlavour() ) << endl
    << "    TFlav:"      << flavourName( getTrafficFlavour() ) << endl
    << "    Prio:"       << getPriority() << endl
    << "    HoldTo:"     << traf.holdTo << endl
    << "    Enable:"     << (iMenable ? "Yes" : "No" ) << endl
    << "    Acceptable:" << (iAcceptable ? "Yes" : "No" ) << endl
    << "    CostFault:"  << (costFault ? "Yes" : "No" ) << endl
    << "    Age:"        << traf.age << endl

    << "    Mail: Crash:" << traf.crashMail << endl
    << "          Normal:" << traf.normalMail << endl
    << "          Immed:"  << traf.immedMail << endl
    << "          Direct:" << traf.directMail << endl
    << "          Hold:"   << traf.holdMail << endl

    << "    XMail: Crash:" << traf.crashXMail << endl
    << "           Normal:" << traf.normalXMail << endl
    << "           Immed:"  << traf.immedXMail << endl
    << "           Direct:" << traf.directXMail << endl
    << "           Hold:"   << traf.holdXMail << endl

    << "    Files: Crash:" << traf.crashFiles << endl
    << "           Normal:" << traf.normalFiles << endl
    << "           Immed:"  << traf.immedFiles << endl
    << "           Direct:" << traf.directFiles << endl
    << "           Hold:"   << traf.holdFiles  << endl

    << "    Request:" << traf.request;
}

char *QueueEntry::nameOfClass() const
{
  return "QueueEntry";
}

char QueueEntry::getPollMode() const
{
  return char( traf.poll!=unknownFlavour);
}

void QueueEntry::setPollMode( Flavour mode )
{
  traf.poll=mode;
}

void QueueEntry::unMap()
{
  mapLine=0;
}

void QueueEntry::setHoldTo( unsigned long delta )
{
  traf.holdTo=delta;
}

uLong QueueEntry::holdSize() const
{
  return traf.holdMail+traf.holdXMail+traf.holdFiles+traf.request;
}

uLong QueueEntry::normalSize() const
{
  return traf.normalMail+traf.normalXMail+traf.normalFiles;
}

uLong QueueEntry::crashSize() const
{
  return traf.crashMail+traf.crashXMail+traf.crashFiles;
}

uLong QueueEntry::immedSize() const
{
  return traf.immedMail+traf.immedXMail+traf.immedFiles;
}

uLong QueueEntry::directSize() const
{
  return traf.directMail+traf.directXMail+traf.directFiles;
}

long QueueEntry::getMail() const
{
  return traf.crashMail+traf.normalMail+traf.holdMail+traf.immedMail+
         traf.directMail;
}

long QueueEntry::getFiles() const
{
  return traf.crashFiles+traf.normalFiles+traf.holdFiles+
         traf.immedFiles+traf.directFiles+
         traf.crashXMail+traf.normalXMail+traf.holdXMail+
         traf.immedXMail+traf.directXMail+
         traf.request;
}

void QueueEntry::loadTry()
{
  delete tryLast;
  tryLast = NULL;
  tryOK = 0;
  tryBad = 0;

  char *full( getTryFileName() );

  if( !iMenable || full==NULL )
  {
    tryLast = newStr( Msg_ScrNoCall );
    delete full;
    return;
  }

  Try a( full, 0 );

  delete full;

  tryOK = a.getOk();
  tryBad = a.getBad();
  tryLast = newStr( a.getLast() );
}

void QueueEntry::saveTry( int bad, int ok )
{
  char *full( getTryFileName() );

  if( full==NULL )
    return;

  Try a( full, 1 );

  delete full;

  a.setInfo( bad, ok, tryLast );
  tryOK = a.getOk();
  tryBad = a.getBad();
}

void QueueEntry::setTryLast( char *msg )
{
  if( msg==NULL )
    return;

  delete tryLast;
  tryLast = newStr( msg );

  char bad(0), ok(0);

  if( begin( tryLast, modemString.connect ) )
  {
    ok = 1;
    strcpy( tryLast, tryLast+strlen(modemString.connect) );

    while( *tryLast==' ' )
      strcpy( tryLast, tryLast+1 );
  }
  else
  {
    char found=0;

    for( ModemStringIterator it( *badCounterString ); it && !found; it++ )
      if( !stricmp( msg, it().getString() ) )
        found=1;

    if( found )
      bad = 1;
  }

  saveTry( bad, ok );

  queue->Lock();

  QueueEntry *found = queue->Find( *this, LockModeNoNeed );

  if( found!=NULL )
  {
    delete found->tryLast;
    found->tryLast = newStr( msg );
    found->tryOK = tryOK;
    found->tryBad = tryBad;
  }

  queue->UnLock();

  if( isMapped() )
    draw( isMapped(), queue->getCurrent() );
}

int QueueEntry::existIn( EnableList *listPtr )
{
  char buf[65];
  int rc = 0;
  char *gettedFlags = getFlags();
  char *flags = gettedFlags==NULL ? (char *)NULL :
                                    (char *)_alloca( strlen(gettedFlags)+2 );

  toString( buf, 64, FIDOAddressShowModeEMSI );

  for( EnableIterator list( *listPtr ); !rc && list; list++ )
  {
    EnableItem& cur = list();

    switch( cur.getType() )
    {
    case EnableItemAddress:
      rc = cur.isUnify( buf );
      break;

    case EnableItemFlag:
      if( gettedFlags!=NULL )
      {
        strcpy( flags, gettedFlags );
	strcat( flags, "," );

	for( char *p=flags; *p && !rc; )
	{
	  char *coma = strchr( p, ',' );

	  if( coma==NULL )
	    break;

	  *coma = 0;

	  rc = cur.isUnify( p );

	  p = coma+1;
	}
      }
      break;
    }
  }

  return rc;
}

void QueueEntry::setAge( char *name, long created )
{
  FileType type;

  if( name )
    msgOfFileType( name, &type );
  else
    type = FileMail;

  if( type==FileMail || type==FileXMail )
    if( (created<traf.age && traf.age) || !traf.age )
      traf.age = created;
}

int QueueEntry::checkHoldState()
{
  if( checkHoldState( getLongBinkName(), ".Freeze" ) ||
      checkHoldState( getBinkName(), ".Hld" ) )
    return 1;
    
  return 0;
}

int QueueEntry::checkHoldState( char *p, char *s )
{
  if( p==NULL )
    return 0;
    
  char *full( (char *)_alloca( strlen(p)+strlen(s)+1 ) );
  
  strcpy( full, p );
  strcat( full, s );

  delete p;
  
  return !access( full, 0 );
}

void QueueEntry::setMailTraf( Flavour fl, long size, long created )
{
  setAge( 0, created );

  switch( fl )
  {
  case immedFlavour:
    traf.immedMail += size;
    break;

  case crashFlavour:
    traf.crashMail += size;
    break;

  case directFlavour:
    traf.directMail += size;
    break;

  case normalFlavour:
    traf.normalMail += size;
    break;

  default:
    traf.holdMail += size;
  }
}

void QueueEntry::addTraffic( Traffic& tr )
{
  traf.holdMail += tr.holdMail;
  traf.crashMail += tr.crashMail;
  traf.normalMail += tr.normalMail;
  traf.immedMail += tr.immedMail;
  traf.directMail += tr.directMail;

  traf.holdXMail += tr.holdXMail;
  traf.crashXMail += tr.crashXMail;
  traf.normalXMail += tr.normalXMail;
  traf.immedXMail += tr.immedXMail;
  traf.directXMail += tr.directXMail;

  traf.holdFiles += tr.holdFiles;
  traf.crashFiles += tr.crashFiles;
  traf.normalFiles += tr.normalFiles;
  traf.immedFiles += tr.immedFiles;
  traf.directFiles += tr.directFiles;

  traf.request += tr.request;

  if( flavourPriority( tr.poll ) > flavourPriority( traf.poll ) )
    traf.poll = tr.poll;

  if( tr.age>traf.age )
    traf.age = tr.age;
    
  if( tr.holdTo )
    traf.holdTo = tr.holdTo;
}

void QueueEntry::addTraffic( char *rez )
{
  if( rez==NULL )
    return;

  char *flav = (char *)_alloca( strlen(rez)+10 );
  unsigned long cm(0), cxm(0), cf(0),
                nm(0), nxm(0), nf(0),
                hm(0), hxm(0), hf(0),
                im(0), ixm(0), iif(0),
                dm(0), dxm(0), df(0),
                req(0);

  sscanf( rez, "%s %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu",
          flav,
          &im, &ixm, &iif,
          &cm, &cxm, &cf,
          &dm, &dxm, &df,
          &nm, &nxm, &nf,
          &hm, &hxm, &hf,
          &req );

  traf.holdMail += hm;
  traf.crashMail += cm;
  traf.normalMail += nm;
  traf.immedMail += im;
  traf.directMail += dm;

  traf.holdXMail += hxm;
  traf.crashXMail += cxm;
  traf.normalXMail += nxm;
  traf.immedXMail += ixm;
  traf.directXMail += dxm;

  traf.holdFiles += hf;
  traf.crashFiles += cf;
  traf.normalFiles += nf;
  traf.immedFiles += iif;
  traf.directFiles += df;

  traf.request += req;

  if( !stricmp( flav, "Crash" ) )
    setFlavour( crashFlavour );
  else if( !stricmp( flav, "Normal" ) )
    setFlavour( normalFlavour );
  else if( !stricmp( flav, "Hold" ) )
    setFlavour( holdFlavour );
  else if( !stricmp( flav, "Direct" ) )
    setFlavour( directFlavour );
  else if( !stricmp( flav, "Immed" ) )
    setFlavour( immedFlavour );
  else if( stricmp( flav, "Default" ) )
    log( Msg_RexxIllegalFlavour, flav );
}

char *centreStr( char *s, int len )
{
  int sLen( strlen(s));

  if( sLen>=len )
    return s;
  else
  {
    char *r( new char [len+1] );

    memset( r, ' ', len );
    strcpy( r+(len-sLen)/2+1, s );

    delete s;
    return r;
  }
}

void QueueEntry::infoShowNdl( void *q )
{
  Window& w = *(Window *)q;

  Buffer b;
  CacheItem *ndl = getNdl();
  char bbb[64];
  int width = w.getWidth()-2;
  char *c;

  if( ndl==NULL )
    if( getPhone()==NULL )
      w.centre( 3, Msg_ScrViewNotListed );
    else
    {
      w.centre( 2, Msg_ScrViewNotListed );
      b.reset();

      int i=getCurrentLine();

      while( 1 )
      {
        c = getTranslatedPhone( getPhone() );

        if( c==NULL )
          c = newStr( Msg_ScrUnpublishedPhone );

        b.putString( c );
        b.putString( ", " );

        delete c;

        setCurrent();

        if( i==getCurrentLine() )
          break;
      }

      if( getWorkTime()!=NULL )
      {
        b.putString( getWorkTime()->getStr() );
        b.putString( ", " );
      }

      if( !b.empty() )
      {
        b.stripByte();
        b.stripByte();
      }

      b.putInto(0);

      if( strlen( b.str )>width )
        b.str[ width ] = 0;

      w.centre( 3, b.str );
    }
  else
  {
    c = ndl->getSystem();

    if( c!=NULL )
      b.putString( c );

    c = ndl->getSysop();

    if( c!=NULL )
    {
      b.putString( ", " );
      b.putString( c );
    }

    c = ndl->getNdlField( 4 );

    if( c!=NULL )
    {
      b.putString( ", " );

      while( *c  )
      {
        if( *c=='_' )
          b.putInto( ' ' );
        else
          b.putInto( *c );
        
        c++;
      }
    }

    b.putInto(0);

    if( strlen( b.str )>width )
      b.str[ width ] = 0;

    w.centre( 2, b.str );
    b.reset();

    c = ndl->getPhone();

    b.putString( c!=NULL ? c : Msg_ScrUnpublishedPhone );

    sprintf( bbb, "%d", ndl->getBaud() );
    b.putString( ", " );
    b.putString( bbb );

    c = ndl->getFlags();

    if( c!=NULL )
    {
      b.putString( ", " );

      while( *c )
      {
        b.putInto( *c );

        if( *c==',' )
          b.putInto( ' ' );

        c++;
      }
    }

    b.putInto(0);

    if( strlen( b.str )>width )
      b.str[ width ] = 0;

    w.centre( 3, b.str );
    b.reset();

    int i=getCurrentLine();

    while( 1 )
    {
      c = getTranslatedPhone( getPhone() );

      if( c==NULL )
        c = newStr( Msg_ScrUnpublishedPhone );

      b.putString( c );
      b.putString( ", " );

      delete c;

      setCurrent();

      if( i==getCurrentLine() )
        break;
    }

    if( !b.empty() )
    {
      b.stripByte();
      b.stripByte();
    }

    if( b.empty() )
      b.putString( Msg_ScrViewNoPhone );

    b.putInto(0);

    if( strlen( b.str )>width )
      b.str[ width ] = 0;

    w.centre( 4, b.str );
    b.reset();
  }

  b.reset();

  if( getPassword()!=NULL )
  {
    b.putString( Msg_ScrViewPassword1 );
    b.putString( getPassword() );
    b.putString( Msg_ScrViewPassword2 );
  }
  else
    b.putString( Msg_ScrViewPassword3 );

  b.putInto(0);

  if( strlen( b.str )>width )
    b.str[ width ] = 0;

  w.centre( 5, b.str );
}

void QueueEntry::info()
{
  char ttl[64];

  toString( ttl+1, sizeof(ttl)-2 );
  *ttl = ' ';
  strcat( ttl, " " );

  const widthWin = 62,
        height = 22;

  Window w( (Z_ScreenWidth-widthWin)/2+1, (Z_ScreenHeight-height)/2+1,
            widthWin, height,
            popUpColor,
            ttl,
            BS_SINGLELINE,
            DS_LOWERRIGHT );

  infoShowNdl( (void *)&w );

  w( 3, 7 )  << Msg_ScrQEntInfo01;
  w( 3, 8 )  << Msg_ScrQEntInfo02;
  w( 3, 9 )  << Msg_ScrQEntInfo03;
  w( 3, 10 ) << Msg_ScrQEntInfo04;
  w( 3, 11 ) << Msg_ScrQEntInfo05;
  w( 3, 12 ) << Msg_ScrQEntInfo06;
  w( 3, 13 ) << Msg_ScrQEntInfo07;
  w( 3, 14 ) << Msg_ScrQEntInfo08;
  w( 3, 15 ) << Msg_ScrQEntInfo09;
  w( 3, 17 ) << Msg_ScrQEntInfo10;
  w( 3, 18 ) << Msg_ScrQEntInfo11;

  time_t nowIs( time( NULL ) );
  struct tm *tm( localtime( &nowIs ) );

  if( tm == NULL )
The_BUG:
    w( 17, 29 ) << "BUG";
  else
  {
    if( tm->tm_wday == 0 )
      tm->tm_wday = 7;

    nowIs -= (tm->tm_wday-1)*24*60*60;
    tm = localtime( &nowIs );

     if( tm == NULL )
       goto The_BUG;

    tm->tm_hour = 0;
    tm->tm_min = 0;
    tm->tm_sec = 0;

    nowIs = mktime( tm );

    for( int i=0; i<7; i++ )
      for( int j=0; j<48; j++ )
      {
        w( 10+j, 7+i ) << char( isWorked( nowIs ) ? '' : ' ');
        nowIs += 30*60;
      }
  }

  w.centre( 20, Msg_ScrViewPressAnyKey );

  +w;

  while( kbhit() )
    Z_GetKey();

  while( Z_GetKey()!=27 ) continue;
}

void QueueEntry::show()
{
  char ttl[64];

  toString( ttl+1, sizeof(ttl)-2 );
  *ttl = ' ';
  strcat( ttl, " " );

  const widthWin = 62,
        height = 21;

  Window w( (Z_ScreenWidth-widthWin)/2+1, (Z_ScreenHeight-height)/2+1,
            widthWin, height,
            popUpColor,
            ttl,
            BS_SINGLELINE,
            DS_LOWERRIGHT );

  char *fla = Msg_ScrEvenNone;

  switch( traf.poll )
  {
  case normalFlavour:
    fla = Msg_FlavourNormal;
    break;

  case crashFlavour:
    fla = Msg_FlavourCrash;
    break;

  case holdFlavour:
    fla = Msg_FlavourHold;
    break;

  case immedFlavour:
    fla = Msg_FlavourImmed;
    break;

  case directFlavour:
    fla = Msg_FlavourDirect;
    break;
  }

  Buffer b;
  char bbb[64];
  int width = w.getWidth()-2;

  infoShowNdl( (void *)&w );

  b.reset();
  b.putString( Msg_ScrViewPoll );
  b.putString( fla );
  b.putString( ", " );
  b.putString( Msg_ScrViewCallable );
  b.putString( iMenable ? Msg_ScrViewYes : Msg_ScrViewNo );
  b.putString( ", " );
  b.putString( Msg_ScrViewAcceptable );
  b.putString( iAcceptable ? Msg_ScrViewYes : Msg_ScrViewNo );
  b.putInto( 0 );

  if( strlen( b.str )>width )
    b.str[ width ] = 0;

  capitalize( b.str );
  w.centre( 7, b.str );

  b.reset();

  b.putString( Msg_ScrViewLastTry );
  sprintf( bbb, "%d/%d", tryOK, tryBad );
  b.putString( bbb );
  b.putString( ", " );
  b.putInto( ' ' );

  if( tryExceeded() )
  {
    b.putString( Msg_ScrUnDial );
    b.putString( ", " );
    b.putInto( ' ' );
  }

  sprintf( bbb, "%d", traf.age/60/60 );
  b.putString( Msg_ScrViewAge );
  b.putString( bbb );
  b.putInto( '[' );
  sprintf( bbb, "%d", traf.age/60/60/24 );
  b.putString( bbb );
  b.putInto( ']' );
  b.putString( ", " );
  b.putInto( ' ' );

  b.putInto( (char)4 );
  b.putInto( ' ' );
  b.putString( tryLast!=NULL ? tryLast : Msg_ScrNoCall );
  b.putInto( ' ' );
  b.putInto( (char)4 );
  b.putInto( 0 );

  if( strlen( b.str )>width )
    b.str[ width ] = 0;

  capitalize( b.str );
  w.centre( 9, b.str );

  b.reset();

  w( 10, 11 ) <<  Msg_ScrViewHdrImmed;
  w( 20, 11 ) <<  Msg_ScrViewHdrCrash;
  w( 30, 11 ) <<  Msg_ScrViewHdrDirect;
  w( 40, 11 ) <<  Msg_ScrViewHdrNormal;
  w( 50, 11 ) <<  Msg_ScrViewHdrHold;
  w( 10, 12 ) <<  Msg_ScrViewHdrLine;
  w( 20, 12 ) <<  Msg_ScrViewHdrLine;
  w( 30, 12 ) <<  Msg_ScrViewHdrLine;
  w( 40, 12 ) <<  Msg_ScrViewHdrLine;
  w( 50, 12 ) <<  Msg_ScrViewHdrLine;

  w( 2, 14 )  << Msg_ScrViewMails;
  w( 2, 15 )  << Msg_ScrViewEchos;
  w( 2, 16 )  << Msg_ScrViewFiles;
  w( 2, 17 )  << Msg_ScrViewFREQs;

  w( 10, 14 )  << traf.immedMail;
  w( 10, 15 )  << traf.immedXMail;
  w( 10, 16 )  << traf.immedFiles;

  w( 20, 14 )  << traf.crashMail;
  w( 20, 15 )  << traf.crashXMail;
  w( 20, 16 )  << traf.crashFiles;

  w( 30, 14 ) << traf.directMail;
  w( 30, 15 ) << traf.directXMail;
  w( 30, 16 ) << traf.directFiles;

  w( 40, 14 ) << traf.normalMail;
  w( 40, 15 ) << traf.normalXMail;
  w( 40, 16 ) << traf.normalFiles;

  w( 50, 14 ) << traf.holdMail;
  w( 50, 15 ) << traf.holdXMail;
  w( 50, 16 ) << traf.holdFiles;
  w( 50, 17 ) << traf.request;

  w.centre( 19, Msg_ScrViewPressAnyKey );

  +w;

  while( kbhit() )
    Z_GetKey();

  while( Z_GetKey()!=27 ) continue;
}

void QueueEntry::format( char *buf, int flag )
{
  flag <<= 1;

  long  sMail = traf.crashMail+traf.normalMail+traf.holdMail+
                traf.immedMail+traf.directMail,
       sFiles = traf.crashXMail+traf.normalXMail+traf.holdXMail+
                traf.immedXMail+traf.directXMail +
                traf.crashFiles+traf.normalFiles+traf.holdFiles+
                traf.immedFiles+traf.directFiles +
                traf.request;

  char adr[40],
       *em( "-" ),
       *mail(  centreStr( sMail  ? strBig( sMail )  : newStr(em), 5 ) ),
       *files( centreStr( sFiles ? strBig( sFiles ) : newStr(em), 5 ) );
  int myage = int( traf.age/60/60/24 );

  if( !myage )
    myage = ' ';
  else if( myage>9 )
  {
    myage += 'A'-10;

    if( myage>'Z' )
      myage = '*';
  }
  else
    myage += '0';

  sprintf( buf, "%s%-27s %3d/%-3d %-5s %-5s %c................ ",
           flag==0 ? "  " : "",
	   toString( adr, 28 ),
	   tryOK%1000,
	   tryBad%1000,
	   mail, files,
           char( myage ) );

  if( traf.immedMail )
    buf[ 51-flag ] = 'M';
  if( traf.immedXMail )
    buf[ 52-flag ] = 'X';
  if( traf.immedFiles )
    buf[ 53-flag ] = 'F';

  if( traf.crashMail )
    buf[ 54-flag ] = 'M';
  if( traf.crashXMail )
    buf[ 55-flag ] = 'X';
  if( traf.crashFiles )
    buf[ 56-flag ] = 'F';

  if( traf.directMail )
    buf[ 57-flag ] = 'M';
  if( traf.directXMail )
    buf[ 58-flag ] = 'X';
  if( traf.directFiles )
    buf[ 59-flag ] = 'F';

  if( traf.normalMail )
    buf[ 60-flag ] = 'M';
  if( traf.normalXMail )
    buf[ 61-flag ] = 'X';
  if( traf.normalFiles )
    buf[ 62-flag ] = 'F';

  if( traf.holdMail )
    buf[ 63-flag ] = 'M';
  if( traf.holdXMail )
    buf[ 64-flag ] = 'X';
  if( traf.holdFiles )
    buf[ 65-flag ] = 'F';

  if( traf.request )
    buf[ 66-flag ] = 'R';

  if( !iMenable )
    strcpy( buf+68-flag, Msg_ScrDisable );
  else if( tryExceeded() )
    strcpy( buf+68-flag, Msg_ScrUnDial );
  else if( isOurs() )
    strcpy( buf+68-flag, Msg_ScrOurs );
  else if( getQueuedFlavour()==holdFlavour )
  {
    if( traf.holdTo )
      if( traf.holdTo==(unsigned long)-1l )
        strcpy( buf+68-flag, Msg_ScrHoldForever );
      else
      {
        struct tm *tm = localTime( (time_t *)&traf.holdTo );

        sprintf( buf+68-flag, Msg_ScrHoldOn, tm->tm_hour, tm->tm_min );
      }
    else
      strcpy( buf+68-flag, Msg_ScrHold );
  }
  else if( !isWorked( time( NULL ) ) || !isFlavourCallable() )
    strcpy( buf+68-flag, Msg_ScrTimeWait );
  else if( costFault )
    strcpy( buf+68-flag, Msg_ScrCost );
  else
    strcpy( buf+68-flag, tryLast );

  delete mail;
  delete files;
}

void QueueEntry::draw( int line, QueueEntry *cur )
{
  if( line>wndQueue.getHeight() )
  {
    mapLine = 0;
    return;
  }

  int isCur( cur!=NULL && *cur==*this );

  if( !detachedSession ) //&& (mapLine!=line || imCur!=isCur) )
  {
    char *buf = (char *)_alloca( 128 + 
                                 (tryLast==NULL ? 80 : strlen( tryLast )) );

    format( buf );

    Color cl = wndQueue.getColor();
    Color cl1( colorSet.queuePtr );
    char c;

    switch( getQueuedFlavour() )
    {
    case holdFlavour:
      cl = Color( colorSet.queueHold );
      c = 'h';
      break;

    case crashFlavour:
      cl = Color( colorSet.queueCrash );
      c = 'c';
      break;

    case normalFlavour:
      cl = Color( colorSet.queueNormal );
      c = 'n';
      break;

    case immedFlavour:
      cl = Color( colorSet.queueImmed );
      c = 'i';
      break;

    case directFlavour:
      cl = Color( colorSet.queueDirect );
      c = 'd';
      break;

    default:
      cl = Color( RED, wndQueue.getColor().back );
      c = '?';
      break;
    }

    buf[0] = checkBsy() ? 'b' : c;

    if( traf.busy )
      buf[0] = char( toupper( buf[0] ) );

    int windowLen = wndQueue.getWidth();

    buf[ windowLen ] = 0;

    for( int len = strlen( buf ); len<windowLen; len++ )
      buf[ len ] = ' ';

    buf[ len ] = 0;

    Z_TextOut( wndQueue.getWindow(), buf, 0, line-1, cl.fore, cl.back );

    if( isCur )
      Z_CharOut( wndQueue.getWindow(), 16, 1, line-1, cl1.fore, cl1.back );
  }

  mapLine = (char)line;
  imCur = char( isCur ? 1 : 0 );
}

int QueueEntry::tryExceeded() const
{
  int i = getMaxTry();

  if( i && tryBad>i )
    return 1;

  return 0;
}

void QueueEntry::rescan()
{
  int statusDelta = (miscInfo.state1&StatusLine)==StatusLine ? 0 : 1;
  char buf[ 128 ];
  Window *wnd;

  toString( buf, sizeof buf );

  if( !detachedSession )
  {
    wnd = new Window( 3, Z_ScreenHeight-1+statusDelta,
                      strlen(buf)+4, 1, popUpColor );
    (*wnd).centre( 1, buf );
    +(*wnd);
  }

  preRescan();

  if( getRexxHook( "BeforeScanNode" )!=NULL )
  {
    char *rez = processRexxHook( "BeforeScanNode", buf, NULL, NULL,
                                 NULL, NULL, NULL, 1 );

    addTraffic( rez );
    delete rez;
  }

  rescanCfg( ScanTypeWithFiles );
  rescanBink( ScanTypeWithFiles );
  rescanLongBink( ScanTypeWithFiles );
  rescanTBox( ScanTypeWithFiles );
  rescanLongBox( ScanTypeWithFiles );

  postRescan();

  if( getRexxHook( "AfterScanNode" )!=NULL )
  {
    char *rez = processRexxHook( "AfterScanNode", buf, NULL, NULL,
                                                  NULL, NULL, NULL, 1 );

    addTraffic( rez );
    delete rez;
  }

  if( !detachedSession )
    delete wnd;
}

void QueueEntry::preRescan()
{
  traf.busy = 0;
  clear();
}

void QueueEntry::postRescan()
{
  if( traf.age )
    traf.age = time( NULL ) - traf.age;

  if( miscInfo.maxCost )
    costFault = char( calculateCost( this ) > miscInfo.maxCost ? 1 : 0 );
}

void QueueEntry::processHold( char *path, char *name )
{
  char *hldPath = (char *)_alloca( strlen(path)+strlen(name)+1 );

  strcpy( hldPath, path );
  strcat( hldPath, name );

  FILE *hld = _fsopen( hldPath, "r" );

  if( hld==NULL )
    log( Msg_ErrUnableOpen, hldPath, strError() );
  else
  {
    long to( -1 );
    char remove( 0 );

    fscanf( hld, "%lu", &to );

    if( ferror( hld ) )
      log( Msg_ErrUnableRead, hldPath, strError() );
    else if( to!=-1 && time( NULL )>=to )
      remove = 1;
    else
      setHoldTo( to );

    fclose( hld );

    if( remove )
      if( unlink( hldPath ) )
        log( Msg_ErrUnableDelete, hldPath, strError() );
  }
}

void QueueEntry::processLO( ScanType type, char *path, char *name,
                            Flavour myFlavour,
                            long created )
{
  long xloLen;

  switch( myFlavour )
  {
  case immedFlavour:
    xloLen = rescanLO( type, path, name, traf.immedMail, traf.immedXMail, traf.immedFiles, created );
    break;

  case crashFlavour:
    xloLen = rescanLO( type, path, name, traf.crashMail, traf.crashXMail, traf.crashFiles, created );
    break;

  case directFlavour:
    xloLen = rescanLO( type, path, name, traf.directMail, traf.directXMail, traf.directFiles, created );
    break;

  case normalFlavour:
    xloLen = rescanLO( type, path, name, traf.normalMail, traf.normalXMail, traf.normalFiles, created );
    break;

  default:
    xloLen = rescanLO( type, path, name, traf.holdMail, traf.holdXMail, traf.holdFiles, created );
  }

  if( !xloLen && flavourPriority( traf.poll ) < flavourPriority( myFlavour ) )
    traf.poll = myFlavour;
}

void QueueEntry::processShortBink( ScanType type, char *name, char *path,
                                   long size, long created )
{
  if( strlen(name)!=12 )
    return;

// mail
    
  if( toupper(name[10])=='U' && toupper(name[11])=='T' )
    switch( toupper(name[9]) )
    {
    case 'I': setMailTraf( immedFlavour,  size, created );  break;
    case 'C': setMailTraf( crashFlavour,  size, created );  break;
    case 'D': setMailTraf( directFlavour, size, created );  break;
    case 'O':
    case 'N': setMailTraf( normalFlavour, size, created );  break;
    case 'H': setMailTraf( holdFlavour,   size, created );  break;
    }

// lists

  else if( toupper(name[10])=='L' && toupper(name[11])=='O' )
    switch( toupper(name[9]) )
    {
    case 'I': processLO( type, path, name, immedFlavour,  created ); break;
    case 'C': processLO( type, path, name, crashFlavour,  created ); break;
    case 'D': processLO( type, path, name, directFlavour, created ); break;
    case 'F':
    case 'N': processLO( type, path, name, normalFlavour, created ); break;
    case 'H': processLO( type, path, name, holdFlavour,   created ); break;
    }
    
// special
    
  else if( toupper(name[9])=='H' && toupper(name[10])=='L' &&
           toupper(name[11])=='D' )
    processHold( path, name );
  else if( !stricmp( name+8, ".Req" ) )
    traf.request += size;
}

void QueueEntry::rescanBink( ScanType type )
{
  char *binkOut( getBinkName() );

  if( binkOut==NULL )
    return;

  char searchMask[] = ".???",
       *sear  = (char *)_alloca( strlen(binkOut) + sizeof searchMask );

  strcpy( sear, binkOut );
  strcat( sear, searchMask );

  for( Directory rc( sear ); rc; rc++ )
    processShortBink( type, rc.getName(), rc.getPath(), rc.getSize(),
                      rc.getUNIXCreate() );

  delete binkOut;
}

long QueueEntry::rescanLO( ScanType type, char *pathPref, char *pathSuf,
                           uLong& mail, uLong& xmail,
                           uLong& files, long created )
{
  if( pathPref==NULL || pathSuf==NULL )
    return 0;

  char *path( (char *)_alloca( strlen(pathPref)+strlen(pathSuf)+1 ) );

  strcpy( path, pathPref );
  strcat( path, pathSuf );

  if( (miscInfo.state3&FastScan)==FastScan && type!=ScanTypeWithFiles )
  {
FastOnly:
    struct stat st;

    if( !stat( path, &st ) )
      files += st.st_size;

    setAge( 0, created );
    return st.st_size;
  }

  if( type!=ScanTypeWithFiles )
    if( makeBsy(1) )
      goto FastOnly;

  FILE *xlo( _fsopen( path, "r" ) );

  if( xlo==NULL )
  {
    log( Msg_ErrUnableOpen, path, strError() );
    return 0;
  }

  char name[_MAX_PATH];
  struct stat st;
  long xloLen( filelength( fileno( xlo ) ) );

  if( !xloLen )
    setAge( 0, created );
  else
    while( !feof( xlo ) )
    {
      char *test( name );

      *name=0;
      fgets( name, _MAX_PATH, xlo );

      if( !*name )
        break;

      if( name[0]=='#' || name[0]=='^' || name[0]=='~' )
        test++;

      register int len( strlen(test)-1 );

      if( test[ len ]=='\n' )
        test[ len ] = 0;

      if( *test && !stat( test, &st ) )
      {
        FileType t;
        msgOfFileType( test, &t );

        setAge( test, st.st_ctime );

        switch( t )
        {
        case FileMail:
	  mail += st.st_size;
  	  break;

        case FileXMail:
	  xmail += st.st_size;
	  break;

        default:
	  files += st.st_size;
        }
      }
    }

  fclose( xlo );
  
  if( type!=ScanTypeWithFiles )
    makeBsy(0);

  return xloLen;
}

void QueueEntry::rescanPath( ScanType scanType, char *path, Flavour fl )
{
  if( path==NULL )
    return;

  static char mask[]( "*" );
  int len( strlen(path) );
  char *full( (char *)_alloca( len+sizeof(mask)+1 ) );
  char *test( (char *)_alloca( len+_MAX_PATH+1 ) );

  strcpy( full, path );
  --len;

  if( *full && full[ len ]!='\\' && full[ len ]!='/' && full[ len ]!=':' )
    strcat( full, "\\" );

  strcat( full, mask );

  uLong mail(0),
        xmail(0),
        files(0);

  for( Directory dir( full ); dir; dir++ )
  {
    register long size( dir.getSize() );

    if( size )
    {
      FileType t;
      char *name( dir.getName() );

      msgOfFileType( name, &t );
      strcpy( test, dir.getPath() );
      strcat( test, name );

      setAge( test, dir.getUNIXCreate() );

      switch( t )
      {
      case FileMail:
	mail += size;
	break;

      case FileXMail:
	xmail += size;
	break;

      default:
	files += size;
      }

      if( (miscInfo.state3&FastScan)==FastScan && scanType!=ScanTypeWithFiles )
        break;
    }
  }

  switch( fl )
  {
  case immedFlavour:
    traf.immedMail += mail;
    traf.immedXMail += xmail;
    traf.immedFiles += files;
    break;

  case crashFlavour:
    traf.crashMail += mail;
    traf.crashXMail += xmail;
    traf.crashFiles += files;
    break;

  case directFlavour:
    traf.directMail += mail;
    traf.directXMail += xmail;
    traf.directFiles += files;
    break;

  case normalFlavour:
    traf.normalMail += mail;
    traf.normalXMail += xmail;
    traf.normalFiles += files;
    break;

  default:
    traf.holdMail += mail;
    traf.holdXMail += xmail;
    traf.holdFiles += files;
  }
}

void QueueEntry::rescanCfg( ScanType type )
{
  rescanPath( type, getImmedBox(),  immedFlavour );
  rescanPath( type, getCrashBox(),  crashFlavour );
  rescanPath( type, getDirectBox(), directFlavour );
  rescanPath( type, getNormalBox(), normalFlavour );
  rescanPath( type, getHoldBox(),   holdFlavour );
}

void QueueEntry::processTBox( ScanType type, char *name )
{
  if( miscInfo.tbox==NULL )
    return;

  char *real( (char *)_alloca( strlen(miscInfo.tbox)+strlen(name)+1 ) );

  strcpy( real, miscInfo.tbox );
  strcat( real, name );

  switch( toupper(name[11]) )
  {
  case 'I': rescanPath( type, real, immedFlavour );  break;
  case 'C': rescanPath( type, real, crashFlavour );  break;
  case 'D': rescanPath( type, real, directFlavour ); break;
  case 0:
  case 'N':
  case 'F': rescanPath( type, real, normalFlavour ); break;
  case 'H': rescanPath( type, real, holdFlavour );   break;
  }
}


void QueueEntry::rescanTBox( ScanType type )
{
  if( miscInfo.tbox==NULL )
    return;

  char *tBox( getTBoxName() );

  if( tBox==NULL )
    return;

  char *path = (char *)_alloca( strlen(miscInfo.tbox)+strlen(tBox)+2 );

  strcpy( path, miscInfo.tbox );
  strcat( path, tBox );
  strcat( path, "?" );

  for( Directory rc( path,
                    SearchAttributes( MustBeDirectory|MayBeReadOnly|MayBeArchived ) );
                rc; rc++ )
  {
    char *name = rc.getName();

    if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory &&
        (name[11]==0||name[12]==0) )
      processTBox( type, name );
  }

  delete tBox;
}

int isIdentical( char *prefName, char *pref, char *name, char *suf1, char *suf2 )
{
  char *full( (char *)_alloca( strlen(pref)+strlen(suf1)+
                               (suf2!=NULL ? strlen(suf2) : 0)+
                               3
                             )
            ),
       *real( (char *)_alloca( strlen(prefName)+strlen(name)+1 ) );

  strcpy( full, pref );
  strcat( full, "." );
  strcat( full, suf1 );

  if( suf2!=NULL )
  {
    strcat( full, "." );
    strcat( full, suf2 );
  }

  strcpy( real, prefName );
  strcat( real, name );

  return !stricmp( full, real );
}

void QueueEntry::processLongBinkName( ScanType type, 
                                      char *path, char *bink, char *name,
                                      long size, long created )
{
  static char mailStr[]( "Mail" ),
              listStr[]( "List" );

// mail

  if( isIdentical( path, bink, name, flavourString( normalFlavour ), mailStr ) )
    setMailTraf( normalFlavour, size, created );
  else if( isIdentical( path, bink, name, flavourString( directFlavour ), mailStr ) )
    setMailTraf( directFlavour, size, created );
  else if( isIdentical( path, bink, name, flavourString( crashFlavour ), mailStr ) )
    setMailTraf( crashFlavour, size, created );
  else if( isIdentical( path, bink, name, flavourString( holdFlavour ), mailStr ) )
    setMailTraf( holdFlavour, size, created );
  else if( isIdentical( path, bink, name, flavourString( immedFlavour ), mailStr ) )
    setMailTraf( immedFlavour, size, created );

// lists

  else if( isIdentical( path, bink, name, flavourString( immedFlavour ), listStr ) )
    processLO( type, path, name, immedFlavour, created );
  else if( isIdentical( path, bink, name, flavourString( crashFlavour ), listStr ) )
    processLO( type, path, name, crashFlavour, created );
  else if( isIdentical( path, bink, name, flavourString( directFlavour ), listStr ) )
    processLO( type, path, name, directFlavour, created );
  else if( isIdentical( path, bink, name, flavourString( normalFlavour ), listStr ) )
    processLO( type, path, name, normalFlavour, created );
  else if( isIdentical( path, bink, name, flavourString( holdFlavour ), listStr ) )
    processLO( type, path, name, holdFlavour, created );
    
// special 
    
  else if( isIdentical( path, bink, name, "Freeze", 0 ) )
    processHold( path, name );
  else if( isIdentical( path, bink, name, flavourString( requestFlavour ), 0 ) )
    traf.request += size;
}

void QueueEntry::rescanLongBink( ScanType type )
{
  char *bink( getLongBinkName() );

  if( bink==NULL )
    return;

  char mask[]( ".*" ),
       *path = (char *)_alloca( strlen(bink)+sizeof(mask) );

  strcpy( path, bink );
  strcat( path, mask );

  for( Directory rc( path,
                    SearchAttributes( MayBeReadOnly|MayBeArchived ) );
                rc; rc++ )
    processLongBinkName( type, rc.getPath(), bink, rc.getName(),
                         rc.getSize(), rc.getUNIXCreate() );

  delete bink;
}

void QueueEntry::rescanLongBox( ScanType type )
{
  char *longBox( getLongBox() );

  if( longBox==NULL )
    return;

  char mask[]( ".*" ),
       *path = (char *)_alloca( strlen(longBox)+sizeof(mask) ),
       *real = (char *)_alloca( strlen(longBox)+_MAX_PATH );

  strcpy( path, longBox );
  strcat( path, mask );
  strcpy( real, longBox );

  char *namePos( fileName( real ) );
  int len( strlen(namePos) );

  for( Directory rc( path,
                    SearchAttributes( MustBeDirectory|MayBeReadOnly|MayBeArchived ) );
                rc; rc++ )
  {
    char *name = rc.getName();

    strcpy( namePos, name );

    if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory && strlen(name)>=len )
      if(      isIdentical( miscInfo.longBox, longBox, name,
                            flavourString( normalFlavour ), 0 ) )
        rescanPath( type, real, normalFlavour );
      else if( isIdentical( miscInfo.longBox, longBox, name, 
                            flavourString( directFlavour ), 0 ) )
        rescanPath( type, real, directFlavour );
      else if( isIdentical( miscInfo.longBox, longBox, name, 
                            flavourString( immedFlavour ), 0 ) )
        rescanPath( type, real, immedFlavour );
      else if( isIdentical( miscInfo.longBox, longBox, name, 
                            flavourString( crashFlavour ), 0 ) )
        rescanPath( type, real, crashFlavour );
      else if( isIdentical( miscInfo.longBox, longBox, name, 
                            flavourString( holdFlavour ), 0 ) )
        rescanPath( type, real, holdFlavour );
  }

  delete longBox;
}

int QueueEntry::isFlavourCallable()
{
  time_t cur = time( NULL );
  Flavour fl = getQueuedFlavour();

  if( miscInfo.callOnlyImmed!=NULL &&
      miscInfo.callOnlyImmed->isWorked( cur ) &&
      fl!=immedFlavour )
    return 0;

  if( miscInfo.callOnlyCrash!=NULL &&
      miscInfo.callOnlyCrash->isWorked( cur ) &&
      fl!=crashFlavour )
    return 0;

  if( miscInfo.callOnlyDirect!=NULL &&
      miscInfo.callOnlyDirect->isWorked( cur ) &&
      fl!=directFlavour )
    return 0;

  if( miscInfo.callOnlyNormal!=NULL &&
      miscInfo.callOnlyNormal->isWorked( cur ) &&
      fl!=normalFlavour )
    return 0;

  return 1;
}

int QueueEntry::isCallable( int flag )
{
  char buf[65];
  
  if( flag )
    toString( buf, sizeof buf );
  
  if( getPhone()==NULL && getRexxDial()==NULL && getDo()==NULL )
  {
    if( flag )
      log( "FROZEN: no phone - %s", buf );
    return 0;
  }

  char node[65];
  int iii = processLogicalRexxHook( "CallableNode",
                                    toString( node, sizeof(node)-1,
                                              FIDOAddressShowModeEMSI ) );
  char callable = 1;
  time_t tt( time( NULL ) );

  if( iii==0 )
  {
    if( flag )
      log( "FROZEN: hook say NO - %s", buf );
    return 0;
  }

  if( iii==1 )
    goto checkBsy;

  if( !iMenable )
  {
    if( flag )
      log( "FROZEN: not enabled - %s", buf );
    return 0;
  }

  if( isOurs() )
  {
    if( flag )
      log( "FROZEN: is ours - %s", buf );
    return 0;
  }

  if( costFault )
  {
    if( flag )
      log( "FROZEN: cost fault - %s", buf );
    return 0;
  }

  if( !isFlavourCallable() )
  {
    if( flag )
      log( "FROZEN: no callable flavours - %s", buf );
    return 0;
  }

  loadTry();

  if( tryExceeded() )
  {
    if( flag )
      log( "FROZEN: try exceeded - %s", buf );
    return 0;
  }

  if( getNdl()!=NULL )
    callable = getNdl()->isCallable();

  switch( getQueuedFlavour() )
  {
  case immedFlavour:
    break;

  case crashFlavour:
  case directFlavour:
  case normalFlavour:
    if( !callable )
    {
      if( flag )
        log( "FROZEN: ndl is not callable - %s", buf );
      return 0;
    }

    if( !isWorked( tt ) )
    {
      if( flag )
        log( "FROZEN: not worked - %s", buf );
      return 0;
    }

    if( miscInfo.zmh!=NULL && miscInfo.zmh->isWorked( tt ) )
      if( traf.normalMail || traf.poll==normalFlavour ||
          traf.directMail || traf.poll==directFlavour ||
          traf.crashMail  || traf.poll==crashFlavour  ||
          traf.immedMail  || traf.poll==immedFlavour )
        break;
      else
      {
        if( flag )
          log( "FROZEN: ZMH - %s", buf );
        return 0;
      }
    else
      break;

  default:
    if( flag )
      log( "FROZEN: HOLD - %s", buf );
    return 0;
  }

checkBsy:
  if( checkBsy() )
  {
    if( flag )
      log( "FROZEN: busy - %s", buf );
    return 0;
  }

  return 1;
}

void QueueEntry::addTraffic( unsigned long *tr )
{
  traf.immedMail   += tr[0];
  traf.immedXMail  += tr[1];
  traf.immedFiles  += tr[2];
  traf.crashMail   += tr[3];
  traf.crashXMail  += tr[4];
  traf.crashFiles  += tr[5];
  traf.directMail  += tr[6];
  traf.directXMail += tr[7];
  traf.directFiles += tr[8];
  traf.normalMail  += tr[9];
  traf.normalXMail += tr[10];
  traf.normalFiles += tr[11];
  traf.holdMail    += tr[12];
  traf.holdXMail   += tr[13];
  traf.holdFiles   += tr[14];
  traf.request     += tr[15];
}

void QueueEntry::clear()
{
  memset( &traf, 0, sizeof(traf) );
  traf.poll = unknownFlavour;
}

int QueueEntry::isEmpty() const
{
  return holdSize()   || normalSize() || crashSize() ||
         immedSize()  || directSize() ||
         (traf.poll!=unknownFlavour) ? 0 : 1;
}

int QueueEntry::getPriority() const
{
  register int prio = NodeConfig::getPriority();

  if( prio )
    return prio;
   
  Flavour fl = getCfgFlavour();

  if( fl!=unknownFlavour )
    return flavourPriority( fl );

  return flavourPriority( getTrafficFlavour() );
}

Flavour QueueEntry::getQueuedFlavour()
{
  if( traf.holdTo )
    if( traf.holdTo==0xffffffffl || traf.holdTo > time( NULL ) )
      return holdFlavour;
    else
    {
      rescan();

      if( isMapped() )
        draw( isMapped(), queue->getCurrent() );
    }

  if( traf.holdTo )
    return holdFlavour;

  Flavour fl = getCfgFlavour();

  if( fl!=unknownFlavour )
    return fl;

  return getTrafficFlavour();
}

Flavour QueueEntry::getTrafficFlavour() const
{
  if( immedSize() || traf.poll==immedFlavour )
    return immedFlavour;

  if( crashSize() || traf.poll==crashFlavour )
    return crashFlavour;

  if( directSize() || traf.poll==directFlavour )
    return directFlavour;

  if( normalSize() || traf.poll==normalFlavour )
    return normalFlavour;

  if( holdSize() || traf.poll==holdFlavour )
    return holdFlavour;

  return unknownFlavour;
}

