#define INCL_DOS
#define INCL_DOSERRORS
#define INCL_WINSHELLDATA
#include <os2.h>

#include <stdlib.h>
#include <io.h>
#include <ctype.h>
#include <fstream.h>
#include <time.h>
#include <process.h>
#include <iostream.h>
#include <new.h>

#include "language.h"

#include "heap.h"
#include "dinamic.h"
#include "misc.h"
#include "queue.h"
#include "translat.h"
#include "32base.h"
#include "key.h"
#include "fidoaddr.h"
#include "session.h"
#include "rexxhook.h"
#include "direct.h"
#include "addinqu.h"

Queue::Queue( SemaphoreMode sema ) :
  current( -1 ),
  scrollCurrent( -1 ),
  flavour( unknownFlavour ),
  first( 1 ),
  wasCurrent( -2 ),
  wasScrollCurrent( -2 ),
  lastRead( 0 ),
  lastWas( NULL ),
  lastWasUsed( NULL ),
  lastIn( 0 ),
  priority( 0 ),
  Collection( 100, 50, sema, DeleteModeDelete )
{}

Queue::~Queue()
{
  delete lastWas;
  delete lastWasUsed;
  UnLock();
}

QueueEntry* Queue::Find( QueueEntry& item, LockMode lock ) const
{
  return (QueueEntry*)Collection::Find( item, lock );
}
    
QueueEntry& Queue::operator [] ( const int pos ) const
{
  return *(QueueEntry *)&(Collection::operator[]( pos ));
}

QueueEntry *Queue::getCurrent() const
{
  return current==-1 ? (QueueEntry *)NULL : (QueueEntry *)&(*this)[current];
}

ostream& operator << ( ostream& s, const Queue& c )
{
   s
    << "  *Shared:     " << c.sharedQueueHeader << " ["
                         << *(long *)(((char *)c.sharedQueueHeader)-4)
                         << char(']') << endl
    << "  *LastScaned: " << c.sharedQueueHeader->lastScaned << endl
    << "  *Count:      " << c.sharedQueueHeader->count << endl
    << "  *InScan:     " << (int)c.sharedQueueHeader->inScan << endl
    << "  *NeedScan:   " << (int)c.sharedQueueHeader->needScan << endl
    << "  Current:     " << c.current << endl
    << "  ScrCurrent:  " << c.scrollCurrent << endl
    << "  First:       " << (int)c.first << endl
    << "  Flavour:     " << (int)c.flavour << endl
    << "  Priority:    " << (int)c.priority << endl
    << "  LastRead:    " << c.lastRead << endl;

  return s << *(const Collection *)&c;
}

void Queue::add( QueueEntry& cur )
{
  Flavour fl = cur.getQueuedFlavour();

  if( fl>flavour )
    flavour = fl;

  int prio = cur.getPriority();

  if( prio>priority )
    priority = prio;

  AddAtTail( cur );
}

void Queue::rescanAllType()
{
  rescanCfg();
  rescanBink();
  rescanLongBink();
  rescanTBox();
  rescanLongBox();
  addInQueueList->scan( this );
}

void Queue::rescanQueue()
{
  heapState( "qscnb" );
  long start = getTimeMark();
  
  Queue *newQueue = new Queue( SemaphoreModeNoNeed );

  if( getRexxHook( "BeforeScan" )!=NULL )
    processRexxHook( "BeforeScan" );

  newQueue->rescanAllType();

  for( int done(0); done==0; )
  {
    done = 1;

    for( QueueIterator q( *newQueue ); q; q++ )
      if( q().isEmpty() )
      {
        done=0;
        newQueue->Remove( q() );
        break;
      }
  }

  for( QueueIterator q( *newQueue ); q; q++ )
    q().postRescan();

  newQueue->sort();

  if( getRexxHook( "AfterScan" )!=NULL )
    processRexxHook( "AfterScan" );

  newQueue->write();
  sharedQueueHeader->inScan = 0;

  start = getTimeMark()-start;
  log( Msg_InfoRescaned, newQueue->GetItemsInContainer(),
       start/60, start%60 );

  delete newQueue;
  heapState( "qscne" );
}

void Queue::rescanBink()
{
  static char mask[]( ".*" );

  for( DomainIterator list( *outboundInfo.domains ); list; list++ )
  {
    Domain *dom = &list();
    char *outbound( dom->getOutbound() );
    int outLen( strlen(outbound) );
    char *path( new char[ outLen+sizeof(mask) ] );

    strcpy( path, outbound );
    strcpy( path+outLen-1, mask );

    for( Directory rc( path,
                       SearchAttributes( MustBeDirectory|MayBeReadOnly|MayBeArchived ) );
                   rc; rc++ )
    {
      char *name = rc.getName();

      if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory )
        if( strcmp( name, "." ) && strcmp( name, ".." ) )
	  rescanBinkZone( outLen, name, dom );
    }

    delete outbound;
    delete path;
  }
}

QueueEntry *Queue::createEntry( uShort zone, uShort net, uShort node, uShort point,
                                char *domain, Domain *dom )
{
  FIDOAddress *addr( new FIDOAddress( zone, net, node, point, domain, dom ) );
  NodeConfig  *addNode( new NodeConfig( addr ) ),
              *inList( systemList->Find( *addNode ) );

  if( inList!=NULL )
  {
    delete addNode;
    addNode = new NodeConfig( inList );
  }

  QueueEntry *q( new QueueEntry( *addNode ) );

  delete addNode;
  delete addr;

  return checkQEntry( q );
}


QueueEntry* Queue::checkQEntry( QueueEntry *q )
{
  q->preRescan();

  return q;
}

void Queue::addIfNeedOrDelete( QueueEntry *q )
{
  if( q==NULL )
    return;

  QueueEntry *old( Find( *q ) );

  if( old==NULL )
    add( *q );
  else
  {
    old->addTraffic( *q->getTrafficPointer() );
    delete q;
  }
}

void Queue::updateEntry( QueueEntry& q )
{
  QueueEntry *old( Find( q ) );

  if( old==NULL )
    log( "*BUG* Queue::updateEntry() no entry exist" );
  else
    old->rescan();
}

void Queue::rescanBinkEntry( uShort zone, uShort net, uShort node, uShort point,
                             char *domain, Domain *dom, char *path, char *name,
                             long size, long create )
{
  QueueEntry *q( createEntry( zone, net, node, point, domain, dom ) );

  if( q==NULL )
    return;

  q->processShortBink( ScanTypeWithoutFiles, name, path, size, create );
  addIfNeedOrDelete( q );
}

void Queue::rescanBinkZone( int outLen, char *zoneStr, Domain *dom )
{
  uShort zone( 0 );
  char ext[_MAX_EXT];
  uShort realZone( 0 );

  {
    char drive[_MAX_DRIVE],
	 dir[_MAX_DIR],
	 file[_MAX_FNAME];

    _splitpath( zoneStr, drive, dir, file, ext );

    if( *ext )
    {
      if( val16base( ext+1, zone ) )
	return;

      realZone = zone;
    }
    else
      zone = dom->getDefZone();
  }

  if( (miscInfo.state&Full5D)!=Full5D )
  {
    char inList(0);

    if( !realZone && dom->getDefZone() )
      inList = 1;

    for( int i(0); i<dom->getZLSize() && !inList; i++ )
      if( zone == dom->getZList()[i] )
        inList = 1;

    if( !inList )
      return;
  }

  char *mask( "\\*" ),
    *path = (char *)_alloca( outLen+strlen(ext)+strlen( mask ) ),
    *full = (char *)_alloca( outLen+strlen(ext)+20 ),
    *outbound( dom->getOutbound() );

  strcpy( path, outbound );
  strcpy( path+outLen-1, ext );
  strcat( path, mask );

  strcpy( full, outbound );
  strcpy( full+outLen-1, ext );
  strcat( full, "\\" );

  uShort net, node;

  for( Directory rc( path,
                     SearchAttributes( MayBeDirectory|MayBeReadOnly|MayBeArchived ) );
                 rc; rc++ )
  {
    char *name = rc.getName();

    if( strlen( name )==12 && split16base( name, net, node ) )
      if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory &&
          !stricmp( name+8, ".pnt" ) )
	rescanBinkPoint( zone, net, node, full, name, dom );
      else
	if( isBundle( name ) || (strlen(name)==12 && !stricmp( name+8, ".hld" )) )
          rescanBinkEntry( zone, net, node, 0, NULL, dom, rc.getPath(), name,
                           rc.getSize(), rc.getUNIXCreate() );
  }

  delete outbound;
}

void Queue::rescanBinkPoint( uShort zone, uShort net, uShort node,
			     char *fullPath, char *nodePath,
                             Domain *dom )
{
  char *mask( "\\*" ),
       *path = (char *)_alloca( strlen(fullPath)+strlen(nodePath)+
                                strlen( mask )+1 );

  strcpy( path, fullPath );
  strcat( path, nodePath );
  strcat( path, mask );

  for( Directory rc( path ); rc; rc++ )
  {
    uShort point, fict;
    char *name = rc.getName();

    if( strlen( name )==12 && split16base( name, fict, point ) )
      if( isBundle( name ) || !stricmp( name+8, ".hld" ) )
        rescanBinkEntry( zone, net, node, point, NULL, dom, rc.getPath(), name,
                         rc.getSize(), rc.getUNIXCreate() );
  }
}

void Queue::rescanLongBink()
{
  if( miscInfo.longOutbound==NULL )
    return;

  static char mask[]( "*" );
  char *path = (char *)_alloca( strlen(miscInfo.longOutbound)+sizeof(mask) );

  strcpy( path, miscInfo.longOutbound );
  strcat( path, mask );

  for( Directory rc( path, SearchAttributes( MayBeReadOnly|MayBeArchived ) );
      rc; rc++ )
  {
    char *name( newStr( rc.getName() ) );
    uShort zone, net, node, point;
    char *domain;
    LongNameType type;
    Flavour fl;

    if( splitLongName( name, zone, net, node, point, &domain, fl, type ) &&
        (type==mailLongName || type==listLongName ||
         type==holdLongName || type==requestLongName) )
    {
      QueueEntry *q( createEntry( zone, net, node, point, domain, NULL ) );

      if( q!=NULL )
      {
        char *bink( q->getLongBinkName() );

        if( bink!=NULL )
        {
          q->processLongBinkName( ScanTypeWithoutFiles, rc.getPath(),
                                  bink,
                                  rc.getName(),
                                  rc.getSize(), rc.getUNIXCreate() );
          addIfNeedOrDelete( q );
          delete bink;
        }
        else
          delete q;
      }
    }

    delete name;
  }
}

void Queue::rescanTBox()
{
  if( miscInfo.tbox==NULL )
    return;

  static char mask[]( "*" );
  char *path = (char *)_alloca( strlen(miscInfo.tbox)+sizeof(mask) );

  strcpy( path, miscInfo.tbox );
  strcat( path, mask );

  uShort zone, net, node, point;

  for( Directory rc( path,
                     SearchAttributes( MustBeDirectory|MayBeReadOnly|MayBeArchived ) );
                 rc; rc++ )
    if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory )
    {
      char *name = rc.getName();

      if( split32base( name, zone, net, node, point ) )
      {
        QueueEntry *q( createEntry( zone, net, node, point, NULL, NULL ) );

        if( q!=NULL )
        {
          q->processTBox( ScanTypeWithoutFiles, name );
          addIfNeedOrDelete( q );
        }
      }
    }
}

void Queue::rescanLongBox()
{
  if( miscInfo.longBox==NULL )
    return;

  static char mask[]( "*" );
  char *path = (char *)_alloca( strlen(miscInfo.longBox)+sizeof(mask) );

  strcpy( path, miscInfo.longBox );
  strcat( path, mask );

  for( Directory rc( path,
                     SearchAttributes( MustBeDirectory|MayBeReadOnly|MayBeArchived ) );
                 rc; rc++ )
    if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory )
    {
      char *name( newStr( rc.getName() ) );
      uShort zone, net, node, point;
      char *domain;
      LongNameType type;
      Flavour fl;

      if( splitLongName( name, zone, net, node, point, &domain, fl, type ) &&
          type==boxLongName )
      {
        QueueEntry *q( createEntry( zone, net, node, point, domain, NULL ) );

        if( q!=NULL )
        {
          char *path( rc.getFSPoint() );

          q->rescanPath( ScanTypeWithoutFiles, path, fl );
          addIfNeedOrDelete( q );
          delete path;
        }
      }

      delete name;
    }
}

void Queue::rescanCfg()
{
  for( NodeConfigIterator list( *systemList ); list; list++ )
  {
    NodeConfig& cur( list() );
    QueueEntry *q( checkQEntry( new QueueEntry( cur ) ) );

    if( q!=NULL )
    {
      q->rescanCfg( ScanTypeWithoutFiles );
      addIfNeedOrDelete( q );
    }
  }
}

void Queue::view()
{
  if( current==-1 )
  {
    execSchedule( melodies.alarm );
    return;
  }

  const width = Z_ScreenWidth,
        height = Z_ScreenHeight-2;

  PICKLIST pick =
  {
    (Z_ScreenWidth-width)/2,
    (Z_ScreenHeight-height)/2,
    height,
    width,
    width-2,
    PLF_VERTSCROLL,
    popUpColor.fore, popUpColor.back,
    stdColor.fore, stdColor.back,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    ((((popUpColor.back<<4)|popUpColor.fore)<<8)|0x20)&0xffff,
    DS_NONE, DARKGREY,
    Msg_ScrQueView
  };

  for( PICKLIST_ITEM *item=(PICKLIST_ITEM *)queue; item!=NULL; )
  {
    long sXMail(0), sMail(0), sFiles(0);

    for( QueueIterator list( *queue ); list; list++ )
    {
      QueueEntry& cur = list();

      char *tryLast = cur.getTryLast();
      char *buf = new char [ 128 + 
                             (tryLast==NULL ? 80 : strlen( tryLast )) ];

      cur.format( buf, 1 );
      
      if( strlen(buf)>width-2 )
        buf[width-2]=0;
        
      Z_AddPickListItem( &pick, buf );

      sMail += cur.getHoldMail() + cur.getNormalMail() + cur.getCrashMail() +
               cur.getDirectMail() + cur.getImmedMail();

      sXMail += cur.getHoldXMail() + cur.getNormalXMail() + cur.getCrashXMail() +
                cur.getDirectXMail() + cur.getImmedXMail();

      sFiles += cur.getHoldFiles() + cur.getNormalFiles() + cur.getCrashFiles() +
                cur.getDirectFiles() + cur.getImmedFiles() +
                cur.getRequest();

      delete buf;
    }

    char buf[128];

    sprintf( buf, Msg_ScrViewTotal, GetItemsInContainer(),
                  sMail, sXMail, sFiles );

    char buf1[ sizeof(buf) ];

    for( int i=width-2-strlen(buf); i>0; i-=2 )
    {
      strcpy( buf1, " " );
      strcat( buf1, buf );
      strcpy( buf, buf1 );
    }

    Z_AddPickListItem( &pick, buf );

    item = Z_HandlePickList( &pick );

    if( item )
    {
      char *text = item->ItemText,
           *p = strchr( text, ' ' );

      if( p!=NULL )
        *p = 0;

      Lock();

      NodeConfig node( text );
      QueueEntry test( node ),
                 *found( Find( test, LockModeNoNeed ) ),
                 *ready( NULL );

      if( found!=NULL )
        ready = new QueueEntry( *found );

      UnLock();

      if( node.isValid() && ready!=NULL )
        ready->show();

      delete ready;
    }

    Z_PurgePickList( &pick );
  }

  Z_DestroyPickList( &pick );
}

void Queue::checkFlavour()
{
  flavour = unknownFlavour;
  priority = 0;

  for( QueueIterator q( *this ); q; q++ )
  {
    QueueEntry &cur = q();
    Flavour fl = cur.getQueuedFlavour();
    int prio = cur.getPriority();

    if( fl>flavour && cur.isFlavourCallable() && cur.isCallable() )
      flavour = fl;

    if( prio>priority && cur.isFlavourCallable() && cur.isCallable() )
      priority = prio;
  }

  if( flavour==unknownFlavour )
    flavour = holdFlavour;

  if( !priority )
    priority = flavourPriority( holdFlavour );
}

void Queue::nextEntry()
{
  if( current==-1 )
    return;
  
  Lock();

  if( current+1<GetItemsInContainer() )
    current++;
  else
    current = 0;

  adjust();

  UnLock();
}

void Queue::adjust()
{
  scrollCurrent = current;

  if( current==-1 )
    return;

  while( scrollCurrent+wndQueue.getHeight()-1 > GetItemsInContainer() && scrollCurrent )
    scrollCurrent--;
}

static char *lastWasKey     = "Last Node",
            *lastInKey      = "Last Mode",
            *lastWasUsedKey = "Last Used";

extern ULONG iniFileHandle;

void Queue::setLast( char *w, char *w1, char i )
{
  delete lastWas;
  lastWas=newStr(w);
  delete lastWasUsed;
  lastWasUsed=newStr(w1);
  lastIn=i;

  redrawLastConnect();

  if( !noPMFeatures && !detachedSession )
  {
    BrakePrfWriteProfileString( iniFileHandle, pszAppBrake,
			        (PSZ)lastWasKey,
			        (PSZ)lastWas );

    BrakePrfWriteProfileString( iniFileHandle, pszAppBrake,
			        (PSZ)lastWasUsedKey,
			        (PSZ)lastWasUsed );

    BrakePrfWriteProfileData( iniFileHandle, pszAppBrake,
		              (PSZ)lastInKey,
		              (PVOID)&lastIn,
		              (ULONG)sizeof( lastIn ) );
  }
}

void Queue::getLast()
{
  const LastWasLen = 256;

  delete lastWas;
  lastWas = new char[ LastWasLen ];

  delete lastWasUsed;
  lastWasUsed = new char[ LastWasLen ];

  ULONG len( sizeof lastIn );
  BOOL rc( FALSE );

  if( !noPMFeatures && !detachedSession )
  {
    ULONG rc1;

    rc = BrakePrfQueryProfileData( iniFileHandle, pszAppBrake,
                                   (PSZ)lastInKey,
                                   (PVOID)&lastIn,
                                   &len );

    if( rc!=FALSE )
      rc1 = BrakePrfQueryProfileString( iniFileHandle, pszAppBrake,
                                        (PSZ)lastWasKey,
                                        NULL,
                                        (PVOID)lastWas,
                                        (ULONG)LastWasLen );
    else
      rc1 = 0;

    if( rc1 )
      rc1 = BrakePrfQueryProfileString( iniFileHandle, pszAppBrake,
                                        (PSZ)lastWasUsedKey,
                                        NULL,
                                        (PVOID)lastWasUsed,
                                        (ULONG)LastWasLen );
    else
      rc1 = 0;

    if( !rc1 )
      rc = FALSE;
  }

  if( rc==FALSE )
  {
    lastIn = 0;
    delete lastWas;
    lastWas = NULL;
    delete lastWasUsed;
    lastWasUsed = NULL;
  }
  else
  {
    register char *c = lastWas;

    lastWas = newStr( lastWas );
    delete c;

    c = lastWasUsed;
    lastWasUsed = newStr( lastWasUsed );
    delete c;

    redrawLastConnect();
  }
}

void Queue::redrawLastConnect()
{
  if( lastWas==NULL )
    return;

  char *p = lastIn ? Msg_ScrWasAnswer : Msg_ScrWasCall;
  int lenp = strlen( p );
  Color bold( colorSet.actionBold );

  wndStat.drawTransact( 2, 4, stdColor, p );

  int maxLastWidth = wndStat.getWidth()-lenp-1;

  if( strlen( lastWas )>maxLastWidth )
    lastWas[ maxLastWidth ] = 0;

  wndStat.drawTransact( 2+lenp, 4, bold, lastWas );

  wndStat.drawTransact( 2, 5, stdColor, Msg_ScrWasTime );
  lenp = strlen( Msg_ScrWasTime );

  maxLastWidth = wndStat.getWidth()-lenp-1;

  if( strlen( lastWasUsed )>maxLastWidth )
    lastWasUsed[ maxLastWidth ] = 0;

  wndStat.drawTransact( 2+lenp, 5, bold, lastWasUsed );
}

void Queue::read()
{
  if( lastRead==sharedQueueHeader->lastScaned || sharedQueueHeader->inScan )
    return;

  heapState( "qreab" );
  showAction( Msg_ScrReadQueue );

// get old current

  FIDOAddress *oldCurrentItem( NULL );

  Lock();

  if( current!=-1 )
    oldCurrentItem = new FIDOAddress( getCurrent()->getZone(),
                                      getCurrent()->getNet(),
                                      getCurrent()->getNode(),
                                      getCurrent()->getPoint() );

  UnLock();

// flush queue

  Flush();

// read queue

  lockMutexSemaphore( globalSemaphore );

  flavour = unknownFlavour;
  priority = 0;
  shMem->seek( sizeof( SharedQueueHeader ) );

  for( int i(0); i<sharedQueueHeader->count; i++ )
  {
    SharedQueueEntry entry;

    if( shMem->read( &entry, sizeof(entry) ) )
    {
QueueError:
      log( "~BUG~ Queue::read unable read queue" );
      break;
    }

    short len = 0;

    if( shMem->read( &len, sizeof(len) ) )
      goto QueueError;

    if( len )
    {
      entry.addr.domain = new char [len+1];

      if( shMem->read( entry.addr.domain, len ) )
      {
        delete entry.addr.domain;
        goto QueueError;
      }

      entry.addr.domain[len]=0;
    }
    else
      entry.addr.domain = NULL;

    FIDOAddress addr( entry.addr );
    NodeConfig *addObj( new NodeConfig( &addr ) ),
               *inList( systemList->Find( *addObj ) );

    if( inList!=NULL )
    {
      delete addObj;
      addObj = new NodeConfig( inList );
    }

    QueueEntry *addQEntry( new QueueEntry( *addObj ) );

    memcpy( addQEntry->getTrafficPointer(), &entry.traf, sizeof( entry.traf ) );

    add( *addQEntry );

    delete addObj;
    delete entry.addr.domain;
  }

  lastRead=sharedQueueHeader->lastScaned;
  unlockMutexSemaphore( globalSemaphore );

// set old current

  Lock();

  current = -1;

  if( GetItemsInContainer() && oldCurrentItem != NULL )
  {
    i = 0;

    for( QueueIterator list( *this ); list && current==-1; list++, i++ )
    {
      QueueEntry& posQue( list() );
      FIDOAddress pos( posQue.getZone(),
                       posQue.getNet(),
                       posQue.getNode(),
                       posQue.getPoint() );

      if( oldCurrentItem->equal( pos ) || 
          oldCurrentItem->less( pos ) )
	current = i;
    }
  }

  if( current==-1 && GetItemsInContainer() )
    current = 0;

  adjust();

  wasCurrent = -2;
  wasScrollCurrent = -2;

  UnLock();

  delete oldCurrentItem;

  checkFlavour();
  draw();
  heapState( "qreae" );
}

void Queue::Remove( QueueEntry& item, DeleteMode del, LockMode lock )
{
  if( lock==LockModeNeed )
    Lock();

  Collection::Remove( item, del, LockModeNoNeed );

  if( !GetItemsInContainer() )
    current = -1;
  else
    if( current>GetItemsInContainer() )
      current = GetItemsInContainer();

  adjust();

  if( lock==LockModeNeed )
    UnLock();
}

void Queue::write()
{
  lockMutexSemaphore( globalSemaphore );
  Lock();

  sharedQueueHeader->lastScaned = getTimeMark();
  sharedQueueHeader->count = 0;

  shMem->seek( sizeof( SharedQueueHeader ) );

  for( QueueIterator list( *this ); list; list++ )
  {
    QueueEntry& cur( list() );
    SharedQueueEntry entry;

    memcpy( &entry.addr, cur.getAddressPointer(), sizeof( entry.addr ) );
    memcpy( &entry.traf, cur.getTrafficPointer(), sizeof( entry.traf ) );

    if( shMem->write( &entry, sizeof(entry) ) )
    {
QueueFull:
      log( Msg_InfoToManyQueueEntry );
      break;
    }

    short len = 0;

    if( entry.addr.domain!=NULL )
      len = (short)strlen( entry.addr.domain );

    if( shMem->write( &len, sizeof(len) ) )
      goto QueueFull;

    if( len )
      if( shMem->write( entry.addr.domain, len ) )
        goto QueueFull;

    ++sharedQueueHeader->count;
  }

  UnLock();
  unlockMutexSemaphore( globalSemaphore );
}

int Queue::initStatic( char *sharedQueueName )
{
  char *mutexTplName = "\\Sem32\\Brake!\\Task\\Shared.Queue.Access.Flag\\",
       *mutexName = (char *)_alloca( strlen(sharedQueueName)+strlen(mutexTplName)+1 );

  strcpy( mutexName, mutexTplName );
  strcat( mutexName, sharedQueueName );

  if( createNamedMutex( mutexName, &globalSemaphore ) )
    return 1;

  lockMutexSemaphore( globalSemaphore );

  char *shMemTplName = "\\ShareMem\\Brake!\\Shared.Queue\\",
       *shMemName = (char *)_alloca( strlen(sharedQueueName)+strlen(shMemTplName)+1 );

  strcpy( shMemName, shMemTplName );
  strcat( shMemName, sharedQueueName );

  shMem = new SharedMemory( shMemName, sharedMemoryQueueSize*1024 );

  unlockMutexSemaphore( globalSemaphore );

  sharedQueueHeader = (SharedQueueHeader *)shMem->getBase();

  if( shMem->isOk() && shMem->isAllocated() )
      memset( sharedQueueHeader, 0, sizeof(SharedQueueHeader) );

  return shMem->isOk() ? 0 : 1;
}

void Queue::queueName( char *msg, char *disabled )
{
  if( detachedSession )
    return;

  static char *oldMsg = "",
              *oldFl  = newStr( "" );
  static int oldLen   = 0;
  static char *oldDisabled = NULL;

  char *fl = msg==Msg_WndTitleQue ? flavourNameByPriority( priority ) : "";

  if( !stricmp( oldMsg, msg ) && !stricmp( oldFl, fl ) && oldDisabled==disabled )
    return;

  oldMsg = msg;
  delete oldFl;
  oldFl  = newStr( fl );
  oldDisabled = disabled;

  char m[]    = ": ",
       dele[] = ", ";
  int len = strlen(msg)+strlen(fl)+sizeof(m)+3+
            ( disabled==NULL ? 0 : strlen(disabled)+sizeof(dele) ),
      alloc = (len<oldLen ? oldLen : len)+4;

  char *who = (char *)_alloca( alloc );

  strcpy( who, " " );
  strcat( who, msg );

  if( *fl )
  {
    strcat( who, m );
    strcat( who, fl );
  }

  capitalize( who );

  if( disabled!=NULL )
  {
    strcat( who, dele );
    strcat( who, disabled );
  }

  strcat( who, " " );

  int delta = (miscInfo.state1&StatusLine)==StatusLine ? 0 : 1;

  masterScreen.drawTransact( 4, 12-delta, brdColor, who, 0 );

  int len1 = strlen(who);

  if( len1<oldLen)
  {
    for( int i=len1, j=0; i<oldLen; i++, j++ )
      who[j] = '';

    who[j]=0;
    masterScreen.drawTransact( 4+len1, 12-delta, stdColor, who, 0 );
  }

  oldLen = len;
}

void Queue::needForRescan()
{
  lockMutexSemaphore( globalSemaphore );

  if( !Queue::sharedQueueHeader->inScan )
    sharedQueueHeader->needScan = 1;

  unlockMutexSemaphore( globalSemaphore );
}

void Queue::sort()
{
  if( miscInfo.sortMode==QueueSortNoSort )
    return;

  int i;

  Lock();

  if( miscInfo.sortMode!=QueueSortAddr &&
      miscInfo.sortMode!=QueueSortAddrDec )
    for( i=1; i<used; i++ )
    {
      QueueEntry *x = (QueueEntry *)list[i];
      int j = i-1;

      while( j!=-1 && ((FIDOAddress *)x)->FIDOAddress::less( *(FIDOAddress *)(list[j]) ) )
      {
        list[j+1] = list[j];
        --j;
      }

      list[j+1] = x;
    }

  for( i=1; i<used; i++ )
  {
    QueueEntry *x = (QueueEntry *)list[i];
    int j = i-1;

    while( j!=-1 && x->less( *(QueueEntry *)list[j] ) )
    {
      list[j+1] = list[j];
      --j;
    }

    list[j+1] = x;
  }

  UnLock();
}

void Queue::draw()
{
  queueName( Msg_WndTitleQue, answerDisabled );

  if( first )
  {
    wndQueue = stdColor;
    wndQueue( 1, 1 ) << Msg_ScrQueTitle;
    first=0;
  }

  if( current!=wasCurrent || scrollCurrent!=wasScrollCurrent )
  {
    int height = wndQueue.getHeight();

    if( current!=-1 )
    {
      int line( 2 );
      QueueEntry *cur = getCurrent();
      int items = GetItemsInContainer();

      for( long pos=scrollCurrent; line<=height && pos<items; pos++, line++ )
        (*this)[ int(pos) ].draw( line, cur );

      while( line<=height )
      {
        wndQueue( 1, line ).clearLine();
        line++;
      }

      char m[3];

      m[0] = char( scrollCurrent!=0 ? 30 : ' ' );
      m[1] = char( pos<items        ? 31 : ' ' );
      m[2] = 0;

      wndQueue = Color( colorSet.queuePtr );
      wndQueue( 1, 1 ) << m;
    }
    else
    {
      wndQueue = Color( colorSet.queuePtr );
      wndQueue( 1, 1 ) << "  ";
      wndQueue = stdColor;

      for( int i=2; i<=height; i++ )
        wndQueue( 1, i ).clearLine();

      wndQueue.centre( height/2+1, Msg_ScrQueEmpty );
    }

    wasCurrent = current;
    wasScrollCurrent = scrollCurrent;
  }
}

void Queue::scroll( int queueDelta )
{
  if( current==-1 )
    return;

  int wndHeight = wndQueue.getHeight()-1,
      totalEntry = GetItemsInContainer();

  if( current==0 && queueDelta<0 )
    return;

  if( current==totalEntry-1 && queueDelta>0 )
    return;

  switch( queueDelta )
  {
  case -2:
    queueDelta = -wndHeight+1;
    break;

  case 2:
    queueDelta = wndHeight-1;
    break;
  }

  current += queueDelta;

  if( current >= totalEntry )
    current = totalEntry-1;
  else if( current<0 )
    current = 0;

  int so = scrollCurrent;

  if( current>=scrollCurrent+wndHeight )
    while( current>=scrollCurrent+wndHeight )
      scrollCurrent += wndHeight-1;
  else if( current<scrollCurrent )
    while( current<scrollCurrent )
      scrollCurrent -= wndHeight-1;

  if( scrollCurrent<0 )
    scrollCurrent = 0;

  if( so!=scrollCurrent )
    while( scrollCurrent+wndHeight > totalEntry && scrollCurrent )
      scrollCurrent--;

  draw();
}

QueueIterator::QueueIterator( Queue& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}

QueueEntry& QueueIterator::operator () () const
{
  return *(QueueEntry *)&(CollectionIterator::operator()());
}

QueueEntryList::QueueEntryList() :
  Collection( 10, 5, SemaphoreModeNoNeed )
{
}

QueueEntry* QueueEntryList::Find( QueueEntry& item, LockMode lock ) const
{
  return (QueueEntry*)Collection::Find( item, lock );
}
    
QueueEntry& QueueEntryList::operator [] ( const int pos ) const
{
  return *(QueueEntry *)&(Collection::operator[]( pos ));
}

QueueEntryIterator::QueueEntryIterator( QueueEntryList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}
  
QueueEntry& QueueEntryIterator::operator () () const
{
  return *(QueueEntry *)&(CollectionIterator::operator()());
}
