#define INCL_DOSPROCESS
#define INCL_DOSEXCEPTIONS
#define INCL_DOSMEMMGR
#define INCL_DOS
#include <os2.h>

#include <process.h>
#include <sys\stat.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <share.h>
#include <fcntl.h>
#include <io.h>
#include <time.h>

#include "except.h"

#include "language.h"

#include "misc.h"
#include "execute.h"
#include "sched.h"
#include "config.h"
#include "rexx.h"
#include "keycd.h"
#include "timer.h"
#include "session.h"
#include "pair.h"
#include "fidoaddr.h"
#include "queue.h"
#include "direct.h"
#include "zmodem.h"
#include "ea.h"
#include "freq.h"

static const long SchedBinaryType = 0x1502aa55;
struct SchedDate schedDate;

int need( char *who );

// ------------------------  Schedule manipulator

void Schedule::operator = ( Schedule& src )
{
  memcpy( this, &src, sizeof(Schedule) );
}

void execSchedule( Schedule *sh )
{
  if( sh!=NULL )
    doEventKey( *sh, -1 );
}

void stringSchAddrFlav( Schedule *sh, Buffer *ret )
{
  char buf[ 128 ];
  
  ret->putString( realFlavour( sh->flavour ) );
  ret->putInto( ' ' );
  ret->putString( sh->adr->toString( buf, sizeof buf, FIDOAddressShowModeEMSI ) );
}

char *stringSch( Schedule *sh, char *retBuf, int len, int level )
{
  if( !level )
    *retBuf = 0;
    
  if( sh==NULL )
  {
    if( !level )
      strncat( retBuf, "*INTERNAL*", len-1 );

    return retBuf;
  }
    
  Buffer *ret( new Buffer );
  char buf[128];
  
  if( level )
    ret->putString( " And " );
  
  if( int(sh->action)&SchedActionThread )
    ret->putString( "Thread " );
    
  SchedAction action( SchedAction( int(sh->action)&~SchedActionThread ) );
    
  switch( action )
  {
  case SchedActionWait:
    ret->putString( "Wait " );
    sprintf( buf, "%d", sh->rc );
    ret->putString( buf );
    break;

  case SchedActionOpen:
    ret->putString( "Open \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;
    
  case SchedActionCall:
    ret->putString( "Call \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionPlay:
    ret->putString( "Play \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;
    
  case SchedActionIndexFREQ:
    ret->putString( "IndexFREQ" );
    break;
    
  case SchedActionNothing:
    ret->putString( "Nothing" );
    break;
    
  case SchedActionInit:
    ret->putString( "InitModem" );
    break;

  case SchedActionExit:
    ret->putString( "Exit " );
    sprintf( buf, "%d", sh->rc );
    ret->putString( buf );
    break;

  case SchedActionCreate:
    ret->putString( "Create \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionRemove:
    ret->putString( "Remove \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionHangUp:
    ret->putString( "HangUp" );
    break;

  case SchedActionLog:
    ret->putString( "Log \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionReScan:
    ret->putString( "ReScan" );
    break;

  case SchedActionSplit:
    ret->putString( "Split" );
    break;

  case SchedActionReset:
    ret->putString( "Reset" );
    break;

  case SchedActionBSY:
    ret->putString( "CheckBSY" );
    break;

  case SchedActionSemCreate:
    ret->putString( "Semaphore Create \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionSemRemove:
    ret->putString( "Semaphore Remove \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionSemPost:
    ret->putString( "Semaphore Post \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionSemReset:
    ret->putString( "Semaphore Reset \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionSemWait:
    ret->putString( "Semaphore Wait \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionCompile:
    ret->putString( "Compile" );
    break;

  case SchedActionSync:
    ret->putString( "Sync" );
    break;

  case SchedActionShutdown:
    ret->putString( "Shutdown" );
    break;

  case SchedActionReboot:    
    ret->putString( "Reboot" );
    break;
    
  case SchedActionCopy:
    ret->putString( "Copy " );
    ret->putString( sh->adr->toString( buf, sizeof buf, FIDOAddressShowModeEMSI ) );
    ret->putString( " \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;
    
  case SchedActionCatch:
    ret->putString( "Catch \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;
    
  case SchedActionSound:
    {
      ret->putString( "Sound " );
      
      for( int i=1; i<sh->sndCount; i++ )
      {
        sprintf( buf, "%d ", sh->sound[i] );
        ret->putString( buf );
      }
      
      ret->stripByte();
    }
    break;

  case SchedActionPoll:
  case SchedActionDrop:
    ret->putString( action==SchedActionPoll ? "Poll " : "Drop " );
    stringSchAddrFlav( sh, ret );
    break;

  case SchedActionSend:
  case SchedActionGet:
    ret->putString( action==SchedActionSend ? "Send " : "Get " );
    stringSchAddrFlav( sh, ret );
    ret->putString( " \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionRexx:
    ret->putString( "Rexx " );
    
    if( sh->preloaded )
      ret->putString( "/PreLoad " );
    
    ret->putInto( '\"' );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionEval:
    ret->putString( "Eval \"" );
    ret->putString( sh->params );
    ret->putInto( '\"' );
    break;

  case SchedActionExec:
    ret->putString( "Exec " );

    if( sh->rOpt&RunOptClose )
      ret->putString( "/Close " );
    
    if( sh->rOpt&RunOptClear )
      ret->putString( "/Clear " );
    
    if( sh->rOpt&RunOptNoClip )
      ret->putString( "/NoClip " );
    
    if( sh->rOpt&RunOptPort )
      ret->putString( "/Port " );
      
    if( sh->dosSetting!=NULL )
    {
      ret->putString( "/Env:" );
      ret->putString( sh->dosSetting );
      ret->putInto( ' ' );
    }    

    ret->putString( sh->params );
    break;
    
  case SchedActionRun:
    ret->putString( "Run " );

    if( sh->rOpt&RunOptChild )
      ret->putString( "/Child " );
    
    if( sh->rOpt&RunOptWait )
      ret->putString( "/Wait " );
    
    if( sh->rOpt&RunOptFore )
      ret->putString( "/Fore " );
    
    if( sh->rOpt&RunOptParent )
      ret->putString( "/Parent " );
      
    if( sh->rOpt&RunOptInVisible )
      ret->putString( "/InVisible " );
      
    if( sh->rOpt&RunOptMax )
      ret->putString( "/Max " );
      
    if( sh->rOpt&RunOptMin )
      ret->putString( "/Min " );
      
    if( sh->rOpt&RunOptNoClose )
      ret->putString( "/NoClose " );
      
    if( sh->rOpt&RunOptFs )
      ret->putString( "/FS " );
      
    if( sh->rOpt&RunOptWin )
      ret->putString( "/Win " );
      
    if( sh->rOpt&RunOptDos )
      ret->putString( "/DOS " );
      
    if( sh->rOpt&RunOptNoClip )
      ret->putString( "/NoClip " );
      
    if( sh->dosSetting!=NULL )
    {
      ret->putString( "/Env:" );
      ret->putString( sh->dosSetting );
      ret->putInto( ' ' );
    }    

    if( sh->icon!=NULL )
    {
      ret->putString( "/Icon:" );
      ret->putString( sh->icon );
      ret->putInto( ' ' );
    }    

    if( sh->title!=NULL )
    {
      ret->putString( "/Title:" );
      ret->putString( sh->title );
      ret->putInto( ' ' );
    }    

    if( sh->prior )
    {
      ret->putString( "/Prio:" );
      sprintf( buf, "%d ", sh->prior );
      ret->putString( buf );
    }
    
    if( sh->xPos || sh->yPos || sh->wPos || sh->hPos )
    {
      ret->putString( "/Pos:" );
      sprintf( buf, "%d,%d,%d,%d ", sh->xPos, sh->yPos, sh->wPos, sh->hPos );
      ret->putString( buf );
    }

    ret->putString( sh->params );
    break;
  }
  
  ret->putInto( 0 );
  
  strncat( retBuf, ret->str, len-1-strlen(retBuf) );
  
  delete ret;
  
  return stringSch( sh->next, retBuf, len, ++level );
}

ostream& operator << ( ostream& o, Schedule& sh )
{
  o << "{Schedule:Type:";

  if( !sh.isSchedule )
    o << "Action";
  else
    switch( sh.type )
    {
    case SchedTypeUnknown:
      o << "SchedTypeUnknown";
      break;

    case SchedTypeEvent:
      o << "SchedTypeEvent";
      break;

    case SchedTypeForce:
      o << "SchedTypeForce";
      break;

    default:
      o << "*Unknown*";
    }

  o << ",Do:";

  if( int(sh.action)&SchedActionThread )
    o << "Thread:";

  switch( SchedAction( int(sh.action)&~SchedActionThread ) )
  {
  case SchedActionOpen:
    o << "SchedActionOpen";
    break;

  case SchedActionIndexFREQ:
    o << "SchedActionIndexFREQ";
    break;

  case SchedActionCopy:
    o << "SchedActionCopy";
    break;

  case SchedActionCatch:
    o << "SchedActionCatch";
    break;

  case SchedActionNothing:
    o << "SchedActionNothing";
    break;

  case SchedActionWait:
    o << "SchedActionWait";
    break;

  case SchedActionInit:
    o << "SchedActionInit";
    break;

  case SchedActionCall:
    o << "SchedActionCall";
    break;

  case SchedActionPlay:
    o << "SchedActionPlay";
    break;

  case SchedActionPoll:
    o << "SchedActionPoll";
    break;

  case SchedActionExit:
    o << "SchedActionExit";
    break;

  case SchedActionExec:
    o << "SchedActionExec";
    break;

  case SchedActionEval:
    o << "SchedActionEval";
    break;

  case SchedActionSound:
    o << "SchedActionSound";
    break;

  case SchedActionRun:
    o << "SchedActionRun";
    break;

  case SchedActionCreate:
    o << "SchedActionCreate";
    break;

  case SchedActionRemove:
    o << "SchedActionRemove";
    break;

  case SchedActionSend:
    o << "SchedActionSend";
    break;

  case SchedActionGet:
    o << "SchedActionGet";
    break;

  case SchedActionHangUp:
    o << "SchedActionHangUp";
    break;

  case SchedActionLog:
    o << "SchedActionLog";
    break;

  case SchedActionReScan:
    o << "SchedActionReScan";
    break;

  case SchedActionSplit:
    o << "SchedActionSplit";
    break;

  case SchedActionReset:
    o << "SchedActionReset";
    break;

  case SchedActionDrop:
    o << "SchedActionDrop";
    break;

  case SchedActionBSY:
    o << "SchedActionBSY";
    break;

  case SchedActionRexx:
    o << "SchedActionRexx";
    break;

  case SchedActionSemCreate:
    o << "SchedActionSemCreate";
    break;

  case SchedActionSemRemove:
    o << "SchedActionSemRemove";
    break;

  case SchedActionSemPost:
    o << "SchedActionSemPost";
    break;

  case SchedActionSemReset:
    o << "SchedActionSemReset";
    break;

  case SchedActionSemWait:
    o << "SchedActionSemWait";
    break;

  case SchedActionCompile:
    o << "SchedActionCompile";
    break;

  case SchedActionSync:
    o << "SchedActionSync";
    break;

  case SchedActionShutdown:
    o << "SchedActionShutdown";
    break;

  case SchedActionReboot:
    o << "SchedActionReboot";
    break;

  default:
    o << "*Unknown*";
  }

  o << ",Done/Code:" << sh.done
    << ",DayOfWeek:" << sh.dayOfWeek
    << ",DayOfMonth:" << sh.dayOfMonth
    << ",Start:"  << sh.start/60/60 << char(':')
                   << (sh.start/60)%60 << char(':')
                   << sh.start%60;

  o << ",Gmt:" << (sh.isGmt ? "YES" : "NO")
    << ",Name:";
  strItem( o, sh.name );

  o << ",Token:" << sh.tokenLen << char(':') << sh.token;
  o << ",ROpt:" << sh.rOpt;

  o << ",Params:";
  strItem( o, sh.params );

  o << ",PreLoad:";
  strItem( o, sh.preloaded );

  o << ",Sound:" << sh.sndCount << "->";

  if( sh.sndCount )
    if( sh.preloaded )
      for( int i=0; i<sh.sndCount; i++ )
        o << ((char *)sh.sound)[i];
    else
      for( int i=0; i<sh.sndCount; i++ )
        o << sh.sound[i] << char(' ');
  else
    o << "None";

  o << ",Flavour:" << flavourName( sh.flavour )
    << ",Address:";

  if( sh.adr==NULL )
    o << "<NULL>";
  else
  {
    char b[65];

    o << sh.adr->toString( b, 64, FIDOAddressShowModeEMSI );
  }

  o << ",RC:" << sh.rc;

  o << ",DOS:";
  strItem( o, sh.dosSetting );

  o << ",Title:";
  strItem( o, sh.title );

  o << ",Icon:";
  strItem( o, sh.icon );

  o << ",Prio:" << sh.prior
    << ",Pos:" << sh.xPos << char(',') << sh.yPos << char(',')
                << sh.wPos << char(',') << sh.hPos;

  o << ",Next:" << sh.next;

  if( sh.next!=0 )
    o << endl << ",DoNext:" << *sh.next;

  o << char('}');

  return o;
}

Schedule *newSched( Schedule *init )
{
  if( init==NULL )
    return NULL;

  Schedule *ret = new Schedule;

  memcpy( ret, init, sizeof( Schedule ) );

  ret->name = newStr( init->name );
  ret->params = newStr( init->params );
  ret->preloaded = newStr( init->preloaded );
  ret->dosSetting = newStr( init->dosSetting );
  ret->title = newStr( init->title );
  ret->icon = newStr( init->icon );
  ret->token = 0;
  ret->tokenLen = 0;

  if( init->preloaded )
  {
    ret->sound = (int *)new char [ ret->sndCount ];
    memcpy( ret->sound, init->sound, ret->sndCount );
  }
  else if( init->sndCount && init->sound!=NULL )
  {
    ret->sound = new int [ ret->sndCount ];

    for( int i=0; i<ret->sndCount; i++ )
      ret->sound[i] = init->sound[i];
  }
  else
  {
    init->sndCount = 0;
    init->sound    = NULL;
  }

  if( init->adr==NULL )
    ret->adr = NULL;
  else
    ret->adr = new FIDOAddress( init->adr );

  if( init->next!=NULL )
    ret->next = newSched( init->next );

  return ret;
}

void deleteSched( Schedule *sh, int delThis )
{
  if( sh==NULL )
    return;

  delete sh->name;
  delete sh->params;
  delete sh->sound;
  delete sh->adr;
  delete sh->preloaded;
  delete sh->dosSetting;
  delete sh->title;
  delete sh->icon;

  if( sh->token )
    DosFreeMem( sh->token );

  if( sh->next!=NULL )
    deleteSched( sh->next );

  if( delThis )
    delete sh;
}

void sortSchedList( int count, Schedule *a )
{
  for( int i=1; i<count; i++ )
  {
    Schedule x = a[i];
    int j = i-1;

    while( j!=-1 && x.start<a[j].start )
    {
      a[j+1] = a[j];
      --j;
    }

    a[j+1] = x;
  }
}

void moveNodeFiles( FIDOAddress *adr, char *path )
{
  showAction( Msg_ScrCopy );

  char buf[65];

  log ( "" );
  log( Msg_LogMovingFiles, adr->toString( buf, sizeof(buf) ), path );

  NodeConfig test( adr ),
             *found = systemList->Find( test );

  if( found==NULL )
    found = &test;

  QueueEntry testQue( *found ),
             *que = queue->Find( testQue );

  if( que==NULL )
    que = &testQue;

  int fsent( 0 );

  sendAka( *que, 0, fsent, COPY, NULL, path, 0, NULL, 0 );
  Queue::needForRescan();
  log ( "" );
}

void getRemoteFiles( char *from )
{
  showAction( Msg_ScrFrom );

  log ( "" );
  log( Msg_LogMovingFilesFrom, from );

  char mask[]="*",
       *path = (char *)_alloca( strlen(from)+sizeof(mask) ),
       *full = (char *)_alloca( strlen(from)+_MAX_PATH );

  strcpy( path, from );
  strcat( path, mask );

  for( Directory rc( path, MayBeArchived ); rc; rc++ )
  {
    strcpy( full, rc.getPath() );
    strcat( full, rc.getName() );

    if( CopyFileToPath( full, NULL, miscInfo.protectedInbound!=NULL ?
                                      miscInfo.protectedInbound :
                                      miscInfo.inbound )==OK )
      if( unlink( full ) )
        log( Msg_ErrUnableDelete, full, strError() );
  }

  Queue::needForRescan();
  log ( "" );
}

int reboot()
{
  ULONG action;
  ULONG handle;
  PSZ dev( "/dev/dos$" );

  APIRET rc = DosOpen( dev, &handle, &action,
                  0,
                  FILE_NORMAL,
                  OPEN_ACTION_OPEN_IF_EXISTS,
                  OPEN_FLAGS_FAIL_ON_ERROR|OPEN_FLAGS_SEQUENTIAL|
                  OPEN_SHARE_DENYNONE|OPEN_ACCESS_READWRITE,
                  0 );
  if( rc )
  {
    log( Msg_ErrUnableOpenOS, dev, rc );
    os2Message( rc );
    return int( rc );
  }

  DosShutdown( 1 );

  rc = DosDevIOCtl( handle, 0xd5, 0xab,
                    0, 0, 0,
                    0, 0, 0 );

  DosClose( handle );

  if( rc )
  {
    log( Msg_ErrUnableShutdown, rc );
    os2Message( rc );
    return int( rc );
  }

  return 0;
}

int doEventKeyTrivial( Schedule& sh, int current,
                       FIDOAddress *fAddr, long elapsedTime,
                       char *fileName )
{
  void splitterForce();

  int waitMode( P_WAIT );

  switch( sh.action )
  {
  case SchedActionShutdown:
    DosShutdown( 0 );
    break;

  case SchedActionReboot:
    reboot();
    break;

  case SchedActionSync:
    DosShutdown( 1 );
    break;

  case SchedActionCopy:
    moveNodeFiles( sh.adr, sh.params );
    break;

  case SchedActionCatch:
    getRemoteFiles( sh.params );
    break;

  case SchedActionCompile:
    compiler();
    break;

  case SchedActionSemCreate:
    brakeSemaphoreCreate( sh.params );
    break;

  case SchedActionSemRemove:
    brakeSemaphoreRemove( sh.params );
    break;

  case SchedActionSemPost:
    brakeSemaphorePost( sh.params );
    break;

  case SchedActionSemReset:
    brakeSemaphoreReset( sh.params );
    break;

  case SchedActionSemWait:
    brakeSemaphoreWait( sh.params );
    break;

  case SchedActionNothing:
    log( Msg_InfoEmptyAction );
    break;

  case SchedActionLog:
    log( sh.params==NULL ? "<Not specified>" : sh.params );
    break;

  case SchedActionOpen:
    openObject( sh.params );
    break;

  case SchedActionIndexFREQ:
    buildFreqIdx();
    break;

  case SchedActionHangUp:
    comPort->lockSema();
    log( Msg_InfoActionHang );
    comPort->hangUp();
    comPort->unlockSema();
    break;

  case SchedActionWait:
    log( Msg_InfoActionWait, sh.rc );
    sleep( sh.rc*1000 );
    break;

  case SchedActionInit:
    needInit = 1;
    break;

  case SchedActionExit:
    if( terminalActive )
      log( Msg_InfoActiveTerminalExit );

    if( comPort->isLocked() && getMyTid()!=mainTid )
      log( Msg_InfoActiveModemUsedExit );

    if( sessionStart && getMyTid()!=mainTid )
      log( Msg_InfoActiveSessionExit );

    return sh.rc;

  case SchedActionRun:
    waitMode = P_NOWAIT;

  // fall

  case SchedActionExec:

    if( sh.params!=NULL )
    {
      if( fAddr==NULL )
        fAddr = (FIDOAddress *) queue->getCurrent();

      if( fAddr==NULL )
        fAddr = defaultAddress;

      makeCmd( *fAddr, elapsedTime, fileName,
               sh.params, waitMode, sh.rOpt, sh.name,
               sh.dosSetting, &sh );

      if( modemOs2Rc )
        return int( modemOs2Rc );
    }
    break;

  case SchedActionCreate:
    {
      makeFilePath( sh.params );

      FILE *fl = _fsopen( sh.params, "a" );

      if( fl==NULL )
        log( Msg_ErrUnableCreateFlag, sh.params, strError() );
      else
      {
        if( miscInfo.flags==NULL || !begin( sh.params, miscInfo.flags ) )
          log( Msg_InfoFlagCreated, sh.params );
        else
          log( Msg_InfoFlagCreated, sh.params+strlen(miscInfo.flags) );

        fclose( fl );
      }
    }
    break;

  case SchedActionRemove:
    if( unlink( sh.params ) )
      log( Msg_ErrUnableDeleteFlag, sh.params, strError() );
    else
      if( miscInfo.flags==NULL || !begin( sh.params, miscInfo.flags ) )
        log( Msg_InfoFlagDeleted, sh.params );
      else
        log( Msg_InfoFlagDeleted, sh.params+strlen(miscInfo.flags) );
    break;

  case SchedActionSend:
    if( sh.adr!=NULL && sh.params!=NULL )
    {
      sh.adr->makeSend( sh.params, sh.flavour );
      Queue::needForRescan();
    }
    break;

  case SchedActionGet:
    if( sh.adr!=NULL && sh.params!=NULL )
    {
      sh.adr->makeFreq( sh.params, sh.flavour );
      Queue::needForRescan();
    }
    break;

  case SchedActionPoll:
    if( sh.adr!=NULL )
    {
      sh.adr->makeLO( sh.flavour );
      Queue::needForRescan();
    }
    break;

  case SchedActionDrop:
    if( sh.adr!=NULL )
    {
      sh.adr->dropLO( sh.flavour );
      Queue::needForRescan();
    }
    break;

  case SchedActionRexx:
    {
      if( fAddr==NULL )
        fAddr = (FIDOAddress *) queue->getCurrent();

      if( fAddr==NULL )
        fAddr = defaultAddress;

      if( sh.preloaded!=NULL )
        doEvalRexxParam( sh, *fAddr, elapsedTime, fileName, -1, sh.name  );
      else if( sh.params!=NULL )
        doParamsRexx( *fAddr, elapsedTime, fileName, sh.params, -1, sh.name );
    }
    break;

  case SchedActionEval:
    doEvalRexx( sh );
    break;

  case SchedActionSound:
    playMelody( (int *)sh.sound );
    break;

  case SchedActionPlay:
    playWaveFile( sh.params );
    break;

  case SchedActionReScan:
    log( Msg_InfoActionRescan );
    Queue::needForRescan();
    break;

  case SchedActionReset:
    log( Msg_InfoActionTry );
    resetTry();
    break;

  case SchedActionSplit:
    log( Msg_InfoActionSpl );
    splitterForce();
    break;

  case SchedActionCall:
    {
      Schedule *doItem = NULL;

      if( actionList!=NULL )
        for( int i=0; i<actionCount && doItem==NULL; i++ )
           if( sh.params!=NULL && actionList[i].name!=NULL &&
               !stricmp( actionList[i].name, sh.params ) )
             doItem = actionList+i;

      if( doItem==NULL )
        log( Msg_LogNoSuchAction, sh.params );
      else
      {
        int rc = doEventKey( *doItem, current, fAddr,
                             elapsedTime, fileName );

        if( rc )
          return rc;
      }
    }
    break;

  case SchedActionBSY:
    {
      void resetBsy( int );

      resetBsy( 0 );
    }
    break;

  default:
    log( "~BUG~ doEventKeyTrivial() unknown action %lu",
         (unsigned long)sh.action );
  }

  return 0;
}

class DoActionParams
{
public:
  Schedule *sh;
  int current;
  FIDOAddress *fAddr;
  long elapsedTime;
  char *fileName;

  DoActionParams( int _c, long _e, char *_f, FIDOAddress *_fa,
                  Schedule& _s ) :
    current( _c ),
    elapsedTime( _e ),
    fileName( newStr( _f ) ),
    fAddr( new FIDOAddress( _fa ) ),
    sh( newSched( &_s ) )
  {}

  ~DoActionParams()
  {
    delete fileName;
    delete fAddr;
    deleteSched( sh, 1 );
  }
};

void _Optlink doEventKeyThread( void *params )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  ((DoActionParams *)params)->sh->action =
    SchedAction( int(((DoActionParams *)params)->sh->action) & ~SchedActionThread );

  int rc = doEventKeyTrivial( *((DoActionParams *)params)->sh,
                               ((DoActionParams *)params)->current,
                               ((DoActionParams *)params)->fAddr,
                               ((DoActionParams *)params)->elapsedTime,
                               ((DoActionParams *)params)->fileName );

  if( rc>OSReturnCode )
    OSReturnCode = rc;

  delete (DoActionParams *)params;

  deRegisterExceptionHandler
}

int doEventKey( Schedule& shi, int current, FIDOAddress *fAddr,
                long elapsedTime, char *fileName )
{
  for( Schedule* sh( &shi ); sh; sh=sh->next )
    if( (sh->action&SchedActionThread)!=SchedActionThread )
    {
      int rc( doEventKeyTrivial( *sh, current, fAddr, elapsedTime, fileName ) );

      if( rc )
        return rc;
    }
    else
    {
      DoActionParams *params = new DoActionParams( current, elapsedTime,
                                                   fileName, fAddr, *sh );

      long tid = _beginthread( doEventKeyThread, 0, 40960, (void *)params );

      if( tid==-1 )
      {
        log( Msg_ErrActionThreadError, strError() );
        delete params;
      }
      else
        log( Msg_ErrActionThreadStarted, tid );
    }

  return 0;
}

// ---------------------- Save/Restore events

int writeString( FILE *out, char *str )
{
  if( str==NULL )
    return 0;

  short len( short( strlen( str ) ) );

  if( fwrite( &len, 1, sizeof(len), out ) != sizeof(len) )
    return 1;

  if( fwrite( str, 1, len, out ) != len )
    return 1;

  return 0;
}

int saveEvents()
{
  time_t tim = time( NULL );
  struct tm *tm = localTime( &tim );
  SchedDate schedDate = {
                          short( tm->tm_year+1900 ),
                          char( tm->tm_mday ),
                          char( tm->tm_mon+1 ),
			  char( tm->tm_wday )
                        };

  FILE *out( _fsopen( miscInfo.schedComp, "wb" ) );

  if( out==NULL )
  {
    log( Msg_ErrUnableCreate, miscInfo.schedComp, strError() );
    return 1;
  }

  char ab( 0 );

  if( fwrite( &SchedBinaryType, 1, sizeof(SchedBinaryType), out ) != sizeof(SchedBinaryType) )
    ab = 1;
  else if( fwrite( &schedDate, 1, sizeof(schedDate), out ) != sizeof(schedDate) )
    ab = 1;
  else if( fwrite( &schedCount, 1, sizeof(schedCount), out ) != sizeof(schedCount) )
    ab = 1;
  else
    for( int i( 0 ); i<schedCount && !ab; i++ )
      for( Schedule *sh = schedList+i; sh!=NULL && !ab; sh=sh->next )
      {
        if( fwrite( sh, 1, sizeof(Schedule), out ) != sizeof(Schedule) )
        {
          ab = 1;
          break;
        }

        if( writeString( out, sh->name ) )
        {
          ab = 1;
          break;
        }

        if( writeString( out, sh->preloaded ) )
        {
          ab = 1;
          break;
        }

        if( writeString( out, sh->params ) )
        {
          ab = 1;
          break;
        }

        if( writeString( out, sh->dosSetting ) )
        {
          ab = 1;
          break;
        }

        if( writeString( out, sh->title ) )
        {
          ab = 1;
          break;
        }

        if( writeString( out, sh->icon ) )
        {
          ab = 1;
          break;
        }

        if( sh->preloaded==0 )
        {
          int size1 = sizeof(sh->sound[0])*sh->sndCount;

          if( sh->sndCount && sh->sound!=NULL )
            if( fwrite( sh->sound, 1, size1, out ) != size1 )
            {
              ab = 1;
              break;
            }
        }

        if( sh->adr!=0 )
        {
          unsigned short ad;

          ad = sh->adr->getZone();

          if( fwrite( &ad, 1, sizeof(ad), out ) != sizeof(ad) )
          {
            ab = 1;
            break;
          }

          ad = sh->adr->getNet();

          if( fwrite( &ad, 1, sizeof(ad), out ) != sizeof(ad) )
          {
            ab = 1;
            break;
          }

          ad = sh->adr->getNode();

          if( fwrite( &ad, 1, sizeof(ad), out ) != sizeof(ad) )
          {
            ab = 1;
            break;
          }

          ad = sh->adr->getPoint();

          if( fwrite( &ad, 1, sizeof(ad), out ) != sizeof(ad) )
          {
            ab = 1;
            break;
          }
        }
      }

  if( ab )
    log( Msg_ErrUnableWrite, miscInfo.schedComp, strError() );

  fclose( out );
  setExtendedAttribute( miscInfo.schedComp, ".TYPE", Msg_EASCH );

  if( ab )
    if( unlink( miscInfo.schedComp ) )
      log( Msg_ErrUnableDelete, miscInfo.schedComp, strError() );

  return 0;
}

void prepareRexxText( char *str )
{
  for( unsigned char *pp=(unsigned char *)str, *pp1=(unsigned char *)str; *pp; pp++ )
    if( *pp>=' ' )
    {
      *pp1 = *pp;
      pp1++;
    }

  *pp1 = 0;
}

void loadRexxProcedure( Schedule &sh, int line )
{
  FILE *rexx = _fsopen( sh.preloaded, "rb" );

  if( rexx == NULL )
  {
    log( Msg_ErrUnableOpen, sh.preloaded, strError() );
FileError:

    if( line )
      log( Msg_ErrAtLine, line );

    delete sh.preloaded;
    sh.preloaded = NULL;
  }
  else
  {
    sh.sndCount = int( filelength( fileno( rexx ) )+2 );
    sh.sound = (int *)new char [ sh.sndCount ];

    int total = fread( sh.sound, 1, sh.sndCount, rexx );

    if( ferror( rexx ) )
    {
      log( Msg_ErrUnableRead, sh.preloaded, strError() );
      sh.sndCount = 0;
      delete (char *)sh.sound;
      sh.sound = NULL;
      goto FileError;
    }

    ((char *)sh.sound)[total] = 'Z'-'@';
    ((char *)sh.sound)[total+1] = 0;
    prepareRexxText( (char *)sh.sound );

    fclose( rexx );
  }
}

int readString( FILE *out, char **str )
{
  if( *str==NULL )
    return 0;

  short len;

  if( fread( &len, 1, sizeof(len), out ) != sizeof(len) )
    return 1;

  *str = new char [len+1];

  if( fread( *str, 1, len, out ) != len )
    return 1;

  (*str)[len] = 0;

  return 0;
}

int loadEvents()
{
  schedCurrent = 0;

  if( schedList!=NULL )
  {
    for( int i(0); i<schedCount; i++ )
      deleteSched( &schedList[i], 0 );

    delete schedList;
    schedList = NULL;
    schedCount = 0;
  }

  if( access( miscInfo.schedComp, 0 ) )
    return 0;

  log( Msg_InfoLoading, miscInfo.schedComp );

  FILE *out( _fsopen( miscInfo.schedComp, "rb", SH_DENYWR ) );

  if( out==NULL )
  {
    log( Msg_ErrUnableOpen, miscInfo.schedComp, strError() );
    return 1;
  }

  char ab( 0 );
  long version;
  int rc(0);

  if( fread( &version, 1, sizeof(version), out ) != sizeof(version) )
    ab = 1;
  else if( version!=SchedBinaryType )
  {
    log( Msg_ErrIncompat, miscInfo.schedComp );
    ab = 1;
    rc = 1;
  }
  else if( fread( &schedDate, 1, sizeof(schedDate), out ) != sizeof(schedDate) )
    ab = 1;
  else if( fread( &schedCount, 1, sizeof(schedCount), out ) != sizeof(schedCount) )
    ab = 1;
  else
  {
    schedList = new Schedule [ schedCount ];

    for( int i( 0 ); i<schedCount && !ab; i++ )
      for( Schedule *sh=schedList+i; sh!=NULL && !ab; sh = sh->next )
      {
        if( fread( sh, 1, sizeof(Schedule), out ) != sizeof(Schedule) )
        {
          ab = 1;
          break;
        }

        sh->token = 0;
        sh->tokenLen = 0;

        if( readString( out, &sh->name ) )
        {
          ab = 1;
          break;
        }

        if( readString( out, &sh->preloaded ) )
        {
          ab = 1;
          break;
        }

        if( readString( out, &sh->params ) )
        {
          ab = 1;
          break;
        }

        if( readString( out, &sh->dosSetting ) )
        {
          ab = 1;
          break;
        }

        if( readString( out, &sh->title ) )
        {
          ab = 1;
          break;
        }

        if( readString( out, &sh->icon ) )
        {
          ab = 1;
          break;
        }

        if( sh->preloaded==0 )
        {
          int size1 = sizeof(sh->sound[0])*sh->sndCount;

          if( sh->sndCount && sh->sound!=NULL )
          {
            sh->sound = new int [ sh->sndCount ];

            if( fread( sh->sound, 1, size1, out ) != size1 )
            {
              ab = 1;
              break;
             }
          }
        }

        if( sh->adr!=0 )
        {
          unsigned short zone, net, node, point;

          if( fread( &zone, 1, sizeof(zone), out ) != sizeof(zone) )
          {
            ab = 1;
            break;
          }

          if( fread( &net, 1, sizeof(net), out ) != sizeof(net) )
          {
            ab = 1;
            break;
          }

          if( fread( &node, 1, sizeof(node), out ) != sizeof(node) )
          {
            ab = 1;
            break;
          }

          if( fread( &point, 1, sizeof(point), out ) != sizeof(point) )
          {
            ab = 1;
            break;
          }

          sh->adr = new FIDOAddress( zone, net, node, point );
        }

        if( sh->preloaded!=0 )
          loadRexxProcedure( *sh, 0 );

        if( sh->next!=NULL )
        {
          sh->next = new Schedule;
          memset( sh->next, 0, sizeof( Schedule ) );
        }
      }
  }

  if( ab && !rc )
    log( Msg_ErrUnableRead, miscInfo.schedComp, strError() );

  fclose( out );

  if( ab )
  {
    schedCount = 0;
    delete schedList;
    schedList = NULL;

    if( unlink( miscInfo.schedComp ) )
      log( Msg_ErrUnableDelete, miscInfo.schedComp, strError() );
  }

  return rc;
}

// ------------------------ compile


int evtAddress( ifstream **cfg, Schedule &sh )
{
  if( getWord( cfg ) )
    return 1;

  unsigned short zone, net, node, point;
  char *domain( NULL );

  if( !checkAddress( inputWord.str, zone, net, node, point, &domain ) )
  {
    log( Msg_ErrCfgIllAddr, inputWord.str, cfgLine );
    return 2;
  }

  sh.adr = new FIDOAddress( zone, net, node, point, domain );
  return 0;
}

int evtFlavourAddress( ifstream **cfg, Schedule &sh )
{
  if( getWord( cfg ) )
    return 1;

  sh.flavour = unknownFlavour;

  if( !stricmp( inputWord.str, "Normal" ) )
    sh.flavour = normalFlavour;
  else if( !stricmp( inputWord.str, "Crash" ) )
    sh.flavour = crashFlavour;
  else if( !stricmp( inputWord.str, "Immed" ) )
    sh.flavour = immedFlavour;
  else if( !stricmp( inputWord.str, "Direct" ) )
    sh.flavour = directFlavour;
  else if( !stricmp( inputWord.str, "Hold" ) )
    sh.flavour = holdFlavour;

  if( sh.flavour == unknownFlavour )
    return 3;

  return evtAddress( cfg, sh );
}

int evtCopy( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionCopy;

  register int rc = evtAddress( cfg, sh );

  if( rc )
    return rc;

  return storePath( cfg, &sh.params, "Path", 0 );
}

int evtCatch( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionCatch;

  return storePath( cfg, &sh.params, "Path", 0 );
}

int evtPoll( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionPoll;

  return evtFlavourAddress( cfg, sh );
}

int evtDrop( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionDrop;

  return evtFlavourAddress( cfg, sh );
}

int evtExit( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionExit;

  if( getWord( cfg ) )
    return 1;

  char *n;

  sh.rc = short( strtoul( inputWord.str, &n, 10 ) );

  if( *n || sh.rc>255 || sh.rc<10 )
  {
    log( Msg_ErrCfgIllNumbExit, inputWord.str, cfgLine );
    return 2;
  }

  return 0;
}

int evtWait( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionWait;

  if( getWord( cfg ) )
    return 1;

  char *n;

  sh.rc = short( strtoul( inputWord.str, &n, 10 ) );

  if( *n )
  {
    log( Msg_ErrCfgIllNumb, inputWord.str, cfgLine );
    return 2;
  }

  return 0;
}

int evtRun( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionRun;

  for( char done=0; !done; )
  {
    if( getWord( cfg ) )
      return 1;

    if( *inputWord.str!='/' )
      done = 1;
    else if( !stricmp( inputWord.str+1, "Child" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptChild );
    else if( !stricmp( inputWord.str+1, "Wait" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptWait | RunOptChild );
    else if( !stricmp( inputWord.str+1, "Fore" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptFore );
    else if( !stricmp( inputWord.str+1, "Parent" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptParent );
    else if( !stricmp( inputWord.str+1, "InVisible" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptInVisible );
    else if( !stricmp( inputWord.str+1, "Max" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptMax );
    else if( !stricmp( inputWord.str+1, "Min" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptMin );
    else if( !stricmp( inputWord.str+1, "NoClose" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptNoClose );
    else if( !stricmp( inputWord.str+1, "Fs" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptFs );
    else if( !stricmp( inputWord.str+1, "Win" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptWin );
    else if( !stricmp( inputWord.str+1, "Dos" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptDos );
    else if( !stricmp( inputWord.str+1, "NoClip" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptNoClip );
    else if( !strnicmp( inputWord.str+1, "Env:", 4 ) )
      sh.dosSetting = newStr( inputWord.str+5 );
    else if( !strnicmp( inputWord.str+1, "Icon:", 5 ) )
      sh.icon = newStr( inputWord.str+6 );
    else if( !strnicmp( inputWord.str+1, "Title:", 6 ) )
      sh.title = newStr( inputWord.str+7 );
    else if( !strnicmp( inputWord.str+1, "Prio:", 5 ) )
    {
      char *e;

      sh.prior = (int)strtoul( inputWord.str+6, &e, 0 );

      if( *e )
      {
        log( Msg_CfgErrIllegalPrio, cfgLine, inputWord.str );
        return 2;
      }
    }
    else if( !strnicmp( inputWord.str+1, "Pos:", 4 ) )
    {
      char *e,
           *b = inputWord.str+5;

      for( int i=0; i<4; i++ )
      {
        int val = (int)strtoul( b, &e, 10 );

        switch( i )
        {
        case 0:
          sh.xPos = val;
          break;

        case 1:
          sh.yPos = val;
          break;

        case 2:
          sh.wPos = val;
          break;

        case 3:
          sh.hPos = val;
          break;
        }

        if( !*e )
          break;

        if( *e!=',' )
          goto Illegal_Pos;

        b = e+1;
      }

      if( *e )
      {
Illegal_Pos:
        log( Msg_CfgErrIllegalPos, cfgLine, inputWord.str );
        return 2;
      }
    }
    else
      log( Msg_ErrIllegalModifier, inputWord.str, cfgLine );
  }

  sh.params = newStr( inputWord.str );

  return 0;
}

int evtRexx( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionRexx;

  if( getWord( cfg ) )
    return 1;

  char preLoad = 0;

  if( !stricmp( inputWord.str, "/PreLoad" ) )
  {
    if( getWord( cfg ) )
      return 1;

    preLoad = 1;
  }

  sh.params = newStr( inputWord.str );

  if( preLoad )
  {
    char *bl = strchr( inputWord.str, ' ' );

    if( bl!=NULL )
      *bl = 0;

    sh.preloaded = rexxProcedureName( inputWord.str );

    loadRexxProcedure( sh, cfgLine );
  }

  return 0;
}

int evtCall( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionCall;

  if( getWord( cfg ) )
    return 1;

  sh.params = newStr( inputWord.str );
  return 0;
}

int evtSemaphore( ifstream **cfg, Schedule &sh )
{
  if( getWord( cfg ) )
    return 1;

  if( !stricmp( inputWord.str, "Create" ) )
    sh.action = SchedActionSemCreate;
  else if( !stricmp( inputWord.str, "Remove" ) )
    sh.action = SchedActionSemRemove;
  else if( !stricmp( inputWord.str, "Post" ) )
    sh.action = SchedActionSemPost;
  else if( !stricmp( inputWord.str, "Reset" ) )
    sh.action = SchedActionSemReset;
  else if( !stricmp( inputWord.str, "Wait" ) )
    sh.action = SchedActionSemWait;
  else
    return need( "Create/Remove/Post/Reset/Wait" );

  if( getWord( cfg ) )
    return 1;

  if( *inputWord.str=='\\' )
    sh.params = newStr( inputWord.str );
  else
  {
    sh.params = new char [ strlen(miscInfo.semaphore)+strlen(inputWord.str)+1 ];
    strcpy( sh.params, miscInfo.semaphore );
    strcat( sh.params, inputWord.str );
  }

  capitalize( sh.params );
  return 0;
}

int evtCreateRemove( ifstream **cfg, Schedule &sh, SchedAction act )
{
  sh.action = (short)act;

  if( getWord( cfg ) )
    return 1;

  char drive[_MAX_DRIVE],
       dir  [_MAX_DIR],
       name [_MAX_FNAME],
       ext  [_MAX_EXT];

  _splitpath( inputWord.str, drive, dir, name, ext );

  if( *dir || *drive )
    sh.params = newStr( inputWord.str );
  else
    if( miscInfo.flags==NULL )
    {
      log( Msg_ErrCfgIllFlagLoc, cfgLine );
      return 2;
    }
    else
    {
      sh.params = new char [ strlen(miscInfo.flags)+strlen(inputWord.str)+1 ];
      strcpy( sh.params, miscInfo.flags );
      strcat( sh.params, inputWord.str );
    }

  return 0;
}

int evtExec( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionExec;

  while( 1 )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "/Close" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptClose );
    else if( !stricmp( inputWord.str, "/Clear" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptClear );
    else if( !stricmp( inputWord.str, "/NoClip" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptNoClip );
    else if( !stricmp( inputWord.str, "/Port" ) )
      sh.rOpt = RunOption( sh.rOpt | RunOptPort );
    else if( !strnicmp( inputWord.str, "/Env:", 5 ) )
      sh.dosSetting = newStr( inputWord.str+5 );
    else
      break;
  }

  sh.params = newStr( inputWord.str );

  return 0;
}


int storeMelodyItem( ifstream **cfg, Buffer *buf, char *who,
                     int *num, char needMsg )
{
  if( getWord( cfg ) )
    if( needMsg )
    {
      inputWord.putString( "EOF" );
      return need( who );
    }
    else
      return 1;

  char *e;
  int n;

  n = (int)strtol( inputWord.str, &e, 10 );

  if( *e )
  {
    if( needMsg )
      log( Msg_ErrCfgExpect, who, inputWord.str, cfgLine );

    return 1;
  }

  if( num!=NULL )
    *num = n;

  if( buf!=NULL )
  {
    e=(char *)&n;

    for( int i=0; i<sizeof(n); i++ )
      buf->putInto( e[i] );
  }

  return 0;
}

int evtSound( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionSound;

  Buffer buf;
  int count;

  if( storeMelodyItem( cfg, NULL, "Repetition Count", &count, 1 ) )
    return 2;

  char first(1);

  while( 1 )
  {
    if( storeMelodyItem( cfg, &buf, "Tone", NULL, first ) )
      if( first )
        return 2;
      else
        break;

    first = 0;

    if( storeMelodyItem( cfg, &buf, "Tone Delta", NULL, 1 ) )
      return 2;

    if( storeMelodyItem( cfg, &buf, "Duration", NULL, 1 ) )
      return 2;

    if( storeMelodyItem( cfg, &buf, "Tackts Count", NULL, 1 ) )
      return 2;
  }

  int len = buf.length();

  if( len%(sizeof(int)*4) )
  {
    log( "~BUG~ storeSound() illegal count\n" );
    return 2;
  }

  delete sh.sound;
  sh.sndCount = len/sizeof(int)+2;
  sh.sound = new int [ sh.sndCount ];

  memcpy( sh.sound+2, buf.str, len );
  sh.sound[0] = len/sizeof(int)/4;
  sh.sound[1] = count;

  parserNoRead = 1;
  return 0;
}

int evtEval( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionEval;

  if( getWord( cfg ) )
    return 1;

  prepareRexxText( inputWord.str );
  sh.params = newStr( inputWord.str );

  return 0;
}

int evtOpen( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionOpen;

  if( getWord( cfg ) )
    return 1;

  sh.params = newStr( inputWord.str );

  return 0;
}

int evtLog( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionLog;

  if( getWord( cfg ) )
    return 1;

  sh.params = newStr( inputWord.str );

  return 0;
}

int evtPlay( ifstream **cfg, Schedule &sh )
{
  sh.action = SchedActionPlay;

  if( getWord( cfg ) )
    return 1;

  sh.params = newStr( inputWord.str );

  return 0;
}

int evtGetSend( ifstream **cfg, Schedule &sh,
				   SchedAction act )
{
  sh.action = short(act);

  if( getWord( cfg ) )
    return 1;

  sh.params = newStr( inputWord.str );

  return evtFlavourAddress( cfg, sh );
}

int processAction( ifstream **cfg, Schedule& sh )
{
  if( getWord( cfg ) )
    return 1;

  int thread = 0,
      rc;

  if( !stricmp( inputWord.str, "Thread" ) )
  {
    thread = 1;

    if( getWord( cfg ) )
      return 1;
  }

  if( !stricmp( inputWord.str, "Poll" ) )
  {
    if( rc=evtPoll( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Copy" ) )
  {
    if( rc=evtCopy( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Catch" ) )
  {
    if( rc=evtCatch( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Drop" ) )
  {
    if( rc=evtDrop( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Exit" ) )
  {
    if( rc=evtExit( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Wait" ) )
  {
    if( rc=evtWait( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Semaphore" ) )
  {
    if( rc=evtSemaphore( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Exec" ) )
  {
    if( rc=evtExec( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Eval" ) )
  {
    if( rc=evtEval( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Sound" ) )
  {
    needSoundThread |= 1;

    if( rc=evtSound( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Play" ) )
  {
    needSoundThread |= 2;

    if( rc=evtPlay( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Open" ) )
  {
    if( rc=evtOpen( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Run" ) )
  {
    if( rc=evtRun( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Rexx" ) )
  {
    if( rc=evtRexx( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Call" ) )
  {
    if( rc=evtCall( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Create" ) )
  {
    if( rc=evtCreateRemove( cfg, sh, SchedActionCreate ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Remove" ) )
  {
    if( rc=evtCreateRemove( cfg, sh, SchedActionRemove ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Send" ) )
  {
    if( rc=evtGetSend( cfg, sh, SchedActionSend ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "Get" ) )
  {
    if( rc=evtGetSend( cfg, sh, SchedActionGet ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "HangUp" ) )
    sh.action = SchedActionHangUp;
  else if( !stricmp( inputWord.str, "InitModem" ) )
    sh.action = SchedActionInit;
  else if( !stricmp( inputWord.str, "Nothing" ) )
    sh.action = SchedActionNothing;
  else if( !stricmp( inputWord.str, "Log" ) )
  {
    if( rc=evtLog( cfg, sh ), rc )
      return rc;
  }
  else if( !stricmp( inputWord.str, "ReScan" ) )
    sh.action = SchedActionReScan;
  else if( !stricmp( inputWord.str, "Split" ) )
    sh.action = SchedActionSplit;
  else if( !stricmp( inputWord.str, "Reset" ) )
    sh.action = SchedActionReset;
  else if( !stricmp( inputWord.str, "CheckBSY" ) )
    sh.action = SchedActionBSY;
  else if( !stricmp( inputWord.str, "Compile" ) )
    sh.action = SchedActionCompile;
  else if( !stricmp( inputWord.str, "IndexFREQ" ) )
    sh.action = SchedActionIndexFREQ;
  else if( !stricmp( inputWord.str, "Sync" ) )
    sh.action = SchedActionSync;
  else if( !stricmp( inputWord.str, "Shutdown" ) )
    sh.action = SchedActionShutdown;
  else if( !stricmp( inputWord.str, "Reboot" ) )
    sh.action = SchedActionReboot;
  else
    return 3;

  if( thread )
    sh.action = SchedAction( int(sh.action)|SchedActionThread );

  if( getWord( cfg ) )
    return 0;

  if( !stricmp( inputWord.str, "And" ) )
  {
    rc = processSch( cfg, &sh.next );

    if( rc )
      return rc;
  }
  else
    parserNoRead = 1;

  return 0;
}

int checkTime( unsigned long& start, short& week, short& month,
               int weekMon, int *isGmt )
{
  start = 0;
  week = 0;
  month = 0;

  if( !*inputWord.str )
    return 1;

  unsigned long tm[3] ={ 0, 0, 0 };
  int i( 0 );

  if( isGmt )
  {
    int strLen = strlen( inputWord.str )-3;

    *isGmt = 0;

    if( strLen>0 && !stricmp( inputWord.str+strLen, "GMT" ) )
    {
      *isGmt = 1;
      inputWord.str[ strLen ] = 0;
    }
  }

  for( char *end( inputWord.str ); *end; )
  {
    unsigned long num( strtoul( end, &end, 10 ) );

    switch( *end )
    {
    case '.':
      if( !weekMon || tm[0] || tm[1] || tm[2] || !num || week )
        return 1;

      week = short( num );
      break;

    case '-':
      if( !weekMon || tm[0] || tm[1] || tm[2] || !num || month )
        return 1;

      month = short( num );
      break;

    case ':':
    case 0:
      if( i==3 )
	return 1;

      tm[i++] = num;
      break;

    default:
      return 1;
    }

    if( *end )
      end++;
  }

  start = (tm[0]*60+tm[1])*60+tm[2];

  if( tm[0]>23 || tm[1]>59 || tm[2]>59 ||
      week>7 || week<0 ||
      month>31 || month<0 )
    return 1;

  return 0;
}

void appendToScheduleList( Schedule *sh, int& schCnt )
{
  Schedule *sched = newSched( sh );

  schedList = (Schedule *)realloc( schedList, sizeof(Schedule)*(schedCount+1) );
  memcpy( schedList+schedCount, sched, sizeof(Schedule) );
  schedCount++;

  if( ++schCnt%10==0 )
  {
    printf( "%5d\r", schCnt );
    fflush( stdout );
  }

  delete sched;
}

int processEvent( ifstream **cfg, int& schCnt )
{
  if( getWord( cfg ) )
    return 255;

  if( !*inputWord.str )
    return 255;

  if( !stricmp(inputWord.str, "Complete.") )
    return 255;

  Schedule sh;
  memset( &sh, 0, sizeof(Schedule) );

  sh.type = SchedTypeUnknown;

  if( !stricmp( inputWord.str, "Event" ) )
    sh.type = SchedTypeEvent;
  else if( !stricmp( inputWord.str, "Force" ) )
    sh.type = SchedTypeForce;

  if( sh.type==SchedTypeUnknown )
    return need( "Event/Force/Complete." );

  if( getWord( cfg ) )
    return 1;

  sh.name = newStr( inputWord.str );
  sh.isSchedule = 1;

  if( getWord( cfg ) )
    return 1;

  if( checkTime( sh.start, sh.dayOfWeek, sh.dayOfMonth, 1, &sh.isGmt ) )
  {
    log( Msg_ErrCfgIllTime, inputWord.str, cfgLine );
    return 2;
  }

  unsigned long period(0);
  unsigned long times(0);

  if( getWord( cfg ) )
    return 1;

  if( !stricmp( inputWord.str, "Every" ) )
  {
    short f1, f2;

    if( getWord( cfg ) )
      return 1;

    if( checkTime( period, f1, f2, 0, 0 ) )
    {
      log( Msg_ErrCfgIllTime, inputWord.str, cfgLine );
      return 2;
    }
  }
  else
    parserNoRead = 1;

  if( getWord( cfg ) )
    return 1;

  if( !stricmp( inputWord.str, "Times" ) )
  {
    long timesTmp;
    int rc = storeLong( cfg, timesTmp, "Times" );

    if( rc )
      return rc;

    times = (unsigned long)timesTmp;
  }
  else
    parserNoRead = 1;

  int rc = processAction( cfg, sh );

  if( rc )
    return rc;

  if( !times && !period )
  {
    appendToScheduleList( &sh, schCnt );
    return 0;
  }

  if( times && !period )
    period = 24*60*60 / times;
  else if( !times && period )
    for( unsigned long st(0); st<24*60*60; st += period )
      ++times;

  if( !times )
    times = 1;

  unsigned long startAt( sh.start );

  while( times && sh.start<60*60*24 )
  {
    appendToScheduleList( &sh, schCnt );

    sh.start += period;
    --times;
  }

  if( times )
  {
    sh.start -= 24*60*60;

    while( times && sh.start<startAt )
    {
      appendToScheduleList( &sh, schCnt );

      sh.start += period;
      --times;
    }
  }

/*
  if( !times && !period )
    times = 1;
  else if( !times && period )
    times = (24*60*60-sh.start) / period;
  else if( times && !period )
    period = (24*60*60-sh.start) / times;

  while( times-- && sh.start<24*60*60 )
  {
    Schedule *sched = newSched( &sh );

    schedList = (Schedule *)realloc( schedList, sizeof(Schedule)*(schedCount+1) );
    memcpy( schedList+schedCount, sched, sizeof(Schedule) );
    schedCount++;

    delete sched;
    sh.start += period;

    if( ++schCnt%10==0 )
    {
      printf( "%5d\r", schCnt );
      fflush( stdout );
    }
  }
*/

  return 0;
}

int realCompile()
{
  log( Msg_InfoProcessing, miscInfo.schedule );

  schedCurrent = 0;

  ifstream *cfg = new ifstream( miscInfo.schedule, ios::in|ios::nocreate );

  if( !cfg->good() )
  {
    log( Msg_ErrUnableOpen, miscInfo.schedule, strError() );
    return 1;
  }

  currentCfg = cfg;
  cfgLine = 1;

  for( int rc(0), schCnt(0); 
       !cfgStack->end() && !rc;
       rc = processEvent( &cfg, schCnt ) )

  printf( "%8c\r", ' ' );
  fflush( stdout );

  closeAllConfig( cfg );

  switch( rc)
  {
  case 1:
    inputWord.reset();
    inputWord.putString( "EOF" );
    inputWord.putInto( 0 );
// fall!
  case 3:
    log( Msg_ErrCfgNoNeed, inputWord.str, cfgLine );
// fall
  default:
    return 1;

  case 255:
  case 0:
    break;
  }

  if( schedList!=NULL )
  {
    for( int i=0; i<schedCount; i++ )
      if( schedList[i].isGmt )
        schedList[i].start -= getTimeZone();

    sortSchedList( schedCount, schedList );
    saveEvents();
  }

  return 0;
}

int compileSchedule()
{
  struct stat st1, st2;
  char need( 0 );
  int rc( 0 );

  cfgFileList.registerFile( miscInfo.schedule );

  if( !stat( miscInfo.schedule, &st1 ) )
    if( stat( miscInfo.schedComp, &st2 ) )
      need = 1;
    else
      if( st1.st_mtime>st2.st_mtime )
	need = 1;

  if( need )
  {
Compile:

    if( realCompile() )
      rc = 1;
  }
  else
    if( loadEvents() )
      goto Compile;

  return rc;
}

// -------------------------- Compile schedule related

int processKeys( ifstream **cfg )
{
  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {

    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;

    Schedule sh;
    memset( &sh, 0, sizeof(Schedule) );

    int len = strlen( inputWord.str );

    for( int i(0); i<codeTableSize; i++ )
    {
      int yes;

      if( len!=1 )
        yes = !stricmp( inputWord.str, codeTable[i].name );
      else
        yes = !strcmp( inputWord.str, codeTable[i].name );

      if( yes )
      {
	sh.done = codeTable[i].code;
	break;
      }
    }

    if( !sh.done )
    {
      log( Msg_ErrCfgIllKey, inputWord.str, cfgLine );
      return 2;
    }

    rc = processAction( cfg, sh );

    if( rc )
      break;

    keyList = (Schedule *)realloc( keyList, sizeof(Schedule)*(keyCount+1) );
    memcpy( keyList+keyCount, &sh, sizeof(Schedule) );
    keyCount++;

  }

  return rc;
}

int processBothFlag( ifstream **cfg, Schedule **list, int *count )
{
  if( getWord( cfg ) )
    return 1;

  Schedule sh;
  memset( &sh, 0, sizeof(Schedule) );

  char drive[_MAX_DRIVE],
       dir  [_MAX_DIR],
       name [_MAX_FNAME],
       ext  [_MAX_EXT];

  _splitpath( inputWord.str, drive, dir, name, ext );

  if( *dir || *drive )
    sh.name = newStr( inputWord.str );
  else
    if( miscInfo.flags==NULL )
    {
      log( Msg_ErrCfgIllFlagLoc, cfgLine );
      return 2;
    }
    else
    {
      sh.name = new char [ strlen(miscInfo.flags)+strlen(inputWord.str)+1 ];
      strcpy( sh.name, miscInfo.flags );
      strcat( sh.name, inputWord.str );
    }

  int rc = processAction( cfg, sh );

  if( !rc )
  {
    *list = (Schedule *)realloc( *list, sizeof(Schedule)*((*count)+1) );
    memcpy( (*list)+(*count), &sh, sizeof(Schedule) );
    (*count)++;
  }
  
  return rc;
}

int processExternalItem( ifstream **cfg, Pair** list, int& count )
{
  if( getWord( cfg ) )
    return 1;

  char *patt = newStr( inputWord.str );
  Schedule* sh = new Schedule;
  memset( sh, 0, sizeof(Schedule) );

  int rc = processAction( cfg, *sh );

  if( !rc )
  {
    Pair pp( patt, sh );

    *list = (Pair *)realloc( *list, sizeof(Pair)*(count+1) );
    memcpy( (*list)+count, &pp, sizeof(Pair) );
    count++;
  }

  return rc;
}

int processBBSRun( ifstream **cfg )
{
  processBBS = new Schedule;
  memset( processBBS, 0, sizeof(Schedule) );

  return processAction( cfg, *processBBS );
}

int processSch( ifstream **cfg, Schedule **sh )
{
  *sh = NULL;

  Schedule *sh1 = new Schedule;
  memset( sh1, 0, sizeof(Schedule) );

  int rc = processAction( cfg, *sh1 );

  if( rc )
  {
    delete sh1;
    return rc;
  }

  *sh = sh1;
  return 0;
}

int processExitStartUp( ifstream **cfg, Schedule **list, int *count )
{
  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;

    if( stricmp( inputWord.str, "Do" ) )
      return need( "End/Do" );

    Schedule sh;
    memset( &sh, 0, sizeof(Schedule) );

    rc = processAction( cfg, sh );

    if( rc )
      break;

    *list = (Schedule *)realloc( *list, sizeof(Schedule)*((*count)+1) );
    memcpy( (*list)+(*count), &sh, sizeof(Schedule) );
    (*count)++;
  }

  return rc;
}

int processActions( ifstream **cfg )
{
  for( int rc( 0 ); !cfgStack->end() && !rc; )
  {
    if( getWord( cfg ) )
      return 1;

    if( !stricmp( inputWord.str, "End" ) )
      break;

    if( stricmp( inputWord.str, "Routine" ) )
      return need( "End/Routine" );

    Schedule sh;
    memset( &sh, 0, sizeof(Schedule) );

    if( getWord( cfg ) )
      return 1;

    sh.name = newStr( inputWord.str );

    rc = processAction( cfg, sh );

    if( rc )
      break;

    actionList = (Schedule *)realloc( actionList, sizeof(Schedule)*(actionCount+1) );
    memcpy( actionList+actionCount, &sh, sizeof(Schedule) );
    actionCount++;
  }

  return rc;
}
