#ifndef __SCHEDULE_H

#define __SCHEDULE_H

#include <iostream.h>
#include <fstream.h>
#include <string.h>

#include "fidoaddr.h"

int compileSchedule();
int loadEvents();
int saveEvents();

#pragma pack(1)
struct SchedDate
{
  short year;
  char day;
  char mon;
  char wday;
};
#pragma pack()

enum SchedType
{
  SchedTypeUnknown=0,
  SchedTypeEvent,
  SchedTypeForce
};

enum SchedAction
{
  SchedActionNothing=0,
  SchedActionPoll,
  SchedActionExit,
  SchedActionExec,
  SchedActionRun,
  SchedActionCreate,
  SchedActionSend,
  SchedActionGet,
  SchedActionHangUp,
  SchedActionLog,
  SchedActionReScan,
  SchedActionSplit,
  SchedActionDrop,
  SchedActionReset,
  SchedActionBSY,
  SchedActionRexx,
  SchedActionSemCreate,
  SchedActionSemRemove,
  SchedActionSemPost,
  SchedActionSemReset,
  SchedActionSemWait,
  SchedActionRemove,
  SchedActionCompile,
  SchedActionWait,
  SchedActionInit,
  SchedActionEval,
  SchedActionSound,
  SchedActionCall,
  SchedActionPlay,
  SchedActionCopy,
  SchedActionCatch,
  SchedActionOpen,
  SchedActionIndexFREQ,
  SchedActionSync,
  SchedActionShutdown,
  SchedActionReboot,
  SchedActionThread = 0x8000
};

enum RunOption
{
  RunOptChild     = 0x0001,
  RunOptFore      = 0x0002,
  RunOptParent    = 0x0004,
  RunOptInVisible = 0x0008,
  RunOptMax       = 0x0010,
  RunOptMin       = 0x0020,
  RunOptNoClose   = 0x0040,
  RunOptFs        = 0x0080,
  RunOptWin       = 0x0100,
  RunOptDos       = 0x0200,
  RunOptNoClip    = 0x0400,
  RunOptWait      = 0x0800,
  RunOptPort      = 0x1000,
  RunOptClose     = 0x2000,
  RunOptClear     = 0x4000
};

#pragma pack(1)
class Schedule
{
public:
  short type;                    //SchedType type;
  short action;                  //SchedAction action;
  short done;
  short dayOfWeek,
	dayOfMonth;
  unsigned long start;
  char *name;
  char *params;
  char *preloaded;
  int sndCount,
      *sound;
  void *token;
  int tokenLen;
  RunOption rOpt;
  Flavour flavour;
  FIDOAddress *adr;
  short rc;
  char *dosSetting;
  char *title;
  char *icon;
  int xPos, yPos, wPos, hPos;
  int prior;
  int isGmt;
  char isSchedule;

  Schedule *next;

  void operator = ( Schedule& src );

  friend ostream& operator << ( ostream&, Schedule& );
};
#pragma pack()

ostream& operator << ( ostream&, Schedule& );

#ifndef _SCHEDULE_MANIP
#define _SCHEDULE_MANIP
void deleteSched( Schedule *sh, int delThis=1 );
Schedule * newSched( Schedule *init );
#endif

void sortSchedList( short count, Schedule *list );

extern Schedule *schedList;

extern int schedCount,
	   schedCurrent;

extern struct SchedDate schedDate;

extern Schedule *keyList;
extern int keyCount;

extern Schedule *startUpList;
extern int startUpCount;

extern Schedule *exitList;
extern int exitCount;

extern Schedule *flagList;
extern int flagCount;

extern Schedule *noFlagList;
extern int noFlagCount;

extern Schedule *actionList;
extern int actionCount;

extern Schedule *processBBS;

int doEventKey( Schedule& sh, int current,
                       FIDOAddress *fAddr=NULL, long elapsedTime=0,
                       char *fileName=NULL);

int processSch( ifstream **cfg, Schedule **sh );
char *stringSch( Schedule *, char *, int, int =0 );

void execSchedule( Schedule *sh );

#endif
