#include <string.h>
#include <ctype.h>
#include <time.h>
#include <stdlib.h>
#include <share.h>
#include <fcntl.h>
#include <io.h>
#include <sys\stat.h>
#include <process.h>
#include <direct.h>
#include <builtin.h>

#define INCL_DOSFILEMGR
#define INCL_DOSERRORS
#define INCL_DOSPROCESS
#define INCL_DOSEXCEPTIONS
#define INCL_DOSSEMAPHORES
#include <os2.h>

#include "except.h"

#include "language.h"

#include "session.h"
#include "misc.h"
#include "buffer.h"
#include "process.h"
#include "freq.h"
#include "filetype.h"
#include "rexxhook.h"
#include "rexx.h"
#include "sched.h"
#include "cost.h"
#include "direct.h"
#include "ea.h"
#include "boxitem.h"

#include "hydra.h"
#include "zmodem.h"
#include "zinter.h"
#include "zdialog.h"
#include "fax.h"

FIDOAddress *primaryAka( NULL );

char *hisPassword = NULL,
     *mailerProductCode = NULL,
     *mailerName = NULL,
     *mailerOwner = NULL,
     *mailerVersion = NULL,
     *mailerSerialNumber = NULL,
     *hisName = NULL,
     *hisPlace = NULL,
     *hisSysOp = NULL,
     *hisPhone = NULL,
     *hisBaud = NULL,
     *hisFlags = NULL,
     *sessionCost = NULL;
int remoteMailerCode = 0;

unsigned long hisTime = 0,
              hisMOH = 0;

unsigned long ourMail = 0,
              ourFiles = 0,
              hisMail = 0,
              hisFiles = 0;

long hisTZ = 0;

static long sessionLimit = 0;
volatile long sessionStart = 0;

SessionInfo sessionBinaryInfo;

// ======================== sessionLimitExceeded ===========================

int sessionLimitExceeded()
{
  if( sessionLimit &&
      getTimeMark() >= (unsigned long)sessionStart+sessionLimit )
  {
    log( Msg_SesTimeExceeded, sessionLimit/60 );
    return 1;
  }

  return 0;
}

// =========================== CopyFileToPath ==============================
int CopyFileToPath( char *fname, char *alias, char *path )
{
  char selfCopy(0);

  {
    char *tmp( tmpStr( fname ) ),
         *tmpFname( fileName( tmp ) );

    if( tmpFname==tmp )
      tmp = ".\\";
    else
      *tmpFname=0;

    char *wrkIn( _fullpath( NULL, tmp, _MAX_PATH ) );
    char *wrkOut( _fullpath( NULL, path, _MAX_PATH ) );

    if( wrkIn!=NULL && wrkOut!=NULL && !stricmp( wrkIn, wrkOut ) )
      selfCopy = 1;

    free( wrkIn );
    free( wrkOut );
  }

  if( selfCopy )
  {
    log( Msg_SelfCopy, fname, path );
    return VERY_SPEC_COND;
  }

  char *fileName;

  if( alias==NULL )
  {
    char *file( ::fileName( fname ) );

    fileName = (char *)_alloca( strlen(file)+strlen(path)+1 );
    strcpy( fileName, path );
    strcat( fileName, file );
  }
  else
  {
    fileName = (char *)_alloca( strlen(alias)+strlen(path)+1 );
    strcpy( fileName, path );
    strcat( fileName, alias );
  }

  makeFilePath( fileName );

  char *wasName = tmpStr( fileName );
  struct stat st;

  if( !stat( fname, &st ) && st.st_size )
    while( 1 )
    {
      APIRET rc( DosCopy( fname, fileName, 0 ) );
    
      switch( rc )
      {
      case 0:
        {
          char *sz = strBig( st.st_size );
          log( Msg_LogFileCopied, fname, sz, fileName );
          delete sz;
        }
        return OK;

      case ERROR_ACCESS_DENIED:
        addOneToBigNum( fileName );

        if( !stricmp( fileName, wasName ) )
        {
          log( Msg_ErrUnableCopy1, fname, path );
          return VERY_SPEC_COND;
        }

        continue;

      default:
        log( Msg_ErrUnableCopy, fname, fileName, rc );
        os2Message( rc );
        return VERY_SPEC_COND;
      }
    }
  
  return OK;
}

// ============================ RealSendFile ===============================

int RealSendFile( char *fname, char *alias, int& fsent, int wazoo, int trunc,
                  FILE *xfer, int del, int isFreq, int isOnline, char **msg,
                  Compatibility hisCode, char *password,
                  Flavour flavour, char *outboundType, QueueEntry& aka,
                  int answer )
{
  if( hisCode!=COPY && sessionLimitExceeded() )
    return ERROR;

  int rc;
  long wasBytes = zOutBytes;
  FileType type;
  char *who;

  if( fname!=NULL )
  {
    if( hisCode!=COPY && msg!=NULL && *msg!=NULL )
    {
      log( *msg );
      *msg = NULL;
    }

    who = msgOfFileType( alias==NULL ? fname : alias, &type );

    if( getRexxHook( "SendThisFile" ) != NULL )
    {
      char *flName = realFlavour( flavour );
      char buf[128],
           *hookIni = processRexxHook( "SendThisFile",
                                       fname, flName, who,
                                       outboundType,
                                       aka.toString( buf, sizeof(buf)-1,
                                                     FIDOAddressShowModeEMSI ),
                                       answer ? "IN" : "OUT",
                                       1 ),
           *hook = tmpStr( hookIni );

      delete hookIni;

      if( hook!=NULL && stricmp( hook, "YourSelf" ) )
        if( !stricmp( hook, "Yes" ) )
          goto PleaseSend;
        else if( !stricmp( hook, "No" ) )
        {
          log( Msg_SesHookNo, fname, who );
          return VERY_SPEC_COND;
        }
        else
        {
          log( Msg_SesHookIll, hook, fname, who );
          return VERY_SPEC_COND;
        }
    }

    if( type!=FileMail && (hisCode&HXT)==HXT )
    {
      log( Msg_SesNoSendHXT, fname, who );
      return VERY_SPEC_COND;
    }

    if( hisCode!=COPY && miscInfo.zmh!=NULL && miscInfo.zmh->isWorked( time( NULL ) ) )
    {
      if( type!=FileMail )
      {
        log( Msg_SesNoSendZMH, fname, who );
        return VERY_SPEC_COND;
      }

      if( isFreq )
      {
        log( Msg_SesNoSendOnlineZMH );
        generateFreqAnswer( Msg_SesNoSendOnlineZMHAns, *primaryAka, hisSysOp,
                            password, NULL );
        return VERY_SPEC_COND;
      }

      if( isOnline )
      {
        log( Msg_SesNoSendFREQZMH );
        generateFreqAnswer( Msg_SesNoSendFREQZMHAns, *primaryAka, hisSysOp,
                            password, NULL );
        return VERY_SPEC_COND;
      }
    }

    struct stat st;

    if( stat( fname, &st ) )
    {
      log( Msg_ErrUnableFind, fname, strError() );
      return SPEC_COND;
    }

    if( !st.st_size )
    {
      log( Msg_SesNoSendEmpty, fname, who );
      return OK;
    }
  }

PleaseSend:
  if( hisCode==COPY )
    if( password==NULL )
    {
      log( "*BUG* sendFile: path is <NULL>" );
      rc = ERROR;
    }
    else if( fname==NULL )
    {
      log( "*BUG* sendFile: fname is <NULL>" );
      rc = ERROR;
    }
    else
      rc = CopyFileToPath( fname, alias, password );
  else if( wazoo!=HydraSes )
    rc = Send_Zmodem( fname, alias, fsent, wazoo, trunc, xfer );
  else
  {
    rc = Hydra( fname, alias );

    switch( rc )
    {
    case XFER_OK:
      rc = OK;
      break;

    case XFER_SKIP:
      rc = VERY_SPEC_COND;
      break;

    default:
      rc = ERROR;
      break;
    }
  }

  if( fname!=NULL )
  {
    switch( type )
    {
    case FileMail:
      sessionBinaryInfo.netSend += zOutBytes-wasBytes;
      break;

    case FileXMail:
      sessionBinaryInfo.xmSend += zOutBytes-wasBytes;
      break;

    default:
      sessionBinaryInfo.filesSend += zOutBytes-wasBytes;
      break;
    }

    if( del )
      switch( rc )
      {
      case OK:
        chmod( fname, S_IREAD|S_IWRITE );

        if( unlink( fname ) )
         log( Msg_ErrUnableDelete, fname, strError() );
        break;

      case VERY_SPEC_COND:
        rc = OK;
        break;

      default:
        rc = ERROR;
      }
  }

  fsent++;
  return rc;
}

// ============================= SendFile =================================

char *sentFileResult( int rc )
{
  switch( rc )
  {
  case SPEC_COND:      return "SKIP";
  case OK:             return "OK";
  case VERY_SPEC_COND: return "DELAY";
  }

  return "FAIL";
}

int SendFile( char *fname, char *alias, int& fsent, int wazoo, int trunc,
              FILE *xfer, int del, int isFreq, int isOnline, char **msg,
              Compatibility hisCode, char *password,
              Flavour flavour, char *outboundType, QueueEntry& aka,
              int answer )
{
  int rc( RealSendFile( fname, alias, fsent, wazoo, trunc,
                        xfer, del, isFreq, isOnline, msg,
                        hisCode, password, flavour, outboundType, 
                        aka, answer ) );
              
  if( fname != NULL && getRexxHook( "FileSent" ) != NULL )
    delete processRexxHook( "FileSent", fname, sentFileResult( rc ) );
  
  return rc;
}

// ============================== sendOut =================================

int sendBothOut( char *bink, char *flavour,
                 int wazoo, int& fsent, char **msg, Compatibility hisCode,
                 FILE *xfer, char *password, Flavour flavourType,
                 QueueEntry& aka, int answer )
{
  if( !bink )
    return OK;

  int zret( OK );
  char *path = (char *)_alloca( strlen(bink)+strlen(flavour)+2 );

  strcpy( path, bink );
  strcat( path, "." );
  strcat( path, flavour );
  makeFileCase( path );

//  log( "?Session out: %s", path );

  if( !access( path, 0 ) )
  {
    char alias[20];

    sprintf( alias, "%08lx.pkt", getTimeMark() );
    makeFileCase( alias );

    if( flavourType==onlineFlavour )
      log( Msg_SesSendOnlineRep );
    else if( flavourType==requestFlavour )
      log( Msg_SesSendFREQRep );

    zret = SendFile( path, alias, fsent, wazoo, (hisCode&FNC)==FNC, xfer,
                     1, 0, 0, msg, hisCode, password,
                     flavourType, "Outbound", aka, answer );
  }

  return zret;
}

int sendShortOut( char *bink, char flavour,
                  int wazoo, int& fsent, char **msg, Compatibility hisCode,
                  FILE *xfer, char *password, Flavour flavourType,
                  QueueEntry& aka, int answer )
{
  static char ext[]( "?ut" );

  ext[0] = flavour;
  return sendBothOut( bink, ext, wazoo, fsent, msg, hisCode, xfer, password,
                      flavourType, aka, answer );
}

int sendLongOut( char *bink, char *flavour,
                 int wazoo, int& fsent, char **msg, Compatibility hisCode,
                 FILE *xfer, char *password, Flavour flavourType,
                 QueueEntry& aka, int answer )
{
  static char ext[]( ".mail" );

  char *path = (char *)_alloca( strlen( flavour ) + sizeof( ext ) );

  strcpy( path, flavour );
  strcat( path, ext );

  return sendBothOut( bink, path, wazoo, fsent, msg, hisCode, xfer, password,
                      flavourType, aka, answer );
}


// ============================== sendTM ==================================

int sendTM( char *tmIni, char flavour, int wazoo,
	    int& fsent,  Compatibility hisCode, char **msg,
	    int deleteBox, FILE *xfer, char *password,
            Flavour flavourType, char *boxType, QueueEntry& aka, int answer )
{
  if( !tmIni )
    return OK;

  int zret( OK );

  char *tm = tmpStr( tmIni ),
       mask[]( "\\*" ),
       *fullBox = (char *)_alloca( strlen(tmIni)+3 );

  if( tm[ strlen(tm)-1 ]=='\\' )
    tm[ strlen(tm)-1 ] = 0;

  char *path = (char *)_alloca( strlen(tm)+_MAX_PATH );

  int len( strlen(tm) );

  strcpy( fullBox, tm );
  fullBox[ len ] = flavour;
  fullBox[ len+1 ] = 0;

  len = strlen( tm );
  strcpy( path, tm );
  path[ len ] = flavour;
  path[ len+1 ] = 0;
  strcat( path, mask );
  len = strlen(path) - sizeof(mask) + 2;

  BoxItemList fileList;

  for( Directory rc( path ); rc; rc++ )
  {
    strcpy( path+len, rc.getName() );
    makeFileCase( path );

    if( !rc.getSize() )
      continue;

    FileType type;
    Priority prio;

    msgOfFileType( path, &type );

    switch( type )
    {
    case FileMail:
      prio = NetMail;
      break;

    case FileXMail:
      prio = EchoMail;
      break;

    case FileReq:
      prio = Request;
      break;

    default:
      prio = File;
      break;
    }

    fileList.Add( *new BoxItem( path, prio, rc.getUNIXCreate() ) );
  }

  if( fileList.GetItemsInContainer() )
  {
    fileList.sort();

    for( BoxItemIterator list( fileList ); zret!=ERROR && list; list++ )
      zret = SendFile( list().getName(), NULL, fsent, wazoo, (hisCode&FNC)==FNC,
                       xfer,
                       1, 0, 0, msg, hisCode, password,
                       flavourType, boxType, aka, answer );
  }

  if( deleteBox )
    rmdir( fullBox );

  return zret;
}

int sendLongBox( char *tmIni, char *flavour, int wazoo,
                 int& fsent,  Compatibility hisCode, char **msg,
                 int deleteBox, FILE *xfer, char *password,
                 Flavour flavourType, QueueEntry& aka, int answer )
{
  if( !tmIni )
    return OK;

  char *tm = (char *)_alloca( strlen(tmIni)+strlen(flavour)+2 );

  strcpy( tm, tmIni );
  strcat( tm, "." );
  strcat( tm, flavour );

  int rc = sendTM( tm, 0, wazoo, fsent, hisCode, msg, deleteBox, xfer, password,
                   flavourType, "Long-Box", aka, answer );

  return rc;
}

// ============================= sendFromLO ================================

struct SendFromLOArgs
{
  char first;
  int wazoo;
  Compatibility hisCode;
  FILE *xfer;
  char **msg;
  char *password;
  Flavour flavourType;
  QueueEntry& aka;
  int answer;
  int& fsent;
};

int sendFromLO( char *name, char, void *args )
{
  SendFromLOArgs *p( (SendFromLOArgs *)args );

  char isOnline( char( p->flavourType==onlineFlavour ) ),
       isFreq( char( p->flavourType==requestFlavour ) );

  if( p->first )
  {
    p->first = 0;

    if( isOnline )
      log( Msg_SesSendOnline );

    if( isFreq )
      log( Msg_SesSendFREQ );
  }

  long FreqTransferTime = getTimeMark();
  long FreqTransferSize = zOutBytes;

//    -     FREQed 

  int zret( SendFile( name, NULL, p->fsent, p->wazoo,
                      (p->hisCode&FNC)==FNC, p->xfer,
                      0, isFreq, isOnline, p->msg,
                      p->hisCode, p->password,
                      p->flavourType, "Outbound",
                      p->aka, p->answer ) );

  if( isFreq && zret!=SPEC_COND )
    if( freqBaseUpDate( getTimeMark()-FreqTransferTime,
                        zOutBytes-FreqTransferSize,
                        name, p->password ) )
      zret = ERROR;

  return zret;
}

// =============================== sendLO ==================================

int sendBothLO( char *bink, char *flavour, int wazoo,
                int& fsent, Compatibility hisCode, char **msg,
                char *password, FILE *xfer, Flavour flavourType,
                QueueEntry& aka, int answer )
{
  if( !bink )
    return OK;

  char *path = (char *)_alloca( strlen( bink )+strlen( flavour )+2 );

  strcpy( path, bink );
  strcat( path, "." );
  strcat( path, flavour );
  capitalize( path );

//  log( "?Session flo: %s", path );

  FLO xlo( path );
  SendFromLOArgs args =
  {
    1,
    wazoo, hisCode, xfer, msg, password,
    flavourType, aka, answer,
    fsent
  };

  return xlo.process( 1, sendFromLO, &args );
}

int sendShortLO( char *bink, char flavour, int wazoo,
                 int& fsent, Compatibility hisCode, char **msg,
                 char *password, FILE *xfer, Flavour flavourType,
                 QueueEntry& aka, int answer )
{
  static char ext[]( "?lo" );

  ext[0]=flavour;
  return sendBothLO( bink, ext, wazoo, fsent, hisCode, msg, password,
                     xfer, flavourType, aka, answer );
}

int sendLongLO( char *bink, char *flavour, int wazoo,
                int& fsent, Compatibility hisCode, char **msg,
                char *password, FILE *xfer, Flavour flavourType,
                QueueEntry& aka, int answer )
{
  static char ext[]( ".list" );

  char *path = (char *)_alloca( strlen( flavour ) + sizeof( ext ) );

  strcpy( path, flavour );
  strcat( path, ext );

  return sendBothLO( bink, path, wazoo, fsent, hisCode, msg, password,
                     xfer, flavourType, aka, answer );
}


// ============================ sendRequest ===============================

static char freqWasSent = 0;

int sendBothRequest( char *suf, char *bink,
                     QueueEntry& aka, int wazoo, int& fsent,
                     Compatibility hisCode, char **msg,
                     char *password, int answer, FILE *xfer )
{
  if( !bink )
    return OK;

  char *req = (char *)_alloca( strlen(bink)+strlen(suf)+1 );

  strcpy( req, bink );
  strcat( req, suf );
  makeFileCase( req );

//  log( "?Session req: %s", req );

  int q( OK );

  char *alias( (char *)_alloca( 16 ) );
  int net,
      node;

  if( aka.is4D() )
  {
    net = 0;
    node = aka.getPoint();
  }
  else
  {
    net = aka.getNet();
    node = aka.getNode();
  }

  sprintf( alias, "%04x%04x.req", net, node );

  if( !access( req, 0 ) )
    if( (hisCode&NRQ)==NRQ )
      log( Msg_SesFREQNotSupp );
    else if( (hisCode&HRQ)==HRQ )
      log( Msg_SesFREQNotAtThis );
    else if( miscInfo.zmh!=NULL && miscInfo.zmh->isWorked( time( NULL ) ) )
      log( Msg_SesFREQNotZMH );
    else
    {
      q = SendFile( req, alias, fsent, wazoo, (hisCode&FNC)==FNC, xfer,
                    0, 0, 0, msg, hisCode, password,
                    requestFlavour, "Outbound", aka, answer );
      freqWasSent = 1;
    }

  return q;
}

int sendShortRequest( char *bink, QueueEntry& aka, int wazoo, int& fsent,
                      Compatibility hisCode, char **msg,
                      char *password, int answer, FILE *xfer )
{
  return sendBothRequest( ".req", bink, aka, wazoo, fsent, hisCode, msg,
                          password, answer, xfer );
}

int sendLongRequest( char *bink, QueueEntry& aka, int wazoo, int& fsent,
                     Compatibility hisCode, char **msg,
                     char *password, int answer, FILE *xfer )
{
  return sendBothRequest( ".request", bink, aka, wazoo, fsent, hisCode, msg,
                          password, answer, xfer );
}

// ============================== sendAka ==================================

int sendAka( QueueEntry& aka, int wazoo, int& fsent,
             Compatibility hisCode, char **msg,
             char *password, int answer, FILE *xfer, int needSendFreq )
{
  int zret( OK );

  char *bink( aka.getBinkName() ),
       *tmIni( aka.getTBoxName() ),
       *tm( NULL ),
       *longBox( NULL );

  char needSendHold( char( answer ||
                           (!answer && aka.getCfgFlavour()!=unknownFlavour
                                    && aka.getCfgFlavour()!=holdFlavour ) ||
                           (!answer && miscInfo.state1&HoldOnCall ) ) );

  if( miscInfo.tbox!=NULL && tmIni!=NULL )
  {
    tm = (char *)_alloca( strlen(miscInfo.tbox)+strlen(tmIni)+1 );
    strcpy( tm, miscInfo.tbox );
    strcat( tm, tmIni );
  }

  delete tmIni;

  if( miscInfo.longBox!=NULL )
  {
    char *i( aka.getLongName() );

    longBox = (char *)_alloca( strlen(miscInfo.longBox)+strlen(i)+1 );
    strcpy( longBox, miscInfo.longBox );
    strcat( longBox, i );
    delete i;
  }

  char requestFirst( char( hisCode&RH1 && wazoo==HydraSes ? 1 : 0 ) );
  char *longBink( aka.getLongBinkName() );

  if( hisCode!=COPY && illegalAka!=NULL )
  {
    if( sendLongOut( longBink, flavourString( requestFlavour ),
                     wazoo, fsent, msg, hisCode,
                     xfer, password,
                     immedFlavour, aka, answer )==ERROR )
      goto Error;

    if( sendShortOut( bink, 'R', wazoo, fsent, msg, hisCode, xfer, password,
                      immedFlavour, aka, answer )==ERROR )
      goto Error;

    goto OkComplete;
  }

  if( hisCode!=COPY && requestFirst && needSendFreq )
  {
    if( sendLongRequest( longBink, aka, wazoo, fsent, hisCode,
                         msg, password,
                         answer, xfer )==ERROR )
      goto Error;

    if( sendShortRequest( bink, aka, wazoo, fsent, hisCode, msg, password,
                          answer, xfer )==ERROR )
      goto Error;

    goto OkComplete;
  }

// long Mail

  if( sendLongOut( longBink, flavourString( immedFlavour ),
                   wazoo, fsent, msg,
                   hisCode, xfer, password,
                   immedFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendLongOut( longBink, flavourString( crashFlavour ),
                   wazoo, fsent, msg,
                   hisCode, xfer, password,
                   crashFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendLongOut( longBink, flavourString( directFlavour ),
                   wazoo, fsent, msg,
                   hisCode, xfer, password,
                   directFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendLongOut( longBink, flavourString( normalFlavour ),
                   wazoo, fsent, msg,
                   hisCode, xfer, password,
                   normalFlavour, aka, answer )==ERROR )
    goto Error;

  if( hisCode==COPY || needSendHold )
    if( sendLongOut( longBink, flavourString( holdFlavour ),
                     wazoo, fsent, msg,
                     hisCode, xfer, password,
                     holdFlavour, aka, answer )==ERROR )
      goto Error;

// short Mail

  if( sendShortOut( bink, 'I', wazoo, fsent, msg, hisCode, xfer, password,
                    immedFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendShortOut( bink, 'C', wazoo, fsent, msg, hisCode, xfer, password,
                    crashFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendShortOut( bink, 'D', wazoo, fsent, msg, hisCode, xfer, password,
                    directFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendShortOut( bink, 'O', wazoo, fsent, msg, hisCode, xfer, password,
                    normalFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendShortOut( bink, 'N', wazoo, fsent, msg, hisCode, xfer, password,
                    normalFlavour, aka, answer )==ERROR )
    goto Error;

  if( hisCode==COPY || needSendHold )
    if( sendShortOut( bink, 'H', wazoo, fsent, msg, hisCode, xfer, password,
                      holdFlavour, aka, answer )==ERROR )
      goto Error;

// Immediate

  if( sendTM( aka.getImmedBox(), 0, wazoo, fsent, hisCode, msg, 0, xfer, password,
              immedFlavour, "Cfg-Box", aka, answer )==ERROR )
    goto Error;

  if( sendTM( tm, 'I', wazoo, fsent, hisCode, msg, 1, xfer, password,
              immedFlavour, "T-Box", aka, answer )==ERROR )
    goto Error;

  if( sendLongBox( longBox, flavourString( immedFlavour ),
                   wazoo, fsent, hisCode, msg, 1, xfer, password,
                   immedFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendLongLO( longBink, flavourString( immedFlavour ),
                  wazoo, fsent, hisCode,
                  msg, password, xfer,
                  immedFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendShortLO( bink, 'I', wazoo, fsent, hisCode, msg, password, xfer,
                   immedFlavour, aka, answer )==ERROR )
    goto Error;

// Crash

  if( sendTM( aka.getCrashBox(), 0, wazoo, fsent, hisCode, msg, 0, xfer, password,
              crashFlavour, "Cfg-Box", aka, answer )==ERROR )
    goto Error;

  if( sendTM( tm, 'C', wazoo, fsent, hisCode, msg, 1, xfer, password,
              crashFlavour, "T-Box", aka, answer )==ERROR )
    goto Error;

  if( sendLongBox( longBox, flavourString( crashFlavour ),
                   wazoo, fsent, hisCode, msg, 1, xfer, password,
                   crashFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendLongLO( longBink, flavourString( crashFlavour ),
                  wazoo, fsent, hisCode,
                  msg, password, xfer,
                  crashFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendShortLO( bink, 'C', wazoo, fsent, hisCode, msg, password, xfer,
                   crashFlavour, aka, answer )==ERROR )
    goto Error;

// Direct

  if( sendTM( aka.getDirectBox(), 0, wazoo, fsent, hisCode, msg, 0, xfer, password,
              directFlavour, "Cfg-Box", aka, answer )==ERROR )
    goto Error;

  if( sendTM( tm, 'D', wazoo, fsent, hisCode, msg, 1, xfer, password,
              directFlavour, "T-Box", aka, answer )==ERROR )
    goto Error;

  if( sendLongBox( longBox, flavourString( directFlavour ),
                   wazoo, fsent, hisCode, msg, 1, xfer, password,
                   directFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendLongLO( longBink, flavourString( directFlavour ),
                  wazoo, fsent, hisCode,
                  msg, password, xfer,
                  directFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendShortLO( bink, 'D', wazoo, fsent, hisCode, msg, password, xfer,
                   directFlavour, aka, answer )==ERROR )
    goto Error;

// Normal

  if( sendTM( aka.getNormalBox(), 0, wazoo, fsent, hisCode, msg, 0, xfer, password,
              normalFlavour, "Cfg-Box", aka, answer )==ERROR )
    goto Error;

  if( sendTM( tm, 0, wazoo, fsent, hisCode, msg, 1, xfer, password,
              normalFlavour, "T-Box", aka, answer )==ERROR )
    goto Error;

  if( sendTM( tm, 'N', wazoo, fsent, hisCode, msg, 1, xfer, password,
              normalFlavour, "T-Box", aka, answer )==ERROR )
    goto Error;

  if( sendTM( tm, 'F', wazoo, fsent, hisCode, msg, 1, xfer, password,
              normalFlavour, "T-Box", aka, answer )==ERROR )
    goto Error;

  if( sendLongBox( longBox, flavourString( normalFlavour ),
                   wazoo, fsent, hisCode, msg, 1, xfer, password,
                   normalFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendLongLO( longBink, flavourString( normalFlavour ),
                  wazoo, fsent, hisCode,
                  msg, password, xfer,
                  normalFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendShortLO( bink, 'F', wazoo, fsent, hisCode, msg, password, xfer,
                   normalFlavour, aka, answer )==ERROR )
    goto Error;

  if( sendShortLO( bink, 'N', wazoo, fsent, hisCode, msg, password, xfer,
                   normalFlavour, aka, answer )==ERROR )
    goto Error;

// FREQ

  if( hisCode!=COPY && !requestFirst && needSendFreq )
  {
    if( sendLongRequest( longBink, aka, wazoo, fsent, hisCode, msg, password,
                         answer, xfer )==ERROR )
      goto Error;

    if( sendShortRequest( bink, aka, wazoo, fsent, hisCode, msg, password,
                          answer, xfer )==ERROR )
      goto Error;
  }

// Hold

  if( hisCode==COPY || needSendHold )
  {
    if( sendTM( aka.getHoldBox(), 0, wazoo, fsent, hisCode, msg, 0, xfer, password,
                holdFlavour, "Cfg-Box", aka, answer )==ERROR )
      goto Error;

    if( sendTM( tm, 'H', wazoo, fsent, hisCode, msg, 1, xfer, password,
                holdFlavour, "T-Box", aka, answer )==ERROR )
      goto Error;

    if( sendLongBox( longBox, flavourString( holdFlavour ),
                     wazoo, fsent, hisCode, msg, 1, xfer, password,
                     holdFlavour, aka, answer )==ERROR )
      goto Error;

    if( sendLongLO( longBink, flavourString( holdFlavour ),
                    wazoo, fsent, hisCode,
                    msg, password, xfer,
                    holdFlavour, aka, answer )==ERROR )
      goto Error;

    if( sendShortLO( bink, 'H', wazoo, fsent, hisCode, msg, password, xfer,
                     holdFlavour, aka, answer )==ERROR )
      goto Error;
  }

  if( hisCode!=COPY )
  {

// ONLINE request

    if( sendLongOut( longBink, flavourString( onlineFlavour ),
                     wazoo, fsent, msg, hisCode,
                     xfer, password,
                     onlineFlavour, aka, answer )==ERROR )
      goto Error;

    if( sendLongLO( longBink, flavourString( onlineFlavour ),
                    wazoo, fsent, hisCode,
                    msg, password, xfer,
                    onlineFlavour, aka, answer )==ERROR )
      goto Error;

    if( sendShortOut( bink, 'Q', wazoo, fsent, msg, hisCode, xfer, password,
                      onlineFlavour, aka, answer )==ERROR )
      goto Error;

    if( sendShortLO( bink, 'Q', wazoo, fsent, hisCode, msg, password, xfer,
                     onlineFlavour, aka, answer )==ERROR )
      goto Error;

// FREQed files

    waitFreqProcessor();

    if( sendLongLO( longBink, flavourString( requestFlavour ),
                    wazoo, fsent, hisCode,
                    msg, password, xfer,
                    requestFlavour, aka, answer )==ERROR )
      goto Error;

    if( sendLongOut( longBink, flavourString( requestFlavour ),
                     wazoo, fsent, msg, hisCode,
                     xfer, password,
                     requestFlavour, aka, answer )==ERROR )
      goto Error;

    if( sendShortLO( bink, 'R', wazoo, fsent, hisCode, msg, password, xfer,
                     requestFlavour, aka, answer )==ERROR )
      goto Error;

    if( sendShortOut( bink, 'R', wazoo, fsent, msg, hisCode, xfer, password,
                      requestFlavour, aka, answer )==ERROR )
      goto Error;
  }

  goto OkComplete;

Error:
  zret = ERROR;

OkComplete:
  delete longBink;
  delete bink;

  return zret;
}

// ============================= sendBatch =================================

int sendBatch( QueueEntryList *akaList, int wazoo,
               Compatibility hisCode, char *password,
               int answer, FILE *xfer, int needSendFreq,
               int needNothingFor, int sendOnce )
{
  int fsent( 0 ),
      zret( OK ),
      j( 0 );

  if( wazoo!=HydraSes )
  {
    infoDisplayStart();
    zSendInitBlock = 0;
    zTransStart = getTimeMark();
  }
  
  int sendCount( sendOnce ? 1 : miscInfo.sendTry );

  if( (hisCode&NPU)==NPU )
    log( Msg_SesNoSendNPU );
  else if( (hisCode&HAT)!=HAT )
    for( int sendTry(0); sendTry<sendCount && zret!=ERROR; sendTry++ )
    {
//      log( "?Session, try #%d", sendTry );

      for( QueueEntryIterator list( *akaList ); list && zret!=ERROR; list++ )
      {
        if( (hisCode&PUP)==PUP && ++j!=1 )
        {
          log( Msg_SesNoSendPUP );
          break;
        }

        QueueEntry& cur( list() );

        int oldFsent( fsent );
        char *msg( NULL );
        char buf[65];
        cur.toString( buf, sizeof buf );

        if( akaList->GetItemsInContainer()>1 )
        {
          msg = new char [ strlen(Msg_SesSendFor)+strlen(buf)+1 ];
          strcpy( msg, Msg_SesSendFor );
          strcat( msg, buf );
        }

        zret = sendAka( cur, wazoo, fsent, hisCode, &msg, password, answer,
                        xfer, sendTry ? 0 : needSendFreq );

        delete msg;

        if( !sendTry && fsent==oldFsent )
          if( wazoo!=HydraSes ||
              (wazoo==HydraSes && needNothingFor) )
            log( Msg_SesNothingFor, buf );

        if( illegalAka!=NULL )
          break;
      }
    }

  if( zret != ERROR )
  {
    fsent = fsent ? END_BATCH : NOTHING_TO_DO;
    zret = SendFile( NULL, NULL, fsent,
                     wazoo, (hisCode&FNC)==FNC, NULL, 0, 0, 0, NULL,
                     hisCode, password, unknownFlavour, "?", (*akaList)[0],
                     answer );
  }

  if( wazoo!=HydraSes )
    infoDisplayEnd();

  return zret!=ERROR ? OK : ERROR;
}

// ============================ deleteQlo ==============================

int deleteFromQLO( char *, char, void * )
{
  return OK;
}

void deleteQlo( char *inList )
{
  FLO *xlo( new FLO( inList ) );

  xlo->process( 1, deleteFromQLO, 0 );
  delete xlo;

  if( unlink( inList ) )
    log( Msg_ErrUnableDelete, inList, strError() );
}

// ============================ removeFreqBsy ==============================

void removeFreqBsy( QueueEntryList *akaList, int rc )
{
  char inList[ _MAX_PATH*2 ];

  for( QueueEntryIterator list( *akaList ); list; list++ )
  {
    QueueEntry& cur( list() );

    if( rc==OK )
      cur.dropFreq();

    char *path( cur.getBinkName() ),
         *pathl( cur.getLongBinkName() );

    if( pathl!=NULL )
    {
// long online

      strcpy( inList, pathl );
      strcat( inList, ".online.list" );

      capitalize( inList );

      if( !access( inList, 0 ) )
	deleteQlo( inList );

      strcpy( inList, pathl );
      strcat( inList, ".online.mail" );

      capitalize( inList );

      if( !access( inList, 0 ) )
	if( unlink( inList ) )
	  log( Msg_ErrUnableDelete, inList, strError() );

// long request

      strcpy( inList, pathl );
      strcat( inList, ".request.list" );

      capitalize( inList );

      if( !access( inList, 0 ) )
	if( unlink( inList ) )
	  log( Msg_ErrUnableDelete, inList, strError() );

      strcpy( inList, pathl );
      strcat( inList, ".request.mail" );

      capitalize( inList );

      if( !access( inList, 0 ) )
	if( unlink( inList ) )
	  log( Msg_ErrUnableDelete, inList, strError() );

      delete pathl;
    }

    if( path!=NULL )
    {
// short online

      strcpy( inList, path );
      strcat( inList, ".qlo" );

      capitalize( inList );

      if( !access( inList, 0 ) )
	deleteQlo( inList );

      strcpy( inList, path );
      strcat( inList, ".qut" );

      capitalize( inList );

      if( !access( inList, 0 ) )
	if( unlink( inList ) )
	  log( Msg_ErrUnableDelete, inList, strError() );

// short request

      strcpy( inList, path );
      strcat( inList, ".rut" );

      capitalize( inList );

      if( !access( inList, 0 ) )
	if( unlink( inList ) )
	  log( Msg_ErrUnableDelete, inList, strError() );

      strcpy( inList, path );
      strcat( inList, ".rlo" );

      capitalize( inList );

      if( !access( inList, 0 ) )
	if( unlink( inList ) )
	  log( Msg_ErrUnableDelete, inList, strError() );

      delete path;
    }

    cur.makeBsy( 0 );
  }
}

// ============================= doSession =================================

int doSession( Compatibility hisCode,
               QueueEntryList *akaList, int answer,
  	       QueueEntry *called,
	       Compatibility myCode,
	       char *password,
               int& needRescan )
{
  char buf[ 128 ];
  int delta = (miscInfo.state1&StatusLine)==StatusLine ? 0 : 1;

  needRescan = 0;

  primaryAka->toString( buf, sizeof(buf) );
  setTitleMode( called==NULL ? Msg_TtlIn : Msg_TtlOut );
  addTitleMode( buf );
  changeTitle( NULL );
  execSchedule( melodies.begSess );
  showAction( Msg_ScrSession );

  wndQueue = brdColor;

  char *scrOurMail = strNumber( ourMail ),
       *scrOurFiles = strNumber( ourFiles );
  char *scrHisMail = strNumber( hisMail ),
       *scrHisFiles = strNumber( hisFiles );

  wndQueue( 15, 10+delta ) << scrOurMail;
  wndQueue( 25, 10+delta ) << scrOurFiles;
  wndQueue( 15,  9+delta ) << scrHisMail;
  wndQueue( 25,  9+delta ) << scrHisFiles;

  delete scrOurMail;
  delete scrOurFiles;
  delete scrHisMail;
  delete scrHisFiles;

  if( called!=NULL )
  {
    if( (hisCode&HAT)==HAT )
    {
      log( Msg_InfoDetectedHAT );
      called->makeHold( miscInfo.holdTime );
      called->setHoldTo( time( NULL )+miscInfo.holdTime );
    }

    char zmh( 0 );

    if( miscInfo.zmh!=NULL && miscInfo.zmh->isWorked( time( NULL ) ) )
      zmh = 1;

    if( (hisCode&HXT)==HXT && !zmh )
    {
      log( Msg_InfoDetectedHXT );
      called->makeHold( miscInfo.holdTime );
      called->setHoldTo( time( NULL )+miscInfo.holdTime );
    }
  }

  int wazoo( 0 );

  if( (hisCode&HYD)==HYD && (miscInfo.state2&HydraAvail)==HydraAvail )
    wazoo = HydraSes+1;
  else if( (hisCode&DZA)==DZA && comPort->getHardWareMode() &&
           (miscInfo.state2&DirZapAvail)==DirZapAvail )
    wazoo = DirZap+1;
  else if( (hisCode&ZAP)==ZAP && (miscInfo.state2&ZedZapAvail)==ZedZapAvail )
    wazoo = ZedZap+1;
  else if( (hisCode&ZMO)==ZMO )
    wazoo = Zmodem+1;

  if( !wazoo )
  {
    log( Msg_SesNoProtocols );
    execSchedule( melodies.fail );
    return ERROR;
  }

  if( wazoo!=DirZap+2 )
    hisCode = Compatibility( hisCode & (~RH1) );

  --wazoo;

  log( Msg_SesProtocol, isProtocol( wazoo ),
                        hisCode&RH1 && wazoo==HydraSes ? Msg_HydraInRH1 : "" );

  char *inbound( NULL );
  int maxInComing(0);

  if( called==NULL )
    for( QueueEntryIterator qIter( *akaList ); qIter && inbound==NULL; qIter++ )
    {
      QueueEntry& cur = qIter();
      NodeConfig test( (FIDOAddress *)&cur ),
                 *found( systemList->Find( test ) );

      if( found!=NULL )
      {
        Domain *dom = found->getDomainPtr();

        inbound = found->getInbound();

        if( inbound==NULL && dom!=NULL )
          if( sessionProtected )
            inbound = dom->getProtected();
          else
            inbound = dom->getInbound();

        if( !maxInComing )
          maxInComing = found->getMaxInComing();
      }
    }
  else
  {
    Domain *dom = called->getDomainPtr();

    inbound = called->getInbound();

    if( inbound==NULL && dom!=NULL )
      if( sessionProtected )
        inbound = dom->getProtected();
      else
        inbound = dom->getInbound();

    maxInComing = called->getMaxInComing();
  }

  if( inbound==NULL && sessionProtected )
    inbound = miscInfo.protectedInbound;

  if( inbound==NULL )
    inbound = miscInfo.inbound;

  if( called==NULL )
    if( maxInComing )
      sessionLimit = maxInComing;
    else
      sessionLimit = miscInfo.maxIn;
  else
    sessionLimit = called->getMaxOutGoing();

  sessionLimit *= 60;

  ReceivedFileList inFiles;
  int rc( OK );

  wndQueue = brdColor;

  initFreqProcessor();
  freqBaseInfoClear();

  char *proc = getRexxHook( "BeforeSession" ),
       addr[ 128 ],
       *sesType = called==NULL ? "INCOMING" : "OUTGOING";

  primaryAka->toString( addr, sizeof(addr)-1, FIDOAddressShowModeEMSI );

  if( proc!=NULL )
  {
    showAction( fileName( proc ) );
    processRexxHook( "BeforeSession", sesType, addr );
  }

  FILE *xfer = NULL;

  if( miscInfo.sentList!=NULL )
  {
    makeFilePath( miscInfo.sentList );
    xfer = _fsopen( miscInfo.sentList, "a" );

    if( xfer==NULL )
      log( Msg_ErrUnableOpen, miscInfo.sentList, strError() );
  }

  freqThrInfo.myCode =  myCode;
  freqThrInfo.password = newStr( password );

  TimeOutList *searchCpsList =
                  wazoo==HydraSes ? ( called==NULL ? hMinInCps : hMinOutCps )
                                  : ( called==NULL ? zMinInCps : zMinOutCps );

  long baudForCpsLim( comPort->getLastBaud() );

  if( searchCpsList!=NULL )
  {
    TimeOut test( baudForCpsLim, 0 );
    register TimeOut *found( searchCpsList->Find( test ) );

    if( found!=NULL )
      sesCpsMinimum = found->getTimeOut();
  }
  else
    sesCpsMinimum = 0;

  log( "?sesCpsMinimum=%d baud=%lu searchCpsList=%08p called=%08p hMinInCps=%08p hMinOutCps=%08p zMinInCps=%08p zMinOutCps=%08p",
       sesCpsMinimum, baudForCpsLim, searchCpsList, called,
       hMinInCps, hMinOutCps, zMinInCps, zMinOutCps );

  if( wazoo==HydraSes )
  {
    changeTitle( Msg_TrmUpLoadWithHydra );

    waZooType = HydraSes;
    freqWasSent = 0;
    HydraInit( 0, called!=NULL, (hisCode&FNC)==FNC, inbound,
               inFiles, illegalAka==NULL ? 0 : 1, xfer );

    freqThrInfo.ignore = !answer;

    char requestFirst( char( hisCode&RH1 && wazoo==HydraSes ) );

    rc = sendBatch( akaList, wazoo, hisCode, password, answer, xfer, 1,
                    requestFirst ? 0 : 1, requestFirst ? 1 : 0 );

    if( rc==OK )
    {
      if( illegalAka==NULL )
        processInbound( inFiles, onlineTable, *primaryAka, getTimeMark()-sessionStart );

      rc = sendBatch( akaList, wazoo, hisCode, password, answer, xfer,
                      requestFirst ? 0 : ( freqWasSent ? 0 : 1 ),
                      requestFirst ? 1 : 0, 0 );
    }

    HydraDeInit();
  }
  else if( answer )
  {
    changeTitle( Msg_TtlReceive );

    freqThrInfo.ignore = 0;
    rc = Get_Zmodem( inbound, inFiles, wazoo, illegalAka==NULL ? 0 : 1 );

    if( rc==OK )
    {
      if( illegalAka==NULL )
	processInbound( inFiles, onlineTable, *primaryAka,
                        getTimeMark()-sessionStart );

      changeTitle( Msg_TtlSend );
      rc = sendBatch( akaList, wazoo, hisCode, password, 1, xfer, 0, 1, 0 );
    }
  }
  else
  {
    changeTitle( Msg_TtlSend );
    rc = sendBatch( akaList, wazoo, hisCode, password, 0, xfer, 1, 1, 0 );

    if( rc==OK )
    {
      changeTitle( Msg_TtlReceive );

      freqThrInfo.ignore = 1;
      rc = Get_Zmodem( inbound, inFiles, wazoo );
    }
  }

  killFreq();
  delete freqThrInfo.password;

  if( xfer!=NULL )
  {
    fclose( xfer );
    setExtendedAttribute( miscInfo.sentList, ".TYPE", Msg_EASentList, 0 );
  }

  proc = getRexxHook( "AfterSession" );

  if( proc!=NULL )
  {
    showAction( fileName( proc ) );
    processRexxHook( "AfterSession", sesType, addr, rc==OK ? "OK" : "FAIL" );
  }

  if( called!=NULL )
  {
    sessionCost = showCost( comPort->getTranxMark(), getTimeMark(), called );

    if( rc!=OK )
      called->makeLO( called->getTrafficFlavour(), 0 );
  }

  if( (hisCode&HAT)==HAT || (hisCode&HXT)==HXT ||
      (hisCode&HRQ)==HRQ || (hisCode&NRQ)==NRQ ||
      (hisCode&NPU)==NPU ||
      (miscInfo.zmh!=NULL && miscInfo.zmh->isWorked( time( NULL ) )) )
    rc = ~OK;

  waitFreqProcessor();
  freqBaseInfoSave();

  processInbound( inFiles, processTable, *primaryAka, getTimeMark()-sessionStart );

  delete primaryAka;
  primaryAka = NULL;

  if( rc )
    execSchedule( melodies.fail );
  else
    execSchedule( melodies.endSess );

  needRescan = 1;

  return rc;
}

// =============================== sessClean ===============================

void sessClean( QueueEntryList **akaList, char **password, int rc )
{
  if( *akaList!=NULL )
  {
    removeFreqBsy( *akaList, rc );

    delete *akaList;
    *akaList = NULL;
  }

  if( illegalAka!=NULL )
  {
    for( QueueEntryIterator i( *illegalAka ); i; i++ )
      i().makeBsy( 0 );

    delete illegalAka;
    illegalAka = NULL;
  }

  delete *password;
  *password = NULL;

  delete hisPassword;
  delete mailerProductCode;
  delete mailerName;
  delete mailerOwner;
  delete mailerVersion;
  delete mailerSerialNumber;
  delete hisName;
  delete hisPlace;
  delete hisSysOp;
  delete hisPhone;
  delete hisBaud;
  delete hisFlags;
  delete sessionCost;

  hisPassword = NULL;
  mailerProductCode = NULL;
  remoteMailerCode = 0;
  mailerName = NULL;
  mailerOwner = NULL;
  mailerVersion = NULL;
  mailerSerialNumber = NULL;
  hisName = NULL;
  hisPlace = NULL;
  hisSysOp = NULL;
  hisPhone = NULL;
  hisBaud = NULL;
  hisFlags = NULL;
  sessionCost = NULL;

  zInCount = 0;
  zOutCount = 0;
  zInBytes = 0;
  zOutBytes = 0;

  bbsAtThisCall = 0;
  sessionProtected = 0;
  sessionListed = 0;
  wasBusyEncountred = 0;
  sesCpsMinimum = 0;
}

// ================================ session ================================

static char node[128];

void _Optlink syncThread( void *delta )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, PRTYC_TIMECRITICAL, PRTYD_MAXIMUM, 0 );

  long curTime = time( NULL ) - (long)delta;
  struct tm *t = localTime( &curTime );

  DATETIME dateTime;
  APIRET rc = DosGetDateTime( &dateTime );

  dateTime.hours   = (UCHAR)t->tm_hour;
  dateTime.minutes = (UCHAR)t->tm_min;
  dateTime.seconds = (UCHAR)t->tm_sec;
  dateTime.day     = (UCHAR)t->tm_mday;
  dateTime.month   = (UCHAR)(t->tm_mon+1);
  dateTime.year    = (USHORT)(t->tm_year+1900);
  dateTime.weekday = (UCHAR)t->tm_wday;

  rc = DosSetDateTime( &dateTime );

  if( rc )
  {
    log( Msg_ErrUnableSetDateTime, rc );
    os2Message( rc );
    log( Msg_InfoTimeSyncUnable, node );
  }
  else
    log( Msg_InfoTimeSync, node, (long)delta );

  deRegisterExceptionHandler
}

void sinchronizeTime( QueueEntryList *akaList )
{
  if( akaList==NULL || !akaList->GetItemsInContainer() || timeSyncList==NULL )
    return;

  for( QueueEntryIterator i( *akaList ); i; i++ )
  {
    i().toString( node, sizeof(node)-1, FIDOAddressShowModeEMSI );

    for( EnableIterator list( *timeSyncList ); list; list++ )
      if( list().isUnify( node ) )
      {
        if( hisTime )
        {
          long hisGmt = hisTime + 
                        (hisTZ ? hisTZ*60 : getTimeZone() );
          long delta = comPort->getTranx()-hisGmt;

          if( abs(int(delta))<miscInfo.timeSyncDelta )
            log( Msg_InfoTimeSyncNoNeed, node, (long)delta );
          else
          {
            long tid = _beginthread( syncThread, 0, 40960, (void *)delta );

            if( tid==-1 )
              log( Msg_ErrUnableTimeSyncThr, strError() );
            else
              DosWaitThread( (PTID)&tid, DCWW_WAIT );
          }
        }
        else
          log( Msg_InfoTimeSyncUnable, node );

        return;
      }				// if  isUnify
  }				// for akaList
}

int session( QueueEntry *called )
{
  execSchedule( melodies.connect );
  addTitleMode( Msg_TtlConnected );
  changeTitle( NULL );

  zInCount = 0;
  zOutCount = 0;
  zInBytes = 0;
  zOutBytes = 0;
  sessionStart = getTimeMark();

  long sessionBegin( time( 0 ) );

  wndQueue.save();
  wndQueue = stdColor;
  --wndQueue;
  queue->queueName( Msg_ScrSession, answerDisabled );
  showAction( Msg_ScrSession );

  logPtr->empty();

  if( called==NULL )
    if( comPort->getCallersId()!=NULL )
      log( Msg_SesIncomingFrom, comPort->getCallersId(), comPort->getSuff() );
    else
      log( Msg_SesIncoming, comPort->getSuff() );
  else
    log( Msg_SesOutgoing, comPort->getSuff() );

  logPtr->empty();

  char *buf;
  int delta = (miscInfo.state1&StatusLine)==StatusLine ? 0 : 1;

  wndQueue = scrColor;

  buf = (char *)_alloca( 80 );

  memset( buf, ' ', 80 );
  buf[ 76 ] = 0;
  wndQueue( 2, 8+delta ) << buf;
  buf[ 10 ] = 0;
  wndQueue( 2, 9+delta ) << buf;
  wndQueue( 2, 10+delta ) << buf;

  wndQueue( 15, 8+delta ) << Msg_ScrSesMail;
  wndQueue( 25, 8+delta ) << Msg_ScrSesFiles;
  wndQueue( 42, 8+delta ) << Msg_ScrSesTotal;
  wndQueue( 68, 8+delta ) << Msg_ScrSesCPS;

  wndQueue( 3, 9+delta )  << Msg_ScrSesForUs;
  wndQueue( 3, 10+delta ) << Msg_ScrSesFromUs;

  wndQueue = brdColor;

  wndQueue( 15, 9+delta ) << char('0');
  wndQueue( 25, 9+delta ) << char('0');
  wndQueue( 40, 9+delta ) << "  0/0";
  wndQueue( 60, 9+delta ) << "0%";
  wndQueue( 68, 9+delta ) << char('0');
  wndQueue( 74, 9+delta ) <<  "0%";

  wndQueue( 15, 10+delta ) << char('0');
  wndQueue( 25, 10+delta ) << char('0');
  wndQueue( 40, 10+delta ) << "  0/0";
  wndQueue( 60, 10+delta ) << "0%";
  wndQueue( 68, 10+delta ) << char('0');
  wndQueue( 74, 10+delta ) <<  "0%";

  buf = (char *)_alloca( strlen(comPort->getSuff())+80 );

  sprintf( buf, "%c %s %s%s%s: %s %c", 4,
		called==NULL ? Msg_ScrSesIncoming : Msg_ScrSesOutgoing,
                Msg_ScrSesCall,	
                comPort->getCallersId()!=NULL ? Msg_ScrSesFrom : "",
                comPort->getCallersId()!=NULL ? comPort->getCallersId() : "",
                comPort->getSuff(), 4 );

  if( strlen(buf)>wndQueue.getWidth()-2 )
    buf[ wndQueue.getWidth()-2 ]=0;

  wndQueue = stdColor;
  wndQueue.centre( 2+delta, buf );

Emsi_Again:

  QueueEntryList *akaList( NULL );

  Compatibility myCode( Compatibility( ZMO|XMA ) ),
		hisCode( Compatibility(0) );

  if( (miscInfo.state3&HPFS)!=HPFS )
    myCode = Compatibility( myCode|FNC );

  if( called==NULL )
    if( !miscInfo.freqMode || miscInfo.freqTime==NULL )
      myCode = Compatibility( myCode|NRQ );
    else
    {
      if( !miscInfo.freqTime->isWorked() )
	myCode = Compatibility( myCode|HRQ );
    }
  else
    if( miscInfo.freqMode || miscInfo.freqTime==NULL )
      myCode = Compatibility( myCode|NRQ );
    else
      myCode = Compatibility( myCode|HRQ );

  char noDZA( 0 );

  if( called!=NULL && called->disableDZA() )
    noDZA = 1;

  if( comPort->getHardWareMode() &&
      (miscInfo.state2&DirZapAvail)==DirZapAvail &&
      !noDZA )
    myCode = Compatibility( myCode|DZA );

  char noZAP( 0 );

  if( called!=NULL && called->disableZAP() )
    noZAP = 1;

  if( (miscInfo.state2&ZedZapAvail)==ZedZapAvail && !noZAP )
    myCode = Compatibility( myCode|ZAP );

  char noHydra( 0 );

  if( called!=NULL && called->disableHydra() )
    noHydra = 1;

  if( !noHydra && (miscInfo.state2&HydraAvail)==HydraAvail )
    myCode = Compatibility( myCode|HYD|RH1 );

  if( called!=NULL )
    myCode = Compatibility( myCode|PUA );

  if( miscInfo.zmh!=NULL && miscInfo.zmh->isWorked( time( NULL ) ) )
  {
    myCode = Compatibility( myCode|HXT );

    if( (myCode&NRQ)!=NRQ )
      myCode = Compatibility( myCode|HRQ );
  }

  char *password( NULL );

  memset( &sessionBinaryInfo, 0, sizeof( sessionBinaryInfo ) );
  sessionBinaryInfo.type = 2;

  int rc1( ERROR );
  int rc( emsiHandShake( &akaList, myCode, hisCode, called, &password ) );
  char isBBS( 0 );
  void executeBBS();
  void internalLogin();
  int needRescan( 0 );
//      synchronize( 0 );

  switch( rc )
  {
  case 255:
    if( EMSIuserAction!=NULL &&
        ( EMSIuserAction->getFunc()==executeBBS ||
          EMSIuserAction->getFunc()==internalLogin
      ) )
    {
      if( called!=NULL )
      {
        EMSIuserAction = NULL;
        goto Emsi_Again_Clean;
      }

      setTitleMode( NULL );
      changeTitle( Msg_TtlBBS );
      EMSIuserAction->process();
      EMSIuserAction = NULL;
      isBBS = 1;
    }
    else
    {
      {
        char *oldTtl = newStr( titleMode );

        log( Msg_ScrSesExtern );
        addTitleMode( Msg_TtlExternal );
        changeTitle( NULL );

        if( EMSIuserAction!=NULL )
          EMSIuserAction->process();

        EMSIuserAction = NULL;

        setTitleMode( oldTtl );
        changeTitle( NULL );
        delete oldTtl;
      }

      if( (miscInfo.state1&RetryAfterExtern)==RetryAfterExtern &&
        comPort->isOnline() )  //check
      {
Emsi_Again_Clean:
        sessClean( &akaList, &password, rc1 );
        goto Emsi_Again;
      }
    }
    break;

  case 0:
//    synchronize = 1;

    if( illegalAka!=NULL && (miscInfo.state&PasswordAbort)!=PasswordAbort )
    {
      if( akaList==NULL )
	akaList = new QueueEntryList;

      QueueEntry &f = (*illegalAka)[0];

      FIDOAddress *adr = new FIDOAddress( f.getZone(), f.getNet(),
					  f.getNode(), f.getPoint(),
                                          f.getDomain() );
      NodeConfig cfg( adr );
      QueueEntry *que = new QueueEntry( cfg );

      akaList->AddAtHead( *que );
      primaryAka = adr;

      Buffer b;

      b.putString( Msg_SesNeedPassword );

      if( hisPassword==NULL )
	b.putString( Msg_SesNeedPassword1 );
      else
      {
	b.putString( Msg_SesNeedPassword2 );
	b.putString( hisPassword );
	b.putString( Msg_SesNeedPassword3 );
      }

      b.putString( Msg_SesNeedPassword4 );

      char buf[65];

      for( QueueEntryIterator list( *illegalAka ); list; list++ )
      {
	b.putString( list().toString( buf, 65 ) );
	b.putString( ", " );
      }

      b.stripByte();
      b.stripByte();

      b.putString( Msg_SesNeedPassword5 );
      b.putString( systemInfo.sysopName );
      b.putString( Msg_SesNeedPassword6 );

      b.putString( "\r\n\r\n" );
      b.putInto( 0 );

      generateFreqAnswer( b.str, *primaryAka, hisSysOp, NULL, NULL );
      goto need_session;
    }
    else
      if( akaList!=NULL && akaList->GetItemsInContainer() )
      {
	{
	  QueueEntry *first = &(*akaList)[0];

	  primaryAka = new FIDOAddress( first->getZone(),
					first->getNet(), first->getNode(),
					first->getPoint(), first->getDomain() );

	  int ok( 1 );

          if( called==NULL )
SearchForNotAcceptedAgain:
	    for( QueueEntryIterator i( *akaList ); i; i++ )
	    {
	      QueueEntry& cur = i();
	      char node[128];

	      cur.toString( node, sizeof(node)-1, FIDOAddressShowModeEMSI );

	      switch( processLogicalRexxHook( "AcceptableNode", node ) )
	      {
              default:
                if( !cur.isAcceptable() )
                {
              case 0:
	          log( Msg_SesNonAccept, node );

                  if( (miscInfo.state1&NotAcceptAbort)==NotAcceptAbort )
  	            ok = 0;
                  else
                  {
                    cur.makeBsy( 0 );
                    akaList->Remove( cur );
                    goto SearchForNotAcceptedAgain;
                  }
                }
                break;

              case 1:
                break;
	      }
	    }

	  if( !ok )
	  {
	    rc = 99;
	    break;
	  }

          if( !akaList->GetItemsInContainer() )
            goto NoAvailableAKA;
	}
need_session:
        if( !(wasBusyEncountred && (miscInfo.state3&BsyAbort)==BsyAbort) )
        {
  	  rc1 = doSession( hisCode, akaList, called==NULL ? 1 : 0, called,
                           myCode, password, needRescan );

  	  sessionBinaryInfo.elapsed = getTimeMark() - sessionStart;
 	  sessionBinaryInfo.protocol = char( waZooType );
	  sessionBinaryInfo.crc = char( Crc32 ? 32 : 16 );

	  QueueEntry &pri = (*akaList)[0];

	  sessionBinaryInfo.zone = pri.getZone();
	  sessionBinaryInfo.net = pri.getNet();
	  sessionBinaryInfo.node = pri.getNode();
	  sessionBinaryInfo.point = pri.getPoint();
          strncpy( sessionBinaryInfo.domain, pri.getDomain(),
                   sizeof(sessionBinaryInfo.domain)-1 );
          sessionBinaryInfo.domain[ sizeof(sessionBinaryInfo.domain)-1 ] = 0;

	  writeToBinaryLog( char( called==NULL ? 5 : 6 ),
			    &sessionBinaryInfo, sizeof( sessionBinaryInfo ),
			    sessionBegin );
        }
        else
          log( Msg_InfoBusyEncountered );
      }
      else
NoAvailableAKA:
	log( Msg_SesNoAvailAka );
    break;

  case 1:
    execSchedule( melodies.fail );
    log( Msg_SesNoHandshake );
    break;

  case 250:
    execSchedule( melodies.fail );
    log( Msg_SesHandshakeErr );
    break;

  case 3:
    execSchedule( melodies.fail );
    log( Msg_ZErrCarrierLost );
    break;

  default:
    execSchedule( melodies.fail );
    log( Msg_SesUnkReject );
  }

  char *t( strTime( getTimeMark()-sessionStart, 0 ) ),
       *adr( NULL );

  if( akaList!=NULL && akaList->GetItemsInContainer() )
  {
    adr = (char *)_alloca( 66 );
    (*akaList)[0].toString( adr, 63 );

    char *t1 = (char *)_alloca( strlen(t)+32 ),
         *t2 = tmpStr( adr ),
         *t3 = strchr( t2, '@' );

    if( t3!=NULL )
      *t3 = 0;

    sprintf( t1, "%lu, %s", comPort->getLastBaud(), t );
    queue->setLast( t2, t1, char(called==NULL) );

    strcat( adr, ", " );
  }

  if( !isBBS )
    if( rc )
      if( rc==99 )
        log( Msg_SesNonAccept1, adr==NULL? "" : adr, t );
      else
        log( Msg_SesFailed, adr==NULL ? "" : adr,
                            t,
                            sessionCost==NULL ? "" : sessionCost );
    else
    {
      char *ssi = strNumber( zInBytes );
      char *sso = strNumber( zOutBytes );

      log( Msg_SesCompleted,
           adr==NULL? "" : adr,
           t,
           zInCount, zOutCount,
           ssi, sso,
           sessionCost==NULL ? "" : sessionCost );
      delete ssi;
      delete sso;
    }

  delete t;

  delete primaryAka;
  primaryAka = NULL;

// statis

  changeTitle( Msg_InfoActionHang );
  showAction( Msg_InfoActionHang );
  comPort->hangUp();

  char *proc = getRexxHook( "Statistics" );

  if( proc!=NULL )
  {
    showAction( fileName( proc ) );
    processRexxHook( "Statistics" );
  }

  comPort->statistics();

  logPtr->empty();
  logPtr->flush();

  sinchronizeTime( akaList );
  sessClean( &akaList, &password, rc1 );

  sessionStart = 0;

  queue->queueName( Msg_WndTitleQue, answerDisabled );
  wndQueue.restore();

  setTitleMode( NULL );

  if( needRescan )
    Queue::needForRescan();

  return rc;
}

// =========================== externalConnect =============================

inline void externalConnect()
{
  connectList[ currentConnectItem ].process();
}

void cleanPort()
{
  comPort->clearin();
  comPort->clearout();
  sleep( 250 );
  comPort->clearin();
  comPort->clearout();
}


// ================================= answer ================================

int answer( int manualAnswer, int connectString )
{
  registerUsageTime( 1 );

  comPort->lock( Msg_InfoModemLockAnswer );

  int rc( 1 );
  int rc1( 0 );
  char *hook = NULL;

  if( connectString )
    goto DoSession;

  if( manualAnswer==2 )
  {
    showAction( Msg_SesDetectDCD );
    setTitleMode( Msg_TtlDCD );
    changeTitle( NULL );

    rc = session( NULL );

    if( rc==255 )
      rc = 0;
  }
  else
  {
    setTitleMode( Msg_TtlAnswer );
    changeTitle( NULL );

    hook = processRexxHook( "Answer", NULL, NULL, NULL,
                            NULL, NULL, NULL, 1 );
    rc1 = 0;

    if( hook!=NULL )
      if( !stricmp( hook, "No" ) )
        rc1 = 1; 
      else if( !stricmp( hook, "Yes" ) )
        rc1 = 2; 
      else if( !stricmp( hook, "YourSelf" ) )
        rc1 = 0; 
      else
      {
        rc1 = 1;
        log( Msg_ErrIllegalAnswerHook, hook );
      }

    delete hook;

    if( !rc1 )
      rc1 = comPort->answer( manualAnswer!=1 ? 1 : 0 );

    switch( rc1 )
    {
    case 2:
DoSession:
      rc = session( NULL );

      if( rc==255 )
	rc = 0;
      break;

    case 8:
      externalConnect();
      break;

    case 5:
      receiveFax();
      break;
    }
  }

  cleanPort();
  comPort->unlock();

  registerUsageTime( 0 );

  return rc;
}

// ============================== formatCallInfo ===========================

void formatCallInfo( Buffer& who, QueueEntry& current, char *buffer,
                     char *phone )
{
  who.putString( buffer );

  if( current.getRexxDial()!=NULL )
  {
    who.putString( ", " );
    who.putString( current.getRexxDial() );
  }

  if( current.getHiddenList()!=NULL )
  {
    who.putString( ", " );

    char rr[16];

    strcpy( rr, "[" );

    if( current.getCurrentLine()==-1 )
      strcpy( rr+1, "P" );
    else
    {
      strcpy( rr+1, "H" );

      if( current.getHiddenList()->GetItemsInContainer()>1 )
        sprintf( rr+2, "%d", current.getCurrentLine()+1 );
    }

    strcat( rr, "]" );

    who.putString( phone );
    who.putString( rr );
  }
  else if( phone!=NULL )
  {
    who.putString( ", " );
    who.putString( phone );
  }

  if( current.isListed() )
  {
    char *c;

    c = current.getNdl()->getSystem();

    if( c!=NULL )
    {
      who.putString( ", " );
      who.putString( c );
    }

    c = current.getNdl()->getSysop();

    if( c!=NULL )
    {
      who.putString( ", " );
      who.putString( c );
    }
  }
  else
    who.putString( Msg_SesUnknownSys );

  who.putInto( 0 );
}

// ================================== callNode ==============================

int callNode( QueueEntry& current, int flag )
{
  registerUsageTime( 1 );

  int initModem();
  int rc( 1 ),
      rc1( 0 );

  char buffer[65];

  current.toString( buffer, sizeof(buffer) );

  if( current.makeBsy( 1 ) )
  {
    log( Msg_SesOnAnother, buffer );
    return 252;
  }

  if( flag && current.isMapped() )
    current.draw( current.isMapped(), &current );

  if( current.getBefore()!=NULL )
  {
    registerUsageTime( 1 );
    doEventKey( *current.getBefore(), -1 );
    registerUsageTime( 0 );
  }

  if( current.getDo()!=NULL )
  {
    registerUsageTime( 1 );
    doEventKey( *current.getDo(), -1 );

    if( current.getAfter()!=NULL )
      doEventKey( *current.getAfter(), -1 );

    registerUsageTime( 0 );
    current.makeBsy( 0 );
    return 252;
  }

  comPort->lock( Msg_InfoModemLockCall );
  setTitleMode( Msg_TtlCall );
  addTitleMode( buffer );
  changeTitle( NULL );

  current.findAvailable();
  int startedAt = current.getCurrentLine();

  for( char done=0; !done; )
  {
    char *phone = current.getTranslatedPhone( current.getPhone() );

    current.loadTry();

    Buffer who;

    formatCallInfo( who, current, buffer, phone );
    log( Msg_SesCalling, flavourName( current.getQueuedFlavour() ), who.str );

    if( current.getRexxDial()!=NULL )
    {
      char buf[65];

      current.toString( buf, sizeof(buf), FIDOAddressShowModeEMSI );
      showAction( fileName( current.getRexxDial() ) );
                                                     
      rc1 = doRexxWithNumericRC( current.getRexxDial(), buf, phone );
      needInit = 1;
    }
    else
    {
      rc1 = comPort->dial( phone, current.getDial(), current.getDialSuff(),
                           current.getWaitCarrier(), 1 );

      if( current.getDial()!=NULL )
        needInit = (char)stricmp( current.getDial(), comPort->getPrefix() );
    }

    current.setCurrent();

    extern Buffer *fromModem;

    fromModem->putInto( 0 );
    current.setTryLast( fromModem->str );

    switch( rc1 )
    {
    case 2:
      rc = session( &current );

      if( rc==255 )
        rc = 0;

      done = 1;
      break;

    case 5:
      log( Msg_SesCallFAX );
      showAction( Msg_InfoActionHang );
      comPort->hangUp();
      done = 1;
      break;

    case 8:
      externalConnect();

    //fall

    case 4:
    case 3:
      done = 1;
      break;

    default:
      if( current.getChain() )
      {
        current.findAvailable();

        if( startedAt == current.getCurrentLine() )
          done = 1;
        else if( needInit )
        {
          needInit = 0;

          if( initModem() )
            done = 255;
        }
      }
      else
        done = 1;
    }

    delete phone;
  }

  if( current.getAfter()!=NULL )
    doEventKey( *current.getAfter(), -1 );

  current.makeBsy( 0 );

  registerUsageTime( 0 );

  switch( rc1 )
  {
  case 4:
    rc = 253;
    break;

  case 3:
    rc = 254;
    break;

  case 5:
    rc = 1;
    break;

  case 2:
    break;

  default:
    rc = rc1;
  }

  if( rc1!=4 )
    cleanPort();

  comPort->unlock();
  return rc;
}
