#include <ctype.h>
#include <string.h>

#include "misc.h"
#include "translat.h"

TransItem::TransItem( char *ip, char *ia, char *is, int cs ) :
  pref( newStr(ip) ),
  suf( newStr(is) ),
  as( newStr(ia) ),
  cost( cs )
{}

TransItem::~TransItem() 
{
  delete pref;
  delete as;
  delete suf;
}

int TransItem::equal( const BaseObject &t ) const
{
  if( pref==NULL || (*(const TransItem *)&t).pref==NULL )
    return 0;
  else
    return !strcmp( pref, (*(const TransItem *)&t).pref );
}

void TransItem::print( ostream& s ) const
{
  s << "Pref:";
  strItem( s, pref );

  s << ",As:";
  strItem( s, as );

  s << ",Suf:";
  strItem( s, suf );

  s << ",Cost:" << cost;
}

char *TransItem::nameOfClass() const
{
  return "TransItem";
}

TranslationTable::TranslationTable() : 
  countryCode( NULL ),
  internationalCall( NULL ),
  localCode( NULL ),
  Collection( 5, 5, SemaphoreModeNoNeed )
{}

TranslationTable::~TranslationTable()
{
   freePvt();
}

TransItem* TranslationTable::Find( TransItem& item, LockMode lock ) const
{
  return (TransItem*)Collection::Find( item, lock );
}
    
TransItem& TranslationTable::operator [] ( const int pos ) const
{
  return *(TransItem *)&(Collection::operator[]( pos ));
}

void TranslationTable::append( char *pref, char *as, char *suf, int cost )
{
  AddAtTail( *new TransItem( pref, as, suf, cost ) );
}

void TranslationTable::freePvt()
{
  delete countryCode;
  delete internationalCall;
  delete localCode;

  countryCode = NULL;
  internationalCall = NULL;
  localCode = NULL;
}

void TranslationTable::setCountry( char *code, char *call, char *l )
{
  if( code!=NULL )
  {
    delete countryCode;
    countryCode = code;
  }

  if( call!=NULL )
  {
    delete internationalCall;
    internationalCall = call;
  }

  if( l!=NULL )
  {
    delete localCode;
    localCode = l;
  }
}

void TranslationTable::translate( char *who, char **rez)
{
  if( *who=='#' )
  {
    *rez = newStr( who+1 );
    return;
  }

  if( localCode!=NULL && begin( who, localCode ) )
  {
     *rez = construct( who, localCode, "", 0 );
     return;
  }

  for( TranslationIterator list( *this ); list; list++ )
  {
    TransItem& cur( list() );
    register char *pref( cur.getPreffix() );

    if( begin( who, pref ) )
    {
      *rez = construct( who, pref, cur.getTranslate(), cur.getSuffix() );
      return;
    }
  }

  if( countryCode!=NULL && internationalCall!=NULL &&
      !begin( who, countryCode ) )
    *rez = construct( who, "", internationalCall, 0 );
  else
    *rez = newStr( who );
}

ostream& operator << ( ostream& o, const TranslationTable& c )
{
  o << "{{Country:";
  strItem( o, c.countryCode );

  o << "}{Call:";
  strItem( o, c.internationalCall );

  o << "}{Local:";
  strItem( o, c.localCode );

  o << "}\n";

  o << *(const Collection *)&c;

  o << "}\n";
  return o;
}

char *TranslationTable::construct( char *who, char *pref, char *as, char *suf )
{
  int delta( strlen( pref ) );
  char *ret( new char [ strlen(as)+strlen(who)-delta+1+
                        (suf==0 ? 0 : strlen(suf)) ] );

  strcpy( ret, as );
  strcat( ret, who+delta );

  if( suf!=0 )
    strcat( ret, suf );

  return ret;
}

int TranslationTable::findCost( char *phone )
{
  if( phone!=NULL )
    for( TranslationIterator list(*this); list; list++ )
    {
      TransItem& cur = list();

      if( begin( phone, cur.getPreffix() ) )
        return cur.getCost();
    }

  return 0;
}

TranslationIterator::TranslationIterator( TranslationTable& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}

TransItem& TranslationIterator::operator () () const
{
  return *(TransItem *)&(CollectionIterator::operator()());
}
