//---------------------------------------------------------------------------
//
// Copyright (c) 1995 by Bad Data, INC.
// All rights reserved.
//
// Written by John A. Gladkih.
//
// Class window - interface.
//

#if !defined(__WINDOW_CLASS__)

#define __WINDOW_CLASS__

#include <stddef.h>
#include <stdio.h>

#include "zforms.h"
#include "semaphor.h"

class Color {
public:

 unsigned char fore, back;

 Color( const Color& col );
 Color( int _fore, int _back );
 Color( unsigned char cl );
 Color( int cl );
 Color( unsigned int cl );

 Color& operator=( int _fore );
 Color& operator=( const Color& col );

 operator unsigned char() const;
 operator int() const;
};

extern char detachedSession;

class Window
{
  int saveX, saveY;
  int physX, physY;
  WINDOW *wnd, *pWin;
  Color color;
  char visible, scrolled;
  int delta, width, height, left, top;
  int *saveWnd;
  static unsigned long mutex;
  static int needLock;

public:

  static void lock();
  static void unlock();

protected:

  void putStringNoLock( const char * const str );
  int putFieldNoLock( const char * const str );
  void putChar( char ch, int lock=1 );
  void scrollUp();
  void clearLineNoLock();
  void gotoXY( int x, int y );
  void putString( const char * const str );

public:

  static void setLockMode( int lock );
  static int getLockMode() { return needLock; }

  Window( int _left, int _top, int _width, int _height, const Color& _col,
	         char* _title=NULL, int _brd=BS_NOBORDER, int _shd=DS_NONE,
  	         const Color _dropClr=Color( WHITE, BLACK ) );

  ~Window();

  Window& operator+();
  Window& operator-();

  Window& operator--();

  Window& operator<<( signed char c );
  Window& operator<<( char c );

  Window& operator<<( const signed char * const  c );
  Window& operator<<( const char * const  c );

  Window& operator<<( long c );
  Window& operator<<( unsigned long c );

  Window& operator<<( int c );
  Window& operator<<( unsigned int c );

  Window& operator<<( void* c );

  Window& operator()( int x, int y );
  Window& operator()( int y );

  Window& field( const char * const str );

  void drawTransact( int x, int y, const Color &cl, const char * const str, int clear=1 );
  void drawTransact( int x, int y, const Color &cl, unsigned long num, int clear=1 );

  void drawTransact( int x, int y, const Color &cl, long num, int clear=1 );
  void drawTransact( int x, int y, const Color &cl, int num, int clear=1 );

  void drawTransactField( int x, int y, const Color &cl, const char * const str, int clear=1 );
  void drawTransactField( int x, int y, const Color &cl, unsigned long num, int clear=1 );

  void drawTransactField( int x, int y, const Color &cl, long num, int clear=1 );
  void drawTransactField( int x, int y, const Color &cl, int num, int clear=1 );

  void pop();
  void push();

  Window& operator=( const Color& _col ){ color = _col; return *this; }

  Window& clearLine();

  Window& move( int x, int y );

  Window& centre( int row, char *str );

  void setScroll( char s );
  void setPhysicalCursor( int force=0 );

  void save();
  void restore();

  int getWidth() { return width; }
  int getHeight() { return height; }
  int getLeft() { return left; }
  int getTop() { return top; }
  char getScroll() { return scrolled; }
  Color& getColor() { return color; }
  int getCurrentX() { return saveX+1; }
  int getCurrentY() {  return saveY+1; }
  WINDOW *getWindow() { return pWin; }
};

void VioScreenDisable( void );
void VioScreenEnable( void );

#endif				// !defined(__WINDOW_CLASS__)
