#include <time.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <strstrea.h>
#include <iomanip.h>
#include <builtin.h>

#include "misc.h"

int WorkTime::equal( const BaseObject &t ) const
{
  return !memcmp( mask, (*(const WorkTime *)&t).mask, sizeof(mask) );
}

void WorkTime::print( ostream& s ) const
{
  s << (valid ? "Valid" : "NotValid")
    << char(',') << (isGmt ? "GMT" : "Local")
    << ",Str:";
  strItem( s, wtStr );
}

char *WorkTime::nameOfClass() const
{
  return "WorkTime";
}

void WorkTime::mark( int dayOfWeek, int hour, int mins )
{
  register int index( makeIndex( hour, mins ) );

  if( dayOfWeek==7 )
    dayOfWeek = 0;

  if( dayOfWeek>6 )
    valid = 0;
  else
    mask[dayOfWeek][index/8] |= char( 1 << (index%8) );
}

void WorkTime::markInterval( int wFrom, int wTo, int from, int to )
{
  for( int i=wFrom; i<wTo; i++ )
    if( to>from )
      for( int j=from; j<to; j++ )
	mark( i, j/60, j%60 );
    else
    {
      for( int j=from; j<24*60; j++ )
	mark( i, j/60, j%60 );

      for( j=0; j<to; j++ )
	mark( i, j/60, j%60 );
    }
}

void WorkTime::initNDL( const char *s )
{
  char *pat="AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXx",
       *s1( strchr( pat, s[1] ) ),
       *s2( strchr( pat, s[2] ) );

  isGmt = 1;

  if( s1==NULL || s2==NULL )
    valid = 0;
  else
  {
    long from( int(s1-pat)*30 ),
	 to( int(s2-pat)*30 );
    int wFrom( 0 ),
        wTo( 7 );

    if( s[0]=='W' )
    {
      wFrom = 6;
      wTo = 8;
      memset( mask[0], 0, sizeof(mask[0]) );
      memset( mask[6], 0, sizeof(mask[6]) );
    }

    markInterval( wFrom, wTo, int(from), int(to) );
  }
}

int decPart( register const char *s, int& day, int& time )
{
  if( !*s )
    return 1;

  int hour(0), minm(0), num(0);
  char mode( 0 );

  while( *s )
  {
    switch( *s )
    {
    case '.':

      if( mode )
	return 1;

      if( num>7 )
	return 1;

      if( !num )
	day = 7;
      else
	day = num;

      num = 0;
      mode = 1;
      break;

    case ':':

      if( mode==2 )
	return 1;

      hour = num;
      num = 0;
      mode = 2;
      break;

    default:
      if( isdigit(*s) )
	num = num*10 + (*s - '0');
      else
	return 1;
    }

    s++;
  }

  switch( mode )
  {

  case 0:
  case 1:
    hour = num;
    break;

  case 2:
    minm = num;
    break;

  default:
    return 1;
  }

  if( hour>23 || minm>59 )
    return 1;

  time = hour*60+minm;
  return 0;
}

void WorkTime::initSubst( const char *sIni )
{
  int from( 0 ),
      to( 0 ),
      wFrom( 0 ),
      wTo( 7 );

  char *s = tmpStr( sIni ),
       *minus = strchr( s, '-' );

  int strLen = strlen( s )-3;

  if( strLen>0 && !stricmp( s+strLen, "GMT" ) )
  {
    isGmt = 1;
    s[ strLen ] = 0;
  }

  if( minus==NULL )
    valid = 0;
  else
  {

    *(minus++) = 0;

    if( decPart( s, wFrom, from ) || decPart( minus, wTo, to ) )
      valid = 0;
    else
    {
      wTo++;

      if( wFrom<wTo )
	markInterval( wFrom, wTo, from, to );
      else
      {
	markInterval( wFrom, 8, from, to );
	markInterval( 1, wTo, from, to );
      }
    }
  }
}

void WorkTime::init( const char *s )
{
  if( s==NULL )
    return;

  char *work = (char *)_alloca( strlen(s)+2 );

  strcpy( work, s );
  work[ strlen(work)+1 ] = 0;

  for( char *subStr( work ); *subStr; )
  {
    char *comma( strchr( subStr, ',' ) );

    if( comma==NULL )
      comma = subStr+strlen( subStr );

    *comma = 0;

    if( strlen(subStr)==3 && (subStr[0]=='T'||subStr[0]=='W') )
      initNDL( subStr );
    else
      if( strlen(subStr)==4 && subStr[0]=='U' && (subStr[1]=='T'||subStr[1]=='W') )
	initNDL( subStr+1 );
      else
	if( !stricmp( subStr, "CM" ) )
	  memset( mask, 0xff, sizeof(mask) );
	else
	  initSubst( subStr );

    subStr = comma+1;
  }
}

WorkTime::WorkTime( const char *s ) :
  wtStr( NULL ),
  valid( 1 ),
  isGmt( 0 )
{
  memset( mask, 0, sizeof(mask) );

  if( s!=NULL )
  {
    wtStr = newStr( s );
    init( s );
  }
}

WorkTime::WorkTime( WorkTime *w )
{
  if( w==NULL )
  {
    wtStr = NULL;
    valid = 1;
    isGmt = 0;
    memset( mask, 0, sizeof(mask) );
  }
  else
  {
    wtStr = newStr( w->wtStr );
    valid = w->valid;
    isGmt = w->isGmt;
    memcpy( mask, w->mask, sizeof(mask) );
  }
}

int WorkTime::makeIndex( int hour, int mins ) const
{
  return hour*60 + mins;
}

WorkTime::~WorkTime()
{
  if( isPresent() )
    delete wtStr;
}

int WorkTime::isWorked( uLong stamp ) const
{
  if( !stamp )
    stamp = time( NULL );

  struct tm *tim;

  if( isGmt )
    tim = gmTime( (long *)&stamp );
  else
    tim = localTime( (long *)&stamp );

  register int index( makeIndex( tim->tm_hour, tim->tm_min ) );
  int dayOfWeek( tim->tm_wday );

  if( dayOfWeek==7 )
    dayOfWeek = 0;

  if( dayOfWeek>6 )
    return 0;

  return ( mask[tim->tm_wday][index/8] & (1 << (index%8)) ) ? 1 : 0;
}

char *stringWorkTime( WorkTime *wt, char *buf, int len )
{
  if( wt==NULL || !wt->isValid() || !wt->isPresent() )
    *buf=0;
  else   
    strncpy( buf, wt->getStr(), len-1 );
    
  buf[ len-1 ]=0;
  
  return buf;
}

