#include <share.h>
#include <fcntl.h>
#include <stdio.h>
#include <io.h>
#include <ctype.h>
#include <sys\stat.h>
#include <stdlib.h>
#include <stdlib.h>

#include "language.h"

#include "filetype.h"
#include "misc.h"
#include "timer.h"
#include "direct.h"
#include "session.h"

#include "zinter.h"
#include "zdialog.h"

/*--------------------------------------------------------------------------*/
/* Private data                                                             */
/*--------------------------------------------------------------------------*/

static FILE *Infile;                        // Handle of file being sent
static long LastZRpos;                      // Last error location
static int ZRPosCount;                     // ZRPOS repeat count
static long Txpos;                          // Transmitted file position
static int Rxbuflen;                        // Receiver's max buffer
					    //   length
static int Rxflags;                         // Receiver's flags

/*--------------------------------------------------------------------------*/
/* SEND ZMODEM (send a file)                                                */
/*   returns TRUE (1) for good xfer, FALSE (0) for bad                      */
/*   sends one file per call; 'fsent' flags start and end of batch          */
/*--------------------------------------------------------------------------*/

int Send_Zmodem( char *fname, char *alias, int fsent, int wazoo, int trunc,
                 FILE *xferinfo )
{
  register byte *p;
  register byte *q;
  int i;
  int rc = OK;  //john TRUE
  Crc32 = 0;

  zReceiveMode = 0;
  zmodemActive = 1;
  waZooType = wazoo;
  telebitFuckinCorrectionTimes = 0;
  timeDelimiter = '/';

  Infile = NULL;

  Rxtimeout = zGetTimeOut();

  infoDisplayMsg( Msg_ZScrWait, Msg_ZScrSend );

  log( "?Send_Zmodem: xferinfo=%p, Timeout=%d, sesCpsMinimum=%d, sent=%d, wazoo=%d",
       xferinfo, Rxtimeout*10, sesCpsMinimum, fsent, wazoo );

  switch( fsent )
  {
  case 0:
    Z_PutString( (byte *)"rz\r" );
    Z_PutLongIntoHeader( 0L );
    Z_SendHexHeader( ZRQINIT, (byte *)Txhdr );

  // fall

  case NOTHING_TO_DO:
    {
//      int saveTimeout = Rxtimeout;			//10/10/97

//      Rxtimeout = 300;				//10/10/97

      int i = ZS_GetReceiverInfo();

      if( i == ERROR )
        goto Err_Out;

      if( i==USER_ABORT)
        goto UserAbort;

//      Rxtimeout = saveTimeout;			//10/10/97
      log( "?Send_Zmodem: ZS_GetReceiverInfo()=%d", i );
    }
  }

  if( fname == NULL || !*fname )
    goto Done;

  Filename = fname;
  realname = Filename;

  Infile = _fsopen( Filename, "rb", SH_DENYWR );

  if( Infile==NULL )
  {
    log( Msg_ErrUnableOpen, Filename, strError() );
    goto Done;
  }

  if( isatty( fileno( Infile ) ) )
  {
    log( Msg_ZErrIsDev, fileName( Filename ) );
    goto Done;
  }

  if( miscInfo.inFileBufferSize )
    setvbuf( Infile, NULL, _IOFBF, miscInfo.inFileBufferSize*1024 );

  struct stat f;
  fstat( fileno( Infile ), &f );

  f.st_mtime -= getTimeZone();

  if( (miscInfo.state1&AdjustTime)==AdjustTime )
  {
    f.st_mtime &= ~1;
    log( "?Send_Zmodem: MSDOS file-time adjust done" );
  }

  file_length = f.st_size;

  for( p = (byte *)((alias != NULL) ? alias : Filename), q = Txbuf; *p; )
  {
    if( (*p == '/') || (*p == '\\') || (*p == ':') )
      q = Txbuf;
    else
      *q++ = char( tolower( *p ) );

    p++;
  }

  *q++ = '\0';

  zModemFNC( (char *)Txbuf, trunc );

  p = q = Txbuf+strlen( (char *)Txbuf )+1;

  while( q < (Txbuf + KSIZE) )
    *q++ = '\0';

  sprintf( (char *)p, "%lu %lo %o", f.st_size, f.st_mtime, f.st_mode );

  infoDisplayName( Crc32 ? 32 : 16, isProtocol( waZooType ), Filename,
		   file_length, "", alias, isProtocolShort( waZooType ) );

  i = ZS_SendFile( 1 + strlen ((char *)p) + (int) (p - Txbuf), wazoo );

  switch( i )
  {
UserAbort:
  case USER_ABORT:
    log( Msg_ZInfoESC );
    goto Err_Out;

  case OK:
    if( xferinfo!=NULL && fname!=NULL )
    {
      fprintf( xferinfo, "%s %s as %s at %lu %d %s\n",
               msgOfFileType( alias!=NULL ? alias : fname, NULL ),
               fname,
               fileName( alias!=NULL ? alias : fname ),
               time( NULL ),
               miscInfo.taskNumber,
               isProtocol( wazoo ) );
      real_flush( xferinfo );
    }

    goto Done;

  case ZSKIP:
    log( Msg_ZInfoSkipped, Filename);
    rc = SPEC_COND;                         // Success but don't postProcess!
    goto Done;

  case ZFERR:
    log( Msg_ZInfoZFERR, Filename );
    rc = VERY_SPEC_COND;                         // Success but don't postProcess!
    goto Done;

  case RCDO:
    log( Msg_ZErrCarrierLost );
    break;

  case ERROR:
   break;

  default:
    log( "?Z-BUG: not this case handled by Send_Zmodem: %d", i );
    break;
  }

Err_Out:
  send_can(); //john
  rc = ERROR;

Done:
  if( rc!=ERROR && fsent < 0 )
    ZS_EndSend();

  if( Infile )
    fclose( Infile );

  zmodemActive = 0;
  Filename = NULL;
  realname = NULL;

  return rc;
}

/*--------------------------------------------------------------------------*/
/* ZS SEND BINARY HEADER                                                    */
/* Send ZMODEM binary header hdr of type type                               */
/*--------------------------------------------------------------------------*/
void ZS_SendBinaryHeader( word type, register byte *hdr )
{
  comPort->bufferByte( ZPAD );
  comPort->bufferByte( ZDLE );

  if( Crc32 )
    ZS_32SendBinaryHeader( type, hdr );
  else
  {
    log( "?Z_SendBinaryHeader: %s", headerName( type ) );
    if( type==ZDATA )
      log( "?Z_SendBinaryHeader: send data from %lu",
		      _Z_PullLongFromHeader(hdr) );

    comPort->bufferByte( ZBIN );
    ZS_SendByte( (byte) type );

    register word crc;

    crc = word(Z_UpdateCRC( type, 0 ));

    for( int n(4); --n >= 0; )
    {
      ZS_SendByte( *hdr );
      crc = word(Z_UpdateCRC (((unsigned short) (*hdr++)), crc));
    }

    ZS_SendByte( (byte)(crc >> 8) );
    ZS_SendByte( (byte)crc );
  }

  comPort->unBufferBytes();

  if( type != ZDATA )
  {
    while( comPort->isOnline() && !comPort->outEmpty() )
      checkPoint();

    if( !comPort->isOnline() )
      comPort->clearout();
   }
}

/*--------------------------------------------------------------------------*/
/* ZS SEND BINARY HEADER 32 CRC                                             */
/* Send ZMODEM binary header hdr of type type                               */
/*--------------------------------------------------------------------------*/
void ZS_32SendBinaryHeader( unsigned short type, register byte *hdr )
{
  log( "?Z_32SendBinaryHeader: %s", headerName( type ) );
  if( type==ZDATA )
    log( "?Z_32SendBinaryHeader: send data from %lu",
		    _Z_PullLongFromHeader(hdr) );

  comPort->bufferByte( ZBIN32 );
  ZS_SendByte( (byte)type );

  unsigned long crc( 0xFFFFFFFFl );
  crc = Z_32UpdateCRC( type, crc );

  for( int n(4); --n >= 0; )
  {
    ZS_SendByte( *hdr );
    crc = Z_32UpdateCRC( (word)(*hdr++), crc );
  }

  crc = ~crc;
  for( n=4; --n >= 0; )
  {
    ZS_SendByte( byte(crc&0xff) );
    crc >>= 8;
  }
}

inline void ZS_SendDataDisplayPos( int how )
{
  infoDisplayPos( Txpos+how );
}

/*--------------------------------------------------------------------------*/
/* ZS SEND DATA                                                             */
/* Send binary array buf with ending ZDLE sequence frameend                 */
/*--------------------------------------------------------------------------*/
int ZS_SendData( register byte *buf, int length, unsigned short frameend, int disp )
{
  int rc = 0;

  if( Crc32 )
    rc = ZS_32SendData( buf, length, frameend, disp );
  else
  {
    log( "?ZS_SendData: send %d at %lu(0x%lx), %d (%s)",
		    length, Txpos, Txpos, frameend, headerName(frameend) );

    register word crc( 0 );
    int was = length;

    for( ; --length >= 0; ++buf )
    {
      ZS_SendByte( *buf );
      crc = word(Z_UpdateCRC( (word)*buf, crc ));

      if( disp )
        ZS_SendDataDisplayPos( was-length );
    }

    comPort->bufferByte( ZDLE );
    comPort->bufferByte( (byte)frameend );

    crc = word(Z_UpdateCRC( frameend, crc ));
    ZS_SendByte( byte(crc>>8) );
    ZS_SendByte( byte(crc&0xff) );
  }

  if( rc )
    goto Done;

  if( frameend == ZCRCW )
  {
    log( "?ZS_SendData: waiting to end transmition" );

    comPort->unBufferBytes();

    if( waZooType!=DirZap )
      comPort->putWait( XON, 1 );

    Timer t( Rxtimeout*10 );

    while( comPort->isOnline() && !comPort->outEmpty() && !t.isTimeOut() )
      checkPoint();

    if( !comPort->isOnline() )
      comPort->clearout();
  }

Done:

  if( Zabort()==27 )
    rc = USER_ABORT;

//  checkPoint();
  return rc;
}

/*--------------------------------------------------------------------------*/
/* ZS SEND DATA with 32 bit CRC                                             */
/* Send binary array buf with ending ZDLE sequence frameend                 */
/*--------------------------------------------------------------------------*/
int ZS_32SendData( register byte *buf, int length, word frameend, int disp )
{
  log( "?ZS_32SendData: send %d at %lu(0x%lx), %d (%s)",
		  length, Txpos, Txpos, frameend, headerName(frameend) );
  unsigned long crc( 0xFFFFFFFFl );

  int was = length;

  for( ; --length >= 0; ++buf )
  {
    ZS_SendByte( *buf );
    crc = Z_32UpdateCRC( (word)(*buf), crc );

    if( disp )
      ZS_SendDataDisplayPos( was-length );
  }

  comPort->bufferByte( ZDLE );
  comPort->bufferByte( (byte)frameend );
  crc = Z_32UpdateCRC( frameend, crc );

  crc = ~crc;

  for( length=4; --length >= 0; )
  {
    ZS_SendByte( byte(crc&0xff) );
    crc >>= 8;
  }

  return 0;
}

/*--------------------------------------------------------------------------*/
/* ZS SEND BYTE                                                             */
/* Send character c with ZMODEM escape sequence encoding.                   */
/* Escape XON, XOFF. Escape CR following @ (Telenet net escape)             */
/*--------------------------------------------------------------------------*/
void ZS_SendByte( register byte c )
{
  static byte lastsent( 0 );

  switch( c )
  {
  case 015:
  case 0215:

    if( (lastsent & 0x7F) != '@' )
      goto SendIt;

  case 021:
  case 023:
  case 0221:
  case 0223:
  case 020:
  case 0220:
  case ZDLE|0x80:

    if( waZooType==DirZap )
      goto SendIt;

  case ZDLE:
    comPort->bufferByte( ZDLE );
    c ^= 0x40;

  default:
SendIt:
    comPort->bufferByte( lastsent = c );
  }
}

/*--------------------------------------------------------------------------*/
/* ZS GET RECEIVER INFO                                                     */
/* Get the receiver's init parameters                                       */
/*--------------------------------------------------------------------------*/
int ZS_GetReceiverInfo()
{
  log( "?ZS_GetReceiverInfo started" );
  int c;

  for( int n(10); --n >= 0; )
  {
    c = Z_GetHeader( (byte *)Rxhdr );

    switch( c )
    {
    case ZCHALLENGE:
	    /*--------------------------------------*/
	    /* Echo receiver's challenge number     */
	    /*--------------------------------------*/
      Z_PutLongIntoHeader( Rxpos );
      Z_SendHexHeader( ZACK, (byte *)Txhdr );
      break;

    case TIMEOUT:
    case ERROR:
    case ZCOMMAND:
	    /*--------------------------------------*/
	    /* They didn't see our ZRQINIT          */
	    /*--------------------------------------*/
      Z_PutLongIntoHeader( 0L );
      Z_SendHexHeader( ZRQINIT, (byte *)Txhdr );
      break;

    case ZRINIT:
	    /*--------------------------------------*/
	    /* */
	    /*--------------------------------------*/
      Rxflags = 0377 & Rxhdr[ZF0];
      Rxbuflen = ((word) Rxhdr[ZP1] << 8) | Rxhdr[ZP0];
      Crc32 = Rxflags & CANFC32;
      return OK;

    case ZCAN:
      log( Msg_ZInfoCanceledRemote );
      return ERROR;

    case RCDO:
      log( Msg_ZErrCarrierLost );
      return ERROR;

    case USER_ABORT:
      return USER_ABORT;

    case ZRQINIT:
      if( Rxhdr[ZF0] == ZCOMMAND )
	break;

    // fall

    default:
      Z_SendHexHeader( ZNAK, (byte *)Txhdr );
    }

    infoDisplayTry( n+1 );
  }

  log( Msg_ZErrSyncTimeOut );
  return ERROR;
}

/*--------------------------------------------------------------------------*/
/* ZS SEND FILE                                                             */
/* Send ZFILE frame and begin sending ZDATA frame                           */
/*--------------------------------------------------------------------------*/
int ZS_SendFile( int blen, int wazoo )
{
  log( "?ZS_SendFile started" );

  register int c;
  Timer t( 20000l );

  for( ;; )
  {
    if( !comPort->isOnline() )
      return RCDO;

    Txhdr[ZF0] = LZCONV;                      // Default file conversion mode
    Txhdr[ZF1] = LZMANAG;                     // Default file management mode
    Txhdr[ZF2] = LZTRANS;                     // Default file transport mode
    Txhdr[ZF3] = 0;
    ZS_SendBinaryHeader( ZFILE, (byte *)Txhdr );
    ZS_SendData( Txbuf, blen, ZCRCW );

Again:

    if( !comPort->isOnline() )
      return RCDO;

    switch( c = Z_GetHeader( (byte *)Rxhdr ) )
    {
    case ZRINIT:
/*
      while( (c = Z_GetByte( 50 )) > 0 )
	if( c == ZPAD )
*/
	  goto Again;

    /* Fall thru to */

    default:
      continue;

    case RCDO:
      return RCDO;

    case USER_ABORT:
      return USER_ABORT;

    case ZFIN:
    case ZABORT:
      log( Msg_ZErrAbortedRemote );
      return ERROR;

    case ZFERR:
      return c;

    case TIMEOUT:
      if( !t.isTimeOut() )
      {
	infoDisplayTry( t.secsLeft() );
	continue;
      }
      else
      {
        log( Msg_ZErrSendStartTimeOut );
	return ERROR;
      }

    case ZCAN:
    case ZSKIP:
	     /*-----------------------------------------*/
	     /* Other system wants to skip this file    */
	     /*-----------------------------------------*/
      return c;

    case ZRPOS:
	     /*-----------------------------------------*/
	     /* Resend from this position...            */
	     /*-----------------------------------------*/
      fseek( Infile, Rxpos, SEEK_SET );

      if( Rxpos != 0L )
      {
        char *ss = strNumber( Rxpos );

        log( Msg_ZInfoSendFrom, ss );
        delete ss;
      }

      comPort->clearout();              	// Get rid of queued data
      comPort->putWait( XON, 1 );               // Send XON to remote

      LastZRpos = Strtpos = Txpos = Rxpos;
      ZRPosCount = 10;
      comPort->clearin();

      return ZS_SendFileData( wazoo );
    }
  }
}


/*--------------------------------------------------------------------------*/
/* ZS Special SEND DATA                                                     */
/* Send the data block 4 telebit modem                                      */
/*--------------------------------------------------------------------------*/

int ZS_SpecialSendData( byte *Txbuf, int len, unsigned short frameend, int disp=0 )
{
  log( "?ZS_SpecialSendData:  send %d bytes, frame %d (%s)",
		    len, frameend, headerName(frameend) );
  int pCount( 0 );

  for( byte *ptr( Txbuf ), *prevPtr( Txbuf ); Txbuf+len!=ptr; )
    switch( byte(*(ptr++)) )
    {
    case '+':
    case '+'|0x80:
      pCount++;
      break;

    case 'a':
    case 'A':
    case 'a'|0x80:
    case 'A'|0x80:
      if( pCount>=3 )
      {
	--ptr;
        telebitFuckinCorrectionTimes = 1;

	int i = ZS_SendData( prevPtr, int(ptr-prevPtr), ZCRCG, disp );

        if( i )
          return i;

	prevPtr = ptr;
      }

    // fall

    default:
      pCount = 0;
    }

  return ZS_SendData( prevPtr, int(ptr-prevPtr), frameend, disp );
}


/*--------------------------------------------------------------------------*/
/* ZS SEND FILE DATA                                                        */
/* Send the data in the file                                                */
/*--------------------------------------------------------------------------*/
int ZS_SendFileData( int wazoo )
{
  log( "?ZS_SendFileData started" );

  register short c, e;
  long newcnt;
  long blklen;
  long goodblks( 0 );
  long goodneeded( 1 );

  long maxblklen( comPort->getLastBaud() / 300l * 256l );

  if( maxblklen < 128 )
    maxblklen = 128;
  else
    if( maxblklen > WAZOOMAX )
      maxblklen = WAZOOMAX;

  if( !wazoo && maxblklen > KSIZE )
    maxblklen = KSIZE;

  if( Rxbuflen && maxblklen > (unsigned)Rxbuflen )
    maxblklen = Rxbuflen;

  if( zSendInitBlock )
  {
    blklen = zSendInitBlock;
    goodneeded = 8;
  }
  else
  {
    blklen = (miscInfo.zBlock != 0) ? miscInfo.zBlock : maxblklen;
    zSendInitBlock = blklen;
    goodneeded = (miscInfo.zBlock != 0) ? 8 : 1;
  }

  infoDisplayBlock( word( blklen > file_length ? file_length : blklen ) );
  infoDisplayPos( Txpos );
  infoDisplayMarkTime( Txpos );

  zTransferStart = getTimeMark();

//SomeMore:

  if( comPort->haveInput() )
  {
WaitAck:

    switch( c = word(ZS_SyncWithReceiver( 1 )) )
    {
    case ZSKIP:
	    /*-----------------------------------------*/
	    /* Skip this file                          */
	    /*-----------------------------------------*/
      goto skip_rest;

    case ZACK:
      break;

    case ZRPOS:
	    /*-----------------------------------------*/
	    /* Resume at this position                 */
	    /*-----------------------------------------*/
Zrpoz:
      blklen = ((blklen >> 2) > 64) ? blklen >> 2 : 64;
      zSendInitBlock = blklen;
      goodblks = 0;
      goodneeded = ((goodneeded << 1) > 16) ? 16 : goodneeded << 1;
      comPort->clearout();
      comPort->putWait( XON, 1 );

      if( remoteMailerCode==0xe0 && Txpos==file_length )
      {
        log( "?ZS_SendFileData: t-m skip on the fly" );
        goto Send_EOF;
      }
      break;

    case ZRINIT:
	    /*-----------------------------------------*/
	    /* Receive init                            */
	    /*-----------------------------------------*/
isOK:
      c = OK;
Done:
      zOutCount++;
      zOutBytes += Txpos - Strtpos;
      infoDisplayDone( Txpos - Strtpos, Strtpos, 0 );
      return c;

    case ZFERR:
remote_file_error:
      goto Done;

    case USER_ABORT:
      goto Done;

    case ERROR:
      goto Done;

    default:
      log( "?Z-BUG: not this case handled by ZS_SendFileData: %d", c );
      break;
    }

      /*
       * Noise probably got us here. Odds of surviving are not good. But we
       * have to get unstuck in any event.
       *
       */

    c = (short)Z_UncorkTransmitter();

    if( c==ERROR )
      log( Msg_ZErrUncorkFail );

    if( c!=OK )
      if( !comPort->isOnline() )
      {
Carrier:
        log( Msg_ZErrCarrierLost );
        c = ERROR;
        goto Done;
      }
      else
        goto Done;

    if( waZooType!=DirZap )
      comPort->putWait( XON, 1 );                   // Send an XON to release other side

    while( comPort->haveInput() )
    {
      if( !comPort->isOnline() )
        goto Carrier;

      switch( comPort->get() )
      {
	case CAN:
	case ZPAD:
	  goto WaitAck;
      }
    }
  }

  newcnt = Rxbuflen;
  Z_PutLongIntoHeader( Txpos );
  ZS_SendBinaryHeader( ZDATA, (byte *)Txhdr );

  do
  {
    if( !comPort->isOnline() )
      goto Carrier;

    c = word( fread( Txbuf, 1, word(blklen), Infile ) );

    if( ferror( Infile ) )
    {
      log( Msg_ErrUnableRead, Filename, strError() );
      c = ERROR;
      goto Done;
    }

    if( (unsigned)c < blklen )
      e = ZCRCE;
    else if( Rxbuflen && (newcnt -= c) <= 0 )
      e = ZCRCW;
    else
      e = ZCRCG;

    infoDisplayPos( Txpos );
    infoDisplayBlock( c );

    int cps = infoDisplayCps();

    if( !terminalActive && !rexxTransfer )
    {
      if( getTimeMark() > zTransferStart+zTransferCpsAfter )
      {
        timeDelimiter = '|';

        if( sesCpsMinimum && cps<sesCpsMinimum )
        {
          log( Msg_ZInfoLowCPS );
          c = ERROR;
          goto Done;
        }
      }

      if( sessionLimitExceeded() )
      {
        c = ERROR;
	goto Done;
      }
    }

    int sndStat;

    if( (miscInfo.state1&Telebit)==Telebit && c )
      sndStat = ZS_SpecialSendData( Txbuf, c, e, 1 );
    else
      sndStat = ZS_SendData( Txbuf, c, e, 1 );

    if( sndStat )
    {
      c = (short)sndStat;
      goto Done;
    }

    Txpos += c;
    if( blklen < maxblklen && ++goodblks > goodneeded )
    {
      blklen = ((blklen << 1) < maxblklen) ? blklen << 1 : maxblklen;
      zSendInitBlock = blklen;
      goodblks = 0;
    }

    if( e == ZCRCW )
      goto WaitAck;

    while( comPort->haveInput() )
    {
      if( !comPort->isOnline() )
        goto Carrier;

      switch( comPort->get() )
      {
      case ZPAD:
	       /*--------------------------------------*/
	       /* Interruption detected;               */
	       /* stop sending and process complaint   */
	       /*--------------------------------------*/
        log( "?Remote was ANY problem" );
	comPort->clearout();
/* very new
	ZS_SendData( Txbuf, 0, ZCRCE );  // 嬬,  祬 ᫠ ⮩ p?????????
*/

      case CAN:
	goto WaitAck;
      }
    }

  } while (e == ZCRCG);

Send_EOF:

  for( ;; )
  {
    Z_PutLongIntoHeader( Txpos );
    ZS_SendBinaryHeader( ZEOF, (byte *)Txhdr );

    switch( (c = word(ZS_SyncWithReceiver( 8 )), c) )
    {
    case ZACK:
      continue;

    case ZRPOS:
	    /*-----------------------------------------*/
	    /* Resume at this position...              */
	    /*-----------------------------------------*/
      goto Zrpoz;

    case ZRINIT:
	    /*-----------------------------------------*/
	    /* Receive init                            */
	    /*-----------------------------------------*/
      goto isOK;

    case ZSKIP:
	    /*-----------------------------------------*/
	    /* Request to skip the current file        */
	    /*-----------------------------------------*/
      goto skip_rest;

    case ZFERR:
      goto remote_file_error;

    case RCDO:
    case ERROR:
    case USER_ABORT:
      goto Done;

    default:
      log( "?Z-BUG: not this case handled by 2 ZS_SendFileData: %d", c );
      c = ERROR;
      goto Done;
    }
  }

skip_rest:
  log( Msg_ZInfoSkipRemote );
  goto isOK;
}

/*--------------------------------------------------------------------------*/
/* ZS SYNC WITH RECEIVER                                                    */
/* Respond to receiver's complaint, get back in sync with receiver          */
/*--------------------------------------------------------------------------*/
int ZS_SyncWithReceiver( int errs )
{
  log( "?ZS_SyncWithReceiver( %d )", errs );

  for( ;; )
  {
    register int c( Z_GetHeader( (byte *)Rxhdr ) );

//    comPort->clearin();

    switch( c )
    {
    case ZFERR:
      return ZFERR;

    case USER_ABORT:
      return USER_ABORT;

    case TIMEOUT:
      infoDisplayTry( errs );

      if( --errs<=0 )
      {
        log( Msg_ZErrSyncTimeOut );
	return ERROR;
      }

      break;

    case ZCAN:
      log( Msg_ZInfoCanceledRemote );
      return ERROR;

    case ZABORT:
      log( Msg_ZErrSyncAborted );
      return ERROR;

    case ZFIN:
      log( Msg_ZErrSyncFinished );
      return ERROR;

    case RCDO:
      log( Msg_ZErrCarrierLost );
      return ERROR;

    case ZRPOS:
      if( Rxpos == LastZRpos )              // Same as last time?
      {
	if( !(--ZRPosCount) )               // Yup, 10 times yet?
	{
          log( Msg_ZErrManyError );
	  return ERROR;                     // Too many, get out
	}

	infoDisplayTry( ZRPosCount );
      }

      LastZRpos = Rxpos;                    // Keep track of this

      {
        char *ss = strNumber( Rxpos );

        fseek( Infile, Rxpos, SEEK_SET );
        Txpos = Rxpos;
        log( Msg_ZInfoResendFrom, ss );
        delete ss;
        return c;
      }

    case ZSKIP:
    case ZRINIT:
    case ZACK:
      return c;

    default:
//      log( Msg_ZErrDebries );
      ZS_SendBinaryHeader( ZNAK, (byte *)Txhdr );
    }
  }
}

/*--------------------------------------------------------------------------*/
/* ZS END SEND                                                              */
/* Say BIBI to the receiver, try to do it cleanly                           */
/*--------------------------------------------------------------------------*/
void ZS_EndSend()
{
  log( "?ZS_EndSend started" );

  for( Timer t(5000); !t.isTimeOut() &&
                      comPort->isOnline() &&
                      !comPort->outEmpty(); )
    checkPoint();

  if( !comPort->isOnline() )
    return;

  for( int i=0; i<5; i++ )
  {
    log( "?ZS_EndSend try=%d", i );

    Z_PutLongIntoHeader( 0L );
    ZS_SendBinaryHeader( ZFIN, (byte *)Txhdr );

    register int c = Z_GetHeader( (byte *)Rxhdr );

    switch( c )
    {
    case TIMEOUT:
      log( "?ZS_EndSend timeout" );
      break;

    case ZFIN:
      comPort->putWait( 'O', 10 );
      comPort->putWait( 'O', 10 );
      log( "?ZS_EndSend OK" );
      return;

    case ZCAN:
    case RCDO:
    case USER_ABORT:
      log( "?ZS_EndSend c=%d", c );
      return;
    }
  }
  log( "?ZS_EndSend retry count exceeded" );
}

int ZSend_Batch( char *name, int wazoo,
                 CallBackFunc callBack,
                 ZModemRexxCallBackContext* context )
{
  char *trnName = (char *)_alloca( strlen(name)+2 ),
       *ori( trnName );

  changeBlankWithZero( trnName, name );

  int zret( OK ),
      wasSent( 0 );

  char drive[_MAX_DRIVE ],
       dir  [_MAX_DIR ],
       fname[_MAX_FNAME],
       ext  [_MAX_EXT],
       full [_MAX_PATH];

  infoDisplayStart();

  zSendInitBlock = 0;

  for( trnName=ori; *trnName && zret!=ERROR; trnName += strlen(trnName)+1 )
  {
    char action( 0 );
    
    if( *trnName=='^' )
    {
      action = '^';
      trnName++;
    }
    else if( *trnName=='#' )
    {
      action = '#';
      trnName++;
    }
    else if( *trnName=='~' )
      continue;

    Directory rc( trnName );

    if( !rc )
      log( Msg_TrmUpLoadNotFound, trnName );
    else
      do
      {
        _splitpath( trnName, drive, dir, fname, ext );
	strcpy( full, drive);
	strcat( full, dir );
	strcat( full, rc.getName() );
	makeFileCase( full );

	zret= Send_Zmodem( full, NULL, wasSent, wazoo, 0, NULL );

        if( zret==OK || zret==SPEC_COND )
        {
          wasSent++;

          if( callBack )
            callBack( context, full );

          postProcessAttach( action, full );
        }
        
        rc++;

      } while( zret!=ERROR && rc );
  }

  if( zret != ERROR && wasSent )
    Send_Zmodem( NULL, NULL, END_BATCH, wazoo, 0, NULL );

  if( zret == OK && !wasSent )
    Send_Zmodem( NULL, NULL, NOTHING_TO_DO, wazoo, 0, NULL );

  infoDisplayEnd();

  return wasSent ? zret : ERROR;
}
