/************************ SWin ***************************\
 *
 *	Copyright (C) 1996 by Dmitry Zavalishin
 *
 *	Module 	: OS/2 specific device classes impl.
 *
 *      $Log: device_os2.c $
 *      Revision 1.1  1996/10/10 10:46:53  dz
 *      Initial revision
 *
 *
 *      
 *
\*/


#include "sw_device.h"
#include <io.h>

#define INCL_MOU
#define INCL_KBD
#define INCL_VIO
#include <os2.h>



static sw_size get_vio_size()
    {
    VIOMODEINFO mi;
    mi.cb = sizeof(mi);
    int rc = VioGetMode( &mi, 0);
    if( rc )
        {
        printf("VioGetMode error: %d\n", rc );
        throw sw_Fail("get_vio_size","VioGetMode failed","");
        }
    return sw_size(mi.col, mi.row);
    }
#if 0
sw_vio_base::sw_vio_base() : sw_device( get_vio_size() )
    {
    ULONG buf;
    USHORT bsize;
    if( VioGetBuf( &buf, &bsize, 0 ) )
        throw sw_Fail("sw_vio_base::sw_vio_base","VioGetBuf failed","");
    unsigned short sel = buf >> 16;
    unsigned short off = buf & 0xFFFFu;
    viobuf = (char *) MAKEP(sel,off);

    int len = xs()*ys()*2;
    saved_screen = new char[len];
    memcpy( saved_screen, viobuf, len ); // save user's screen
    }

sw_vio_base::~sw_vio_base()
    {
    int len = xs()*ys()*2;
    memcpy( viobuf, saved_screen, len ); // restore user's screen
    VioShowBuf( 0, len, 0 );
    delete [] saved_screen;
    }
#else
sw_vio_base::sw_vio_base() : sw_2b_video_memory_device( get_vio_size() )
    {
    saved_screen = 0;
    vm_alloc();
    }

sw_vio_base::~sw_vio_base()
    {
    vm_free();
    }

void sw_vio_base::vm_alloc()
    {
    ULONG buf;
    USHORT bsize;
    if( VioGetBuf( &buf, &bsize, 0 ) )
        throw sw_Fail("sw_vio_base::sw_vio_base","VioGetBuf failed","");
    unsigned short sel = buf >> 16;
    unsigned short off = buf & 0xFFFFu;
    video_memory = (char *) MAKEP(sel,off);
    
    int len = xs()*ys()*2;
    saved_screen = new char[len];
    memcpy( saved_screen, video_memory, len ); // save user's screen
    }

void sw_vio_base::vm_free()
    {
    if(saved_screen)
        {
        int len = xs()*ys()*2;
        memcpy( video_memory, saved_screen, len ); // restore user's screen
        VioShowBuf( 0, len, 0 );
        delete [] saved_screen;
        saved_screen = 0;
        }
    }


#endif

void sw_vio_base::do_flush(const sw_rect &_p)
    {
    sw_rect p;
    if( !p.intersection( _p, sw_rect(0,0,xs(),ys()) ) )        return;

    int off = p.y()  * xs() * 2;
    int len = p.ys() * xs() * 2;

    if( VioShowBuf( off, len, 0 ) )
        throw sw_Fail( "sw_vio_base::do_flush", "VioShowBuf error", "" );
    }

#if 0
void sw_vio_base::fill_cw( const sw_abstract_win &who, sw_color col, char ch, sw_rect _p )
    {
    invalidate(_p);
    
    sw_rect p;
    if( !p.intersection( _p, sw_rect(0,0,xs(),ys()) ) )        return;
    
    //int inp_bpl = _p.xs() * 2;
    int scr_bpl = xs() * 2;
    int map_bpl = xs();
    
    char *vio_line = viobuf+(p.y()*scr_bpl);
    const sw_abstract_win * const * map_line = map.mapp()+(p.y()*map_bpl);
    
    int pxshift = p.x()*2;
    char cv = col.value();
    
    for( int vcount = p.ys(); vcount--; )
        {
        //const char *lp = in_line;
        //in_line += inp_bpl;
        
        char *vp = vio_line+pxshift;
        vio_line += scr_bpl;
        
        const sw_abstract_win * const * mp = map_line+p.x();
        map_line += map_bpl;
        
        for( int hcount = p.xs(); hcount--; )
            {
            if( &who == *mp++ )
                {
                *vp++ = ch;
                *vp++ = cv;
                }
            else
                {
                vp+=2;
                }
            }
        }
    }

void sw_vio_base::border_cw( const sw_abstract_win &who, sw_color col, sw_border_t t )
    {
    }
void sw_vio_base::put_cw( const sw_abstract_win &who, const sw_rect &_p, const char *in_line )
    {
    invalidate(_p);
    
    sw_rect p;
    if( !p.intersection( _p, sw_rect(0,0,xs(),ys()) ) )        return;

    int inp_bpl = _p.xs() * 2;
    int scr_bpl = xs() * 2;
    int map_bpl = xs();

    char *vio_line = viobuf+(p.y()*scr_bpl);
    const sw_abstract_win * const * map_line = map.mapp()+(p.y()*map_bpl);

    int pxshift = p.x()*2;
    
    for( int vcount = p.ys(); vcount--; )
        {
        const char *lp = in_line;
        in_line += inp_bpl;

        char *vp = vio_line+pxshift;
        vio_line += scr_bpl;

        const sw_abstract_win * const * mp = map_line+p.x();
        map_line += map_bpl;

        for( int hcount = p.xs(); hcount--; )
            {
            if( &who == *mp++ )
                {
                *vp++ = *lp++;
                *vp++ = *lp++;
                }
            else
                {
                vp+=2; lp+=2;
                }
            }
        }
    }
#endif
  // -----------------------------------------------------------------------

sw_vio_device::sw_vio_device()  {}
sw_vio_device::~sw_vio_device() {}

const char *sw_vio_device::name() { return "vio"; }

  // -----------------------------------------------------------------------

sw_avio_device::sw_avio_device()  {}
sw_avio_device::~sw_avio_device() {}

const char *sw_avio_device::name() { return "avio"; }

  // -----------------------------------------------------------------------

