/************************ SWin ***************************\
 *
 *	Copyright (C) 1996 by Dmitry Zavalishin
 *
 *	Module 	: map classes impl.
 *
 *      $Log: map.c $
 *      Revision 1.5  1996/09/28 22:58:52  dz
 *      menu item classes written (wrong way yet)
 *
 *      Revision 1.4  1996/09/27 00:33:20  dz
 *      viobuf, optimization (sort of:)
 *
 *      Revision 1.3  1996/09/26 22:07:12  dz
 *      faster map rebuild. not much, though.
 *
 *      Revision 1.2  1996/09/20 14:40:18  dz
 *      update
 *
 *      Revision 1.1  1996/09/12 01:12:51  dz
 *      Initial revision
 *
 *
 *      
 *
\*/


#include "sw_win.h"
#include "sw_map.h"

  // --------------------------- sw_map -----------------------------


void sw_map::allocme()
    {
    int bytes = sw_size::xs() * sw_size::ys();
    v = new sw_abstract_win * [ bytes ];
    if( v == NULL ) throw sw_Out_Of_Memory("sw_map::allocme");
    clear(NULL);
    }

void sw_map::freeme()
    {
    delete [] v;
    v = NULL;
    }

sw_map::sw_map( const sw_size &_s ) : sw_size(_s)
    {
    allocme();
    }

sw_map::~sw_map()
    {
    freeme();
    }

void sw_map::resize( const sw_size &delta )
    {  freeme();  sw_size::resize( delta ); allocme();  }

void sw_map::xs( int _ )
    {  freeme();  sw_size::xs( _ );         allocme();  }

void sw_map::ys( int _ )
    {  freeme();  sw_size::ys( _ );         allocme();  }



void sw_map::clear( sw_abstract_win *w )
    {
    for( int i = sw_size::xs()*sw_size::ys(); i--; )
        v[i] = w;
    }

void sw_map::load( const sw_map &m, const sw_point &p )
    {
    sw_rect is;

    if( !is.intersection(
                         sw_rect( sw_point(0,0), *this ),
                         sw_rect( p, m )
                        ) )
        return;

#if 1
    int y2x1 = is.y()*sw_size::xs();
    int y2x2 = ((is.y()-p.y())*m.sw_size::xs())-p.x();
    int y2x1_inc = sw_size::xs();
    int y2x2_inc = m.sw_size::xs();
    int isyend = is.yend();
    int isxend = is.xend();
    
    for( int yi = is.y(); yi < isyend; yi++ )
        {
        for( int xi = is.x(); xi < isxend; xi++ )
            v[xi+y2x1] = m.v[xi+y2x2];
        y2x1 += y2x1_inc;
        y2x2 += y2x2_inc;
        }
#else
    for( int yi = is.y(); yi < is.yend(); yi++ )
        for( int xi = is.x(); xi < is.xend(); xi++ )
            v[xi+(yi*sw_size::xs())] =
            m.v[xi-p.x()+((yi-p.y())*m.sw_size::xs())];
#endif
    }

void sw_map::dump() const
    {
    cerr.flags( ios::hex );
    cerr.width( 6 );
    cerr << "sw_map::dump ->\n";
    for( int yi = 0; yi < sw_size::ys(); yi++ )
        {
        cerr << " ";
        for( int xi = 0; xi < sw_size::xs(); xi++ )
            cerr << (int)v[xi+(yi*sw_size::xs())] << " ";
        cerr << "\n";
        }
    }

int sw_map::belongs( const sw_abstract_win *w, int x, int y ) const
    {
    return w == v[x+(y*sw_size::xs())];
    }

int sw_map::belong_count( const sw_abstract_win *w, int x, int y ) const
    {
    int count = 0;
    sw_abstract_win **vp = v + x + (y*sw_size::xs());
    int max = sw_size::xs()-x;
    while( max-- && w == *vp++)
        count++;
    return count;
    }

int sw_map::not_belong_count( const sw_abstract_win *w, int x, int y ) const
    {
    int count = 0;
    sw_abstract_win **vp = v + x + (y*sw_size::xs());
    int max = sw_size::xs()-x;
    while( max-- && w != *vp++)
        count++;
    return count;
    }

