
#ifndef SW_EVENT
#define SW_EVENT

#include <style.h>

class sw_key // used as a namespace for key definitions
    {
    public:
        enum k
            {
            tab = 0x0F, ins = 0x52, del = 0x53, bs = 0x0e,
            space = 0x39, enter = 0x1C, esc = 0x01,
            
            left = 0x4B, right = 0x4D, up = 0x48, down = 0x50,
            cleft = 0x73, cright = 0x74, cup = 0x8D, cdown = 0x91,

            pgup = 0x49, pgdn = 0x51, home = 0x47, end = 0x4F,
            
            f1 = 0x3B, f2 = 0x3C, f3 = 0x3D, f4 = 0x3E, f5 = 0x3F, f6 = 0x40,
            f7 = 0x41, f8 = 0x42, f9 = 0x43, f10 = 0x44, f11 = 0x85, f12 = 0x86,

            cf1 = 0x5E,
            cf11 = 0x89,

            sf1 = 0x54,
            sf11 = 0x87,

            af1 = 0x68,
            af11 = 0x8B

            
            };
    };
    


class sw_event
    {
    public:
          // Delivery:
          //   To focus until confirmed
          //   To focus, forced
          //   Broadcast until confirmed
          //   Broadcast, forced
        enum md { Focus = 0x000, FFocus = 0x001, Broad = 0x020, FBroad = 0x021 };
        enum et { Key, Mouse, Info };
        enum it { NotInFocus, InFocus };
        enum st
            {
            RShift = 1, LShift = 2,  Control = 0x0004, Alt = 0x0008,
            Scrolllock_on = 0x0010, Numlock_on = 0x0020, Capslock_on = 0x0040,
            Insert_on = 0x0080, LControl = 0x0100, LAlt = 0x0200,
            RControl = 0x0400, RAlt = 0x0800,
            Scrolllock = 0x1000, Numlock = 0x2000,
            Capslock = 0x4000, Sysreq = 0x8000
            };

    private:
        md             mode_v;
        et             type_v;
        it             info_v;
        char           akey_v;
        short          vkey_v;
        st             shifts_v;
    public:
        sw_event( it info, md mode = Focus )
            : info_v(info), type_v(Info), mode_v(mode) {}
        
        sw_event( char a, short v, short s, md mode = Focus )
            : akey_v(a), vkey_v(v), type_v(Key), shifts_v((st)s), mode_v(mode) {}


        int operator == ( const sw_event &e ) const;

        md             mode()   const { return mode_v; }
        int            broad()  const { return ((int)mode_v) & 0x02; }
        int            forced() const { return ((int)mode_v) & 0x01; }
        
        et             type() const { return type_v; }

        it             info() const { return info_v; }

        virtual short  fkey() const
            { return type_v == Key ? vkey_v : 0; }

        virtual char   akey() const 
            { return type_v == Key ? akey_v : 0; }

        virtual bool   shift() const 
            { return shifts_v & (RShift|LShift) ? Yes : No; }

    };


#endif // SW_EVENT

