#ifndef __MY_CLASS_LIB_Imp

#define  __MY_CLASS_LIB_Imp

#ifdef __MULTI__
#ifdef __DEMO_SEMAPHORE__
#ifdef _INCLUDE_DEMO_SEMAPHORE_ROUTINE_

#include <time.h>

void SemaphoreCreate( unsigned long *sem )
{
  char sname[ 128 ];
  APIRET rc;

  for( rc=ERROR_DUPLICATE_NAME; rc==ERROR_DUPLICATE_NAME;
       rc = DosCreateMutexSem( (PSZ)sname, sem, DC_SEM_SHARED, 0 ) )
  {
    strcpy( sname, "\\SEM32\\ClassLib\\" );
    ltoa( time( NULL ), sname+strlen(sname), 2 );
  }

  if( rc )
  {
    cerr << "Unable create semaphore: " << rc << endl;
    *sem = 0xffffffff;
  }
}

inline void SemaphoreRemove( unsigned long sem )
{
  APIRET rc;

  if( sem!=0xffffffff )
    if( rc=DosCloseMutexSem( sem ), rc )
      cerr << "Unable remove semaphore: " << rc << endl;
}

inline void SemaphoreRequest( unsigned long sem )
{
  APIRET rc;

  if( sem!=0xffffffff )
    if( rc=DosRequestMutexSem( sem, SEM_INDEFINITE_WAIT ), rc )
      cerr << "Unable request semaphore: " << rc << endl;
}

void SemaphoreRelease( unsigned long sem )
{
  APIRET rc;

  if( sem!=0xffffffff )
    if( rc=DosReleaseMutexSem( sem ), rc )
      cerr << "Unable release semaphore: " << rc << endl;
}

#endif // _INCLUDE_DEMO_SEMAPHORE_ROUTINE_

#else // __DEMO_SEMAPHORE__

unsigned long  createMutexSemaphore( char *tag );
void           deleteMutexSemaphore( unsigned long sema );
void           lockMutexSemaphore( unsigned long sema );
void           unlockMutexSemaphore( unsigned long sema );

#endif // __DEMO_SEMAPHORE__

#endif // __MULTI__

#ifdef __MULTI__
extern char classLibLockNeed;
#endif

template <class T> OrdinalCollection<T>::OrdinalCollection( const int isize, const int idelta,
						     const SemaphoreMode sema,
						     const DeleteMode del ) :
  used( 0 ),
  size( isize ),
  delta( idelta ),
  deleteMode( del ),
  semaphore( 0xffffffffl ),
  needSema( sema )
{
  list = Allocate( size );

#ifdef __MULTI__
  if( sema==SemaphoreModeNeed )
#ifdef __DEMO_SEMAPHORE__
    SemaphoreCreate( &semaphore );
#else
  semaphore = createMutexSemaphore( T::nameOfClass() );
#endif
#endif
}

template <class T> OrdinalCollection<T>::~OrdinalCollection()
{
  Lock();

  if( deleteMode==DeleteModeDelete )
    for( int i=0; i<used; i++ )
      delete list[i];

  delete list;
  UnLock();

#ifdef __MULTI__
  if( needSema==SemaphoreModeNeed )
#ifdef __DEMO_SEMAPHORE__
    SemaphoreRemove( semaphore );
#else
    deleteMutexSemaphore( semaphore );
#endif
#endif
}

template <class T> void OrdinalCollection<T>::Lock() const
{
#ifdef __MULTI__
  if( needSema==SemaphoreModeNeed && classLibLockNeed )
#ifdef __DEMO_SEMAPHORE__
    SemaphoreRequest( semaphore );
#else
    lockMutexSemaphore( semaphore );
#endif
#endif
}

template <class T> void OrdinalCollection<T>::UnLock() const
{
#ifdef __MULTI__
  if( needSema==SemaphoreModeNeed && classLibLockNeed )
#ifdef __DEMO_SEMAPHORE__
    SemaphoreRelease( semaphore );
#else
    unlockMutexSemaphore( semaphore );
#endif
#endif
}

template <class T> T** OrdinalCollection<T>::Allocate( const int size ) const
{
  checkCondition( size!=0 )
  T** newList = new T* [size];
  memset( newList, 0, sizeof(T**)*size );
  return newList;
}

template <class T> void OrdinalCollection<T>::Expand()
{
  size += delta;
  T **newList = Allocate( size );

  memcpy( newList, list, sizeof(T**)*used );
  delete list;
  list = newList;
}

template <class T> void OrdinalCollection<T>::MoveRight( const int pos )
{
  checkCondition( pos<used && pos>=0 )

  memmove( list+pos+1, list+pos, sizeof(T**)*(used-pos) );
}

template <class T> void OrdinalCollection<T>::Add( T& item, const int pos )
{
  checkCondition( pos<=used && pos>=0 )
  checkCondition( used+1<=size )
  checkCondition( size!=0 )

  Lock();

  if( used+1==size )
    Expand();

  if( list[pos]!=0 )
    MoveRight( pos );

  list[pos] = &item;
  used++;
  UnLock();
}

template <class T> void OrdinalCollection<T>::Replace( T& item, DeleteMode del )
{
  Lock();

  T *found = Find( item, LockModeNoNeed );

  if( found!=NULL )
    Remove( *found, del );

  AddAtTail( item );
  UnLock();
}

template <class T> T& OrdinalCollection<T>::operator [] ( const int pos ) const
{
  checkCondition( pos<used && pos>=0 )
  return *list[pos];
}

template <class T> int OrdinalCollection<T>::GetIndex( T& item,
                                                              LockMode lock ) const
{
  if( !used )
    return -1;

  if( lock==LockModeNeed )
    Lock();

  int iter;

  for( iter=0; iter<used; iter++ )
    if( *(list[iter]) == item )
      break;

  if( lock==LockModeNeed )
    UnLock();

  return iter==used ? -1 : iter;
}

template <class T> void OrdinalCollection<T>::MoveLeft( const int ind )
{
  checkCondition( ind<used && ind>=0 )

  if( ind!=used-1 )
    memmove( list+ind, list+ind+1, sizeof(T**)*(used-1-ind) );
}

template <class T> void OrdinalCollection<T>::Remove( int ind,
                                                             DeleteMode del,
                                                             LockMode lock )
{
  checkCondition( ind>=0 && ind<used )

  if( lock==LockModeNeed )
    Lock();

  if( del==DeleteModeDelete ||
      (del==DeleteModeDefault && deleteMode==DeleteModeDelete ) )
    delete list[ind];

  MoveLeft( ind );

  --used;
  list[used] = 0;

  if( lock==LockModeNeed )
    UnLock();
}

template <class T> void OrdinalCollection<T>::Remove( T& item,
                                                             DeleteMode del,
                                                             LockMode lock )
{
  if( used )
  {
    if( lock==LockModeNeed )
      Lock();

    int ind = GetIndex( item, LockModeNoNeed );

    if( ind!=-1 )
    {
      if( del==DeleteModeDelete ||
          (del==DeleteModeDefault && deleteMode==DeleteModeDelete ) )
        delete list[ind];

      MoveLeft( ind );

      --used;
      list[used] = 0;
    }

    if( lock==LockModeNeed )
      UnLock();
  }
}

template <class T> void OrdinalCollection<T>::Flush( DeleteMode del )
{
  if( used )
  {
    Lock();

    while( used )
      Remove( used-1, del, LockModeNoNeed );

    UnLock();
  }
}
template <class T> T* OrdinalCollection<T>::Find( T& item,
                                                                LockMode lock ) const
{
  int ind = GetIndex( item, lock );

  return ind==-1 ? (T *)0 : list[ind];
}

template <class T> ostream & operator << ( ostream& o, const OrdinalCollection<T>& c )
{
  c.Lock();
  o << "  Total " << c.used << " items (sem=" ;

  if( c.semaphore==0xffffffffl )
    o << "none";
  else
    o << c.semaphore;

  o << char(')') << endl;

  for( int i=0; i<c.used; i++ )
    o << "  "
      << i
      << "> "
      << *(c.list[i])
      << endl;

  c.UnLock();
  return o;
}

#endif				//__MY_CLASS_LIB_Imp