#ifndef __MY_CLASS_LIB

#define  __MY_CLASS_LIB

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <builtin.h>

#ifdef __MULTI__
#define INCL_DOSSEMAPHORES
#define INCL_DOSERRORS
#define INCL_DOSPROCESS
#include <os2.h>
#endif

//#define __DEMO_SEMAPHORE__
//#define _INCLUDE_DEMO_SEMAPHORE_ROUTINE_

#ifdef __CLASS_DEBUG__
#ifndef __DEMO_SEMAPHORE__
#include "log.h"
#define checkCondition(p) if(!(p)) \
                  { \
                    void log( char * ... ); \
                    extern LOG *logPtr; \
                    log("*BUG* %d: %s", __LINE__, #p );\
                    if( logPtr!=NULL ) logPtr->flush(); \
                    __interrupt( 0x3 ); \
                  }
#else
#define checkCondition(p) \
                  if(!(p)) \
                  { \
                    printf("*BUG* %d: %s\n", __LINE__, #p );\
                    __interrupt( 0x3 ); \
                  }
#endif
#else
#define checkCondition(p) {}
#endif

#ifdef __MULTI__
extern char classLibLockNeed;
#endif

enum SemaphoreMode
{
  SemaphoreModeNoNeed = 0,
  SemaphoreModeNeed = 1
};

enum LockMode
{
  LockModeNoNeed = 0,
  LockModeNeed = 1
};

enum DeleteMode
{
  DeleteModeNoDelete = 0,
  DeleteModeDelete = 1,
  DeleteModeDefault = 2
};

template <class T> class OrdinalCollectionIterator;

template <class T> class OrdinalCollection
{
protected:

  T **list;
  int used,
      size,
      delta;
  DeleteMode deleteMode;
  SemaphoreMode needSema;
  unsigned long semaphore;

private:

  T** Allocate( const int size ) const;
  void Expand();
  void MoveRight( const int pos );
  void MoveLeft( const int ind );

public:

  OrdinalCollection( const int size=10, const int delta=5,
#ifdef __MULTI__
	      const SemaphoreMode needSema=SemaphoreModeNeed,
#else
	      const SemaphoreMode needSema=SemaphoreModeNoNeed,
#endif
	      const DeleteMode del=DeleteModeDelete );
  ~OrdinalCollection();

  void Lock() const;
  void UnLock() const;

  T* Find( T& item,
#ifdef __MULTI__
                  LockMode lock=LockModeNeed
#else
                  LockMode lock=LockModeNoNeed
#endif
                ) const;

  int GetIndex( T& item,
#ifdef __MULTI__
                  LockMode lock=LockModeNeed
#else
                  LockMode lock=LockModeNoNeed
#endif
                ) const;

  T& operator [] ( const int pos ) const;

  int HasMember( T& item ) const { return Find( item )!=0; }

  void Add( T& item, const int pos );
  void AddAtHead( T& item ) { Add( item, 0 ); }
  void AddAtTail( T& item ) { Lock(); Add( item, used ); UnLock(); }
  void Add( T& item ) { AddAtTail( item ); }

  void Replace( T& item, DeleteMode del=DeleteModeDefault );

  void Remove( T& item, DeleteMode del=DeleteModeDefault,
#ifdef __MULTI__
                      LockMode lock=LockModeNeed
#else
                      LockMode lock=LockModeNoNeed
#endif
                    );

  void Remove( int item, DeleteMode del=DeleteModeDefault,
#ifdef __MULTI__
                      LockMode lock=LockModeNeed
#else
                      LockMode lock=LockModeNoNeed
#endif
                    );

  void Flush( DeleteMode del=DeleteModeDefault );

  int GetItemsInContainer() const { return used; }

  friend ostream & operator << ( ostream& o, const OrdinalCollection<T>& c );
  friend OrdinalCollectionIterator<T>;
};

template <class T> class OrdinalCollectionIterator
{
private:
  int current;
  OrdinalCollection<T>& c;

public:

  OrdinalCollectionIterator( OrdinalCollection<T>& _c ) : current(0), c( _c ) { _c.Lock(); }
  ~OrdinalCollectionIterator() { c.UnLock(); }

  operator int() const { return current<c.used; }
  T& operator () () const { return c[current]; }
  void operator ++() { ++current; }
  void operator ++( int ) { ++current; }
  void restart() { current=0; }
};

#endif

