#ifndef __REREXP_H

#define __REREXP_H

#include <stddef.h>

/*
 * Definitions etc. for regexp(3) routines.
 *
 * Caveat:  this is V8 regexp(3) [actually, a reimplementation thereof],
 * not the System V one.
 */

#define RegExp_NSUBEXP 128

typedef struct regexp
{
  char *startp[RegExp_NSUBEXP];
  char *endp[RegExp_NSUBEXP];
  char regstart;		/* Internal use only. */
  char reganch;			/* Internal use only. */
  char *regmust;		/* Internal use only. */
  int regmlen;			/* Internal use only. */
  char program[1];		/* Unwarranted chumminess with compiler. */
} regexp;

class RegularExpresion
{
private:
  char *regparse;		/* Input-scan pointer. */
  int regnpar;			/* () count. */
  char regdummy;
  char *regcode;		/* Code-emit pointer; &regdummy = don't. */
  long regsize;			/* Code size. */
  char *regErrorMsg;
  char *reginput;		/* String-input pointer. */
  char *regbol;			/* Beginning of input, for ^ check. */
  char **regstartp;		/* Pointer to startp array. */
  char **regendp;		/* Ditto for endp. */
  regexp *mask;

  regexp *regcomp( char * );
  int regexec( regexp *, char * );
  void regsub(  regexp *, char *, char * );
  void regerror( char * );

  char *reg( int, int * );
  char *regbranch( int * );
  char *regpiece( int * );
  char *regatom( int * );
  char *regnode( char );
  char *regnext( char * );
  void reginsert( char, char * );
  void regtail( char *, char * );
  void regoptail( char *, char * );
  int regtry( regexp *, char * );
  int regmatch( char * );
  int regrepeat( char * );
  char *regprop();

  void RegularExpresion::regc( char b ) // emit (if appropriate) a byte of code
  {
    if( regcode != &regdummy )
      *regcode++ = b;
    else
      regsize++;
  }

public:
  RegularExpresion( char *str ) :
    regparse( NULL ),
    regnpar( 0 ),
    regdummy( 0 ),
    regcode( NULL ),
    regsize( 0 ),
    reginput( NULL ),
    regbol( NULL ),
    regstartp( NULL ),
    regendp( NULL ),
    regErrorMsg( NULL )
    { mask = regcomp( str ); }

  RegularExpresion( RegularExpresion& );

  ~RegularExpresion()
    { delete mask; }

  int maskMach( char *addr );
  char *matched( char *src );
  char *getRegErrorMsg() const { return regErrorMsg; }
  int isValid() const { return mask!=NULL; }
};

extern char *regErrorMsg;

#endif
