// p 㭪樨: ( PSZ name, ULONG nargs, PRXSTRING args, PSZ queue, PRXSTRING ret )

#define INCL_DOSMEMMGR
#define INCL_DOSMODULEMGR
#define INCL_DOSERRORS
#define INCL_DOSSEMAPHORES
#define INCL_DOSMISC
#define INCL_DOSMEMMGR
#define INCL_DOSPROCESS
#define INCL_DOSEXCEPTIONS
#include <os2.h>

#define INCL_REXXSAA
#define INCL_RXSHV
#include <rexxsaa.h>

#include <stdio.h>
#include <share.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <stdlib.h>
#include <process.h>
#include <builtin.h>

#include "language.h"

#include "misc.h"
#include "buffer.h"
#include "key.h"
#include "pair.h"
#include "timer.h"
#include "queue.h"
#include "zmodem.h"
#include "zinter.h"
#include "hydra.h"
#include "semaphor.h"
#include "translat.h"
#include "execute.h"
#include "rexx.h"
#include "rexxhook.h"
#include "winhand.h"
#include "session.h"
#include "globlvar.h"
#include "crc.h"
#include "except.h"
#include "fax.h"
#include "calls.h"
#include "keycd.h"
#include "process.h"
#include "nolog.h"
#include "addinqu.h"
#include "freq.h"
#include "heap.h"

char volatile rexxTransfer = 0;
char zeroString[2] = "0";

char *str( char *out )
{
  return out==NULL ? "<NONE>" : out;
}

static WindowHandleList * volatile wndList = NULL;
static GlobalVariableList * volatile globalVariableList = NULL;

char *rexxProcedureName( const char *prog )
{
  char *progFull = newStr( prog );

  for( int i=0; prog[i]; i++ )
    if( prog[i]==':' || prog[i]=='\\' || prog[i]=='/' )
      break;

  if( !prog[i] )
  {
    delete progFull;
    progFull = new char [ strlen(prog)+strlen(miscInfo.rexxPath)+1 ];
    strcpy( progFull, miscInfo.rexxPath );
    strcat( progFull, prog );
  }

  char drive[_MAX_DRIVE],
       dir[_MAX_DIR],
       name[_MAX_FNAME],
       ext[_MAX_EXT],
       *ret;

  _splitpath( progFull, drive, dir, name, ext );

  if( !*ext )
    strcat( ext, ".Brk" );

  ret = new char [ strlen(drive)+strlen(dir)+strlen(name)+strlen(ext)+1 ];

  strcpy( ret, drive );
  strcat( ret, dir );
  strcat( ret, name );
  strcat( ret, ext );

  delete progFull;

  capitalize( ret );
  return ret;
}

char *RXSIOHandlerString( PEXIT ParamBlock )
{
  register char *str( ((RXSIOSAY_PARM *)ParamBlock)->rxsio_string.strptr );
  int len( strlen(str) );

  if( len>=2 && str[len-1]=='\n' && str[len-2]=='\r' )
    str[len-2] = 0;

  return str;
}

void RXSIOHandlerStringEnter( char *title, PEXIT ParamBlock )
{
  register char *str( ((RXSIOSAY_PARM *)ParamBlock)->rxsio_string.strptr );
  register ULONG& len( ((RXSIOSAY_PARM *)ParamBlock)->rxsio_string.strlength );
  
  memset( str, 0, len );

  if( !terminalActive )
    +screenCursor;

  int i( stringEnter( len, str, title ) );

  if( !terminalActive )
    -screenCursor;

  str[len-1] = 0;

  if( i )
  {
    *str = 0;
    len = 0;
  }
  else
  {
    if( !strcmp( str, "  " ) )
      *str = 0;
      
    len = strlen( str );
  }
}

APIRET EXPENTRY RXSIOHandler( LONG ExitNumber, LONG Subfunction, PEXIT ParamBlock )
{
  registerExceptionHandler
  
  APIRET rc( RXEXIT_NOT_HANDLED );
  
  if( ExitNumber==RXSIO )
    switch( Subfunction )
    {
    case RXSIOSAY:
      log( "RexxSay: %s", RXSIOHandlerString( ParamBlock ) );
      rc = RXEXIT_HANDLED;
      break;

    case RXSIOTRC:
      log( "RexxError: %s", RXSIOHandlerString( ParamBlock ) );
      rc = RXEXIT_HANDLED;
      break;

    case RXSIOTRD:
      RXSIOHandlerStringEnter( Msg_RexxRXSIOTRD, ParamBlock );
      rc = RXEXIT_HANDLED;
      break;

    case RXSIODTR:
      RXSIOHandlerStringEnter( Msg_RexxRXSIODTR, ParamBlock );
      rc = RXEXIT_HANDLED;
      break;
    }
    
  deRegisterExceptionHandler
  
  return rc;
}

void RexxAddVariable( char *nameIni, char *val )
{
  SHVBLOCK request;
  char *name( tmpStr( nameIni ) );
  
  if( val==NULL )
    val = "";

  strupr( name );
  request.shvnext = NULL;
  request.shvname.strptr = name;
  request.shvname.strlength = strlen( name );
  request.shvvalue.strptr = val;
  request.shvvalue.strlength = strlen( val );
  request.shvnamelen = request.shvname.strlength;
  request.shvvaluelen = request.shvvalue.strlength;
  request.shvcode = RXSHV_SET;
  request.shvret = 0;

  APIRET rc = RexxVariablePool( &request );

  if( rc!=RXSHV_OK && rc!=RXSHV_NEWV )
    log( Msg_ErrRexxPoolAdd, name, val, rc );
}

void RexxAddVariable( char *name, long val )
{
  RexxAddVariable( name, _ltoa( val, (char *)_alloca(32), 10 ) );
}

void RexxAddVariable( char *name, unsigned long val )
{
  RexxAddVariable( name, _ultoa( val, (char *)_alloca(32), 10 ) );
}

inline void RexxAddVariable( char *name, int val )
{
  RexxAddVariable( name, (long)val );
}

inline void RexxAddVariable( char *name, unsigned val )
{
  RexxAddVariable( name, (unsigned long)val );
}

inline void RexxAddVariable( char *name, char *name1, char *val )
{
  RexxAddVariable( strcat( strcat( strcpy( (char *)_alloca( strlen(name)+
                                                            strlen(name1)+2 ), 
                                           name ), 
                                   "." ), 
                           name1 ),
                   val );
}

inline void RexxAddVariable( char *name, char *name1, unsigned long val )
{
  RexxAddVariable( name, name1, _ultoa( val, (char *)_alloca(32), 10 ) );
}

inline void RexxAddVariable( char *name, char *name1, long val )
{
  RexxAddVariable( name, name1, _ltoa( val, (char *)_alloca(32), 10 ) );
}

inline void RexxAddVariable( char *name, char *name1, int val )
{
  RexxAddVariable( name, name1, (long)val );
}

inline void RexxAddVariable( char *name, char *name1, unsigned val )
{
  RexxAddVariable( name, name1, (unsigned long)val );
}

void RexxAddVariable( char *name, int pos, char *val )
{
  RexxAddVariable( strcat( strcat( strcpy( (char *)_alloca( strlen(name)+32+2 ), 
                                           name ), 
                                   "." ),
                           _itoa( pos, (char *)_alloca(32), 10 ) ),
                   val );
}

void RexxAddVariable( char *name, int pos, char *val1, char *val2 )
{
  RexxAddVariable( name, pos, strcat( strcat( strcpy( 
                                                      (char *)_alloca( strlen(val1)+strlen(val2)+2 ), 
                                                      val1 ), 
                                              " " ),
                                      val2 ) );
}

void RexxAddVariable( char *name, int pos, int val )
{
  RexxAddVariable( name, pos, _itoa( val, (char *)_alloca(32), 10 ) );
}

char *rexxExportConfigHexString( char *buf, int attr )
{
  sprintf( buf, "%X", attr );
  return buf;
}

char *rexxExportConfigYesNoString( char *buf, int attr )
{
  return strcpy( buf, attr ? "YES" : "NO" );
}

char *rexxExportConfigStringPtr( char *buf, char *ptr )
{
  return strcpy( buf, ptr==NULL ? "" : ptr );
}

char *rexxExportConfigWorkTime( char *buf, WorkTime *wt )
{
  return rexxExportConfigStringPtr( buf, wt==NULL ? NULL : wt->getStr() );
}

char *rexxExportConfigFlavour( Flavour fl )
{
  return fl==unknownFlavour ? NULL : flavourName( fl );
}
void rexxExportConfigTimeoutCps( TimeOutList *timeoutList, char *scope, char *buf )
{
  if( timeoutList==NULL )
    return;
    
  int i(1);
  char val[65];
  
  strcpy( buf, "cfg.protocol." );
  strcat( buf, scope );
    
  if( timeoutList==NULL )
    RexxAddVariable( buf, 0, zeroString );
  else
  {
    for( TimeOutIterator list(*timeoutList); list; i++, list++ )
    {
      sprintf( val, "%d %d", list().getSpeed(), list().getTimeOut() );
      RexxAddVariable( buf, i, val );
    }
    
    RexxAddVariable( buf, 0, i-1 );
  }
}

void rexxExportConfigProcessFlag( char *pref, Schedule *list, int count, char *buf, size_t size )
{
  RexxAddVariable( pref, 0, count );
  
  if( list!=NULL )
    for( int i(0); i<count; i++ )
    {
      char ppp[80];
      
      sprintf( ppp, "%s.%d", pref, i+1 );
      RexxAddVariable( ppp, "action", stringSch( list+i, buf, size ) );
      RexxAddVariable( ppp, "path", list[i].name );
    }
}

void rexxExportConfigProcessTable( char *pref, ProcessList *table, char *buf, size_t size )
{
  int i(1);
  
  if( table!=NULL )
    for( ProcessIterator list( *table ); list; list++, i++ )
    {
      char ppp[80];
      SessionItem &cur( list() );
      
      sprintf( ppp, "%s.%d", pref, i );
      RexxAddVariable( ppp, "for", cur.toString( buf, size ) );
      RexxAddVariable( ppp, "action", stringSch( cur.getSchedule(), buf, size ) );
    }
  
  RexxAddVariable( pref, 0, i-1 );
}

void rexxExportConfigExitStart( char *pref, Schedule *list, int count, char *buf, int size )
{
  RexxAddVariable( pref, 0, count );
  
  if( list!=NULL )
    for( int i(0); i<count; i++ )
      RexxAddVariable( pref, i+1, stringSch( list+i, buf, size ) );
}

void rexxExportConfigExternal( char *pref, Pair *table, int count, char *buf, size_t size )
{
  RexxAddVariable( pref, 0, count );
  
  if( table!=NULL )
    for( int i(0); i<count; i++ )
      if( table[i].getSchedule() != NULL )
      {
        char ppp[80];
      
        sprintf( ppp, "%s.%d", pref, i+1 );
        RexxAddVariable( ppp, "for", table[i].getPattern() );
        RexxAddVariable( ppp, "action", stringSch( table[i].getSchedule(), buf, size ) );
      }
}

void rexxExportConfigAcceptCall( char *pref, EnableList *table, char *buf, int size )
{
  int i(1);
  
  if( table!=NULL )
    for( EnableIterator list( *table ); list; list++, i++ )
    {
      *buf = 0;
      RexxAddVariable( pref, i, list().toString( buf, size ) );
    }
  
  RexxAddVariable( pref, 0, i-1 );
}

void rexxExportConfigXlat( char *pref, short *xlt, char *buf, int )
{
  *buf = 0;

  for( int i(0); i<256; i++ )
  {
    register int n( xlt[ i ] );

    if( n != -1 )
      sprintf( buf+strlen(buf), "%02x ", n&0xff );
  }

  if( *buf )
    buf[ strlen(buf) ] = 0;

  RexxAddVariable( pref, "to", buf );

  *buf = 0;

  for( i=0; i<256; i++ )
    if( int( xlt[ i ] ) != -1 )
      sprintf( buf+strlen(buf), "%02x ", i );

  if( *buf )
    buf[ strlen(buf) ] = 0;

  RexxAddVariable( pref, "from", buf );
}

void rexxExportConfigXlatWords( char *pref, Pair *table, int how, Pair::DoType type )
{
  int cnt(0);

  while( how )
  {
    if( table->getType()==type )
      RexxAddVariable( pref, ++cnt, table->getOut() );

    table++;
    how--;
  }

  RexxAddVariable( pref, zeroString, cnt );
}

void rexxExportConfigPriority( char *buf, char *pref, unsigned long prio )
{
  strcpy( buf, "cfg.miscellaneous." );
  strcat( buf, pref );
  strcat( buf, "priority" );
  RexxAddVariable( buf, prio );
}

char *rexxExportConfigSortMode( char *buf, QueueSortMode mode )
{
  register char *p( "*Unknown*" );

  switch( mode )
  {
  case QueueSortNoSort:      p = "NoSort";
  case QueueSortAddr:        p = "+Address";
  case QueueSortAddrDec:     p = "-Address";
  case QueueSortSize:        p = "+Size";
  case QueueSortSizeDec:     p = "-Size";
  case QueueSortPriority:    p = "+Priority";
  case QueueSortPriorityDec: p = "-Priority";
  case QueueSortAge:         p = "+Age";
  case QueueSortAgeDec:      p = "-Age";
  }

  return strcpy( buf, p );
}

inline char *rexxExportConfigSeparator( char *buf, int separator )
{
  if( separator )
  {
    buf[0] = char( separator );
    buf[1] = 0;
  }
  else
    buf[0] = 0;

  return buf;
}

void rexxExportConfigLogScr( char *pref, NoLogList *noList )
{
  int i(1);
  
  if( noList!=NULL )
    for( NoLogIterator list( *noList ); list; list++, i++ )
      RexxAddVariable( pref, i, list().getStr() );
  
  RexxAddVariable( pref, 0, i-1 );
}

char *rexxExportConfigFileCase( char *buf, FileCase mode )
{
  register char *p( "*Unknown*" );

  switch( mode )
  {
  case FileCaseLower: p = "Lower";
  case FileCaseUpper: p = "Upper";
  case FileCaseMixed: p = "Mixed";
  case FileCaseAsIs:  p = "AsIs";
  }

  return strcpy( buf, p );
}

char *rexxExportConfigFreqMode( char *buf, FreqMode freqMode )
{
  char *p( "*Unknown*" );

  switch( freqMode )
  {
  case AllFreq:       p="YES";       break;
  case NoFreq:        p="NO";        break;
  case ListedFreq:    p="Listed";    break;
  case ProtectedFreq: p="Protected"; break;
  }

  return strcpy( buf, p );
}

void rexxExportConfigFreqPath( char *pref, int flag )
{
  int i(1);
  
  if( freqList!=NULL )
    for( AliasIterator list( *freqList ); list; list++ )
      if( list().isScan()==flag )
      {
        RexxAddVariable( pref, i, list().getPath() );
        i++;
      }
  
  RexxAddVariable( pref, 0, i-1 );
}

char *rexxExportConfigSchedule( Schedule *sh, char *buf, int bufSize )
{
  return strcpy( buf, sh==NULL ? "" : stringSch( sh, buf, bufSize ) );
}

void rexxExportConfigEvent( int i, Schedule& sh, char *buf, int bufSize )
{
  char pref[80];

  sprintf( pref, "cfg.event.%d", i );

  RexxAddVariable( pref, "type", sh.type==SchedTypeForce ? "Force" : "Event" );
  RexxAddVariable( pref, "name", sh.name );

  *buf = 0;

  if( sh.dayOfMonth )
    sprintf( buf+strlen(buf), "%d-", sh.dayOfMonth );
  
  if( sh.dayOfWeek )
    sprintf( buf, "%d.", sh.dayOfWeek );
  
  sprintf( buf+strlen(buf), "%02d:%02d:%02d", sh.start/60/60, (sh.start/60)%60,
                                              sh.start%60 );
    
  if( sh.isGmt )
    strcat( buf, "GMT" );
      
  RexxAddVariable( pref, "time", buf );
  RexxAddVariable( pref, "action", stringSch( &sh, buf, bufSize ) );
}

void rexxExportConfigNodeConfigList( char *stemPref, NodeConfigList *thelist,
                                     char *buf, int bufSize )
{
  if( !thelist || !thelist->GetItemsInContainer() )
  {
    RexxAddVariable( stemPref, zeroString, zeroString );
    return;
  }

  RexxAddVariable( stemPref, zeroString, thelist->GetItemsInContainer() );

  char *pref( (char *)_alloca( strlen( stemPref )+64 ) );
  int i( 1 );

  for( NodeConfigIterator list( *thelist ); list; list++, i++ )
  {
    NodeConfig& cur( list() );

    sprintf( pref, "%s.%d", stemPref, i );
    cur.toString( buf, bufSize, FIDOAddressShowModeEMSI );
    RexxAddVariable( pref, "node", buf );

    RexxAddVariable( pref, "immed", rexxExportConfigStringPtr( buf, cur.getImmedBox() ) );
    RexxAddVariable( pref, "hold", rexxExportConfigStringPtr( buf, cur.getHoldBox() ) );
    RexxAddVariable( pref, "direct", rexxExportConfigStringPtr( buf, cur.getDirectBox() ) );
    RexxAddVariable( pref, "normal", rexxExportConfigStringPtr( buf, cur.getNormalBox() ) );
    RexxAddVariable( pref, "crash", rexxExportConfigStringPtr( buf, cur.getCrashBox() ) );
    RexxAddVariable( pref, "inbound", rexxExportConfigStringPtr( buf, cur.getInbound() ) );
    RexxAddVariable( pref, "password", rexxExportConfigStringPtr( buf, cur.getPassword() ) );
    RexxAddVariable( pref, "rexxdial", rexxExportConfigStringPtr( buf, cur.getRexxDial() ) );
    RexxAddVariable( pref, "alias", rexxExportConfigStringPtr( buf, cur.getAlias() ) );
    RexxAddVariable( pref, "time", rexxExportConfigWorkTime( buf, cur.getWorkTime() ) );
    RexxAddVariable( pref, "priority", cur.getPriority() );
    RexxAddVariable( pref, "phone", rexxExportConfigStringPtr( buf, cur.getPhone() ) );
    RexxAddVariable( pref, "dial", rexxExportConfigStringPtr( buf, cur.getDial() ) );
    RexxAddVariable( pref, "dialsuf", rexxExportConfigStringPtr( buf, cur.getDialSuff() ) );
    RexxAddVariable( pref, "carrier", cur.getWaitCarrier() );
    RexxAddVariable( pref, "mode", rexxExportConfigFlavour( cur.getCfgFlavour() ) );

    char *newBuf( (char*)_alloca( strlen(pref)+80 ) );

    {
      FIDOAddressList *akaList( cur.getAkaList() );
      static char zeroAddress[]( "address.0" );

      if( akaList==NULL )
        RexxAddVariable( pref, zeroAddress, zeroString );
      else
      {
        strcpy( newBuf, pref );
        strcat( newBuf, ".address" );

        char *ptr( newBuf+strlen(newBuf) );
        int i( 1 );

        for( FIDOAddressIterator akas( *akaList ); akas; akas++, i++ )
        {
          sprintf( ptr, ".%d", i );
          RexxAddVariable( newBuf, akas().toString( buf, bufSize ) );
        }

        RexxAddVariable( pref, zeroAddress, i-1 );
      }
    }

    {
      HiddenList *hiddenList( cur.getHiddenList() );
      static char zeroHidden[]( "hidden.0" );

      if( hiddenList==NULL )
        RexxAddVariable( pref, zeroHidden, zeroString );
      else
      {
        strcpy( newBuf, pref );
        strcat( newBuf, ".hidden" );

        char *ptr( newBuf+strlen(newBuf) );
        int i( 1 );

        for( HiddenIterator hidden( *hiddenList ); hidden; hidden++, i++ )
        {
          Hidden& cur( hidden() );

          sprintf( ptr, ".%d", i );

          RexxAddVariable( newBuf, "carrier", cur.getWait() );
          RexxAddVariable( newBuf, "phone",
            rexxExportConfigStringPtr( buf, cur.getPhone() ) );
          RexxAddVariable( newBuf, "dial",
            rexxExportConfigStringPtr( buf, cur.getDial() ) );
          RexxAddVariable( newBuf, "dialsuf",
            rexxExportConfigStringPtr( buf, cur.getDialSuff() ) );
          RexxAddVariable( newBuf, "Time",
            rexxExportConfigWorkTime( buf, cur.getWorkTime() ) );
        }

        RexxAddVariable( pref, zeroHidden, i-1 );
      }
    }

    RexxAddVariable( pref, "flags", rexxExportConfigStringPtr( buf, cur.getFlags() ) );
    RexxAddVariable( pref, "CPS", cur.getBaud() );
    RexxAddVariable( pref, "chain", rexxExportConfigYesNoString( buf, cur.getChain() ) );
    RexxAddVariable( pref, "maxincoming", cur.getMaxInComing() );
    RexxAddVariable( pref, "maxoutgoing", cur.getMaxOutGoing() );
    RexxAddVariable( pref, "maxtry", cur.getMaxTry() );
    RexxAddVariable( pref, "cost", cur.getCost() );
    RexxAddVariable( pref, "dirzap", rexxExportConfigYesNoString( buf, !cur.disableDZA() ) );
    RexxAddVariable( pref, "zedzap", rexxExportConfigYesNoString( buf, !cur.disableZAP() ) );
    RexxAddVariable( pref, "hydra", rexxExportConfigYesNoString( buf, !cur.disableHydra() ) );
    RexxAddVariable( pref, "before", rexxExportConfigSchedule( cur.getBefore(), buf, bufSize ) );
    RexxAddVariable( pref, "after", rexxExportConfigSchedule( cur.getAfter(), buf, bufSize ) );
    RexxAddVariable( pref, "do", rexxExportConfigSchedule( cur.getDo(), buf, bufSize ) );
  }
}


void rexxExportConfig()
{
  char buf[4097];

  heapState( "rexxb" );
    
  RexxAddVariable( "cfg.emsi.sysop", systemInfo.sysopName );
  RexxAddVariable( "cfg.emsi.system", systemInfo.name );
  RexxAddVariable( "cfg.emsi.place", systemInfo.place );
  RexxAddVariable( "cfg.emsi.phone", systemInfo.phone );
  RexxAddVariable( "cfg.emsi.speed", systemInfo.baud );
  RexxAddVariable( "cfg.emsi.flags", systemInfo.flags );
    
  {
    int i(1);
    
    for( FIDOAddressIterator list(*systemInfo.addr); list; i++, list++ )
      RexxAddVariable( "cfg.emsi.address", i, list().toString( buf, sizeof buf ) );
    
    RexxAddVariable( "cfg.emsi.address", 0, i-1 );
  }
    
  RexxAddVariable( "cfg.fax.inbound", faxConfig.inbound );
  RexxAddVariable( "cfg.fax.baud", faxConfig.baud );
    
  sprintf( buf, "%d %d", colorSet.x, colorSet.y );
  RexxAddVariable( "cfg.video.videomode", buf );
  sprintf( buf, "%d %d %d %d", miscInfo.x, miscInfo.y, miscInfo.cx, miscInfo.cy );
  RexxAddVariable( "cfg.video.frame", buf );
    
  extern char *switchListTitle;

  RexxAddVariable( "cfg.video.title", switchListTitle );
  RexxAddVariable( "cfg.video.status", rexxExportConfigHexString( buf, colorSet.status ) );
  RexxAddVariable( "cfg.video.statuslampon", rexxExportConfigHexString( buf, colorSet.statusLampOn ) );
  RexxAddVariable( "cfg.video.statuslampoff", rexxExportConfigHexString( buf, colorSet.statusLampOff ) );
  RexxAddVariable( "cfg.video.border", rexxExportConfigHexString( buf, colorSet.border ) );
  RexxAddVariable( "cfg.video.action", rexxExportConfigHexString( buf, colorSet.action ) );
  RexxAddVariable( "cfg.video.actionbold", rexxExportConfigHexString( buf, colorSet.actionBold ) );
  RexxAddVariable( "cfg.video.abort", rexxExportConfigHexString( buf, colorSet.abort ) );
  RexxAddVariable( "cfg.video.popup", rexxExportConfigHexString( buf, colorSet.popUp ) );
  RexxAddVariable( "cfg.video.popupitem", rexxExportConfigHexString( buf, colorSet.popUpItem ) );
  RexxAddVariable( "cfg.video.popupbold", rexxExportConfigHexString( buf, colorSet.popUpBold ) );
  RexxAddVariable( "cfg.video.terminal", rexxExportConfigHexString( buf, colorSet.terminal ) );
  RexxAddVariable( "cfg.video.help", rexxExportConfigHexString( buf, colorSet.help ) );
  RexxAddVariable( "cfg.video.helpbold", rexxExportConfigHexString( buf, colorSet.helpBold ) );
  RexxAddVariable( "cfg.video.queuecrash", rexxExportConfigHexString( buf, colorSet.queueCrash ) );
  RexxAddVariable( "cfg.video.queuenormal", rexxExportConfigHexString( buf, colorSet.queueNormal ) );
  RexxAddVariable( "cfg.video.queuehold", rexxExportConfigHexString( buf, colorSet.queueHold ) );
  RexxAddVariable( "cfg.video.queueimmed", rexxExportConfigHexString( buf, colorSet.queueImmed ) );
  RexxAddVariable( "cfg.video.queuedirect", rexxExportConfigHexString( buf, colorSet.queueDirect ) );
  RexxAddVariable( "cfg.video.queueptr", rexxExportConfigHexString( buf, colorSet.queuePtr ) );
  RexxAddVariable( "cfg.video.modemto", rexxExportConfigHexString( buf, colorSet.modemTo ) );
  RexxAddVariable( "cfg.video.modemfrom", rexxExportConfigHexString( buf, colorSet.modemFrom ) );
  RexxAddVariable( "cfg.video.modemctlto", rexxExportConfigHexString( buf, colorSet.modemCtlTo ) );
  RexxAddVariable( "cfg.video.modemctlfrom", rexxExportConfigHexString( buf, colorSet.modemCtlFrom ) );
  RexxAddVariable( "cfg.video.color", rexxExportConfigYesNoString( buf, colorSet.isColor ) );
  RexxAddVariable( "cfg.video.screensaver", miscInfo.saverInterval );
  RexxAddVariable( "cfg.video.screensaverModule", miscInfo.saverModule );
    
  RexxAddVariable( "cfg.security.passwordabort", rexxExportConfigYesNoString( buf, miscInfo.state&PasswordAbort ) );
    
  RexxAddVariable( "cfg.protocol.chattimeout", miscInfo.chatTimeout );
  RexxAddVariable( "cfg.protocol.zstartblock", miscInfo.zBlock );
  RexxAddVariable( "cfg.protocol.zbasetimeout", miscInfo.timeOut );
  RexxAddVariable( "cfg.protocol.zcheckcpsafter", zTransferCpsAfter );
  RexxAddVariable( "cfg.protocol.dirzap", rexxExportConfigYesNoString( buf, miscInfo.state2&DirZapAvail ) );
  RexxAddVariable( "cfg.protocol.hydra", rexxExportConfigYesNoString( buf, miscInfo.state2&HydraAvail ) );
  RexxAddVariable( "cfg.protocol.hydrachat", rexxExportConfigYesNoString( buf, miscInfo.state&HydraChat ) );
  RexxAddVariable( "cfg.protocol.hbasetimeout", hydra_timeout );
  RexxAddVariable( "cfg.protocol.hcheckcpsafter", hTransferCpsAfter );
  RexxAddVariable( "cfg.protocol.htxwindow", hydra_txwindow );
  RexxAddVariable( "cfg.protocol.hrxwindow", hydra_rxwindow );
    
  hydraPutFlags( buf, hydra_options );
  RexxAddVariable( "cfg.protocol.hoption", buf );
    
  if( noHydraChatList==NULL )
    RexxAddVariable( "cfg.protocol.nostarthydrachat.0", zeroString );
  else
  {
    int i(1);
    
    for( NoHydraChatIterator list(*noHydraChatList); list; i++, list++ )
      RexxAddVariable( "cfg.protocol.nostarthydrachat", i, list().getStr() );
      
    RexxAddVariable( "cfg.protocol.nostarthydrachat", 0, i-1 );
  }
    
  rexxExportConfigTimeoutCps( timeOuts, "ztimeout", buf );
  rexxExportConfigTimeoutCps( zMinOutCps, "zminoutcps", buf );
  rexxExportConfigTimeoutCps( zMinInCps, "zminincps", buf );
  rexxExportConfigTimeoutCps( timeOutsHydra, "htimeout", buf );
  rexxExportConfigTimeoutCps( hMinOutCps, "hminoutcps", buf );
  rexxExportConfigTimeoutCps( hMinInCps, "hminincps", buf );
  
  if( schedList==NULL )
    RexxAddVariable( "cfg.event.0", zeroString );
  else
  {
    RexxAddVariable( "cfg.event.0", schedCount );
    
    for( int i(0); i<schedCount; i++ )
      rexxExportConfigEvent( i+1, schedList[i], buf, sizeof buf );
  }
  
  RexxAddVariable( "cfg.melody.connect", stringSch( melodies.connect, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.beginsession", stringSch( melodies.begSess, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.endsession", stringSch( melodies.endSess, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.badsession", stringSch( melodies.fail, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.endtransfer", stringSch( melodies.abort, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.human", stringSch( melodies.human, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.errorabort", stringSch( melodies.error, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.ring", stringSch( melodies.ring, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.alarm", stringSch( melodies.alarm, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.start", stringSch( melodies.start, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.stop", stringSch( melodies.stop, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.bell", stringSch( melodies.bell, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.chat", stringSch( melodies.chat, buf, sizeof buf ) );
  RexxAddVariable( "cfg.melody.audiopath", miscInfo.wavePath );
  RexxAddVariable( "cfg.melody.volume", miscInfo.volume );
  RexxAddVariable( "cfg.melody.nosound", stringWorkTime( miscInfo.noSound, buf, sizeof buf ) );
  
  RexxAddVariable( "cfg.split.circle", miscInfo.splitCircle );
  RexxAddVariable( "cfg.split.minechosize", miscInfo.splitSize );
  RexxAddVariable( "cfg.split.minpktsize", miscInfo.splitPktSize );
  RexxAddVariable( "cfg.split.maxfilesize", miscInfo.splitFileSize );
  RexxAddVariable( "cfg.split.delete", rexxExportConfigYesNoString( buf, miscInfo.state&SplitDelete ) );
  RexxAddVariable( "cfg.split.pktfromlo", rexxExportConfigYesNoString( buf, miscInfo.state&SplitPKTFromLO ) );
  RexxAddVariable( "cfg.split.longnames", rexxExportConfigYesNoString( buf, miscInfo.state2&SplitterLong ) );
  RexxAddVariable( "cfg.split.removenotfound", rexxExportConfigYesNoString( buf, miscInfo.state2&SplitDelNoFound ) );

  RexxAddVariable( "cfg.modem.port", miscInfo.port );
  RexxAddVariable( "cfg.modem.maxrring", miscInfo.maxRRing );
  RexxAddVariable( "cfg.modem.autoinit", miscInfo.autoInit );
  RexxAddVariable( "cfg.modem.ringcount", miscInfo.ringCount );
  RexxAddVariable( "cfg.modem.ringwait", miscInfo.ringWait );
  
  if( arqList==NULL )
    RexxAddVariable( "cfg.modem.arq.0", zeroString );
  else
  {
    int i(1);
    
    for( CallsIterator list(*arqList); list; i++, list++ )
      RexxAddVariable( "cfg.modem.arq", i, list().getFlag() );
      
    RexxAddVariable( "cfg.modem.arq", 0, i-1 );
  }
  
  if( halfDuplexList==NULL )
    RexxAddVariable( "cfg.modem.halfduplex.0", zeroString );
  else
  {
    int i(1);
    
    for( CallsIterator list(*halfDuplexList); list; i++, list++ )
      RexxAddVariable( "cfg.modem.halfduplex", i, list().getFlag() );
      
    RexxAddVariable( "cfg.modem.halfduplex", 0, i-1 );
  }
  
  RexxAddVariable( "cfg.modem.waitcall", portInfo.callWait );
  RexxAddVariable( "cfg.modem.waitanswer", portInfo.answerWait );
  
  if( ignoreString==NULL )
    RexxAddVariable( "cfg.modem.ignore.0", zeroString );
  else
  {
    int i(1);
    
    for( ModemStringIterator list(*ignoreString); list; i++, list++ )
      RexxAddVariable( "cfg.modem.ignore", i, list().getString() );
      
    RexxAddVariable( "cfg.modem.ignore", 0, i-1 );
  }
  
  if( badCounterString==NULL )
    RexxAddVariable( "cfg.modem.badcounter.0", zeroString );
  else
  {
    int i(1);
    
    for( ModemStringIterator list(*badCounterString); list; i++, list++ )
      RexxAddVariable( "cfg.modem.badcounter", i, list().getString() );
      
    RexxAddVariable( "cfg.modem.badcounter", 0, i-1 );
  }
  
  RexxAddVariable( "cfg.modem.failsafe", portInfo.failSafe );
  RexxAddVariable( "cfg.modem.delay", portInfo.pause );
  RexxAddVariable( "cfg.modem.baud", portInfo.speed );
  RexxAddVariable( "cfg.modem.init", portInfo.init );
  RexxAddVariable( "cfg.modem.answer", portInfo.answer );
  RexxAddVariable( "cfg.modem.hangup", portInfo.hangup );
  RexxAddVariable( "cfg.modem.dialprefix", portInfo.pref );
  RexxAddVariable( "cfg.modem.dialsuffix", portInfo.suff );
  RexxAddVariable( "cfg.modem.afterdial", portInfo.after );
  RexxAddVariable( "cfg.modem.abortdial", portInfo.dialterm );
  RexxAddVariable( "cfg.modem.onexit", portInfo.onExit );
  RexxAddVariable( "cfg.modem.statistics.command", modemString.statCmd );
  RexxAddVariable( "cfg.modem.statistics.stop", modemString.statStop );
  RexxAddVariable( "cfg.modem.fax", modemString.fax );
  RexxAddVariable( "cfg.modem.ring", modemString.ring );
  RexxAddVariable( "cfg.modem.busy", modemString.busy );
  RexxAddVariable( "cfg.modem.ok", modemString.ok );
  RexxAddVariable( "cfg.modem.connect", modemString.connect );
  RexxAddVariable( "cfg.modem.callersid", modemString.callersId );
  RexxAddVariable( "cfg.modem.nodialtone", modemString.nodialtone );
  RexxAddVariable( "cfg.modem.rring", modemString.rring );
  RexxAddVariable( "cfg.modem.nocarrier", modemString.nocarrier );
  RexxAddVariable( "cfg.modem.error", modemString.error );
  RexxAddVariable( "cfg.modem.voice", modemString.voice );
  RexxAddVariable( "cfg.modem.noanswer", modemString.noanswer );
  RexxAddVariable( "cfg.modem.noconnect", modemString.noconnect );
  RexxAddVariable( "cfg.modem.monitor", portInfo.carrierDelay );
  RexxAddVariable( "cfg.modem.flowcontrol", portInfo.hardwareFlow  ? "HardWare" : "SoftWare" );
  RexxAddVariable( "cfg.modem.breakmode", rexxExportConfigYesNoString( buf, portInfo.breakMode ) );
  RexxAddVariable( "cfg.modem.modemreport", rexxExportConfigYesNoString( buf, miscInfo.state&ModemReport ) );
  RexxAddVariable( "cfg.modem.virtualdcd", rexxExportConfigYesNoString( buf, portInfo.virtualDCD ) );
  RexxAddVariable( "cfg.modem.window", rexxExportConfigYesNoString( buf, miscInfo.state&ModemWindow ) );
  
  extern char startInputQueueMonitor;
  
  RexxAddVariable( "cfg.modem.queuestats", rexxExportConfigYesNoString( buf, startInputQueueMonitor ) );
  RexxAddVariable( "cfg.modem.collision", rexxExportConfigYesNoString( buf, miscInfo.state&Collision ) );
  
  if( preffixList==NULL )
    RexxAddVariable( "cfg.modem.special.0", zeroString );
  else
  {
    int i(1);
    
    for( CallsIterator list(*preffixList); list; i++, list++ )
      RexxAddVariable( "cfg.modem.special", i, list().getFlag(), list().getPref() );
      
    RexxAddVariable( "cfg.modem.special", 0, i-1 );
  }
  
  if( suffixList==NULL )
    RexxAddVariable( "cfg.modem.specialsuf.0", zeroString );
  else
  {
    int i(1);
    
    for( CallsIterator list(*suffixList); list; i++, list++ )
      RexxAddVariable( "cfg.modem.specialsuf", i, list().getFlag(), list().getPref() );
      
    RexxAddVariable( "cfg.modem.specialsuf", 0, i-1 );
  }
  
  sprintf( buf, "%d %d", miscInfo.wndModemX, miscInfo.wndModemY );
  RexxAddVariable( "cfg.modem.windowat", buf );
  RexxAddVariable( "cfg.modem.windowoff", miscInfo.wndModemOff );
  RexxAddVariable( "cfg.modem.mininbound", miscInfo.minIn );
  RexxAddVariable( "cfg.modem.minoutbound", miscInfo.minOut );
  RexxAddVariable( "cfg.modem.fuckingtelebit", rexxExportConfigYesNoString( buf, miscInfo.state1&Telebit ) );
  RexxAddVariable( "cfg.modem.hardwarering", rexxExportConfigYesNoString( buf, miscInfo.state3&HardRing ) );
  
  RexxAddVariable( "cfg.nodelist.path", outboundInfo.path );
  RexxAddVariable( "cfg.nodelist.outboundroot", outboundInfo.root );
  RexxAddVariable( "cfg.nodelist.indexes", outboundInfo.index );
  RexxAddVariable( "cfg.nodelist.defaultdomain", outboundInfo.defDomain->getName() );
  
  {
    char *exclStr="";
  
    if( miscInfo.exclude!=NULL )
    {
      exclStr = tmpStr( miscInfo.exclude+1 );
      exclStr[ strlen(exclStr)-1 ] = 0;
    }
    
    RexxAddVariable( "cfg.nodelist.exclude", exclStr );
  }
  
  {
    int i(1);
  
    for( DomainIterator list( *outboundInfo.domains ); list; list++, i++ )
    {
      Domain &cur( list() );
      char pref[80];
    
      sprintf( pref, "cfg.nodelist.domain.%d", i );
    
      RexxAddVariable( pref, "name", cur.getName() );
      RexxAddVariable( pref, "outbound", 
                       rexxExportConfigStringPtr( buf, cur.getOutboundRel() ) );
      RexxAddVariable( pref, "inbound", 
                       rexxExportConfigStringPtr( buf, cur.getInbound() ) );
      RexxAddVariable( pref, "protected", 
                       rexxExportConfigStringPtr( buf, cur.getProtected() ) );
      RexxAddVariable( pref, "mainzone", cur.getDefZone() );
      
      Buffer zones;
      
      for( int j(0); j<cur.getZLSize(); j++ )
      {
        char b[16];
        
        sprintf( b, "%d,", cur.getZList()[j] );
        zones.putString( b );
      }
      
      if( zones.length() )
        zones.stripByte();
        
      zones.putInto( 0 );
      RexxAddVariable( pref, "zones", zones.str );
      
      strcat( pref, ".nodelist" );
      
      if( !cur.getNdl() )
        RexxAddVariable( pref, 0, zeroString );
      else
      {
        int j(1);
        
        for( NDLIterator ndl( *cur.getNdl() ); ndl; ndl++, j++ )
          RexxAddVariable( pref, j, ndl().toString( buf, sizeof buf ) );
          
        RexxAddVariable( pref, 0, j-1 );
      }
    }
    
    RexxAddVariable( "cfg.nodelist.domain", 0, i-1 );
  }
  
  if( translationTable==NULL )
  {
    RexxAddVariable( "cfg.translate.local", "" );
    RexxAddVariable( "cfg.translate.country", "" );
    RexxAddVariable( "cfg.translate.international", "" );
    RexxAddVariable( "cfg.translate.0", zeroString );
  }
  else
  {
    RexxAddVariable( "cfg.translate.local", 
                     rexxExportConfigStringPtr( buf, translationTable->getLocal() ) );
    RexxAddVariable( "cfg.translate.country", 
                     rexxExportConfigStringPtr( buf, translationTable->getCountry() ) );
    RexxAddVariable( "cfg.translate.international", 
                     rexxExportConfigStringPtr( buf, translationTable->getInternational() ) );
                     
    int i(1);
    
    for( TranslationIterator list( *translationTable ); list; list++, i++ )
    {
      TransItem &cur( list() );
      char pref[80];
    
      sprintf( pref, "cfg.translate.%d", i );
      
      RexxAddVariable( pref, "area", rexxExportConfigStringPtr( buf, cur.getPreffix() ) );
      RexxAddVariable( pref, "prefix", rexxExportConfigStringPtr( buf, cur.getTranslate() ) );
      RexxAddVariable( pref, "suffix", rexxExportConfigStringPtr( buf, cur.getSuffix() ) );
      RexxAddVariable( pref, "cost", cur.getCost() );
    }
    
    RexxAddVariable( "cfg.translate.0", i-1 );
  }
  
  {
    int i(0);
    
    while( i<keyCount )
    {
      char pref[80];
    
      sprintf( pref, "cfg.keys.%d", i+1 );
      
      int j(0);
      
      while( j<codeTableSize ) 
      {
        if( keyList[i].done == codeTable[j].code )
          break;
          
        j++;
      }
      
      RexxAddVariable( pref, "key", j!=codeTableSize ? codeTable[j].name : "Undefined" );
      RexxAddVariable( pref, "action", stringSch( keyList+i, buf, sizeof buf ) );
      i++;
    }
    
    RexxAddVariable( "cfg.keys.0", i );
  }
  
  {
    char *tmpExitFlag( tmpStr( miscInfo.exitFlag ) );
    
    if( tmpExitFlag!=NULL )
    {
      char *ptr( strrchr( tmpExitFlag, '.' ) );
      
      if( ptr!=NULL )
        *ptr = 0;
    }
    
    RexxAddVariable( "cfg.process.exitflag", rexxExportConfigStringPtr( buf, tmpExitFlag ) );
  }
  
  RexxAddVariable( "cfg.process.multy", rexxExportConfigYesNoString( buf, !(miscInfo.state&ProcessOnce) ) );
  rexxExportConfigProcessTable( "cfg.process.online", onlineTable, buf, sizeof buf );
  rexxExportConfigProcessTable( "cfg.process.after", processTable, buf, sizeof buf );
  rexxExportConfigProcessTable( "cfg.process.immediate", immediateTable, buf, sizeof buf );
  rexxExportConfigProcessFlag( "cfg.process.flag", flagList, flagCount, buf, sizeof buf );
  rexxExportConfigProcessFlag( "cfg.process.noflag", noFlagList, noFlagCount, buf, sizeof buf );
  
  rexxExportConfigExitStart( "cfg.exit", exitList, exitCount, buf, sizeof buf );
  rexxExportConfigExitStart( "cfg.startup", startUpList, startUpCount, buf, sizeof buf );

  rexxExportConfigExternal( "cfg.external.connect", connectList, connectCount, buf, sizeof buf );
  rexxExportConfigExternal( "cfg.external.handshake", handShakeList, handShakeCount, buf, sizeof buf );
  RexxAddVariable( "cfg.external.remoteuser", rexxExportConfigStringPtr( buf, miscInfo.remoteName ) );
  RexxAddVariable( "cfg.external.remotepassword", miscInfo.remotePassword );
  RexxAddVariable( "cfg.external.bbs", rexxExportConfigYesNoString( buf, miscInfo.state1&BBSenable ) );
  RexxAddVariable( "cfg.external.autobbs", rexxExportConfigYesNoString( buf, miscInfo.state&AutoBBS ) );
  RexxAddVariable( "cfg.external.minbaud", miscInfo.minBBSBaud );
  RexxAddVariable( "cfg.external.time", stringWorkTime( miscInfo.BBSTime, buf, sizeof buf ) );
  RexxAddVariable( "cfg.external.runbbs", stringSch( processBBS, buf, sizeof buf ) );
  RexxAddVariable( "cfg.external.twiceesc", rexxExportConfigStringPtr( buf, bbsInfo.twiceEsc ) );
  RexxAddVariable( "cfg.external.mailonly", rexxExportConfigStringPtr( buf, bbsInfo.mailOnly ) );
  RexxAddVariable( "cfg.external.slowmodem", rexxExportConfigStringPtr( buf, bbsInfo.slowModem ) );
  RexxAddVariable( "cfg.external.mailnow", rexxExportConfigStringPtr( buf, bbsInfo.mailNow ) );
  RexxAddVariable( "cfg.external.loading", rexxExportConfigStringPtr( buf, bbsInfo.loading ) );
  RexxAddVariable( "cfg.external.goodbye", rexxExportConfigStringPtr( buf, bbsInfo.goodBye ) );
  
  {
    int i(0);
    
    while( i<actionCount )
    {
      char pref[80];
    
      sprintf( pref, "cfg.actions.%d", i+1 );
      
      RexxAddVariable( pref, "routine", actionList[i].name );
      RexxAddVariable( pref, "action", stringSch( actionList+i, buf, sizeof buf ) );
      i++;
    }
    
    RexxAddVariable( "cfg.actions.0", i );
  }

  rexxExportConfigAcceptCall( "cfg.calls.allowed", enabledList, buf, sizeof buf );
  rexxExportConfigAcceptCall( "cfg.calls.except", disabledList, buf, sizeof buf );

  rexxExportConfigAcceptCall( "cfg.accept.allowed", acceptedList, buf, sizeof buf );
  rexxExportConfigAcceptCall( "cfg.accept.except", unAcceptedList, buf, sizeof buf );
  RexxAddVariable( "cfg.accept.abort", rexxExportConfigYesNoString( buf, miscInfo.state1&NotAcceptAbort ) );

  if( rexxHookList==NULL )
    RexxAddVariable( "cfg.rexxhook.0", zeroString );
  else
  {
    int i(1);
    
    for( HookIterator list(*rexxHookList); list; i++, list++ )
    {
      sprintf( buf, "cfg.rexxhook.%d", i );
      RexxAddVariable( buf, "name", list().getHook() );
      RexxAddVariable( buf, "procedure", list().getProc() );
    }
      
    RexxAddVariable( "cfg.rexxhook", 0, i-1 );
  }

  rexxExportConfigXlat( "cfg.xlat.in", terminalInXLAT, buf, sizeof buf );
  rexxExportConfigXlat( "cfg.xlat.out", terminalOutXLAT, buf, sizeof buf );
  rexxExportConfigXlatWords( "cfg.xlat.answer",
                             ansiTab+ansiHowInternal, ansiHow-ansiHowInternal,
                             Pair::doModem );
  rexxExportConfigXlatWords( "cfg.xlat.display",
                             ansiTab+ansiHowInternal, ansiHow-ansiHowInternal,
                             Pair::doOut );

  sprintf( buf, "%d %d", miscInfo.circleMax, miscInfo.circleMin );
  RexxAddVariable( "cfg.miscellaneous.circle", buf );
  {
    int i(1);
    
    for( AddInQueueIterator list(*addInQueueList); list; i++, list++ )
    {
      sprintf( buf, "cfg.miscellaneous.addinqueue.%d", i );
      RexxAddVariable( buf, "name", list().getName() );
      RexxAddVariable( buf, "params", list().getParams() );
    }
      
    RexxAddVariable( "cfg.miscellaneous.addinqueue", 0, i-1 );
  }
  {
    int i(1);
  
    if( timeSyncList!=NULL )
      for( EnableIterator list( *timeSyncList ); list; list++, i++ )
      {
        *buf = 0;
        RexxAddVariable( "cfg.miscellaneous.timesync", i, list().toString( buf, sizeof buf ) );
      }
  
    RexxAddVariable( "cfg.miscellaneous.timesync.0", i-1 );
  }
  RexxAddVariable( "cfg.miscellaneous.trytosend", miscInfo.sendTry );
  RexxAddVariable( "cfg.miscellaneous.infilebuffersize", miscInfo.inFileBufferSize );
  RexxAddVariable( "cfg.miscellaneous.outfilebuffersize", miscInfo.outFileBufferSize );
  RexxAddVariable( "cfg.miscellaneous.holdtime", miscInfo.holdTime );
  RexxAddVariable( "cfg.miscellaneous.timesyncdelta", miscInfo.timeSyncDelta );
  RexxAddVariable( "cfg.miscellaneous.autorescan", miscInfo.rescan );
  RexxAddVariable( "cfg.miscellaneous.rescansleep", miscInfo.rescanPeriod );
  RexxAddVariable( "cfg.miscellaneous.immedpriority", immedFlavourPriority );
  RexxAddVariable( "cfg.miscellaneous.crashpriority", crashFlavourPriority );
  RexxAddVariable( "cfg.miscellaneous.directpriority", directFlavourPriority );
  RexxAddVariable( "cfg.miscellaneous.normalpriority", normalFlavourPriority );
  RexxAddVariable( "cfg.miscellaneous.holdpriority", holdFlavourPriority );
  RexxAddVariable( "cfg.miscellaneous.configcheck", miscInfo.configCheck/1000 );
  RexxAddVariable( "cfg.miscellaneous.logbuffer", miscInfo.logBuffer>>10 );
  RexxAddVariable( "cfg.miscellaneous.bsymaxage", miscInfo.bsyAge/60/60 );
  RexxAddVariable( "cfg.miscellaneous.bsycheck", miscInfo.bsyInterval );
  RexxAddVariable( "cfg.miscellaneous.maxcallcost", miscInfo.maxCost );
  RexxAddVariable( "cfg.miscellaneous.terminaltimeout", miscInfo.terminalTimeOut );
  RexxAddVariable( "cfg.miscellaneous.flagscheck", miscInfo.checkFlags );
  RexxAddVariable( "cfg.miscellaneous.schedulecheck", miscInfo.checkSched );
  RexxAddVariable( "cfg.miscellaneous.inbound", rexxExportConfigStringPtr( buf, miscInfo.inbound ) );
  RexxAddVariable( "cfg.miscellaneous.bsysemapath", rexxExportConfigStringPtr( buf, miscInfo.bsySemaPath ) );
  RexxAddVariable( "cfg.miscellaneous.protected", rexxExportConfigStringPtr( buf, miscInfo.protectedInbound ) );
  RexxAddVariable( "cfg.miscellaneous.flags", rexxExportConfigStringPtr( buf, miscInfo.flags ) );
  RexxAddVariable( "cfg.miscellaneous.rexxpath", rexxExportConfigStringPtr( buf, miscInfo.rexxPath ) );
  RexxAddVariable( "cfg.miscellaneous.terminalinbound", rexxExportConfigStringPtr( buf, miscInfo.terminalInbound ) );
  RexxAddVariable( "cfg.miscellaneous.sentList", rexxExportConfigStringPtr( buf, miscInfo.sentList ) );
  RexxAddVariable( "cfg.miscellaneous.tboxes", rexxExportConfigStringPtr( buf, miscInfo.tbox ) );
  RexxAddVariable( "cfg.miscellaneous.longboxes", rexxExportConfigStringPtr( buf, miscInfo.longBox ) );
  RexxAddVariable( "cfg.miscellaneous.longoutbound", rexxExportConfigStringPtr( buf, miscInfo.longOutbound ) );
  RexxAddVariable( "cfg.miscellaneous.hpfs", rexxExportConfigYesNoString( buf, miscInfo.state3&HPFS ) );
  RexxAddVariable( "cfg.miscellaneous.4dcompatible", rexxExportConfigYesNoString( buf, miscInfo.state&Full5D ) );
  RexxAddVariable( "cfg.miscellaneous.fastscan", rexxExportConfigYesNoString( buf, miscInfo.state3&FastScan ) );
  RexxAddVariable( "cfg.miscellaneous.scanonline", rexxExportConfigYesNoString( buf, miscInfo.state3&ScanOnLine ) );
  RexxAddVariable( "cfg.miscellaneous.scanbefore", rexxExportConfigYesNoString( buf, miscInfo.state&ScanBefore ) );
  RexxAddVariable( "cfg.miscellaneous.ea", rexxExportConfigYesNoString( buf, (miscInfo.state&NoEA)!=NoEA ) );
  RexxAddVariable( "cfg.miscellaneous.binkdrecover", rexxExportConfigYesNoString( buf, miscInfo.state&BinkDRecover ) );
  RexxAddVariable( "cfg.miscellaneous.exportconfig", rexxExportConfigYesNoString( buf, miscInfo.state&ExportConfig ) );
  RexxAddVariable( "cfg.miscellaneous.taskinlog", rexxExportConfigYesNoString( buf, miscInfo.state2&TaskInLog ) );
  RexxAddVariable( "cfg.miscellaneous.fulllog", rexxExportConfigYesNoString( buf, miscInfo.state2&FullLog ) );
  RexxAddVariable( "cfg.miscellaneous.statusline", rexxExportConfigYesNoString( buf, miscInfo.state1&StatusLine ) );
  RexxAddVariable( "cfg.miscellaneous.statusperiod", miscInfo.statusLine );
  RexxAddVariable( "cfg.miscellaneous.rexxdebug", rexxExportConfigYesNoString( buf, miscInfo.state2&RexxDebug ) );
  RexxAddVariable( "cfg.miscellaneous.retryafterextern", rexxExportConfigYesNoString( buf, miscInfo.state1&RetryAfterExtern ) );
  RexxAddVariable( "cfg.miscellaneous.bsysemaphore", rexxExportConfigYesNoString( buf, miscInfo.state2&SemaphoreBsy ) );
  RexxAddVariable( "cfg.miscellaneous.checkname", rexxExportConfigYesNoString( buf, miscInfo.state3&CheckName ) );
  RexxAddVariable( "cfg.miscellaneous.binklog", rexxExportConfigYesNoString( buf, miscInfo.state&BinkLog ) );
  RexxAddVariable( "cfg.miscellaneous.callafteranswer", rexxExportConfigYesNoString( buf, miscInfo.state&CallAfterAnswer ) );
  RexxAddVariable( "cfg.miscellaneous.terminallog", rexxExportConfigYesNoString( buf, miscInfo.state&TerminalLog ) );
  RexxAddVariable( "cfg.miscellaneous.zmh", rexxExportConfigWorkTime( buf, miscInfo.zmh ) );
  RexxAddVariable( "cfg.miscellaneous.calltime", rexxExportConfigWorkTime( buf, miscInfo.callTime ) );
  RexxAddVariable( "cfg.miscellaneous.answertime", rexxExportConfigWorkTime( buf, miscInfo.answerTime ) );
  RexxAddVariable( "cfg.miscellaneous.callonlyimmed", rexxExportConfigWorkTime( buf, miscInfo.callOnlyImmed ) );
  RexxAddVariable( "cfg.miscellaneous.callonlycrash", rexxExportConfigWorkTime( buf, miscInfo.callOnlyCrash ) );
  RexxAddVariable( "cfg.miscellaneous.callonlydirect", rexxExportConfigWorkTime( buf, miscInfo.callOnlyDirect ) );
  RexxAddVariable( "cfg.miscellaneous.callonlynormal", rexxExportConfigWorkTime( buf, miscInfo.callOnlyNormal ) );
  RexxAddVariable( "cfg.miscellaneous.redial", miscInfo.redial );
  RexxAddVariable( "cfg.miscellaneous.maxtry", miscInfo.maxTry );
  RexxAddVariable( "cfg.miscellaneous.waitmailer", miscInfo.EMSItime/1000 );
  RexxAddVariable( "cfg.miscellaneous.maxincoming", miscInfo.maxIn );
  RexxAddVariable( "cfg.miscellaneous.maxoutgoing", miscInfo.maxOut );
  RexxAddVariable( "cfg.miscellaneous.debug", rexxExportConfigYesNoString( buf, miscInfo.state&ZDebug ) );
  RexxAddVariable( "cfg.miscellaneous.emsidebug", rexxExportConfigYesNoString( buf, miscInfo.state&EMSIDebug ) );
  RexxAddVariable( "cfg.miscellaneous.emsilog", rexxExportConfigStringPtr( buf, miscInfo.emsiLog ) );
  RexxAddVariable( "cfg.miscellaneous.logname", rexxExportConfigStringPtr( buf, miscInfo.logName ) );
  RexxAddVariable( "cfg.miscellaneous.pipe", rexxExportConfigStringPtr( buf, miscInfo.pipe ) );
  RexxAddVariable( "cfg.miscellaneous.binarylog", rexxExportConfigStringPtr( buf, miscInfo.binLogName ) );
  RexxAddVariable( "cfg.miscellaneous.schedule", rexxExportConfigStringPtr( buf, miscInfo.schedule ) );
  RexxAddVariable( "cfg.miscellaneous.compiled", rexxExportConfigStringPtr( buf, miscInfo.schedComp ) );
  RexxAddVariable( "cfg.miscellaneous.threshold", miscInfo.trash );
  rexxExportConfigPriority( buf, "compiler", miscInfo.compilerPriority );
  RexxAddVariable( "cfg.miscellaneous.compilerprioritydelta", miscInfo.compilerPriorityDelta );
  rexxExportConfigPriority( buf, "scan", miscInfo.reScanPriority );
  RexxAddVariable( "cfg.miscellaneous.scanprioritydelta", miscInfo.reScanPriorityDelta );
  rexxExportConfigPriority( buf, "", miscInfo.priority );
  RexxAddVariable( "cfg.miscellaneous.foregrounddelta", miscInfo.priorityDelta );
  RexxAddVariable( "cfg.miscellaneous.discardpartialpkt", rexxExportConfigYesNoString( buf, miscInfo.state1&DiscardPKT ) );
  RexxAddVariable( "cfg.miscellaneous.abortonbsy", rexxExportConfigYesNoString( buf, miscInfo.state3&BsyAbort ) );
  RexxAddVariable( "cfg.miscellaneous.emptyexit", miscInfo.emptyExit );
  RexxAddVariable( "cfg.miscellaneous.dostime", rexxExportConfigYesNoString( buf, miscInfo.state1&AdjustTime ) );
  RexxAddVariable( "cfg.miscellaneous.holdoncall", rexxExportConfigYesNoString( buf, miscInfo.state1&HoldOnCall ) );
  RexxAddVariable( "cfg.miscellaneous.semaphore", rexxExportConfigStringPtr( buf, miscInfo.semaphore ) );
  RexxAddVariable( "cfg.miscellaneous.separator", rexxExportConfigSeparator( buf, miscInfo.separator ) );
  RexxAddVariable( "cfg.miscellaneous.queuesort", rexxExportConfigSortMode( buf, miscInfo.sortMode ) );
  RexxAddVariable( "cfg.miscellaneous.filecase", rexxExportConfigFileCase( buf, miscInfo.fileCase ) );
  rexxExportConfigLogScr( "cfg.miscellaneous.nolog", noLogList );
  rexxExportConfigLogScr( "cfg.miscellaneous.noscr", noScrList );
  RexxAddVariable( "cfg.miscellaneous.iconmail", rexxExportConfigStringPtr( buf, miscInfo.iconMail ) );
  RexxAddVariable( "cfg.miscellaneous.iconxmail", rexxExportConfigStringPtr( buf, miscInfo.iconXMail ) );
  RexxAddVariable( "cfg.miscellaneous.iconrip", rexxExportConfigStringPtr( buf, miscInfo.iconRIP ) );
  buf[0] = miscInfo.addrForm[0];
  buf[1] = 0;
  RexxAddVariable( "cfg.miscellaneous.addressform.domain", buf );
  RexxAddVariable( "cfg.miscellaneous.addressform.zeropoint", rexxExportConfigYesNoString( buf, miscInfo.addrForm[1] == 'P' ) );
  
  RexxAddVariable( "cfg.freq.freq", rexxExportConfigFreqMode( buf, miscInfo.freqMode ) );
  RexxAddVariable( "cfg.freq.sessiontime", miscInfo.sessionFreqTime );
  RexxAddVariable( "cfg.freq.sessionsize", miscInfo.sessionFreqSize );
  RexxAddVariable( "cfg.freq.daytime", miscInfo.dayFreqTime );
  RexxAddVariable( "cfg.freq.daysize", miscInfo.dayFreqSize );
  RexxAddVariable( "cfg.freq.weektime", miscInfo.weekFreqTime );
  RexxAddVariable( "cfg.freq.weeksize", miscInfo.weekFreqSize );
  RexxAddVariable( "cfg.freq.monthtime", miscInfo.monFreqTime );
  RexxAddVariable( "cfg.freq.monthsize", miscInfo.monFreqSize );
  RexxAddVariable( "cfg.freq.time", rexxExportConfigWorkTime( buf, miscInfo.freqTime ) );
  RexxAddVariable( "cfg.freq.time", rexxExportConfigWorkTime( buf, miscInfo.freqTime ) );
  rexxExportConfigFreqPath( "cfg.freq.path", 0 );
  rexxExportConfigFreqPath( "cfg.freq.tree", 1 );
  {
    int i(1);
  
    if( freqAlias!=NULL )
      for( AliasIterator list( *freqAlias ); list; list++, i++ )
      {
        sprintf( buf, "cfg.freq.alias.%d", i );
        RexxAddVariable( buf, "alias", list().getAlias() );
        RexxAddVariable( buf, "path", list().getPath() );
      }
  
    RexxAddVariable( "cfg.freq.alias.0", i-1 );
  }
  RexxAddVariable( "cfg.freq.basefile", rexxExportConfigStringPtr( buf, miscInfo.freqBase ) );
  RexxAddVariable( "cfg.freq.minbaud", miscInfo.minFreqBaud );
  RexxAddVariable( "cfg.freq.policy", rexxExportConfigStringPtr( buf, miscInfo.freqPolicy ) );
  RexxAddVariable( "cfg.freq.checkreq", rexxExportConfigYesNoString( buf, miscInfo.state3&CheckREQ ) );
  RexxAddVariable( "cfg.freq.Index", rexxExportConfigStringPtr( buf, miscInfo.freqIdx ) );

  rexxExportConfigNodeConfigList( "cfg.node", systemList, buf, sizeof buf );
  rexxExportConfigNodeConfigList( "cfg.mask", maskList, buf, sizeof buf );

  heapState( "rexxe" );
}

APIRET EXPENTRY RXINIHandler( LONG ExitNumber, LONG Subfunction, PEXIT )
{
  registerExceptionHandler
  
  APIRET rc( RXEXIT_NOT_HANDLED );
  
  if( ExitNumber==RXINI && Subfunction==RXINIEXT )
  {
    if( (miscInfo.state&ExportConfig) == ExportConfig )
      rexxExportConfig();

    rc = RXEXIT_HANDLED;
  }

  deRegisterExceptionHandler
  
  return rc;
}

char RXSIOHandlerName[] = "BrakeRexxIOWrapper";
char RXINIHandlerName[] = "BrakeRexxINIT";

RXSYSEXIT rexxExitList[] = 
{ 
  {RXSIOHandlerName, RXSIO}, 
  {RXINIHandlerName, RXINI}, 
  {NULL, RXENDLST} 
};

APIRET RegisterRexxHandler()
{
  APIRET rc( RexxRegisterExitExe( RXSIOHandlerName, (PFN)RXSIOHandler, NULL ) ? 1 : 0 );
  
  if( rc )
  {
    log( Msg_RexxErrUnableIORegister, rc );
    os2Message( rc );
  }
  
  rc |= RexxRegisterExitExe( RXINIHandlerName, (PFN)RXINIHandler, NULL ) ? 2 : 0;
  
  if( rc&2 )
  {
    log( Msg_RexxErrUnableInitRegister, rc );
    os2Message( rc );
  }
  
  return rc;
}

void DeRegisterRexxHandler( APIRET registerd )
{
  if( (registerd&1)!=1 )
  {
    APIRET rc = RexxDeregisterExit( RXSIOHandlerName, 0 );
  
    if( rc )
    {
      log( Msg_RexxErrUnableIODeRegister, rc );
      os2Message( rc );
    }
  }
  
  if( (registerd&2)!=2 )
  {
    APIRET rc = RexxDeregisterExit( RXINIHandlerName, 0 );
  
    if( rc )
    {
      log( Msg_RexxErrUnableInitDeRegister, rc );
      os2Message( rc );
    }
  }
}

int CallRexx( char *prog, char **out,
              char *a1, char *a2, char *a3,
              char *a4, char *a5, char *a6,
              int needLog )
{
  heapState( "rcalb" );

  RXSTRING arg[6];
  RXSTRING rexxretval;

  APIRET   rc;
  SHORT    rexxrc = 0;

  if( prog == NULL || strlen(prog) == 0 )
    return 0;

  char *progFull = rexxProcedureName( prog );

  if( access( progFull, 0 ) )
  {
    log( Msg_RexxNoFile, prog );
    delete progFull;
    heapState( "rcale" );
    return 0;
  }

  rexxretval.strlength = 0L;
  rexxretval.strptr = NULL;

  if( a1 == NULL )
    a1 = "";

  MAKERXSTRING( arg[0], a1, strlen(a1) )

  if( a2 == NULL )
    a2 = "";

  MAKERXSTRING( arg[1], a2, strlen(a2) )

  if( a3 == NULL )
    a3 = "";

  MAKERXSTRING( arg[2], a3, strlen(a3) )

  if( a4 == NULL )
    a4 = "";

  MAKERXSTRING( arg[3], a4, strlen(a4) )

  if( a5 == NULL )
    a5 = "";

  MAKERXSTRING( arg[4], a5, strlen(a5) )

  if( a6 == NULL )
    a6 = "";

  MAKERXSTRING( arg[5], a6, strlen(a6) )

  if( needLog )
    log( Msg_RexxExecute, prog,
         a1[0] ? " " : "", a1,
         a2[0] ? " " : "", a2,
         a3[0] ? " " : "", a3,
         a4[0] ? " " : "", a4,
         a5[0] ? " " : "", a5,
         a6[0] ? " " : "", a6 );

  if( wndList==NULL )
    wndList = new WindowHandleList;
    
//  APIRET exitNotActive( RegisterRexxHandler() );

  rc = RexxStart( (LONG)       6,             /* number of arguments   */
                  (PRXSTRING)  &arg,          /* array of arguments    */
                  (PSZ)        progFull,      /* name of REXX file     */
                  (PRXSTRING)  0,             /* No INSTORE used       */
                  (PSZ)        "Brake!",      /* Command env. name     */
                  (LONG)       RXSUBROUTINE,  /* Code for how invoked  */
                  (PRXSYSEXIT) rexxExitList,  /* EXITs list            */
                  (PSHORT)     &rexxrc,       /* Rexx program output   */
                  (PRXSTRING)  &rexxretval ); /* Rexx program output   */
	       
//  DeRegisterRexxHandler( exitNotActive );

  delete progFull;

  if( needLog )
    log( Msg_RexxReturned, str( rexxretval.strptr ),
         rc, (int)rexxrc );

  if( out!=NULL )
  {
    *out = newStr( rexxretval.strptr );
    DosFreeMem( rexxretval.strptr );          /* Release storage       */
  }

  heapState( "rcale" );

  return rc == 0 && rexxrc == 0? 1 : 0;
}

void RexxRetString( char *str, PRXSTRING ret, int lenFromArg=0 )
{
  int len = lenFromArg ? lenFromArg : strlen(str);
  APIRET rc;

  if( len>ret->strlength )
// ७ ,  ஥ 㦭 ᢮  
    if( rc=DosAllocMem( (PPVOID)&(ret->strptr), len+1, PAG_READ|PAG_WRITE|PAG_GUARD ), rc )
    {
      log( Msg_RexxErrRetString, rc );
      os2Message( rc );
      sprintf( ret->strptr, "FAIL: DosAllocMem error code %lu", rc );
      ret->strlength = strlen( ret->strptr );
      return;
    }

  if( lenFromArg )
    memcpy( ret->strptr, str, len );
  else
    strcpy( ret->strptr, str );

  ret->strlength = len;
}

void RexxFuncComplete( int ok, PRXSTRING ret )
{
  if( ok )
  {
    strcpy( ret->strptr, "OK" );
    ret->strlength = 2;
  }
  else
  {
    strcpy( ret->strptr, "FAIL" );
    ret->strlength = 4;
  }
}

int RexxCheckArgs( ULONG numarg, PRXSTRING args,
                   PRXSTRING ret, char *routine )
{
  for( int i=0; i<numarg; i++ )
    if( args[i].strptr==NULL )
    {
      log( Msg_RexxNullArg, i+1, routine );
      RexxFuncComplete( 0, ret );
      return 1;
    }

  return 0;
}

APIRET EXPENTRY RexxBrkComPutCode( PRXSTRING, ULONG numarg, PRXSTRING args,
				   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkComPutCode" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  ULONG *arr( NULL );
  APIRET rc( 0 );

  if( numarg )
  {
    arr = new ULONG [ numarg ];

    for( ULONG i=0; i<numarg; i++ )
    {
      char *end;

      arr[i] = strtoul( args[i].strptr, &end, 10 );

      if( *end || arr[i]>255 )
        goto Fail;
    }

    for( i=0; i<numarg; i++ )
      if( !comPort->put( (char)arr[i] ) )
      {
        log( Msg_RexxDebComPutCode, i );
        goto Unsucc;
      }

    log( Msg_RexxDebComPutCode1, i );
    RexxFuncComplete( 1, ret );
  }
  else
  {
Fail:
    log( Msg_RexxErrComPutCode );
    rc = 1;
Unsucc:
    RexxFuncComplete( 0, ret );
  }

  delete arr;
  deRegisterExceptionHandler
  return rc;
}

APIRET EXPENTRY RexxBrkComPutString( PRXSTRING, ULONG numarg, PRXSTRING args,
                                     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkComPutString" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  APIRET rc( 0 );

  if( numarg==1 )
  {
    register char *ptr = args[0].strptr;
    register ULONG len = args[0].strlength;

    for( ULONG i=0; i<len; i++, ptr++ )
      if( !comPort->put( *ptr ) )
      {
        log( Msg_RexxDebComPutString, i );
        goto Unsucc;
      }

    log( Msg_RexxDebComPutString1, i );
    RexxFuncComplete( 1, ret );
  }
  else
  {
    log( Msg_RexxErrComPutStr );
    rc = 1;
Unsucc:
    RexxFuncComplete( 0, ret );
  }

  deRegisterExceptionHandler
  return rc;
}

APIRET EXPENTRY RexxBrkComSetConnectString( PRXSTRING, ULONG numarg, PRXSTRING args,
                                            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkComSetConnectString" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrComSetConnStr );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  comPort->setConnectString( args[0].strptr );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComPutCommand( PRXSTRING, ULONG numarg, PRXSTRING args,
                                      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkComPutCommand" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }
  
  if( numarg!=1 )
  {
    log( Msg_RexxErrComPutCommand );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( comPort->putOffLine( args[0].strptr ) ? 0 : 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComFlushBuffer( PRXSTRING, ULONG, PRXSTRING,
                                       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->flushBufferedData();

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComDTROn( PRXSTRING, ULONG, PRXSTRING,
                                 PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->dtr( 1 );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComDTROff( PRXSTRING, ULONG, PRXSTRING,
                                  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->dtr( 0 );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComXonXoffOn( PRXSTRING, ULONG, PRXSTRING,
                                     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->xonxoff( 1 );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComXonXoffOff( PRXSTRING, ULONG, PRXSTRING,
                                     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->xonxoff( 0 );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComCtsRtsOn( PRXSTRING, ULONG, PRXSTRING,
                                    PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->ctsrts( 1 );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComCtsRtsOff( PRXSTRING, ULONG, PRXSTRING,
                                     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->ctsrts( 0 );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComClose( PRXSTRING, ULONG, PRXSTRING,
                                 PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->closePort();

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComOpen( PRXSTRING, ULONG, PRXSTRING,
                                PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->openPort() ? 0 : 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComIsOK( PRXSTRING, ULONG, PRXSTRING,
                                PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->ok() ? 1 : 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComIsOpened( PRXSTRING, ULONG, PRXSTRING,
                                    PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->isOpened() ? 1 : 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComIsHalfDuplex( PRXSTRING, ULONG, PRXSTRING,
                                        PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->isHalfDuplex() ? 1 : 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetTranx( PRXSTRING, ULONG, PRXSTRING,
                                    PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%d", (int)comPort->getTranx() );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetLastBaud( PRXSTRING, ULONG, PRXSTRING,
                                       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%d", (int)comPort->getLastBaud() );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComHasInput( PRXSTRING, ULONG, PRXSTRING,
                                     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->haveInput() ? 1 : 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComIsOnline( PRXSTRING, ULONG, PRXSTRING,
                                    PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->isOnline() ? 1 : 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComCheckOnline( PRXSTRING, ULONG, PRXSTRING,
                                       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->CheckOnline() ? 1 : 0, ret );  //check
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComIsARQ( PRXSTRING, ULONG, PRXSTRING,
                                 PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->isArq() ? 1 : 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetWaitCallTime( PRXSTRING, ULONG, PRXSTRING,
                                           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%d", (int)comPort->getWaitCallTime() );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetWaitAnswerTime( PRXSTRING, ULONG, PRXSTRING,
                                             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%d", (int)comPort->getWaitAnswerTime() );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComOutIsEmpty( PRXSTRING, ULONG, PRXSTRING,
                                      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->outEmpty() ? 1 : 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetDialPrefix( PRXSTRING, ULONG, PRXSTRING,
                                         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxRetString( comPort->getPrefix(), ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetDialSuffix( PRXSTRING, ULONG, PRXSTRING,
                                         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxRetString( comPort->getSuffix(), ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetConnect( PRXSTRING, ULONG, PRXSTRING,
                                      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  char *conn = comPort->getSuff();

  if( conn==NULL )
    conn = "*No Connect String*";

  RexxRetString( conn, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetSpeed( PRXSTRING, ULONG, PRXSTRING,
                                    PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%lu", (ULONG)comPort->getSpeed() );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetPort( PRXSTRING, ULONG, PRXSTRING,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  ULONG port = (ULONG)comPort->getPort();

  if( port==0 )
    port = (ULONG)-1;

  sprintf( ret->strptr, "%ld", port );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetHandle( PRXSTRING, ULONG, PRXSTRING,
                                     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  ULONG port = (ULONG)comPort->getHandle();

  if( port==0 )
    port = (ULONG)-1;

  sprintf( ret->strptr, "%ld", port );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetCode( PRXSTRING, ULONG, PRXSTRING,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%d", (int)comPort->get() );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetChar( PRXSTRING, ULONG, PRXSTRING,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  *(ret->strptr) = (unsigned char)comPort->get();
  ret->strlength = 1;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetStringTo( PRXSTRING, ULONG numarg, PRXSTRING args,
                                       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkComGetStringTo" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  // BrkComGetStringTo( [secs[, delim]] )

  char *waitStr = numarg>=1 ? args[0].strptr : NULL;
  char lastChar = char( numarg>=2 ? *args[1].strptr : 0 );
  long secs( 10 );

  if( waitStr )
  {
    char *e;

    secs = (int)strtol( args[0].strptr, &e, 10 );

    if( *e )
    {
      log( Msg_RexxErrComGetStringTo1, args[0].strptr );
      RexxFuncComplete( 0, ret );
      deRegisterExceptionHandler
      return 1;
    }
  }

  Buffer b;

  for( Timer t( secs*1000 ); !t.isTimeOut(); )
  {
    int ch = comPort->get();

    if( ch!=-1 )
    {
      b.putInto( char(ch) );

      if( numarg>=2 && ch==lastChar )
        break;
    }
  }

  b.putInto( 0 );
  RexxRetString( b.str, ret, b.length()-1 );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComGetString( PRXSTRING, ULONG numarg, PRXSTRING args,
                                     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkComGetString" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  // BrkComGetString( [secs[, prompt[, password]]] )

  int secs = 0,
      flag = 0;
  char *prompt = numarg>=2 ? args[1].strptr : NULL;

  if( numarg )
  {
    char *e;

    secs = (int)strtol( args[0].strptr, &e, 10 );

    if( *e )
    {
      log( Msg_RexxErrComGetString1, args[0].strptr );
Fail:
      RexxFuncComplete( 0, ret );
      deRegisterExceptionHandler
      return 1;
    }
  }

  if( numarg>=3 )
    if( !stricmp( args[2].strptr, "YES" ) )
      flag = 1;
    else if( !stricmp( args[2].strptr, "NO" ) )
      flag = 0;
    else
    {
      log( Msg_RexxErrComGetString2, args[0].strptr );
      goto Fail;
    }

  if( comPort->promptString( prompt, ret->strptr, int( ret->strlength-1 ),
                             secs*1000, flag ) )
    RexxFuncComplete( 0, ret );
  else
    ret->strlength = strlen( ret->strptr );

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComInit( PRXSTRING, ULONG, PRXSTRING,
                               PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->init() ? 0 : 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComHangUp( PRXSTRING, ULONG, PRXSTRING,
                                  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->hangUp() ? 0 : 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComWaitOK( PRXSTRING, ULONG, PRXSTRING,
                                  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->waitOk() ? 0 : 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComClearIn( PRXSTRING, ULONG, PRXSTRING,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->clearin();
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComClearOut( PRXSTRING, ULONG, PRXSTRING,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->clearout();
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComStartBreak( PRXSTRING, ULONG, PRXSTRING,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->do_break( 1 );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComEndBreak( PRXSTRING, ULONG, PRXSTRING,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  comPort->do_break( 0 );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComUnBufferBytes( PRXSTRING, ULONG, PRXSTRING,
                                         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( comPort->unBufferBytes() ? 0 : 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComBufferByte( PRXSTRING, ULONG numarg, PRXSTRING args,
                                      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkComBufferByte" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }
  
  ULONG *arr( NULL );
  APIRET rc( 0 );

  if( numarg )
  {
    arr = new ULONG [ numarg ];

    for( ULONG i=0; i<numarg; i++ )
    {
      char *end;

      arr[i] = strtoul( args[i].strptr, &end, 10 );

      if( *end || arr[i]>255 )
        goto Fail;
    }

    for( i=0; i<numarg; i++ )
      if( comPort->bufferByte( (char)arr[i] ) )
      {
        log( Msg_RexxDebComBufferByte, i );
        goto Unsucc;
      }

    log( Msg_RexxDebComBufferByte1, i );
    RexxFuncComplete( 1, ret );
  }
  else
  {
Fail:
    log( Msg_RexxErrComBufByte );
    rc = 1;
Unsucc:
    RexxFuncComplete( 0, ret );
  }

  delete arr;
  deRegisterExceptionHandler
  return rc;
}

APIRET EXPENTRY RexxBrkComAnswer( PRXSTRING, ULONG, PRXSTRING,
                                  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%d", (int)comPort->answer( 1 ) );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebComAnswer, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComDial( PRXSTRING, ULONG numarg, PRXSTRING arg,
				PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, arg, ret, "BrkComDial" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

// number, wait, blank, pref, suff
  char *pref  = NULL,
       *suff  = NULL;
  int blank = 0,
      wait  = 0;

  if( numarg>5 )
  {
    log( Msg_RexxErrComDial );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg>4 )
    suff = arg[4].strptr;

  if( numarg>3 )
    pref = arg[3].strptr;

  char *end,
       *msg;
  int num = 3;

  if( numarg>2 )
  {
    msg = arg[2].strptr;
    blank = (int)strtoul( msg, &end, 10 );

    if( *end )
    {
ill_num:
      log( Msg_RexxErrComDial1, num, msg );
      goto Unsucc;
    }
  }

  num = 2;

  if( numarg>1 )
  {
    msg = arg[1].strptr;
    wait = (int)strtoul( msg, &end, 10 );

    if( *end )
      goto ill_num;
  }

  if( !numarg )
  {
    log( Msg_RexxErrComDial2 );
    goto Unsucc;
  }

  sprintf( ret->strptr, "%d", (int)comPort->dial( arg[0].strptr, pref, suff,
						  wait, blank ) );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebComDial, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkComTrace( PRXSTRING, ULONG numarg, PRXSTRING args,
				 PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkComTrace" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  char *end;

  if( numarg<3 )
  {
    log( Msg_RexxErrComTrace );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  int num = 1;
  char *msg = args[0].strptr;
  int wait = (int)strtoul( msg, &end, 10 );

  if( *end )
  {
Unsucc1:
    log( Msg_RexxErrComTrace1, num, msg );
    goto Unsucc;
  }

  num = 2;
  msg = args[1].strptr;
  int dcd = (int)strtoul( msg, &end, 10 );

  if( *end )
    goto Unsucc1;

  Buffer deb;

  deb.putString( Msg_RexxDebComTrace );

  int how( int( numarg-2 ) );
  Pair **tab = new Pair* [ how ];

  for( int i=2; i<numarg; i++ )
  {
    tab[i-2] = new Pair( (char *)args[i].strptr, EMSI_PACKETS( i-1 ) );
    deb.putInto( '\'' );
    deb.putString( args[i].strptr );
    deb.putString( Msg_RexxDebComTrace1 );
  }

  deb.putString( Msg_RexxDebComTrace2 );
  deb.putString( args[0].strptr );
  deb.putString( Msg_RexxDebComTrace3 );

  if( dcd )
    deb.putString( Msg_RexxDebComTrace4 );
  else
    deb.putString( Msg_RexxDebComTrace5 );

  deb.putString( Msg_RexxDebComTrace6 );
  deb.putInto( 0 );

  log( deb.str );

  for( i=0; i<how; i++ )
    tab[i]->reset();

  Timer left( wait*1000l );
  int reti = 0;

  void terminalWriteToScreen( int chr );

  while( !left.isTimeOut() && !reti )
  {
    if( dcd && !comPort->isOnline() )
      break;

    int ch( comPort->get() );

    if( ch!=-1 )
    {
      if( terminalActive )
        terminalWriteToScreen( ch );

      for( i=0; i<how; i++ )
	if( tab[i]->compare( (char)ch ) == Pair::Completed )
	  reti = i + 1;
    }
  }

  for( i=0; i<how; i++ )
    delete tab[i];
    
  delete tab;

  log( Msg_RexxDebComTrace7, reti );

  sprintf( ret->strptr, "%d", reti );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkLog( PRXSTRING, ULONG numarg, PRXSTRING args,
			    PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkLog" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  APIRET rc( 0 );
  Buffer out;

  if( numarg )
  {
    for( ULONG i=0; i<numarg; i++ )
    {
      register char *ptr = args[i].strptr;
      ULONG len = args[i].strlength;

      for( register ULONG j=0; j<len; j++, ptr++ )
        out.putInto( *ptr );

      out.putInto( ' ' );
    }

    out.stripByte();
    out.putInto( 0 );
    log( out.str );

    RexxFuncComplete( 1, ret );
  }
  else
  {
    log( Msg_RexxErrLog );
    rc = 1;
    RexxFuncComplete( 0, ret );
  }

  deRegisterExceptionHandler
  return rc;
}

APIRET EXPENTRY RexxBrkKbdPressed( PRXSTRING, ULONG, PRXSTRING,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( keyPressed() ? 1 : 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkKbdClear( PRXSTRING, ULONG, PRXSTRING,
                                 PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  keyClear();
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkKbdGetCode( PRXSTRING, ULONG, PRXSTRING,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%d", (int)keyGet() );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkKbdGetChar( PRXSTRING, ULONG, PRXSTRING,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  *(ret->strptr) = (unsigned char)keyGet();
  ret->strlength = 1;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxBrkKbdGetString( PRXSTRING, ULONG numargs, PRXSTRING arg,
				     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numargs, arg, ret, "BrkKbdGetString" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  char prompt[65],
       work[81];
  int len( int( arg[0].strlength ) );

  memset( work, 0, sizeof(work) );

  if( len>64 )
    len = 64;

  if( numargs )
  {
    strncpy( prompt, arg[0].strptr, len );
    prompt[ len ] = 0;
  }
  else
    strcpy( prompt, "BrkKbdGetString" );

  if( numargs>1 )
    strncpy( work, arg[1].strptr, sizeof(work)-1 );

  if( !terminalActive )
    +screenCursor;

  int i( stringEnter( 80, work, prompt ) );

  if( !terminalActive )
    -screenCursor;

  work[80] = 0;

  if( i )
  {
    *(ret->strptr) = 0;
    ret->strlength = 0;
  }
  else
  {
    if( !strcmp( work, "  " ) )
      *work = 0;

    RexxRetString( work, ret );
  }

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY U1_Command( PRXSTRING r_cmd, PUSHORT flags, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( r_cmd->strptr==NULL )
  {
    log( Msg_RexxIllegalCmd );
FailCmd:
    *flags = RXSUBCOM_FAILURE;
    strcpy( ret->strptr, "-1" );
    ret->strlength = 2;
    deRegisterExceptionHandler
    return 0;
  }

  char *cmdPtr( r_cmd->strptr ),
       close[] =  "/Close ",
       vio[] =    "/Clear ",
       noclip[] = "/NoClip ",
       port[] =   "/Port ",
       ok( 1 );
  RunOption runOpt( (RunOption)0 );

  while( 1 )
    if( begin( cmdPtr, close ) )
    {
      runOpt = RunOption( runOpt | RunOptClose );
      cmdPtr += sizeof(close)-1;
    }
    else if( begin( cmdPtr, vio ) )
    {
      runOpt = RunOption( runOpt | RunOptClear );
      cmdPtr += sizeof(vio)-1;
    }
    else if( begin( cmdPtr, noclip ) )
    {
      runOpt = RunOption( runOpt | RunOptNoClip );
      cmdPtr += sizeof(noclip)-1;
    }
    else if( begin( cmdPtr, port ) )
    {
      runOpt = RunOption( runOpt | RunOptPort );
      cmdPtr += sizeof(port)-1;
    }
    else
      break;

  long ela = sessionStart ? getTimeMark()-sessionStart : 0;
  FIDOAddress *fAddr = primaryAka!=NULL ? (FIDOAddress *)primaryAka : 
                                          (FIDOAddress *)queue->getCurrent();

  if( fAddr==NULL )
    fAddr = defaultAddress;

  int rc = makeCmd( *fAddr, ela, 0, cmdPtr, 0, runOpt, 0, 0, 0 );

  if( runOpt&RunOptClose && modemOs2Rc )
  {
    OSReturnCode = int( modemOs2Rc );
    ok = 0;
  }

  if( !ok )
    goto FailCmd;

  sprintf( ret->strptr, "%d", rc );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxCursorHide( PRXSTRING, ULONG, PRXSTRING,
				PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  -screenCursor;
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxCursorShow( PRXSTRING, ULONG, PRXSTRING,
				PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  +screenCursor;
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinCreate( PRXSTRING, ULONG numarg, PRXSTRING args,
			       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkWinCreate" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  char *end;

  if( numarg<4 )
  {
    log( Msg_RexxErrWinCreate );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  int coord[4];
  char *msg;

  for( int num=0; num<4; num++ )
  {
    msg = args[num].strptr;
    coord[num] = (int)strtoul( msg, &end, 10 );

    if( *end )
    {
Unsucc1:
      log( Msg_RexxErrWinCreate1, num+1, msg );
      goto Unsucc;
    }
  }

  char *title( NULL );

  if( numarg>4 )
    title = args[4].strptr;

  Color col( stdColor );

  if( numarg>5 )
  {
    msg = args[5].strptr;
    num = 4;

    ULONG cl = strtoul( msg, &end, 10 );

    if( *end )
      goto Unsucc1;

    col = Color( (unsigned char)cl );
  }

  int border( BS_SINGLELINE );

  if( numarg>6 )
  {
    msg = args[6].strptr;

    if( !stricmp( msg, "Single" ) )
      border = BS_SINGLELINE;
    else if( !stricmp( msg, "Double" ) )
      border = BS_DOUBLELINE;
    else if( !stricmp( msg, "None" ) )
      border = BS_NOBORDER;
    else
    {
      log( Msg_RexxErrWinCreate2, msg );
      goto Unsucc;
    }
  }

  int shade( DS_NONE );

  if( numarg>7 )
  {
    msg = args[7].strptr;

    if( !stricmp( msg, "LowerRight" ) )
      shade = DS_LOWERRIGHT;
    else if( !stricmp( msg, "LowerLeft" ) )
      shade = DS_LOWERLEFT;
    else if( !stricmp( msg, "UpperLeft" ) )
      shade = DS_UPPERLEFT;
    else if( !stricmp( msg, "UpperRight" ) )
      shade = DS_UPPERRIGHT;
    else if( !stricmp( msg, "None" ) )
      shade = DS_NONE;
    else
    {
      log( Msg_RexxErrWinCreate3, msg );
      goto Unsucc;
    }
  }

  Color colDrop( WHITE, BLACK );

  if( numarg>8 )
  {
    msg = args[8].strptr;
    num = 7;

    ULONG cl = strtoul( msg, &end, 10 );

    if( *end )
      goto Unsucc1;

    colDrop = Color( (unsigned char)cl );
  }

  Window *w = new Window( coord[0], coord[1], coord[2], coord[3],
			  col, title, border, shade, colDrop );

  wndList->AddAtTail( *new WindowHandle( w ) );

  sprintf( ret->strptr, "%lu", w );
  ret->strlength = strlen( ret->strptr );

  if( (miscInfo.state2&RexxDebug)==RexxDebug )
  {
    Buffer deb;

    deb.putString( Msg_RexxDebWinCreate );

    for( num=0; num<numarg; num++ )
    {
      deb.putInto( '\'' );
      deb.putString( args[num].strptr );
      deb.putString( "' " );
    }

    deb.stripByte();
    deb.putString( Msg_RexxDebWinCreate1 );
    deb.putString( ret->strptr );
    deb.putInto( 0 );

    log( deb.str );
  }

  deRegisterExceptionHandler
  return 0;
}

Window *RexxWinValidateWindow( char *str, char *routine )
{
  char *end;
  ULONG ret = strtoul( str, &end, 10 );

  if( *end )
  {
Fail:
    log( Msg_RexxWinNotValid, routine, str );
    return NULL;
  }

  WindowHandle test( (Window *)ret ),
               *found = wndList->Find( test );

  test.setWindow( NULL );

  if( found==NULL )
    goto Fail;

  return (Window *)ret;
}

APIRET EXPENTRY RexxWinShow( PRXSTRING, ULONG numarg, PRXSTRING args,
			     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkWinShow" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinShow );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinShow" );

  if( win==NULL )
    goto Unsucc;

  +(*win);

  log( Msg_RexxDebWinShow, win );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinHide( PRXSTRING, ULONG numarg, PRXSTRING args,
			     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinHide" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinHide );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinHide" );

  if( win==NULL )
    goto Unsucc;

  -(*win);

  log( Msg_RexxDebWinHide, win );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinCoord( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinCoord" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=3 )
  {
    log( Msg_RexxErrWinCoord );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinCoord" );

  if( win==NULL )
    goto Unsucc;

  char *end;
  int num = 2;
  char *msg = args[1].strptr;

  int x = (int)strtoul( msg, &end, 10 );

  if( *end )
  {
Unsucc1:
    log( Msg_RexxErrWinCoord1, num, msg );
    goto Unsucc;
  }

  num = 3;
  msg = args[2].strptr;

  int y = (int)strtoul( msg, &end, 10 );

  if( *end )
    goto Unsucc1;

  (*win)( x, y );

  log( Msg_RexxDebWinCoord, win, x, y );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinCode( PRXSTRING, ULONG numarg, PRXSTRING args,
			     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinCode" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=2 )
  {
    log( Msg_RexxErrWinCode );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinCode" );

  if( win==NULL )
    goto Unsucc;

  char *end;
  int num = 2;
  char *msg = args[1].strptr;

  int x = (int)strtoul( msg, &end, 10 );

  if( *end || x>255 )
  {
    log( Msg_RexxErrWinCode1, num, msg );
    goto Unsucc;
  }

  (*win) << char(x);

  log( Msg_RexxDebWinCode, win, x );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinString( PRXSTRING, ULONG numarg, PRXSTRING args,
			       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinString" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=2 )
  {
    log( Msg_RexxErrWinString );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinString" );

  if( win==NULL )
    goto Unsucc;

  (*win) << args[1].strptr;

  log( Msg_RexxDebWinString, win, args[1].strptr );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinPop( PRXSTRING, ULONG numarg, PRXSTRING args,
			    PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinPop" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinPop );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinPop" );

  if( win==NULL )
    goto Unsucc;

  win->pop();

  log( Msg_RexxDebWinPop, win );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinPush( PRXSTRING, ULONG numarg, PRXSTRING args,
			     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkWinPush" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinPush );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinPush" );

  if( win==NULL )
    goto Unsucc;

  win->push();

  log( Msg_RexxDebWinPush, win );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinClear( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinClear" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinClear );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinClear" );

  if( win==NULL )
    goto Unsucc;

  --(*win);

  log( Msg_RexxDebWinClear, win );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinClearLine( PRXSTRING, ULONG numarg, PRXSTRING args,
				  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinClearLine" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinClearLine );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinClearLine" );

  if( win==NULL )
    goto Unsucc;

  win->clearLine();

  log( Msg_RexxDebWinClearLine, win );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinColor( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinColor" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=2 )
  {
    log( Msg_RexxErrWinColor );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinColor" );

  if( win==NULL )
    goto Unsucc;

  char *end;
  int num = 2;
  char *msg = args[1].strptr;

  int x = (int)strtoul( msg, &end, 10 );

  if( *end || x>255 )
  {
    log( Msg_RexxErrWinColor1, num, msg );
    goto Unsucc;
  }

  x &= 0xff;
  (*win) = Color( (unsigned char)x );

  log( Msg_RexxDebWinColor, win, x );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinCentre( PRXSTRING, ULONG numarg, PRXSTRING args,
			       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinCentre" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=3 )
  {
    log( Msg_RexxErrWinCentre );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinCentre" );

  if( win==NULL )
    goto Unsucc;

  char *end;
  int num = 2;
  char *msg = args[1].strptr;

  int x = (int)strtoul( msg, &end, 10 );

  if( *end )
  {
    log( Msg_RexxErrWinCentre1, num, msg );
    goto Unsucc;
  }

  win->centre( x, args[2].strptr );

  log( Msg_RexxDebWinCentre, win, args[2].strptr, x );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinMove( PRXSTRING, ULONG numarg, PRXSTRING args,
			     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinMove" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=3 )
  {
    log( Msg_RexxErrWinMove );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinMove" );

  if( win==NULL )
    goto Unsucc;

  char *end;
  int num = 2;
  char *msg = args[1].strptr;

  int x = (int)strtoul( msg, &end, 10 );

  if( *end )
  {
Unsucc1:
    log( Msg_RexxErrWinMove1, num, msg );
    goto Unsucc;
  }

  num = 3;
  msg = args[2].strptr;

  int y = (int)strtoul( msg, &end, 10 );

  if( *end )
    goto Unsucc1;

  win->move( x, y );

  log( Msg_RexxDebWinMove, win, x, y );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinSetScroll( PRXSTRING, ULONG numarg, PRXSTRING args,
				  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinSetScroll" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=2 )
  {
    log( Msg_RexxErrWinSetScroll );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinSetScroll" );

  if( win==NULL )
    goto Unsucc;

  char scroll;

  if( !stricmp( args[1].strptr, "YES" ) )
    scroll = 1;
  else if( !stricmp( args[1].strptr, "NO" ) )
    scroll = 0;
  else  {
    log( Msg_RexxErrWinSetScroll1, args[1].strptr );
    goto Unsucc;
  }

  win->setScroll( scroll );

  log( Msg_RexxDebWinScroll, win,
       scroll ? Msg_RexxDebWinScrollYes : Msg_RexxDebWinScrollNo );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinSetPhCursor( PRXSTRING, ULONG numarg, PRXSTRING args,
				    PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinSetPhCursor" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( !numarg )
  {
    log( Msg_RexxErrWinSetPhCursor );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinSetPhCursor" );

  if( win==NULL )
    goto Unsucc;

  int forced( 0 );

  if( numarg>1 )
    if( !stricmp( args[1].strptr, "Force" ) )
      forced = 1;
    else if( !stricmp( args[1].strptr, "Optimal" ) )
      forced = 0;
    else
    {
      log( Msg_RexxErrWinSetPhCursor1, args[1].strptr );
      goto Unsucc;
    }

  win->setPhysicalCursor( forced );

  log( Msg_RexxDebWinCursor, win,
       forced ? Msg_RexxDebWinCursorForced : Msg_RexxDebWinCursorOptimized );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinGetWidth( PRXSTRING, ULONG numarg, PRXSTRING args,
				 PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinGetWidth" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinGetWidth );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinGetWidth" );

  if( win==NULL )
    goto Unsucc;

  sprintf( ret->strptr, "%d", (int)win->getWidth() );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebWinWidth, win, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinGetHeight( PRXSTRING, ULONG numarg, PRXSTRING args,
				  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinGetHeight" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinGetHeight );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinGetHeight" );

  if( win==NULL )
    goto Unsucc;

  sprintf( ret->strptr, "%d", (int)win->getHeight() );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebWinHeight, win, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinGetLeft( PRXSTRING, ULONG numarg, PRXSTRING args,
				PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinGetLeft" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinGetLeft );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinGetLeft" );

  if( win==NULL )
    goto Unsucc;

  sprintf( ret->strptr, "%d", (int)win->getLeft() );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebWinLeft, win, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinGetTop( PRXSTRING, ULONG numarg, PRXSTRING args,
			       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinGetTop" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinGetTop );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinGetTop" );

  if( win==NULL )
    goto Unsucc;

  sprintf( ret->strptr, "%d", (int)win->getTop() );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebWinTop, win, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinGetColor( PRXSTRING, ULONG numarg, PRXSTRING args,
				 PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinGetColor" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinGetColor );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinGetColor" );

  if( win==NULL )
    goto Unsucc;

  sprintf( ret->strptr, "%d", (int)win->getColor() );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebWinIsColor, win, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinGetCurX( PRXSTRING, ULONG numarg, PRXSTRING args,
				PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinGetCurX" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinGetCurX );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinGetCurX" );

  if( win==NULL )
    goto Unsucc;

  sprintf( ret->strptr, "%d", (int)win->getCurrentX() );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebWinCurX, win, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinGetCurY( PRXSTRING, ULONG numarg, PRXSTRING args,
				PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinGetCurY" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinGetCurY );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinGetCurY" );

  if( win==NULL )
    goto Unsucc;

  sprintf( ret->strptr, "%d", (int)win->getCurrentY() );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebWinCurY, win, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinGetScroll( PRXSTRING, ULONG numarg, PRXSTRING args,
				  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinGetScroll" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinGetScroll );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinGetScroll" );

  if( win==NULL )
    goto Unsucc;

  strcpy( ret->strptr, win->getScroll() ? "YES" : "NO" );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebWinIsScroll, win, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxWinDestroy( PRXSTRING, ULONG numarg, PRXSTRING args,
				PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkWinDestroy" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrWinDestroy );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Window *win = RexxWinValidateWindow( args[0].strptr, "BrkWinDestroy" );

  if( win==NULL )
    goto Unsucc;

  WindowHandle ww( win );

  wndList->Remove( ww );
  ww.setWindow( NULL );

  log( Msg_RexxDebWinDestroy, win );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxEachVars( ULONG numarg, PRXSTRING args, char *funcName,
                              char *maskPar,
                              void (*doIt)( RXSTRING&, RXSTRING&, void* ),
                              void (*doLast)( void* ),
                              void *pass,
                              PRXSTRING ret )
{
  if( RexxCheckArgs( numarg, args, ret, funcName ) )
    return 1;

  char *maskExp( ".*" );
  
  if( maskPar )
    maskExp = maskPar;
  
  RegularExpresion mask( maskExp );

  if( !mask.isValid() )
  {
    log( Msg_RexxErrIllegalRegexp, funcName, maskExp );
    RexxFuncComplete( 0, ret );
    return 1;
  }

  while( 1 )
  {
    SHVBLOCK request;
    
    memset( &request, 0, sizeof request );
    request.shvcode = RXSHV_NEXTV;

    APIRET rc = RexxVariablePool( &request );
    
    if( rc )
      break;
      
    char *name( new char[ request.shvname.strlength+1 ] );
    
    memcpy( name, request.shvname.strptr, (size_t)request.shvname.strlength );
    name[ request.shvname.strlength ] = 0;
      
    if( mask.maskMach( name ) )
      doIt( request.shvname, request.shvvalue, pass );
    
    delete name;

    DosFreeMem( request.shvname.strptr );
    DosFreeMem( request.shvvalue.strptr );
  }

  doLast( pass );

  RexxFuncComplete( 1, ret );
  return 0;
}

void RexxVarsPrint( RXSTRING& name, RXSTRING& val, void * )
{
  Buffer ret( 128, 128 );
    
  ret.putInto( '"' );
    
  for( int i(0); i<name.strlength; i++ )
    ret.putQuotted( ((unsigned)(unsigned char)name.strptr[i])&0xff );
      
  ret.putString( "\"=\"" );
    
  for( i=0; i<val.strlength; i++ )
    ret.putQuotted( ((unsigned)(unsigned char)val.strptr[i])&0xff );
      
  ret.putInto( '"' );
  ret.putInto( 0 );
  log( "RexxVar: %s", ret.str );
}

void RexxVarsPrintLast( void * )
{
}

APIRET EXPENTRY RexxPrintVars( PRXSTRING, ULONG numarg, PRXSTRING args,
			       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  heapState( "prtvb" );

  APIRET rc( RexxEachVars( numarg, args, "BrkPrintVars",
                           numarg ? args[0].strptr : NULL,
                           RexxVarsPrint, RexxVarsPrintLast, NULL,
                           ret ) ? 1 : 0 );

  heapState( "prtve" );
  deRegisterExceptionHandler
  return rc;
}

struct EnumVariablesPars
{
  PRXSTRING stem;
  GlobalVariableList glist;
};

void RexxVarsEnum( RXSTRING& name, RXSTRING& val, void *args )
{
  ((EnumVariablesPars *)args)->glist.AddAtTail( *new GlobalVariable( name.strptr, 0 ) );
}

void RexxVarsToPool( char *stem, GlobalVariableList& varList )
{
  char *str( new char [ strlen( stem ) + 32 ] );
  int i( 0 );

  varList.sort();

  for( GlobalVariableIterator list( varList ); list; list++ )
  {
    sprintf( str, "%s.%d", stem, ++i );

    RexxAddVariable( str, list().getName() );
  }

  RexxAddVariable( stem, zeroString, i );

  delete str;
}

void RexxVarsEnumLast( void *args )
{
  RexxVarsToPool( ((EnumVariablesPars *)args)->stem->strptr,
                  ((EnumVariablesPars *)args)->glist );
}

APIRET EXPENTRY RexxEnumVars( PRXSTRING, ULONG numarg, PRXSTRING args,
                              PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  heapState( "enumb" );

  APIRET rc( 0 );

  if( numarg<1 )
  {
    log( Msg_RexxErrEnumVars );
    RexxFuncComplete( 0, ret );
    rc = 1;
  }
  else
  {
    EnumVariablesPars enumVariablesPars;

    enumVariablesPars.stem = args;

    rc = RexxEachVars( numarg, args, "BrkEnumVars",
                       numarg>1? args[1].strptr : NULL,
                       RexxVarsEnum, RexxVarsEnumLast, &enumVariablesPars,
                       ret );
  }

  heapState( "enume" );
  deRegisterExceptionHandler
  return rc;
}

APIRET EXPENTRY RexxExportConfig( PRXSTRING, ULONG, PRXSTRING,
			          PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  time_t start( time( NULL ) );

  rexxExportConfig();
  start = time( NULL ) - start;

  sprintf( ret->strptr, "%lu", start );
  ret->strlength = strlen( ret->strptr );

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQuoteString( PRXSTRING, ULONG numarg, PRXSTRING args,
			         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkQuoteString" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }
  
  if( numarg!=1 )
  {
    log( Msg_RexxErrQuoteString );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }
  
  Buffer b( 128, 128 );

  for( int i(0); i<args[0].strlength; i++ )
    b.putQuotted( ((unsigned)(unsigned char)args[0].strptr[i])&0xff );

  b.putInto( 0 );

  RexxRetString( b.str, ret, b.length()-1 );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxDeQuoteString( PRXSTRING, ULONG numarg, PRXSTRING args,
		  	           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkDeQuoteString" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrDeQuoteString );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }
  
  Buffer b( 128, 128 );

  for( int i(0); i<args[0].strlength; i++ )
    b.putInto( args[0].strptr[i] );

  b.deQuote();
  b.putInto( 0 );

  RexxRetString( b.str, ret, b.length()-1 );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxRandom( PRXSTRING, ULONG numarg, PRXSTRING args,
			    PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkRandom" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrRandom );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *end;
  int r = (int)strtoul( args[0].strptr, &end, 10 );

  if( *end )
  {
    log( Msg_RexxErrRandom1, args[0].strptr );
    goto Unsucc;
  }
  
  if( !r )
    r = 1;

  sprintf( ret->strptr, "%d", random( r ) );
  ret->strlength = strlen( ret->strptr );

  log( Msg_RexxDebRandom, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxOpenObject( PRXSTRING, ULONG numarg, PRXSTRING args,
		  	        PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkOpenObject" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrOpenObject );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  openObject( args[0].strptr );

  RexxFuncComplete( 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxCharToCode( PRXSTRING, ULONG numarg, PRXSTRING args,
			        PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkCharToCode" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrCharToCode );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  sprintf( ret->strptr, "%d", (int)args[0].strptr[0] );
  ret->strlength = strlen( ret->strptr );

  log( Msg_RexxDebChr2Cd, args[0].strptr[0], ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxCodeToChar( PRXSTRING, ULONG numarg, PRXSTRING args,
			        PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkCodeToChar" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrCodeToChar );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *end;
  int r = (int)strtoul( args[0].strptr, &end, 10 );

  if( *end || r>255)
  {
    log( Msg_RexxErrCodeToChar1, 1, args[0].strptr );
    goto Unsucc;
  }

  ret->strptr[0] = (char)r;
  ret->strptr[1] = 0;
  ret->strlength = 1;

  log( Msg_RexxDebCd2Chr, args[0].strptr, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

NodeConfig *RexxNodeValidate( char *str, char *routine, int fl=1 )
{
  USHORT zone, net, node, point;
  char *domain;

  if( !checkAddress( str, zone, net, node, point, &domain ) )
  {
    if( fl )
      log( Msg_RexxErrAddress, routine, str );

    return NULL;
  }

  FIDOAddress *addr = new FIDOAddress( zone, net, node, point, domain );
  NodeConfig *addNode = new NodeConfig( addr ),
             *inList = systemList->Find( *addNode );

  if( inList!=NULL )
  {
    delete addNode;
    addNode = new NodeConfig( inList );
  }

  delete addr;
  return addNode;
}

int RexxTrafficValidate( PRXSTRING args, ULONG numarg,
                         ULONG from, ULONG max, ULONG *arr, char *routine )
{
  memset( arr, 0, (size_t)(sizeof(*arr)*max) );

  char *end;

  for( ULONG i=0; i<max && from+i<numarg; i++ )
  {
     arr[i] = strtoul( args[from+i].strptr, &end, 10 );

     if( *end )
     {
       log( Msg_RexxErrArgument, routine, from+i, args[from+i].strptr );
       return 1;
     }
  }

  return 0;
}

APIRET EXPENTRY RexxQueueAdd( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkQueueAdd" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( !numarg )
  {
    log( Msg_RexxErrQueueAdd );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *addNode = RexxNodeValidate( args[0].strptr, "BrkQueueAdd" );

  if( addNode==NULL )
    goto Unsucc;

  ULONG traf[16];

  if( RexxTrafficValidate( args, numarg, 1, 10, traf, "BrkQueueAdd" ) )
    goto Unsucc;

  QueueEntry *q( new QueueEntry( *addNode ) );

  q->addTraffic( traf );
  queue->Lock();

  QueueEntry *found( queue->Find( *q, LockModeNoNeed ) );

  if( found==NULL )
  {
    queue->add( *q );
    log( Msg_RexxDebQueueAdd, args[0].strptr );
  }
  else
  {
    delete q;
    found->addTraffic( traf );
    log( Msg_RexxDebQueueAdd1, args[0].strptr );
  }

  queue->UnLock();

  delete addNode;

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQueueRemove( PRXSTRING, ULONG numarg, PRXSTRING args,
			         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkQueueRemove" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrQueueRemove );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *addNode = RexxNodeValidate( args[0].strptr, "BrkQueueRemove" );

  if( addNode==NULL )
    goto Unsucc;

  QueueEntry *q( new QueueEntry( *addNode ) );

  queue->Lock();

  QueueEntry *found( queue->Find( *q, LockModeNoNeed ) );

  if( found!=NULL )
  {
    queue->Remove( *found );
    log( Msg_RexxDebQueueRemove, args[0].strptr );
  }
  else
    log( Msg_RexxDebQueueRemove1, args[0].strptr );

  queue->UnLock();

  delete addNode;
  delete q;

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQueueLock( PRXSTRING, ULONG, PRXSTRING,
			       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  queue->Lock();
  log( Msg_RexxDebQueueLock );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQueueUnLock( PRXSTRING, ULONG, PRXSTRING,
			         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  queue->UnLock();
  log( Msg_RexxDebQueueUnLock );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQueueCount( PRXSTRING, ULONG, PRXSTRING,
				PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%d", (int)queue->GetItemsInContainer() );
  ret->strlength = strlen( ret->strptr );
  log( Msg_RexxDebQueueCount, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQueueFind( PRXSTRING, ULONG numarg, PRXSTRING args,
			       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkQueueFind" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrQueueFind );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *addNode = RexxNodeValidate( args[0].strptr, "BrkQueueFind" );

  if( addNode==NULL )
    goto Unsucc;

  QueueEntry *q( new QueueEntry( *addNode ) );
  QueueEntry *found( queue->Find( *q ) );

  if( found!=NULL )
    log( Msg_RexxDebQueueFind, args[0].strptr );
  else
    log( Msg_RexxDebQueueFind1, args[0].strptr );

  delete addNode;
  delete q;

  RexxFuncComplete( found==NULL ? 0 : 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQueueQueryNode( PRXSTRING, ULONG numarg, PRXSTRING args,
			            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkQueueQueryNode" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrQueueFind );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *addNode = RexxNodeValidate( args[0].strptr, "BrkQueueQueryNode" );

  if( addNode==NULL )
    goto Unsucc;

  QueueEntry *q( new QueueEntry( *addNode ) );

  queue->Lock();

  QueueEntry *found( queue->Find( *q, LockModeNoNeed ) );

  if( found!=NULL )
  {
    sprintf( ret->strptr, "%s %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu",
             flavourName( found->getQueuedFlavour() ),
             long( found->getAge()/60/60 ),
             found->getImmedMail(),  found->getImmedXMail(),  found->getImmedFiles(),
             found->getCrashMail(),  found->getCrashXMail(),  found->getCrashFiles(),
             found->getDirectMail(), found->getDirectXMail(), found->getDirectFiles(),
             found->getNormalMail(), found->getNormalXMail(), found->getNormalFiles(),
             found->getHoldMail(),   found->getHoldXMail(),   found->getHoldFiles(),
             found->getRequest() );
    ret->strlength = strlen( ret->strptr );
    log( Msg_RexxDebQueueQuery, args[0].strptr, ret->strptr );
  }
  else
  {
    log( Msg_RexxDebQueueQuery1, args[0].strptr );
    RexxFuncComplete( 0, ret );
  }

  queue->UnLock();

  delete addNode;
  delete q;

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQueueDraw( PRXSTRING, ULONG, PRXSTRING,
			       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  queue->draw();
  log( Msg_RexxDebQueueDraw );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQueueReScan( PRXSTRING, ULONG, PRXSTRING,
			         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  queue->needForRescan();
  log( Msg_RexxDebQueueRescan );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQueueCurrent( PRXSTRING, ULONG, PRXSTRING,
 			          PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  QueueEntry *cur = queue->getCurrent();

  if( cur==NULL )
    RexxFuncComplete( 0, ret );
  else
  {
    cur->toString( ret->strptr, int( ret->strlength-1 ), FIDOAddressShowModeEMSI );
    ret->strlength = strlen( ret->strptr );
  }

  log( Msg_RexxDebQueueCurrent, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxQueueGet( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkQueueGet" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrQueueGet );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *end;
  int r = (int)strtoul( args[0].strptr, &end, 10 );

  if( *end )
  {
    log( Msg_RexxErrQueueGet1, args[0].strptr );
    goto Unsucc;
  }

  queue->Lock();

  if( r<0 || r>=queue->GetItemsInContainer() )
    RexxFuncComplete( 0, ret );
  else
  {
    (*queue)[r].toString( ret->strptr, int( ret->strlength-1 ),
                          FIDOAddressShowModeEMSI );
    ret->strlength = strlen( ret->strptr );
  }

  queue->UnLock();

  log( Msg_RexxDebQueueGet, args[0].strptr, ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxDebugDump( PRXSTRING, ULONG, PRXSTRING,
			      PUSHORT, PRXSTRING ret )
{
  void debugDump();

  registerExceptionHandler
  debugDump();
  log( Msg_RexxDebDebugInfoDumped );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

void RexxCreateReceivedStem( char *strptr, ReceivedFileList& in )
{
  char *name( (char *)_alloca( strlen(strptr)+32 ) );
  int count(1);
  int needDot( strptr[ strlen(strptr)-1 ]!='.' );

  for( ReceivedFileIterator list( in ); list; list++ )
  {
    strcpy( name, strptr );

    if( needDot )
      strcat( name, "." );

    sprintf( name+strlen(name), "%d", count );
    count++;

    RexxAddVariable( name, list().getName() );
  }

  strcpy( name, strptr );

  if( needDot )
    strcat( name, "." );

  strcat( name, zeroString );

  char buf[32];

  sprintf( buf, "%d", count-1 );
  RexxAddVariable( name, buf );
}

APIRET EXPENTRY RexxZReceive( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkZReceive" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

// [inbound[, WazooType[, stem]]]

  char *inbound=miscInfo.inbound;
  int type=0;

  if( numarg )
    inbound = args[0].strptr;

  if( numarg>1 )
    if( !stricmp( args[1].strptr, "ZMO" ) )
      type = Zmodem;
    else if( !stricmp( args[1].strptr, "ZAP" ) )
      type = ZedZap;
    else if( !stricmp( args[1].strptr, "DZA" ) )
      type = DirZap;
    else
    {
      log( Msg_RexxZReceiveErr, args[1].strptr );
      RexxFuncComplete( 0, ret );
      deRegisterExceptionHandler
      return 1;
    }

  ReceivedFileList in;

  rexxTransfer = 1;
  type = Get_Zmodem( inbound, in, type, 0 );
  rexxTransfer = 0;

  if( numarg>2 )
    RexxCreateReceivedStem( args[2].strptr, in );

  if( type==OK || type==SPEC_COND )
  {
    log( Msg_RexxDebZReceiveOk );
    RexxFuncComplete( 1, ret );
  }
  else
  {
    log( Msg_RexxDebZReceiveFail );
    RexxFuncComplete( 0, ret );
  }

  deRegisterExceptionHandler
  return 0;
}

void callBack( ZModemRexxCallBackContext *context, char *nameFile )
{
  char *name = (char *)_alloca( strlen(context->name)+32 );

  strcpy( name, context->name );

  if( name[ strlen(name)-1 ]!='.' )
    strcat( name, "." );

  sprintf( name+strlen(name), "%d", context->count );
  context->count++;

  RexxAddVariable( name, nameFile );
}

APIRET EXPENTRY RexxZSend( PRXSTRING, ULONG numarg, PRXSTRING args,
			   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkZSend" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

// what[, WazooType[, Stem ]]

  int type=Zmodem;

  if( numarg<1 )
  {
    log( Msg_RexxZSendErr );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg>1 )
    if( !stricmp( args[1].strptr, "ZMO" ) )
      type = Zmodem;
    else if( !stricmp( args[1].strptr, "ZAP" ) )
      type = ZedZap;
    else if( !stricmp( args[1].strptr, "DZA" ) )
      type = DirZap;
    else
    {
      log( Msg_RexxZSendErr1, args[1].strptr );
      RexxFuncComplete( 0, ret );
      deRegisterExceptionHandler
      return 1;
    }

  ZModemRexxCallBackContext context;

  context.count = 1;
  context.name = numarg>2 ? args[2].strptr : 0;

  rexxTransfer = 1;
  type = ZSend_Batch( args[0].strptr, type,
                      numarg>2 ? callBack : (CallBackFunc)0, &context );
  rexxTransfer = 0;

  if( numarg>2 )
  {
    char buf[64];

    sprintf( buf, "%d", context.count-1 );
    context.count = 0;
    callBack( &context, buf );
  }

  if( type==OK || type==SPEC_COND )
  {
    log( Msg_RexxDebZSendOk );
    RexxFuncComplete( 1, ret );
  }
  else
  {
    log( Msg_RexxDebZSendFail );
    RexxFuncComplete( 0, ret );
  }

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxHydra( PRXSTRING, ULONG numarg, PRXSTRING args,
			   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkHydra" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

// [what[, inbound[, stem[, stem]]]

  char *what = "",
       *inbound = miscInfo.inbound;

  if( numarg>=1 )
    what = args[0].strptr;

  if( numarg>=2 )
    inbound = args[1].strptr;

  ZModemRexxCallBackContext context;

  context.count = 1;
  context.name = numarg>=3 ? args[2].strptr : 0;

  ReceivedFileList in;

  rexxTransfer = 1;
  int type = HydraBatch( what, inbound, in,
                         numarg>=3 ? callBack : (CallBackFunc)0, &context );
  rexxTransfer = 0;

  if( numarg>=3 )
  {
    char buf[64];

    sprintf( buf, "%d", context.count-1 );
    context.count = 0;
    callBack( &context, buf );
  }

  if( numarg>=4 )
    RexxCreateReceivedStem( args[3].strptr, in );

  if( type==XFER_OK || type==XFER_SKIP )
  {
    log( Msg_RexxDebHydraOk );
    RexxFuncComplete( 1, ret );
  }
  else
  {
    log( Msg_RexxDebHydraFail );
    RexxFuncComplete( 0, ret );
  }

  deRegisterExceptionHandler
  return 0;
}

Flavour RexxFlavourValidate( char *str, char *name, Flavour& fl )
{
  if( !stricmp( str, "Crash" ) )
    fl = crashFlavour;
  else if( !stricmp( str, "Normal" ) )
    fl = normalFlavour;
  else if( !stricmp( str, "Hold" ) )
    fl = holdFlavour;
  else if( !stricmp( str, "Immed" ) )
    fl = immedFlavour;
  else if( !stricmp( str, "Direct" ) )
    fl = directFlavour;
  else
  {
    fl = unknownFlavour;
    log( Msg_RexxErrFlavour, name, str );
  }

  return fl;
}

APIRET EXPENTRY RexxNodePoll( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodePoll" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodePoll" );

  if( node==NULL )
  {
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Flavour fl( normalFlavour );

  if( numarg>1 &&
      RexxFlavourValidate( args[1].strptr, "BrkNodePoll" , fl ) == unknownFlavour )
    goto Unsucc;

  node->makeLO( fl );

  delete node;
  log( Msg_RexxDebNodePoll, numarg>1 ? args[1].strptr : flavourName( fl ),
       args[0].strptr );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

static Flavour flavourTable[5] =
{
  immedFlavour, crashFlavour, directFlavour,
  normalFlavour, holdFlavour
};

int RexxFlavourValidate1( char *str, char *name, int& fl )
{
  if( !stricmp( str, "Immed" ) )
    fl = 0;
  else if( !stricmp( str, "Crash" ) )
    fl = 1;
  else if( !stricmp( str, "Direct" ) )
    fl = 2;
  else if( !stricmp( str, "Normal" ) )
    fl = 3;
  else if( !stricmp( str, "Hold" ) )
    fl = 4;
  else if( !stricmp( str, "All" ) )
    fl = 5;
  else
  {
    fl = -1;
    log( Msg_RexxErrFlavour, name, str );
  }

  return fl;
}

APIRET EXPENTRY RexxNodeDrop( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeDrop" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeDrop" );

  if( node==NULL )
  {
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  int dropFlavour = 5;

  if( numarg>1 &&
      RexxFlavourValidate1( args[1].strptr, "BrkNodeDrop" , dropFlavour ) == -1 )
    goto Unsucc;

  if( dropFlavour==5 )
    for( int i=0; i<5; i++ )
      node->dropLO( flavourTable[i] );
  else
    if( dropFlavour>=0 && dropFlavour<5 )
      node->dropLO( flavourTable[dropFlavour] );

  delete node;
  log( Msg_RexxDebNodeDrop, numarg>1 ? args[1].strptr :
                                       flavourName(flavourTable[dropFlavour]),
       args[0].strptr );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeIsPoint( PRXSTRING, ULONG numarg, PRXSTRING args,
  			         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeIsPoint" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeIsPoint" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->is4D() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeIsValid( PRXSTRING, ULONG numarg, PRXSTRING args,
  			         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeIsValid" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeIsValid", 0 );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 0;
  }

  RexxFuncComplete( node->isValid() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeIsOurs( PRXSTRING, ULONG numarg, PRXSTRING args,
  			        PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeIsOurs" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeIsOurs", 0 );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 0;
  }

  QueueEntry q( *node );

  RexxFuncComplete( q.isOurs() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

int cnvStrToNum( int& num, char *val, char *who,
                 char *msg=Msg_RexxErrNodeChangeIllNum )
{
  char *end;

  num = int( strtol( val, &end, 10 ) );

  if( *end )
  {
    log( msg, val, who );
    return 1;
  }

  return 0;
}

APIRET EXPENTRY RexxNodeChange( PRXSTRING, ULONG numarg, PRXSTRING args,
  			        PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeChange" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=3 )
  {
    log( Msg_RexxErrNodeChange1 );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 0;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeChange", 0 );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 0;
  }

  char *par = args[1].strptr,
       *val = args[2].strptr;
  char changed( 1 );
  int num;

  if( !stricmp( par, "Time" ) )
  {
    WorkTime *w = new WorkTime( val );

    if( w->isValid() )
      node->changeWork( w );
    else
    {
      changed = 0;
      log( Msg_RexxErrNodeChangeWork, val );
    }

    delete w;
  }
  else if( !stricmp( par, "Hold" ) )
    node->changeHoldBox( val );
  else if( !stricmp( par, "Normal" ) )
    node->changeNormalBox( val );
  else if( !stricmp( par, "Crash" ) )
    node->changeCrashBox( val );
  else if( !stricmp( par, "Immed" ) )
    node->changeImmedBox( val );
  else if( !stricmp( par, "Direct" ) )
    node->changeDirectBox( val );
  else if( !stricmp( par, "Inbound" ) )
    node->changeInbound( val );
  else if( !stricmp( par, "Password" ) )
    node->changePassword( val );
  else if( !stricmp( par, "Phone" ) )
    node->changePhone( val );
  else if( !stricmp( par, "Dial" ) )
    node->changeDial( val );
  else if( !stricmp( par, "DialSuf" ) )
    node->changeDialSuff( val );
  else if( !stricmp( par, "Flags" ) )
    node->changeFlags( val );
  else if( !stricmp( par, "RexxDial" ) )
    node->changeRexxDial( val );
  else if( !stricmp( par, "Alias" ) )
    node->changeAlias( val );
  else if( !stricmp( par, "Mode" ) )
    if(      !stricmp( val, "Hold" ) )
      node->changeFlavour( holdFlavour );
    else if( !stricmp( val, "Normal" ) )
      node->changeFlavour( normalFlavour );
    else if( !stricmp( val, "Crash" ) )
      node->changeFlavour( crashFlavour );
    else if( !stricmp( val, "Immed" ) )
      node->changeFlavour( immedFlavour );
    else if( !stricmp( val, "Direct" ) )
      node->changeFlavour( directFlavour );
    else if( !stricmp( val, "Default" ) )
      node->changeFlavour( unknownFlavour );
    else
    {
      changed = 0;
      log( Msg_RexxErrNodeChangeIllFla, val );
    }
  else if( !stricmp( par, "DirZap" ) )
    if( strIsYes( val ) )
      node->changeNoDZAenabled( 0 );
    else if( strIsNo( val ) )
      node->changeNoDZAenabled( 1 );
    else
    {
      changed = 0;
      log( Msg_RexxErrNodeChangeIllDZA, val, par );
    }
  else if( !stricmp( par, "ZedZap" ) )
    if( strIsYes( val ) )
      node->changeNoZAPenabled( 0 );
    else if( strIsNo( val ) )
      node->changeNoZAPenabled( 1 );
    else
    {
      changed = 0;
      log( Msg_RexxErrNodeChangeIllDZA, val, par );
    }
  else if( !stricmp( par, "Hydra" ) )
    if( strIsYes( val ) )
      node->changeNoHydraEnabled( 0 );
    else if( strIsNo( val ) )
      node->changeNoHydraEnabled( 1 );
    else
    {
      changed = 0;
      log( Msg_RexxErrNodeChangeIllDZA, val, par );
    }
  else if( !stricmp( par, "Chain" ) )
    if( strIsYes( val ) )
      node->changeChain( 1 );
    else if( strIsNo( val ) )
      node->changeChain( 0 );
    else
    {
      changed = 0;
      log( Msg_RexxErrNodeChangeIllDZA, val, par );
    }
  else if( !stricmp( par, "Priority" ) )
    if( cnvStrToNum( num, val, "Priority" ) )
      changed = 0;
    else
      node->changePriority( num );
  else if( !stricmp( par, "Carrier" ) )
    if( cnvStrToNum( num, val, "Carrier" ) )
      changed = 0;
    else
      node->changeWait( num );
  else if( !stricmp( par, "CPS" ) )
    if( cnvStrToNum( num, val, "CPS" ) )
      changed = 0;
    else
      node->changeBaud( num*10 );
  else if( !stricmp( par, "Cost" ) )
    if( cnvStrToNum( num, val, "Cost" ) )
      changed = 0;
    else
      node->changeCost( num );
  else if( !stricmp( par, "MaxTry" ) )
    if( cnvStrToNum( num, val, "MaxTry" ) )
      changed = 0;
    else
      node->changeMaxTry( num );
  else if( !stricmp( par, "MaxInComing" ) )
    if( cnvStrToNum( num, val, "MaxInComing" ) )
      changed = 0;
    else
      node->changeMaxIn( (short)num );
  else if( !stricmp( par, "MaxOutGoing" ) )
    if( cnvStrToNum( num, val, "MaxOutGoing" ) )
      changed = 0;
    else
      node->changeMaxOut( (short)num );
  else
  {
    changed = 0;
    log( Msg_RexxErrNodeChangeIllPar, par );
  }

  if( !changed )
    delete node;
  else
  {
    systemList->Remove( *node );
    systemList->AddAtTail( *node );
  }

  RexxFuncComplete( changed ? 1 : 0, ret );
  deRegisterExceptionHandler
  return changed ? 0 : 1;
}

APIRET EXPENTRY RexxNodeGetZone( PRXSTRING, ULONG numarg, PRXSTRING args,
  			         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetZone" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetZone" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  sprintf( ret->strptr, "%d", (int)node->getZone() );
  ret->strlength = strlen( ret->strptr );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetNet( PRXSTRING, ULONG numarg, PRXSTRING args,
  			        PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetNet" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetNet" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  sprintf( ret->strptr, "%d", (int)node->getNet() );
  ret->strlength = strlen( ret->strptr );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetNode( PRXSTRING, ULONG numarg, PRXSTRING args,
  			         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetNode" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetNode" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  sprintf( ret->strptr, "%d", (int)node->getNode() );
  ret->strlength = strlen( ret->strptr );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetPoint( PRXSTRING, ULONG numarg, PRXSTRING args,
  			          PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetPoint" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetPoint" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  sprintf( ret->strptr, "%d", (int)node->getPoint() );
  ret->strlength = strlen( ret->strptr );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetDomain( PRXSTRING, ULONG numarg, PRXSTRING args,
  			           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetDomain" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetDomain" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *dom = node->getDomain();

  if( dom==NULL )
    RexxFuncComplete( 0, ret );
  else
    RexxRetString( dom, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetBsySemaName( PRXSTRING, ULONG numarg, PRXSTRING args,
  			                PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetBsySemaName" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetBsySemaName" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *dom = node->getBsySemaName();

  if( dom==NULL )
    RexxFuncComplete( 0, ret );
  else
    RexxRetString( dom, ret );

  delete dom;
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetTBoxName( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetTBoxName" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetTBoxName" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *dom = node->getTBoxName();

  if( dom==NULL || miscInfo.tbox==NULL )
    RexxFuncComplete( 0, ret );
  else
  {
    char *dom1 = (char *)_alloca( strlen(dom)+strlen( miscInfo.tbox )+1 );
    
    strcpy( dom1, miscInfo.tbox );
    strcat( dom1, dom );
    RexxRetString( dom1, ret );
  }

  delete dom;
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetFullAddress( PRXSTRING, ULONG numarg, PRXSTRING args,
  			                PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetFullAddress" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetFullAddress" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  node->toString( ret->strptr, int( ret->strlength-1 ), FIDOAddressShowModeEMSI );
  ret->strlength = strlen( ret->strptr );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetBinkName( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetBinkName" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetBinkName" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *dom = node->getBinkName();

  if( dom==NULL )
    RexxFuncComplete( 0, ret );
  else
    RexxRetString( dom, ret );

  delete dom;
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetLongName( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetLongName" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetLongName" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *dom = node->getLongName();

  if( dom==NULL )
    RexxFuncComplete( 0, ret );
  else
    RexxRetString( dom, ret );

  delete dom;
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetLongBoxName( PRXSTRING, ULONG numarg, PRXSTRING args,
  			                PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetLongBoxName" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetLongBoxName" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *dom = node->getLongBox();

  if( dom==NULL )
    RexxFuncComplete( 0, ret );
  else
    RexxRetString( dom, ret );

  delete dom;
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetLongBinkName( PRXSTRING, ULONG numarg, PRXSTRING args,
  			                 PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetLongBinkName" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetLongBinkName" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *dom = node->getLongBinkName();

  if( dom==NULL )
    RexxFuncComplete( 0, ret );
  else
    RexxRetString( dom, ret );

  delete dom;
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeHaveHoldBox( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeHaveHoldBox" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeHaveHoldBox" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->haveHoldBox() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeHaveNormalBox( PRXSTRING, ULONG numarg, PRXSTRING args,
  			               PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeHaveNormalBox" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeHaveNormalBox" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->haveNormalBox() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeHaveCrashBox( PRXSTRING, ULONG numarg, PRXSTRING args,
  			               PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeHaveCrashBox" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeHaveCrashBox" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->haveCrashBox() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeHaveImmedBox( PRXSTRING, ULONG numarg, PRXSTRING args,
  			              PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeHaveImmedBox" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeHaveImmedBox" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->haveImmedBox() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeHaveDirectBox( PRXSTRING, ULONG numarg, PRXSTRING args,
  			               PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeHaveDirectBox" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeHaveDirectBox" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->haveDirectBox() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeHaveInbound( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeHaveInbound" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeHaveInbound" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->haveInbound() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeHavePassword( PRXSTRING, ULONG numarg, PRXSTRING args,
  			              PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeHavePassword" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeHavePassword" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->havePassword() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeIsWorked( PRXSTRING, ULONG numarg, PRXSTRING args,
  			          PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeIsWorked" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeIsWorked" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->isWorked( time( NULL ) ) ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeIsPrimaryWorked( PRXSTRING, ULONG numarg, PRXSTRING args,
  			                 PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeIsPrimaryWorked" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeIsPrimaryWorked" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->isPrimaryWorked( time( NULL ) ) ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeIsListed( PRXSTRING, ULONG numarg, PRXSTRING args,
  			          PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeIsListed" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeIsListed" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->isListed() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetHoldBox( PRXSTRING, ULONG numarg, PRXSTRING args,
  			            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetHoldBox" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetHoldBox" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getHoldBox();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetNormalBox( PRXSTRING, ULONG numarg, PRXSTRING args,
  			              PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetNormalBox" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetNormalBox" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getNormalBox();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );
  
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetCrashBox( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetCrashBox" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetCrashBox" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getCrashBox();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetImmedBox( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetImmedBox" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetImmedBox" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getImmedBox();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetDirectBox( PRXSTRING, ULONG numarg, PRXSTRING args,
  			              PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetDirectBox" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetDirectBox" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getDirectBox();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetInbound( PRXSTRING, ULONG numarg, PRXSTRING args,
  			            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetInbound" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetInbound" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getInbound();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetPassword( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetPassword" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetPassword" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getPassword();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetRexxDial( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetRexxDial" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetRexxDial" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getRexxDial();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetPhone( PRXSTRING, ULONG numarg, PRXSTRING args,
  			          PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetPhone" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetPhone" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getPhone();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetWorkTime( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetWorkTime" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetWorkTime" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  WorkTime *work = node->getWorkTime();

  if( work==NULL || !work->isValid() || !work->isPresent() )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( work->getStr(), ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetDial( PRXSTRING, ULONG numarg, PRXSTRING args,
  			         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetDial" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetDial" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getDial();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetDialSuf( PRXSTRING, ULONG numarg, PRXSTRING args,
  			            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetDialSuffix" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetDialSuffix" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getDialSuff();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetWaitCarrier( PRXSTRING, ULONG numarg, PRXSTRING args,
  			                PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetWaitCarrier" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetWaitCarrier" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  sprintf( ret->strptr, "%d", int(node->getWaitCarrier()) );
  ret->strlength = strlen( ret->strptr );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetFlags( PRXSTRING, ULONG numarg, PRXSTRING args,
  			          PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetFlags" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetFlags" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getFlags();

  if( box==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetAlias( PRXSTRING, ULONG numarg, PRXSTRING args,
  			          PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetAlias" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetAlias" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *box = node->getAlias();

  if( box==NULL )
    RexxFuncComplete( 0, ret );
  else
    RexxRetString( box, ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetNdlString( PRXSTRING, ULONG numarg, PRXSTRING args,
  			              PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetNdlString" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetNdlString" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  CacheItem *ndl = node->getNdl();

  if( ndl==NULL || ndl->getString()==NULL )
    RexxFuncComplete(  0, ret );
  else
    RexxRetString( ndl->getString(), ret );

  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetFlavour( PRXSTRING, ULONG numarg, PRXSTRING args,
  			            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetFlavour" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetFlavour" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxRetString( flavourName( node->getCfgFlavour() ), ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetMaxInComing( PRXSTRING, ULONG numarg, PRXSTRING args,
  			                PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetMaxInComing" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetMaxInComing" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  sprintf( ret->strptr, "%d", int(node->getMaxInComing()) );
  ret->strlength = strlen( ret->strptr );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetMaxOutGoing( PRXSTRING, ULONG numarg, PRXSTRING args,
  			                PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetMaxOutGoing" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetMaxOutGoing" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  sprintf( ret->strptr, "%d", int(node->getMaxOutGoing()) );
  ret->strlength = strlen( ret->strptr );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeGetMaxTry( PRXSTRING, ULONG numarg, PRXSTRING args,
  			           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeGetMaxTry" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeGetMaxTry" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  sprintf( ret->strptr, "%d", int(node->getMaxTry()) );
  ret->strlength = strlen( ret->strptr );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeBsySemaCreate( PRXSTRING, ULONG numarg, PRXSTRING args,
  			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeBsySemaCreate" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeBsySemaCreate" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->makeBsySemaphore( 1 ) ? 0 : 1, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeBsySemaRemove( PRXSTRING, ULONG numarg, PRXSTRING args,
  			               PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeBsySemaRemove" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeBsySemaRemove" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->makeBsySemaphore( 0 ) ? 0 : 1, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeBsyCreate( PRXSTRING, ULONG numarg, PRXSTRING args,
  			           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeBsyCreate" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeBsyCreate" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->makeBsy( 1 ) ? 0 : 1, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeBsyRemove( PRXSTRING, ULONG numarg, PRXSTRING args,
  			           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeBsyRemove" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeBsyRemove" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->makeBsy( 0 ) ? 0 : 1, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeIsBusy( PRXSTRING, ULONG numarg, PRXSTRING args,
  			        PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeIsBusy" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeIsBusy" );

  if( node==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RexxFuncComplete( node->checkBsy() ? 1 : 0, ret );
  delete node;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeSend( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkNodeSend" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeSend" );

  if( node==NULL )
  {
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Flavour fl( normalFlavour );

  if( numarg<2 )
  {
    log( Msg_RexxErrNodeSend );
    goto Unsucc;
  }

  if( numarg>2 &&
      RexxFlavourValidate( args[2].strptr, "BrkNodeSend" , fl ) == unknownFlavour )
    goto Unsucc;

  node->makeSend( args[1].strptr, fl );

  delete node;
  log( Msg_RexxDebNodeSend, args[1].strptr, args[0].strptr );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeFreq( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeFreq" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeFreq" );

  if( node==NULL )
  {
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Flavour fl( normalFlavour );

  if( numarg<2 )
  {
    log( Msg_RexxErrNodeFreq );
    goto Unsucc;
  }

  if( numarg>2 &&
      RexxFlavourValidate( args[2].strptr, "BrkNodeFreq" , fl ) == unknownFlavour )
    goto Unsucc;

  node->makeFreq( args[1].strptr, fl );

  delete node;
  log( Msg_RexxDebNodeFreq, args[1].strptr, args[0].strptr );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxNodeCopy( PRXSTRING, ULONG numarg, PRXSTRING args,
			      PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkNodeCopy" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  NodeConfig *node = RexxNodeValidate( args[0].strptr, "BrkNodeCopy" );

  if( node==NULL )
  {
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=2 )
  {
    log( Msg_RexxErrNodeCopy );
    goto Unsucc;
  }

  char *path = (char *)_alloca( strlen( args[1].strptr )+2 );

  strcpy( path, args[1].strptr );

  if( path[ strlen( path )-1 ] != '\\' )
    strcat( path, "\\" );

  moveNodeFiles( (FIDOAddress *)node, path );

  delete node;
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

//// ErrorCode is not thread safe

static volatile APIRET ErrorCode = 0;

APIRET EXPENTRY RexxErrorCode( PRXSTRING, ULONG, PRXSTRING, PUSHORT,
                               PRXSTRING ret )
{
  sprintf( ret->strptr, "%ld", ErrorCode );
  ret->strlength = strlen( ret->strptr );
  return 0;
}

APIRET EXPENTRY RexxCfgGetTaskNumber( PRXSTRING, ULONG, PRXSTRING, PUSHORT,
                                      PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%lu", miscInfo.taskNumber );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxGetLanguage( PRXSTRING, ULONG, PRXSTRING, PUSHORT,
                                 PRXSTRING ret )
{
  registerExceptionHandler
  RexxRetString( localeName, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxGetVersion( PRXSTRING, ULONG, PRXSTRING, PUSHORT,
                                PRXSTRING ret )
{
  registerExceptionHandler
  RexxRetString( (char *)M_VERSION, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxGetTimeZone( PRXSTRING, ULONG, PRXSTRING, PUSHORT,
                                 PRXSTRING ret )
{
  registerExceptionHandler
  sprintf( ret->strptr, "%ld", getTimeZone() );
  ret->strlength = strlen( ret->strptr );
  deRegisterExceptionHandler
  return 0;
}

char *RexxSemaphoreName(  ULONG numarg, PRXSTRING args, char *who )
{
  if( !numarg )
  {
    log( Msg_RexxErrSemaFnc, who );
    return NULL;
  }

  char *pref="";

  if( args[0].strptr[0]!='\\' )
    pref = "\\Sem32\\Brake!\\Rexx\\";

  char *ret = new char [ strlen(pref)+strlen(args[0].strptr)+1 ];

  strcpy( ret, pref );
  strcat( ret, args[0].strptr );

  return ret;
}

ULONG RexxResourceValidate( ULONG numarg, PRXSTRING args, char *who )
{
  if( !numarg )
  {
Fail:
    log( Msg_RexxErrSemaFnc1, who );
    return 0xffffffffl;
  }

  char *e;
  ULONG handle;

  handle = strtoul( args[0].strptr, &e, 10 );

  if( !handle || *e )
    goto Fail;

  return handle;
}

// ============================== Mutex Semaphores

APIRET EXPENTRY RexxMutexSemaCreate( PRXSTRING, ULONG numarg, PRXSTRING args,
			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "RexxMutexSemaCreate" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  char *sname = RexxSemaphoreName( numarg, args, "RexxMutexSemaCreate" );

  if( sname==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  HMTX sema;

  ErrorCode = DosCreateMutexSem( (PSZ)sname, &sema, DC_SEM_SHARED, 0 );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaCMFail, sname, ErrorCode );
  }
  else
  {
    addResourceItem( mutexSemList, (ULONG)sema, sname );
    sprintf( ret->strptr, "%lu", (ULONG)sema );
    ret->strlength = strlen( ret->strptr );
    log( Msg_RexxDebSemaCMOk, sname, (ULONG)sema );
  }

  delete sname;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxMutexSemaOpen( PRXSTRING, ULONG numarg, PRXSTRING args,
			           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "RexxMutexSemaOpen" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  char *sname = RexxSemaphoreName( numarg, args, "RexxMutexSemaOpen" );

  if( sname==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  HMTX sema( 0 );

  ErrorCode = DosOpenMutexSem( (PSZ)sname, &sema );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaOMFail, sname, ErrorCode );
  }
  else
  {
    addResourceItem( mutexSemList, (ULONG)sema, sname );
    sprintf( ret->strptr, "%lu", (ULONG)sema );
    ret->strlength = strlen( ret->strptr );
    log( Msg_RexxDebSemaOMOk, sname, (ULONG)sema );
  }

  delete sname;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxMutexSemaClose( PRXSTRING, ULONG numarg, PRXSTRING args,
			            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "RexxMutexSemaClose" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  ULONG handle = RexxResourceValidate( numarg, args, "RexxMutexSemaClose" );

  if( handle==0xffffffffL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  ErrorCode = DosCloseMutexSem( (HMTX)handle );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaLMFail, handle, ErrorCode );
  }
  else
  {
    detachResourceItem( mutexSemList, (ULONG)handle );
    RexxFuncComplete( 1, ret );
    log( Msg_RexxDebSemaLMOk, handle );
  }

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxMutexSemaRelease( PRXSTRING, ULONG numarg, PRXSTRING args,
			              PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "RexxMutexSemaRelease" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  ULONG handle = RexxResourceValidate( numarg, args, "RexxMutexSemaRelease" );

  if( handle==0xffffffffl )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  ErrorCode = DosReleaseMutexSem( (HMTX)handle );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaRelMFail, handle, ErrorCode );
  }
  else
  {
    RexxFuncComplete( 1, ret );
    log( Msg_RexxDebSemaRelMOk, handle );
  }

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxMutexSemaRequest( PRXSTRING, ULONG numarg, PRXSTRING args,
			              PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "RexxMutexSemaRequest" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  ULONG handle = RexxResourceValidate( numarg, args, "RexxMutexSemaRequest" );

  if( handle==0xffffffffl )
  {
Fail:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  LONG timeOut( -1 );

  if( numarg>1 )
  {
    char *e,
         *str = args[1].strptr;

    timeOut = strtol( str, &e, 10 );

    if( *e )
      if( !stricmp( str, "Wait" ) )
        timeOut = SEM_INDEFINITE_WAIT;
      else if( !stricmp( str, "NoWait" ) )
        timeOut = SEM_IMMEDIATE_RETURN;
      else
      {
        log( Msg_RexxErrSemaWait, "RexxMutexSemaRequest", str );
        goto Fail;
      }
  }

  ErrorCode = DosRequestMutexSem( (HMTX)handle, timeOut );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaReqMFail, handle, ErrorCode );
  }
  else
  {
    RexxFuncComplete( 1, ret );
    log( Msg_RexxDebSemaReqMOk, handle );
  }

  deRegisterExceptionHandler
  return 0;
}

// ============================== Event Semaphores

APIRET EXPENTRY RexxEventSemaCreate( PRXSTRING, ULONG numarg, PRXSTRING args,
			             PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "RexxEventSemaCreate" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  char *sname = RexxSemaphoreName( numarg, args, "RexxEventSemaCreate" );

  if( sname==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  HEV sema;

  ErrorCode = DosCreateEventSem( (PSZ)sname, &sema, DC_SEM_SHARED, 0 );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaCEFail, sname, ErrorCode );
  }
  else
  {
    addResourceItem( eventSemList, (ULONG)sema, sname );
    sprintf( ret->strptr, "%lu", (ULONG)sema );
    ret->strlength = strlen( ret->strptr );
    log( Msg_RexxDebSemaCEOk, sname, (ULONG)sema );
  }

  delete sname;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxEventSemaOpen( PRXSTRING, ULONG numarg, PRXSTRING args,
			           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "RexxEventSemaOpen" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  char *sname = RexxSemaphoreName( numarg, args, "RexxEventSemaOpen" );

  if( sname==NULL )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  HEV sema( 0 );

  ErrorCode = DosOpenEventSem( (PSZ)sname, &sema );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaOEFail, sname, ErrorCode );
  }
  else
  {
    addResourceItem( eventSemList, (ULONG)sema, sname );
    sprintf( ret->strptr, "%lu", (ULONG)sema );
    ret->strlength = strlen( ret->strptr );
    log( Msg_RexxDebSemaOEOk, sname, (ULONG)sema );
  }

  delete sname;
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxEventSemaClose( PRXSTRING, ULONG numarg, PRXSTRING args,
			            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "RexxEventSemaClose" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  ULONG handle = RexxResourceValidate( numarg, args, "RexxEventSemaClose" );

  if( handle==0xffffffffl )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  ErrorCode = DosCloseEventSem( (HEV)handle );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaLEFail, handle, ErrorCode );
  }
  else
  {
    detachResourceItem( eventSemList, (ULONG)handle );
    RexxFuncComplete( 1, ret );
    log( Msg_RexxDebSemaLEOk, handle );
  }

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxEventSemaWait( PRXSTRING, ULONG numarg, PRXSTRING args,
			           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "RexxEventSemaWait" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  ULONG handle = RexxResourceValidate( numarg, args, "RexxEventSemaWait" );

  if( handle==0xffffffffl )
  {
Fail:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  LONG timeOut( -1 );

  if( numarg>1 )
  {
    char *e,
         *str = args[1].strptr;

    timeOut = strtol( str, &e, 10 );

    if( *e )
      if( !stricmp( str, "Wait" ) )
        timeOut = SEM_INDEFINITE_WAIT;
      else if( !stricmp( str, "NoWait" ) )
        timeOut = SEM_IMMEDIATE_RETURN;
      else
      {
        log( Msg_RexxErrSemaWait, "RexxEventSemaWait", str );
        goto Fail;
      }
  }

  ErrorCode = DosWaitEventSem( (HEV)handle, timeOut );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaReqEFail, handle, ErrorCode );
  }
  else
  {
    RexxFuncComplete( 1, ret );
    log( Msg_RexxDebSemaReqEOk, handle );
  }

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxEventSemaPost( PRXSTRING, ULONG numarg, PRXSTRING args,
			           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "RexxEventSemaPost" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  ULONG handle = RexxResourceValidate( numarg, args, "RexxEventSemaPost" );

  if( handle==0xffffffffl )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  ErrorCode = DosPostEventSem( (HEV)handle );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaPEFail, handle, ErrorCode );
  }
  else
  {
    RexxFuncComplete( 1, ret );
    log( Msg_RexxDebSemaPEOk, handle );
  }

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxEventSemaReset( PRXSTRING, ULONG numarg, PRXSTRING args,
			            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "RexxEventSemaReset" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  ULONG handle = RexxResourceValidate( numarg, args, "RexxEventSemaReset" );

  if( handle==0xffffffffl )
  {
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  ULONG count;

  ErrorCode = DosResetEventSem( (HEV)handle, &count );

  if( ErrorCode )
  {
    RexxFuncComplete( 0, ret );
    log( Msg_RexxDebSemaREFail, handle, ErrorCode );
  }
  else
  {
    sprintf( ret->strptr, "%ld", count );
    ret->strlength = strlen( ret->strptr );
    log( Msg_RexxDebSemaREOk, handle, count );
  }

  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxErrorMessage( PRXSTRING, ULONG numarg, PRXSTRING args,
			          PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "RexxErrorMessage" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
Fail:
    RexxFuncComplete( 1, ret );
    log( Msg_RexxErrErrMsg );
    deRegisterExceptionHandler
    return 1;
  }

  char *e;
  long number = strtol( args[0].strptr, &e, 10 );

  if( *e )
    goto Fail;

  UCHAR DataArea[ 513 ];
  ULONG DataLength;
  ULONG MsgLength;
  char *pp[1];
  char field1[20];
  char file[20];

  memset( DataArea, 0, 513 );

  pp[0] = field1;
  strcpy( field1, "dummy" );

  strcpy( file, "OSO001.MSG" );

  DataLength = 511;
  MsgLength = DataLength;

  DosGetMessage( (PCHAR *)&pp[0], 0, (PCHAR)&DataArea, DataLength,
		 number, (PSZ)file, &MsgLength );

  DataArea[ MsgLength ] = 0;
  RexxRetString( (char *)&DataArea, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxDisplayMode( PRXSTRING, ULONG numarg, PRXSTRING args,
                                 PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkDisplayMode" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( !numarg )
  {
    log( Msg_RexxErrDisplayMode );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  showAction( args[0].strptr );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxCrc16( PRXSTRING, ULONG numarg, PRXSTRING args,
                           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkCrc16" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( !numarg )
  {
    log( Msg_RexxErrCrc16 );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }
  
  register word crc(0);
  register unsigned char *p( (unsigned char *)args[0].strptr );

  for( int n(0); n<args[0].strlength; p++, n++ )
    crc = word( Z_UpdateCRC( (unsigned short)(*p), crc) );

  ret->strlength = sprintf( ret->strptr, "%04X", crc );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxCrc32( PRXSTRING, ULONG numarg, PRXSTRING args,
                           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkCrc32" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( !numarg )
  {
    log( Msg_RexxErrCrc32 );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  unsigned long crc( 0xFFFFFFFFl );
  register unsigned char *p( (unsigned char *)args[0].strptr );

  for( int n(0); n<args[0].strlength; ++p, ++n )
    crc = Z_32UpdateCRC( (unsigned short)(*p), crc );

  ret->strlength = sprintf( ret->strptr, "%08X", ~crc );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxRegExp( PRXSTRING, ULONG numarg, PRXSTRING args,
                            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkRegExp" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=2 )
  {
    log( Msg_RexxErrRegExp );
Abort:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  RegularExpresion mask( args[0].strptr );

  if( !mask.isValid() )
  {
    log( Msg_RexxErrRegExp1, mask.getRegErrorMsg(), args[0].strptr );
    goto Abort;
  }

  RexxFuncComplete( mask.maskMach( args[1].strptr ) ? 1 : 0, ret );
  deRegisterExceptionHandler
  return 0;
}


APIRET EXPENTRY RexxSound( PRXSTRING, ULONG numarg, PRXSTRING args,
                           PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkSound" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=2 )
  {
Fail:
    log( Msg_RexxErrSound );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *e;
  int f, d;

  f = int( strtol( args[0].strptr, &e, 10 ) );

  if( *e )
    goto Fail;

  d = int( strtol( args[1].strptr, &e, 10 ) );

  if( *e )
    goto Fail;

  makeBeep( f, d );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxCfgSetCircle( PRXSTRING, ULONG numarg, PRXSTRING args,
                                  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkCfgSetCircle" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=2 )
  {
Fail:
    log( Msg_RexxErrCircle );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *e;
  int cmax, cmin;

  cmax = int( strtoul( args[0].strptr, &e, 10 ) );

  if( *e )
    goto Fail;

  cmin = int( strtoul( args[1].strptr, &e, 10 ) );

  if( *e )
    goto Fail;

  miscInfo.circleMax = cmax;
  miscInfo.circleMin = cmin;

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxPlay( PRXSTRING, ULONG numarg, PRXSTRING args,
                          PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkPlay" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrPlay );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  playWaveFile( args[0].strptr );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxTranslatePhone( PRXSTRING, ULONG numarg, PRXSTRING args,
                                    PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkTranslatePhone" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( !numarg )
  {
    log( Msg_RexxErrTranslatePhone );
Unsucc:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  char *ne( NULL );

  if( numarg<2 )
    if( translationTable!=NULL )
      translationTable->translate( args[0].strptr, &ne );
    else
      ne = tmpStr( args[0].strptr );
  else
  {
    NodeConfig *addNode = RexxNodeValidate( args[1].strptr, "BrkTranslatePhone" );

    if( addNode==NULL )
      goto Unsucc;

    ne = addNode->getTranslatedPhone( args[0].strptr );
    delete addNode;
  }

  if( ne==NULL )
    ne = tmpStr( Msg_ScrUnpublishedPhone );

  RexxRetString( ne, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxCallAction( PRXSTRING, ULONG numarg, PRXSTRING args,
                                PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkCallAction" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrCallAction );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  Schedule sh;

  memset( &sh, 0, sizeof(sh) );
  sh.action = SchedActionCall;
  sh.params = args[0].strptr;

  int RRC = doEventKey( sh, -1 );

  if( RRC>OSReturnCode )
    OSReturnCode = RRC;

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxMelody( PRXSTRING, ULONG numarg, PRXSTRING args,
                            PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkMelody" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg<5 || (numarg-1)%4!=0 )
  {
    log( Msg_RexxErrMelody, numarg );
Fail:
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  int *melody = (int *)_alloca( (unsigned int)((numarg+1)*sizeof(int)) );

  melody[0] = int( (numarg-1)/4 );

  for( int i=0; i<numarg; i++ )
  {
    char *e;

    melody[i+1] = (int)strtol( args[i].strptr, &e, 10 );

    if( *e )
    {
      log( Msg_RexxErrMelody1, i+1, args[i].strptr );
      delete melody;
      goto Fail;
    }
  }

  playMelody( melody );
  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxGlobalVarEnum( PRXSTRING, ULONG numarg, PRXSTRING args,
                                   PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  int rc( 0 );

  if( RexxCheckArgs( numarg, args, ret, "BrkGlobalVarEnum" ) )
    rc = 1;
  else if( numarg!=1 )
  {
    log( Msg_RexxErrGlobalVarEnum );
    rc = 1;
  }
  else if( !globalVariableList )
    RexxAddVariable( args[0].strptr, zeroString, 0 );
  else
    RexxVarsToPool( args[0].strptr, *globalVariableList );

  RexxFuncComplete( rc ? 0 : 1, ret );

  deRegisterExceptionHandler
  return rc;
}

APIRET EXPENTRY RexxGlobalVarSet( PRXSTRING, ULONG numarg, PRXSTRING args,
                                  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  
  if( RexxCheckArgs( numarg, args, ret, "BrkGlobalVarSet" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=2 )
  {
    log( Msg_RexxErrGlobalVarSet );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  if( !globalVariableList )
    globalVariableList = new GlobalVariableList;

  GlobalVariable *test( new GlobalVariable( args[0].strptr, args[1].strptr ) ),
                 *found( globalVariableList->Find( *test ) );

  if( found )
  {
    delete test;
    found->setValue( args[1].strptr );
  }
  else
    globalVariableList->AddAtTail( *test );

  RexxFuncComplete( 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxGlobalVarGet( PRXSTRING, ULONG numarg, PRXSTRING args,
                                  PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  if( RexxCheckArgs( numarg, args, ret, "BrkGlobalVarGet" ) )
  {
    deRegisterExceptionHandler
    return 1;
  }

  if( numarg!=1 )
  {
    log( Msg_RexxErrGlobalVarGet );
    RexxFuncComplete( 0, ret );
    deRegisterExceptionHandler
    return 1;
  }

  if( !globalVariableList )
    RexxRetString( "", ret );
  else
  {
    GlobalVariable test( args[0].strptr, 0 ),
                   *found( globalVariableList->Find( test ) );

    RexxRetString( found ? found->getValue() : "", ret );
  }
  
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxSysReboot( PRXSTRING, ULONG, PRXSTRING,
  			       PUSHORT, PRXSTRING ret )
{
  int reboot();

  registerExceptionHandler
  reboot();
  RexxFuncComplete( 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxSysShutdown( PRXSTRING, ULONG, PRXSTRING,
  			         PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  DosShutdown( 0 );
  RexxFuncComplete( 0, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxSysSync( PRXSTRING, ULONG, PRXSTRING,
  			     PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler
  RexxFuncComplete( DosShutdown( 1 ) ? 0 : 1, ret );
  deRegisterExceptionHandler
  return 0;
}

APIRET EXPENTRY RexxSysUpTime( PRXSTRING, ULONG, PRXSTRING,
  			       PUSHORT, PRXSTRING ret )
{
  registerExceptionHandler

  ret->strlength = sprintf( ret->strptr, "%d", getMSec() );
  deRegisterExceptionHandler
  return 0;
}

static struct
{
  char *name;
  PFN ptr;
} exportList[] =
{
  { "BrkComPutCode",       (PFN)RexxBrkComPutCode },
  { "BrkComPutString",     (PFN)RexxBrkComPutString },

  { "BrkComGetCode",       (PFN)RexxBrkComGetCode },
  { "BrkComGetChar",       (PFN)RexxBrkComGetChar },
  { "BrkComGetString",     (PFN)RexxBrkComGetString },
  { "BrkComGetStringTo",   (PFN)RexxBrkComGetStringTo },

  { "BrkComHasInput",            (PFN)RexxBrkComHasInput },
  { "BrkComIsOnline",            (PFN)RexxBrkComIsOnline },
  { "BrkComCheckOnline",         (PFN)RexxBrkComCheckOnline },
  { "BrkComGetWaitCallTime",     (PFN)RexxBrkComGetWaitCallTime },
  { "BrkComGetWaitAnswerTime",   (PFN)RexxBrkComGetWaitAnswerTime },
  { "BrkComOutIsEmpty",          (PFN)RexxBrkComOutIsEmpty },
  { "BrkComGetDialPrefix",       (PFN)RexxBrkComGetDialPrefix },
  { "BrkComGetDialSuffix",       (PFN)RexxBrkComGetDialSuffix },
  { "BrkComGetSpeed",            (PFN)RexxBrkComGetSpeed },
  { "BrkComGetConnect",          (PFN)RexxBrkComGetConnect },
  { "BrkComIsARQ",               (PFN)RexxBrkComIsARQ },

  { "BrkComGetPort",       (PFN)RexxBrkComGetPort },
  { "BrkComGetHandle",     (PFN)RexxBrkComGetHandle },

  { "BrkComInit",             (PFN)RexxBrkComInit },
  { "BrkComHangUp",           (PFN)RexxBrkComHangUp },
  { "BrkComWaitOK",           (PFN)RexxBrkComWaitOK },
  { "BrkComDial",             (PFN)RexxBrkComDial },
  { "BrkComAnswer",           (PFN)RexxBrkComAnswer },
  { "BrkComClearIn",          (PFN)RexxBrkComClearIn },
  { "BrkComClearOut",         (PFN)RexxBrkComClearOut },
  { "BrkComStartBreak",       (PFN)RexxBrkComStartBreak },
  { "BrkComEndBreak",         (PFN)RexxBrkComEndBreak },
  { "BrkComBufferByte",       (PFN)RexxBrkComBufferByte },
  { "BrkComUnBufferBytes",    (PFN)RexxBrkComUnBufferBytes },
  { "BrkComSetConnectString", (PFN)RexxBrkComSetConnectString },

  { "BrkComFlushBuffer",   (PFN)RexxBrkComFlushBuffer },
  { "BrkComDTROn",         (PFN)RexxBrkComDTROn },
  { "BrkComDTROff",        (PFN)RexxBrkComDTROff },
  { "BrkComXonXoffOn",     (PFN)RexxBrkComXonXoffOn },
  { "BrkComXonXoffOff",    (PFN)RexxBrkComXonXoffOff },
  { "BrkComCtsRtsOn",      (PFN)RexxBrkComCtsRtsOn },
  { "BrkComCtsRtsOff",     (PFN)RexxBrkComCtsRtsOff },
  { "BrkComClose",         (PFN)RexxBrkComClose },
  { "BrkComOpen",          (PFN)RexxBrkComOpen },
  { "BrkComIsOK",          (PFN)RexxBrkComIsOK },
  { "BrkComIsOpened",      (PFN)RexxBrkComIsOpened },
  { "BrkComIsHalfDuplex",  (PFN)RexxBrkComIsHalfDuplex },
  { "BrkComGetTranx",      (PFN)RexxBrkComGetTranx },
  { "BrkComGetLastBaud",   (PFN)RexxBrkComGetLastBaud },
  { "BrkComPutCommand",    (PFN)RexxBrkComPutCommand },

  { "BrkComTrace",         (PFN)RexxBrkComTrace },

  { "BrkLog",              (PFN)RexxBrkLog },

  { "BrkKbdGetString",     (PFN)RexxBrkKbdGetString },
  { "BrkKbdGetCode",       (PFN)RexxBrkKbdGetCode },
  { "BrkKbdGetChar",       (PFN)RexxBrkKbdGetChar },
  { "BrkKbdPressed",       (PFN)RexxBrkKbdPressed },
  { "BrkKbdClear",         (PFN)RexxBrkKbdClear },

  { "BrkCursorHide",       (PFN)RexxCursorHide },
  { "BrkCursorShow",       (PFN)RexxCursorShow },

  { "BrkWinCreate",        (PFN)RexxWinCreate },
  { "BrkWinDestroy",       (PFN)RexxWinDestroy },
  { "BrkWinShow",          (PFN)RexxWinShow },
  { "BrkWinHide",          (PFN)RexxWinHide },
  { "BrkWinCoord",         (PFN)RexxWinCoord },
  { "BrkWinCode",          (PFN)RexxWinCode },
  { "BrkWinString",        (PFN)RexxWinString },
  { "BrkWinClear",         (PFN)RexxWinClear },
  { "BrkWinClearLine",     (PFN)RexxWinClearLine },
  { "BrkWinPop",           (PFN)RexxWinPop },
  { "BrkWinPush",          (PFN)RexxWinPush },
  { "BrkWinColor",         (PFN)RexxWinColor },
  { "BrkWinCentre",        (PFN)RexxWinCentre },
  { "BrkWinMove",          (PFN)RexxWinMove },
  { "BrkWinSetScroll",     (PFN)RexxWinSetScroll },
  { "BrkWinSetPhCursor",   (PFN)RexxWinSetPhCursor },
  { "BrkWinGetWidth",      (PFN)RexxWinGetWidth },
  { "BrkWinGetHeight",     (PFN)RexxWinGetHeight },
  { "BrkWinGetLeft",       (PFN)RexxWinGetLeft },
  { "BrkWinGetTop",        (PFN)RexxWinGetTop },
  { "BrkWinGetColor",      (PFN)RexxWinGetColor },
  { "BrkWinGetCurX",       (PFN)RexxWinGetCurX },
  { "BrkWinGetCurY",       (PFN)RexxWinGetCurY },
  { "BrkWinGetScroll",     (PFN)RexxWinGetScroll },

  { "BrkQueueAdd",         (PFN)RexxQueueAdd },
  { "BrkQueueRemove",      (PFN)RexxQueueRemove },
  { "BrkQueueLock",        (PFN)RexxQueueLock },
  { "BrkQueueUnLock",      (PFN)RexxQueueUnLock },
  { "BrkQueueCount",       (PFN)RexxQueueCount },
  { "BrkQueueFind",        (PFN)RexxQueueFind },
  { "BrkQueueDraw",        (PFN)RexxQueueDraw },
  { "BrkQueueCurrent",     (PFN)RexxQueueCurrent },
  { "BrkQueueGet",         (PFN)RexxQueueGet },
  { "BrkQueueReScan",      (PFN)RexxQueueReScan },
  { "BrkQueueQueryNode",   (PFN)RexxQueueQueryNode },

  { "BrkZReceive",         (PFN)RexxZReceive },
  { "BrkZSend",            (PFN)RexxZSend },
  { "BrkHydra",            (PFN)RexxHydra },

  { "BrkNodeIsPoint",           (PFN)RexxNodeIsPoint },
  { "BrkNodeIsValid",           (PFN)RexxNodeIsValid },
  { "BrkNodeIsListed",          (PFN)RexxNodeIsListed },
  { "BrkNodeIsWorked",          (PFN)RexxNodeIsWorked },
  { "BrkNodeIsPrimaryWorked",   (PFN)RexxNodeIsPrimaryWorked },
  { "BrkNodeIsBusy",            (PFN)RexxNodeIsBusy },
  { "BrkNodeIsOurs",            (PFN)RexxNodeIsOurs },

  { "BrkNodeHaveHoldBox",       (PFN)RexxNodeHaveHoldBox },
  { "BrkNodeHaveNormalBox",     (PFN)RexxNodeHaveNormalBox },
  { "BrkNodeHaveCrashBox",      (PFN)RexxNodeHaveCrashBox },
  { "BrkNodeHaveImmedBox",      (PFN)RexxNodeHaveImmedBox },
  { "BrkNodeHaveDirectBox",     (PFN)RexxNodeHaveDirectBox },
  { "BrkNodeHaveInbound",       (PFN)RexxNodeHaveInbound },
  { "BrkNodeHavePassword",      (PFN)RexxNodeHavePassword },

  { "BrkNodeGetZone",           (PFN)RexxNodeGetZone },
  { "BrkNodeGetNet",            (PFN)RexxNodeGetNet },
  { "BrkNodeGetNode",           (PFN)RexxNodeGetNode },
  { "BrkNodeGetPoint",          (PFN)RexxNodeGetPoint },
  { "BrkNodeGetDomain",         (PFN)RexxNodeGetDomain },
  { "BrkNodeGetBsySemaName",    (PFN)RexxNodeGetBsySemaName },
  { "BrkNodeGetTBoxName",       (PFN)RexxNodeGetTBoxName },
  { "BrkNodeGetBinkName",       (PFN)RexxNodeGetBinkName },
  { "BrkNodeGetLongName",       (PFN)RexxNodeGetLongName },
  { "BrkNodeGetLongBoxName",    (PFN)RexxNodeGetLongBoxName },
  { "BrkNodeGetLongBinkName",   (PFN)RexxNodeGetLongBinkName },
  { "BrkNodeGetFullAddress",    (PFN)RexxNodeGetFullAddress },

  { "BrkNodeGetHoldBox",        (PFN)RexxNodeGetHoldBox },
  { "BrkNodeGetNormalBox",      (PFN)RexxNodeGetNormalBox },
  { "BrkNodeGetCrashBox",       (PFN)RexxNodeGetCrashBox },
  { "BrkNodeGetImmedBox",       (PFN)RexxNodeGetImmedBox },
  { "BrkNodeGetDirectBox",      (PFN)RexxNodeGetDirectBox },
  { "BrkNodeGetInbound",        (PFN)RexxNodeGetInbound },
  { "BrkNodeGetPassword",       (PFN)RexxNodeGetPassword },
  { "BrkNodeGetRexxDial",       (PFN)RexxNodeGetRexxDial },
  { "BrkNodeGetPhone",          (PFN)RexxNodeGetPhone },
  { "BrkNodeGetDial",           (PFN)RexxNodeGetDial },
  { "BrkNodeGetDialSuffix",     (PFN)RexxNodeGetDialSuf },
  { "BrkNodeGetWorkTime",       (PFN)RexxNodeGetWorkTime },
  { "BrkNodeGetWaitCarrier",    (PFN)RexxNodeGetWaitCarrier },
  { "BrkNodeGetFlags",          (PFN)RexxNodeGetFlags },
  { "BrkNodeGetMaxInComing",    (PFN)RexxNodeGetMaxInComing },
  { "BrkNodeGetMaxOutGoing",    (PFN)RexxNodeGetMaxOutGoing },
  { "BrkNodeGetMaxTry",         (PFN)RexxNodeGetMaxTry },
  { "BrkNodeGetNdlString",      (PFN)RexxNodeGetNdlString },
  { "BrkNodeGetFlavour",        (PFN)RexxNodeGetFlavour },
  { "BrkNodeGetAlias",          (PFN)RexxNodeGetAlias },

  { "BrkNodeBsySemaCreate",     (PFN)RexxNodeBsySemaCreate },
  { "BrkNodeBsySemaRemove",     (PFN)RexxNodeBsySemaRemove },
  { "BrkNodeBsyCreate",         (PFN)RexxNodeBsyCreate },
  { "BrkNodeBsyRemove",         (PFN)RexxNodeBsyRemove },

  { "BrkNodePoll",              (PFN)RexxNodePoll },
  { "BrkNodeDrop",              (PFN)RexxNodeDrop },
  { "BrkNodeSend",              (PFN)RexxNodeSend },
  { "BrkNodeFreq",              (PFN)RexxNodeFreq },

  { "BrkNodeChange",            (PFN)RexxNodeChange },
  { "BrkNodeCopy",              (PFN)RexxNodeCopy },

  { "BrkMutexSemaCreate",       (PFN)RexxMutexSemaCreate },
  { "BrkMutexSemaOpen",         (PFN)RexxMutexSemaOpen },
  { "BrkMutexSemaClose",        (PFN)RexxMutexSemaClose },
  { "BrkMutexSemaRequest",      (PFN)RexxMutexSemaRequest },
  { "BrkMutexSemaRelease",      (PFN)RexxMutexSemaRelease },

  { "BrkEventSemaCreate",       (PFN)RexxEventSemaCreate },
  { "BrkEventSemaOpen",         (PFN)RexxEventSemaOpen },
  { "BrkEventSemaClose",        (PFN)RexxEventSemaClose },
  { "BrkEventSemaWait",         (PFN)RexxEventSemaWait },
  { "BrkEventSemaPost",         (PFN)RexxEventSemaPost },
  { "BrkEventSemaReset",        (PFN)RexxEventSemaReset },

  { "BrkCfgGetTaskNumber",      (PFN)RexxCfgGetTaskNumber },
  { "BrkGetLanguage",           (PFN)RexxGetLanguage },
  { "BrkGetVersion",            (PFN)RexxGetVersion },

  { "BrkExportConfig",          (PFN)RexxExportConfig },
  { "BrkPrintVars",             (PFN)RexxPrintVars },
  { "BrkRandom",                (PFN)RexxRandom },
  { "BrkCodeToChar",            (PFN)RexxCodeToChar },
  { "BrkCharToCode",            (PFN)RexxCharToCode },
  { "BrkOpenObject",            (PFN)RexxOpenObject },
  { "BrkRegExp",                (PFN)RexxRegExp },
  { "BrkCrc16",                 (PFN)RexxCrc16 },
  { "BrkCrc32",                 (PFN)RexxCrc32 },
  { "BrkErrorCode",             (PFN)RexxErrorCode },
  { "BrkErrorMessage",          (PFN)RexxErrorMessage },
  { "BrkDisplayMode",           (PFN)RexxDisplayMode },
  { "BrkSound",                 (PFN)RexxSound },
  { "BrkPlay",                  (PFN)RexxPlay },
  { "BrkMelody",                (PFN)RexxMelody },
  { "BrkCallAction",            (PFN)RexxCallAction },
  { "BrkTranslatePhone",        (PFN)RexxTranslatePhone },
  { "BrkDebugDump",             (PFN)RexxDebugDump },
  { "BrkGlobalVarSet",          (PFN)RexxGlobalVarSet },
  { "BrkGlobalVarGet",          (PFN)RexxGlobalVarGet },
  { "BrkGlobalVarEnum",         (PFN)RexxGlobalVarEnum },
  { "BrkCfgSetCircle",          (PFN)RexxCfgSetCircle },
  { "BrkQuoteString",           (PFN)RexxQuoteString },
  { "BrkDeQuoteString",         (PFN)RexxDeQuoteString },
  { "BrkEnumVars",              (PFN)RexxEnumVars },
  { "BrkGetTimeZone",           (PFN)RexxGetTimeZone },

  { "BrkSysUpTime",             (PFN)RexxSysUpTime },
  { "BrkSysSync",               (PFN)RexxSysSync },
  { "BrkSysShutdown",           (PFN)RexxSysShutdown },
  { "BrkSysReboot",             (PFN)RexxSysReboot }
};

APIRET RexxExitNotActive = 1;

void RexxModuleDeInitialize()
{
  DeRegisterRexxHandler( RexxExitNotActive );

  RexxDeregisterSubcom( "Brake!", 0 );

  for( int i=0; i<sizeof(exportList)/sizeof(exportList[0]); i++ )
    RexxDeregisterFunction( exportList[i].name );
}

void RexxModuleInitialize()
{
  APIRET rc = RexxRegisterSubcomExe( "Brake!", (PFN)U1_Command, 0 );

  if( rc )
    log( Msg_RexxErrRegisterSC, rc );

  for( int i=0; i<sizeof(exportList)/sizeof(exportList[0]); i++ )
  {
    rc=RexxRegisterFunctionExe( exportList[i].name, exportList[i].ptr );

    if( rc )
      log( Msg_RexxErrRegisterFunc, exportList[i].name, rc );
  }

  RexxExitNotActive = RegisterRexxHandler();
}

void doTrivialRexx( char *act )
{
  char *out=NULL;
  CallRexx( act, &out, NULL, NULL, NULL );

  if( out!=NULL )
    delete out;
}

void doEvalRexx( Schedule& sh )
{
  log( Msg_InfoRexxEval, sh.params );

  RXSTRING inStore[2],
           rexxretval;
  APIRET   rc;
  SHORT    rexxrc = 0;
  char     *name = "This-Is-Fictive-Name-For-REXX-Eval";

  inStore[0].strptr = sh.params;
  inStore[0].strlength = strlen( sh.params );
  inStore[1].strptr = (PCHAR)sh.token;
  inStore[1].strlength = sh.tokenLen;

  rexxretval.strlength = 0L;
  rexxretval.strptr = NULL;

  if( wndList==NULL )
    wndList = new WindowHandleList;

//  APIRET exitNotActive( RegisterRexxHandler() );

  rc=RexxStart((LONG)       0,             /* number of arguments   */
	       (PRXSTRING)  0,             /* array of arguments    */
	       (PSZ)        name,          /* name of REXX file     */
	       (PRXSTRING)  &inStore,      /* No INSTORE used       */
	       (PSZ)        "Brake!",      /* Command env. name     */
	       (LONG)       RXSUBROUTINE,  /* Code for how invoked  */
               (PRXSYSEXIT) rexxExitList,  /* EXITs list            */
	       (PSHORT)     &rexxrc,       /* Rexx program output   */
	       (PRXSTRING)  &rexxretval ); /* Rexx program output   */

//  DeRegisterRexxHandler( exitNotActive );
  
  sh.token = inStore[1].strptr;
  sh.tokenLen = int( inStore[1].strlength );

  log( Msg_RexxReturned, str( rexxretval.strptr ),
	  rc, (int)rexxrc );

  if( rexxretval.strptr != NULL )
    DosFreeMem( rexxretval.strptr );          /* Release storage       */
}

char **rexxParam( char **params, FIDOAddress& address, long elapsed,
                  char *name, char *actIni, int wait, char *title )
{
  char **cmd = parseMacro( address, elapsed, name, actIni, wait, title );
  int len = 0;

  for( int i=3; cmd[i]!=NULL; i++ )
    len += strlen( cmd[i] ) + 1;

  *params = new char [ len ? len : 1 ];
  **params = 0;

  if( len )
    for( i=3; cmd[i]!=NULL; i++ )
    {
      strcat( *params, cmd[i] );

      if( cmd[i+1]!=NULL )
        strcat( *params, " " );
    }

  return cmd;
}

void doEvalRexxParam( Schedule& sh, FIDOAddress& address, long elapsed,
                      char *fileName, int wait, char *title )
{
  char *params,
       **cmd = rexxParam( &params, address, elapsed, fileName, sh.params,
                          wait, title );

  log( Msg_InfoRexxEvalParam, sh.preloaded, params );

  RXSTRING inStore[2],
           rexxretval;
  APIRET   rc;
  SHORT    rexxrc = 0;

  inStore[0].strptr = (char *)sh.sound;
  inStore[0].strlength = sh.sndCount;
  inStore[1].strptr = (PCHAR)sh.token;
  inStore[1].strlength = sh.tokenLen;

  rexxretval.strlength = 0L;
  rexxretval.strptr = NULL;

  RXSTRING arg;
  MAKERXSTRING( arg, params, strlen( params ) )

  if( wndList==NULL )
    wndList = new WindowHandleList;

//  APIRET exitNotActive( RegisterRexxHandler() );

  rc=RexxStart((LONG)       0,             /* number of arguments   */
	       (PRXSTRING)  0,             /* array of arguments    */
	       (PSZ)        cmd[2],        /* name of REXX file     */
	       (PRXSTRING)  &inStore,      /* No INSTORE used       */
	       (PSZ)        "Brake!",      /* Command env. name     */
	       (LONG)       RXSUBROUTINE,  /* Code for how invoked  */
               (PRXSYSEXIT) rexxExitList,  /* EXITs list            */
	       (PSHORT)     &rexxrc,       /* Rexx program output   */
	       (PRXSTRING)  &rexxretval ); /* Rexx program output   */

//  DeRegisterRexxHandler( exitNotActive );
  
  sh.token = inStore[1].strptr;
  sh.tokenLen = int( inStore[1].strlength );

  log( Msg_RexxReturned, str( rexxretval.strptr ), rc, (int)rexxrc );

  if( rexxretval.strptr != NULL )
    DosFreeMem( rexxretval.strptr );

  delete params;
  freeMacro( cmd );
}

int doRexxWithNumericRC( char *proc, char *phone, char *addr )
{
  char *out=NULL;
  CallRexx( proc, &out, phone, addr, NULL );

  int rc=0;

  if( out!=NULL )
  {
    char *end;

    rc = (int)strtoul( out, &end, 10 );

    if( *end )
      rc = 0;

    delete out;
  }

  log( Msg_LogDialerCompleted, rc );
  return rc;
}

void doParamsRexx( FIDOAddress& address, long elapsed,
                   char *name, char *actIni, int wait, char *title )
{
  char *params,
       **cmd = rexxParam( &params, address, elapsed, name, actIni, wait, title );

  char *out=NULL;

  CallRexx( cmd[2], &out, params, NULL, NULL );

  if( out!=NULL )
    delete out;

  delete params;
  freeMacro( cmd );
}
